Package: taxize
Title: Taxonomic Information from Around the Web
Description: Interacts with a suite of web application programming interfaces
    (API) for taxonomic tasks, such as getting database specific taxonomic
    identifiers, verifying species names, getting taxonomic hierarchies, fetching
    downstream and upstream taxonomic names, getting taxonomic synonyms, converting
    scientific to common names and vice versa, and more. Some of the services
    supported include 'NCBI E-utilities' (<https://www.ncbi.nlm.nih.gov/books/NBK25501/>),
    'Encyclopedia of Life' (<https://eol.org/docs/what-is-eol/data-services>),
    'Global Biodiversity Information Facility' (<https://techdocs.gbif.org/en/openapi/>),
    and many more. Links to the API documentation for other supported services are
    available in the documentation for their respective functions in this package.
Version: 0.10.1
License: MIT + file LICENSE
URL: https://docs.ropensci.org/taxize/ (website),
        https://github.com/ropensci/taxize (devel)
BugReports: https://github.com/ropensci/taxize/issues
Authors@R: 
    c(person(given = "Scott",
             family = "Chamberlain",
             role = c("aut"),
             email = "myrmecocystus@gmail.com"),
      person(given = "Eduard",
             family = "Szoecs",
             role = "aut"),
      person(given = "Zachary",
             family = "Foster",
             role = c("aut", "cre"),
             email = "zacharyfoster1989@gmail.com"),
      person(given = "Zebulun",
             family = "Arendsee",
             role = "aut"),
      person(given = "Carl",
             family = "Boettiger",
             role = "ctb"),
      person(given = "Karthik",
             family = "Ram",
             role = "ctb"),
      person(given = "Ignasi",
             family = "Bartomeus",
             role = "ctb"),
      person(given = "John",
             family = "Baumgartner",
             role = "ctb"),
      person(given = "James",
             family = "O'Donnell",
             role = "ctb"),
      person(given = "Jari",
             family = "Oksanen",
             role = "ctb"),
      person(given = "Bastian Greshake",
             family = "Tzovaras",
             role = "ctb"),
      person(given = "Philippe",
             family = "Marchand",
             role = "ctb"),
      person(given = "Vinh",
             family = "Tran",
             role = "ctb"),
      person(given = "Maëlle",
             family = "Salmon",
             role = "ctb"),
      person(given = "Gaopeng",
             family = "Li",
             role = "ctb"),
      person(given = "Matthias",
             family = "Grenié",
             role = "ctb"),
      person(given = "rOpenSci",
             role = "fnd"))
LazyLoad: yes
LazyData: true
Encoding: UTF-8
Language: en-US
Depends: R(>= 3.2.1)
Imports: graphics, methods, stats, utils, crul (>= 0.7.0), xml2 (>=
        1.2.0), jsonlite, ape, zoo, data.table, tibble (>= 1.2),
        rredlist, rotl (>= 3.0.0), ritis (>= 0.7.6), worrms (>= 0.4.0),
        natserv (>= 1.0.0), wikitaxa (>= 0.3.0), R6, crayon, cli,
        phangorn, lifecycle, curl, stringi
Suggests: testthat, vegan, vcr
RoxygenNote: 7.3.3
X-schema.org-applicationCategory: Taxonomy
X-schema.org-keywords: taxonomy, biology, nomenclature, JSON, API, web,
        api-client, identifiers, species, names
X-schema.org-isPartOf: https://ropensci.org
NeedsCompilation: no
Packaged: 2026-02-13 20:17:14 UTC; fosterz
Author: Scott Chamberlain [aut],
  Eduard Szoecs [aut],
  Zachary Foster [aut, cre],
  Zebulun Arendsee [aut],
  Carl Boettiger [ctb],
  Karthik Ram [ctb],
  Ignasi Bartomeus [ctb],
  John Baumgartner [ctb],
  James O'Donnell [ctb],
  Jari Oksanen [ctb],
  Bastian Greshake Tzovaras [ctb],
  Philippe Marchand [ctb],
  Vinh Tran [ctb],
  Maëlle Salmon [ctb],
  Gaopeng Li [ctb],
  Matthias Grenié [ctb],
  rOpenSci [fnd]
Maintainer: Zachary Foster <zacharyfoster1989@gmail.com>
Repository: CRAN
Date/Publication: 2026-02-14 14:40:02 UTC
Built: R 4.6.0; ; 2026-04-29 21:35:31 UTC; unix

Package: taxize
Title: Taxonomic Information from Around the Web
Description: Interacts with a suite of web application programming interfaces
    (API) for taxonomic tasks, such as getting database specific taxonomic
    identifiers, verifying species names, getting taxonomic hierarchies, fetching
    downstream and upstream taxonomic names, getting taxonomic synonyms, converting
    scientific to common names and vice versa, and more. Some of the services
    supported include 'NCBI E-utilities' (<https://www.ncbi.nlm.nih.gov/books/NBK25501/>),
    'Encyclopedia of Life' (<https://eol.org/docs/what-is-eol/data-services>),
    'Global Biodiversity Information Facility' (<https://techdocs.gbif.org/en/openapi/>),
    and many more. Links to the API documentation for other supported services are
    available in the documentation for their respective functions in this package.
Version: 0.10.1
License: MIT + file LICENSE
URL: https://docs.ropensci.org/taxize/ (website),
        https://github.com/ropensci/taxize (devel)
BugReports: https://github.com/ropensci/taxize/issues
Authors@R: 
    c(person(given = "Scott",
             family = "Chamberlain",
             role = c("aut"),
             email = "myrmecocystus@gmail.com"),
      person(given = "Eduard",
             family = "Szoecs",
             role = "aut"),
      person(given = "Zachary",
             family = "Foster",
             role = c("aut", "cre"),
             email = "zacharyfoster1989@gmail.com"),
      person(given = "Zebulun",
             family = "Arendsee",
             role = "aut"),
      person(given = "Carl",
             family = "Boettiger",
             role = "ctb"),
      person(given = "Karthik",
             family = "Ram",
             role = "ctb"),
      person(given = "Ignasi",
             family = "Bartomeus",
             role = "ctb"),
      person(given = "John",
             family = "Baumgartner",
             role = "ctb"),
      person(given = "James",
             family = "O'Donnell",
             role = "ctb"),
      person(given = "Jari",
             family = "Oksanen",
             role = "ctb"),
      person(given = "Bastian Greshake",
             family = "Tzovaras",
             role = "ctb"),
      person(given = "Philippe",
             family = "Marchand",
             role = "ctb"),
      person(given = "Vinh",
             family = "Tran",
             role = "ctb"),
      person(given = "Maëlle",
             family = "Salmon",
             role = "ctb"),
      person(given = "Gaopeng",
             family = "Li",
             role = "ctb"),
      person(given = "Matthias",
             family = "Grenié",
             role = "ctb"),
      person(given = "rOpenSci",
             role = "fnd"))
LazyLoad: yes
LazyData: true
Encoding: UTF-8
Language: en-US
Depends: R(>= 3.2.1)
Imports: graphics, methods, stats, utils, crul (>= 0.7.0), xml2 (>=
        1.2.0), jsonlite, ape, zoo, data.table, tibble (>= 1.2),
        rredlist, rotl (>= 3.0.0), ritis (>= 0.7.6), worrms (>= 0.4.0),
        natserv (>= 1.0.0), wikitaxa (>= 0.3.0), R6, crayon, cli,
        phangorn, lifecycle, curl, stringi
Suggests: testthat, vegan, vcr
RoxygenNote: 7.3.3
X-schema.org-applicationCategory: Taxonomy
X-schema.org-keywords: taxonomy, biology, nomenclature, JSON, API, web,
        api-client, identifiers, species, names
X-schema.org-isPartOf: https://ropensci.org
NeedsCompilation: no
Packaged: 2026-02-13 20:17:14 UTC; fosterz
Author: Scott Chamberlain [aut],
  Eduard Szoecs [aut],
  Zachary Foster [aut, cre],
  Zebulun Arendsee [aut],
  Carl Boettiger [ctb],
  Karthik Ram [ctb],
  Ignasi Bartomeus [ctb],
  John Baumgartner [ctb],
  James O'Donnell [ctb],
  Jari Oksanen [ctb],
  Bastian Greshake Tzovaras [ctb],
  Philippe Marchand [ctb],
  Vinh Tran [ctb],
  Maëlle Salmon [ctb],
  Gaopeng Li [ctb],
  Matthias Grenié [ctb],
  rOpenSci [fnd]
Maintainer: Zachary Foster <zacharyfoster1989@gmail.com>
Repository: CRAN
Date/Publication: 2026-02-14 14:40:02 UTC
Built: R 4.6.0; ; 2026-04-26 19:38:35 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.boldid,boldid)
S3method(as.boldid,character)
S3method(as.boldid,data.frame)
S3method(as.boldid,list)
S3method(as.boldid,numeric)
S3method(as.data.frame,boldid)
S3method(as.data.frame,eolid)
S3method(as.data.frame,gbifid)
S3method(as.data.frame,iucn)
S3method(as.data.frame,natservid)
S3method(as.data.frame,nbnid)
S3method(as.data.frame,pow)
S3method(as.data.frame,tolid)
S3method(as.data.frame,tpsid)
S3method(as.data.frame,tsn)
S3method(as.data.frame,uid)
S3method(as.data.frame,wiki)
S3method(as.data.frame,wormsid)
S3method(as.eolid,character)
S3method(as.eolid,data.frame)
S3method(as.eolid,eolid)
S3method(as.eolid,list)
S3method(as.eolid,numeric)
S3method(as.gbifid,character)
S3method(as.gbifid,data.frame)
S3method(as.gbifid,gbifid)
S3method(as.gbifid,list)
S3method(as.gbifid,numeric)
S3method(as.iucn,character)
S3method(as.iucn,data.frame)
S3method(as.iucn,iucn)
S3method(as.iucn,list)
S3method(as.iucn,numeric)
S3method(as.natservid,character)
S3method(as.natservid,data.frame)
S3method(as.natservid,list)
S3method(as.natservid,natservid)
S3method(as.natservid,numeric)
S3method(as.nbnid,character)
S3method(as.nbnid,data.frame)
S3method(as.nbnid,list)
S3method(as.nbnid,nbnid)
S3method(as.pow,character)
S3method(as.pow,data.frame)
S3method(as.pow,list)
S3method(as.pow,pow)
S3method(as.tolid,character)
S3method(as.tolid,data.frame)
S3method(as.tolid,list)
S3method(as.tolid,numeric)
S3method(as.tolid,tolid)
S3method(as.tpsid,character)
S3method(as.tpsid,data.frame)
S3method(as.tpsid,list)
S3method(as.tpsid,numeric)
S3method(as.tpsid,tpsid)
S3method(as.tsn,character)
S3method(as.tsn,data.frame)
S3method(as.tsn,list)
S3method(as.tsn,numeric)
S3method(as.tsn,tsn)
S3method(as.uid,character)
S3method(as.uid,data.frame)
S3method(as.uid,list)
S3method(as.uid,numeric)
S3method(as.uid,uid)
S3method(as.wiki,character)
S3method(as.wiki,data.frame)
S3method(as.wiki,list)
S3method(as.wiki,numeric)
S3method(as.wiki,wiki)
S3method(as.wormsid,character)
S3method(as.wormsid,data.frame)
S3method(as.wormsid,list)
S3method(as.wormsid,numeric)
S3method(as.wormsid,wormsid)
S3method(cbind,classification)
S3method(cbind,classification_ids)
S3method(children,boldid)
S3method(children,default)
S3method(children,ids)
S3method(children,tsn)
S3method(children,uid)
S3method(children,wormsid)
S3method(classification,boldid)
S3method(classification,default)
S3method(classification,eolid)
S3method(classification,gbifid)
S3method(classification,ids)
S3method(classification,natservid)
S3method(classification,nbnid)
S3method(classification,pow)
S3method(classification,tolid)
S3method(classification,tpsid)
S3method(classification,tsn)
S3method(classification,uid)
S3method(classification,wiki)
S3method(classification,wormsid)
S3method(comm2sci,default)
S3method(comm2sci,tsn)
S3method(comm2sci,uid)
S3method(downstream,boldid)
S3method(downstream,default)
S3method(downstream,gbifid)
S3method(downstream,ids)
S3method(downstream,tsn)
S3method(downstream,uid)
S3method(downstream,wormsid)
S3method(id2name,boldid)
S3method(id2name,default)
S3method(id2name,gbifid)
S3method(id2name,tolid)
S3method(id2name,tsn)
S3method(id2name,uid)
S3method(id2name,wormsid)
S3method(iucn_status,default)
S3method(iucn_status,iucn_summary)
S3method(iucn_summary,character)
S3method(iucn_summary,default)
S3method(lowest_common,default)
S3method(lowest_common,gbifid)
S3method(lowest_common,tolid)
S3method(lowest_common,tsn)
S3method(lowest_common,uid)
S3method(plot,classtree)
S3method(print,classtree)
S3method(print,tax_agg)
S3method(rbind,classification)
S3method(rbind,classification_ids)
S3method(sci2comm,default)
S3method(sci2comm,iucn)
S3method(sci2comm,tsn)
S3method(sci2comm,uid)
S3method(sci2comm,wormsid)
S3method(synonyms,default)
S3method(synonyms,ids)
S3method(synonyms,iucn)
S3method(synonyms,nbnid)
S3method(synonyms,pow)
S3method(synonyms,tpsid)
S3method(synonyms,tsn)
S3method(synonyms,wormsid)
S3method(synonyms_df,default)
S3method(synonyms_df,synonyms)
S3method(tax_rank,character)
S3method(tax_rank,default)
S3method(tax_rank,numeric)
S3method(upstream,default)
S3method(upstream,ids)
S3method(upstream,tsn)
export(apgFamilies)
export(apgOrders)
export(apg_lookup)
export(as.boldid)
export(as.colid)
export(as.eolid)
export(as.gbifid)
export(as.iucn)
export(as.natservid)
export(as.nbnid)
export(as.pow)
export(as.tolid)
export(as.tpsid)
export(as.tsn)
export(as.ubioid)
export(as.uid)
export(as.wiki)
export(as.wormsid)
export(bold_children)
export(bold_downstream)
export(bold_ping)
export(bold_search)
export(children)
export(class2tree)
export(classification)
export(col_children)
export(col_classification)
export(col_downstream)
export(col_ping)
export(col_search)
export(comm2sci)
export(downstream)
export(eol_dataobjects)
export(eol_hierarchy)
export(eol_invasive)
export(eol_pages)
export(eol_ping)
export(eol_search)
export(eubon)
export(eubon_capabilities)
export(eubon_children)
export(eubon_hierarchy)
export(eubon_search)
export(fg_all_updated_names)
export(fg_author_search)
export(fg_deprecated_names)
export(fg_epithet_search)
export(fg_name_by_key)
export(fg_name_full_by_lsid)
export(fg_name_search)
export(fg_ping)
export(gbif_downstream)
export(gbif_name_usage)
export(gbif_parse)
export(gbif_ping)
export(genbank2uid)
export(get_boldid)
export(get_boldid_)
export(get_colid)
export(get_colid_)
export(get_eolid)
export(get_eolid_)
export(get_gbifid)
export(get_gbifid_)
export(get_genes)
export(get_genes_avail)
export(get_ids)
export(get_ids_)
export(get_iucn)
export(get_natservid)
export(get_natservid_)
export(get_nbnid)
export(get_nbnid_)
export(get_pow)
export(get_pow_)
export(get_seqs)
export(get_tolid)
export(get_tolid_)
export(get_tpsid)
export(get_tpsid_)
export(get_tsn)
export(get_tsn_)
export(get_ubioid)
export(get_ubioid_)
export(get_uid)
export(get_uid_)
export(get_wiki)
export(get_wiki_)
export(get_wormsid)
export(get_wormsid_)
export(getkey)
export(gisd_isinvasive)
export(gna_data_sources)
export(gna_parse)
export(gna_search)
export(gna_verifier)
export(gni_details)
export(gni_parse)
export(gni_seach)
export(gnr_datasources)
export(gnr_resolve)
export(id2name)
export(ion)
export(iplant_resolve)
export(ipni_ping)
export(ipni_search)
export(itis_acceptname)
export(itis_downstream)
export(itis_getrecord)
export(itis_hierarchy)
export(itis_kingdomnames)
export(itis_lsid)
export(itis_name)
export(itis_native)
export(itis_ping)
export(itis_refs)
export(itis_taxrank)
export(itis_terms)
export(iucn_getname)
export(iucn_id)
export(iucn_status)
export(iucn_summary)
export(lowest_common)
export(names_list)
export(nbn_classification)
export(nbn_ping)
export(nbn_search)
export(nbn_synonyms)
export(ncbi_children)
export(ncbi_downstream)
export(ncbi_get_taxon_summary)
export(ncbi_getbyid)
export(ncbi_getbyname)
export(ncbi_ping)
export(ncbi_search)
export(phylomatic_format)
export(phylomatic_tree)
export(plantminer)
export(pow_lookup)
export(pow_search)
export(pow_synonyms)
export(rankagg)
export(resolve)
export(sci2comm)
export(scrapenames)
export(status_codes)
export(synonyms)
export(synonyms_df)
export(tax_agg)
export(tax_name)
export(tax_rank)
export(taxize_capwords)
export(taxize_cite)
export(taxize_ldfast)
export(taxize_options)
export(taxon_clear)
export(taxon_last)
export(tnrs)
export(tnrs_sources)
export(tol_resolve)
export(tp_acceptednames)
export(tp_accnames)
export(tp_classification)
export(tp_dist)
export(tp_namedistributions)
export(tp_namereferences)
export(tp_refs)
export(tp_search)
export(tp_summary)
export(tp_synonyms)
export(tpl_families)
export(tpl_get)
export(tpl_search)
export(tropicos_ping)
export(ubio_classification)
export(ubio_classification_search)
export(ubio_id)
export(ubio_ping)
export(ubio_search)
export(ubio_synonyms)
export(upstream)
export(use_entrez)
export(use_eol)
export(use_iucn)
export(use_tropicos)
export(vascan_ping)
export(vascan_search)
export(worms_downstream)
importFrom(R6,R6Class)
importFrom(ape,as.phylo.hclust)
importFrom(ape,plot.phylo)
importFrom(ape,read.tree)
importFrom(cli,cat_line)
importFrom(cli,rule)
importFrom(cli,symbol)
importFrom(crayon,style)
importFrom(crul,AsyncVaried)
importFrom(crul,HttpClient)
importFrom(crul,HttpRequest)
importFrom(crul,upload)
importFrom(data.table,as.data.table)
importFrom(data.table,dcast)
importFrom(data.table,melt)
importFrom(data.table,rbindlist)
importFrom(data.table,setDF)
importFrom(data.table,transpose)
importFrom(graphics,plot)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(methods,as)
importFrom(methods,is)
importFrom(stats,aggregate)
importFrom(stats,as.dist)
importFrom(stats,complete.cases)
importFrom(stats,hclust)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(utils,URLencode)
importFrom(utils,browseURL)
importFrom(utils,citation)
importFrom(utils,download.file)
importFrom(utils,read.delim)
importFrom(utils,tail)
importFrom(utils,write.table)
importFrom(xml2,as_list)
importFrom(xml2,read_xml)
importFrom(xml2,xml_children)
importFrom(xml2,xml_find_all)
importFrom(xml2,xml_find_first)
importFrom(xml2,xml_name)
importFrom(xml2,xml_ns)
importFrom(xml2,xml_text)
importFrom(zoo,na.locf)

# taxize

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![rworkflows](https://github.com/ropensci/taxize/workflows/rworkflows/badge.svg)](https://github.com/ropensci/taxize/actions/)
[![codecov](https://codecov.io/gh/ropensci/taxize/branch/master/graph/badge.svg)](https://app.codecov.io/gh/ropensci/taxize)
[![rstudio mirror
downloads](https://cranlogs.r-pkg.org/badges/taxize)](https://github.com/r-hub/cranlogs.app)
[![cran
version](https://www.r-pkg.org/badges/version/taxize)](https://cran.r-project.org/package=taxize)

`taxize` allows users to search over many taxonomic data sources for
species names (scientific and common) and download up and downstream
taxonomic hierarchical information - among other things.

Package documentation: <https://docs.ropensci.org/taxize/>

## Installation

### Stable version from CRAN

``` r
install.packages("taxize")
```

### Development version from GitHub

Windows users install Rtools first.

``` r
install.packages("remotes")
remotes::install_github("ropensci/taxize")
```

``` r
library('taxize')
```

## Screencast

<a href="https://vimeo.com/92883063"><img src="man/figures/screencast.png" width="400"></a>

## Contributing

See our
[CONTRIBUTING](https://github.com/ropensci/taxize/blob/master/.github/CONTRIBUTING.md)
document.

## Contributors

Collected via GitHub Issues: honors all contributors in alphabetical
order. Code contributors are in bold.

afkoeppel - ahhurlbert - albnd - Alectoria - andzandz11 - **anirvan** -
antagomir - **arendsee** - ArielGreiner - arw36 - ashenkin -
**ashiklom** - benjaminschwetz - **benmarwick** - bienflorencia -
binkySallly - bomeara - BridgettCollis - bw4sz - **cboettig** -
cdeterman - ChrKoenig - chuckrp - clarson2191 - claudenozeres -
cmzambranat - cparsania - daattali - DanielGMead - DarrenObbard -
davharris - davidvilanova - diogoprov - **dlebauer** - dlenz1 -
dougwyu - dschlaep - **EDiLD** - edwbaker - emhart - eregenyi -
fdschneider - fgabriel1891 - fischhoff - **fmichonneau** - **fozy81** -
**gedankenstuecke** - gimoya - git-og - glaroc - **gpli** - gustavobio -
hlapp - **ibartomeus** - **Ironholds** - jabard89 - jangorecki -
**jarioksa** - jebyrnes - **jeroen** - **jimmyodonnell** - joelnitta -
johnbaums - jonmcalder - jordancasey - **josephwb** - jsgosnell -
JulietteLgls - **jwilk** - kamapu - **karthik** - **katrinleinweber** -
KevCaz - kgturner - kmeverson - Koalha - **ljvillanueva** - **maelle** -
Markus2015 - matutosi - mcsiple - MikkoVihtakari - millerjef -
miriamgrace - mpnelsen - MUSEZOOLVERT - nate-d-olson - nmatzke - npch -
ocstringham - p-neves - p-schaefer - padpadpadpad - paternogbc -
**patperu** - pederengelstad - philippi - Phylloxera - **pmarchand1** -
pozsgaig - pssguy - **raredd** - rec3141 - **Rekyt** - RodgerG -
rossmounce - sariya - sastoudt - scelmendorf - **sckott** -
SimonGoring - snsheth - snubian - Squiercg - sunray1 - **taddallas** -
tdjames1 - tmkurobe - toczydlowski - tpaulson1 - tpoisot -
**TrashBirdEcology** - **trvinh** - **vijaybarve** - wcornwell -
willpearse - wpetry - yhg926 - **zachary-foster**

## Road map

Check out our
[milestones](https://github.com/ropensci/taxize/milestones) to see what
we plan to get done for each version.

## Meta

- Please [report any issues or
  bugs](https://github.com/ropensci/taxize/issues).
- License: MIT
- Get citation information for `taxize` in R doing
  `citation(package = 'taxize')`
- Please note that this package is released with a [Contributor Code of
  Conduct](https://ropensci.org/code-of-conduct/). By contributing to
  this project, you agree to abide by its terms.

[![rofooter](https://ropensci.org/public_images/github_footer.png)](https://ropensci.org)

#' read newick files, copied directly from phytools
#' @param file file path
#' @param text text string
#' @export
#' @keywords internal
read_newick <- function (file = "", text) 
{
    if (file != "") 
        text <- scan(file, sep = "\n", what = "character")
    if (length(text) > 1) {
        tree <- lapply(text, newick)
        class(tree) <- "multiPhylo"
    }
    else tree <- newick(text)
    return(tree)
}


#' Newick function required by read_newick
#' @param text input text
#' @export
#' @keywords internal
newick <- function (text) 
{
    text <- unlist(strsplit(text, NULL))
    tip.label <- vector(mode = "character")
    node.label <- vector(mode = "character")
    edge <- matrix(c(1, NA), 1, 2)
    edge.length <- vector()
    currnode <- 1
    Nnode <- currnode
    i <- j <- k <- 1
    while (text[i] != ";") {
        if (text[i] == "(") {
            if (j > nrow(edge)) 
                edge <- rbind(edge, c(NA, NA))
            edge[j, 1] <- currnode
            i <- i + 1
            if (is.na(match(text[i], c("(", ")", ",", ":", ";")))) {
                temp <- get_label(text, i)
                tip.label[k] <- temp$label
                i <- temp$end
                edge[j, 2] <- -k
                k <- k + 1
                if (text[i] == ":") {
                  temp <- getEdgeLength(text, i)
                  edge.length[j] <- temp$edge.length
                  i <- temp$end
                }
            }
            else if (text[i] == "(") {
                Nnode <- Nnode + 1
                currnode <- Nnode
                edge[j, 2] <- currnode
            }
            j <- j + 1
        }
        else if (text[i] == ")") {
            i <- i + 1
            if (is.na(match(text[i], c("(", ")", ",", ":", ";")))) {
                temp <- get_label(text, i)
                node.label[currnode] <- temp$label
                i <- temp$end
            }
            if (text[i] == ":") {
                temp <- getEdgeLength(text, i)
                if (currnode > 1) {
                  ii <- match(currnode, edge[, 2])
                  edge.length[ii] <- temp$edge.length
                }
                else root.edge <- temp$edge.length
                i <- temp$end
            }
            if (currnode > 1) 
                currnode <- edge[match(currnode, edge[, 2]), 
                  1]
        }
        else if (text[i] == ",") {
            if (j > nrow(edge)) 
                edge <- rbind(edge, c(NA, NA))
            edge[j, 1] <- currnode
            i <- i + 1
            if (is.na(match(text[i], c("(", ")", ",", ":", ";")))) {
                temp <- get_label(text, i)
                tip.label[k] <- temp$label
                i <- temp$end
                edge[j, 2] <- -k
                k <- k + 1
                if (text[i] == ":") {
                  temp <- getEdgeLength(text, i)
                  edge.length[j] <- temp$edge.length
                  i <- temp$end
                }
            }
            else if (text[i] == "(") {
                Nnode <- Nnode + 1
                currnode <- Nnode
                edge[j, 2] <- currnode
            }
            j <- j + 1
        }
    }
    Ntip <- k - 1
    edge[edge > 0] <- edge[edge > 0] + Ntip
    edge[edge < 0] <- -edge[edge < 0]
    edge.length[is.na(edge.length)] <- 0
    if (length(edge.length) == 0) 
        edge.length <- NULL
    node.label[is.na(node.label)] <- ""
    if (length(node.label) == 0) 
        node.label <- NULL
    tree <- list(edge = edge, Nnode = as.integer(Nnode), tip.label = tip.label, 
        edge.length = edge.length, node.label = node.label)
    class(tree) <- "phylo"
    return(tree)
}

#' get label function from phytools
#' @param text input text
#' @param start start at
#' @export
#' @keywords internal
get_label <- function (text, start) 
{
    i <- 0
    label <- vector()
    while (is.na(match(text[i + start], c(",", ":", ")")))) {
        label[i + 1] <- text[i + start]
        i <- i + 1
    }
    return(list(label = paste(label, collapse = ""), end = i + 
        start))
}

out <- make_taxon(genus="Poa", epithet="annua", authority="L.")
out <- make_taxon(genus="Poa", epithet="annua", authority="L.", family='Poaceae', clazz='Poales', kingdom='Plantae', variety='annua')
out@binomial
out@binomial@canonical
out@binomial@species
out@binomial@authority
out@classification
out@classification@family
out[['family']] # get a single rank
out['family','kingdom'] # get a range of ranks
out['variety','genus']
gethier(out) # get hierarchy as data.frame

make_taxon <- function(genus, epithet, authority, ...){
  res <- new("binomial", 
             genus=genus, epithet=epithet, canonical=paste(genus,epithet,collapse=" "), 
             species=paste(genus,epithet,authority,collapse=" "), authority=authority)
  hier <- new("classification", genus=genus, species=paste(genus,epithet,collapse=" "), ...)
  new("taxon", binomial=res, classification=hier)
}

setClass("binomial", slots = c(genus="character", epithet="character", canonical="character", species="character", authority="character"))

setClass("classification", slots = c(
  superkingdom="character",
  kingdom="character",
  subkingdom="character",
  infrakingdom="character",
  division="character",
  phylum="character",
  subdivision="character",
  infradavision="character",
  superclass="character",
  clazz="character",
  subclass="character",
  infraclass="character",
  superorder="character",
  order="character",
  suborder="character",
  infraorder="character",
  superfamily="character",
  family="character",
  subfamily="character",
  tribe="character",
  subtribe="character",
  genus="character",
  subgenus="character",
  section="character",
  subsection="character",
  species="character",
  subspecies="character",
  variety="character", 
  race="character",
  subvariety="character",
  stirp="character",
  morph="character", 
  form="character", 
  aberration="character", 
  subform="character", 
  unspecified="character"
), prototype = prototype(superkingdom="",
                         kingdom="",
                         subkingdom="",
                         infrakingdom="",
                         division="",
                         phylum="",
                         subdivision="",
                         infradavision="",
                         superclass="",
                         clazz="",
                         subclass="",
                         infraclass="",
                         superorder="",
                         order="",
                         suborder="",
                         infraorder="",
                         superfamily="",
                         family="",
                         subfamily="",
                         tribe="",
                         subtribe="",
                         genus="",
                         subgenus="",
                         section="",
                         subsection="",
                         species="",
                         subspecies="",
                         variety="", 
                         race="",
                         subvariety="",
                         stirp="",
                         morph="", 
                         form="", 
                         aberration="", 
                         subform="", 
                         unspecified=""))
setClass("taxon", slots = c(binomial = 'binomial', classification = 'classification'))
# setClass("taxon", contains = c("binomial","classification"))

setMethod("[[", "taxon", function(x, i, ...){
  tmp <- x@classification
  slot(tmp, i)
})

setMethod("[", "taxon", function(x, i, j, ...){
  tmp <- x@classification
  nn <- slotNames(tmp)
  from <- match(j, nn)
  to <- match(i, nn)
  vapply(nn[to:from], function(g) slot(tmp, g), "")
})

setGeneric("gethier", function(x) standardGeneric("gethier"))
setMethod("gethier", "taxon", function(x){
  tmp <- x@classification
  nn <- slotNames(tmp)
  vals <- vapply(nn, function(g) slot(tmp, g), "", USE.NAMES = FALSE)
  data.frame(rank=nn, value=vals, stringsAsFactors = FALSE)
})

library(taxize)
orders <- apgOrders()
orders <- tibble::as_tibble(orders)
orders
apg_orders <- dplyr::arrange(orders, order) %>% 
  dplyr::mutate(accepted_name = ifelse(is.na(synonym), order, synonym))

families <- apgFamilies()
families$family <- gsub("\"", "", families$family)
families <- tibble::as_tibble(families)
families
apg_families <- dplyr::arrange(families, family) %>% 
  dplyr::mutate(accepted_name = ifelse(is.na(synonym), family, synonym))

save(apg_orders, file = "data/apg_orders.RData", version = 2)
save(apg_families, file = "data/apg_families.RData", version = 2)

#' NatureServe
#'
#' @export
#' @param x (character) A name to query. Required.
#' @param key (character) API key. Required.
#' @param ... Curl options passed on to \code{\link[httr]{GET}}
#' @references \url{https://services.natureserve.org/index.jsp}
#' @return A data.frame, with columns:
#' \itemize{
#'  \item sciname - Scientfic name
#'  \item comname - Common name
#'  \item uid - UID - the taxonomic identifier NatureServe uses
#'  \item uri - URL to get to info online for the taxon
#'  \item taxcomment - comments about the taxon, if any
#' }
#' @examples \dontrun{
#' head(ns_search("Ruby*"))
#' ns_search("Helianthus annuus")
#'
#' ns_data(uid = 'ELEMENT_GLOBAL.2.100925')
#' }
ns_search <- function(x, key = "72ddf45a-c751-44c7-9bca-8db3b4513347", ...) {
  x <- GET(paste0(ns_base(), '/v1/globalSpecies/list/nameSearch'),
           query = list(name = x, NSAccessKeyId = key))
  stop_for_status(x)
  xml <- xml2::read_xml(content(x, "text"))
  kids <- xml2::xml_children(xml2::xml_children(xml)[[2]])
  dat <- lapply(kids, function(z) {
    data.frame(sapply(xml_children(z), function(x) {
      as.list(setNames(xml_text(x), xml_name(x)))
    }), stringsAsFactors = FALSE)
  })
  df <- setNames(rbind.fill(dat),
                 c('uid', 'uri', 'sciname', 'comname', 'taxcomment'))
  df <- move_col2(df, "uid")
  df <- move_col2(df, "uri")
  move_col2(df, "taxcomment")
}

#' @export
#' @rdname nature_serve
ns_data <- function(uid, key = "72ddf45a-c751-44c7-9bca-8db3b4513347", ...) {
  x <- GET(paste0(ns_base(), '/v1.1/globalSpecies/comprehensive'),
           query = list(uid = uid, NSAccessKeyId = key))
  stop_for_status(x)
  xml <- xml2::read_xml(content(x, "text"))
  kids <- xml2::xml_children(xml2::xml_children(xml)[[1]])
  dat <- lapply(kids, function(z) {
    data.frame(sapply(xml_children(z), function(x) {
      as.list(setNames(xml_text(x), xml_name(x)))
    }), stringsAsFactors = FALSE)
  })
  df <- setNames(rbind.fill(dat),
                 c('uid', 'uri', 'sciname', 'comname', 'taxcomment'))
  df <- move_col2(df, "uid")
  df <- move_col2(df, "uri")
  move_col2(df, "taxcomment")
}

ns_base <- function() 'https://services.natureserve.org/idd/rest/ns'

#' Format tree string, submit to Phylomatic, get newick or nexml tree.
#' 
#' Newer input format to the API, doesn't work that well. Do use phylomatic_tree()
#' 
#' @import httr ape
#' @param taxa Phylomatic format input of taxa names.
#' @param get 'GET' or 'POST' format for submission to the website.
#' @param informat One of newick, nexml, or cdaordf. If using a stored tree, 
#'    informat should always be newick.
#' @param method One of phylomatic or convert
#' @param storedtree One of R20120829 (Phylomatic tree R20120829 for plants), 
#' 		smith2011 (Smith 2011, plants), or binindaemonds2007 (Bininda-Emonds 2007, 
#'   	mammals).
#' @param outformat One of newick, nexml, or fyt.
#' @param clean Return a clean tree or not.
#' @details Use the web interface here http://phylodiversity.net/phylomatic/
#' @return Newick formatted tree, or nexml xml object.
#' @noRd
#' @examples \donttest{ 
#' # Input taxonomic names
#' taxa <- c("Collomia grandiflora", "Lilium lankongense", "Helianthus annuus")
#' tree <- phylomatic_tree2(taxa=taxa, get = 'POST', informat='newick', 
#'    method = "phylomatic", storedtree = "smith2011", 
#'    outformat = "newick", clean = "true")
#' plot(tree)
#' 
#' # Lots of names
#' taxa <- c("Collomia grandiflora", "Lilium lankongense", "Phlox diffusa", 
#'           "Iteadaphne caudata", "Nicotiana tomentosa", "Gagea sarmentosa")
#' tree <- phylomatic_tree2(taxa=taxa, get = 'POST', informat='newick', 
#'                         method = "phylomatic", storedtree = "smith2011",
#'                         outformat = "newick", clean = "true")
#' plot(tree)
#' 
#' # Output NeXML format
#' taxa <- c("Gonocarpus leptothecus", "Gonocarpus leptothecus", "Impatiens davidis")
#' out <- phylomatic_tree2(taxa=taxa, get = 'POST', informat='newick', method = "phylomatic", 
#'    storedtree = "smith2011", outformat = "nexml", clean = "true")
#' library(RNeXML)
#' read.nexml(out, type="nexml", asText=TRUE)
#' }

phylomatic_tree2 <- function(taxa, get = 'GET', informat = "newick", method = "phylomatic", 
  storedtree = "smith2011", outformat = "newick", clean = "true")
{
  url <- "http://phylodiversity.net/phylomatic/pmws"
  taxa <- sapply(taxa, function(x) gsub("\\s", "_", x), USE.NAMES=FALSE)
  if (length(taxa) > 1) { taxa <- paste(taxa, collapse = "\n") } else { taxa <- taxa }  
  args <- compact(list(taxa = taxa, informat = informat, method = method, 
                       storedtree = storedtree, outformat = outformat, clean = clean))
  get <- match.arg(get, choices=c("GET",'POST'))
  out <- eval(parse(text=get))(url, query=args)
  stop_for_status(out)
  tt <- content(out, as="text")
  
  outformat <- match.arg(outformat, choices=c("nexml",'newick'))
  getnewick <- function(x){
    tree <- gsub("\n", "", x[[1]])
    read.tree(text = colldouble(tree))
  }

  switch(outformat, 
         nexml = tt,
         newick = getnewick(tt))
}

(out <- make_taxon(genus="Poa", epithet="annua", authority="L."))
(out <- make_taxon(genus="Poa", epithet="annua", authority="L.", 
                  family='Poaceae', clazz='Poales', kingdom='Plantae', variety='annua'))
(out <- make_taxon(genus="Poa"))
make_taxon(epithet="annua") # errors
out@binomial
out@binomial@canonical
out@binomial@species
out@binomial@authority
out@classification
out@classification@family
out[['family']] # get a single rank
out['family','kingdom'] # get a range of ranks
out['variety','genus'] # get a range of ranks
gethier(out) # get hierarchy as data.frame


df <- data.frame(rank=c('family','tribe','subtribe','genus','subgenus','species'), 
                 name=c('Helianthi','Helianthi','Helianthi','Poa','Festuci','Poa annua'), 
                 id=c(1,2,3,4,5,6),
                 stringsAsFactors = FALSE)

apply(df, 1, make_taxon_fromclass)

make_taxon_fromclass <- function(x, authority="none", ...){
  rank <- x[['rank']]
  name <- x[['name']]
  id <- as.numeric(x[['id']])
  res <- list(new("taxonref", rank=rank, name=name, id=id))
  names(res) <- rank
  res[1]
}

#   species <- x[ x$rank %in% 'species', 'name']
#   epithet <- strsplit(species, " ")[[1]][[2]]
#   res <- new("binomial", 
#              genus=genus, 
#              epithet=epithet, 
#              canonical=paste(genus,epithet,collapse=" "), 
#              species=paste(genus,epithet,authority,collapse=" "), 
#              authority=authority)
#   hier <- new("classification", 
#               genus=new("taxonref", name=genus), 
#               species=new("taxonref", name=paste(genus,epithet,collapse=" ")))
#   if(length(input) > 0){
#     output <- list()
#     for(i in seq_along(input)){
#       slot(hier, names(input)[i]) <- new("taxonref", name=input[[i]])
#     }
#   }
#   new("taxon", binomial=res, classification=hier)
# }

make_taxon <- function(genus="none", epithet="none", authority="none", ...){
  if(genus=='none') stop("You must supply at least genus")
  res <- new("binomial", 
             genus=genus, 
             epithet=epithet, 
             canonical=paste(genus,epithet,collapse=" "), 
             species=paste(genus,epithet,authority,collapse=" "), 
             authority=authority)
  input <- list(...)
  hier <- new("classification", 
              genus=new("taxonref", name=genus), 
              species=new("taxonref", name=paste(genus,epithet,collapse=" ")))
  if(length(input) > 0){
    output <- list()
    for(i in seq_along(input)){
      slot(hier, names(input)[i]) <- new("taxonref", name=input[[i]])
    }
  }
  new("taxon", binomial=res, classification=hier)
}

setClass("binomial", slots = c(genus="character", epithet="character", canonical="character", species="character", authority="character"))

setClass("taxonref", slots = c(rank='character', name='character', id='numeric', source='character'), 
         prototype = prototype(rank='none', name='none', id=NaN, source='none'))

setClass("ListOfTaxonRefs", slots = c(names="character"), contains="list")

setAs("character", "taxonref", function(from) new("taxonref", name=from))

# setMethod("combinetaxonref", "ListOfTaxonRefs", function(...){
#   input <- list(...)
#   assert_that(all(sapply(input, is, "ListOfTaxonRefs")))
#   
# })

# new("ListOfTaxonRefs", lapply(list(kingdom="adfa", family="adf"), as, "taxonref"))

setClass("classification", slots = c(
  kingdom="taxonref",
  subkingdom="taxonref",
  infrakingdom="taxonref",
  division="taxonref",
  phylum="taxonref",
  subdivision="taxonref",
  infradavision="taxonref",
  superclass="taxonref",
  clazz="taxonref",
  subclass="taxonref",
  infraclass="taxonref",
  superorder="taxonref",
  order="taxonref",
  suborder="taxonref",
  infraorder="taxonref",
  superfamily="taxonref",
  family="taxonref",
  subfamily="taxonref",
  tribe="taxonref",
  subtribe="taxonref",
  genus="taxonref",
  subgenus="taxonref",
  section="taxonref",
  subsection="taxonref",
  species="taxonref",
  subspecies="taxonref",
  variety="taxonref", 
  race="taxonref",
  subvariety="taxonref",
  stirp="taxonref",
  morph="taxonref", 
  form="taxonref", 
  aberration="taxonref", 
  subform="taxonref", 
  unspecified="taxonref"
), contains = "list")
# , contains = "list")
setClass("taxon", slots = c(binomial = 'binomial', classification = 'classification'))
setClass("ListOfTaxa", slots = c(taxon = 'taxon'), prototype = prototype(list()), contains = 'list')
# 
# new('ListOfTaxa', taxon=out)
# out2 <- list(make_taxon(genus="Poa", epithet="annua", authority="L."), 
#              make_taxon(genus="Helianthus", epithet="annuus", authority="Baker"))
# gg <- new('ListOfTaxa', lapply(out2, as, "taxon"))
# class(gg)
# length(gg@.Data)

setMethod("[[", "taxon", function(x, i, ...){
  tmp <- x@classification
  slot(tmp, i)@name
})

setMethod("[", "taxon", function(x, i, j, ...){
  tmp <- x@classification
  nn <- slotNames(tmp)
  from <- match(j, nn)
  to <- match(i, nn)
  vapply(nn[to:from], function(g) slot(tmp, g)@name, "")
})

setGeneric("gethier", function(x) standardGeneric("gethier"))
setMethod("gethier", "taxon", function(x){
  tmp <- x@classification
  nn <- slotNames(tmp)[-1]
  vals <- vapply(nn, function(g) slot(tmp, g)@name, "", USE.NAMES = FALSE)
  data.frame(rank=nn, value=vals, stringsAsFactors = FALSE)
})



####
# subset data.frame using taxonomy
df <- data.frame(family=c('Asteraceae','Asteraceae','Asteraceae','Poaceae','Poaceae','Poaceae'), 
                 tribe=c('Helianthi','Helianthi','Helianthi','Poaeae','Festuci','Poaeae'), 
                 genus=c('Helianthus','Helianthus','Madia','Poa','Festuca','Holodiscus'), stringsAsFactors = FALSE)
df_tax <- new('taxonDataFrame', data=df)
df_tax['family','Asteraceae']
df_tax['genus','Madia']
df_tax['tribe','Helianthi']

setClass("taxonDataFrame", slots = c(data = 'data.frame'))
setMethod("[", "taxonDataFrame", function(x, i, j, ...){
  tmp <- x@data
  tmp[ tmp[[i]] %in% j, ]
})

#  from https://www.birdpop.org/pages/birdSpeciesCodes.php
download.file("http://www.birdpop.org/docs/misc/IBPAOU.zip", destfile = "~/IBPAOU.zip")
unzip("~/IBPAOU.zip", exdir = "~/")
res <- read.csv("~/IBP-Alpha-Codes20.csv")
head(res)

# checklist from http://checklist.aou.org/taxa/
chklst <- read.csv('http://checklist.aou.org/taxa.csv', header = TRUE, sep = ",", stringsAsFactors = FALSE)
head(chklst)

pryr::object_size(res)
pryr::object_size(chklst)

#' Run Phylocom's bladj from R
#' 
#' @import ape
#' @importFrom phytools read_newick
#' @param tree (phylo/character) If left NULL, it is expected that you already have a 
#' phylo file with your newick tree in your directory with your phylocom executable
#' @param ages (data.frame) If left NULL, it is expected that you already have an 
#' ages file with your node names and ages in your directory with your Phylocom executable
#' @param path (character) Path to the folder with at least the Phylocom executable
#' @param fixroot (logical) If TRUE root name is changed to the oldest root in your ages file.
#' If FALSE (default), 
#' @export
#' @examples \donttest{
#' taxa <- names_list("species", 15)
#' tree <- phylomatic_tree(taxa=taxa, get = 'POST')
#' # set path to where your dir is with phylocom executable, and optionally phylo 
#' # and/or ages files
#' path = "path to phylocom executable directory" 
#' g <- rbladj(tree, path=path, fixroot=TRUE)
#' g$edge.length
#' plot(g)
#' }
rbladj <- function(tree=NULL, ages=NULL, path=NULL, fixroot=FALSE)
{
  #download phylocom
  download.file("http://phylodiversity.net/phylocom/phylocom-4.2.zip")
  
  #set path
  if(!is.null(path)){
    setwd(path)
  } else
  { stop("You must provide a path to where phylocom executable is stored on your machine") }
  
  #phylo
  if(!is.null(tree)){
    if(any(class(tree) == "character")){ tree <- read_newick(file=paste(path,"phylo",sep="")) } else
      if(any(class(tree) == "phylo")) { tree <- tree } else {
        stop("need newick tree as text string or phylo object")
      }
    setwd(path)
    write.tree(tree, "phylo")
  }
  
  #ages
  if(is.null(ages)){
    tmp <- tryCatch(read.table(paste(path, "ages", sep="")), error = function(e) e)
    if(any(class(tmp) == "error")){
      stop("You must provide an ages data.frame or a path to the ages file")
    }
  } else
  {
    if(is(ages, "data.frame")){
      write.table(ages, file = paste(path,"ages",sep=""), sep = "\t", col.names = FALSE, row.names = FALSE, quote = FALSE)      
    } else {
      stop("ages, it not a path, must be a data.frame with two columns, node name, and node age")
    }
  }
  
  # check if root node in phylogeny is in the ages file
  if(is.null(tree)){ tree <- read_newick(file=paste(path,"phylo",sep="")) }
  if(is.null(ages)){ 
    ages <- read.table(paste(path, "ages", sep="")) 
    names(ages) <- c("names","ages")
  }
  rootnode <- as.character(tree$node.label[1])
  if(!rootnode %in% as.character(ages[,1])){
    if(fixroot){
      ages_sorted <- sort_df(ages, "ages")
      newroot <- as.character(ages_sorted[nrow(ages_sorted),1])
      tree <- read.tree(text=sub(rootnode, newroot, write.tree(tree)))
      setwd(path)
      write.tree(tree, "phylo")
    } else
    {
      stop("Your root node in your phylogeny must be in your ages file - names must match exactly")
    }
  }
  
  #bladj
  if(.Platform$OS.type != "unix"){
    setwd(path)
    shell("phylocom bladj > phyloout.txt")
  } else
  {
    setwd(path)
    system("./phylocom bladj > phyloout.txt")    
  }
  
  #read tree back in
  tr <- read_newick(paste(path, "phyloout.txt", sep=""))
  return( tr )
}

#' Function to allow user to set timeout on a function call.
#'
#' This is separate from the curl timeout.ms parameter
#'
#' @importFrom R.utils evalWithTimeout
#' @param test_fn The function call to test
#' @param tlimit = 120 A timeout in seconds
#' @param defaultvalue Default value
#' @return If in a loop, passes on to next thing, or if not, then gives back message
#' @examples \donttest{
#' library("taxize")
#' try_with_timeout(eol_search('Salix'))
#' try_with_timeout(eol_search('Salix'), tlimit=10)
#' }

try_with_timeout <- function(test_fn, tlimit = 30, defaultvalue = "TimedOut") {
  results <- tryCatch(expr = evalWithTimeout(test_fn, timeout = tlimit),
                      TimeoutException = function(ex) defaultvalue)
  if(is(results, "TimedOut")){
    return( defaultvalue )
  } else { results }
}

library('R6')

MultiTaxonId <- R6Class(
  "MultiTaxonId",
  #inherit = TaxonId,
  public = list(
    ids = NA,
    print = function() {
      cat(paste0("<NCBI IDs> ", paste0(self$fetch_ids(), collapse = ", ")), sep = "")
      invisible(self)
    },
    get_ids = function(x) {
      tmp <- unclass(get_uid(x))
      taxa <- list()
      for (i in seq_along(tmp)) {
        taxa[[i]] <- TaxonId$new(tmp[i], x[i], attr(tmp, "uri")[i], attr(tmp, "match")[i])
      }
      self$ids <- taxa
    },
    fetch_uris = function() {
      res <- c()
      for (i in seq_along(self$ids)) res[i] <- self$ids[[i]]$uri
      return(res)
    },
    fetch_matches = function() {
      res <- c()
      for (i in seq_along(self$ids)) res[i] <- self$ids[[i]]$match
      return(res)
    },
    fetch_taxa = function() {
      res <- c()
      for (i in seq_along(self$ids)) res[i] <- self$ids[[i]]$taxa
      return(res)
    },
    fetch_ids = function() {
      res <- c()
      for (i in seq_along(self$ids)) res[i] <- self$ids[[i]]$id
      return(res)
    }
  )
)

TaxonId <- R6Class(
  "TaxonId",
  public = list(
    id = NA,
    taxa = NA,
    uri = NA,
    match = NA,
    initialize = function(id, taxa, uri, match) {
      if (!missing(id)) self$id <- id
      if (!missing(taxa)) self$taxa <- taxa
      if (!missing(uri)) self$uri <- uri
      if (!missing(match)) self$match <- match
    },
    print = function() {
      cat(paste0("<NCBI ID> ", self$id), "\n")
      cat(paste0("  taxon: ", self$taxa), "\n")
      cat(paste0("  match: ", self$match), "\n")
      cat(paste0("  uri: ", self$uri), "\n")
      invisible(self)
    }
  )
)

`[.MultiTaxonId` <- function(x, i) {
  x$ids[i]
}

`[[.MultiTaxonId` <- function(x, i) {
  x$ids[[i]]
}

`[.TaxonId` <- function(x, i) {
  x$id[i]
}

x <- MultiTaxonId$new()
x$get_ids(c("Chironomus riparius", "Quercus douglasii", "Poa annua"))
x
x$ids
x$fetch_ids()
x$fetch_uris()
x$fetch_matches()
x$fetch_taxa()

# subset
x$ids[[1]]
x$ids[3]
x$ids[[3]]
x$ids[2:3]

# tests for classification fxn in taxize
context("classification")

# is_up <- function(seconds=3, which="itis"){
#   itisfxn <- function(x) tryCatch(itis_ping(config=timeout(x)), error=function(e) e)
#   switch(which,
#          itis = !is(itisfxn(seconds), "OPERATION_TIMEDOUT")
#   )
# }
#
# is_up()
# is_up(which = "ncbi")
#
# skip_if <- function(x){
#   if(!res) skip("API down")
#
#   expect_is(classification(c("Chironomus riparius", "aaa vva"), db = 'itis', messages = FALSE), "classification")
# }

# tpsids <- get_tpsid(sciname=c("Helianthus excubitor", "aaa vva"), messages = FALSE)
# clas_tpids <- classification(tpsids, messages = FALSE)

# clas_tp <- suppressMessages(classification(c("Helianthus excubitor", "aaa vva"), db = 'tropicos'))
# names(clas_tp) <- NULL

test_that("classification returns the correct values and classes", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("classification", {
    clas_ncbi <- classification(c("Chironomus riparius", "aaa vva"), db = 'ncbi',
                                messages = FALSE)
    names(clas_ncbi) <- NULL

    clas_itis <- classification(c("Chironomus riparius", "aaa vva"), db = 'itis',
                                messages = FALSE)
    names(clas_itis) <- NULL
  })

	expect_that(clas_ncbi[[2]], equals(NA))
	expect_that(clas_itis[[2]], equals(NA))
# 	expect_that(clas_tp[[2]], equals(NA))

	expect_is(clas_ncbi, "classification")
	expect_is(clas_ncbi[[1]], "data.frame")
	expect_equal(length(clas_ncbi), 2)

  expect_is(clas_itis, "classification")
	expect_is(clas_itis[[1]], "data.frame")
  expect_equal(length(clas_itis), 2)

# 	expect_that(clas_tp, is_a("classification"))
# 	expect_that(clas_tp[[1]], is_a("data.frame"))
# 	expect_that(length(clas_tp), equals(2))

  vcr::use_cassette("classification_get_fxn", {
    uids <- get_uid(c("Chironomus riparius", "aaa vva"), messages = FALSE)
    tsns <- get_tsn(c("Chironomus riparius", "aaa vva"), messages = FALSE)
    clas_uids <- classification(uids, messages = FALSE)
    names(clas_uids) <- NULL
    clas_tsns <- classification(tsns, messages = FALSE)
    names(clas_tsns) <- NULL
  })
  
  expect_identical(clas_uids, clas_ncbi)
  expect_equal(clas_tsns, clas_itis)
  #### FIX THESE TWO, SHOULD BE MATCHING
#   expect_identical(clas_tpids, clas_tp)
})

test_that("passing in an id works", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("classification_passing_id", {
    fromid_ncbi <- classification(9606, db = 'ncbi')
    fromid_itis <- classification(129313, db = 'itis')
    fromid_gbif <- classification(c(2704179, 2441176), db = 'gbif')
  })
  #fromid_nbn <- classification("NBNSYS0000004786", db = 'nbn')

  expect_is(fromid_ncbi, "classification")
  expect_equal(attr(fromid_ncbi, "db"), "ncbi")

  expect_is(fromid_itis, "classification")
  expect_equal(attr(fromid_itis, "db"), "itis")

  expect_is(fromid_gbif, "classification")
  expect_equal(attr(fromid_gbif, "db"), "gbif")

  #expect_is(fromid_nbn, "classification")
  #expect_equal(attr(fromid_nbn, "db"), "nbn")
})

test_that("rbind and cbind work correctly", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("classification_cbind_rbind", {
    out <- get_ids(c("Puma concolor", "Accipiter striatus"),
                   db = 'ncbi', messages = FALSE, suppress = TRUE)
    cl <- classification(out)
  })

  # rbind
  clr <- rbind(cl)
  expect_is(clr, "data.frame")
  expect_named(clr, c("name", "rank", "id", "query", "db"))

  # cbind
  clc <- cbind(cl)
  expect_is(clc, "data.frame")
  expect_gt(length(names(clc)), 50)
})

set.seed(1)
df <- theplantlist[sample(1:nrow(theplantlist), 50), ]
nn <- apply(df, 1, function(x) paste(x["genus"], x["species"], collapse = " "))

test_that("works on a variety of names", {
  skip_on_cran()
  # skip_on_ci() # keeps timing out on GH actions for unknown reason
  
  vcr::use_cassette("classification_variety_of_names", {
    x <- classification(nn[5], db = "ncbi", messages = FALSE)
    z <- classification(nn[6], db = "ncbi", messages = FALSE)
  })

	expect_is(x, "classification")
	expect_is(z, "classification")
})

test_that("queries with no results fail well", {
  skip_on_cran()
  vcr::use_cassette("classification_no_results", {
    aa <- classification("foobar", db = "itis", messages = FALSE)
    bb <- classification(get_tsn("foobar", messages = FALSE), messages = FALSE)
  })

  expect_true(is.na(unclass(aa)[[1]]))
  expect_identical(unname(aa), unname(bb))
})

test_that("all rank character strings are lower case (all letters)", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("classification_rank_is_lowercase", {
    aa <- classification(9606, db = 'ncbi', messages = FALSE)
    bb <- classification(129313, db = 'itis', messages = FALSE)
    dd <- classification(2441176, db = 'gbif', messages = FALSE)
    #ee <- classification(25509881, db = 'tropicos', messages = FALSE)
    #ff <- classification("NBNSYS0000004786", db = 'nbn', messages = FALSE)
  })

  expect_false(all(grepl("[A-Z]", aa[[1]]$rank)))
  expect_false(all(grepl("[A-Z]", bb[[1]]$rank)))
  expect_false(all(grepl("[A-Z]", dd[[1]]$rank)))
  #expect_false(all(grepl("[A-Z]", ee[[1]]$rank)))
  #expect_false(all(grepl("[A-Z]", ff[[1]]$rank)))
})


test_that("rows parameter, when used, works", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("classification_rows_param", {
    # a <- classification("Asdfafsfd", db = 'ncbi', rows = 1, messages = FALSE)
    b <- classification("Asdfafsfd", db = 'itis', rows = 1, messages = FALSE)
    d <- classification("Asdfafsfd", db = 'gbif', rows = 1, messages = FALSE)
    g <- classification("Asdfafsfd", db = 'tropicos', rows = 1, messages = FALSE)
    h <- classification("Asdfafsfd", db = 'nbn', rows = 1, messages = FALSE)
  })

  # expect_is(a, "classification")
  expect_is(b, "classification")
  expect_is(d, "classification")
  expect_is(g, "classification")
  expect_is(h, "classification")
})

test_that("warn on mismatch 'db'", {
  skip_on_cran()
  vcr::use_cassette("classification_warn_on_db_mismatch", {
    expect_warning(
      classification(
        get_uid("Chironomus riparius", messages = FALSE), db = "itis"))
  })
})

# see https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=668400
test_that("ncbi classification - taxon merged", {
  skip_on_cran()
  vcr::use_cassette("classification_fetches_merged_taxon", {
    a <- classification(668400, db = "ncbi")
  })

  expect_is(a, "classification")
  expect_named(a, "668400")
  # input id is different from id returned
  expect_false(identical("668400", a[[1]]$id[NROW(a[[1]])]))
})

context("classification.pow")
test_that("classification works for POW data", {
  skip_on_cran()
  vcr::use_cassette("classification_pow", {
    clas_pow <- classification(c("Quercus douglasii", "aaa vva"), db = "pow",
                                messages = FALSE)
    names(clas_pow) <- NULL
  })

  expect_that(clas_pow[[2]], equals(NA))
  expect_is(clas_pow, "classification")
  expect_is(clas_pow[[1]], "data.frame")
  expect_equal(length(clas_pow), 2)
  # rank is lowercase
  expect_false(all(grepl("[A-Z]", clas_pow[[1]]$rank)))

  vcr::use_cassette("classification_pow_get_fxn", {
    pows <- get_pow(c("Quercus douglasii", "aaa vva"), messages = FALSE)
    clas_pows <- classification(pows, messages = FALSE)
    names(clas_pows) <- NULL
  })

  expect_identical(clas_pows, clas_pow)
})

test_that("passing in an id works", {
  skip_on_cran()
  vcr::use_cassette("classification_pow_passing_id", {
    fromid_pow <- classification("urn:lsid:ipni.org:names:119003-2",
      db = "pow")
  })

  expect_is(fromid_pow, "classification")
  expect_equal(attr(fromid_pow, "db"), "pow")
})

test_that("rbind and cbind work correctly", {
  skip_on_cran()
  vcr::use_cassette("classification_pow_cbind_rbind", {
    out <- get_ids( 
      c("Brillantaisia vogeliana", "Aphelandra aurantiaca"),
      db = "pow", messages = FALSE, suppress = TRUE)
    cl <- classification(out)
  }, preserve_exact_body_bytes = TRUE)

  # rbind
  clr <- rbind(cl)
  expect_is(clr, "data.frame")
  expect_named(clr, c("name", "rank", "id", "query", "db"))

  # cbind
  clc <- cbind(cl)
  expect_is(clc, "data.frame")
  expect_gt(length(names(clc)), 5)
})

test_that("works on a variety of names", {
  skip_on_cran()
  vcr::use_cassette("classification_pow_more_name_egs", {
    x <- classification("Mimosa weberbaueri", db = "pow", messages = FALSE)
    z <- classification("Cyperus zollingeri", db = "pow", messages = FALSE)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(x, "classification")
  expect_is(z, "classification")
})

test_that("queries with no results fail well", {
  skip_on_cran()
  vcr::use_cassette("classification_pow_no_results", {
    aa <- classification("foobar", db = "pow", messages = FALSE)
    bb <- classification(get_pow("foobar", messages = FALSE), messages = FALSE)
  })

  expect_true(is.na(unclass(aa)[[1]]))
  expect_identical(unname(aa), unname(bb))
})

context("downstream")

test_that("downstream basic usage works", {
  skip_on_cran()
  vcr::use_cassette("downstream", {
    cc <- downstream("Ursus", db = "gbif", downto = "Species",
      messages = FALSE)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(cc, "downstream")
  expect_named(cc, "Ursus")
  expect_is(cc$Ursus$rank, "character")
})

test_that("downstream - taxonomic id input", {
  skip_on_cran()
  vcr::use_cassette("downstream_id_input", {
    aa <- downstream(get_gbifid("Ursus", messages = FALSE),
      db = "gbif", downto = "Species")
  }, preserve_exact_body_bytes = TRUE)

  expect_is(aa, "downstream")
  expect_is(aa[[1]], "data.frame")
  expect_is(aa[[1]]$name_type, "character")
})

test_that("downstream - multiple data sources", {
  skip_on_cran()
  vcr::use_cassette("downstream_multiple_data_sources", {
    ids <- sw(get_ids("Ursus", db = c("gbif", 'itis'), rows = 1,
      messages = FALSE, suppress = TRUE))
    aa <- downstream(ids, downto = "Species")
  }, preserve_exact_body_bytes = TRUE)

  expect_is(aa, "downstream_ids")
  expect_is(aa[[1]], "downstream")
  expect_is(aa[[1]]$Ursus, "data.frame")
})

test_that("downstream - Use the rows parameter", {
  skip_on_cran()
  vcr::use_cassette("downstream_rows_param", {
    aa <- downstream("Pinus", db = 'gbif', downto = "species",
      rows = 1, messages = FALSE)
  })

  expect_is(aa, "downstream")
  expect_is(aa[[1]], "data.frame")
  expect_is(aa[[1]]$name, "character")
})

test_that("downstream fails well", {
  skip_on_cran()

  expect_error(downstream("adfaf"), "Must specify downto")
  expect_error(downstream("Ursus", downto = "adfasdf"), "Must specify db")
  expect_error(downstream("Ursus", downto = "Species", db = "asdfdsf"),
               "the provided db value was not recognised")
})

test_that("warn on mismatch 'db'", {
  skip_on_cran()
  vcr::use_cassette("downstream_warn_on_db_mismatch", {
    expect_warning(
      downstream(
        get_gbifid('Apis', messages = FALSE, rows = 1), downto = "species",
        db = "itis"))
  })
})

test_that("downstream basic usage works", {
  skip_on_cran()
  vcr::use_cassette("downstream_bold", {
    cc <- downstream("Ursus", db = "bold", downto = "species",
      messages = FALSE)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(cc, "downstream")
  expect_named(cc, "Ursus")
  expect_is(cc$Ursus$rank, "character")
})

test_that("worms - can pass through marine_only parameter", {
  skip_on_cran()
  vcr::use_cassette("downstream_worms_marine_only_para", {
    # should return one row, Didelphimorphia
    aa <- downstream(1837, db = "worms", downto = "order")
    # should return 10 rows
    bb <- downstream(1837, db = "worms", downto = "order", marine_only=FALSE)
  })

  expect_is(aa, "downstream")
  expect_is(bb, "downstream")
  expect_gt(NROW(bb[[1]]), NROW(aa[[1]]))
})

# tests for get_ids fxn in taxize
context("get_ids")


test_that("get_ids returns the correct values and classses", {
  skip_on_cran()

  tt <- get_ids("Chironomus riparius", db = "ncbi",
    messages = FALSE, suppress = TRUE)

  expect_equal(tt[[1]][[1]], "315576")
  expect_is(tt, "ids")
  expect_is(tt[[1]], "uid")
  expect_is(tt[[1]][[1]], "character")
})

test_that("get_ids accepts ask and verbose arguments", {
  skip_on_cran()

  expect_message(get_ids("Pinus contorta", db = "ncbi",
    suppress = TRUE))
  expect_message(get_ids("Pinus contorta", db = "ncbi",
    messages = FALSE, suppress = TRUE), NA)
})

nn <- c('Imperata brasiliensis','Hylebates cordatus','Apocopis intermedius',
  'Paspalum subciliatum','Bromus nottowayanus','Chimonobambusa marmorea',
  'Panicum adenophorum','Otatea glauca','Himalayacalamus falconeri',
  'Briza lamarckiana','Trisetum turcicum','Brachiaria subulifolia',
  'Boissiera squarrosa','Arthrostylidium pubescens','Neyraudia reynaudiana'
  ,'Bromus gunckelii','Poa sudicola','Pentameris thuarii',
  'Calamagrostis inexpansa','Willkommia texana','Helictotrichon cantabricum',
  'Muhlenbergia tenuifolia','Sporobolus ioclados','Bambusa cerosissima',
  'Axonopus flabelliformis','Glyceria lithuanica','Pentaschistis malouinensis',
  'Perrierbambus madagascariensis','Hierochloe alpina','Hemarthria compressa',
  'Zizania latifolia','Festuca altaica','Gigantochloa wrayi','Festuca alpina',
  'Aegilops caudata','Elymus cognatus','Agrostis gracililaxa',
  'Gymnopogon foliosus')

test_that("works on a variety of names", {
  skip_on_cran()
  skip_on_ci()

  expect_is(sw(get_ids(nn[13], db = c("ncbi", "itis", "tropicos"),
    suppress = TRUE, ask = FALSE, messages = FALSE)), "ids")
  expect_is(sw(get_ids(nn[14], db = c("ncbi", "tropicos"),
    suppress = TRUE, ask = FALSE, messages = FALSE)), "ids")
  expect_is(sw(get_ids(nn[15], db = c("ncbi", "itis", "tropicos"),
    suppress = TRUE, ask = FALSE, messages = FALSE)), "ids")
})

ids <- c(
  '01', # domain
  '05','10','20','25', # kingdom
  '30','40','45', # phylum/division
  '50','60','70','80','81','82', # class
  '83','84','85','86','87', # cohort
  '90','100','110','120','125', # order
  '130','140','150', # family
  '155','160','170', # tribe
  '180','190','200','210','215','217', # genus
  '220','225','230', # species/etc.
  '240','250','255','260','265', # variety/sub/form/etc.
  '280', # genetic variants
  '300' # unspecified
)
ranks <- c(
  'domain', # domain
  'superkingdom','kingdom','subkingdom','infrakingdom,superphylum', # kingdom
  'phylum,division','subphylum,subdivision','infradivision', # phylum/division
  'superclass','class','subclass','infraclass','subterclass','parvclass', # class
  'megacohort','supercohort','cohort','subcohort','infracohort', # cohort
  'superorder','order','suborder','infraorder','parvorder', # order
  'superfamily','family','subfamily', # family
  'supertribe','tribe','subtribe', # tribe
  'genus','subgenus','section','subsection','species group,series','species subgroup', # genus
  'species','infraspecies','subspecies,forma specialis', # species/etc.
  'variety,varietas','subvariety,race','stirp','form,forma,morph','subform', # variety/sub/form/etc.
  'biotype,isolate,pathogroup,serogroup,serotype,strain,aberration', # genetic variants
  'unspecified,no rank,unranked,clade' # unspecified
)
rank_ref <- data.frame(
  rankid = ids,
  ranks = ranks
)
# NOTE: "version = 2" is so that we don't have to require R > 3.5
save(rank_ref, file = "data/rank_ref.RData", version = 2)
# NOTES: 
# - genetic variants: placed slightly higher above 'unspecified' just to denote
#   they're not no rank, but they're not really a taxonomic rank either


#### rank_ref_zoo
ids_zoo <- c(
  '01', # domain
  '05','10','20','25', # kingdom
  '30','40','45', # phylum/division
  '50','60','70','80','81','82', # class
  '83','84','85','86','87', # cohort
  '90','100','110','120','125', # order
  '126','127', # section/subsection
  '130','135','140','150', # family
  '155','160','170', # tribe
  '180','190','215','217', # genus
  '220','225','230', # species/etc.
  '240','250','255','260','265', # variety/sub/form/etc.
  '280', # genetic variants
  '300' # unspecified
)
ranks_zoo <- c(
  'domain', # domain
  'superkingdom','kingdom','subkingdom','infrakingdom,superphylum', # kingdom
  'phylum,division','subphylum,subdivision','infraphylum,infradivision', # phylum/division
  'superclass','class','subclass','infraclass','subterclass','parvclass', # class
  'megacohort','supercohort','cohort','subcohort','infracohort', # cohort
  'superorder','order','suborder','infraorder','parvorder', # order
  'section','subsection', # section/subsection
  'superfamily','epifamily','family','subfamily', # family
  'supertribe','tribe','subtribe', # tribe
  'genus','subgenus','species group,series','species subgroup', # genus
  'species','infraspecies','subspecies,forma specialis', # species/etc.
  'variety,varietas','subvariety,race','stirp','form,forma,morph','subform', # variety/sub/form/etc.
  'biotype,isolate,pathogroup,serogroup,serotype,strain,aberration', # genetic variants
  'unspecified,no rank,unranked,clade' # unspecified
)
rank_ref_zoo <- data.frame(
  rankid = ids_zoo,
  ranks = ranks_zoo
)
# NOTE: "version = 2" is so that we don't have to require R > 3.5
save(rank_ref_zoo, file = "data/rank_ref_zoo.RData", version = 2)
# NOTES: 
# - genetic variants: placed slightly higher above 'unspecified' just to denote
#   they're not no rank, but they're not really a taxonomic rank either

#' Retrieve all taxonomic names downstream in hierarchy for GBIF
#'
#' @export
#' @param id A taxonomic serial number.
#' @param downto The taxonomic level you want to go down to. See examples below.
#' 		The taxonomic level IS case sensitive, and you do have to spell it
#' 		correctly. See `data(rank_ref)` for spelling.
#' @param intermediate (logical) If TRUE, return a list of length two with
#' target taxon rank names, with additional list of data.frame's of
#' intermediate taxonomic groups. Default: FALSE
#' @param limit Number of records to return. default: 100. max: 1000. use in
#' combination with the `start` parameter
#' @param start Record number to start at. default: 0. use in combination 
#' with the `limit` parameter
#' @param key Deprecated, see `id`
#' @param ... Further args passed on to [gbif_name_usage()]
#' @return data.frame of taxonomic information downstream to family from e.g.,
#' 		Order, Class, etc., or if `intermediated=TRUE`, list of length two,
#'   	with target taxon rank names, and intermediate names.
#' @author Scott Chamberlain
#' @details Sometimes records don't have a `canonicalName` entry which is 
#' what we look for. In that case we grab the `scientificName` entry. 
#' You can see the type of name colleceted in the column `name_type`
#' @examples \dontrun{
#' ## the plant class Bangiophyceae
#' gbif_downstream(id = 198, downto="genus")
#' gbif_downstream(id = 198, downto="genus", intermediate=TRUE)
#'
#' # families downstream from the family Strepsiptera (twisted wing parasites)
#' gbif_downstream(id = 1227, "family")
#' ## here, intermediate leads to the same result as the target
#' gbif_downstream(id = 1227, "family", intermediate=TRUE)
#' 
#' if (interactive()) {
#' # Lepidoptera
#' gbif_downstream(id = 797, "family")
#'
#' # get species downstream from the genus Ursus
#' gbif_downstream(id = 2433406, "species")
#'
#' # get tribes down from the family Apidae
#' gbif_downstream(id = 7799978, downto="species")
#' gbif_downstream(id = 7799978, downto="species", intermediate=TRUE)
#' 
#' # names that don't have canonicalname entries for some results
#' # Myosotis: key 2925668
#' key <- 2925668
#' res <- gbif_downstream(key, downto = "species")
#' res2 <- downstream(key, db = "gbif", downto = "species")
#' 
#' # Pagination
#' # GBIF limits queries to a maximum of 1000 records per request, so if
#' # there's more than 1000, use the start parameter
#' # Piper, taxonKey = 3075433
#' x1 <- gbif_downstream(id = 3075433, downto = "species", limit=1000)
#' x2 <- gbif_downstream(id = 3075433, downto = "species", limit=1000,
#'   start=1000)
#' x3 <- gbif_downstream(id = 3075433, downto = "species", limit=1000,
#'   start=2000)
#' x4 <- gbif_downstream(id = 3075433, downto = "species", limit=1000,
#'   start=3000)
#' rbind(x1, x2, x3, x4)
#' }
#' }

gbif_downstream <- function(id, downto, intermediate = FALSE, limit = 100,
  start = NULL, key = NULL, ...) {

  if (!is.null(key)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "gbif_downstream(key)", with = "gbif_downstream(id)")
    id <- key
  }
  
  should_be('intermediate', intermediate, 'logical')

  downto <- tolower(downto)
  poss_ranks <- unique(do.call(c,
    sapply(taxize_ds$rank_ref$ranks, strsplit, split = ",",
      USE.NAMES = FALSE)))
  downto <- match.arg(downto, choices = poss_ranks)
  torank <- sapply(taxize_ds$rank_ref[which_rank(downto), "ranks"],
    function(x) strsplit(x, ",")[[1]][[1]], USE.NAMES = FALSE)

  stop_ <- "not"
  notout <- data.frame(rank = "", stringsAsFactors = FALSE)
  out <- list()
  if (intermediate) intermed <- list()
  iter <- 0
  while (stop_ == "not") {
    iter <- iter + 1
    temp <- dt2df(lapply(id, function(x) gbif_name_usage_clean(x)),
      idcol = FALSE)
    tt <- dt2df(lapply(temp$key, function(x) gbif_name_usage_children(x,
      limit = limit, start = start, ...)), idcol = FALSE)
    tt <- prune_too_low(tt, downto)

    if (NROW(tt) == 0) {
      out[[iter]] <- data.frame(stringsAsFactors = FALSE)
      stop_ <- "nodata"
    } else {
      if (intermediate) intermed[[iter]] <- tt
      if (NROW(tt[tt$rank == downto, ]) > 0)
        out[[iter]] <- tt[tt$rank == downto, ]
      if (NROW(tt[!tt$rank == downto, ]) > 0) {
        notout <- tt[!tt$rank %in% torank, ]
      } else {
        notout <- data.frame(rank = downto, stringsAsFactors = FALSE)
      }

      if (all(notout$rank == downto)) {
        stop_ <- "fam"
      } else {
        id <- notout$key
        stop_ <- "not"
      }
    }
    if (intermediate) intermed[[iter]] <- intermed[[iter]]
  } # end while loop

  tmp <- dt2df(out, idcol = FALSE)
  if (intermediate) {
    list(target = tmp, intermediate = intermed)
  } else {
    tmp
  }
}

gbif_name_usage_clean <- function(x, ...) {
  tt <- gbif_name_usage(x, ...)
  tt <- tt[sapply(tt, length) != 0]
  tt$rank <- tolower(tt$rank)
  tt <- setNames(tt, tolower(names(tt)))
  data.frame(tt, stringsAsFactors = FALSE)[, c('canonicalname', 'rank', 'key')]
}

gbif_name_usage_children <- function(x, limit = 100, start = NULL, ...) {
  tt <- gbif_name_usage(x, data = 'children', limit = limit, start = start, ...)$results
  dt2df(lapply(tt, function(z) {
    z <- z[sapply(z, length) != 0]
    df <- data.frame(z, stringsAsFactors = FALSE)
    df$rank <- tolower(df$rank)
    df <- stats::setNames(df, tolower(names(df)))
    nms <- c('rank', 'key')
    if ('canonicalname' %in% names(df)) {
      nms <- c('canonicalname', nms) 
      type <- "canonicalname"
    } else {
      nms <- c('scientificname', nms)
      type <- "scientificname"
    }
    dd <- df[, nms]
    dd <- stats::setNames(dd, c('name', 'rank', 'key'))
    dd$name_type <- type
    dd
  }), idcol = FALSE)
}

context("gbif_downstream")

test_that("gbif_downstream works", {
  skip_on_cran()
  vcr::use_cassette("gbif_downstream", {
    aa <- gbif_downstream(id = 198, downto = "Genus")
    bb <- gbif_downstream(id = 1227, "Family")
  }, preserve_exact_body_bytes = TRUE)

  expect_is(aa, "data.frame")
  expect_is(aa$name, "character")
  expect_is(aa$rank, "character")
  expect_is(aa$key, "integer")
  expect_equal(unique(aa$rank), "genus")

  expect_is(bb, "data.frame")
  expect_is(bb$name, "character")
  expect_is(bb$rank, "character")
  expect_is(bb$key, "integer")
  expect_equal(unique(bb$rank), "family")
})

test_that("gbif_downstream intermediate param works", {
  skip_on_cran()
  vcr::use_cassette("gbif_downstream_intermediate", {
    cc <- gbif_downstream(id = 198, downto = "genus", intermediate = TRUE)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(cc, "list")
  expect_is(cc$target, "data.frame")
  expect_is(cc$intermediate, "list")

  expect_is(cc$target$name, "character")
  expect_is(cc$target$rank, "character")
  expect_is(cc$target$key, "integer")

  expect_equal(unique(cc$target$rank), "genus")
})

test_that("gbif_downstream - limit and start params", {
  skip_on_cran()
  vcr::use_cassette("gbif_downstream_pagination", {
    aa <- gbif_downstream(2978223, "species", limit = 3)
    bb <- gbif_downstream(2978223, "species", limit = 3, start = 3)
  })

  expect_is(aa, "data.frame")
  expect_is(bb, "data.frame")
  expect_true(!all(aa$key %in% bb$key))
})

test_that("gbif_downstream fails well", {
  skip_on_cran()

  expect_error(gbif_downstream(198, "adfadf"), "'arg' should be one of")
  expect_error(gbif_downstream(198, "Genus", intermediate = "adf"),
               "'intermediate' should be of class 'logical'")
})

#' Taxonomic IDs to taxonomic names
#'
#' @export
#' @param id vector of taxonomic IDs (character or numeric)
#' @param db (character) database to query. One or more of `tol`, `itis`,
#' `ncbi`, `worms`, `gbif`, or `bold`. Note that each
#' taxonomic data source has their own  identifiers, so that if you provide
#' the wrong `db` value for the identifier you could get a result,
#' but it will likely be wrong (not what you were expecting). If using ncbi
#' we recommend getting API keys; see [taxize-authentication]
#' @param x Deprecated, see `id`
#' @param ... Further args passed on to `tol_id2name` or
#' [itis_getrecord], or other internal functions.
#' See those functions for what parameters can be passed on.
#'
#' @return A named list of data.frames, named by the input taxonomic ids
#' 
#' @section HTTP version for NCBI requests:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')` 
#'
#' @examples \dontrun{
#' # ITIS
#' id2name(19322, db = "itis")
#'
#' # TOL
#' id2name(515698, db = "tol")
#' # get NCBI ID and pass to classification()
#' x <- id2name(515698, db = "tol")
#' classification(as.uid(x[[1]]$tax_sources_ncbi))
#'
#' # NCBI
#' id2name(315567, db = "ncbi")
#' id2name(3339, db = "ncbi")
#' id2name(9696, db = "ncbi")
#' id2name(c(9695, 9696), db = "ncbi")
#'
#' # WORMS
#' id2name(105706, db = "worms")
#'
#' # GBIF
#' id2name(2441176, db = "gbif")
#'
#' # BOLD
#' id2name(88899, db = "bold")
#' }
id2name <- function(id, db = NULL, x = NULL, ...) UseMethod("id2name")

#' @export
#' @rdname id2name
id2name.default <- function(id, db = NULL, x = NULL, ...) {
  nstop(db)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "id2name(x)", with = "id2name(sci)")
    sci <- x
  }
  
  if (!is.null(x)) id <- x
  if (!db %in% id2name_sources) {
    stop("'db' must be one of ", paste(id2name_sources, collapse = ", "))
  }
  w <- process_idn_ids(id, db)
  stats::setNames(id2name(w, ...), id)
}

id2name_sources <- c('tol', 'itis', 'ncbi', 'worms', 'gbif', 'bold')

process_idn_ids <- function(input, db) {
  as_fxn <- switch(db, tol = as.tolid, itis = as.tsn, ncbi = as.uid,
    worms = as.wormsid, gbif = as.gbifid,
    bold = as.boldid)
  as_fxn(input, check = FALSE)
}

# TOL
#' @export
#' @rdname id2name
id2name.tolid <- function(id, ...) {
  warn_db(list(...), "tol")
  fun <- function(y) {
    if (is.na(y)) NA_character_ else tol_id2name(as.numeric(y))
  }
  out <- lapply(id, fun, ...)
  names(out) <- id
  class(out) <- 'id2name'
  attr(out, 'db') <- 'tol'
  return(out)
}


# ITIS
itis_id2name <- function(x, ...) {
  z <- itis_getrecord(x, ...)
  if (NROW(z) == 0) return(id2name_blanks$itis)
  data.frame(id = x, name = z$scientificName$combinedName,
    rank = strtrim(z$taxRank$rankName),
    status = z$coreMetadata$taxonUsageRating,
    parent_tsn = z$parentTSN$parentTsn,
    stringsAsFactors = FALSE)
}

#' @export
#' @rdname id2name
id2name.tsn <- function(id, ...) {
  warn_db(list(...), "itis")
  fun <- function(y) {
    if (is.na(y)) NA_character_ else itis_id2name(y, ...)
  }
  out <- lapply(id, fun, ...)
  names(out) <- id
  class(out) <- 'id2name'
  attr(out, 'db') <- 'tsn'
  return(out)
}


# NCBI
ncbi_id2name <- function(x, ...) {
  key <- getkey(NULL, "ENTREZ_KEY")
  cli <- crul::HttpClient$new(url = ncbi_base(), headers = tx_ual,
    opts = list(http_version = 2L, ...))
  args <- tc(list(db = "taxonomy", id = x, api_key = key))
  res <- cli$get("entrez/eutils/esummary.fcgi", query = args)
  res$raise_for_status()
  tt <- xml2::read_xml(res$parse("UTF-8"))
  toget <- c("ScientificName", "Rank", "Status")
  res <- vapply(toget, function(z) {
    xml2::xml_text(xml2::xml_find_first(tt, sprintf("//Item[@Name=\"%s\"]", z)))
  }, "")
  if (all(is.na(res))) return(id2name_blanks$ncbi)
  tmp <- rbind.data.frame(c(x, unname(res)), stringsAsFactors = FALSE)
  names(tmp) <- c("id", tolower(names(res)))
  names(tmp)[2] <- "name"
  return(tmp)
}

#' @export
#' @rdname id2name
id2name.uid <- function(id, ...) {
  warn_db(list(...), "ncbi")
  fun <- function(y, ...) {
    if (is.na(y)) NA_character_ else ncbi_id2name(y, ...)
  }
  out <- lapply(id, fun, ...)
  names(out) <- id
  class(out) <- 'id2name'
  attr(out, 'db') <- 'ncbi'
  return(out)
}


# WORMS
worms_id2name <- function(x, ...) {
  res <- worrms::wm_record(as.numeric(x), ...)
  if (length(res) == 0) return(id2name_blanks$ncbi)
  data.frame(id = x, name = res$scientificname,
    rank = res$rank, status = res$status,
    stringsAsFactors = FALSE)
}

#' @export
#' @rdname id2name
id2name.wormsid <- function(id, ...) {
  warn_db(list(...), "worms")
  fun <- function(y) {
    if (is.na(y)) NA_character_ else worms_id2name(y, ...)
  }
  out <- lapply(id, fun, ...)
  names(out) <- id
  class(out) <- 'id2name'
  attr(out, 'db') <- 'worms'
  return(out)
}


# GBIF
gbif_id2name <- function(x, ...) {
  z <- tryCatch(gbif_name_usage(x, ...), error = function(e) e)
  if (inherits(z, "error") || length(z) == 0) return(id2name_blanks$ncbi)
  data.frame(id = x, name = z$canonicalName,
    rank = tolower(z$rank), status = tolower(z$taxonomicStatus),
    stringsAsFactors = FALSE)
}

#' @export
#' @rdname id2name
id2name.gbifid <- function(id, ...) {
  warn_db(list(...), "gbif")
  fun <- function(y) {
    if (is.na(y)) NA_character_ else gbif_id2name(y, ...)
  }
  out <- lapply(id, fun, ...)
  names(out) <- id
  class(out) <- 'id2name'
  attr(out, 'db') <- 'gbif'
  return(out)
}

# BOLD
bold_id2name <- function(x, ...) {
  z <- bold_search(id = x, ...)
  if ("noresults" %in% names(z)) return(id2name_blanks$bold)
  data.frame(id = x, name = z$taxon,
    rank = tolower(z$tax_rank),
    stringsAsFactors = FALSE)
}

#' @export
#' @rdname id2name
id2name.boldid <- function(id, ...) {
  warn_db(list(...), "bold")
  fun <- function(y) {
    if (is.na(y)) NA_character_ else bold_id2name(y, ...)
  }
  out <- lapply(id, fun, ...)
  names(out) <- id
  class(out) <- 'id2name'
  attr(out, 'db') <- 'bold'
  return(out)
}


######## blank data.frames
id2name_blanks <- list(
  itis = data.frame(
    id = character(0),
    name = character(0),
    rank = character(0),
    status = character(0),
    parent_tsn = character(0),
    stringsAsFactors = FALSE
  ),
  ncbi = data.frame(
    id = character(0),
    name = character(0),
    rank = character(0),
    status = character(0),
    stringsAsFactors = FALSE
  ),
  tol = data.frame(
    id = character(0),
    name = character(0),
    rank = character(0),
    tax_sources_ncbi = character(0),
    tax_sources_gbif = character(0),
    tax_sources_irmng = character(0),
    stringsAsFactors = FALSE
  ),
  bold = data.frame(
    id = character(0),
    name = character(0),
    rank = character(0),
    stringsAsFactors = FALSE
  )
)

context("id2name")

test_that("id2name: itis", {
  skip_on_cran()

  aa <- id2name(19322, db = "itis")

  expect_is(aa, "id2name")
  expect_equal(attr(aa, "db"), "tsn")
  expect_is(unclass(aa), "list")
  expect_is(aa[[1]], "data.frame")
  expect_named(aa[[1]], c('id', 'name', 'rank', 'status', 'parent_tsn'))
  expect_is(aa[[1]]$name, "character")
})

test_that("id2name: tol", {
  skip_on_cran()

  aa <- id2name(515698, db = "tol")

  expect_is(aa, "id2name")
  expect_equal(attr(aa, "db"), "tol")
  expect_is(unclass(aa), "list")
  expect_is(aa[[1]], "data.frame")
  expect_is(aa[[1]]$name, "character")
})

test_that("missing/wrong data given returns result", {
  skip_on_cran()

  expect_error(id2name(), "Must specify db")
  expect_error(id2name(db = "tol"), "argument \"id\" is missing")
  expect_error(id2name(5, db = "stuff"), "'db' must be one of")
  # requires HTTP request
  expect_warning(id2name(55555555, db = "tol"), "Bad Request")
  expect_equal(NROW(suppressWarnings(id2name(55555555, db = "tol"))[[1]]), 0)
})

test_that("warn on mismatch 'db'", {
  skip_on_cran()
  vcr::use_cassette("id2name_warn_on_db_mismatch", {
    expect_warning(
      id2name(
        get_uid('Apis', messages = FALSE), db = "gbif"))
  })
})

#' Get references related to a ITIS TSN.
#'
#' @export
#' @param tsn One or more TSN's (taxonomic serial number) for a
#' taxonomic group (numeric)
#' @param ... Further arguments passed on to getpublicationsfromtsn
#' @examples \dontrun{
#' itis_refs(202385)
#' itis_refs(c(202385, 70340))
#' }
itis_refs <- function(tsn, ...) {
  lapply(tsn, ritis::publications, ...)
}

context("itis_refs")

test_that("itis_refs returns the correct class", {
  skip_on_cran()
  vcr::use_cassette("itis_refs", {
    one <- itis_refs(202385, verbose=FALSE)
    two <- itis_refs(c(202385,70340), verbose=FALSE)
  })

  expect_that(one, is_a("list"))
  expect_that(one[[1]], is_a("data.frame"))
  expect_that(two, is_a("list"))
  expect_that(two[[1]], is_a("data.frame"))
})

#' Get the EOL ID from Encyclopedia of Life from taxonomic names.
#'
#' Note that EOL doesn't expose an API endpoint for directly querying for EOL
#' taxon ID's, so we first use the function [eol_search()]
#' to find pages that deal with the species of interest, then use
#' [eol_pages()] to find the actual taxon IDs.
#'
#' @export
#' @param sci_com character; one or more scientific or common names. Or,
#' a `taxon_state` object (see [taxon-state])
#' @param ask logical; should get_eolid be run in interactive mode?
#' If TRUE and more than one ID is found for the species, the user is asked for
#' input. If FALSE NA is returned for multiple matches.
#' @param ... Further args passed on to [eol_search()]
#' @param messages logical; If `TRUE` the actual taxon queried is printed
#' on the console.
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this function still only gives back a eolid
#' class object with one to many identifiers. See
#' [get_eolid_()] to get back all, or a subset, of the raw
#' data that you are presented during the ask process.
#' @param rank (character) A taxonomic rank name. See [rank_ref]
#' for possible options. Though note that some data sources use atypical ranks,
#' so inspect the data itself for options. Optional. See `Filtering` below.
#' @param data_source (character) A data source inside of EOL. These are
#' longish names like e.g., "Barcode of Life Data Systems" or
#' "USDA PLANTS images". Optional. See `Filtering` below.
#' @param x Input to [as.eolid()]
#' @param check logical; Check if ID matches any existing on the DB, only
#' used in [as.eolid()]
#' @param sciname Deprecated, see `sci_com`
#' @template getreturn
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @author Scott Chamberlain
#'
#' @details EOL is a bit odd in that they have page IDs for each taxon, but
#' then within that, they have taxon ids for various taxa within that page
#' (e.g., GBIF and NCBI each have a taxon they refer to within the
#' page \[i.e., taxon\]). And we need the taxon ids from a particular data
#' provider (e.g, NCBI) to do other things, like get a higher classification
#' tree. However, humans want the page id, not the taxon id. So, the
#' id returned from this function is the taxon id, not the page id. You can
#' get the page id for a taxon by using [eol_search()] and
#' `[eol_pages()], and the URI returned in the attributes for a
#' taxon will lead you to the taxon page, and the ID in the URL is the
#' page id.
#'
#' @section Filtering:
#' The parameters `rank` and `data_source` are not used in the
#' search to the data provider, but are used in filtering the data down to a
#' subset that is closer to the target you want.  For all these parameters,
#' you can use regex strings since we use [grep()] internally to
#' match. Filtering narrows down to the set that matches your query, and
#' removes the rest.
#'
#' @examples \dontrun{
#' get_eolid(sci_com='Pinus contorta')
#' get_eolid(sci_com='Puma concolor')
#'
#' get_eolid(c("Puma concolor", "Pinus contorta"))
#'
#' # specify rows to limit choices available
#' get_eolid('Poa annua')
#' get_eolid('Poa annua', rows=1)
#' get_eolid('Poa annua', rows=2)
#' get_eolid('Poa annua', rows=1:2)
#'
#' # When not found
#' get_eolid(sci_com="uaudnadndj")
#' get_eolid(c("Chironomus riparius", "uaudnadndj"))
#'
#' # filter results to a rank or data source, or both
#' get_eolid("Satyrium")
#' get_eolid("Satyrium", rank = "genus")
#' get_eolid("Satyrium", data_source = "INAT")
#' get_eolid("Satyrium", rank = "genus",
#'   data_source = "North Pacific Species List")
#'
#' # Convert a eolid without class information to a eolid class
#' # already a eolid, returns the same
#' as.eolid(get_eolid("Chironomus riparius"))
#' # same
#' as.eolid(get_eolid(c("Chironomus riparius","Pinus contorta")))
#' # numeric
#' as.eolid(10247706)
#' # numeric vector, length > 1
#' as.eolid(c(6985636,12188704,10247706))
#' # character
#' as.eolid("6985636")
#' # character vector, length > 1
#' as.eolid(c("6985636","12188704","10247706"))
#' # list, either numeric or character
#' as.eolid(list("6985636","12188704","10247706"))
#' ## dont check, much faster
#' as.eolid("6985636", check=FALSE)
#' as.eolid(6985636, check=FALSE)
#' as.eolid(c("6985636","12188704","10247706"), check=FALSE)
#' as.eolid(list("6985636","12188704","10247706"), check=FALSE)
#'
#' (out <- as.eolid(c(6985636,12188704,10247706)))
#' data.frame(out)
#' as.eolid( data.frame(out) )
#'
#' # Get all data back
#' get_eolid_("Poa annua")
#' get_eolid_("Poa annua", rows=2)
#' get_eolid_("Poa annua", rows=1:2)
#' get_eolid_(c("asdfadfasd", "Pinus contorta"))
#' }

get_eolid <- function(sci_com, ask = TRUE, messages = TRUE,
  rows = NA, rank = NULL, data_source = NULL, sciname = NULL, ...) {

  assert(sci_com, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(messages, "logical")
  assert(rank, "character")
  assert(data_source, "character")
  assert_rows(rows)
  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_eolid(sciname)", with = "get_eolid(sci_com)")
    sci_com <- sciname
  }
  

  if (inherits(sci_com, "character")) {
    tstate <- taxon_state$new(class = "eolid", names = sci_com)
    items <- sci_com
  } else {
    assert_state(sci_com, "eolid")
    tstate <- sci_com
    sci_com <- tstate$taxa_remaining()
    items <- c(sci_com, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci_com)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci_com[i], "'\n")
    tmp <- eol_search(sci = sci_com[i], ...)
    datasource <- NA_character_
    if (all(is.na(tmp))) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      page_id <- NA_character_
      att <- "not found"
      mm <- FALSE
    } else {
      pageids <- tmp[grep(tolower(sci_com[i]), tolower(tmp$name)), "pageid"]

      if (length(pageids) == 0) {
        if (nrow(tmp) > 0)
        mssg(messages, paste(m_not_found_sp_altclass, sprintf("\nDid find: %s",
                                        paste(tmp$name, collapse = "; "))))
        id <- NA_character_
      } else {
        dfs <- lapply(pageids, function(x) {
          y <- tryCatch(eol_pages(x), error = function(e) e)
          if (is(y, "error")) NULL else y$scinames
        })
        names(dfs) <- pageids
        dfs <- tc(dfs)
        if (length(dfs) > 1) dfs <- dfs[!sapply(dfs, nrow) == 0]
        df <- dt2df(dfs)
        df <- rename(df, c(".id" = "pageid", "identifier" = "eolid",
          "scientificname" = "name", "nameaccordingto" = "source",
          "taxonrank" = "rank"))

        # drop columns
        df$canonicalform <- df$sourceidentifier <- NULL
        mm <- NROW(df) > 1
        df <- sub_rows(df, rows)

        if (nrow(df) == 0) {
          mssg(messages, m_not_found_sp_altclass)
          id <- NA_character_
          page_id <- NA_character_
        } else{
          id <- df$eolid
        }
        names(id) <- df$pageid
      }
    }

    # not found on eol
    if (length(id) == 0 || all(is.na(id))) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      page_id <- NA_character_
			mm <- FALSE
      att <- "not found"
    }
    # only one found on eol
    if (length(id) == 1 & !all(is.na(id))) {
      id <- df$eolid
      page_id <- df$pageid
      datasource <- df$source
      direct <- TRUE
      att <- "found"
    }

    # filter by rank, if given
    if (!is.null(rank) || !is.null(data_source)) {
      df <- filt(df, "rank", rank)
      df <- filt(df, "source", data_source)
      id <- df$eolid
      page_id <- df$pageid
      datasource <- df$source
      att <- "found"
    }

    if (length(id) == 0 || all(is.na(id))) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      page_id <- NA_character_
      datasource <- NA_character_
      mm <- FALSE
      att <- "not found"
    }

    if (length(id) > 1) {
      matchtmp <- df[tolower(df$name) %in% tolower(sci_com[i]), ]
      if (NROW(matchtmp) == 1) {
        id <- matchtmp$eolid
        direct <- TRUE
        page_id <- matchtmp$pageid
        datasource <- matchtmp$source
        att <- "found"
      }
    }

    # more than one found on eol -> user input
    if (length(id) > 1) {
      if (ask) {
        rownames(df) <- 1:nrow(df)
        # prompt
        message("\n\n")
        message("\nMore than one eolid found for taxon '", sci_com[i], "'!\n
            Enter rownumber of taxon (other inputs will return 'NA'):\n")
        message(paste0(utils::capture.output(df), collapse = "\n"))
        take <- scan(n = 1, quiet = TRUE, what = "raw")

        if (length(take) == 0) {
          take <- "notake"
        }
        if (take %in% seq_len(nrow(df))) {
          take <- as.numeric(take)
          message("Input accepted, took eolid '",
                  as.character(df$eolid[take]), "'.\n")
          id <- as.character(df$eolid[take])
          names(id) <- as.character(df$pageid[take])
          page_id <- as.character(df$pageid[take])
          datasource <- as.character(df$source[take])
          att <- "found"
        } else {
          id <- NA_character_
          page_id <- NA_character_
          att <- "not found"
          mssg(messages, "\nReturned 'NA'!\n\n")
        }
      } else {
        if (length(id) != 1) {
          warning(sprintf(m_more_than_one_found, "eolid", sci_com[i]),
            call. = FALSE)
          id <- NA_character_
          page_id <- NA_character_
          att <- m_na_ask_false
        }
      }
    }
    res <- list(id = as.character(id), page_id = page_id, source = datasource, 
      att = att, multiple = mm, direct = direct)
    prog$completed(sci_com[i], att)
    prog$prog(att)
    tstate$add(sci_com[i], res)
  }
  out <- tstate$get()
  page_ids <- pluck_un(out, "page_id", "")
  ids <- structure(
    pluck_un(out, "id", ""), class = "eolid",
    pageid = page_ids,
    provider = pluck_un(out, "source", ""),
    match = pluck_un(out, "att", ""),
    multiple_matches = pluck_un(out, "multiple", logical(1)),
    pattern_match = pluck_un(out, "direct", logical(1))
  )
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$eol, page_ids)
}

taxize_sort_df <- function(data, vars = names(data)) {
  if (length(vars) == 0 || is.null(vars))
    return(data)
  data[do.call("order", data[, vars, drop = FALSE]), , drop = FALSE]
}

#' @export
#' @rdname get_eolid
as.eolid <- function(x, check=TRUE) {
  UseMethod("as.eolid")
}

#' @export
#' @rdname get_eolid
as.eolid.eolid <- function(x, check=TRUE) {
  x
}

#' @export
#' @rdname get_eolid
as.eolid.character <- function(x, check=TRUE) {
  if (length(x) == 1) {
    make_eolid(x, check)
  } else {
    collapse(x, fxn = make_eolid, class = "eolid", check = check)
  }
}

#' @export
#' @rdname get_eolid
as.eolid.list <- function(x, check=TRUE) {
  if (length(x) == 1) {
    make_eolid(x, check)
  } else {
    collapse(x, make_eolid, "eolid", check = check)
  }
}

#' @export
#' @rdname get_eolid
as.eolid.numeric <- function(x, check=TRUE) {
  as.eolid(as.character(x), check)
}

#' @export
#' @rdname get_eolid
as.eolid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class = "eolid", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri = x$uri)
}

#' @export
#' @rdname get_eolid
as.data.frame.eolid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "eolid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_eolid <- function(x, check=TRUE) {
  tmp <- make_generic(x, 'https://eol.org/pages/%s/overview', "eolid", check)
  if (!check) {
    attr(tmp, "uri") <- NA_character_
  } else {
    z <- get_eol_pageid(x)
    if (!is.na(z)) {
      attr(tmp, "uri") <- sprintf('https://eol.org/pages/%s/overview', z)
    } else {
      attr(tmp, "uri") <- NA_character_
    }
  }
  tmp
}

check_eolid <- function(x) {
  url <- sprintf("https://eol.org/api/hierarchy_entries/1.0/%s.json", x)
  tryid <- tax_GET_nocheck(url)
  if (tryid$status_code == 200) TRUE else FALSE
}

get_eol_pageid <- function(x) {
  url <- sprintf("https://eol.org/api/hierarchy_entries/1.0/%s.json", x)
  tt <- tax_GET_nocheck(url)
  if (tt$status_code == 200) {
    jsonlite::fromJSON(tt$parse("UTF-8"), FALSE)$taxonConceptID
  } else {
    NA
  }
}

#' @export
#' @rdname get_eolid
get_eolid_ <- function(sci_com, messages = TRUE, rows = NA, sciname = NULL,
  ...) {
  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_eolid_(sciname)", with = "get_eolid_(sci_com)")
    sci_com <- sciname
  }
  
  stats::setNames(lapply(sci_com, get_eolid_help, messages = messages,
                  rows = rows, ...), sci_com)
}

get_eolid_help <- function(sci_com, messages, rows, ...) {
  mssg(messages, "\nRetrieving data for taxon '", sci_com, "'\n")
  tmp <- eol_search(sci_com, ...)

  if (all(is.na(tmp))) {
    NULL
  } else {
    pageids <- tmp[grep(tolower(sci_com), tolower(tmp$name)), "pageid"]
    if (length(pageids) == 0) {
      NULL
    } else {
      dfs <- lapply(pageids, function(x) {
        y <- tryCatch(eol_pages(x), error = function(e) e)
        if (inherits(y, "error")) NULL else y$scinames
      })
      names(dfs) <- pageids
      dfs <- tc(dfs)
      if (length(dfs) > 1) dfs <- dfs[!sapply(dfs, nrow) == 0]
      dfs <- dt2df(dfs)
      df <- dfs[,c('.id','identifier','scientificname','nameaccordingto')]
      names(df) <- c('pageid','eolid','name','source')
      if (NROW(df) == 0) NULL else sub_rows(df, rows)
    }
  }
}

context("get_eolid")

test_that("get_eolid works", {
    skip_on_cran()

  vcr::use_cassette("get_eolid", {
    a <- get_eolid(sci_com = "Puma concolor", messages = FALSE, rows = 1)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(a, "eolid")
  expect_is(a[[1]], "character")
  expect_is(as.numeric(a[[1]]), "numeric")
  expect_is(attr(a, "pageid"), "character")
  expect_is(attr(a, "provider"), "character")
  expect_is(attr(a, "match"), "character")
  expect_true(attr(a, "multiple_matches"))
  expect_true(attr(a, "pattern_match"))
  expect_is(attr(a, "uri"), "character")
})

test_that("get_eolid accepts ask-argument", {
  skip_on_cran()

  vcr::use_cassette("get_eolid_ask_param", {
    expect_true(is.na(get_eolid(sci_com = 'adsf asdf asdf', ask = FALSE,
      messages = FALSE)[[1]]))
  })
})

test_that("get_eolid fails as expected", {
  skip_on_cran()

  expect_error(get_eolid(), "argument \"sci_com\" is missing")
  expect_error(get_eolid("Poa annua", ask = 4, messages = FALSE),
               "ask must be of class logical")

  Sys.sleep(1)
  # rows param
  expect_error(get_eolid("Poa annua", rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_eolid("Poa annua", rows = 0, messages = FALSE),
               "all\\(rows > 0\\) is not TRUE")
})

#' Use Catalogue of Life to get downstream taxa to a given taxonomic level.
#'
#' @param name The string to search for. Only exact matches found the name given
#'     will be returned, unless one or wildcards are included in the search
#'   	string. An * (asterisk) character denotes a wildcard; a % (percentage)
#'    character may also be used. The name must be at least 3 characters long,
#'    not counting wildcard characters.
#' @param id The record ID of the specific record to return (only for scientific
#'   	names of species or infraspecific taxa)
#' @param downto The taxonomic level you want to go down to. See examples below.
#' 		The taxonomic level IS case sensitive, and you do have to spell it
#' 		correctly. See \code{data(rank_ref)} for spelling.
#' @param format The returned format (default = NULL). If NULL xml is used.
#'    Currently only xml is supported.
#' @param start  The first record to return (default = NULL). If NULL, the
#'    results are returned from the first record (start=0). This is useful if
#'    the total number of results is larger than the maximum number of results
#'    returned by a single Web service query (currently the maximum number of
#'    results returned by a single query is 500 for terse queries and 50 for
#'    full queries).
#' @param checklist The year of the checklist to query, if you want a specific
#' 		year's checklist instead of the lastest as default (numeric).
#' @param verbose Print or suppress messages.
#' @param intermediate (logical) If TRUE, return a list of length two with target
#'    taxon rank names, with additional list of data.frame's of intermediate
#'    taxonomic groups. Default: FALSE
#' @param ... Curl options passed on to \code{\link[httr]{GET}}
#' @details Provide only names instead of id's
#' @return A list of data.frame's.
#' @export
#' @examples \dontrun{
#' # Some basic examples
#' col_downstream(name="Apis", downto="species")
#' col_downstream(name="Bryophyta", downto="family")
#'
#' # get classes down from the kingdom Animalia
#' col_downstream(name="Animalia", downto="class")
#' col_downstream(name="Animalia", downto="class", intermediate=TRUE)
#'
#' # An example that takes a bit longer
#' col_downstream(name=c("Plantae", "Animalia"), downto="class")
#'
#' # Using a checklist from a specific year
#' col_downstream(name="Bryophyta", downto="family", checklist=2009)
#'
#' # By id
#' col_downstream(id='576d098d770a39d09e2bcfa1c0896b26', downto="species",
#'   checklist=2012)
#' }

col_downstream <- function(name = NULL, id = NULL, downto, format = NULL,
  start = NULL, checklist = NULL, verbose = TRUE, intermediate = FALSE, ...) {

  downto <- tolower(downto)
  poss_ranks <- unique(do.call(c, sapply(rank_ref$ranks, strsplit, split = ",",
                                         USE.NAMES = FALSE)))
  downto <- match.arg(downto, choices = poss_ranks)

  func <- function(x=NULL, y=NULL, checklist, format, start, ...) {
    url <- make_url(checklist)
    torank <- sapply(rank_ref[which_rank(downto), "ranks"],
                function(x) strsplit(x, ",")[[1]][[1]], USE.NAMES = FALSE)

    toget <- ifelse(is.null(y), x, y)
    stop_ <- "not"
    notout <- data.frame(rankName = "")
    out <- list()
    if (intermediate) intermed <- list()
    iter <- 0
    while (stop_ == "not") {
      iter <- iter + 1
      if (is.null(x)) {
        tt <- ldply(toget, function(z) {
          search_col_safe(name = NULL, id = z, checklist = checklist,
                          format = format, start = start, ...)
        })
      } else {
        tt <- ldply(toget, function(z) {
          search_col_safe(name = z, id = NULL, checklist = checklist,
                          format = format, start = start, ...)
        })
      }
      # prune if too low
      tt <- prune_too_low_col(tt, downto)

      if (NROW(tt) == 0) {
        out[[iter]] <- data.frame(stringsAsFactors = FALSE)
        stop_ <- "nodata"
      } else {
        # remove
        if (NROW(tt[tt$childtaxa_rank == downto, ]) > 0) {
          out[[iter]] <- tt[tt$childtaxa_rank == downto, ]
        }

        if (NROW(tt[!tt$childtaxa_rank == downto, ]) > 0) {
          notout <- tt[!tt$childtaxa_rank %in% torank, ]
        } else {
          notout <- data.frame(rankName = downto)
        }

        if (all(notout$childtaxa_rank == downto)) {
          stop_ <- "fam"
        } else {
          if (intermediate) intermed[[iter]] <- notout
          x <- NULL
          toget <- as.character(notout$childtaxa_id)
          stop_ <- "not"
        }
      }
    } # end while loop

    if (length(out) == 0) {
      ret <-  data.frame(childtaxa_id = NA, childtaxa_name = NA,
                         childtaxa_rank = NA)
    } else {
      res <- tc(out)
      ret <- do.call(rbind.fill, res)
    }
    if (intermediate) list(target = ret, intermediate = intermed) else ret
  } # end fxn func

  safe_func <- plyr::failwith(NULL, func)
  if (is.null(id)) {
    temp <- setNames(lapply(name, safe_func, y = NULL,
                            checklist = checklist, format = format,
                            start = start, ...), name)
  } else {
    temp <- setNames(lapply(id, function(z) {
      safe_func(x = NULL, y = id, checklist = checklist, format = format,
                start = start, ...)
    }), id)
  }

  nas <- sapply(temp, function(z)
    NROW(na.omit( if (intermediate) z$target else z )))
  if (verbose)
    message(
      sprintf('These taxa with no data: %s\nTry adjusting input parameters',
              names(nas[nas == 0])))
  return( temp )
}

# tests for col_downstream fxn in taxize
context("col_downstream")

test_that("col_downstream returns the correct class", {
  skip_on_cran()

  temp4 <- col_downstream(name="Animalia", downto = "Phylum", verbose = FALSE)
  temp5 <- col_downstream(name="Plantae", downto = "Phylum", verbose = FALSE)
  temp6 <- col_downstream(name="Salicaceae", downto = "Genus", verbose = FALSE)

  expect_is(temp4, "list")
  expect_is(temp5, "list")
  expect_is(temp6, "list")
  expect_is(temp4$Animalia, "data.frame")
  expect_is(temp5[[1]], "data.frame")
  expect_equal(as.character(temp4[[1]][,3][[1]]), "phylum")
})

test_that("gives what's expected on input errors", {
  library("plyr")
  expect_message(col_downstream(name="Pinus contorta", downto = "Species")[[1]], "Try adjusting")
})

#' Get metadata about GNA data sources
#'
#' Downloads metadata about Global Names Architecture (GNA) data sources
#' available to be used in other GNA functions.
#'
#' @param output_type What format of output to return. Either `'json'`,
#'   `'list'`, or `'table'`.
#' @param ... Passed to [crul::HttpClient].
#'
#' @author Zachary S.L. Foster
#'
#' @examples \dontrun{
#' 
#' gna_data_sources()
#' }
#' 
#' @export
gna_data_sources <- function(output_type = 'table', ...) {
  gna_data_sources_url <- 'https://verifier.globalnames.org/api/v1/data_sources'
  
  # Check arguments
  possible_output_types <- c('json', 'table', 'list')
  if (! output_type %in% possible_output_types) {
    stop(call. = FALSE, 'The `output_type` value must be one of: ', paste0(possible_output_types, collapse = ', '))
  }
  
  # Make and parse API call
  api <- crul::HttpClient$new(gna_data_sources_url, headers = tx_ual, opts = list(...))
  response <- api$get()
  response$raise_for_status()
  response_json <- response$parse("UTF-8")
  if (output_type == 'json') {
    return(response_json)
  }
  response_data <- jsonlite::fromJSON(response_json, FALSE)
  
  # Reformat response data to a list
  if (output_type == 'list') {
    return(response_data)
  }
  
  # Reformat response data to a table
  if (output_type == 'table') {
    used_cols <- c(
      'id',
      'titleShort',
      'title',
      'description',
      'curation',
      'hasTaxonData',
      'recordCount',
      'updatedAt',
      'homeURL',
      'uuid',
      'version',
      'isOutlinkReady',
      'doi'
    )
    response_table <- do.call(rbind, lapply(response_data, function(x) {
      out <- x[used_cols]
      names(out) <- used_cols
      out[unlist(lapply(out, is.null))]  <- NA
      as.data.frame(out)
    }))
    response_table <- tibble::as_tibble(response_table)
  }
  return(response_table)
}

context("gna_data_sources")

test_that("gna_data_sources returns a table", {
  skip_on_cran()
  vcr::use_cassette("gna_data_sources", {
    tmp <- gna_data_sources()
  })
  expect_is(tmp, "data.frame")
  expect_equal(NCOL(tmp), 13)
})

#' Search for pages in EOL database using a taxonconceptID.
#'
#' @export
#' @param taxonconceptID (numeric) a taxonconceptID, which is also the page
#' number
#' @param images_per_page (integer) number of returned image objects (0-75)
#' @param images_page (integer) images page
#' @param videos_per_page (integer) number of returned video objects (0-75)
#' @param videos_page (integer) videos page
#' @param sounds_per_page (integer) number of returned sound objects (0-75)
#' @param sounds_page (integer) sounds page
#' @param maps_per_page (integer) number of returned map objects (0-75)
#' @param maps_page (integer) maps page
#' @param texts_per_page (integer) number of returned text objects (0-75)
#' @param texts_page (integer) texts page
#' @param subjects 'overview' (default) to return the overview text
#' (if exists), a pipe | delimited list of subject names from the list of EOL
#' accepted subjects (e.g. TaxonBiology, FossilHistory), or 'all' to get text
#' in any subject. Always returns an overview text as a first result (if one
#' exists in the given context).
#' @param licenses A pipe | delimited list of licenses or 'all' (default) to
#' get objects under any license. Licenses abbreviated cc- are all Creative
#' Commons licenses. Visit their site for more information on the various
#' licenses they offer.
#' @param details Include all metadata for data objects. (Default: `FALSE`)
#' @param common_names Return all common names for the page's taxon
#' (Default: `FALSE`)
#' @param synonyms Return all synonyms for the page's taxon
#' (Default: `FALSE`)
#' @param references Return all references for the page's taxon
#' (Default: `FALSE`)
#' @param taxonomy (logical) Whether to return any taxonomy details from
#' different taxon hierarchy providers, in an array named `taxonconcepts`
#' (Default: `TRUE`)
#' @param vetted If 'vetted' is given a value of '1', then only trusted
#' content will be returned. If 'vetted' is '2', then only trusted and
#' unreviewed content will be returned (untrusted content will not be returned).
#' The default is to return all content. (Default: `FALSE`)
#' @param cache_ttl The number of seconds you wish to have the response cached.
#' @param ... Curl options passed on to [crul::HttpClient]
#' @details It's possible to return JSON or XML with the EOL API. However,
#' this function only returns JSON for now.
#' @return JSON list object, or data.frame.
#' @examples \dontrun{
#' (pageid <- eol_search('Pomatomus')$pageid[1])
#' x <- eol_pages(taxonconceptID = pageid)
#' x
#' x$scinames
#'
#' z <- eol_pages(taxonconceptID = pageid, synonyms = TRUE)
#' z$synonyms
#'
#' z <- eol_pages(taxonconceptID = pageid, common_names = TRUE)
#' z$vernacular
#' }

eol_pages <- function(taxonconceptID, images_per_page=NULL, images_page=NULL,
  videos_per_page=NULL, videos_page=NULL, sounds_per_page=NULL, sounds_page=NULL,
  maps_per_page=NULL, maps_page=NULL, texts_per_page=NULL, texts_page=NULL,
  subjects='overview', licenses='all', details=FALSE,
  common_names=FALSE, synonyms=FALSE, references=FALSE, taxonomy=TRUE, vetted=0,
  cache_ttl=NULL, ...) {

  details <- tolower(details)
  common_names <- tolower(common_names)
  synonyms <- tolower(synonyms)
  references <- tolower(references)

	args <- tc(list(images_per_page=images_per_page, images_page=images_page,
    videos_per_page=videos_per_page, videos_page=videos_page,
    sounds_per_page=sounds_per_page, sounds_page=sounds_page,
    maps_per_page=maps_per_page, maps_page=maps_page,
    texts_per_page=texts_per_page, texts_page=texts_page,
    subjects=subjects, licenses=licenses, details=details,
    common_names=common_names, synonyms=synonyms, references=references,
    taxonomy=taxonomy, vetted=vetted, cache_ttl=cache_ttl))
  cli <- crul::HttpClient$new(
    url = file.path(eol_url("pages"), paste0(taxonconceptID, ".json")),
    headers = tx_ual,
    opts = list(...)
  )
  res <- cli$get(query = argsnull(args))
  res$raise_for_status()
  tt <- res$parse("UTF-8")
  stopifnot(res$response_headers$`content-type` ==
    "application/json; charset=utf-8")
  res <- jsonlite::fromJSON(tt, FALSE)

  # scinames <- lapply(res$taxonConcept$taxonConcepts, function(z) {
  scinames <- lapply(res$taxonConcept$taxonConcepts, function(z) {
    z[vapply(z, class, "") == "NULL"] <- NA_character_
    z
  })
  scinames <- dt2tibble(scinames)
  if (!is.null(scinames)) {
    names(scinames) <- tolower(names(scinames))
    if ("taxonrank" %in% names(scinames)) {
      scinames$taxonrank <- tolower(scinames$taxonrank)
    }
    scinames$name <- NULL
  }
  syns <- parseeoldata('synonyms', res)
  vernac <- parseeoldata('vernacularNames', res)
  refs <- parseeoldata('references', res)
  dataobj <- parseeoldata('dataObjects', res)
  list(scinames=scinames, synonyms=syns, vernacular=vernac, refs=refs,
    data_objects=dataobj)
}

parseeoldata <- function(x, y) {
  xx <- y$taxonConcept[[x]]
  if (length(xx) == 0) {
    NULL
  } else {
    tmp <- lapply(xx, data.frame)
    if (length(tmp) == 1) {
      tmp2 <- tmp[[1]]
    } else {
      tmp2 <- dt2df(tmp, idcol = FALSE)
    }
    names(tmp2) <- tolower(names(tmp2))
    tmp2
  }
}

context("eol_pages")

test_that("eol_pages returns the correct value and classes", {
  skip_on_cran()

  vcr::use_cassette("eol_pages", {
    pageid <- suppressMessages(eol_search('Pomatomus'))$pageid[1]
    pageid2 <- suppressMessages(eol_search('Helianthus'))$pageid[1]
    aa <- suppressMessages(eol_pages(taxonconceptID = pageid))
  }, preserve_exact_body_bytes = TRUE)

  expect_is(aa, "list")
  expect_is(aa$scinames, "data.frame")
  expect_null(aa$synonyms)
  expect_named(aa, c('scinames', 'synonyms', 'vernacular', 'refs', 'data_objects'))

  vcr::use_cassette("eol_pages2", {
  	bb <- suppressMessages(eol_pages(taxonconceptID = pageid2))
  }, preserve_exact_body_bytes = TRUE)

	expect_is(bb, "list")
	expect_is(bb$scinames, "data.frame")
	expect_null(bb$synonyms)
  expect_named(bb, c('scinames', 'synonyms', 'vernacular', 'refs', 'data_objects'))
})

#' @title NCBI taxon information from uids
#'
#' @description Downloads summary taxon information from the NCBI taxonomy 
#' databases for a set of taxonomy UIDs using eutils esummary.
#'
#' @export
#' @param id (character) NCBI taxonomy uids to retrieve information for. 
#' See Details.
#' @param key (character) NCBI Entrez API key. optional. See Details.
#' @param ... Curl options passed on to [crul::verb-GET]
#' @return A `data.frame` with the following columns:
#' * `uid` The uid queried for
#' * `name` The name of the taxon; a binomial name if the taxon is of rank species
#' * `rank` The taxonomic rank (e.g. 'Genus')
#' @author Zachary Foster \email{zacharyfoster1989@@Sgmail.com}
#' @details If your input vector or list of NCBI IDs is longer than about 
#' 2500 characters (use `nchar(paste(ids, collapse = "+"))`), split 
#' the list up into chunks since at about that number of characters you 
#' will run into the HTTP 414 error "Request-URI Too Long".
#' 
#' @section HTTP version:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')` 
#' 
#' @section Authentication:
#' See [taxize-authentication] for help on authentication. 
#' We strongly recommend getting an API key
#' 
#' @examples \dontrun{
#' ncbi_get_taxon_summary(c(1430660, 4751))
#'
#' # use curl options
#' ncbi_get_taxon_summary(c(1430660, 4751), verbose = TRUE)
#' }
ncbi_get_taxon_summary <- function(id, key = NULL, ...) {
  # Argument validation -------------------------------------------------------
  if (is.null(id)) return(NULL)
  if (length(id) <= 1 && is.na(id)) return(NA)
  id <- as.character(id)
  toolong <- nchar(paste(id, collapse = "+")) > 2500L
  if (toolong) message(sub("\n", "", "Number of ids long; we're splitting up into 
chunks for multiple HTTP requests"))
  # Make eutils esummary query ------------------------------------------------
  key <- getkey(key, "ENTREZ_KEY")
  # split into chunks if needed
  if (toolong) {
    ids_list <- lapply(split(id, ceiling(seq_along(id) / 250)), paste, collapse = "+")
  } else {
    ids_list <- list(paste(id, collapse = "+"))
  }

  cli <- crul::HttpClient$new(url = ncbi_base(), headers = tx_ual,
    opts = list(http_version = 2L, ...))
  out <- list()
  for (i in seq_along(ids_list)) {
    query <- tc(list(db = "taxonomy", id = ids_list[[i]], api_key = key))
    # Search ncbi taxonomy for uid ----------------------------------------------
    rr <- cli$get("entrez/eutils/esummary.fcgi", query = query)
    if (!rr$success()) {
      warning("query failed, proceeding to next if there is one")
      out[[i]] <- paste0(rr$status_code, ": ", rr$status_http()$message)
    } else {
      out[[i]] <- parse_ncbi_gts(rr)
    }
    if (is.null(key)) Sys.sleep(0.33)
  }
  # filter to only data.frame's
  out <- Filter(is.data.frame, out)

  df <- dt2df(out)
  df$.id <- NULL
  return(df)
}

parse_ncbi_gts <- function(x) {    
  raw_results <- x$parse("UTF-8")
  results <- xml2::read_xml(raw_results)
  output <- data.frame(stringsAsFactors = FALSE,
    uid = xml_text_all(results, "/eSummaryResult//DocSum/Id"),
    name = xml_text_all(results, 
      "/eSummaryResult//DocSum/Item[@Name='ScientificName']"),
    rank = xml_text_all(results, "/eSummaryResult//DocSum/Item[@Name='Rank']")
  )
  output$rank[output$rank == ''] <- "no rank"
  return(output)
}

context("ncbi_get_taxon_summary")

test_that("ncbi_get_taxon_summary returns correct class and result", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("ncbi_get_taxon_summary", {
    tt <- ncbi_get_taxon_summary(c(4751))
    tt2 <- ncbi_get_taxon_summary(NA)
    tt3 <- ncbi_get_taxon_summary(id = NULL)
  })

  expect_is(tt, "data.frame")
  expect_equal(ncol(tt), 3)
  expect_equal(tt[1, 3], 'kingdom')
  expect_equal(tt2, NA)
  expect_equal(tt3, NULL)
})


test_that("ncbi_get_taxon_summary behaves correctly when very large ID vector", {
  skip_on_cran() # uses secrets

  # short UID's are okay at larger quantities
  ids <- rep(4751, 550)
  vcr::use_cassette("ncbi_get_taxon_summary_many_ids_short", {
    x <- ncbi_get_taxon_summary(id = ids)
  })
  expect_is(x, "data.frame")

  # but longer IDs add up of course
  ids <- rep(1430660, 1100)
  vcr::use_cassette("ncbi_get_taxon_summary_many_ids_long", {
    expect_message(ncbi_get_taxon_summary(id = ids), "splitting up")
  })
})

taxize_ds <- new.env()
.onLoad <- function(libname, pkgname) {
  utils::data("rank_ref", package = "taxize", envir = taxize_ds)
  utils::data("rank_ref_zoo", package = "taxize", envir = taxize_ds)
  utils::data("theplantlist", package = "taxize", envir = taxize_ds)
  utils::data("apg_families", package = "taxize", envir = taxize_ds)
  utils::data("apg_orders", package = "taxize", envir = taxize_ds)
  utils::data("worrms_ranks", package = "taxize", envir = taxize_ds)
}

#' Resolve names using Open Tree of Life (OTL) resolver
#'
#' @export
#' @param names (character) taxon names to be queried
#' @param context_name name of the taxonomic context to be searched
#' (length-one character vector). Must match (case sensitive) one
#' of the values returned by [rotl::tnrs_contexts()].
#' @param do_approximate_matching (logical) A logical indicating whether or not
#' to perform approximate string (a.k.a. \dQuote{fuzzy}) matching. Using `FALSE`
#' will greatly improve speed. Default: `TRUE`
#' @param ids An array of OTL ids to use for identifying names. These
#' will be assigned to each name in the names array. If ids is
#' provided, then ids and names must be identical in length.
#' @param include_suppressed (logical) Ordinarily, some quasi-taxa, such as
#' incertae sedis buckets and other non-OTUs, are suppressed from TNRS
#' results. If this parameter is true, these quasi-taxa are allowed as
#' possible TNRS results. Default: `FALSE`
#'
#' @return A data frame summarizing the results of the query. The
#' original query output is appended as an attribute to the
#' returned object (and can be obtained using
#' `attr(object, "original_response")`).
#'
#' @param ... Curl options passed on to `httr::POST` within
#' [rotl::tnrs_match_names()]
#' @author Francois Michonneau \email{francois.michonneau@@gmail.com}
#' Scott Chamberlain
#' @seealso [gnr_resolve()], [tnrs()]
#' @references
#' https://github.com/OpenTreeOfLife/germinator/wiki/TNRS-API-v3#match_names
#' @keywords resolve names taxonomy
#' @examples \dontrun{
#' tol_resolve(names=c("echinodermata", "xenacoelomorpha",
#'   "chordata", "hemichordata"))
#' tol_resolve(c("Hyla", "Salmo", "Diadema", "Nautilus"))
#' tol_resolve(c("Hyla", "Salmo", "Diadema", "Nautilus"),
#'   context_name = "Animals")
#'
#' turducken_spp <- c("Meleagris gallopavo", "Anas platyrhynchos",
#'   "Gallus gallus")
#' tol_resolve(turducken_spp, context_name="Animals")
#' }
tol_resolve <- function(names = NULL, context_name = NULL,
  do_approximate_matching = TRUE, ids = NULL, include_suppressed = FALSE, ...) {

  rotl::tnrs_match_names(names = names, context_name = context_name,
    do_approximate_matching = do_approximate_matching, ids = ids, 
    include_suppressed = include_suppressed, ...)
}

context("tol_resolve")

test_that("tol_resolve basic usage works", {
  skip_on_cran()
  
  vcr::use_cassette("tol_resolve", {
    xx <- c("echinodermata", "xenacoelomorpha", "chordata", "hemichordata")
    tmp <- tol_resolve(names = xx)
  })

  expect_is(tmp, "data.frame")
  expect_is(tmp$search_string, "character")
  expect_type(tmp$number_matches, "integer")
  expect_equal(NROW(tmp), 4)

  # query gives back expected output
  expect_equal(
    xx,
    tolower(tmp$unique_name)
  )
})

test_that("context_name works correctly", {
  skip_on_cran()

  vcr::use_cassette("tol_resolve_context_name", {
    aa <- tol_resolve(c("Hyla", "Salmo", "Diadema", "Nautilus"),
                context_name = "Animals")
  })

  expect_is(aa, "data.frame")
  expect_is(aa$search_string, "character")
  expect_type(aa$number_matches, "integer")

  # errors when not in accepted set
  expect_error(
    tol_resolve(c("Hyla", "Salmo", "Diadema", "Nautilus"),
                context_name = "stuff"),
    "is not valid. Check possible values using tnrs_contexts"
  )
})

test_that("do_approximate_matching works correctly", {
  skip_on_cran()

  vcr::use_cassette("tol_resolve_do_approximate_matching", {
    aa <- tol_resolve("Nautilas", do_approximate_matching = TRUE)
  })

  expect_is(aa, "data.frame")
  expect_equal(aa$unique_name, "Nautilus")

  # FIXME: bring back when new rotl ver on CRAN
  vcr::use_cassette("tol_resolve_do_approximate_matching_false", {
    expect_warning(
      tol_resolve("Nautilas", do_approximate_matching = FALSE),
      "not matched"
    )
  })
})

test_that("fails well", {
  skip_on_cran()
  
  expect_error(tol_resolve(), "You must supply")
  expect_error(tol_resolve(ids = 5), "You must supply a")
  expect_error(tol_resolve(context_name = "stuff"), "is not valid")
  expect_error(tol_resolve(names = 5), "must be of class")
})

#' Search uBio for taxonomic synonyms by hierarchiesID.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @rdname ubio_synonyms-defunct
#' @export
#' @param ... Parameters, ignored
ubio_synonyms <- function(...) {
  .Defunct(msg = "the uBio API is down, for good as far as we know, see `?taxize-defunct`")
}

# tests for ubio_synonyms fxn in taxize
context("ubio_synonyms")

out <- ubio_synonyms(hierarchiesID = 4091702)

test_that("ubio_synonyms returns the correct value", {
	expect_that(dim(out), equals(c(1,9)))
})

test_that("ubio_synonyms returns the correct class", {
	expect_that(out, is_a("data.frame"))	
})

#' @title TNRS sources
#' @description Defunct - service was down far too much to be reliable
#' @keywords internal 
#' @rdname tnrs_sources-defunct
#' @export
#' @param ... ignored
tnrs_sources <- function(...) {
  .Defunct(msg = "This function is defunct. See ?tnrs_sources")
}

context("tnrs_sources")

test_that("tnrs_sources returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("tnrs_sources", {
    out <- tnrs_sources()
  })

  expect_that(length(out), equals(3))
  expect_that(out, is_a("character"))
})

#' Search Barcode of Life for taxonomic IDs
#'
#' @export
#' @param sci (character) One or more scientific names.
#' @param id (integer) One or more BOLD taxonomic identifiers.
#' @param fuzzy (logical) Whether to use fuzzy search or not (default: `FALSE`).
#' Only used if `name` passed.
#' @param dataTypes (character) Specifies the datatypes that will be returned.
#' See Details for options. This variable is ignored if `name` parameter is passed,
#' but is used if the `id` parameter is passed.
#' @param includeTree (logical) If TRUE (default: FALSE), returns a list containing
#' information for parent taxa as well as the specified taxon. Only used if `id`
#' passed.
#' @param response (logical) Note that response is the object that returns from the
#' curl call, useful for debugging, and getting detailed info on the API call.
#' @param name Deprecated, see `sci`
#' @param ... named curl options passed on to [crul::verb-GET]
#' @details You must provide one of `sci` or `id` to this function. The other
#' parameters are optional. Note that when passing in `sci`, `fuzzy` can be used
#' as well, while if `id` is passed, then `fuzzy` is ignored, and `dataTypes`
#' `includeTree` can be used.
#'
#' Options for `dataTypes` parameter:
#'
#' * all returns all data
#' * basic returns basic taxon information
#' * images returns specimen image. Includes copyright information, image URL,
#' image metadata.
#' * stats Returns specimen and sequence statistics. Includes public species
#' count, public BIN count, public marker counts, public record count,
#' specimen count, sequenced specimen count, barcode specimen count, species
#' count, barcode species count.
#' * geo Returns collection site information. Includes country, collection
#' site map.
#' * sequencinglabs Returns sequencing labs. Includes lab name, record count.
#' * depository Returns specimen depositories. Includes depository name,
#' record count.
#' * thirdparty Returns information from third parties. Includes wikipedia
#' summary, wikipedia URL, GBIF map.
#'
#' @references http://www.boldsystems.org/index.php/resources/api
#' @return A list of data.frame's.
#' @examples \dontrun{
#' # A basic example
#' bold_search(sci="Apis")
#' bold_search(sci="Agapostemon")
#' bold_search(sci="Poa")
#'
#' # Fuzzy search
#' head(bold_search(sci="Po", fuzzy=TRUE))
#' head(bold_search(sci="Aga", fuzzy=TRUE))
#'
#' # Many names
#' bold_search(sci=c("Apis","Puma concolor"))
#' nms <- names_list('species')
#' bold_search(sci=nms)
#'
#' # Searching by ID - dataTypes can be used, and includeTree can be used
#' bold_search(id=88899)
#' bold_search(id=88899, dataTypes="stats")
#' bold_search(id=88899, dataTypes="geo")
#' bold_search(id=88899, dataTypes="basic")
#' bold_search(id=88899, includeTree=TRUE)
#' }
bold_search <- function(sci = NULL, id = NULL, fuzzy = FALSE,
  dataTypes = 'basic', includeTree=FALSE, response=FALSE, name = NULL, ...) {

  if (!is.null(name)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "bold_search(name)", with = "bold_search(sci)")
    sci <- name
  }
  
  stopifnot(!is.null(sci) | !is.null(id))
  type <- if (is.null(sci)) "id" else "sci"
  tmp <- switch(type,
         sci = bold_tax_name(name = sci, fuzzy = fuzzy, response = response, ...),
         id = bold_tax_id2(id = id, dataTypes = dataTypes, includeTree = includeTree,
                           response = response, ...)
  )
  return(tmp)
}

#' Barcode of Life taxonomic children
#' 
#' BEWARE: this function scrapes data from the BOLD website, so may 
#' be unstable. That is, one day it may work, and the next it may fail.
#' Open an issue if you encounter an error: 
#' https://github.com/ropensci/taxize/issues
#'
#' @export
#' @keywords internal
#' @param id (integer) A BOLD taxonomic identifier. length=1. required
#' @param ... named curl options passed on to [crul::verb-GET]
#' debugging
#' @return list of data.frame's
#' @examples \dontrun{
#' # Osmia (genus): 253 children
#' bold_children(id = 4940)
#' # Momotus (genus): 3 children
#' bold_children(id = 88899)
#' # Momotus aequatorialis (species): no children
#' bold_children(id = 115130)
#' # Osmia sp1 (species): no children
#' bold_children(id = 293378)
#' # Arthropoda (phylum): 27 children
#' bold_children(id = 82)
#' # Psocodea (order): 51 children
#' bold_children(id = 737139)
#' # Megachilinae (subfamily): 2 groups (tribes: 3, genera: 60)
#' bold_children(id = 4962)
#' # Stelis (species): 78 taxa
#' bold_children(id = 4952)
#' }
bold_children <- function(id, ...) {
  stopifnot(length(id) == 1)
  x <- crul::HttpClient$new("https://v4.boldsystems.org", opts = list(...))
  res <- x$get("index.php/Taxbrowser_Taxonpage", query = list(taxid = id))
  res$raise_for_status()
  html <- xml2::read_html(res$parse("UTF-8"))
  nodes <- xml2::xml_find_all(html,
    '//div[@class = "row"]//div[@class = "ibox float-e-margins"]//ol')
  if (length(nodes) == 0) {
    message("no children found")
    return(list(tibble::tibble()))
  }
  group_nmz <- xml2::xml_find_all(html,
    '//div[@class = "row"]//div[@class = "ibox float-e-margins"]//lh')
  bb <- lapply(nodes, bold_children_each_node)
  if (length(group_nmz) > 0) {
    lst_nmz <- tolower(gsub("\\([0-9]+\\)|\\s", "", xml2::xml_text(group_nmz)))
    for (i in seq_along(lst_nmz)) {
      ranknm <- bold_make_rank(lst_nmz[i])
      bb[[i]]$rank <- ranknm
      if (!is.na(ranknm)) names(bb)[i] <- ranknm
    }
  }
  return(bb)
}

bold_children_each_node <- function(x) {
  out <- lapply(xml2::xml_find_all(x, ".//a"), function(w) {
    nm <- gsub("\\s\\[[0-9]+\\]$", "", xml2::xml_text(w))
    id <- strextract(xml2::xml_attr(w, "href"), "[0-9]+$")
    data.frame(name = nm, id = id, stringsAsFactors = FALSE)
  })
  tibble::as_tibble(data.table::rbindlist(out))
}

bold_make_rank <- function(z) {
  if (grepl("order", z)) return("order")
  if (grepl("class", z)) return("class")
  if (grepl("fam", z)) return("family")
  if (grepl("tribe", z)) return("tribe")
  if (grepl("gen", z)) return("genus")
  if (grepl("spec", z)) return("species")
  return(NA_character_)
}

test_that("bold_children", {
  skip_on_cran()
  vcr::use_cassette("bold_children", {
    # Momotus (genus): 3 children
    aa <- bold_children(id = 88899)
    # Momotus aequatorialis (species): no children
    bb <- bold_children(id = 115130)
    # Megachilinae (subfamily): 2 groups (tribes: 3, genera: 60)
    cc <- bold_children(id = 4962)
  })

  expect_is(aa, "list")
  expect_named(aa, "species")
  expect_is(aa[[1]], "data.frame")
  expect_named(aa[[1]], c("name", "id", "rank"))

  expect_is(bb, "list")
  expect_null(names(bb))
  expect_is(bb[[1]], "data.frame")
  expect_equal(NROW(bb[[1]]), 0)

  expect_is(cc, "list")
  expect_named(cc, c("tribe", "genus"))
  expect_is(cc[[1]], "data.frame")
  expect_is(cc[[2]], "data.frame")
  expect_named(cc[[1]], c("name", "id", "rank"))
  expect_named(cc[[2]], c("name", "id", "rank"))
})

test_that("bold_children fails well", {
  skip_on_cran()

  # no inputs
  expect_error(bold_children())
  # id length must == 1
  expect_error(bold_children(id = 1:3))

  vcr::use_cassette("bold_children_no_results", {
    x <- bold_children("Adfadfasfadsfasddfads")
  })
  expect_equal(NROW(x[[1]]), 0)
})

test_that("bold_search returns the correct value, dimensions, and classes", {
  skip_on_cran()
  vcr::use_cassette("bold_search", {
    a <- bold_search(sci = "Apis")
    b <- bold_search(sci = "Aga", fuzzy = TRUE)
    c <- bold_search(sci = c("Apis", "Puma concolor"))
    d <- bold_search(id = 88899)
  })

	expect_equal(names(a)[2], "taxid")
	expect_that(a$taxon, equals("Apis"))

  expect_equal(NROW(a), 1)
  expect_gt(NROW(b), 10)
  expect_gt(NROW(c), 1)
  expect_equal(NROW(d), 1)

	expect_is(a, "data.frame")
	expect_is(b, "data.frame")

	expect_that(a$tax_rank, is_a("character"))
	expect_that(d$parentname, is_a("character"))
})

test_that("bold_search is robust to user error", {
  skip_on_cran()

  expect_is(bold_search(sci = "asdfsdf"), "data.frame")
  expect_is(bold_search(sci = ""), "data.frame")
  expect_is(bold_search(id = "asdfsdf"), "data.frame")
  expect_error(bold_search())
})

#' @title taxize authentication
#' 
#' @description Help on authentication
#'
#' @name taxize-authentication
#' @aliases authentication
#' @seealso [key_helpers()]
#' 
#' @section What is an API?:
#' An API is an Application Programming Interface. The term "API" can be used
#' for lots of scenarios, but in this case we're talking about web APIs,
#' or APIs (interfaces) to web resources. \pkg{taxize} interacts with
#' remote databases on the web via their APIs. You don't need to worry
#' about the details of how that all works; just know that some of them
#' require authentication and some do not.
#'
#' @section What are API keys?:
#' For those APIs that require authentication, the way that's typically done
#' is through API keys: alphanumeric strings of variable lengths that are
#' supplied with a request to an API.
#'
#' \pkg{taxize} won't get these keys for you; rather, you have to
#' go get a key for each service, but we do provide information on how
#' to get those keys. See [key_helpers()] for help on how to
#' obtain keys for this package.
#'
#' @section Using API keys:
#' You can store API keys as R options in your `.Rprofile` file, or
#' as environment variables in either your `.Renviron` file or
#' `.bash_profile` file, o`.zshrc` file (if you use oh-my-zsh) or
#' similar. See [Startup] for help on R options and environment
#' variables.
#'
#' Save your API keys with the following names:
#' * Tropicos: R option or env var as 'TROPICOS_KEY'
#' * IUCN: R option or env var as 'IUCN_REDLIST_KEY'
#' * ENTREZ: R option or env var as 'ENTREZ_KEY'
#'
#' If you save in .Renviron it looks like: `ENTREZ_KEY=somekey`
#'
#' If you save in a .bash_profile, .zshrc, or similar file it looks like:
#' `export ENTREZ_KEY=somekey`
#'
#' If you save in a .Rprofile it looks like: `options(ENTREZ_KEY = "somekey")`
#'
#' Remember to restart your R session (and to start a new shell window/tab
#' if you're using the shell) to take advantage of the new R options
#' or environment variables.
#'
#' We strongly recommend using environment variables
#' (https://en.wikipedia.org/wiki/Environment_variable) over R options
#' because environment variables are widely used across programming
#' languages, operating systems, and computing environments; whereas
#' R options are specific to R.
#'
#' Note that NCBI Entrez doesn't require that you use an API key,
#' but you do get a higher rate limit with a key (more requests per
#' time period), from 3 to 10 requests per second, so do get one.
NULL

#' @title Details on `get_*()` functions
#' @description Including outputs from `get_*()` functions, as well as their
#' attributes, and all exception behaviors.
#'
#' @name get_id_details
#'
#' @details This document applies to the following functions:
#' * [get_boldid()]
#' * [get_eolid()]
#' * [get_gbifid()]
#' * [get_ids()]
#' * [get_iucn()]
#' * [get_natservid()]
#' * [get_nbnid()]
#' * [get_tolid()]
#' * [get_tpsid()]
#' * [get_tsn()]
#' * [get_ubioid()]
#' * [get_uid()]
#' * [get_wiki()]
#' * [get_wormsid()]
#'
#' @section attributes:
#' Each output from `get_*()` functions have the following attributes:
#'
#' * *match* (character) - the reason for NA, either 'not found', 'found' or
#' if `ask = FALSE` then 'NA due to ask=FALSE')
#' * *multiple_matches* (logical) - Whether multiple matches were returned by
#' the data source.
#' This can be `TRUE`, even if you get 1 name back because we try to pattern
#' match the name to see if there's any direct matches. So sometimes this
#' attribute is `TRUE`, as well as `pattern_match`, which then returns 1
#' resulting name without user prompt.
#' * *pattern_match* (logical) - Whether a pattern match was made.
#' If `TRUE` then`multiple_matches` must be `TRUE`, and we found a perfect match
#' to your name, ignoring case. If `FALSE`, there wasn't a direct match, and
#' likely you need to pick from many choices or further parameters can be used
#' to limit results
#' * *uri* (character) - The URI where more information can be read on the taxon
#' - includes the taxonomic identifier in the URL somewhere. This may be missing
#' if the value returned is `NA`
#' 
#'
#' @section exceptions:
#' The following are the various ways in which `get_*()` functions behave:
#'
#'  * success - the value returned is a character string or numeric
#'  * no matches found - you'll get an NA, refine your search or possible the
#'  taxon searched for does not exist in the database you're using
#'  * more than on match and `ask = FALSE` - if there's more than one matching
#'  result, and you have set `ask = FALSE`, then we can't determine the single
#'  match to return, so we give back `NA`. However, in this case we do set the
#'  `match` attribute to say `NA due to ask=FALSE & > 1 result` so it's very
#'  clear what happened - and you can even programatically check this as well
#'  * NA due to some other reason - some `get_*()` functions have additional
#'  parameters for filtering taxa. It's possible that even though there's
#'  results (that is, `found` will say `TRUE`), you can get back an NA. This is
#'  most likely if the parameter filters taxa after they are returned from the
#'  data provider and the value passed to the parameter leads to no matches.
NULL

sw <- function(x) suppressWarnings(x)
sm <- function(x) suppressMessages(x)

# set up vcr
library("vcr")
invisible(vcr::vcr_configure(
  dir = "../fixtures",
  filter_sensitive_data = list(
    "<<rredlist_api_token>>" = Sys.getenv("IUCN_REDLIST_KEY"),
    "<<entrez_api_token>>" = Sys.getenv("ENTREZ_KEY"),
    "<<tropicos_api_token>>" = Sys.getenv("TROPICOS_KEY"),
    "<<natureserve_api_token>>" = Sys.getenv("NATURE_SERVE_KEY")
  ),
  record = 'once'
))
vcr::check_cassette_names()

# suppress messages in progressor class in get_* fxns
taxize_options(TRUE, quiet = TRUE)

has_internet <- function() {
  z <- try(suppressWarnings(readLines('https://www.google.com', n = 1)), silent = TRUE)
  !inherits(z, "try-error")
}

skip_if_net_down <- function() {
  if (has_internet()) {
    return()
  }
  testthat::skip("no internet")
}

#' EUBON taxonomy search
#'
#' @export
#' @param query (character) The scientific name to search for. For example:
#' "Bellis perennis", "Prionus" or "Bolinus brandaris". This is an exact search
#' so wildcard characters are not supported
#' @param providers (character) A list of provider id strings concatenated by
#' comma characters. The default : "pesi,bgbm-cdm-server\[col\]" will be
#' used if this parameter is not set. A list of all available provider ids
#' can be obtained from the '/capabilities' service end point. Providers can be
#' nested, that is a parent provider can have sub providers. If the id of the
#' parent provider is supplied all subproviders will be queried. The query
#' can also be restricted to one or more subproviders by using the following
#' syntax: parent-id\[sub-id-1,sub-id2,...\]
#' @param searchMode (character) Specifies the searchMode. Possible search
#' modes are: `scientificNameExact`, `scientificNameLike` (begins with),
#' `vernacularNameExact`, `vernacularNameLike`
#' (contains), `findByIdentifier`. If the a provider does not support the
#' chosen searchMode it will be skipped and the status message in the
#' tnrClientStatus will be set to 'unsupported search mode' in this case.
#' @param addSynonymy (logical) Indicates whether the synonymy of the accepted
#' taxon should be included into the response. Turning this option on may
#' cause an increased response time. Default: `FALSE`
#' @param addParentTaxon (logical) Indicates whether the the parent taxon of
#' the accepted taxon should be included into the response. Turning this option
#' on may cause a slightly increased response time. Default: `FALSE`
#' @param timeout (numeric) The maximum of milliseconds to wait for responses
#' from any of the providers. If the timeout is exceeded the service will just
#' return the responses that have been received so far. The default timeout is
#' 0 ms (wait for ever)
#' @param dedup (character) Allows to deduplicate the results by making use of
#' a deduplication strategy. The deduplication is done by comparing
#' specific properties of the taxon:
#' * id: compares 'taxon.identifier'
#' * id_name: compares 'taxon.identifier' AND
#'  'taxon.taxonName.scientificName'
#' * name: compares 'taxon.taxonName.scientificName' Using the pure
#'  'name' strategy is not recommended.
#'
#' @param limit (numeric/integer) number of records to retrieve. default: 20.
#' This only affects the search mode `scientificNameLike` and
#' `vernacularNameLike`; other search modes are expected to return only one
#' record per check list
#' @param page (numeric/integer) page to retrieve. default: 1. This only
#' affects the search mode `scientificNameLike` and `vernacularNameLike`; other
#' search modes are expected to return only one record per check list
#' @param ... Curl options passed on to [crul::verb-GET]
#' @references https://cybertaxonomy.eu/eu-bon/utis/1.3/doc.html
#' @family eubon-methods
#' @examples \dontrun{
#' eubon_search("Prionus")
#' eubon_search("Salmo", "pesi")
#' eubon_search("Salmo", c("pesi", "worms"))
#' eubon_search("Salmo", "worms", "scientificNameLike")
#' eubon_search("Salmo", "worms", "scientificNameLike", limit = 3)
#' eubon_search("Salmo", "worms", "scientificNameLike", limit = 20, page = 2)
#' eubon_search("Salmo", "worms", addSynonymy = TRUE)
#' eubon_search("Salmo", "worms", addParentTaxon = TRUE)
#' }
eubon_search <- function(query, providers = "pesi",
  searchMode = "scientificNameExact", addSynonymy = FALSE,
  addParentTaxon = FALSE, timeout = 0,
  dedup = NULL, limit = 20, page = 1, ...) {

  assert(providers, 'character')
  assert(searchMode, 'character')
  assert(addSynonymy, 'logical')
  assert(addParentTaxon, 'logical')
  assert(timeout, 'numeric')
  assert(dedup, 'character')
  assert(limit, c('numeric', 'integer'))
  assert(page, c('numeric', 'integer'))
  page <- page - 1
  args <- tc(list(query = query, providers = paste0(providers, collapse = ","),
                  searchMode = searchMode, addSynonymy = as_l(addSynonymy),
                  addParentTaxon = as_l(addParentTaxon), timeout = timeout,
                  pageSize = limit, pageIndex = page))
  cli <- crul::HttpClient$new(file.path(eubon_base(), "search"),
    headers = tx_ual, opts = list(...))
  res <- cli$get(query = args)
  eubon_error(res)
  tmp <- jsonlite::fromJSON(res$parse("UTF-8"), TRUE, flatten = TRUE)
  tmp$query$response[[1]]
}

#' DEFUNCT
#' @export
#' @keywords internal
#' @rdname eubon-defunct
eubon <- function(...) {
  .Defunct(msg = "use eubon_search()")
}

# helpers
eubon_base <- function() "https://cybertaxonomy.eu/eu-bon/utis/1.3"

eubon_error <- function(x) {
  if (grepl("json", x$response_headers$`content-type`)) {
    cs <- jsonlite::fromJSON(
      x$parse("UTF-8"), FALSE)$query[[1]]$clientStatus[[1]]
    if (x$status_code > 201 || cs$statusMessage != "ok") {
      stop(cs$statusMessage)
    }
  } else if (grepl("html", x$response_headers$`content-type`)) {
    if (x$status_code > 201) {
      mssg <- xml2::xml_text(
        xml2::xml_find_first(
          xml2::read_html(x$parse("UTF-8")), "//title"))
      stop(mssg)
    }
  } else {
    stop(x$status_http()$message)
  }
}

context("eubon_search")

test_that("eubon_search works", {
  skip_on_cran()
  vcr::use_cassette("eubon_search", match_requests_on = c('method', 'path'), {
    aa <- eubon_search("Prionus")
    # bb <- eubon_search("Salmo", providers = 'worms') # The worms data source for Eubon does not seem to be working 
    cc <- eubon_search("Salmo", providers = c('pesi', 'worms'))
  })

  expect_is(aa, "data.frame")
  # expect_is(bb, "data.frame")
  expect_is(cc, "data.frame")

  expect_is(aa$otherNames, "list")
  expect_match(aa$taxon.taxonName.scientificName, "Prionus")
  expect_equal(NROW(aa), 1)

  # expect_is(bb$checklistId, "character")
  # expect_match(bb$taxon.taxonName.scientificName, "Salmo")
  # expect_equal(NROW(bb), 1)

  expect_is(cc$taxon.accordingTo, "character")
  expect_match(cc$taxon.taxonName.scientificName, "Salmo")
  expect_equal(NROW(cc), 1)
})

test_that("eubon_search fails well", {
  skip_on_cran()
  
  expect_error(eubon_search("Prionus", limit = "asdf"), 
    "limit must be of class numeric, integer", class = "error")
  expect_error(eubon_search("Prionus", page = "asdf"), 
    "page must be of class numeric, integer", class = "error")
  expect_error(eubon_search("Prionus", timeout = "adf"), 
    "timeout must be of class numeric", class = "error")
  expect_error(eubon_search("Prionus", addParentTaxon = 5), 
    "addParentTaxon must be of class logical", class = "error")
})

# vcr::use_cassette("eubon_search_fails", {
#   test_that("eubon_search fails well", {
#     expect_error(eubon_search("Salmo", 'asdfdf'),
#                  "Error 400 invalid value for request parameter 'providers'")

#     expect_error(eubon_search("Salmo", searchMode = "adfdf"),
#                  "Error 400 Bad Request")
#   })
# }, preserve_exact_body_bytes = TRUE)

test_that("eubon and eubon_search are no longer aliases - eubon is defunct", {
  skip_on_cran()

  expect_error(eubon(), "use eubon_search", class = "error")
})

#' Retrieve all taxa names downstream in hierarchy for NCBI
#'
#' @export
#' @param id (numeric/integer) An NCBI taxonomic identifier
#' @param downto The taxonomic level you want to go down to. See examples below.
#' 		The taxonomic level IS case sensitive, and you do have to spell it
#' 		correctly. See `data(rank_ref)` for spelling.
#' @param intermediate (logical) If `TRUE`, return a list of length two
#' with target taxon rank names, with additional list of data.frame's of
#' intermediate taxonomic groups. Default: `FALSE`
#' @param ... Further args passed on to [ncbi_children()]
#' @return Data.frame of taxonomic information downstream to family from e.g.,
#' 		Order, Class, etc., or if `intermediate=TRUE`, list of length two,
#'   	with target taxon rank names, and intermediate names.
#' @author Scott Chamberlain
#' @section No Rank:
#' A sticky point with NCBI is that they can have designation for taxonomic
#' rank of "No Rank". So we have no way of programatically knowing what to
#' do with that taxon. Of course one can manually look at a name and perhaps
#' know what it is, or look it up on the web - but we can't do anything
#' programatically. So, no rank things will sometimes be missing.
#' 
#' @section Authentication:
#' See [taxize-authentication()] for help on authentication. 
#' We strongly recommend getting an API key
#' 
#' @examples \dontrun{
#' ## genus Apis
#' ncbi_downstream(id = 7459, downto="species")
#'
#' ## get intermediate taxa as a separate object
#' ncbi_downstream(id = 7459, downto="species", intermediate = TRUE)
#'
#' ## Lepidoptera
#' ncbi_downstream(id = 7088, downto="superfamily")
#'
#' ## families in the ferns (Moniliformopses)
#' (id <- get_uid("Moniliformopses"))
#' ncbi_downstream(id = id, downto = "order")
#' }
ncbi_downstream <- function(id, downto, intermediate = FALSE, ...) {

  should_be('intermediate', intermediate, 'logical')

  downto <- tolower(downto)
  poss_ranks <- unique(do.call(c,
    sapply(taxize_ds$rank_ref$ranks, strsplit, split = ",",
      USE.NAMES = FALSE)))
  downto <- match.arg(downto, choices = poss_ranks)
  torank <- sapply(taxize_ds$rank_ref[which_rank(downto), "ranks"],
    function(x) strsplit(x, ",")[[1]][[1]], USE.NAMES = FALSE)

  stop_ <- "not"
  notout <- data.frame(rank = "", stringsAsFactors = FALSE)
  out <- list()
  if (intermediate) intermed <- list()
  iter <- 0
  while (stop_ == "not") {
    iter <- iter + 1
    tt <- dt2df(lapply(id, function(x) ncbi_children(id = x, ...)[[1]]))
    tt$.id <- NULL
    tt <- remove_self_ids(tt, id)
    tt <- rename(tt, c('childtaxa_rank' = 'rank'))
    tt <- prune_too_low(tt, downto, ignore_no_rank = TRUE)

    if (NROW(tt) == 0) {
      out[[iter]] <- data.frame(stringsAsFactors = FALSE)
      stop_ <- "nodata"
    } else {
      if (intermediate) intermed[[iter]] <- tt
      if (NROW(tt[tt$rank == downto, ]) > 0) {
        out[[iter]] <- tt[tt$rank == downto, ]
      }
      if (NROW(tt[!tt$rank == downto, ]) > 0) {
        notout <- tt[!tt$rank %in% torank, ]
      } else {
        notout <- data.frame(rank = downto, stringsAsFactors = FALSE)
      }

      if (all(notout$rank == downto)) {
        stop_ <- "fam"
      } else {
        id <- notout$childtaxa_id
        stop_ <- "not"
      }
    }
  } # end while loop

  tmp <- unique(dt2df(out, idcol = FALSE))
  if (intermediate) {
    list(target = tmp, intermediate = intermed)
  } else {
    tmp
  }
}

remove_self_ids <- function(x, id) {
  x[!as.character(x$childtaxa_id) %in% as.character(id), ]
}

context("ncbi_downstream")

test_that("ncbi_downstream returns correct structure", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("ncbi_downstream", {
    tt <- ncbi_downstream(id = 7459, downto="species")
  })

  expect_is(tt, "data.frame")
  expect_equal(NCOL(tt), 3)
  for (i in seq_along(tt)) expect_is(tt[[i]], "character")
})

test_that("ncbi_downstream does remove some ambiguous taxa", {
  skip_on_cran()
  
  # 590 = "Salmonella"
  ## DOES remove "subsp."
  vcr::use_cassette("ncbi_downstream_ambiguous_false", {
    amb_no <- ncbi_downstream(id = '590', downto = "species", ambiguous = FALSE)
  })

  ## DOES NOT remove "subsp."
  vcr::use_cassette("ncbi_downstream_ambiguous_true", {
    amb_yes <- ncbi_downstream(id = '590', downto = "species", ambiguous = TRUE)
  })

  expect_is(amb_no, "data.frame")
  expect_is(amb_yes, "data.frame")
  expect_gt(NROW(amb_yes), NROW(amb_no))
})

test_that("ncbi_downstream handles when taxa searches return themselves", {
  skip_on_cran() # uses secrets
  # eg.., with `Euchloe` ncbi_downstream was fetching 2 subgenus rank children
  # which return data that had the ids from those subgenera within it
  # fix for https://github.com/ropensci/taxize/issues/807 to remove "self ids"
  # and remove any duplicate records resulting
  vcr::use_cassette("ncbi_downstream_handles_self_ids", {
    tt <- downstream("Euchloe", downto = "species", db = "ncbi",
                     row = 2, messages = FALSE)
  })

  expect_named(tt, "Euchloe")
  expect_is(tt, "downstream")
  expect_is(tt[[1]], "data.frame")
  expect_equal(attr(tt, "db"), "ncbi")
})

test_that("ncbi_downstream doesn't fail on no intermediate data", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("ncbi_downstream_intermediate", {
    tt <- ncbi_downstream(1398485, downto = "no rank", intermediate = TRUE)
  })

  expect_is(tt, "list")
  expect_is(tt$target, "data.frame")
  expect_equal(NROW(tt$target), 0)
  expect_is(tt$intermediate, "list")
  expect_length(tt$intermediate, 0)
})

#' Retrieve all taxa names downstream in hierarchy for BOLD
#'
#' @export
#' @param id (integer) One or more BOLD taxonomic identifiers
#' @param downto (character) The taxonomic level you want to go down to. 
#' See examples below. The taxonomic level IS case sensitive, and you do have 
#' to spell it correctly. See `data(rank_ref)` for spelling.
#' @param intermediate (logical) If `TRUE`, return a list of length two 
#' with target taxon rank names, with additional list of data.frame's of
#' intermediate taxonomic groups. Default: `FALSE`
#' @param ... crul options passed on to [crul::verb-GET]
#' @return data.frame of taxonomic information downstream to family from e.g.,
#' Order, Class, etc., or if `intermediated=TRUE`, list of length two,
#' with target taxon rank names, and intermediate names.
#' @details
#' BEWARE: This function scrapes the BOLD website, so may be unstable. That is,
#' one day it may work, and the next it may fail. Open an issue if you
#' encounter an error: https://github.com/ropensci/taxize/issues
#' @examples \dontrun{
#' ## the genus Gadus
#' bold_downstream(id = 3451, downto="species")
#' 
#' bold_downstream(id = 443, downto="genus")
#' bold_downstream(id = 443, downto="genus", intermediate=TRUE)
#' }
bold_downstream <- function(id, downto, intermediate = FALSE, ...) {
  should_be('intermediate', intermediate, 'logical')

  downto <- tolower(downto)
  poss_ranks <- unique(do.call(c,
    sapply(taxize_ds$rank_ref$ranks, strsplit, split = ",",
      USE.NAMES = FALSE)))
  downto <- match.arg(downto, choices = poss_ranks)
  torank <- sapply(taxize_ds$rank_ref[which_rank(downto), "ranks"],
    function(x) strsplit(x, ",")[[1]][[1]], USE.NAMES = FALSE)

  stop_ <- "not"
  notout <- data.frame(rank = "", stringsAsFactors = FALSE)
  out <- list()
  if (intermediate) intermed <- list()
  iter <- 0
  while (stop_ == "not") {
    iter <- iter + 1
    tt <- dt2df(unlist(lapply(id, bold_children), FALSE), idcol = FALSE)
    tt <- prune_too_low(tt, downto)

    if (NROW(tt) == 0) {
      out[[iter]] <- data.frame(stringsAsFactors = FALSE)
      stop_ <- "nodata"
    } else {
      if (intermediate) intermed[[iter]] <- tt
      if (NROW(tt[tt$rank == downto, ]) > 0) {
        out[[iter]] <- tt[tt$rank == downto, ]
      }
      if (NROW(tt[!tt$rank == downto, ]) > 0) {
        notout <- tt[!tt$rank %in% torank, ]
      } else {
        notout <- data.frame(rank = downto, stringsAsFactors = FALSE)
      }

      if (all(notout$rank == downto)) {
        stop_ <- "fam"
      } else {
        id <- notout$id
        stop_ <- "not"
      }
    }
  } # end while loop

  tmp <- dt2df(out, idcol = FALSE)
  if (intermediate) {
    list(target = tmp, intermediate = intermed)
  } else {
    tmp
  }
}

test_that("bold_downstream", {
  skip_on_cran()
  vcr::use_cassette("bold_downstream", {
    aa <- bold_downstream(id = 3451, downto="species")
  }, preserve_exact_body_bytes = TRUE)

  expect_is(aa, "data.frame")
  expect_is(aa$name, "character")
  expect_is(aa$rank, "character")
  expect_is(aa$id, "character")
  expect_equal(unique(aa$rank), "species")
})

test_that("bold_downstream intermediate param works", {
  skip_on_cran()
  vcr::use_cassette("bold_downstream_intermediate", {
    cc <- bold_downstream(id = 443, downto="genus", intermediate=TRUE)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(cc, "list")
  expect_is(cc$target, "data.frame")
  expect_is(cc$intermediate, "list")

  expect_is(cc$target$name, "character")
  expect_is(cc$target$rank, "character")
  expect_is(cc$target$id, "character")

  expect_equal(unique(cc$target$rank), "genus")
})

test_that("bold_downstream fails well", {
  skip_on_cran()

  expect_error(bold_downstream(198, "adfadf"), "'arg' should be one of")
  expect_error(bold_downstream(198, "Genus", intermediate = "adf"),
               "'intermediate' should be of class 'logical'")
})

#' Parse taxon names using the GBIF name parser.
#'
#' @export
#' @param scientificname (character) scientific names
#' @param ... Further args passed on to [crul::verb-POST]
#' @return A `data.frame` containing fields extracted from parsed
#' taxon names. Fields returned are the union of fields extracted from
#' all species names in `scientificname`.
#' @author John Baumgartner \email{johnbb@@student.unimelb.edu.au}
#' @references https://www.gbif.org/tools/name-parser/about
#' @seealso [gni_parse()], [gna_parse()]
#' @examples \dontrun{
#' gbif_parse(scientificname='x Agropogon littoralis')
#' gbif_parse(c('Arrhenatherum elatius var. elatius',
#'              'Secale cereale subsp. cereale', 'Secale cereale ssp. cereale',
#'              'Vanessa atalanta (Linnaeus, 1758)'))
#' }
gbif_parse <- function(scientificname, ...) {
  cli <- crul::HttpClient$new("https://api.gbif.org", 
    headers = c(list('Content-Type' = "application/json"), tx_ual),
    opts = list(...))
  tt <- cli$post("v1/parser/name", body = jsonlite::toJSON(scientificname))
  tt$raise_for_status()
  stopifnot(tt$response_headers$`content-type` == 'application/json')
  res <- jsonlite::fromJSON(tt$parse("UTF-8"), FALSE)
  res <- lapply(res, function(x) Map(function(z) if (is.null(z)) NA else z, x))
  (tmp <- data.table::setDF(
    data.table::rbindlist(res, fill = TRUE, use.names = TRUE)))
  setNames(tmp, tolower(names(tmp)))
}

context("gbif_parse")

test_that("gbif_parse returns the correct values and dimensions classes", {
  skip_on_cran()
  vcr::use_cassette("gbif_parse", {
    tt <- gbif_parse(scientificname = 'x Agropogon littoralis L.')
  }, preserve_exact_body_bytes = TRUE)

  expect_match(as.character(tt[1,1]), 'x Agropogon littoralis')
	expect_match(as.character(tt[1,2]), 'SCIENTIFIC')

  expect_that(dim(tt), equals(c(1,12)))

	expect_is(tt, "data.frame")
	expect_is(tt$specificepithet, "character")
	expect_is(tt$authorship, "character")
})

#' Search the CANADENSYS Vascan API.
#'
#' @export
#' @param q (character) Can be a scientific name, a vernacular name or a VASCAN
#'    taxon identifier (e.g. 861)
#' @param format (character) One of json (default) or xml.
#' @param raw (logical) If TRUE, raw json or xml returned, if FALSE, parsed
#' data returned.
#' @param ... (list) Further args passed on to [crul::verb-GET]
#' @author Scott Chamberlain
#' @return json, xml or a list.
#' @references API docs https://data.canadensys.net/vascan/api
#' @keywords names taxonomy
#' @details Note that we lowercase all outputs in data.frame's, but when a
#' list is given back, we don't touch the list names.
#' @examples \dontrun{
#' vascan_search(q = "Helianthus annuus")
#' vascan_search(q = "Helianthus annuus", raw=TRUE)
#' vascan_search(q = c("Helianthus annuus", "Crataegus dodgei"), raw=TRUE)
#'
#' # format type
#' ## json
#' c <- vascan_search(q = "Helianthus annuus", format="json", raw=TRUE)
#' library("jsonlite")
#' fromJSON(c, FALSE)
#'
#' ## xml
#' d <- vascan_search(q = "Helianthus annuus", format="xml", raw=TRUE)
#' library("xml2")
#' xml2::read_xml(d)
#'
#' # lots of names, in this case 50
#' splist <- names_list(rank='species', size=50)
#' vascan_search(q = splist)
#'
#' # Curl options
#' invisible(vascan_search(q = "Helianthus annuus", verbose = TRUE))
#' }
vascan_search <- function(q, format='json', raw=FALSE, ...) {
  url <- sprintf("https://data.canadensys.net/vascan/api/0.1/search.%s",
    format)
  cli <- crul::HttpClient$new(url = url, headers = tx_ual,
    opts = list(...))

  if (!length(q) > 1) {
    temp <- cli$get(query = list(q = q))
    temp$raise_for_status()
    out <- temp$parse("UTF-8")
  } else {
    args <- paste(q, collapse = '\n')
    temp <- cli$post(body = list(q = args), encode = "form", ...)
    temp$raise_for_status()
    out <- temp$parse("UTF-8")
  }
  if (raw) {
    return( out )
  } else {
    tmp <- jsonlite::fromJSON(out, FALSE)
    res <- tmp$results[[1]]
    lapply(tmp$results, vascan_parse)
  }
}

vascan_parse <- function(x){
  parsed <- lapply(x$matches, function(y) {
    taxass <- nmslwr(data.frame(y$taxonomicAssertions,
      stringsAsFactors = FALSE))
    dist <- nmslwr(data.frame(rbindlist(y$distribution)))
    vern <- nmslwr(data.frame(rbindlist(y$vernacularNames)))
    list(taxonomicassertions = taxass, distribution = dist,
      vernacularnames = vern)
  })
  if (length(parsed) == 0) {
    nmslwr(data.frame(searchedterm = x$searchedTerm, nummatches = x$numMatches,
                      matches = NA, stringsAsFactors = FALSE))
  } else {
    list(searchedterm = x$searchedTerm, nummatches = x$numMatches,
      matches = parsed)
  }
}

context("vascan_search")

test_that("vascan_search returns the correct class", {
  skip_on_cran()
  vcr::use_cassette("vascan_search", {
    aa <- vascan_search(q = "Helianthus annuus")
    bb <- vascan_search(q = c("Helianthus annuus", "Crataegus dodgei"), raw=TRUE)
    splist <- names_list(rank='species', size=50)
    cc <- vascan_search(q = splist)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(aa, "list")
  expect_is(bb, "character")
  expect_is(cc, "list")
  expect_is(aa[[1]]$matches[[1]]$taxonomicassertions, "data.frame")

  expect_equal(NCOL(aa[[1]]$matches[[1]]$taxonomicassertions), 7)
})

#' Search for presence of taxonomic names in EOL invasive species databases.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @rdname eol_invasive-defunct
#' @keywords internal
eol_invasive <- function(...) {
  .Defunct("eol", "originr", msg = "This function is defunct. See originr::eol()")
}

#' Get taxonomic names for a given taxonomic name query.
#'
#' @param query TSN number (taxonomic serial number).
#' @param get The rank of the taxonomic name to get.
#' @return Taxonomic name for the searched taxon.
#' @examples \dontrun{
#' itis_name(query="Helianthus annuus", get="family")
#' }
#' @export
#' @rdname itis_name-deprecated
itis_name <- function(query = NULL, get = NULL) {
  .Deprecated("tax_name", "taxize", msg="This function is deprecated. This function has changed name to tax_name.")
}

mssg <- function(v, ...) if (v) message(...)
tc <- function(l) Filter(Negate(is.null), l)
tcnull <- function(x) if (all(sapply(x, is.null))) NULL else x

pluck <- function(x, name, type) {
  if (missing(type)) {
    lapply(x, "[[", name)
  } else {
    vapply(x, "[[", name, FUN.VALUE = type)
  }
}

# pluck with unname
pluck_un <- function(x, name, type) unname(pluck(x, name, type))

collapse <- function(x, fxn, class, match=TRUE, ...) {
  tmp <- lapply(x, fxn, ...)
  if (match) {
    structure(sapply(tmp, unclass), class = class,
              name = unlist(sapply(tmp, attr, which = "name")),
              match = sapply(tmp, attr, which = "match"),
              multiple_matches = sapply(tmp, attr, which = "multiple_matches"),
              pattern_match = sapply(tmp, attr, which = "pattern_match"),
              uri = tcnull(sapply(tmp, attr, which = "uri")))
  } else {
    structure(sapply(tmp, unclass), class = class,
              uri = tcnull(sapply(tmp, attr, which = "uri")))
  }
}

evalfxn <- function(x) eval(parse(text = paste0("check", "_", x)))

toid <- function(x, url, class, ...) {
  uri <- sprintf(url, x)
  structure(x, class = class, match = "found", multiple_matches = FALSE,
            pattern_match = FALSE, uri = uri, ...)
}

add_uri <- function(ids, url, z = NULL){
  if ( !all(is.na(ids)) ) {
    uri_ids <- if (!is.null(z)) z else ids
    attr(ids, 'uri') <- sapply(uri_ids, function(x) {
      if (!is.na(x)) sprintf(url, x) else NA
    }, USE.NAMES = FALSE)
  }
  ids
}

assert_rows <- function(rows) {
  if (!all(is.na(rows))) {
    assert(rows, c("numeric", "integer"))
    stopifnot(all(rows > 0))
  }
}

check_rows <- function(z) {
  if (!is.numeric(z) && !any(is.na(z))) {
    stop("'rows' must be numeric or NA", call. = FALSE)
  }
  if (is.numeric(z)) {
    if (length(z) == 1) {
      if (z < 1) {
        stop("'rows' value must be an integer 1 or greater",
             call. = FALSE)
      }
    }
  }
}

sub_rows <- function(x, rows){
  check_rows(rows)
  if ( any(is.na(rows)) ) {
    x
  } else {
    # subset
    if (NROW(x) == 0) {
      x
    } else {
      # check that vector is = or > nrow of data.frame
      if (NROW(x) < max(rows)) rows <- min(rows):NROW(x)
      x[rows, ]
    }
  }
}

sub_vector <- function(x, rows){
  rows <- check_rows(rows)
  if (length(x) < max(rows)) rows <- min(rows):length(x)
  if ( any(is.na(rows)) ) x else x[rows]
}

nstop <- function(x, arg='db') {
  if (is.null(x)) stop(sprintf("Must specify %s!", arg), call. = FALSE)
}

colClasses <- function(d, colClasses) {
  colClasses <- rep(colClasses, len=length(d))
  d[] <- lapply(seq_along(d), function(i) switch(colClasses[i],
    numeric=as.numeric(d[[i]]),
    character=as.character(d[[i]]),
    Date=as.Date(d[[i]], origin='1970-01-01'),
    POSIXct=as.POSIXct(d[[i]], origin='1970-01-01'),
    factor=as.factor(d[[i]]),
    as(d[[i]], colClasses[i]) ))
  d
}

strtrim <- function(str) {
  gsub("^\\s+|\\s+$", "", str)
}

# function to help filter get_*() functions for a rank name or rank itself ----
filt <- function(df, rank, z) {
  if (NROW(df) == 0) {
    df
  } else {
    if (!is.null(z)) {
      mtch <- grep(sprintf("%s", tolower(z)), tolower(df[,rank]))
      if (length(mtch) != 0) {
        df[mtch, ]
      } else {
        data.frame(NULL)
      }
    } else {
      df
    }
  }
}

argsnull <- function(x) {
  if (length(x) == 0) {
    NULL
  } else {
    x
  }
}

as_l <- function(z) {
  if (is.logical(z) || tolower(z) == "true" || tolower(z) == "false") {
    if (z) {
      return('true')
    } else {
      return('false')
    }
  } else {
    return(z)
  }
}

should_be <- function(arg_name, x, class) {
  if (!is(x, class)) {
    stop(sprintf("'%s' should be of class '%s'", arg_name, class), call. = FALSE)
  }
}

move_col <- function(tt, y){
  tt[ c(names(tt)[ -sapply(y, function(m) grep(m, names(tt))) ], y) ]
}

move_col2 <- function(x, y) x[ c(names(x)[-grep(y, names(x))], y) ]

move_col_begin <- function(tt, y){
  tt[ c(y, names(tt)[ -sapply(y, function(m) grep(m, names(tt))) ]) ]
}

pop <- function(x, nms) {
  x[ !names(x) %in% nms ]
}

assert <- function(x, y) {
  if (!is.null(x) && !is_na(x)) {
    if (!inherits(x, y)) {
      stop(deparse(substitute(x)), " must be of class ",
           paste0(y, collapse = ", "), call. = FALSE)
    }
  }
}

is_na <- function(x) {
  if (is.list(x)) return(FALSE)
  if (is.environment(x)) return(FALSE)
  return(all(is.na(x)))
}

assert_state <- function(x, y) {
  if (!is.null(x) && inherits(x, "taxon_state")) {
    if (x$class != y) {
      stop("taxon_state class must match the get_* function called ",
        call. = FALSE)
    }
  }
}

dt2df <- function(x, idcol = TRUE) {
  (data.table::setDF(
    data.table::rbindlist(x, use.names = TRUE, fill = TRUE, idcol = idcol)))
}

dt2tibble <- function(x) {
  tibble::as_tibble(data.table::setDF(
    data.table::rbindlist(x, use.names = TRUE, fill = TRUE))
  )
}

dbswap <- function(x) {
  switch(
    x,
    boldid = "bold",
    eolid = "eol",
    gbifid = "gbif",
    natservid = "natserv",
    nbnid = "nbn",
    tolid = "tol",
    tpsid = "tropicos",
    tsn = "itis",
    uid = "ncbi",
    wormsid = "worms",
    stop("'db' not recognized", call. = FALSE)
  )
}

check_entrez_key <- function (x) {
  tmp <- if (is.null(x)) Sys.getenv("ENTREZ_KEY", "") else x
  if (tmp == "") {
    getOption("entrez_key",
      warning("no API key found for Entrez, proceeding w/o key",
        call. = FALSE))
  } else {
    tmp
  }
}

taxize_ua <- function(x) {
  versions <- c(
    `r-curl` = as.character(utils::packageVersion("curl")),
    crul = as.character(utils::packageVersion("crul")),
    taxize = as.character(utils::packageVersion("taxize"))
  )
  paste0(names(versions), "/", versions, collapse = " ")
}

`%||%` <- function (x, y) if (is.null(x) || length(x) == 0) y else x

# a common set of functions used together, so a helper fxn to do that
xml_text_all <- function(x, xpath) {
  xml2::xml_text(xml2::xml_find_all(x, xpath))
}

tx_ua <- function() {
  versions <- c(paste0("r-curl/", utils::packageVersion("curl")),
    paste0("crul/", utils::packageVersion("crul")),
    sprintf("rOpenSci(taxize/%s)", utils::packageVersion("taxize")))
  paste0(versions, collapse = " ")
}

tx_ual <- list(`User-Agent` = tx_ua(), `X-USER-AGENT` = tx_ua())

warn_db <- function(x, type) {
  if ("db" %in% names(x)) {
    if (!is.null(x$db)) {
      if (x$db != type) {
      warning(
      sprintf("'db' value '%s' ignored; does not match dispatched method '%s'",
            x$db, type),
          call. = FALSE, immediate. = TRUE)
      }
    }
  }
}

strextract <- function(str, pattern) regmatches(str, regexpr(pattern, str))
strexec <- function(str, pattern) regmatches(str, regexec(pattern, str))


ncbi_rate_limit_pause <- function(key) {
  # NCBI limits requests to three per second when no key and ten per second with key
  ncbi_sleep <- taxize_env$options$ncbi_sleep_sec
  if (is.null(key)) {
    tmz <- c(ncbi_sleep, 0.334)
    time_sleep <- tmz[which.max(tmz)]
    Sys.sleep(time_sleep)
  } else {
    tmz <- c(ncbi_sleep, 0.101)
    time_sleep <- tmz[which.max(tmz)]
    Sys.sleep(tmz)
  }
}

test_that("warn_db", {
  expect_null(warn_db(list(), "foo"))
  expect_null(warn_db(NULL, "foo"))
  expect_null(warn_db(list(a = 5), "foo"))
  expect_null(warn_db(list(db = NULL), "foo"))
  expect_null(warn_db(list(db = "foo"), "foo"))
  expect_warning(warn_db(list(db = "bar"), "foo"),
    "'db' value 'bar' ignored; does not match dispatched method 'foo'")
})

skip_on_cran()

took <- function(x) round(system.time(x)[[3]], 1)

test_that("ncbi_rate_limit_pause", {
  taxize_options(quiet=TRUE) # reset
  
  # key is missing
  expect_error(ncbi_rate_limit_pause(), "missing")

  # ncbi_sleep NOT set, key IS NULL: should be around 0.3
  expect_equal(took(ncbi_rate_limit_pause(NULL)), 0.3)

  # ncbi_sleep NOT set, key NOT NULL: should be around 0.1
  expect_equal(took(ncbi_rate_limit_pause("foobar")), 0.1)

  # ncbi_sleep IS set, key IS NULL: should be around 0.3
  taxize_options(ncbi_sleep = 0.6)
  expect_equal(took(ncbi_rate_limit_pause(NULL)), 0.6)

  # ncbi_sleep IS set, key NOT NULL: should be around 0.1
  taxize_options(ncbi_sleep = 0.7)
  expect_equal(took(ncbi_rate_limit_pause("foobar")), 0.7)
})

iucn_base_url <- "https://www.iucnredlist.org/details/%s/0"


#' Get a IUCN Redlist taxon
#'
#' @export
#' @param sci (character) A vector of scientific names. Or, a
#' `taxon_state` object (see [taxon-state])
#' @param messages logical; should progress be printed?
#' @param key (character) required. you IUCN Redlist API key. See
#' [rredlist::rredlist-package] for help on authenticating with
#' IUCN Redlist
#' @param check (logical) Check if ID matches any existing on the DB, only
#' used in [as.iucn()]
#' @param x For `get_iucn()`: Deprecated, see `sci`. For `as.iucn()`, various,
#' see examples
#' @param ... Ignored
#'
#' @return A vector of taxonomic identifiers as an S3 class.
#'
#' Comes with the following attributes:
#' 
#' * *match* (character) - the reason for NA, either 'not found',
#'  'found' or if `ask = FALSE` then 'NA due to ask=FALSE')
#' * *name* (character) - the taxonomic name, which is needed in
#'  [synonyms()] and [sci2comm()] methods since they
#'  internally use \pkg{rredlist} functions which require the taxonomic name,
#'  and not the taxonomic identifier
#' * *ri* (character) - The URI where more information can be
#'  read on the taxon - includes the taxonomic identifier in the URL somewhere
#'
#' *multiple_matches* and *pattern_match* do not apply here as in other `get_*`
#' methods since there is no IUCN Redlist search, so you either get a match or
#' you do not get a match.
#'
#' @details There is no underscore method, because there's no real
#' search for IUCN, that is, where you search for a string, and get back
#' a bunch of results due to fuzzy matching. If that exists in the future
#' we'll add an underscore method here.
#'
#' IUCN ids only work with [synonyms()] and [sci2comm()]
#' methods.
#'
#' @family taxonomic-ids
#'
#' @examples \dontrun{
#' get_iucn("Branta canadensis")
#' get_iucn("Branta bernicla")
#' get_iucn("Panthera uncia")
#'
#' }
get_iucn <- function(sci, messages = TRUE, key = NULL, x = NULL, ...) {

  assert(sci, c("character", "taxon_state", "numeric"))
  assert(messages, "logical")
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_iucn(x)", with = "get_iucn(sci)")
    sci <- x
  }
  
  raw_data <- get_iucn_data(names_or_ids = sci, messages = messages, key = key, ...)
  
  out <- lapply(names(raw_data), function(input_name) {
    result <- raw_data[[input_name]]
    if (length(result) == 1 && is.na(result)) {
      id <- NA_character_
      direct <- NA
      att <- "not found"
    } else {
      id <- result$taxon$sis_id
      direct <- tolower(result$taxon$scientific_name) == tolower(input_name)
      att <- "found"
    }
    list(id = id, name = input_name, att = att, direct = direct)
  })

  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "iucn",
                   match = pluck_un(out, "att", ""),
                   name = pluck_un(out, "name", ""))
  add_uri(ids, get_url_templates$iucn)
}


#' Get a IUCN Redlist taxon data
#' 
#' Used to get IUCN data for other functions to use.
#' 
#' @param latest If `TRUE`, latest use [rredlist::rl_species_latest()] instead of [rredlist::rl_species()]
#' 
#' @keywords internal
get_iucn_data <- function(names_or_ids, messages = TRUE, key = NULL, latest = FALSE, ...) {
  
  if (inherits(names_or_ids, "character") || inherits(names_or_ids, "numeric")) {
    tstate <- taxon_state$new(class = "iucn", names = names_or_ids)
    items <- names_or_ids
  } else {
    assert_state(names_or_ids, "iucn")
    tstate <- names_or_ids
    names_or_ids <- tstate$taxa_remaining()
    items <- c(names_or_ids, tstate$taxa_completed())
  }
  
  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()
  
  
  for (i in seq_along(names_or_ids)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", names_or_ids[i], "'\n")
    result <- tryCatch(
      {
        if (is.numeric(names_or_ids[i]) || grepl(names_or_ids[i], pattern = '^[0-9]+$')) {
          if (latest) {
            tmp <- rredlist::rl_sis_latest(names_or_ids[i], key = key, ...)
          } else {
            tmp <- rredlist::rl_sis(as.numeric(names_or_ids[i]), key = key, ...)
          }
        } else {
          parts <- strsplit(names_or_ids[i], split = ' +')[[1]]
          if (latest) {
            rl_func_to_use <- rredlist::rl_species_latest
          } else {
            rl_func_to_use <- rredlist::rl_species
          }
          if (length(parts) == 2) {
            tmp <- rl_func_to_use(genus = parts[1], species = parts[2], key = key, ...)
          } else {
            tmp <- rl_func_to_use(genus = parts[1], species = parts[2], infra = parts[3], key = key, ...)
          }
        }
        tmp
      },
      error = function(e) {
        NA_integer_
      }
    ) 
    
    if (length(result) == 1 && is.na(result)) {
      att <- "not found"
    } else {
      att <- "found"
    }
    prog$completed(names_or_ids[i], att)
    prog$prog(att)
    tstate$add(as.character(names_or_ids[i]), result)
  }
  out <- tstate$get()
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  return(out)
}



#' @export
#' @rdname get_iucn
as.iucn <- function(x, check = TRUE, key = NULL) {
  UseMethod("as.iucn")
}

#' @export
#' @rdname get_iucn
as.iucn.iucn <- function(x, check = TRUE, key = NULL) x

#' @export
#' @rdname get_iucn
as.iucn.character <- function(x, check = TRUE, key = NULL) {
  make_iucn(x, check = check, key = key)
}

#' @export
#' @rdname get_iucn
as.iucn.list <- function(x, check = TRUE, key = NULL) {
  as.list(make_iucn(unlist(x), check = check, key = key))
}

#' @export
#' @rdname get_iucn
as.iucn.numeric <- function(x, check=TRUE, key = NULL) {
  make_iucn(x, check = check, key = key)
}

#' @export
#' @rdname get_iucn
as.iucn.data.frame <- function(x, check=TRUE, key = NULL) {
  make_iucn(x$ids, check = check, key = key)
}

#' @export
#' @rdname get_iucn
as.data.frame.iucn <- function(x, ...){
  data.frame(ids = unclass(x),
             class = "iucn",
             name = attr(x, "name"),
             match = attr(x, "match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

#' @keywords internal
make_iucn <- function(x, uu = iucn_base_url, clz = "iucn", check = TRUE, key = NULL) {
  if (check) {
    res <- get_iucn(x, key = key)
  } else {
    toid(x, uu, clz)
  }
}

test_that("synonyms returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("synonyms_itis", {
    tt <- sw(synonyms("Poa annua", db = "itis", rows = 1,
      messages = FALSE))
  })

	expect_match(names(tt), "Poa annua")
	expect_match(tt[[1]][1, "syn_name"], "Poa annua var. aquatica")

	expect_is(tt, "synonyms")
	expect_equal(attr(tt, "db"), "itis")
	expect_is(tt[[1]], "data.frame")

	expect_gt(NROW(tt[[1]]), 1)
})

test_that("synonyms works with worms data", {
  skip_on_cran()
  # VCR bugging out on this one?
  # vcr::use_cassette("synonyms_worms", {
  #   tt <- synonyms('Pomatomus saltatrix', db = "worms",
  #     messages = FALSE)
  # })
  tt <- synonyms('Pomatomus saltatrix', db = "worms",
                 messages = FALSE)
  
  expect_match(names(tt), 'Pomatomus saltatrix')
  expect_match(tt$`Pomatomus saltatrix`$valid_name[1], 'Pomatomus saltatrix')

  expect_is(tt, "synonyms")
  expect_equal(attr(tt, "db"), "worms")
  expect_is(tt[[1]], "data.frame")
  expect_is(tt[[1]], "tbl_df")
})

# test_that("synonyms: data sources return consistent outputs", {
#   skip_on_cran()
#   if (Sys.getenv('IUCN_REDLIST_KEY') == "") {
#     skip("No IUCN api key so test not run.")
#   }
#   if (Sys.getenv('TROPICOS_KEY') == "") {
#     skip("No Tropicos api key so test not run.")
#   }
#   
#   # when name not found, returns NA
#   vcr::use_cassette("synonyms_name_not_found", {
#     aa <- synonyms("Foo bar", db="itis", messages = FALSE)
#     bb <- synonyms("Foo bar", db="tropicos", messages = FALSE)
#     cc <- synonyms("Foo barasdfasdf", db="nbn", messages = FALSE)
#     ee <- synonyms("Foo bar", db="worms", messages = FALSE)
#     gg <- synonyms("Foo bar", db="pow", messages = FALSE)
#   })
# 
#   vcr::use_cassette("synonyms_name_not_found_redlist", {
#     ff <- synonyms("Foo bar", db="iucn", messages = FALSE)
#   }, match_requests_on = c("method", "query"))
# 
#   for (i in list(aa, bb, cc, ee, ff, gg)) expect_true(is.na(i[[1]]))
#   
# 
#   # when name found, but no synonyms found, returns empty data.frame
#   vcr::use_cassette("synonyms_name_found_but_no_synonyms", {
#     gg <- synonyms("Epigonus thai", db="worms", messages = FALSE)
#     hh <- synonyms("Arctoa andersonii", db="nbn", rank = "species", rows=1, messages = FALSE)
#     # kk <- synonyms("Pinus balfouriana", db="itis", messages = FALSE)
#     ll <- synonyms("Pinus contorta", db="tropicos", messages = FALSE)
#     mm <- synonyms("Quercus ransomii", db="pow", accepted=FALSE, messages = FALSE)
#   })
# 
#   vcr::use_cassette("synonyms_name_found_but_no_synonyms_redlist", {
#     ii <- synonyms(get_iucn('Ursus americanus'), db="iucn", messages = FALSE)
#   }, match_requests_on = c("method", "query"))
#   
#   # type is data.frame
#   for (i in list(gg, hh, ii, ll, mm)) expect_is(i[[1]], "data.frame")
#   # no rows
#   for (i in list(gg, hh, ii, ll, mm)) expect_equal(NROW(i[[1]]), 0)
#   # no names
#   for (i in list(gg, hh, ii, ll, mm)) expect_equal(length(names(i[[1]])), 0)
# })

test_that("warn on mismatch 'db'", {
  skip_on_cran()
  vcr::use_cassette("synonyms_warn_on_db_mismatch", {
    expect_warning(
      synonyms(
        get_tsn("Poa annua"), db = "tropicos"))
  })
})

#' Search for taxonomy data from Plantminer.com
#'
#' @export
#' @param plants (character) Vector of plant species names. Required.
#' @param from (character) One of tpl (for theplantlist.com data), or
#'    flora (for Brazilian Flora Checklist). Required. Default: `tpl`
#' @param messages (logical) informative messages or not. Default: `TRUE`
#' @param ... curl options passed on to [crul::HttpClient]
#' @return data.frame of results.
#' @note you used to need an API key for Plantminer; it's no longer needed
#' @examples \dontrun{
#' # A single taxon
#' plantminer("Ocotea pulchella")
#'
#' # Many taxa
#' plants <- c("Myrcia lingua", "Myrcia bella", "Ocotea pulchella",
#' 		"Miconia", "Coffea arabica var. amarella", "Bleh")
#' plantminer(plants)
#'
#' # By deafult, tpl is used, for Theplantlist data,
#' # toggle the from parameter here
#' plantminer("Ocotea pulchella", from = "flora")
#' }
plantminer <- function(plants, from = "tpl", messages = TRUE, ...) {
  if (!from %in% c("tpl", "flora")) {
    stop("'from' must be one of 'tpl' or 'flora'")
  }
  do.call(rbind, lapply(seq_along(plants), function(i) {
    mssg(messages, paste(plants[i], collapse = " "))
    cli <- crul::HttpClient$new(url = pmbase(), headers = tx_ual,
                                opts = list(...))
    sp <- cli$get(from, query = list(taxon = plants[i]))
    sp$raise_for_status()
    jsonlite::fromJSON(sp$parse("UTF-8"))
  }))
}

pmbase <- function() "http://www.plantminer.com"

context("plantminer")

test_that("plantminer returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("plantminer", {
    plants <- c("Myrcia lingua", "Myrcia bella", "Ocotea pulchella",
                "Miconia", "Coffea arabica var. amarella", "Bleh")
    df <- plantminer(plants, messages = FALSE)
  })

	expect_is(df, "data.frame")
	expect_is(df$id, "character")

	expect_equal(df$original.search[1], plants[1])
})

test_that("plantminer fails well", {
  skip_on_cran()

  expect_error(plantminer(), "argument \"plants\" is missing")
  
  vcr::use_cassette("plantminer_not_found", {
    expect_equal(plantminer("foo bar", messages=FALSE)$note, "not found")
  })
})

#' Get a random vector of species names.
#'
#' Family and order names come from the APG plant names list. Genus and
#' species names come from Theplantlist.org.
#'
#' @export
#' @param rank (character) Taxonomic rank, one of species, genus (default),
#' family, order
#' @param size (integer/numeric) Number of names to get. Maximum depends on
#' the rank
#' @return character vector of taxonomic names
#' @author Scott Chamberlain 
#' @examples
#' names_list()
#' names_list('species')
#' names_list('genus')
#' names_list('family')
#' names_list('order')
#' names_list('order', 2)
#' names_list('order', 15)
#'
#' # You can get a lot of genus or species names if you want
#' nrow(theplantlist)
#' names_list('genus', 500)
names_list <-  function(rank='genus', size=10) {
  assert(rank, "character")
  assert(size, c('integer', 'numeric'))
  getsp <- function(size){
    tmp <- apply(taxize_ds$theplantlist[
      sample(1:nrow(taxize_ds$theplantlist), size), c('genus','species')], 1,
      function(y) paste(y, collapse = " "))
    names(tmp) <- NULL
    tmp
  }
  switch(rank,
    family = as.character(sample(taxize_ds$apg_families$accepted_name, size)),
    order = as.character(sample(taxize_ds$apg_orders$accepted_name, size)),
    genus = sample(unique(taxize_ds$theplantlist$genus), size),
    species = getsp(size)
  )
}

context("names_list")

test_that("names_list returns the correct dimensions", {
	expect_equal(length(names_list('species')), 10)
	expect_equal(length(names_list('species', size=20)), 20)
	expect_equal(length(names_list('order', size=2)), 2)
})

test_that("names_list returns the correct class", {
	expect_is(names_list('species'), "character")
})

#' @title Resolve names from different data sources
#'
#' @description Resolve names from iPlant's name resolver, and 
#' the Global Names Resolver (GNR)
#'
#' @export
#' @param sci Vector of one or more taxonomic names (common names not 
#' supported)
#' @param db Source to check names against. One of iplant or gnr. 
#' Default: gnr. Note that each taxonomic data source has their own 
#' identifiers, so that if you provide the wrong `db` value for the 
#' identifier you could get a result, but it will likely be wrong (not 
#' what you were expecting).
#' @param query Deprecated, see `sci`
#' @param ... Curl options passed on to [crul::verb-GET] or 
#' [crul::verb-POST]. In addition, further named args passed 
#' on to  each respective function. See examples
#' @return A list with length equal to length of the db parameter (number 
#' of sources requested), with each element being a data.frame or list 
#' with results from that source.
#' @examples \dontrun{
#' resolve(sci=c("Helianthus annuus", "Homo sapiens"))
#' resolve(sci="Quercus keloggii", db='gnr')
#' resolve(sci=c("Helianthus annuus", "Homo sapiens"), db=c('iplant', 'gnr'))
#' resolve(sci="Quercus keloggii", db=c('iplant', 'gnr'))
#'
#' # pass in options specific to each source
#' resolve("Helianthus annuus", db = 'gnr', preferred_data_sources = c(3, 4))
#' resolve("Helianthus annuus", db = 'iplant', retrieve = 'best')
#' identical(
#'  resolve("Helianthus annuus", db = 'iplant', retrieve = 'best')$iplant,
#'  iplant_resolve("Helianthus annuus", retrieve = 'best')
#' )
#'
#' # pass in curl options
#' resolve(sci="Qercuss", db = "iplant", verbose = TRUE)
#' }
resolve <- function(sci, db = 'gnr', query = NULL, ...) {
  if (!is.null(query)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "resolve(query)", with = "resolve(sci)")
    sci <- query
  }
  
  if (!is.null(query)) sci <- query
  db <- match.arg(db, choices = c('iplant', 'gnr'), 
    several.ok = TRUE)
  foo <- function(x, y, ...){
    res <- switch(x,
      gnr = tryDefault(gnr_resolve(sci = y, ...)),
      iplant = tryDefault(iplant_resolve(sci = y, ...))
    )
    if (is.null(res)) "Error: no data found" else res
  }
  stats::setNames(lapply(db, function(z) foo(z, sci, ...)), db)
}

tryDefault <- function(expr, default = NULL, quiet = TRUE) {
  result <- default
  if (quiet) {
    tryCatch(result <- expr, error = function(e) {
    })
  } else {
    try(result <- expr)
  }
  result
}

#' Get an ID for a IUCN listed taxon
#'
#' @export
#' @param sciname character; Scientific name. Should be cleaned and in the
#' format `*<Genus> <Species>*`. One or more.
#' @param key (character) required. you IUCN Redlist API key. See
#' [rredlist::rredlist-package] for help on authenticating with
#' IUCN Redlist
#' @param ... Curl options passed on to [crul::HttpClient]
#' @return A named list (names are input taxa names) of one or more IUCN IDs.
#' Taxa that aren't found are silently dropped.
#' @author Scott Chamberlain, 
#' @examples \dontrun{
#' iucn_id("Branta canadensis")
#' iucn_id("Branta bernicla")
#' iucn_id("Panthera uncia")
#' iucn_id("Lynx lynx")
#'
#' # many names
#' iucn_id(c("Panthera uncia", "Lynx lynx"))
#'
#' # many names, some not found
#' iucn_id(c("Panthera uncia", "Lynx lynx", "foo bar", "Gorilla gorilla gorilla"))
#'
#' # a name not found
#' iucn_id("Foo bar")
#' }
iucn_id <- function(sciname, key = NULL, ...) {
  out <- list()
  parts <- strsplit(sciname, split = ' +')
  lengths <- vapply(parts, length, FUN.VALUE = numeric(1))
  invalid_lengths <- lengths <= 1 | lengths > 3
  if (any(invalid_lengths)) {
    error_list <- paste0('    ', which(invalid_lengths), ': "', sciname[invalid_lengths], '"')
    if (length(error_list) > 100) {
      error_list <- c(error_list[1:100], '   ...')
    }
    stop(
      'The following inputs have the incorrect number of elements in their species name:\n',
      paste0(error_list, collapse = '\n')
    )
  }
  
  for (i in seq_along(parts)) {
    out[[i]] <- get_iucn_id(parts[[i]], key = key, ...)
  }
  unlist(out)
}

get_iucn_id <- function(parts, key = NULL, ...) {
  tryCatch(
    {
      if (length(parts) == 2) {
        tmp <- rredlist::rl_species(genus = parts[1], species = parts[2], key = key, ...)
      } else {
        tmp <- rredlist::rl_species(genus = parts[1], species = parts[2], infra = parts[3], key = key, ...)
      }
      tmp$taxon$sis_id
    },
    error = function(e) {
      NA_integer_
    }
  ) 
}

context("iucn_id")

test_that("iucn_id returns the correct class", {
  skip_on_cran()

  if (Sys.getenv('IUCN_REDLIST_KEY') == "") {
    skip("No IUCN api key so test not run.")
  }

  vcr::use_cassette("iucn_id", {
    aa <- iucn_id("Branta canadensis")
    bb <- iucn_id("Panthera uncia")
    cc <- iucn_id("Lynx lynx")
  },
    preserve_exact_body_bytes = TRUE,
    match_requests_on = c("method", "query")
  )

  expect_is(aa, "integer")
  expect_is(bb, "integer")
  expect_is(cc, "integer")

  expect_equal(aa, 22679935)
  expect_equal(bb, 22732)
  expect_equal(cc, 12519)
})

test_that("iucn_id fails well", {
  skip_on_cran()
  
  if (Sys.getenv('IUCN_REDLIST_KEY') == "") {
    skip("No IUCN api key so test not run.")
  }

  expect_error(iucn_id(), "argument \"sciname\" is missing")
  
  vcr::use_cassette("iucn_id_fail", {
    expect_equal(suppressWarnings(iucn_id("foo bar")), NA_integer_)
  }, match_requests_on = c("method", "query"))
})

#' Function to get API key.
#'
#' Checks first to get key from your .Rprofile or .Renviron (or similar) file
#'
#' @export
#' @param x (character) An API key, defaults to `NULL`
#' @param service (character) The API data provider, used to match to
#' default guest key (for Tropicos; there's no guest
#' key for NCBI or IUCN, for which you have to get your own)
#' @examples \dontrun{
#' getkey(service="tropicos")
#' getkey(service="iucn")
#' getkey(service="entrez")
#' }
getkey <- function(x = NULL, service) {
  if (is.null(x)) {
    keynames <- c("TROPICOS_KEY", "IUCN_REDLIST_KEY", "ENTREZ_KEY")
    service <- grep(service, keynames, ignore.case = TRUE, value = TRUE)
    key <- getOption(service)
    if (is.null(key)) key <- Sys.getenv(service, "")

    # if Entrez, return either way as no key actually required
    if (service == "ENTREZ_KEY") {
      if (is.null(key) || !nzchar(key)) {
        message("No ENTREZ API key provided\n Get one via taxize::use_entrez()\nSee https://ncbiinsights.ncbi.nlm.nih.gov/2017/11/02/new-api-keys-for-the-e-utilities/")
        return(NULL)
      } else {
        return(key)
      }
    }

    # if IUCN, stop if no key as a key is required
    if (service == "IUCN_REDLIST_KEY") {
      if (is.null(key) || !nzchar(key)) {
        stop("No IUCN API key provided\nSee taxize::use_iucn()")
      }
      return(key)
    }

    if (is.null(key) || !nzchar(key)) {
      keys <- c("00ca3d6a-cbcc-4924-b882-c26b16d54446",
                "44f1a53227f1c0b6238a997fcfe7513415f948d2")
      names(keys) <- keynames[1:2]
      key <- keys[[service]]
      key_helpers <- c("taxize::use_tropicos()", "taxize::use_iucn()")
      names(key_helpers) <- keynames[1:2]
      message(paste("Using default key: Please get your own API key via ",
                    key_helpers[service], sep = ""))
    } else if (inherits(key, "character")) {
      key <- key
    } else {
      stop("check your key - it should be a character string",
        call. = FALSE)
    }
  } else {
    key <- x
  }

  return(key)
}

#' Lookup in the APGIII taxonomy and replace family names
#'
#' @export
#' @param taxa (character) Taxonomic name to lookup a synonym for
#' in APGIII taxonomy.
#' @param rank (character) Taxonomic rank to lookup a synonym for. One of
#' family or order.
#' @return A APGIII family or order name, the original name if the name
#' is the same as APG has, or NA if no match found
#' @details Internally in this function, we use the datasets [apg_families]
#' and [apg_orders] - see their descriptions for the data in them.
#' The functions [apgOrders()] [apgFamilies()] are for scraping
#' current content from the http://www.mobot.org/MOBOT/research/APweb/
#' website
#'
#' The datasets used in this function are from the most recent version of
#' APGIII, Version 14 (http://www.mobot.org/MOBOT/research/APweb/)
#' @examples
#' # New name found
#' apg_lookup(taxa = "Hyacinthaceae", rank = "family")
#' # Name is the same
#' apg_lookup(taxa = "Poaceae", rank = "family")
#' apg_lookup(taxa = "Asteraceae", rank = "family")
#' # Name not found
#' apg_lookup(taxa = "Foobar", rank = "family")
#' 
#' # New name found
#' apg_lookup(taxa = "Acerales", rank = "order")
#' # Name is the same
#' apg_lookup(taxa = "Acorales", rank = "order")
#' # Name not found
#' apg_lookup(taxa = "Foobar", rank = "order")
apg_lookup <- function(taxa, rank = "family") {
  if (!rank %in% c('family', 'order')) {
    stop("rank must be one of family or order")
  }
  af <- taxize_ds$apg_families
  if (rank == "family") {
    x <- af[af$family %in% taxa,]$synonym
    if (length(x) == 0) {
      message("no match found...")
      out <- NA_character_
    } else {
      if (is.na(x)) {
        message("name is the same...")
        out <- taxa
      } else {
        message("new name...")
        out <- x
      }
    }
  } else {
    ao <- taxize_ds$apg_orders
    x <- ao[ao$order %in% taxa,]$synonym
    if (length(x) == 0) {
      message("no match found...")
      out <- NA_character_
    } else {
      if (is.na(x)) {
        message("name is the same...")
        out <- taxa
      } else {
        message("new name...")
        out <- x
      }
    }
  }
  return(out)
}

#' Ping an API used in taxize to see if it's working.
#'
#' @name ping
#' @param what (character) One of status (default), content, or an HTTP status
#' code. If status, we just check that the HTTP status code is 200, or similar
#' signifying the service is up. If content, we do a simple, quick check to
#' determine if returned content matches what's expected. If an HTTP status
#' code, it must match an appropriate code. See [status_codes()].
#' @param key (character) NCBI Entrez API key. optional. See [get_uid()]
#' @param ... Curl options passed on to [crul::verb-GET]
#' @return A logical, TRUE or FALSE
#' @details For ITIS, see [ritis::description], which provides number of
#' scientific and common names in a character string.
#' @section HTTP version for NCBI requests:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')` 
#' @examples \dontrun{
#' col_ping()
#' col_ping("content")
#' col_ping(200)
#' col_ping("200")
#' col_ping(204)
#'
#' itis_ping()
#' eol_ping()
#' ncbi_ping()
#' tropicos_ping()
#' nbn_ping()
#'
#' gbif_ping()
#' gbif_ping(200)
#'
#' bold_ping()
#' bold_ping(200)
#' bold_ping("content")
#'
#' ipni_ping()
#' ipni_ping(200)
#' ipni_ping("content")
#'
#' vascan_ping()
#' vascan_ping(200)
#' vascan_ping("content")
#'
#' # curl options
#' vascan_ping(verbose = TRUE)
#' eol_ping(500, verbose = TRUE)
#' }

#' @export
#' @rdname ping
col_ping <- function(what = "status", ...) {
  res <- pGET('https://www.catalogueoflife.org/col/webservice?name=Apis', ...)
  switch(
    matchwhat(what),
    status = match_status(res),
    code = match_code(res, what),
    content = {
      identical(
        "Apis",
        xml_text(xml_find_first(xml_children(xml2::read_xml(res$parse("UTF-8")))[[1]], "name")))
    }
  )
}

#' @export
#' @rdname ping
eol_ping <- function(what = "status", ...) {
  res <- pGET('https://eol.org/api/ping/1.0.json', ...)
  switch(
    matchwhat(what),
    status = match_status(res),
    code = match_code(res, what),
    content = grepl("success", 
      jsonlite::fromJSON(res$parse("UTF-8"))$response$message,
      ignore.case = TRUE)
  )
}

#' @export
#' @rdname ping
itis_ping <- function(what = "status", ...) {
  res <- pGET("https://www.itis.gov/ITISWebService/services/ITISService/getDescription", ...)
  switch(matchwhat(what),
         status = match_status(res),
         code = match_code(res, what),
         content = {
           tt <- xml_text(xml2::xml_children(xml2::read_xml(res$parse("UTF-8")))[[1]])
           grepl("this is the itis web service", tt, ignore.case = TRUE)
         })
}

#' @export
#' @rdname ping
ncbi_ping <- function(what = "status", key = NULL, ...) {
  key <- getkey(key, "ENTREZ_KEY")
  args <- tc(list(api_key = key))
  res <- pGET(paste0(ncbi_base(), 
    "/entrez/eutils/efetch.fcgi?db=taxonomy&ID=4232"), args = args, 
    http_version = 2L, ...)
  switch(
    matchwhat(what),
    status = match_status(res),
    code = match_code(res, what),
    content = identical("4232",
      xml_text(xml_find_first(
        xml_children(read_xml(res$parse("UTF-8")))[[1]], "TaxId")))
  )
}

#' @export
#' @rdname ping
tropicos_ping <- function(what = "status", ...) {
  res <- pGET("http://services.tropicos.org/Name/25509881?apikey=f3e499d4-1519-42c9-afd1-685a16882f5a&format=xml", ...)
  error = FALSE
  if (grepl("exception occurred", res$parse("UTF-8"), ignore.case = TRUE)) error = TRUE
  switch(
    matchwhat(what),
    status = if (error) TRUE else match_status(res),
    code = if (error) TRUE else match_code(res, what),
    content = {
      if (error)
        TRUE
      else
        grepl(
          25509881,
          xml2::as_list( xml2::read_xml(res$parse("UTF-8")))$Name$NameId[[1]]
        )
    }
  )
}

#' @export
#' @rdname ping
nbn_ping <- function(what = "status", ...) {
  res <- pGET("https://species-ws.nbnatlas.org/species/NHMSYS0000502940", ...)
  switch(
    matchwhat(what),
    status = match_status(res),
    code = match_code(res, what),
    content = {
      jsonlite::fromJSON(res$parse("UTF-8"))$taxonConcept$guid == 'NHMSYS0000502940'
    }
  )
}

#' @export
#' @rdname ping
gbif_ping <- function(what = "status", ...) {
  res <- pGET("https://api.gbif.org/v1/species/1", ...)
  switch(matchwhat(what),
         status = match_status(res),
         code = match_code(res, what),
         content = grepl("1", jsonlite::fromJSON(res$parse("UTF-8"))$key, 
          ignore.case = TRUE))
}

#' @export
#' @rdname ping
bold_ping <- function(what = "status", ...) {
  res <- pGET("http://v4.boldsystems.org/index.php/API_Tax/TaxonData?taxId=88899&dataTypes=basic&includeTree=FALSE", ...)
  switch(
    matchwhat(what),
    status = match_status(res),
    code = match_code(res, what),
    content = grepl("88899", jsonlite::fromJSON(res$parse("UTF-8"))$taxid,
                    ignore.case = TRUE)
  )
}

#' @export
#' @rdname ping
ipni_ping <- function(what = "status", ...) {
  res <- pGET("https://www.ipni.org/ipni/advPlantNameSearch.do?find_genus=Brintonia&output_format=delimited-minimal", ...)
  switch(matchwhat(what),
         status = match_status(res),
         code = match_code(res, what),
         content = {
           txt <- res$parse("UTF-8")
           dat <- read.delim(text = txt, sep = "%", 
            stringsAsFactors = FALSE)
           grepl("Asteraceae", dat$Family[1], ignore.case = TRUE)
        })
}

#' @export
#' @rdname ping
vascan_ping <- function(what = "status", ...) {
  res <- pGET("https://data.canadensys.net/vascan/api/0.1/search.json?q=Crataegus", ...)
  switch(matchwhat(what),
         status = match_status(res),
         code = match_code(res, what),
         content = grepl("Crataegus", jsonlite::fromJSON(res$parse("UTF-8"))$results[[1]], 
          ignore.case = TRUE)
  )
}

#' @export
#' @rdname ping
fg_ping <- function(what = "status", ...) {
  res <- pGET("http://www.indexfungorum.org/ixfwebservice/fungus.asmx/NameFullByKey?NameLsid=urn:lsid:indexfungorum.org:names:17703", ...)
  switch(matchwhat(what),
         status = match_status(res),
         code = match_code(res, what),
         content = grepl("Gymnopus", xml_text(read_xml(res$parse("UTF-8"))), 
          ignore.case = TRUE)
  )
}


matchwhat <- function(x){
  x <- as.character(x)
  if ( x %in% c("status", "content") ) x else "code"
}

match_status <- function(x){
  stopifnot(inherits(x, "HttpResponse"))
  x$status_code == 200
}

match_code <- function(x, y){
  stopifnot(inherits(x, "HttpResponse"))
  x$status_code == y
}

pGET <- function(url, args = list(), ...) {
  cli <- crul::HttpClient$new(url, headers = tx_ual, opts = list(...))
  cli$get(query = args)
}

#' uBio ping
#'
#' @export
#' @rdname ubio_ping-defunct
ubio_ping <- function() {
  .Defunct(msg = "the uBio API is down, for good as far as we know")
}

context("ping")

test_that("ncbi_ping returns the correct value", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("ncbi_ping", {
    expect_true(ncbi_ping())
    Sys.sleep(1)
    expect_false(ncbi_ping(503))
    Sys.sleep(1)
    expect_true(ncbi_ping("content"))
  })
})

test_that("trpicos_ping returns the correct value", {
  skip_on_cran() # uses secrets
  skip_on_ci()
  vcr::use_cassette("trpicos_ping", {
    expect_true(tropicos_ping())
    expect_true(tropicos_ping("content"))
  }, preserve_exact_body_bytes = TRUE)
})

test_that("nbn_ping returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("nbn_ping", {
    expect_true(nbn_ping())
    expect_false(nbn_ping(503))
    expect_true(nbn_ping("content"))
  })
})

test_that("gbif_ping returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("gbif_ping", {
    expect_true(gbif_ping())
    expect_false(gbif_ping(503))
    expect_true(gbif_ping("content"))
  })
})

test_that("bold_ping returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("bold_ping", {
    expect_true(bold_ping())
    expect_false(bold_ping(503))
    expect_true(bold_ping("content"))
  }, match_requests_on = c("method", "query"))
})

# test_that("ipni_ping returns the correct value", {
#   skip_on_cran()
#
#   expect_true(ipni_ping())
#   expect_false(ipni_ping(503))
#   expect_true(ipni_ping("content"))
# })

test_that("vascan_ping returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("vascan_ping", {
    expect_true(vascan_ping())
    expect_false(vascan_ping(503))
    expect_true(vascan_ping("content"))
  }, preserve_exact_body_bytes = TRUE, match_requests_on = c("method", "query"))
})

context("itis_ping")

test_that("itis_ping returns the correct class", {
  skip_on_cran()
  vcr::use_cassette("itis_ping", {
    one <- itis_ping()
  })

  expect_is(one, "logical")
})

test_that("itis_ping returns correct things", {
  skip_on_cran()
  vcr::use_cassette("itis_ping_http_code_503", {
    expect_false(itis_ping(503))
  })

  vcr::use_cassette("itis_ping_http_code_200", {
    expect_true(itis_ping(200))
  })
})

context("eol_ping")

test_that("eol_ping returns the correct value", {
  skip_on_cran()
  skip_on_ci()

  Sys.sleep(0.5)
  expect_true(eol_ping())
  expect_false(eol_ping(503))
  Sys.sleep(0.5)
  expect_true(eol_ping("content"))
})

#' Get NCBI taxonomy UID from GenBankID
#'
#' @export
#' @param id A GenBank accession alphanumeric string, or a gi numeric string.
#' @param batch_size The number of queries to submit at a time.
#' @param key (character) NCBI Entrez API key. optional. See Details.
#' @param ... Curl args passed on to [crul::HttpClient]
#' @details See https://www.ncbi.nlm.nih.gov/Sitemap/sequenceIDs.html for
#' help on why there are two identifiers, and the difference between them.
#' 
#' @section Authentication:
#' See [taxize-authentication] for help on authentication. We
#' recommend getting an API key.
#' 
#' @section HTTP version:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')` 
#' 
#' @section Rate limits:
#' In case you run into errors due to your rate limit being exceeded, see
#' [taxize_options()], where you can set `ncbi_sleep`.
#'
#' @return one or more NCBI taxonomic IDs
#' @examples \dontrun{
#' # with accession numbers
#' genbank2uid(id = 'AJ748748')
#' genbank2uid(id = 'Y13155')
#' genbank2uid(id = 'X78312')
#' genbank2uid(id = 'KM495596')
#'
#' # with gi numbers
#' genbank2uid(id = 62689767)
#' genbank2uid(id = 22775511)
#' genbank2uid(id = 156446673)
#'
#' # pass in many accession or gi numbers
#' genbank2uid(c(62689767,156446673))
#' genbank2uid(c('X78312','KM495596'))
#' genbank2uid(list('X78312',156446673))
#'
#' # curl options
#' res <- genbank2uid(id = 156446673, verbose = TRUE)
#' }
genbank2uid <- function(id, batch_size = 100, key = NULL, ...) {
  assert(batch_size, c("integer", "numeric"))
  process_batch <- function(id, key, ...) {
    #removes version number of accession ids
    id <- gsub(pattern = "\\.[0-9]+$", "", id)
    
    key <- getkey(key, "ENTREZ_KEY")
    
    # Make NCBI eutils query
    query <- tc(list(db = "nucleotide", id = paste(id, collapse = ","), api_key = key))

    # Execute query
    cli <- crul::HttpClient$new(url = ncbi_base(), headers = tx_ual,
      opts = list(http_version = 2L, ...))
    res <- cli$get("entrez/eutils/esummary.fcgi", query = query)
    res$raise_for_status()
    parsed_xml <- read_xml(res$parse('UTF-8'))
    
    # Extract taxon ID and sequences name
    taxon_ids <- xml_text(
      xml_find_all(parsed_xml,
        "//eSummaryResult//DocSum//Item[@Name='TaxId']"))
    titles <- xml_text(
      xml_find_all(parsed_xml,
        "//eSummaryResult//DocSum//Item[@Name='Title']"))
    
    # Add NAs for failed queries
    raw_errors <- xml_text(xml_find_all(parsed_xml, "//eSummaryResult//ERROR"))
    if (length(raw_errors) > 0) {
      error_regexes <- c('^Invalid uid (.+) at position=[0-9]+$', 
                         '^Failed uid="(.+)"$')
      error_ids <- lapply(error_regexes, function(r) {
        vapply(raw_errors, function(z) strexec(z, r)[[1]][2], "",
          USE.NAMES=FALSE)
      })
      error_ids <- unlist(error_ids)
      error_ids <- error_ids[!is.na(error_ids)]
      error_ids <- unique(error_ids)
      add_error_na <- function(values) {
        output <- rep(NA, length(id))
        output[- match(error_ids, id)] <- values
        return(output)
      }
      taxon_ids <- add_error_na(taxon_ids)
      titles <- add_error_na(titles)
    }
    
    # combine results into a data.frame and return
    result <- data.frame(id = taxon_ids, name = titles,
                         stringsAsFactors = FALSE)
    # NCBI limits requests to three per second when no key
    ncbi_rate_limit_pause(key)
    return(result)
  }
  # Run each batch and combine
  batches <- split(id, ceiling(seq_along(id) / batch_size))
  batch_results <- lapply(batches,
                          function(x) map_unique(x, process_batch, 
                            key = key, ...))
  result <- do.call(rbind, batch_results)
  
  # Convert to list format
  output <- lapply(seq_len(nrow(result)), function(i) {
    my_uid <- result[i, "id"]
    my_uid <- as.uid(result[i, "id"], check = FALSE)
    if (is.na(my_uid)) {
      attr(my_uid, "match") <- "not found"
      attr(my_uid, "name") <- "unknown"
    } else {
      attr(my_uid, "match") <- "found"
      attr(my_uid, "name") <- result[i, "name"]
    }
    return(my_uid)
  })
  
  # Alert user to errors
  if (any(is.na(result$id))) {
    failed <- paste0("[", which(is.na(result$id)), "] ", id[is.na(result$id)])
    warning(paste0("The following ", sum(is.na(result$id)), " of ", nrow(result),
                   " queries could not be found:\n  ",
                   limited_print(failed, type = "silent")), call. = FALSE)
   }
  return(output)
}

is_acc <- function(x){
  gg <- suppressWarnings(as.numeric(x))
  is.na(gg)
}

#=====================================================================
# get indexes of a unique set of the input
unique_mapping <- function(input) {
  unique_input <- unique(input)
  vapply(input, function(x) which(x == unique_input), numeric(1))
}

#=====================================================================
# run a function on unique values of a iterable
map_unique <- function(input, func, ...) {
  input_class <- class(input)
  unique_input <- unique(input)
  class(unique_input) <- input_class
  func(unique_input, ...)
  # func(unique_input, ...)[unique_mapping(input)]
}

context("genbank2uid")

test_that("genbank2uid - with accession numbers", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("genbank2uid", {
    aa <- genbank2uid(id = 'AJ748748')
    bb <- genbank2uid(c('X78312','KM495596'))
  })

  expect_is(aa, "list")
  expect_is(aa[[1]], "uid")
  expect_is(unclass(aa[[1]]), "character")
  expect_is(attr(aa[[1]], "match"), "character")
  expect_false(attr(aa[[1]], "multiple_matches"))
  expect_false(attr(aa[[1]], "pattern_match"))
  expect_is(attr(aa[[1]], "uri"), "character")
  expect_match(attr(aa[[1]], "uri"), "http")

  expect_is(bb, "list")
  expect_is(bb[[1]], "uid")
  expect_is(unclass(bb[[1]]), "character")
  expect_is(attr(bb[[1]], "match"), "character")
  expect_equal(vapply(bb, attr, TRUE, which = "multiple_matches"), c(FALSE, FALSE))
  expect_equal(vapply(bb, attr, TRUE, which = "pattern_match"), c(FALSE, FALSE))
  expect_is(attr(bb[[1]], "uri"), "character")
  expect_match(attr(bb[[1]], "uri"), "http")
})

test_that("genbank2uid - with gi numbers", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("genbank2uid_gi_numbers", {
    aa <- genbank2uid(id = 62689767)
    bb <- genbank2uid(c(62689767,156446673))
  })

  expect_is(aa, "list")
  expect_is(aa[[1]], "uid")
  expect_is(unclass(aa[[1]]), "character")
  expect_is(attr(aa[[1]], "match"), "character")
  expect_false(attr(aa[[1]], "multiple_matches"))
  expect_false(attr(aa[[1]], "pattern_match"))
  expect_is(attr(aa[[1]], "uri"), "character")
  expect_match(attr(aa[[1]], "uri"), "http")

  expect_is(bb, "list")
  expect_is(bb[[1]], "uid")
  expect_is(unclass(bb[[1]]), "character")
  expect_is(attr(bb[[2]], "match"), "character")
  expect_equal(vapply(bb, attr, TRUE, which = "multiple_matches"), c(FALSE, FALSE))
  expect_equal(vapply(bb, attr, TRUE, which = "pattern_match"), c(FALSE, FALSE))
  expect_is(attr(bb[[1]], "uri"), "character")
  expect_match(attr(bb[[1]], "uri"), "http")
})

### THIS ONLY RETURNS 1 RESULT NOW
# test_that("genbank2uid - where ID has more than one taxon associated", {
#   skip_on_cran()
#
#   aa <- genbank2uid(id = "AM420293")
#
#   expect_is(aa, "list")
#   expect_is(aa[[1]], "uid")
#   expect_is(aa[[2]], "uid")
#   expect_is(aa[[3]], "uid")
#   expect_match(attr(aa[[1]], "name"), "Saccharopolyspora")
#   expect_match(attr(aa[[2]], "name"), "Saccharopolyspora")
#   expect_match(attr(aa[[3]], "name"), "Saccharopolyspora")
# })

test_that("genbank2uid - fails well", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("genbank2uid_fails", {
    # one Id not found
    expect_warning(genbank2uid(id = "gwa2_scaffold_1731_16S_1"),
                   "The following 1 of 1 queries could not be found")
    aa <- suppressWarnings(genbank2uid(id = "gwa2_scaffold_1731_16S_1"))
    expect_is(aa[[1]], "uid")
    expect_true(is.na(unclass(aa[[1]])))

    # many Ids not found
    expect_warning(genbank2uid(id = c("gwa2_scaffold_1731_16S_1", "asdfadfs")),
                   "The following 2 of 2 queries could not be found")
    expect_warning(genbank2uid(id = c("gwa2_scaffold_1731_16S_1", "asdfadfs")),
                   "The following 2 of 2 queries could not be found")
  })

  skip_on_cran()
  aa <- suppressWarnings(genbank2uid(id = c("gwa2_scaffold_1731_16S_1",
                                            "asdfadfs")))
  expect_is(aa[[1]], "uid")
  expect_true(all(is.na(unclass(aa[[1]]))))

  # id not given
  expect_error(genbank2uid(), "argument \"id\" is missing")

  # id not given
  expect_error(genbank2uid("Asdfdf", batch_size = "Asdf"),
               "batch_size must be of class integer, numeric")
})

# param: x (character) a rank name (e.g., "variety")
which_rank <- function(x, zoo = FALSE) {
  rr <- if (zoo) taxize_ds$rank_ref_zoo else taxize_ds$rank_ref
  which(sapply(rr$ranks, function(z) {
    any(unlist(strsplit(z, split = ",")) == x)
  }, USE.NAMES = FALSE))
}

prune_too_low <- function(x, rank, ignore_no_rank = FALSE, zoo = FALSE,
  ignore_missing_rank = FALSE) {

  # ignore_missing_rank
  if (ignore_missing_rank) x$rank[is.na(x$rank)] <- "no rank"

  # the pruning
  rr <- if (zoo) taxize_ds$rank_ref_zoo else taxize_ds$rank_ref
  rank_target_no <- as.numeric(rr[which_rank(rank, zoo = zoo), "rankid"])
  rank_nos <- as.numeric(
    rr[vapply(x$rank, function(z) which_rank(z, zoo = zoo), 1),
    "rankid"])
  if (ignore_no_rank) rank_nos[rank_nos %in% c(300, 400)] <- 0
  x[!rank_nos > rank_target_no, ]
}

#' Get any matching IUCN species names
#'
#' @export
#' @param name character; taxon name
#' @param verbose logical; should messages be printed?
#' @param ... Further arguments passed on to [iucn_summary()], note
#' that you'll need an API key.
#' @seealso [iucn_summary()] [iucn_status()]
#' @return Character vector of names that matched in IUCN
#' @details Beware: IUCN functions can give back incorrect data. This isn't
#' our fault. We do our best to get you the correct data quickly, but sometimes
#' IUCN gives back the wrong data, and sometimes Global Names gives back the
#' wrong data. We will fix these as soon as possible. In the meantime, just
#' make sure that the data you get back is correct.
#' @examples \dontrun{
#' iucn_getname(name = "Cyanistes caeruleus")
#' iucn_getname(name = "Panthera uncia")
#'
#' # not found in IUCN search
#' iucn_getname(name = "Acacia allenii")
#' }
iucn_getname <- function(name, verbose = TRUE, ...) {
  mssg(verbose, "searching Global Names ...")
  all_names <- gna_search(sci = name, parse_names = TRUE)
  if (NROW(all_names) == 0) {
    stop("No names found matching ", name, call. = FALSE)
  }
  mssg(verbose, "searching IUCN...")
  out <- suppressWarnings(iucn_summary(all_names$canonical, ...))
  x <- all_names$canonical[!sapply(out, function(x) x$status) %in% NA]
  unique(as.character(x))
}

context("iucn_getname")

# test_that("iucn_getname returns the correct value", {
#   skip_on_cran()
#      
#   vcr::use_cassette("iucn_getname", {
#     temp <- sm(iucn_getname("Cyanistes caeruleus", verbose = FALSE))
#   })
#
#   expect_equal(temp, "Cyanistes caeruleus")
#   expect_is(temp, "character")
#   expect_equal(length(temp), 1)
# })

#' Get the UID codes from NCBI for taxonomic names.
#'
#' Retrieve the Unique Identifier (UID) of a taxon from NCBI taxonomy browser.
#'
#' @export
#' @param sci_com character; scientific or common name. Or, a `taxon_state`
#' object (see [taxon-state])
#' @param ask logical; should get_uid be run in interactive mode? If TRUE and
#' more than one TSN is found for the species, the user is asked for input. If
#' FALSE NA is returned for multiple matches.
#' @param messages logical; If `TRUE` (default) the actual taxon queried is
#' printed on the console.
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this function still only gives back a uid
#' class object with one to many identifiers. See [get_uid_()] to get back
#' all, or a subset, of the raw data that you are presented during the ask
#' process.
#' @param modifier (character) A modifier to the `sci_com` given. Options
#' include: Organism, Scientific Name, Common Name, All Names, Division,
#' Filter, Lineage, GC, MGC, Name Tokens, Next Level, PGC, Properties, Rank,
#' Subtree, Synonym, Text Word. These are not checked, so make sure they are
#' entered correctly, as is.
#' @param rank_query (character) A taxonomic rank name to modify the query sent
#' to NCBI. See [rank_ref] for possible options. Though note that
#' some data sources use atypical ranks, so inspect the data itself for
#' options. Optional. See `Querying` below.
#' @param division_filter (character) A division (aka phylum) name to filter
#' data after retrieved from NCBI. Optional. See `Filtering` below.
#' @param rank_filter (character) A taxonomic rank name to filter data after
#' retrieved from NCBI. See [rank_ref] for possible options.
#' Though note that some data sources use atypical ranks, so inspect the data
#' itself for options. Optional. See `Filtering` below.
#' @param key (character) NCBI Entrez API key. optional. See Details.
#' @param x Input to [as.uid()]
#' @param sciname Deprecated, see `sci_com`
#' @param ... Ignored
#' @param check logical; Check if ID matches any existing on the DB, only used
#'   in [as.uid()]
#' @template getreturn
#' 
#' @section Rate limits:
#' In case you run into errors due to your rate limit being exceeded, see
#' [taxize_options()], where you can set `ncbi_sleep`.
#'
#' @section Querying: The parameter `rank_query` is used in the search sent
#'   to NCBI, whereas `rank_filter` filters data after it comes back. The
#'   parameter `modifier` adds modifiers to the name. For example,
#'   `modifier="Organism"` adds that to the name, giving e.g.,
#'   `Helianthus[Organism]`.
#'
#' @section Filtering: The parameters `division_filter` and
#'   `rank_filter` are not used in the search to the data provider, but are
#'   used in filtering the data down to a subset that is closer to the target
#'   you want. For all these parameters, you can use regex strings since we use
#'   [grep()] internally to match. Filtering narrows down to the set
#'   that matches your query, and removes the rest.
#'
#' @section Beware: NCBI does funny things sometimes. E.g., if you search on
#'   Fringella morel, a slight misspelling of the genus name, and a non-existent
#'   epithet, NCBI gives back a morel fungal species. In addition, NCBI doesn't
#'   really do fuzzy searching very well, so if there is a slight mis-spelling
#'   in your names, you likely won't get what you are expecting. The lesson:
#'   clean your names before using this function. Other data sources are better
#'   about fuzzy matching.
#'
#' @section Authentication:
#' See [taxize-authentication] for help on authentication
#'
#' Note that even though you can't pass in your key to `as.uid` functions,
#' we still use your Entrez API key if you have it saved as an R option
#' or environment variable.
#' 
#' @section HTTP version:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')`
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @author Eduard Szoecs, \email{eduardszoecs@@gmail.com}
#'
#' @examples \dontrun{
#' get_uid(c("Chironomus riparius", "Chaetopteryx"))
#' get_uid(c("Chironomus riparius", "aaa vva"))
#'
#' # When not found
#' get_uid("howdy")
#' get_uid(c("Chironomus riparius", "howdy"))
#'
#' # Narrow down results to a division or rank, or both
#' ## By modifying the query
#' ### w/ modifiers to the name
#' get_uid(sci_com = "Aratinga acuticauda", modifier = "Organism")
#' get_uid(sci_com = "bear", modifier = "Common Name")
#'
#' ### w/ rank query
#' get_uid(sci_com = "Pinus", rank_query = "genus")
#' get_uid(sci_com = "Pinus", rank_query = "subgenus")
#' ### division query doesn't really work, for unknown reasons, so not available
#'
#' ## By filtering the result
#' ## Echinacea example
#' ### Results w/o narrowing
#' get_uid("Echinacea")
#' ### w/ division
#' get_uid(sci_com = "Echinacea", division_filter = "eudicots")
#' get_uid(sci_com = "Echinacea", division_filter = "sea urchins")
#'
#' ## Satyrium example
#' ### Results w/o narrowing
#' get_uid(sci_com = "Satyrium")
#' ### w/ division
#' get_uid(sci_com = "Satyrium", division_filter = "monocots")
#' get_uid(sci_com = "Satyrium", division_filter = "butterflies")
#'
#' ## Rank example
#' get_uid(sci_com = "Pinus")
#' get_uid(sci_com = "Pinus", rank_filter = "genus")
#' get_uid(sci_com = "Pinus", rank_filter = "subgenus")
#'
#' # Fuzzy filter on any filtering fields
#' ## uses grep on the inside
#' get_uid("Satyrium", division_filter = "m")
#'
#' # specify rows to limit choices available
#' get_uid('Dugesia') # user prompt needed
#' get_uid('Dugesia', rows=1) # 2 choices, so returns only 1 row, so no choices
#' get_uid('Dugesia', ask = FALSE) # returns NA for multiple matches
#'
#' # Go to a website with more info on the taxon
#' res <- get_uid("Chironomus riparius")
#' browseURL(attr(res, "uri"))
#'
#' # Convert a uid without class information to a uid class
#' as.uid(get_uid("Chironomus riparius")) # already a uid, returns the same
#' as.uid(get_uid(c("Chironomus riparius","Pinus contorta"))) # same
#' as.uid(315567) # numeric
#' as.uid(c(315567,3339,9696)) # numeric vector, length > 1
#' as.uid("315567") # character
#' as.uid(c("315567","3339","9696")) # character vector, length > 1
#' as.uid(list("315567","3339","9696")) # list, either numeric or character
#' ## dont check, much faster
#' as.uid("315567", check=FALSE)
#' as.uid(315567, check=FALSE)
#' as.uid(c("315567","3339","9696"), check=FALSE)
#' as.uid(list("315567","3339","9696"), check=FALSE)
#'
#' (out <- as.uid(c(315567,3339,9696)))
#' data.frame(out)
#' as.uid( data.frame(out) )
#'
#' # Get all data back
#' get_uid_("Puma concolor")
#' get_uid_("Dugesia")
#' get_uid_("Dugesia", rows=2)
#' get_uid_("Dugesia", rows=1:2)
#' get_uid_(c("asdfadfasd","Pinus contorta"))
#'
#' # use curl options
#' get_uid("Quercus douglasii", verbose = TRUE)
#' }

get_uid <- function(sci_com, ask = TRUE, messages = TRUE, rows = NA,
                    modifier = NULL, rank_query = NULL,
                    division_filter = NULL, rank_filter = NULL,
                    key = NULL, sciname = NULL, ...) {

  assert(sci_com, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(messages, "logical")
  assert(modifier, "character")
  assert(rank_query, "character")
  assert(division_filter, "character")
  assert(rank_filter, "character")
  assert_rows(rows)
  key <- getkey(key, service = "entrez")
  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_uid(sciname)", with = "get_uid(sci_com)")
    sci_com <- sciname
  }
  

  if (inherits(sci_com, "character")) {
    tstate <- taxon_state$new(class = "uid", names = sci_com)
    items <- sci_com
  } else {
    assert_state(sci_com, "uid")
    tstate <- sci_com
    sci_com <- tstate$taxa_remaining()
    items <- c(sci_com, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci_com)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci_com[i], "'\n")
    sci_com[i] <- gsub(" ", "+", sci_com[i])
    if (!is.null(modifier))
      sci_com[i] <- paste0(sci_com[i], sprintf("[%s]", modifier))
    term <- sci_com[i]
    if (!is.null(rank_query))
      term <- paste0(term, sprintf(" AND %s[Rank]", rank_query))
    try_again_errors <- c("Could not resolve host: eutils.ncbi.nlm.nih.gov")
    query_args <- tc(list(db = "taxonomy", term = term, api_key = key))
    raw_xml_result <- repeat_until_it_works(try_again_errors,
                                            "esearch",
                                            query = query_args,
                                            ...)
    if (grepl("error", raw_xml_result)) stop(raw_xml_result, call.=FALSE)
    xml_result <- xml2::read_xml(raw_xml_result)

    # NCBI limits requests to three per second
    ncbi_rate_limit_pause(key)
    uid <- xml2::xml_text(xml2::xml_find_all(xml_result, "//IdList/Id"))
    mm <- length(uid) > 1

    if (length(uid) == 0) { # if taxon name is not found
      uid <- NA_character_
    } else {
      att <- 'found'
    }

    # not found on ncbi
    if (length(uid) == 0 || all(is.na(uid))) {
      mssg(messages, m_not_found_sp_altclass)
      uid <- NA_character_
      att <- 'NA due to not found'
    }
    # more than one found on ncbi -> user input
    if (length(uid) > 1) {
      ID <- paste(uid, collapse = ",")
      try_again_errors <- c("Could not resolve host: eutils.ncbi.nlm.nih.gov")
      query_args <- tc(list(db = "taxonomy", ID = ID, api_key = key))
      tt <- repeat_until_it_works(try_again_errors, "esummary",
                                  query_args, ...)
      ttp <- xml2::read_xml(tt)
      df <- parse_ncbi(ttp)
      rownames(df) <- 1:nrow(df)

      if (!is.null(division_filter) || !is.null(rank_filter)) {
        df <- filt(df, "division", division_filter)
        df <- filt(df, "rank", rank_filter)
      }

      df <- sub_rows(df, rows)
      uid <- df$uid
      if (length(uid) == 1) {
        direct <- TRUE
        att <- "found"
      }
      if (length(uid) == 0) {
        uid <- NA_character_
      }

      if (length(uid) > 1) {
        # check for exact match
        matchtmp <- df[
          tolower(
            as.character(df$scientificname)) %in% tolower(sci_com[i]), "uid"]
        if (length(matchtmp) == 1) {
          uid <- as.character(matchtmp)
          direct <- TRUE
        }
      }

      if (length(uid) > 1) {
        if (!ask) {
          if (length(uid) == 1) {
            att <- "found"
          } else {
            warning(
              sprintf(m_more_than_one_found, "UID", sci_com[i]),
              call. = FALSE
            )
            uid <- NA_character_
            att <- m_na_ask_false
          }
        } else {
          # prompt
          rownames(df) <- 1:nrow(df)
          message("\n\n")
          message("\nMore than one UID found for taxon '", sci_com[i], "'!\n
            Enter rownumber of taxon (other inputs will return 'NA'):\n")
          message(paste0(utils::capture.output(df), collapse = "\n"))
          take <- scan(n = 1, quiet = TRUE, what = 'raw')

          if (length(take) == 0) {
            take <- "notake"
            att <- "nothing chosen"
          }
          if (take %in% seq_len(nrow(df))) {
            take <- as.numeric(take)
            message("Input accepted, took UID '",
                    as.character(df$uid[take]), "'.\n")
            uid <- as.character(df$uid[take])
            att <- 'found'
          } else {
            uid <- NA_character_
            att <- "NA due to user input out of range"
            mssg(messages, "\nReturned 'NA'!\n\n")
          }
        }
      }
    }
    res <- list(id = as.character(uid), att = att, multiple = mm,
      direct = direct)
    prog$completed(sci_com[i], att)
    prog$prog(att)
    tstate$add(sci_com[i], res)
  }
  out <- tstate$get()
  ids <- structure(pluck_un(out, "id", ""), class = "uid",
    match = pluck_un(out, "att", ""),
    multiple_matches = pluck_un(out, "multiple", logical(1)),
    pattern_match = pluck_un(out, "direct", logical(1))
  )
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$ncbi)
}

repeat_until_it_works <- function(catch, path, query, max_tries = 3,
  wait_time = 10, messages = TRUE, ...) {

  error_handler <- function(e) {
    if (e$message %in% catch) {
      if (messages) warning(paste("Caught error:", e$message))
      return(NA)
    } else {
      stop(e$message)
    }
  }
  for (count in 1:max_tries) {
    cli <- crul::HttpClient$new(url = ncbi_base(),
      headers = tx_ual, opts = list(http_version = 2L, ...))
    res <- cli$get(sprintf("entrez/eutils/%s.fcgi", path),
      query = tc(query))
    output <- tryCatch(res$parse("UTF-8"), error = error_handler)
    if (!is.na(output)) return(output)
    Sys.sleep(wait_time * count)
  }
  return(output)
}

#' @export
#' @rdname get_uid
as.uid <- function(x, check=TRUE) UseMethod("as.uid")

#' @export
#' @rdname get_uid
as.uid.uid <- function(x, check=TRUE) x

#' @export
#' @rdname get_uid
as.uid.character <- function(x, check=TRUE) if(length(x) == 1) make_uid(x, check) else collapse(x, make_uid, "uid", check=check)

#' @export
#' @rdname get_uid
as.uid.list <- function(x, check=TRUE) if(length(x) == 1) make_uid(x, check) else collapse(x, make_uid, "uid", check=check)

#' @export
#' @rdname get_uid
as.uid.numeric <- function(x, check=TRUE) as.uid(as.character(x), check)

#' @export
#' @rdname get_uid
as.uid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class="uid", match=x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri=x$uri)
}

#' @export
#' @rdname get_uid
as.data.frame.uid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "uid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_uid <- function(x, check=TRUE) {
  make_generic(x, 'https://www.ncbi.nlm.nih.gov/taxonomy/%s',
    "uid", check)
}

check_uid <- function(x){
  key <- getkey(NULL, "ENTREZ_KEY")
  cli <- crul::HttpClient$new(url = ncbi_base(), headers = tx_ual,
    opts = list(http_version = 2L))
  args <- tc(list(db = "taxonomy", id = x, api_key = key))
  res <- cli$get("entrez/eutils/esummary.fcgi", query = args)
  res$raise_for_status()
  tt <- xml2::read_xml(res$parse("UTF-8"))
  tryid <- xml2::xml_text(xml2::xml_find_all(tt, "//Id"))
  identical(as.character(x), tryid)
}


#' @export
#' @rdname get_uid
get_uid_ <- function(sci_com, messages = TRUE, rows = NA, key = NULL,
  sciname = NULL, ...) {
  
  key <- getkey(key, "ENTREZ_KEY")
  if (!is.null(sciname)) {
  lifecycle::deprecate_warn(when = "v0.9.97", what = "get_uid_(sciname)", with = "get_uid_(sci_com)")
    sci_com <- sciname
  }
  
  stats::setNames(lapply(sci_com, get_uid_help, messages = messages,
    rows = rows, key = key, ...), sci_com)
}

get_uid_help <- function(sci_com, messages, rows, key, ...) {
  mssg(messages, "\nRetrieving data for taxon '", sci_com, "'\n")
  cli <- crul::HttpClient$new(url = ncbi_base(), headers = tx_ual,
    opts = list(http_version = 2L, ...))
  res <- cli$get(
    "entrez/eutils/esearch.fcgi",
    query = tc(list(api_key = key,
      db = "taxonomy", term = gsub(" ", "+", sci_com))))
  res$raise_for_status()
  xml_result <- xml2::read_xml(res$parse("UTF-8"))
  Sys.sleep(0.33)
  uid <- xml_text(xml_find_all(xml_result, "//IdList/Id"))
  if (length(uid) == 0) {
    NULL
  } else {
    res <- cli$get("entrez/eutils/esummary.fcgi",
      query = tc(list(api_key = key, db = "taxonomy",
        ID = paste(uid, collapse = ","))))
    res$raise_for_status()
    ttp <- xml2::read_xml(res$parse("UTF-8"))
    df <- parse_ncbi(ttp)
    sub_rows(df, rows)
  }
}

parse_ncbi <- function(x) {
  mget <- c("Status", "Rank", "Division", "ScientificName",
            "CommonName", "TaxId", "Genus", "Species", "Subsp",
            "ModificationDate")
  nget <- paste0('Item[@Name="', mget, "\"]")
  nodes <- xml_find_all(x, "//DocSum")
  tmp <- taxize_ldfast(lapply(nodes, function(z) {
    data.frame(as.list(
      setNames(sapply(nget, function(w) xml_text(xml_find_all(z, w))), tolower(mget))),
      stringsAsFactors = FALSE)
  }))
  rename(tmp, c('taxid' = 'uid'))
}

ncbi_base <- function() "https://eutils.ncbi.nlm.nih.gov"

context("comm2sci")

test_that("comm2sci returns the correct values and classes", {
  skip_on_cran()

  vcr::use_cassette("comm2sci", {
    tt <- suppressMessages(comm2sci(com='american black bear'))
    uu <- suppressMessages(comm2sci('annual blue grass', 
      db='tropicos'))
    temp2 <- suppressMessages(comm2sci(com='black bear', db='tropicos'))
    # FIXME: skip EOL for now until they're bad cert issue sorted out
    # temp3 <- suppressMessages(comm2sci(com=c('black bear','roe deer'),
    #   db='eol'))
  }, preserve_exact_body_bytes = TRUE)

  expect_equal(names(tt), 'american black bear')
  expect_equal(names(uu), 'annual blue grass')

  expect_identical(
    suppressMessages(
      comm2sci(com='bear', db='itis', itisby = "asfasdf"))[[1]], 
    character(0))
  expect_named(temp2)
  # expect_named(temp3)

  expect_that(tt, is_a("list"))
  expect_that(tt[[1]], is_a("character"))
  expect_that(uu, is_a("list"))
  expect_that(uu[[1]], is_a("character"))
  # expect_is(
  #   suppressMessages(
  #     comm2sci(com='black bear', db='itis', simplify = FALSE))[[1]],
  #   "data.frame")
})

test_that("comm2sci fails well", {
  skip_on_cran()

  expect_error(comm2sci(5), "com must be of class character")
  expect_error(comm2sci(list()), "com must be of class character")
  expect_error(comm2sci(mtcars), "com must be of class character")

  expect_error(comm2sci("bear", db = "adsf"),
               "'db' must be one of 'ncbi', 'itis', 'tropicos', 'eol', 'worms'")

  expect_error(comm2sci("bear", simplify = "Asdf"),
               "simplify must be of class logical")

  # no results if itisby is not a valid value, doesn't error though
  expect_equal(length(comm2sci('bear', db='itis', itisby = "asdff")[[1]]), 0)
})

test_that("warn on mismatch 'db'", {
  skip_on_cran()
  vcr::use_cassette("comm2sci_warn_on_db_mismatch", {
    expect_warning(
      comm2sci(
        get_uid("Chironomus riparius", messages = FALSE), db = "itis"))
  })
})

context("get_uid")

test_that("get_uid returns the correct value", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("get_uid", {
    x <- get_uid(c("Chironomus riparius", "aaa"), messages = FALSE)
    z <- get_uid(c("Chironomus riparius", "Chaetopteryx"),
      messages = FALSE)
  })
  expect_true(is.na(x[2]))
  expect_is(z, "uid")
})

test_that("get_uid accepts ask-argument", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("get_uid_ask_arg", {
    x <- sw(get_uid("Dugesia", ask = FALSE, messages = FALSE))
  })
  expect_true(is.na(x))
})

test_that("get_uid query modifiers work", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("get_uid_name_modifiers", {
    mod1 <- get_uid(sci_com = "Aratinga", messages = FALSE)
    mod2 <- get_uid(sci_com = "Aratinga", modifier = "Organism",
      rows = 1, messages = FALSE)
  })

  expect_is(mod1, "uid")
  expect_is(mod2, "uid")
  expect_equal(mod1[[1]], "12945")
  expect_equal(mod2[[1]], "1230190")

  vcr::use_cassette("get_uid_rank_modifiers", {
    rq1 <- get_uid(sci_com = "Bombus", rank_query = "genus", messages = FALSE)
    rq2 <- get_uid(sci_com = "Bombus", rank_query = "subgenus", messages = FALSE)
  })

  expect_is(rq1, "uid")
  expect_is(rq2, "uid")
  expect_equal(rq1[[1]], "28641")
  expect_equal(rq2[[1]], "144708")
})

test_that("get_uid filtering works", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("get_uid_filtering_division", {
    df1 <- get_uid(sci_com = "Echinacea", division_filter = "eudicots",
      messages = FALSE)
    df2 <- get_uid(sci_com = "Echinacea", division_filter = "sea urchins",
      messages = FALSE)
  })

  expect_is(df1, "uid")
  expect_is(df2, "uid")
  expect_equal(df1[[1]], "53747")
  expect_equal(df2[[1]], "7674")

  vcr::use_cassette("get_uid_filtering_rank", {
    rf1 <- get_uid(sci_com = "Bombus", rank_filter = "genus", rows = 2,
      messages = FALSE)
    rf2 <- get_uid(sci_com = "Bombus", rank_filter = "subgenus",
      messages = FALSE)
  })

  expect_is(rf1, "uid")
  expect_is(rf2, "uid")
  expect_equal(rf1[[1]], "28641")
  expect_equal(rf2[[1]], "144708")
})

test_that("get_uid fails as expected", {
  skip_on_cran()

  expect_error(get_uid(), "argument \"sci_com\" is missing")
  expect_error(get_uid("Satyrium", ask = 4, messages = FALSE),
               "ask must be of class logical")
  expect_error(
    get_uid(sci_com = "Aratinga acuticauda", modifier = 5,
            messages = FALSE),
               "modifier must be of class character")
  expect_error(
    get_uid(sci_com = "Pinus", rank_query = TRUE,
            messages = FALSE),
    "rank_query must be of class character")
  expect_error(
    get_uid(sci_com = "Echinacea", division_filter = 4,
            messages = FALSE),
    "division_filter must be of class character")
  expect_error(
    get_uid(sci_com = "Pinus", rank_filter = 34,
            messages = FALSE),
    "rank_filter must be of class character")

  # rows param
  expect_error(get_uid("Achlya", rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_uid("Achlya", rows = 0, messages = FALSE),
               "all\\(rows > 0\\) is not TRUE")
})

context("children")

test_that("children returns the correct values and classes", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("children_with_name", {
    ch_ncbi <- children("Salmo", db = "ncbi")
  })

  expect_is(ch_ncbi, "children")
  expect_equal(attr(ch_ncbi, "db"), "ncbi")
  expect_named(ch_ncbi, "Salmo")
  expect_is(ch_ncbi$Salmo, "data.frame")
  expect_named(ch_ncbi$Salmo, c('childtaxa_id', 'childtaxa_name', 'childtaxa_rank'))
})

test_that("passing in an id works", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("children_with_id", {
    ch_ncbi <- children(8028, db = "ncbi")
    ch_worms <- sw(children(125732, db='worms'))
  })

  expect_is(ch_worms, "children")
  expect_equal(attr(ch_worms, "db"), "worms")
  expect_named(ch_worms, '125732')
  expect_is(ch_worms$`125732`, "data.frame")
  expect_named(ch_worms$`125732`, c('childtaxa_id', 'childtaxa_name', 'childtaxa_rank'))

  expect_is(ch_ncbi, "children")
  expect_equal(attr(ch_ncbi, "db"), "ncbi")
  expect_named(ch_ncbi, '8028')
  expect_is(ch_ncbi$`8028`, "data.frame")
  expect_named(ch_ncbi$`8028`, c('childtaxa_id', 'childtaxa_name', 'childtaxa_rank'))
})

test_that("queries with no results fail well", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("children_no_results", {
    aa <- children("Saurauia", db = "itis", verbose = FALSE)
  })
  expect_equal(NROW(aa[[1]]), 0)
})
# Commented out due to https://github.com/ropensci/vcr/issues/271
# test_that("itis types are correct", {
#   skip_on_cran()
#   itis_expected_col_types <- c(
#     parentname = 'character',
#     parenttsn  = 'character',
#     rankname   = 'character',
#     taxonname  = 'character',
#     tsn        = 'character'
#   )
#   
#   vcr::use_cassette("children_itis_types", {
#     x <- children(1234123434, "itis")
#     z <- children(161994, "itis")
#   })
# 
#   # for no result
#   expect_equal(
#     sapply(x[['1234123434']], class), itis_expected_col_types 
#   )
#   # for good result
#   expect_equal(
#     sapply(z[['161994']], class), itis_expected_col_types 
#   )
# })

test_that("rows parameter, when used, works", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("children_rows_param", {
    x <- children("Asdfafsfd", db = "ncbi", rows = 1, messages = FALSE)
  })

  expect_is(x, "children")
})

test_that("expected results for no query match when using get_* fxns", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("children_no_results_structure_ncbi_x", {
    ncbi_x <- children(get_uid("23424234234", messages = FALSE))
  })

  vcr::use_cassette("children_no_results_structure_ncbi_y", {
    ncbi_y <- children('dragon', db = "ncbi", messages = FALSE)
  })

  expect_named(ncbi_x, NA_character_)
  expect_named(ncbi_y, "dragon")
})

test_that("expected results for no query match when using get_* fxns", {
  skip_on_cran()
  vcr::use_cassette("children_no_results_structure_x", {
    itis_x <- children(get_tsn("23424234234", messages = FALSE))
    worms_x <- children(get_wormsid("23424234234", messages = FALSE))
  })

  vcr::use_cassette("children_no_results_structure_y", {
    itis_y <- children("23424234234", db = "itis", messages = FALSE)
    worms_y <- children(get_wormsid("23424234234", messages = FALSE))
  })
  
  expect_named(itis_x, NA_character_)
  expect_named(itis_y, '23424234234')

  expect_named(worms_x, NA_character_)
  expect_named(worms_y, NA_character_)
})


test_that("children doesn't remove ambiguous taxa", {
  skip_on_cran()
  vcr::use_cassette("children_ambiguous_ncbi", {
    # 28901 = "Salmonella enterica" - DOES NOT remove "subsp."
    subsp <- children('28901', db = "ncbi")
    # 2508041 = "unclassified Helianthus" - DOES NOT remove "sp."
    sp <- children('2508041', db = "ncbi")
  })

  expect_is(subsp, "children")
  expect_is(subsp[[1]], "data.frame")
  expect_is(sp, "children")
  expect_is(sp[[1]], "data.frame")

  expect_gt(NROW(subsp[[1]]), 3)
  expect_gt(NROW(sp[[1]]), 3)
})

test_that("warn on mismatch 'db'", {
  skip_on_cran()
  vcr::use_cassette("children_warn_on_db_mismatch", {
    expect_warning(
      children(
        get_uid("Chironomus riparius", messages = FALSE), db = "itis"))
  })
})

test_that("works with source bold", {
  skip_on_cran()
  vcr::use_cassette("children_bold", {
    x_id <- children("88899", db = "bold", messages = FALSE)
    x_from_get <- children(get_boldid("Momotus", messages = FALSE))
  })

  expect_named(x_id, "88899")
  expect_named(x_from_get, 'species')

  expect_is(x_id, "children")
  expect_is(x_id[[1]], "data.frame")
  expect_is(x_id[[1]]$name, "character")
  expect_match(x_id[[1]]$name, "Momotus")

  expect_is(x_from_get, "children")
  expect_is(x_from_get[[1]], "data.frame")
  expect_is(x_from_get[[1]]$name, "character")
  expect_match(x_from_get[[1]]$name, "Momotus")
})

test_that("works with worms, passing through marine_only", {
  skip_on_cran()
  vcr::use_cassette("children_worms_marine_only", {
    a <- children("1837", db = "worms")
    b <- children("1837", db = "worms", marine_only = FALSE)
  })

  expect_is(a, "children")
  expect_is(b, 'children')
  expect_gt(NROW(b[[1]]), NROW(a[[1]]))
})

#' Search Catalogue of Life for for direct children of a particular taxon.
#'
#' @export
#' @param name The string to search for. Only exact matches found the name given
#' 		will be returned, unless one or wildcards are included in the search
#'   	string. An * (asterisk) character denotes a wildcard; a % (percentage)
#'    character may also be used. The name must be at least 3 characters long,
#'    not counting wildcard characters.
#' @param id The record ID of the specific record to return (only for scientific
#' 		names of species or infraspecific taxa)
#' @param format format of the results returned. Valid values are format=xml and
#' 		format=php; if the format parameter is omitted, the results are returned
#'   	in the default XML format. If format=php then results are returned as a
#'    PHP array in serialized string format, which can be converted back to an
#'    array in PHP using the unserialize command
#' @param start The first record to return. If omitted, the results are returned
#' 		from the first record (start=0). This is useful if the total number of
#' 		results is larger than the maximum number of results returned by a single
#' 		Web service query (currently the maximum number of results returned by a
#' 		single query is 500 for terse queries and 50 for full queries).
#' @param checklist The year of the checklist to query, if you want a specific
#' 		year's checklist instead of the lastest as default (numeric).
#' @param ... Curl options passed on to \code{\link[httr]{GET}}
#' @details You must provide one of name or id. The other parameters (format
#' 		and start) are optional.
#' @return A list of data.frame's.
#' @examples \dontrun{
#' # A basic example
#' col_children(name="Apis")
#'
#' # An example where there is no classification, results in data.frame with
#' # no rows
#' col_children(id='b2f88f382aa5568f93a97472c6be6516')
#'
#' # Use a specific year's checklist
#' col_children(name="Apis", checklist=2012)
#' col_children(name="Apis", checklist=2009)
#'
#' # Pass in many names or many id's
#' out <- col_children(name=c("Buteo","Apis","Accipiter","asdf"),
#'   checklist = "2012")
#' out$Apis # get just the output you want
#' library("plyr")
#' ldply(out) # or combine to one data.frame
#'
#' # or pass many id's
#' ids <- c('abe977b1d27007a76dd12a5c93a637bf',
#'   'b2f88f382aa5568f93a97472c6be6516')
#' out <- col_children(id = ids, checklist=2012)
#' library("plyr")
#' ldply(out) # combine to one data.frame
#' }
col_children <- function(name = NULL, id = NULL, format = NULL, start = NULL,
                         checklist = NULL, ...) {

	if (is.null(id)) {
		temp <- llply(name, search_col_safe, id = NULL, checklist = checklist,
		              format = format, start = start, ...)
		setNames(temp, name)
	} else {
		temp <- llply(id, search_col_safe, name = NULL, checklist = checklist,
		              format = format, start = start, ...)
		setNames(temp, id)
	}
}

search_col <- function(name, id, checklist, format, start, ...) {
  url <- make_url(checklist)
  args <- tc(list(name = name, id = id, format = format, response = "full",
                  start = start))
  out <- GET(col_base(), query = argsnull(args), ...)
  stop_for_status(out)
  tt <- xml2::read_xml(con_utf8(out))
  search_col_child_df(tt)
}
search_col_safe <- plyr::failwith(NULL, search_col)

search_col_child_df <- function(x) {
  childtaxa_id <- xml_text(xml_find_all(x, "//child_taxa//id"))
  childtaxa_name <- xml_text(xml_find_all(x, "//child_taxa//name"))
  childtaxa_rank <- xml_text(xml_find_all(x, "//child_taxa//rank"))
  data.frame(childtaxa_id, childtaxa_name,
             childtaxa_rank = tolower(childtaxa_rank),
             stringsAsFactors = FALSE)
}

# tests for col_children fxn in taxize
context("col_children")

test_that("col_children returns the correct classes and dimensions", {
  skip_on_cran()

  temp4 <- col_children(name="Animalia")
  temp5 <- col_children(name="Plantae")
  temp11 <- col_children(name="Accipiter striatus")
  temp13 <- col_children(id=c(2346405,2344165,2346405), checklist = '2012')

	expect_is(temp4, "list")
  expect_is(temp5, "list")
  expect_is(temp11, "list")
  expect_is(temp4$Animalia, "data.frame")
  expect_is(temp5$Plantae, "data.frame")
  expect_is(temp11$`Accipiter striatus`, "data.frame")
  expect_is(temp13$`2346405`, "data.frame")

	expect_equal(NCOL(temp4[[1]]), 3)
	expect_equal(NCOL(temp5[[1]]), 3)

	expect_equal(length(temp11), 1)
})

test_that("missing/wrong data given returns result", {
  skip_on_cran()

  expect_equal(nrow(col_children(name="")[[1]]), 0)
  expect_equal(nrow(col_children(name="asdfasdfdf")[[1]]), 0)
  expect_is(col_children(), "list")
})

#' Get The Plant List families.
#'
#' @export
#' @param ... (list) Curl options passed on to [crul::verb-GET]
#' @details Requires an internet connection in order to connect to 
#' <www.theplantlist.org>.
#' @return Returns a `data.frame` including the names of all families
#' indexed by The Plant List, and the major groups into which they fall 
#' (i.e. Angiosperms, Gymnosperms, Bryophytes and Pteridophytes).
#' @author John Baumgartner (johnbb@@student.unimelb.edu.au)
#' @seealso [tpl_get()]
#' @examples \dontrun{
#' # Get a data.frame of plant families, with the group name 
#' # (Angiosperms, etc.)
#' head(tpl_families())
#' }
tpl_families <- function(...) {
  cli <- crul::HttpClient$new('http://www.theplantlist.org/1.1/browse/-/',
    headers = tx_ual, opts = list(...))
  temp <- cli$get()
  temp$raise_for_status()
  temp <- xml2::read_html(temp$parse("UTF-8"), encoding = "UTF-8")
  families <- xml2::xml_text(
    xml2::xml_find_all(temp, "//ul[@id='nametree']//a"))
  groups <- as.character(factor(basename(dirname(
    xml2::xml_attr(
      xml2::xml_find_all(temp, "//ul[@id='nametree']//a"), "href"))),
        levels = c('A', 'B', 'G', 'P'),
        labels = c('Angiosperms', 'Bryophytes',
          'Gymnosperms', 'Pteridophytes')))
  data.frame(group = groups, family = families, 
    stringsAsFactors = FALSE)
}

#' ION - Index to Organism Names
#'
#' @export
#' @param x An LSID number. Required.
#' @param ... Curl options passed on to [crul::verb-GET]
#' @references http://www.organismnames.com
#' @return A data.frame
#' @examples \dontrun{
#' ion(155166)
#' ion(298678)
#' ion(4796748) # ursus americanus
#' ion(1280626) # puma concolor
#' }
ion <- function(x, ...) {
  cli <- crul::HttpClient$new(ion_base(), headers = tx_ual, opts = list(...))
  res <- cli$get(query = list(lsid = x))
  res$raise_for_status()
  xml <- xml2::read_xml(res$parse("UTF-8"))
  dc <- as.list(sapply(c('identifier', 'Title'), function(z) {
    xml_text(xml_find_all(xml, paste0("//dc:", z), xml_ns(xml)))
  }))
  tdwg <- as.list(
    stats::setNames(
      xml_text(xml_find_all(xml, "//tdwg_tn:nameComplete", xml_ns(xml))),
      "nameComplete")
  )
  df <- data.frame(c(dc, tdwg), stringsAsFactors = FALSE)
  stats::setNames(df, tolower(names(df)))
}

ion_base <- function() 'http://www.organismnames.com/lsidmetadata.htm'

context("ion")

test_that("ion returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("ion", {
    aa <- ion(155166)
    bb <- ion(298678)
    cc <- ion(4796748) # ursus americanus
    dd <- ion(1280626) # puma concolor
  })

  expect_is(aa, 'data.frame')
  expect_is(bb, 'data.frame')
  expect_is(cc, 'data.frame')
  expect_is(dd, 'data.frame')

  expect_named(aa, c('identifier', 'title', 'namecomplete'))
})

test_that("ion fails well", {
  skip_on_cran()
  vcr::use_cassette("ion_fails_well", {
    expect_error(ion(2343434434434), class = "error")
    expect_error(ion("asdfasfs"), class = "error")
  })
  
  skip_on_cran()
  expect_error(ion(), "argument \"x\" is missing")
})

make_generic <- function(x, uu, clz, check = TRUE) {
  if (check) {
    if ( evalfxn(clz)(x) ) {
      toid(x, uu, clz)
    } else {
      structure(
        NA, class = clz, match = "not found",
        multiple_matches = FALSE, pattern_match = FALSE, uri = NA
      )
    }
  } else {
    toid(x, uu, clz)
  }
}

make_wiki_generic <- function(x, uu, clz, check = TRUE) {
  if (check) {
    if ( evalfxn(clz)(sprintf(uu, x)) ) {
      toid(x, uu, clz)
    } else {
      structure(
        NA, class = clz, match = "not found",
        multiple_matches = FALSE, pattern_match = FALSE, uri = NA
      )
    }
  } else {
    toid(x, uu, clz)
  }
}

# messages
tx_msg_not_found <-
  "Not found. Consider checking the spelling or alternate classification"

# url templates for uri attributes of get_* functions
get_url_templates <- list(
  gbif = "https://www.gbif.org/species/%s",
  pow = "https://powo.science.kew.org/taxon/%s",
  tol = "https://tree.opentreeoflife.org/opentree/argus/ottol@%s",
  tropicos = "http://tropicos.org/Name/%s",
  worms = "http://www.marinespecies.org/aphia.php?p=taxdetails&id=%s",
  ncbi = "https://www.ncbi.nlm.nih.gov/taxonomy/%s",
  eol = "https://eol.org/pages/%s/",
  bold = "http://boldsystems.org/index.php/Taxbrowser_Taxonpage?taxid=%s",
  itis = "https://www.itis.gov/servlet/SingleRpt/SingleRpt?search_topic=TSN&search_value=%s",
  nbn = "https://species.nbnatlas.org/species/%s",
  natserv = "https://explorer.natureserve.org/Taxon/ELEMENT_GLOBAL.2.%s",
  iucn = iucn_base_url
)

#' Get citations and licenses for data sources used in taxize
#'
#' @export
#' @param fxn Function to search on. A special case is the package name
#' 'taxize' that will give the citations for the package.
#' @param what One of citation (default), license, or both.
#' @examples
#' taxize_cite(fxn='eol_search')
#' taxize_cite(fxn='itis_hierarchy')
#' taxize_cite(fxn='tp_classification')
#' taxize_cite(fxn='gbif_ping')
#' taxize_cite(fxn='plantminer')
#' taxize_cite(fxn='get_natservid_')
#' taxize_cite(fxn='as.natservid')
#' taxize_cite(fxn='get_wormsid')
#' taxize_cite(fxn='as.wormsid')
#'
#' # Functions that use many data sources
#' taxize_cite(fxn='synonyms')
#' taxize_cite(fxn='classification')
#'
#' # Get the taxize citation
#' taxize_cite(fxn='taxize')
#'
#' # Get license information
#' taxize_cite(fxn='taxize', "license")

taxize_cite <- function(fxn = "itis", what = 'citation'){
  what <- match.arg(what, c("citation","license","both"))
  if (what == "citation") {
    out <- data_citations(fxn)
    if (is.null(out)) {
      message(paste0(
        "Nothing found, try different inputs\n\n",
        "Can this citation be improved? https://github.com/ropensci/taxize/issues\n",
        "Please cite taxize in your paper: citation(package = 'taxize')"
      ))
    } else {
      if (fxn == 'taxize') {
        message("The paper:")
        message(format(out[[1]]))
        message("\n")
        message("The software:")
        message(format(out[[2]]))
      } else {
        for (i in seq_along(out)) {
          message(sprintf("Source: %s", names(out)[i]), "\n")
          message(sprintf("  Home page: %s", out[[i]]$url_home), "\n")
          message(sprintf("  API help: %s", out[[i]]$apidocs), "\n")
          message(sprintf("  Citation: %s", out[[i]]$citation), "\n\n")
        }
        message("Can any of these citations be improved? https://github.com/ropensci/taxize/issues")
      }
    }
  } else if (what == "license") {
    out <- data_licenses(fxn)
    if (is.null(out)) {
      message("Unknown...")
    } else{
      message(sprintf("License: %s", out$license), "\n")
      message(sprintf("URL:     %s", out$url))
    }
  }
}

data_citations <- function(x){
  switch(x,
    itis_acceptname = list(itis = c_itis),
    itis_downstream = list(itis = c_itis),
    itis_getrecord = list(itis = c_itis),
    itis_hierarchy = list(itis = c_itis),
    itis_kingdomnames = list(itis = c_itis),
    itis_lsid = list(itis = c_itis),
    itis_name = list(itis = c_itis),
    itis_native = list(itis = c_itis),
    itis_ping = list(itis = c_itis),
    itis_refs = list(itis = c_itis),
    itis_taxrank = list(itis = c_itis),
    itis_terms = list(itis = c_itis),
    apg_lookup = list(apg = c_apg),
    as.boldid = list(bold = c_bold),
    as.eolid = list(eol = c_eol),
    as.gbifid = list(gbif = c_gbif),
    as.natservid = list(natserv = c_natureserve),
    as.nbnid = list(nbn = c_nbn),
    as.tpsid = list(tropicos = c_tropicos),
    as.tsn = list(itis = c_itis),
    as.uid = list(ncbi = c_ncbi),
    as.wormsid = list(worms = c_worms),
    bold_ping = list(bold = c_bold),
    bold_search = list(bold = c_bold),
    cbind.classification = list(none = c_none),
    cbind.classification_ids = list(none = c_none),
    children = list(itis = c_itis, ncbi = c_ncbi, worms = c_worms),
    class2tree = list(none = c_none),
    classification = list(itis = c_itis, ncbi = c_ncbi,
                          gbif = c_gbif, eol = c_eol, troicos = c_tropicos,
                          nbn = c_nbn, worms = c_worms, natserv = c_natureserve),
    comm2sci = list(c_itis, c_ncbi, c_eol, c_tropicos, worms = c_worms),
    downstream = list(itis = c_itis),
    eol_dataobjects = list(eol = c_eol),
    eol_hierarchy = list(eol = c_eol),
    eol_invasive = list(eol = c_eol),
    eol_pages = list(eol = c_eol),
    eol_ping = list(eol = c_eol),
    eol_search = list(eol = c_eol),
    gbif_name_usage = list(gbif = c_gbif),
    gbif_parse = list(gbif = c_gbif),
    gbif_ping = list(gbif = c_gbif),
    genbank2uid = list(genbank = c_genbank),
    get_boldid = list(bold = c_bold),
    get_boldid_ = list(bold = c_bold),
    get_eolid = list(eol = c_eol),
    get_eolid_ = list(eol = c_eol),
    get_gbifid = list(gbif = c_gbif),
    get_gbifid_ = list(gbif = c_gbif),
    get_ids = list(itis = c_itis, ncbi = c_ncbi, gbif = c_gbif, eol = c_eol,
                   tropicos = c_tropicos, nbn = c_nbn),
    get_ids_ = list(itis = c_itis, ncbi = c_ncbi, gbif = c_gbif, eol = c_eol,
                    tropicos = c_tropicos, nbn = c_nbn),
    get_natservid = list(natserv = c_natureserve),
    get_natservid_ = list(natserv = c_natureserve),
    get_nbnid = list(nbn = c_nbn),
    get_nbnid_ = list(nbn = c_nbn),
    get_seqs = list(genbank = c_genbank),
    get_tpsid = list(tropicos = c_tropicos),
    get_tpsid_ = list(tropicos = c_tropicos),
    get_tsn = list(itis = c_itis),
    get_tsn_ = list(itis = c_itis),
    get_uid = list(ncbi = c_ncbi),
    get_uid_ = list(ncbi = c_ncbi),
    get_wormsid = list(worms = c_worms),
    get_wormsid_ = list(worms = c_worms),
    gisd_isinvasive = list(gisd = c_gisd),
    gni_details = list(gni = c_gni),
    gni_parse = list(gni = c_gni),
    gna_search = list(gni = c_gni),
    gnr_datasources = list(global_names = c_gnames),
    gnr_resolve = list(global_names = c_gnames),
    iplant_resolve = list(iplant = c_iplant),
    ipni_ping = list(ipni = c_ipni),
    ipni_search = list(ipni = c_ipni),
    iucn_getname = list(iucn = c_iucn),
    iucn_status = list(iucn = c_iucn),
    iucn_summary = list(iucn = c_iucn),
    names_list = list(theplantlist = c_plist),
    nbn_classification = list(nbn = c_nbn),
    nbn_ping = list(nbn = c_nbn),
    nbn_search = list(nbn = c_nbn),
    nbn_synonyms = list(nbn = c_nbn),
    ncbi_children = list(ncbi = c_ncbi),
    ncbi_get_taxon_summary = list(ncbi = c_ncbi),
    ncbi_getbyid = list(genbank = c_genbank),
    ncbi_getbyname = list(genbank = c_genbank),
    ncbi_ping = list(ncbi = c_ncbi),
    ncbi_search = list(genbank = c_genbank),
    plantminer = list(plantminer = c_plantminer),
    rankagg = list(none = c_none),
    rbind.classification = list(none = c_none),
    rbind.classification_ids = list(none = c_none),
    resolve = list(global_names = c_gnames, iplant = c_iplant),
    sci2comm = list(itis = c_itis, ncbi = c_ncbi, eol = c_eol, worms = c_worms),
    scrapenames = list(global_names = c_gnames),
    synonyms = list(itis = c_itis, tropicos = c_tropicos,
                    nbn = c_nbn, worms = c_worms),
    tax_agg = list(itis = c_itis, ncbi = c_ncbi),
    tax_name = list(itis = c_itis, ncbi = c_ncbi),
    tax_rank = list(itis = c_itis, ncbi = c_ncbi),
    tp_acceptednames = list(tropicos = c_tropicos),
    tp_accnames = list(tropicos = c_tropicos),
    tp_classification = list(tropicos = c_tropicos),
    tp_dist = list(tropicos = c_tropicos),
    tp_namedistributions = list(tropicos = c_tropicos),
    tp_namereferences = list(tropicos = c_tropicos),
    tp_refs = list(tropicos = c_tropicos),
    tp_search = list(tropicos = c_tropicos),
    tp_summary = list(tropicos = c_tropicos),
    tp_synonyms = list(tropicos = c_tropicos),
    tpl_families = list(theplantlist = c_plist),
    tpl_get = list(theplantlist = c_plist),
    tpl_search = list(theplantlist = c_plist),
    tropicos_ping = list(theplantlist = c_plist),
    upstream = list(c_itis),
    vascan_ping = list(vascan = c_vascan),
    vascan_search = list(vascan = c_vascan),
    taxize = citation("taxize")
  )
}

data_licenses <- function(x){
  switch(x,
    taxize = list(license = "MIT", url = "https://opensource.org/licenses/MIT")
  )
}

c_itis_url <- "https://www.itis.gov/"
c_itis_api <- "https://www.itis.gov/ws_description.html"
c_itis_citation <- "Retrieved [month, day, year], from the Integrated Taxonomic Information System on-line database, https://www.itis.gov."
c_itis <- list(url_home = c_itis_url, apidocs = c_itis_api, citation = c_itis_citation)
c_eol <- list(url_home = 'http://eol.org/', apidocs = 'http://eol.org/api/')
c_gbif = list(url_home = 'http://www.gbif.org', apidocs = 'http://www.gbif.org/developer/summary',
              citation = 'GBIF (2013). GBIF (Ed.), Global Biodiversity Information Facility Data Portal (2013)',
              dataset_citation = 'http://www.gbif.org/resources/2381')
c_tropicos <- list(url_home = "http://tropicos.org/", apidocs = "http://services.tropicos.org/")
c_plist <- list(url_home = "http://theplantlist.org/", apidocs = NULL,
                citation = "The Plant List (2013). Version 1.1. Published on the Internet; http://www.theplantlist.org/ ([accessed <date>).")
c_vascan <- list(url_home = "http://data.canadensys.net/explorer/en/search", apidocs = "http://data.canadensys.net/vascan/api",
                 citation = "Brouillet, L., F. Coursol, S.J. Meades, M. Favreau, M. Anions, P. Belisle & P. Desmet. 2010+. VASCAN, the Database of Vascular Plants of Canada. http://data.canadensys.net/vascan/ (consulted on <date>)")
c_plantminer <- list(url_home = "http://www.plantminer.com/", apidocs = "http://www.plantminer.com/help",
                     citation = 'See The Plant List or Tropicos citations')
c_nbn <- list(url_home = "http://www.nbn.org.uk/", apidocs = "https://data.nbn.org.uk/Documentation/Web_Services/",
              citation = NULL)
c_ncbi <- list(url_home = "http://www.ncbi.nlm.nih.gov/taxonomy", apidocs = "http://www.ncbi.nlm.nih.gov/books/NBK25501/",
               citation = "Federhen S: The NCBI Taxonomy database. Nucleic Acids Res 2012, 40 (Database issue): D136-D143.")
c_genbank <- list(url_home = "http://www.ncbi.nlm.nih.gov/genbank/", apidocs = "http://www.ncbi.nlm.nih.gov/books/NBK25501/",
                  citation = 'Bilofsky, Howard S., and Burks Christian. "The GenBank genetic sequence data bank." Nucleic acids research 16.5 (1988): 1861-1863.')
c_iplant <- list(url_home = "http://tnrs.iplantcollaborative.org/", apidocs = "http://tnrs.iplantcollaborative.org/api.html",
              citation = 'Boyle, B. et al. 2013. The taxonomic name resolution service: an online tool for automated standardization of plant names. BMC bioinformatics 14:16. doi:10.1186/1471-2105-14-16')
c_gnames <- list(url_home = "http://gnrd.globalnames.org/", apidocs = "http://gnrd.globalnames.org/api", citation = NULL)
c_gni <- list(url_home = "http://gni.globalnames.org/", apidocs = "https://github.com/dimus/gni/wiki/api", citation = NULL)
c_bold <- list(url_home = "http://www.boldsystems.org/", apidocs = "http://www.boldsystems.org/index.php/resources/api",
               citation = "Ratnasingham, S., & Hebert, P. D. (2007). BOLD: The Barcode of Life Data System (http://www. barcodinglife. org). Molecular ecology notes, 7(3), 355-364.")
c_iucn <- list(url_home = "https://www.iucn.org/", apidocs = NULL,
               citation = "IUCN 2015. The IUCN Red List of Threatened Species. Version 2015.1. <https://www.iucnredlist.org>. Downloaded on <date>.")
c_ipni <- list(url_home = "http://www.ipni.org/", apidocs = "http://www.ipni.org/ipni/advPlantNameSearch.do",
               citation = "The International Plant Names Index (<year>). Published on the Internet http://www.ipni.org [accessed <date>]")
c_apg <- list(url_home = "http://www.mobot.org/MOBOT/research/APweb/", apidocs = NULL,
              citation = "Stevens, P. F. (<year>). Angiosperm Phylogeny Website. Version 13, July 2012.")
c_gisd <- list(url_home = "http://www.issg.org/database/welcome/", apidocs = NULL, citation = NULL)
c_none <- list(url_home = "no data source", apidocs = "no data source", citation = "no data source")
c_worms <- list(url_home = "http://www.marinespecies.org/", apidocs = "http://www.marinespecies.org/rest/", citation = "We ask you to cite the individual global or regional species lists, or species pages as appropriate. Their citations are shown on their web pages. The database as a whole is to be cited as follows:\n\n     WoRMS Editorial Board (2017). World Register of Marine Species. Available from http://www.marinespecies.org at VLIZ. Accessed <date>. doi:10.14284/170")
c_natureserve <- list(url_home = "http://www.natureserve.org/", apidocs = "https://services.natureserve.org/index.jsp", citation = "Citation: Natureserve. 2017. NatureServe Web Service. Arlington, VA. U.S.A. Available http://services.natureserve.org. (Accessed: <date>)")

#' Check invasive species status for a set of species from GISD database
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @rdname gisd_invasive-defunct
#' @keywords internal
gisd_isinvasive <- function(...) {
  .Defunct("gisd", "originr", msg = "This function is defunct. See originr::gisd()")
}

#' @title Find taxon names using Global Names Recognition and Discovery
#'
#' @description Uses the Global Names Recognition and Discovery service, see
#'   http://gnrd.globalnames.org/
#'
#'   NOTE: This function sometimes gives data back and sometimes not. The API
#'   that this function is using is extremely buggy.
#'
#' @export
#'
#' @param url (character) If text parameter is empty, and `url` is given,
#'   GNfinder will process the URL and will find names in the content of its
#'   body.
#' @param text (character) Contains the text which will be checked for
#'   scientific names. If this parameter is not empty, the `url` parameter is
#'   ignored.
#' @param format (character) Sets the output format. It can be set to: `"csv"`
#'   (the default), `"tsv"`, or `"json"`.
#' @param bytes_offset (logical) This changes how the position of a detected
#'   name in text is calculated. Normally a name's start and end positions are
#'   given as the number of UTF-8 characters from the beginning of the text. If
#'   this is `TRUE`, the start and end offsets are recalculated in the number of
#'   bytes.
#' @param return_content (logical) If this is `TRUE`, the text used for the name
#'   detection is returned back. This is especially useful if the input was not
#'   a plain UTF-8 text and had to be prepared for name-finding. Then the
#'   returned content can be used together with start and end fields of detected
#'   name-strings to locate the strings in the text.
#' @param unique_names (logical) If this is `TRUE`, the output returns a list of
#'   unique names, instead of a list of all name occurrences. Unique list of
#'   names does not provide position information of a name in the text.
#' @param ambiguous_names (logical) If this is `TRUE`, strings which are
#'   simultaneously scientific names and "normal" words are not filtered out
#'   from the results. For example, generic names like America, Cancer,
#'   Cafeteria will be returned in the results.
#' @param no_bayes (logical) If this is `TRUE`, only heuristic algorithms are
#'   used for name detection.
#' @param odds_details (logical) If `TRUE`, the result will contain odds of all
#'   features used for calculation of NaiveBayes odds. Odds describe probability
#'   of a name to be 'real'. The higher the odds, the higher the probability
#'   that a detected name is not a false positive. Odds are calculated by
#'   multiplication of the odds of separate features. Odds details explain how
#'   the final odds value is calculated.
#' @param language (character) The language of the text. Language value is used
#'   for calculation of Bayesian odds. If this parameter is not given, `"eng"`
#'   is used by default. Currently only English and German languages are
#'   supported. Valid values are: `"eng"`, `"deu"`, and `"detect"`.
#' @param words_around (integer) Allows to see the context surrounding a
#'   name-string. This sets the number of words located immediately before or
#'   after a detected name. These words are then returned in the output. Default
#'   is 0, maximum value is 5.
#' @param verification (character) When this `TRUE`, there is an additional
#'   verification step for detected names. This step requires internet
#'   connection and uses https://verifier.globalnames.org/api/v1 for
#'   verification queries.
#' @param sources Pipe separated list of data source ids to resolve found names
#'   against. See list of Data Sources
#'   http://resolver.globalnames.org/data_sources
#' @param all_matches When this option is true all found results are returned,
#'   not only the bestResults. The bestResult field in this case is null, and
#'   results field should contain found results of the matches.
#' @param ... Further args passed to [crul::verb-GET]
#' @param detect_language Defunct. See the `language` option.
#' @param data_source_ids Defunct. See the `sources` option.
#' @param file Defunct. If you feel this is important functionality submit an
#'   issue at "https://github.com/ropensci/taxize"
#' @param unique Defunct. See the `unique_names` option.
#' @param engine Defunct. The API used no longer supports this option.
#'
#' @author Scott Chamberlain, Zachary Foster
#'
#' @return A [tibble::tibble()] or list representing parsed JSON output
#'   depending on the value of the `format` option.
#' @examples \dontrun{
#' # Get data from a website using its URL
#' scrapenames('https://en.wikipedia.org/wiki/Spider')
#' scrapenames('https://en.wikipedia.org/wiki/Animal')
#' scrapenames('https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0095068')
#' scrapenames('https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0080498')
#'
#' scrapenames(url = 'https://en.wikipedia.org/wiki/Spider', source=c(1, 169))
#'
#' # Get data from text string
#' scrapenames(text='A spider named Pardosa moesta Banks, 1892')
#'
#' # return OCR content
#' scrapenames(text='A spider named Pardosa moesta Banks, 1892',
#'             return_content = TRUE, format = 'json')
#' }
scrapenames <- function(
    url = NULL,
    text = NULL,
    format = 'csv',
    bytes_offset = FALSE,
    return_content = FALSE,
    unique_names = TRUE,
    ambiguous_names = FALSE,
    no_bayes = FALSE,
    odds_details = FALSE,
    language = 'detect',
    words_around = 0,
    verification = TRUE,
    sources = NULL,
    all_matches = FALSE,
     ...,
    file = NULL,
    unique = NULL,
    engine = NULL,
    detect_language = NULL,
    data_source_ids = NULL
) {
  
  # Error if defunct parameters are used.
  if (!is.null(unique)) {
    stop(call. = FALSE, 'The `unique` option is defunct. See the `unique_names` option. ')
  }
  if (!is.null(engine)) {
    stop(call. = FALSE, 'The `engine` option is defunct. The API no longer supports this option. ')
  }
  if (!is.null(detect_language)) {
    stop(call. = FALSE, 'The `detect_language` option is defunct. See the `language` option. ')
  }
  if (!is.null(data_source_ids)) {
    stop(call. = FALSE, 'The `data_source_ids` option is defunct. See the `source` option. ')
  }
  
  # Validate parameters
  if (! format %in% c('csv', 'tsv', 'json')) {
    stop(call. = FALSE, 'The `format` option must be "csv", "tsv", or "json". "', format, '" was the value given')
  }
  
  # Make query
  base <- "http://gnrd.globalnames.org/api/v1/find"
  args <- tc(list(
    text = text,
    url = url,
    format = format,
    bytesOffset = bytes_offset,
    returnContent = return_content,
    uniqueNames = unique_names,
    ambiguousNames = ambiguous_names,
    noBayes = no_bayes,
    oddsDetails = odds_details,
    language = language,
    wordsAround = words_around,
    verification = verification,
    sources = paste0(sources, collapse = '|'),
    allMatches = all_matches
  ))
  cli <- crul::HttpClient$new(base, headers = tx_ual, opts = list(...))
  response <- cli$post(body = args, encode = "form")
  
  # Check for errors
  if (response$status_code == "500") {
    warning(call. = FALSE, 'The GNR server has encountered an internal error trying to process this request.')
    return(NULL)
  }
  
  # Parse and return results
  raw_output <- response$parse("UTF-8")
  switch (format,
    csv = tibble::as_tibble(utils::read.csv(text = raw_output)),
    tsv = tibble::as_tibble(utils::read.csv(text = raw_output, sep = '\t')),
    json = jsonlite::fromJSON(raw_output),
    other = stop("Invalid 'format' option.")
  )
}

#' Search for names in the International Plant Names Index (IPNI).
#'
#' Note: This data source is also provided in the Global Names Index (GNI)
#' (http://gni.globalnames.org/data_sources). The interface to the data is
#' different among the two services though.
#'
#' @export
#' @param family Family name to search on (Optional)
#' @param infrafamily Infrafamilial name to search on (Optional)
#' @param genus Genus name to search on (Optional)
#' @param infragenus Infrageneric name to search on (Optional)
#' @param species Species name to search on (Optional) - Note, this is the
#' epithet, not the full genus - epithet name combination.
#' @param infraspecies Infraspecies name to search on (Optional)
#' @param publicationtitle Publication name or abbreviation to search on.
#' Again, replace any spaces with a '+' (e.g. 'J.+Bot.') (Optional)
#' @param authorabbrev Author standard form to search on (publishing author,
#' basionym author or both - see below) (Optional)
#' @param includepublicationauthors TRUE (default) to include the taxon author
#' in the search or FALSE to exclude it
#' @param includebasionymauthors TRUE (default) to include the basionum author
#' in the search or FALSE to exclude it
#' @param geounit Country name or other geographical unit to search on (see
#' the help pages for more information and warnings about the use of this
#' option) (Optional)
#' @param addedsince Date to search on in the format 'yyyy-mm-dd', e.g.
#' 2005-08-01 for all records added since the first of August, 2005. (see the
#' help pages for more information and warnings about the use of this option)
#' (Optional. If supplied must be in format YYYY-MM-DD and must be greater
#' than or equal to 1984-01-01.)
#' @param modifiedsince Date to search on in the format 'yyyy-mm-dd', e.g.
#' 2005-08-01 for all records edited since the first of August, 2005. (See
#' the help pages for more information about the use of this option)
#' (Optional. If supplied must be in format YYYY-MM-DD and must be greater
#' than or equal to 1993-01-01.)
#' @param isapnirecord FALSE (default) to exclude records from the Australian
#' Plant Name Index
#' @param isgcirecord FALSE (default) to exclude records from the Gray Cards
#' Index
#' @param isikrecord FALSE (default) to exclude records from the Index Kewensis
#' @param ranktoreturn One of a few options to choose the ranks returned. See
#' details.
#' @param output One of minimal (default), classic, short, or extended
#' @param ... Curl options passed on to [crul::verb-GET]
#' (Optional). Default: returns all ranks.
#' @references
#' https://web.archive.org/web/20190501132148/http://www.ipni.org/link_to_ipni.html
#' @details
#' `ranktoreturn` options:
#'
#' * "all" - all records
#' * "fam" - family records
#' * "infrafam" - infrafamilial records
#' * "gen" - generic records
#' * "infragen" - infrageneric records
#' * "spec" - species records
#' * "infraspec" - infraspecific records
#'
#' @return a tibble (data.frame)
#' @examples \dontrun{
#' ipni_search(genus='Brintonia', isapnirecord=TRUE, isgcirecord=TRUE,
#'   isikrecord=TRUE)
#' ipni_search(genus='Ceanothus')
#' ipni_search(genus='Pinus', species='contorta')
#'
#' # Different output formats
#' ipni_search(genus='Ceanothus')
#' ipni_search(genus='Ceanothus', output='short')
#' ipni_search(genus='Ceanothus', output='extended')
#' }

ipni_search <- function(family=NULL, infrafamily=NULL, genus=NULL,
  infragenus=NULL, species=NULL, infraspecies=NULL, publicationtitle=NULL,
  authorabbrev=NULL, includepublicationauthors=NULL,
  includebasionymauthors=NULL, geounit=NULL, addedsince=NULL,
  modifiedsince=NULL, isapnirecord=NULL, isgcirecord=NULL, isikrecord=NULL,
  ranktoreturn=NULL, output="minimal", ...) {

  output <- match.arg(output, c('minimal','classic','short','extended'), FALSE)
  output_format <- sprintf('delimited-%s', output)
  url <- "https://www.ipni.org/ipni/advPlantNameSearch.do"
  args <- tc(list(output_format=output_format, find_family=family,
          find_infrafamily=infrafamily, find_genus=genus,
          find_infragenus=infragenus, find_species=species,
          find_infraspecies=infraspecies,
          find_publicationTitle=publicationtitle,
          find_authorAbbrev=authorabbrev,
          find_includePublicationAuthors=l2(includepublicationauthors),
          find_includebasionymauthors=l2(includebasionymauthors),
          find_geounit=geounit,
          find_addedSince=addedsince, find_modifiedSince=modifiedsince,
          find_isAPNIRecord=l2(isapnirecord),
          find_isGCIRecord=l2(isgcirecord),
          find_isIKRecord=l2(isikrecord), find_rankToReturn=ranktoreturn))
  cli <- crul::HttpClient$new(url, headers = tx_ual, opts = list(...))
  tt <- cli$get(query = args)
  if (
    tt$status_code > 200 ||
    tt$response_headers$`content-type` != "text/plain;charset=UTF-8"
  ) {
    stop("No results", call. = FALSE)
  }
  res <- tt$parse("UTF-8")
  df1 <- read.delim(text = res, sep = "%", stringsAsFactors = FALSE)
  if (NROW(df1) == 0) {
    warning("No data found")
    df <- tibble::tibble()
  } else {
    names(df1) <- gsub("\\.", "_", tolower(names(df1)))
    df <- tibble::as_tibble(df1)
  }
  return(df)
}

l2 <- function(x) {
  if (!is.null(x)) {
    if (x) {
      "on"
    } else {
      "off"
    }
  } else{
    NULL
  }
}

context("ipni_search")

test_that("ipni_search works", {
  skip_on_cran()
  vcr::use_cassette("ipni_search", {
      aa <- ipni_search(genus = 'Brintonia', isapnirecord = TRUE,
                        isgcirecord = TRUE, isikrecord = TRUE)
      # cc <- ipni_search(genus = 'Pinus', species = 'contorta')
  })

  expect_is(aa, "data.frame")
  # expect_is(cc, "data.frame")
  expect_is(aa$id, "character")
  expect_is(aa$family, "character")

  expect_named(aa, c('id','version','family','full_name_without_family_and_authors','authors'))
  expect_is(aa$family, "character")
})

test_that("ipni_search works with different output formats", {
  skip_on_cran()
  vcr::use_cassette("ipni_search_output_formats", {
    aa <- ipni_search(genus = 'Brintonia')
    bb <- ipni_search(genus = 'Brintonia', output = 'short')
    cc <- ipni_search(genus = 'Brintonia', output = 'classic')
    dd <- ipni_search(genus = 'Brintonia', output = 'extended')
  })

  expect_is(aa, "data.frame")
  expect_is(bb, "data.frame")
  expect_is(cc, "data.frame")
  expect_is(dd, "data.frame")

  expect_lt(NCOL(aa), NCOL(bb))
  expect_lt(NCOL(bb), NCOL(cc))
  expect_lt(NCOL(aa), NCOL(cc))
  expect_lt(NCOL(aa), NCOL(dd))
  expect_lt(NCOL(cc), NCOL(dd))
})

test_that("ipni_search fails correctly", {
  skip_on_cran()
  
  expect_error(ipni_search(output = "foobar"), "'arg' should be one of")

  vcr::use_cassette("ipni_search_no_results", {
    expect_warning(ipni_search(family = 5), "No data")
    expect_warning(ipni_search(genus = "adfasdfasffd"), "No data found")
  })
})

#' Search for gene sequences available for taxa from NCBI.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @rdname ncbi_search-defunct
#' @keywords internal
#' @export
ncbi_search <- function(...) {
  .Defunct("ncbi_searcher", "traits",
    msg = "This function is defunct. See traits::ncbi_searcher()")
}

#' Retrieve gene sequences from NCBI by accession number.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @keywords internal
#' @rdname get_genes_avail-defunct
get_genes_avail <- function(...) {
  .Defunct("ncbi_search", "taxize", "Function name changed. See ncbi_search")
}

#' @title Get a summary from the IUCN Red List
#'
#' @description Get a summary from the IUCN Red List
#'   (https://www.iucnredlist.org/).
#'
#' @export
#' @param x character; Scientific name. Should be cleaned and in the format
#'   `*<Genus> <Species>*`.
#' @param distr_detail logical; If `TRUE`, the geographic distribution is
#'   returned as a list of vectors corresponding to the different range types:
#'   native, introduced, etc.
#' @param key a Redlist API key, get one from
#'   https://apiv3.iucnredlist.org/api/v3/token Required for `iucn_summary`.
#'   Defaults to `NULL` in case you have your key stored (see `Redlist
#'   Authentication` below).
#' @param ... curl options passed on to [crul::verb-GET]
#'
#' @return A list (for every species one entry) of data returned by
#'   [rredlist::rl_species_latest()].
#'
#' @note Not all data types are available for every species
#'   and NA is returned. [iucn_status()] is an extractor function to easily
#'   extract status into a vector.
#'
#' @seealso [iucn_status()]
#'
#' @details Beware: IUCN functions can give back incorrect data. This isn't our
#'   fault. We do our best to get you the correct data quickly, but sometimes
#'   IUCN gives back the wrong data, and sometimes Global Names gives back the
#'   wrong data. We will fix these as soon as possible. In the meantime, just
#'   make sure that the data you get back is correct.
#'
#'   `iucn_summary` has a default method that errors when anything's passed in
#'   that's not `character` or `iucn` class - a `iucn_summary.character` method
#'   for when you pass in taxon names - and a `iucn_summary.iucn` method so you
#'   can pass in iucn class objects as output from [get_iucn()] or [as.iucn()].
#'   If you already have IUCN IDs, coerce them to `iucn` class via `as.iucn(...,
#'   check = FALSE)`
#'
#' @author Eduard Szoecs, \email{eduardszoecs@@gmail.com}
#' @author Philippe Marchand, \email{marchand.philippe@@gmail.com}
#' @author Scott Chamberlain,
#' @author Zachary S.L. Foster
#'
#' @section Redlist Authentication: `iucn_summary` uses the new Redlist API for
#'   searching for a IUCN ID, so we use the [rredlist::rl_species()] function internally.
#'   This function requires an API key. Get the key at
#'   https://apiv3.iucnredlist.org/api/v3/token, and pass it to the `key`
#'   parameter, or store in your `.Renviron` file like
#'   `IUCN_REDLIST_KEY=yourkey` or in your `.Rprofile` file like
#'   `options(iucn_redlist_key="yourkey")`. We strongly encourage you to not
#'   pass the key in the function call but rather store it in one of those two
#'   files. This key will also set you up to use the \pkg{rredlist} package.
#'
#' @examples \dontrun{
#' # if you send a taxon name, an IUCN API key is required
#' ## here, the key is being detected from a .Rprofile file
#' ## or .Renviron file, See "Redlist Authentication" above
#' iucn_summary("Lutra lutra")
#'
#' ia <- iucn_summary(c("Panthera uncia", "Lynx lynx"))
#' ia <- iucn_summary(c("Panthera uncia", "Lynx lynx", "aaa"))
#' iucn_summary("Muntiacus rooseveltorum/truongsonensis")
#' iucn_summary(c("Muntiacus rooseveltorum/truongsonensis", "Lynx lynx"))
#'
#' ## get detailed distribution
#' iac <- iucn_summary(x="Ara chloropterus", distr_detail = TRUE)
#' iac[[1]]$distr
#'
#'
#' # If you pass in an IUCN ID, you don't need to pass in a Redlist API Key
#' # extract status
#' iucn_status(iac)
#' }
iucn_summary <- function(x, distr_detail = FALSE, key = NULL, ...) {
  UseMethod("iucn_summary")
}

#' @export
iucn_summary.default <- function(x, distr_detail = FALSE, key = NULL, ...) {
  stop("no 'iucn_summary' method for ", class(x), call. = FALSE)
}

#' @export
iucn_summary.character <- function(x, distr_detail = FALSE, key = NULL, ...) {
  raw_data <- get_iucn_data(x, key = key, latest = TRUE)
  structure(raw_data, class = "iucn_summary")
}


## helpers --------
try_red <- function(fun, x, key, ...) {
  tryCatch(fun(id = x, key = key, ...), error = function(e) e)
}

null_res <- list(status = NA, history = NA, distr = NA, trend = NA)

#' Extractor functions for `iucn`-class.
#'
#' @export
#' @param x an `iucn`-object as returned b`iucn_summary`ry
#' @param ... Currently not used
#' @return A character vector with the status.
#' @seealso [iucn_summary()]
#' @examples \dontrun{
#' ia <- iucn_summary(c("Panthera uncia", "Lynx lynx"))
#' iucn_status(ia)}
iucn_status <- function(x, ...){
  UseMethod("iucn_status")
}

#' @export
iucn_status.default <- function(x, ...) {
  stop("no method for 'iucn_status' for ", class(x), call. = FALSE)
}

#' @export
iucn_status.iucn_summary <- function(x, ...) {
  unlist(lapply(x, function(x) x$red_list_category$code))
}

context("iucn_summary: basic usage")
test_that("iucn_summary returns the correct value", {
  if (Sys.getenv("IUCN_REDLIST_KEY") == "") {
    skip("No IUCN api key so test not run.")
  }
  skip_on_cran()

  # vcr::use_cassette("iucn_summary", {
  #   temp <- iucn_summary(c("Panthera uncia", "Lynx lynx"))
  # }, preserve_exact_body_bytes = TRUE, match_requests_on = c("method", "query"))
  temp <- iucn_summary(c("Panthera uncia", "Lynx lynx"))

  expect_is(temp, "iucn_summary")
  expect_equal(length(iucn_status(temp)), 2)
})

context("iucn_summary: gives expected result for lots of names")
test_that("iucn_summary gives expected result for lots of names", {
  if (Sys.getenv("IUCN_REDLIST_KEY") == "") {
    skip("No IUCN api key so test not run.")
  }
  skip_on_cran()

  # vcr::use_cassette("iucn_summary_other_egs", {
  #   aa <- iucn_summary("Abies koreana")
  #   bb <- iucn_summary("Xylopia collina")
  #   cc <- iucn_summary("Brugmansia versicolor")
  #   dd <- iucn_summary("Achatinella buddii")
  #   ee <- iucn_summary("Annona hystricoides")
  #   ff <- iucn_summary("Chamaecrista onusta")
  #   gg <- iucn_summary("Cyornis lemprieri")
  #   hh <- iucn_summary("Frailea pumila")
  # }, 
  #   preserve_exact_body_bytes = TRUE, 
  #   match_requests_on = c("method", "query")
  # )
  aa <- iucn_summary("Abies koreana")
  bb <- iucn_summary("Xylopia collina")
  cc <- iucn_summary("Brugmansia versicolor")
  dd <- iucn_summary("Achatinella buddii")
  ee <- iucn_summary("Annona hystricoides")
  ff <- iucn_summary("Chamaecrista onusta")
  gg <- iucn_summary("Cyornis lemprieri")
  hh <- iucn_summary("Frailea pumila")
  
  expect_equal(aa$`Abies koreana`$red_list_category$code, "EN")
  expect_equal(bb$`Xylopia collina`$red_list_category$code, "NT")
  expect_equal(cc$`Brugmansia versicolor`$red_list_category$code, "EW")
  expect_equal(dd$`Achatinella buddii`$red_list_category$code, "EX")
  expect_equal(ee$`Annona hystricoides`$red_list_category$code, "CR")
  expect_equal(ff$`Chamaecrista onusta`$red_list_category$code, "VU")
  expect_equal(gg$`Cyornis lemprieri`$red_list_category$code, "NT")
  expect_equal(hh$`Frailea pumila`$red_list_category$code, "LC")
})

m_not_found_sp_altclass <-
"Not found. Consider checking the spelling or alternate classification"

m_more_than_one_found <-
"More than one %s found for taxon '%s'; refine query or set ask=TRUE"

m_na_ask_false <-
"NA due to ask=FALSE & > 1 result"

m_na_ask_false_no_direct <-
"NA due to ask=FALSE & no direct match found"

#' Search Tropicos by scientific name, common name, or Tropicos ID.
#'
#' @export
#' @param sci A scientific name, e.g., "poa annua". See Details.
#' @param com A common name, e.g., "annual blue grass"
#' @param nameid Your search string. e.g., "25509881"
#' @param orderby Your search string. e.g., "1"
#' @param sortorder Your search string. e.g., "ascending"
#' @param pagesize Your search string. e.g., "100"
#' @param startrow Your search string. e.g., "1"
#' @param type Type of search, "wildcard" (default) will add a wildcard to
#' the end of your search string. "exact" will use your search string exactly.
#' @param key Your Tropicos API key; See [taxize-authentication] 
#' for help on authentication
#' @param name Deprecated, see `sci`
#' @param commonname Deprecated, see `com`
#' @param ... Further args passed on to [crul::HttpClient]
#' @return List or dataframe.
#' @references http://services.tropicos.org/help?method=SearchNameXml
#' @details More details on the `name` parameter: Tropicos will fail
#' if you include a period (`.`) in your name string, e.g., `var.`,
#' so we replace periods before the request is made to the Tropicos web service.
#' In addition, Tropicos for some reason doesn't want to see sub-specific rank
#' names like `var`/`subsp`, so remove those from your query.
#' @examples \dontrun{
#' tp_search(sci = 'Poa annua')
#' tp_search(sci = 'Poa annua subsp. annua')
#' tp_search(sci = 'Poa annua var. annua')
#' tp_search(sci = 'Poa annua var annua')
#' tp_search(sci = 'Poa annua annua')
#' }
tp_search <- function(sci=NULL, com=NULL, nameid=NULL, orderby=NULL,
  sortorder=NULL, pagesize=NULL, startrow=NULL, type=NULL, key=NULL, name=NULL,
  commonname=NULL, ...) {

  if (!is.null(name)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "tp_search(name)", with = "tp_search(sci)")
    sci <- name
  }
  if (!is.null(commonname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "tp_search(commonname)", with = "tp_search(com)")
    com <- commonname
  }
  
  
  url = 'http://services.tropicos.org/Name/Search'
  key <- getkey(key, "TROPICOS_KEY")
  if (!is.null(sci)) {
    if (grepl(paste(sprintf("\\s%s\\.?\\s", subsp_ranks), collapse = "|"), sci)) {
      warning("Tropicos doesn't like sub-specific ranks - remove them in your query",
        call. = FALSE)
    }
    if (grepl("\\.", sci)) {
      warning("'.' detected, being URL encoded prior to data request",
        call. = FALSE)
    }
    sci <- gsub("\\.", "%2E", sci)
  }
  args <- tc(list(format='json', name=sci, nameid=nameid,
                  commonname=com, orderby=orderby, sortorder=sortorder,
                  pagesize=pagesize, startrow=startrow, type=type, apikey=key))
  tt <- tp_GET(url, args, raise = FALSE, ...)
  out <- jsonlite::fromJSON(tt, FALSE)
  tmp <- dt2df(lapply(out, data.frame, stringsAsFactors = FALSE), idcol = FALSE)
  stats::setNames(tmp, tolower(names(tmp)))
}

subsp_ranks <- c('sp', 'ssp', 'subsp', 'subspecies', 'var', 'varietas', 'fo', 'f', 'forma')

context("tp_search")

test_that("tp_search returns the correct class", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tp_search", {
    ttt <- suppressMessages(tp_search(sci = 'Poa annua'))
    uuu <- suppressMessages(tp_search(sci = 'stuff things'))
  }, preserve_exact_body_bytes = TRUE)

  if ("error" %in% names(ttt)) skip("error in tp_search call - skipping")

	expect_that(ttt, is_a("data.frame"))

  expect_that(uuu, is_a("data.frame"))
  expect_that(names(uuu), equals("error"))
  expect_that(as.character(uuu[1,1]), equals("No names were found"))
})

test_that("tp_search behaves correctly on dot inputs", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tp_search_warnings_dots", {
    expect_that(tp_search('Poa annua .annua'),
                gives_warning("detected, being URL encoded"))
    expect_warning(tp_search('Poa annua annua'), NA)
  })
})

test_that("tp_search behaves correctly on subspecific inputs", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tp_search_warnings_subspecific", {
    expect_that(tp_search('Poa annua var annua'),
                gives_warning("Tropicos doesn't like"))
    expect_that(tp_search('Poa annua var. annua'),
                gives_warning("Tropicos doesn't like"))
    expect_that(tp_search('Poa annua sp. annua'),
                gives_warning("Tropicos doesn't like"))
    expect_that(tp_search('Poa annua ssp. annua'),
                gives_warning("Tropicos doesn't like"))
    expect_that(tp_search('Poa annua subspecies annua'),
                gives_warning("Tropicos doesn't like"))

    expect_warning(tp_search('Poa annua foo bar annua'), NA)
  })
})

#' Retrieve immediate children taxa for a given taxon name or ID.
#'
#' This function is different from [downstream()] in that it only
#' collects immediate taxonomic children, while [downstream()]
#' collects taxonomic names down to a specified taxonomic rank, e.g.,
#' getting all species in a family.
#'
#' @export
#' @param sci_id Vector of taxa names (character) or IDs (character or numeric)
#' to query.
#' @param db character; database to query. One or more of `itis`,
#' `ncbi`, `worms`, or `bold`. Note that each taxonomic data
#' source has their own identifiers, so that if you provide the wrong
#' `db` value for the identifier you could get a result, but it will
#' likely be wrong (not what you were expecting). If using ncbi, we recommend
#' getting an API key; see [taxize-authentication]
#' @param rows (numeric) Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this parameter is ignored if you pass in a
#' taxonomic id of any of the acceptable classes: tsn. NCBI has a
#' method for this function but rows doesn't work.
#' @param x Deprecated, see `sci_id`
#' @param ... Further args passed on to [ritis::hierarchy_down()],
#' [ncbi_children()], [worrms::wm_children()], [bold_children()]
#' See those functions for what parameters can be passed on.
#'
#' @section ncbi:
#' note that with `db = "ncbi"`, we set `ambiguous = TRUE`; that is, children
#' taxa with words like "unclassified", "unknown", "uncultured", "sp." are
#' NOT removed
#' 
#' @section bold:
#' BEWARE: `db="bold"` scrapes the BOLD website, so may be unstable. That is,
#' one day it may work, and the next it may fail. Open an issue if you
#' encounter an error: https://github.com/ropensci/taxize/issues
#'
#' @return A named list of data.frames with the children names of every
#' supplied taxa. You get an NA if there was no match in the database.
#'
#' @examples \dontrun{
#' # Plug in taxonomic IDs
#' children(161994, db = "itis")
#' children(8028, db = "ncbi")
#' ## works with numeric if as character as well
#' children(161994, db = "itis")
#' children(88899, db = "bold")
#' children(as.boldid(88899))
#'
#' # Plug in taxon names
#' children("Salmo", db = 'itis')
#' children("Salmo", db = 'ncbi')
#' children("Salmo", db = 'worms')
#' children("Salmo", db = 'bold')
#'
#' # Plug in IDs
#' (id <- get_wormsid("Gadus"))
#' children(id)
#'
#' # Many taxa
#' sp <- c("Tragia", "Schistocarpha", "Encalypta")
#' children(sp, db = 'itis')
#'
#' # Two data sources
#' (ids <- get_ids("Apis", db = c('ncbi','itis')))
#' children(ids)
#' ## same result
#' children(get_ids("Apis", db = c('ncbi','itis')))
#'
#' # Use the rows parameter
#' children("Poa", db = 'itis')
#' children("Poa", db = 'itis', rows=1)
#'
#' # use curl options
#' res <- children("Poa", db = 'itis', rows=1, verbose = TRUE)
#' }

children <- function(...){
  UseMethod("children")
}

#' @export
#' @rdname children
children.default <- function(sci_id, db = NULL, rows = NA, x = NULL, ...) {
  nstop(db)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "children(x)", with = "children(sci_id)")
    sci_id <- x
  }
  
  results <- switch(
    db,
    itis = {
      id <- process_children_ids(sci_id, db, get_tsn, rows = rows, ...)
      stats::setNames(children(id, ...), sci_id)
    },

    ncbi = {
      if (all(grepl("^[[:digit:]]*$", sci_id))) {
        id <- sci_id
        class(id) <- "uid"
        stats::setNames(children(id, ...), sci_id)
      } else {
        out <- ncbi_children(name = sci_id, ...)
        structure(out, class = 'children', db = 'ncbi', .Names = sci_id)
      }
    },

    worms = {
      id <- process_children_ids(sci_id, db, get_wormsid, rows = rows, ...)
      stats::setNames(children(id, ...), sci_id)
    },

    bold = {
      id <- process_children_ids(as.character(sci_id), db, get_boldid,
        rows = rows, ...)
      stats::setNames(children(id, ...), sci_id)
    },

    stop("the provided db value was not recognised", call. = FALSE)
  )

  set_output_types(results, sci_id, db)
}

# Ensure that the output types are consistent when searches return nothing
itis_blank <- data.frame(
  parentname = character(0),
  parenttsn  = character(0),
  rankname   = character(0),
  taxonname  = character(0),
  tsn        = character(0),
  stringsAsFactors = FALSE
)
worms_blank <- ncbi_blank <- bold_blank <-
  data.frame(
    childtaxa_id     = character(0),
    childtaxa_name   = character(0),
    childtaxa_rank   = character(0),
    stringsAsFactors = FALSE
  )

set_output_types <- function(x, x_names, db){
  blank_fun <- switch(
    db,
    itis  = function(w) if (nrow(w) == 0 || all(is.na(w))) itis_blank else w,
    ncbi  = function(w) if (nrow(w) == 0 || all(is.na(w))) ncbi_blank else w,
    worms = function(w) if (nrow(w) == 0 || all(is.na(w))) worms_blank else w,
    bold = function(w) if (nrow(w) == 0 || all(is.na(w))) bold_blank else w
  )

  typed_results <- lapply(seq_along(x), function(i) blank_fun(x[[i]]))
  names(typed_results) <- x_names
  attributes(typed_results) <- attributes(x)
  typed_results
}

process_children_ids <- function(input, db, fxn, ...){
  g <- tryCatch(as.numeric(as.character(input)), warning = function(e) e)
  if (inherits(g, "condition")) return(eval(fxn)(input, ...))
  if (is.numeric(g) || is.character(input) && all(grepl("[[:digit:]]", input))) {
    as_fxn <- switch(db, itis = as.tsn, worms = as.wormsid, bold = as.boldid)
    as_fxn(input, check = FALSE)
  } else {
    eval(fxn)(input, ...)
  }
}

#' @export
#' @rdname children
children.tsn <- function(sci_id, db = NULL, ...) {
  warn_db(list(db = db), "itis")
  fun <- function(y){
    # return NA if NA is supplied
    if (is.na(y)) {
      out <- NA
    } else {
		  out <- ritis::hierarchy_down(y, ...)
    }
  }
  out <- lapply(sci_id, fun)
  names(out) <- sci_id
  class(out) <- 'children'
  attr(out, 'db') <- 'itis'
  return(out)
}

df2dt2tbl <- function(x) {
  tibble::as_tibble(
    data.table::setDF(
      data.table::rbindlist(
        x, use.names = TRUE, fill = TRUE)
    )
  )
}

#' @export
#' @rdname children
children.wormsid <- function(sci_id, db = NULL, ...) {
  warn_db(list(db = db), "worms")
  fun <- function(y){
    # return NA if NA is supplied
    if (is.na(y)) {
      out <- NA
    } else {
      out <- worms_children_all(y, ...)
      stats::setNames(
        out[names(out) %in% c('AphiaID', 'scientificname', 'rank')],
        c('childtaxa_id', 'childtaxa_name', 'childtaxa_rank')
      )
    }
  }
  out <- lapply(sci_id, fun)
  names(out) <- sci_id
  class(out) <- 'children'
  attr(out, 'db') <- 'worms'
  return(out)
}

#' @export
#' @rdname children
children.ids <- function(sci_id, db = NULL, ...) {
  fun <- function(y, ...){
    # return NA if NA is supplied
    if (is.na(y)) {
      out <- NA
    } else {
      out <- children(y, ...)
    }
    return(out)
  }
  out <- lapply(sci_id, fun)
  class(out) <- 'children_ids'
  return(out)
}

#' @export
#' @rdname children
children.uid <- function(sci_id, db = NULL, ...) {
  warn_db(list(db = db), "uid")
  out <- if (is.na(sci_id)) {
    stats::setNames(list(ncbi_blank), sci_id)
  } else {
    ncbi_children(id = sci_id, ambiguous = TRUE, ...)
  }
  class(out) <- 'children'
  attr(out, 'db') <- 'ncbi'
  return(out)
}

#' @export
#' @rdname children
children.boldid <- function(sci_id, db = NULL, ...) {
  warn_db(list(db = db), "bold")
  out <- if (is.na(sci_id)) {
    stats::setNames(list(bold_blank), sci_id)
  } else {
    bold_children(id = sci_id, ...)
  }
  class(out) <- 'children'
  attr(out, 'db') <- 'bold'
  return(out)
}

#' Get family names to make Phylomatic input object, and output input string
#'    to Phylomatic for use in the function phylomatic_tree.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @keywords internal
#' @rdname phylomatic_format-defunct
#' @param ... Parameters, ignored
phylomatic_format <- function(...) {
  .Defunct(msg = "This function is defunct - See ?`taxize-defunct`")
}

#' Search Catalogue of Life for taxonomic IDs
#'
#' @export
#' @param name The string to search for. Only exact matches found the name given
#'   	will be returned, unless one or wildcards are included in the search
#'   	string. An * (asterisk) character denotes a wildcard; a % (percentage)
#'    character may also be used. The name must be at least 3 characters long,
#'    not counting wildcard characters.
#' @param id The record ID of the specific record to return (only for scientific
#' 		names of species or infraspecific taxa)
#' @param start The first record to return. If omitted, the results are returned
#' 		from the first record (start=0). This is useful if the total number of
#' 		results is larger than the maximum number of results returned by a single
#' 		Web service query (currently the maximum number of results returned by a
#' 		single query is 500 for terse queries and 50 for full queries).
#' @param checklist The year of the checklist to query, if you want a specific
#' 		year's checklist instead of the lastest as default (numeric).
#' @param response (character) one of "terse" or "full"
#' @param ... Curl options passed on to \code{\link[httr]{GET}}
#' @details You must provide one of name or id. The other parameters (format
#' 		and start) are optional.
#' @references \url{http://webservice.catalogueoflife.org/}
#' @return A list of data.frame's.
#' @examples \dontrun{
#' # A basic example
#' col_search(name="Apis")
#' col_search(name="Agapostemon")
#' col_search(name="Poa")
#'
#' # Get full response, i.e., more data
#' col_search(name="Apis", response="full")
#' col_search(name="Poa", response="full")
#'
#' # Many names
#' col_search(name=c("Apis","Puma concolor"))
#' col_search(name=c("Apis","Puma concolor"), response = "full")
#'
#' # An example where there is no data
#' col_search(id = "36c623ad9e3da39c2e978fa3576ad415")
#' col_search(id = "36c623ad9e3da39c2e978fa3576ad415", response = "full")
#' col_search(id = "787ce23969f5188c2467126d9a545be1")
#' col_search(id = "787ce23969f5188c2467126d9a545be1", response = "full")
#' col_search(id = c("36c623ad9e3da39c2e978fa3576ad415", "787ce23969f5188c2467126d9a545be1"))
#' ## a synonym
#' col_search(id = "f726bdaa5924cabf8581f99889de51fc")
#' col_search(id = "f726bdaa5924cabf8581f99889de51fc", response = "full")
#' }

col_search <- function(name=NULL, id=NULL, start=NULL, checklist=NULL, response="terse", ...) {
  response <- match.arg(response, c("terse", "full"))
  func <- function(x, y, ...) {
    url <- make_url(checklist)
    args <- tc(list(name = x, id = y, start = start, response = response,
                    format = "json"))
    temp <- GET(url, query = argsnull(args), ...)
    if (temp$status_code >= 300) {
      warning("COL taxon not found", call. = FALSE)
    } else {
      tt <- jsonlite::fromJSON(con_utf8(temp), FALSE)
      switch(response,
             terse = parse_terse(tt),
             full = parse_full(tt))
    }
  }
  safe_func <- plyr::failwith(NULL, func)
  if (is.null(id)) {
    setNames(lapply(name, safe_func, y = NULL, ...), name)
  } else {
    setNames(lapply(id, safe_func, x = NULL, ...), id)
  }
}

make_url <- function(checklist) {
  if (is.null(checklist)) {
    col_base()
  } else {
    cc <- match.arg(as.character(checklist), choices = 2015:2007)
    sprintf("http://catalogueoflife.org/annual-checklist/%s/webservice", cc)
  }
}

col_base <- function() "http://www.catalogueoflife.org/col/webservice"

parse_terse <- function(x) {
  nodes <- x$results
  ldply(nodes, parsecoldata)
}

parsecoldata <- function(x){
  vals <- x[c('id', 'name', 'rank', 'name_status', 'source_database')]
  vals[sapply(vals, is.null)] <- NA
  names(vals) <- c('id', 'name', 'rank', 'name_status', 'source_database')
  bb <- data.frame(vals, stringsAsFactors = FALSE)
  names(bb)[4:5] <- c('status', 'source')
  bb$rank <- tolower(bb$rank)
  acc <- x$accepted_name
  if (is.null(acc)) {
    accdf <- data.frame(acc_id=NA, acc_name=NA, acc_rank=NA, acc_status=NA, acc_source=NA, stringsAsFactors = FALSE)
  } else {
    accdf <- data.frame(acc[c('id','name','rank','name_status','source_database')], stringsAsFactors=FALSE)
    names(accdf) <- c('acc_id','acc_name','acc_rank','acc_status','acc_source')
    accdf$acc_rank <- tolower(accdf$acc_rank)
  }
  cbind(bb, accdf)
}

parse_full <- function(x) {
  tmp <- Filter(length, x$results)
  taxize_ldfast(
    lapply(tmp, function(z) {
      switch(z$name_status,
             `accepted name` = {
               if (length(z$classification) == 0) {
                 h <- parse_one(z)
                 rank <- z$rank
                 id <- z$id
               } else {
                 h <- parse_one(z)
                 h_vals <- pluck(z$classification, "name", "")
                 h_nms <- pluck(z$classification, "rank", "")
                 class <- setNames(rbind.data.frame(h_vals), tolower(h_nms))
                 h <- cbind(h, class)
                 rank <- z$rank
                 id <- z$id
               }
             },
             `provisionally accepted name` = {
               if (length(z$classification) == 0) {
                 h <- parse_one(z)
                 rank <- z$rank
                 id <- z$id
               } else {
                 h <- parse_one(z)
                 h_vals <- pluck(z$classification, "name", "")
                 h_nms <- pluck(z$classification, "rank", "")
                 class <- setNames(rbind.data.frame(h_vals), tolower(h_nms))
                 h <- cbind(h, class)
                 rank <- z$rank
                 id <- z$id
               }
             },
             `common name` = {
               h_vals <- pluck(z$accepted_name$classification, "name", "")
               h_nms <- pluck(z$accepted_name$classification, "rank", "")
               h <- setNames(rbind.data.frame(h_vals), tolower(h_nms))
               rank <- z$accepted_name$rank
               id <- z$accepted_name$id
             },
             `synonym` = {
               h <- parse_one(z)
               name <- z$accepted_name$name
               rank <- z$accepted_name$rank
               id <- z$accepted_name$id
               name_status <- z$accepted_name$name_status
               h <- cbind(h, setNames(data.frame(id, name, rank, name_status, stringsAsFactors = FALSE),
                                      c('acc_id','acc_name','acc_rank','acc_status')))
             },
             `ambiguous synonym` = {
               h <- parse_one(z)
               name <- z$accepted_name$name
               rank <- z$accepted_name$rank
               id <- z$accepted_name$id
               name_status <- z$accepted_name$name_status
               h <- cbind(h, setNames(data.frame(id, name, rank, name_status, stringsAsFactors = FALSE),
                                      c('acc_id','acc_name','acc_rank','acc_status')))
             },
             `misapplied name` = {
               h <- parse_one(z)
               name <- z$accepted_name$name
               rank <- z$accepted_name$rank
               id <- z$accepted_name$id
               name_status <- z$accepted_name$name_status
               h <- cbind(h, setNames(data.frame(id, name, rank, name_status, stringsAsFactors = FALSE),
                                      c('acc_id','acc_name','acc_rank','acc_status')))
             }
      )
      target <- setNames(rbind.data.frame(
        c(z$name,
          rank,
          id,
          z$name_status)),
        c("name", "rank", "id", "name_status"))
      target$rank <- tolower(target$rank)
      tempdf <- cbind(target, h)
      tempdf[] <- lapply(tempdf, as.character)
      tempdf
    })
  )
}

parse_one <- function(z) {
  scrut <- z$record_scrutiny_date
  scrutie <- if (is.null(scrut[[1]]) || scrut[[1]] == FALSE) FALSE else TRUE
  if (scrutie) scrut <- data.frame(record_scrutiny_date = scrut$scrutiny, stringsAsFactors = FALSE)
  refs <- z$references
  refsie <- if (is.null(refs) || length(refs) == 0) FALSE else TRUE
  if (refsie) refs <- data.frame(tc(refs[[1]]), stringsAsFactors = FALSE)
  lst <- pop(z, c("distribution", "classification", "synonyms", "common_names",
                  "record_scrutiny_date", "references", "accepted_name", "child_taxa",
                  "name_html"))
  df <- data.frame(lst, stringsAsFactors = FALSE)
  if (is(scrut, "data.frame")) df <- cbind(df, scrut)
  if (is(refs, "data.frame")) df <- cbind(df, refs)
  df
}

# tests for col_search fxn in taxize
context("col_search")


test_that("col_search returns the correct values, dimensions, and classes", {
  skip_on_cran()

  temp <- col_search(name = "Apis")
  two <- col_search(name = c("Apis","Puma concolor"))
  byid <- col_search(id = 11935941)

	expect_equal(names(temp), "Apis")
	expect_equal(temp[[1]]$name[1], "Apis")

	expect_equal(names(temp), "Apis")
	expect_equal(temp[[1]]$name[1], "Apis")

  expect_that(is.null(dim(temp)), is_true())
  expect_equal(NCOL(temp[[1]]), 10)

  expect_that(length(two), equals(2))
  expect_equal(NCOL(two[[1]]), 10)
  expect_equal(NCOL(do.call(rbind, lapply(two, data.frame))), 10)
  expect_that(dim(byid[[1]]), equals(c(0,0)))

	expect_that(temp, is_a("list"))
	expect_that(temp[[1]], is_a("data.frame"))

	expect_that(two, is_a("list"))
	expect_that(two[[1]], is_a("data.frame"))
})

test_that("col_search is robust to user error", {
  expect_is(col_search(name = "asdfsdf")[[1]], "data.frame")
  expect_is(col_search(name = "")[[1]], "data.frame")
  expect_is(col_search(id = "asdfsdf")[[1]], "data.frame")
  expect_is(col_search(), "list")
  expect_equal(length(col_search()), 0)
})

#' Search uBio by namebank ID.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @rdname ubio_id-defunct
#' @export
#' @param ... Parameters, ignored
ubio_id <- function(...) {
  .Defunct(msg = "the uBio API is down, for good as far as we know")
}

# tests for ubio_id fxn in taxize
context("ubio_id")

out <- ubio_id(namebankID = 2483153)

test_that("ubio_id returns the correct dimensions", {
	expect_that(length(out), equals(5))
	expect_that(dim(out$data), equals(c(1,8)))
	expect_that(dim(out$synonyms), equals(c(7,3)))
	expect_that(dim(out$vernaculars), equals(c(27,5)))
	expect_that(dim(out$cites), equals(c(7,4)))
})

test_that("ubio_id returns the correct class", {
	expect_that(out, is_a("list"))
	expect_that(out$synonyms, is_a("data.frame"))
	expect_that(out$cites, is_a("data.frame"))
})

taxize_env <- new.env()
taxize_env$options <- list()

#' taxize options
#'
#' @export
#' @param taxon_state_messages (logical) suppress messages? default: `NULL`
#' (same as setting `FALSE`). Set to `TRUE` to suppress messages, and `FALSE`
#' to not suppress messages
#' @param ncbi_sleep (numeric/integer) number of seconds to sleep between
#' NCBI ENTREZ http requests. applies to the functions: [classification()],
#' [comm2sci()], [genbank2uid()], [get_uid()] and [ncbi_children()].
#' defaults: 0.334 (without API key) or 0.101 (with API key). minimum value
#' can not be less than 0.101
#' @param quiet (logical) quiet informational output from this function.
#' default: `TRUE`
#' @examples \dontrun{
#' taxize_options()
#' taxize_options(FALSE)
#' taxize_options(TRUE)
#' taxize_options(ncbi_sleep = 0.4)
#' taxize_options(taxon_state_messages = TRUE, ncbi_sleep = 0.4)
#' }
taxize_options <- function(taxon_state_messages = NULL, ncbi_sleep = NULL,
  quiet = FALSE) {

  assert(taxon_state_messages, 'logical')
  taxize_env$options$taxon_state_messages <- taxon_state_messages
  if (!is.null(ncbi_sleep)) {
    assert(ncbi_sleep, c('numeric', 'integer'))
    if (!ncbi_sleep > 0.101) stop("'ncbi_sleep' must be > 0.101")
  }
  taxize_env$options$ncbi_sleep_sec <- ncbi_sleep
  tseo <- taxize_env$options
  if (!quiet) cat("taxize options", sep = "\n")
  if (length(tseo) == 0) return(cat(""))
  for (i in seq_along(tseo)) {
    if (!quiet) cat(sprintf("  %s: %s", names(tseo)[i], tseo[[i]]),
      sep = "\n")
  }
}

skip_on_cran()

test_that("taxize_options", {
  taxize_options(quiet=TRUE) # reset
  expect_is(taxize_env, "environment")
  expect_is(taxize_env$options, "list")
  expect_length(taxize_env$options, 0)
  # fxn returns null itself
  expect_null(taxize_options(quiet=TRUE))
})

test_that("taxize_options fails well", {
  # unused arg
  expect_error(taxize_options(a=5, quiet=TRUE), "unused argument")

  # wrong types
  expect_error(taxize_options(taxon_state_messages = 5, quiet=TRUE), "logical")
  expect_error(taxize_options(ncbi_sleep = 'foo', quiet=TRUE), "numeric")

  # sleep min value
  expect_error(taxize_options(ncbi_sleep = 0.01, quiet=TRUE),
    "must be >")
})

#' Resolve names using Global Names Resolver
#'
#' NOTE: this function is depreciated and will be removed in a future version.
#' The service this function interacts with is no longer maintained and has been
#' replaced by GNA Verifier, which can be used with the [gna_verifier()]
#' function.
#' 
#' See section **Age of datasets in the Global Names Resolver**
#'
#' @export
#' @param sci character; taxonomic names to be resolved. Doesn't work for
#'   vernacular/common names.
#' @param data_source_ids character; IDs to specify what data source is
#'   searched. See [gnr_datasources()].
#' @param resolve_once logical; Find the first available match instead of
#'   matches across all data sources with all possible renderings of a name.
#'   When `TRUE`, response is rapid but incomplete.
#' @param with_context logical; Reduce the likelihood of matches to taxonomic
#'   homonyms. When `TRUE` a common taxonomic context is calculated for all
#'   supplied names from matches in data sources that have classification tree
#'   paths. Names out of determined context are penalized during score
#'   calculation.
#' @param canonical logical; If `FALSE` (default), gives back names with
#'   taxonomic authorities. If `TRUE`, returns canocial names (without tax.
#'   authorities and abbreviations).
#' @param highestscore logical; Return those names with the highest score for
#'   each searched name? Defunct
#' @param best_match_only (logical) If `TRUE`, best match only returned.
#'   Default: `FALSE`
#' @param preferred_data_sources (character) A vector of one or more data source
#'   IDs.
#' @param with_canonical_ranks (logical) Returns names with infraspecific ranks,
#'   if present. If `TRUE`, we force `canonical=TRUE`, otherwise this parameter
#'   would have no effect. Default: `FALSE`
#' @param http The HTTP method to use, one of "get" or "post". Default: "get".
#'   Use `http="post"` with large queries. Queries with > 300 records use "post"
#'   automatically because "get" would fail
#' @param names Deprecated, see `sci`
#' @param ... Curl options passed on to [crul::HttpClient]
#' @param cap_first (logical) For each name, fix so that the first name part is
#'   capitalized, while others are not. This web service is sensitive to
#'   capitalization, so you'll get different results depending on
#'   capitalization. First name capitalized is likely what you'll want and is
#'   the default. If `FALSE`, names are not modified. Default: `TRUE`
#' @param fields (character) One of minimal (default) or all. Minimal gives back
#'   just four fields, whereas all gives all fields back.
#'
#' @author Scott Chamberlain
#' @return A data.frame with one attribute `not_known`: a character vector of
#'   taxa unknown to the Global Names Index. Access like `attr(output,
#'   "not_known")`, or `attributes(output)$not_known`.
#'
#'   Columns of the output data.frame:
#' * user_supplied_name (character) - the name you passed in to the
#'   `names` parameter, unchanged.
#' * submitted_name (character) - the actual name submitted to the GNR
#'   service
#' * data_source_id (integer/numeric) - data source ID
#' * data_source_title (character) - data source name
#' * gni_uuid (character) - Global Names Index UUID (aka identifier)
#' * matched_name (character) - the matched name in the GNR service
#' * matched_name2 (character) - returned if `canonical=TRUE`, in
#'   which case **matched_name** is not returned
#' * classification_path (character) - names of the taxonomic
#'   classification tree, with names separated by pipes (`|`)
#' * classification_path_ranks (character) - ranks of the taxonomic
#'   classification tree, with names separated by pipes (`|`)
#' * classification_path_ids (character) - identifiers of the taxonomic
#'   classification tree, with names separated by pipes (`|`)
#' * taxon_id (character) - taxon identifier
#' * edit_distance (integer/numeric) - edit distance
#' * imported_at (character) - date imported
#' * match_type (integer/numeric) - match type
#' * match_value (character) - description of match type
#' * prescore (character) - pre score
#' * score (numeric) - score
#' * local_id (character) - local identifier
#' * url (character) - URL for taxon
#' * global_id (character) - global identifier
#' * current_taxon_id (character) - current taxon id
#' * current_name_string (character) - current name string
#'
#'   Note that names (i.e. rows) are dropped that are NA, are zero length
#'   strings, are not character vectors, or are not found by the API.
#'
#' @section Age of datasets in the Global Names Resolver: IMPORTANT: Datasets
#'   used in the Global Names Resolver vary in how recently they've been
#'   updated. See the `updated_at` field in the output of [gnr_datasources()]
#'   for dates when each dataset was last updated.
#'
#' @section preferred_data_sources: If `preferred_data_sources` is used, only
#'   the preferred data is returned - if it has any results.
#'
#' @seealso [gnr_datasources()]
#' @keywords resolve names taxonomy
#' @references http://gnrd.globalnames.org/api http://gnrd.globalnames.org/
#' @examples \dontrun{
#' gnr_resolve(sci = c("Helianthus annuus", "Homo sapiens"))
#' gnr_resolve(sci = c("Asteraceae", "Plantae"))
#'
#' # Using data source 12 (Encyclopedia of Life)
#' sources <- gnr_datasources()
#' sources
#' eol <- sources$id[sources$title == 'EOL']
#' gnr_resolve(names=c("Helianthos annuus","Homo sapians"), data_source_ids=eol)
#'
#' # Two species in the NE Brazil catalogue
#' sps <- c('Justicia brasiliana','Schinopsis brasiliensis')
#' gnr_resolve(sci = sps, data_source_ids = 145)
#'
#' # Best match only, compare the two
#' gnr_resolve(sci = "Helianthus annuus", best_match_only = FALSE)
#' gnr_resolve(sci = "Helianthus annuus", best_match_only = TRUE)
#'
#' # Preferred data source
#' gnr_resolve(sci = "Helianthus annuus", preferred_data_sources = c(3,4))
#'
#' # Return canonical names - default is canonical=FALSE
#' head(gnr_resolve(sci = "Helianthus annuus"))
#' head(gnr_resolve(sci = "Helianthus annuus", canonical=TRUE))
#'
#' # Return canonical names with authority stripped but
#' # ranks still present
#' gnr_resolve("Scorzonera hispanica L. subsp. asphodeloides Wallr.")
#' ## vs.
#' gnr_resolve("Scorzonera hispanica L. subsp. asphodeloides Wallr.",
#'    with_canonical_ranks = TRUE)
#' }

gnr_resolve <- function(sci, data_source_ids = NULL, resolve_once = FALSE,
  with_context = FALSE, canonical = FALSE, highestscore = TRUE,
  best_match_only = FALSE, preferred_data_sources = NULL,
  with_canonical_ranks = FALSE, http = "get", cap_first = TRUE,
  fields = "minimal", names = NULL, ...) {

  lifecycle::deprecate_warn(what = "taxize::gnr_resolve()", with = "taxize::gna_verifier()", when = "v0.9.103")
  
  if (!is.null(names)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "gnr_resolve(names)", with = "gnr_resolve(sci)")
    sci <- names
  }
  
  if (!is.null(names)) sci <- names
  fields <- match.arg(fields, c("minimal", "all"))
  http <- match.arg(http, c("get", "post"))
  num <- NULL
  url <- "https://resolver.globalnames.org/name_resolvers.json"
  # clean out zero length strings
  sci <- Filter(function(x) nzchar(x) && !is.na(x) && is.character(x), sci)
  # store original sci supplied by user
  orig_names <- sci
  if (cap_first) sci <- taxize_capwords(sci, onlyfirst = TRUE)
  names2 <- paste0(sci, collapse = "|")
  if (length(sci) > 300 && http == "get") http <- "post"

  data_source_ids <- paste0(data_source_ids, collapse = "|")
  preferred_data_sources <- paste0(preferred_data_sources, collapse = "|")
  if (nchar(preferred_data_sources, keepNA = FALSE) == 0)
    preferred_data_sources <- NULL
  if (with_canonical_ranks) canonical <- TRUE

  args <- tc(list(names = names2, data_source_ids = data_source_ids,
            resolve_once = cv(resolve_once), with_context = cv(with_context),
            best_match_only = cv(best_match_only),
            preferred_data_sources = preferred_data_sources,
            with_canonical_ranks = cv(with_canonical_ranks)))
  args <- argsnull(args)

  cli <- crul::HttpClient$new(url = url,
    headers = tx_ual, opts = list(...))
  if (http == 'get') {
    tmp <- cli$get(query = args)
    tmp$raise_for_status()
    tmp2 <- tmp$parse("UTF-8")
    dat <- jsonlite::fromJSON(tmp2, FALSE)$data
  } else {
    args <- args[!names(args) %in% "names"]
    nms <- split(sci, ceiling(seq_along(sci)/500))
    datbits <- list()
    for (i in seq_along(nms)) {
      tt <- data.frame(paste0(seq_along(nms[[i]]), "|", sci))
      file <- tempfile(fileext = ".txt")
      write.table(tt, file = file, row.names = FALSE,
                  col.names = FALSE, quote = FALSE)
      tmp <- cli$post(query = args, body = list(file = crul::upload(file)))
      tmp$raise_for_status()
      ss <- tmp$parse("UTF-8")
      datbits[[i]] <- jsonlite::fromJSON(ss, FALSE)$data
    }

    dat <- do.call("c", datbits)
  }

  # add original name supplied by user
  dat <- Map(function(x,y) c(original_name = y, x), dat, orig_names)

  to_get <- if (is.null(preferred_data_sources)) "results" else "preferred_results"

  data_ <-
    lapply(dat, function(y) {
      if (!is.null(unlist(y[[to_get]]))) {
        res <- lapply(y[[to_get]], function(x) {
          take_fields <- switch(fields,
            minimal = c("name_string", "data_source_title","score",
                        "canonical_form"),
            all = names(x)
          )
          take <- x[take_fields]
          take[sapply(take, is.null)] <- NA
          return(data.frame(take, stringsAsFactors = FALSE))
        })
      } else {
        res <- NULL
      }
      list(y[c("original_name", "supplied_name_string")], res)
    })
  not_known <- Filter(function(x) is.null(x[[2]]), data_)
  not_known <- sapply(not_known, function(x) x[[1]]$original_name)
  data_ <- Filter(function(x) !is.null(x[[2]]), data_)

  # check for empty data object
  drill <- tryCatch(data_[[1]], error = function(e) e)
  to_rename <- c("original_name", "supplied_name_string", "name_string",
                 "canonical_form")
  if (inherits(drill, "simpleError")) {
    out <- data.frame(NULL)
  } else {
    data_2 <- dt2df(lapply(data_, function(x)
      data.frame(x[[1]], dt2df( if (length(x[[2]]) == 0) {
      list(data.frame(name_string = "", data_source_title = "", score = NaN,
                      canonical_form = ""))
    } else {
      x[[2]]
    }, idcol = FALSE), stringsAsFactors = FALSE)), idcol = FALSE)
    names(data_2)[names(data_2) %in% to_rename] <-
      c("user_supplied_name", "submitted_name",
        "matched_name", "matched_name2")
    data_2$matched_name <- as.character(data_2$matched_name)
    data_2$data_source_title <- as.character(data_2$data_source_title)
    data_2$matched_name2 <- as.character(data_2$matched_name2)

    if (canonical) {
      data_2 <- data_2[ , !names(data_2) %in% "matched_name"]
    } else {
      data_2 <- data_2[ , !names(data_2) %in% "matched_name2"]
    }
    # canonical = TRUE, may result into duplicates
    out <- unique(data_2)
  }

  row.names(out) <- NULL
  structure(tibble::as_tibble(out), not_known = not_known)
}

cv <- function(x) {
  if (x) {
    'true'
  } else {
    NULL
  }
}

# NOTE: gnr_resolve is depreciated and will be removed in future versions.

# context("gnr_resolve")
# 
# test_that("gnr_resolve returns the correct value", {
#   skip_on_cran()
# 
#   tmp <- gnr_resolve(c("Helianthus annuus", "Homo sapiens"))
# 
#   expect_equal(NCOL(tmp), 5)
# 
# 	expect_is(tmp, "data.frame")
# 	expect_is(tmp$matched_name, "character")
# })
# 
# test_that("best_match_only works correctly", {
#   skip_on_cran()
# 
#   x <- "Helianthus annuus"
#   a <- gnr_resolve(x, best_match_only = TRUE)
#   b <- gnr_resolve(x, best_match_only = FALSE)
# 
#   expect_is(a, "data.frame")
#   expect_is(b, "data.frame")
#   expect_equal(NROW(a), 1)
#   expect_true(all(c("names", "row.names", "class", "not_known") %in% names(attributes(a))))
#   expect_is(b$data_source_title, "character")
# 
#   ## same order as user supplied
#   cc <- gnr_resolve(c("Homo sapiens", "Helianthus annuus"), best_match_only = TRUE)
#   expect_identical(cc$user_supplied_name, c("Homo sapiens", "Helianthus annuus"))
# })
# 
# test_that("canonical works correctly", {
#   skip_on_cran()
# 
#   # x = a canse where no canonical names is found
#   x <- gnr_resolve("Metzgeria", data_source_ids = c(12), canonical = TRUE)
#   y <- "Helianthus annuus"
#   w <- gnr_resolve(y, canonical = TRUE)
#   z <- gnr_resolve(y, canonical = FALSE)
# 
#   expect_is(w, "data.frame")
#   expect_is(z, "data.frame")
#   expect_named(w, c("user_supplied_name", "submitted_name", "data_source_title", "score", "matched_name2"))
#   expect_named(z, c("user_supplied_name", "submitted_name", "matched_name", "data_source_title", "score"))
#   expect_equal(NROW(x), 1)
#   expect_true(is.na(x$matched_name2[2]))
# })
# 
# test_that("fields parameter works correctly", {
#   skip_on_cran()
# 
#   tmp1 <- gnr_resolve(c("Asteraceae", "Plantae"), fields = 'all')
#   tmp2 <- gnr_resolve(c("Asteraceae", "Plantae"), fields = 'minimal')
# 
#   expect_is(tmp1, "data.frame")
#   expect_is(tmp1$matched_name, "character")
#   expect_true(any(grepl("Asteraceae", tmp1$matched_name)))
#   expect_is(tmp2, "data.frame")
#   expect_is(tmp2$matched_name, "character")
#   expect_true(any(grepl("Asteraceae", tmp2$matched_name)))
#   expect_false(identical(tmp1$matched_name, tmp2$matched_name))
# 
#   expect_lt(NCOL(tmp2), NCOL(tmp1))
# })
# 
# test_that("works correctly when no data found for preferred data source", {
#   skip_on_cran()
# 
#   aa <- gnr_resolve("Scabiosa triandra", preferred_data_sources = c(21,24),
#     best_match_only = TRUE)
# 
#   expect_is(aa, "data.frame")
#   expect_equal(NROW(aa), 0)
#   expect_equal(length(attributes(aa)$not_known), 1)
# })

#' Retrieve the downstream taxa for a given taxon name or ID.
#'
#' This function uses a while loop to continually collect children taxa down
#' to the taxonomic rank that you specify in the `downto` parameter. You
#' can get data from ITIS (itis), GBIF (gbif), NCBI (ncbi), WORMS (worms),
#' or BOLD (bold). There is no method exposed by these four
#' services for getting taxa at a specific taxonomic rank, so we do it
#' ourselves here.
#'
#' @export
#' @param sci_id Vector of taxa names (character) or IDs (character or numeric)
#' to query.
#' @param db character; database to query. One or more of `itis`, `gbif`,
#' `ncbi`, `worms`, or `bold`. Note that each taxonomic  data source has
#' their own identifiers, so that if you provide the wrong `db` value for
#' the identifier
#' you could get a result, but it will likely be wrong (not what you were
#' expecting). If using ncbi, we recommend getting an API key; see
#' [taxize-authentication]
#' @param downto What taxonomic rank to go down to. One of: 'superkingdom',
#' 'kingdom', 'subkingdom','infrakingdom','phylum','division','subphylum',
#' 'subdivision','infradivision', 'superclass','class','subclass','infraclass',
#' 'superorder','order','suborder','infraorder','superfamily','family',
#' 'subfamily','tribe','subtribe','genus','subgenus','section','subsection',
#' 'species group','species','subspecies','variety','form','subvariety','race',
#' 'stirp', 'morph','aberration','subform', 'unspecified', 'no rank'
#' @param intermediate (logical) If `TRUE`, return a list of length two
#' with target taxon rank names, with additional list of data.frame's of
#' intermediate taxonomic groups. Default: `FALSE`
#' @param rows (numeric) Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this parameter is ignored if you pass in a
#' taxonomic id of any of the acceptable classes: tsn.
#' @param limit Number of records to return. Applies to gbif only.
#' default: 100. max: 1000. use in combination with the `start` parameter
#' @param start Record number to start at. Applies to gbif only. default: 0.
#' use in combination with the `limit` parameter
#' @param x Deprecated, see `sci_id`
#' @param ... Further args passed on to [itis_downstream()],
#' [gbif_downstream()], [ncbi_downstream()],
#' [worms_downstream()], or [bold_downstream()]
#'
#' @return A named list of data.frames with the downstream names of every
#' supplied taxa. You get an NA if there was no match in the database.
#'
#' @section Authentication:
#' See [taxize-authentication] for help on authentication
#' 
#' @section bold:
#' BEWARE: `db="bold"` scrapes the BOLD website, so may be unstable. That is,
#' one day it may work, and the next it may fail. Open an issue if you
#' encounter an error: https://github.com/ropensci/taxize/issues
#'
#' @examples \dontrun{
#' # Plug in taxon IDs
#' downstream(125732, db = 'worms', downto = 'species')
#' downstream(3451, db = 'bold', downto = 'species')
#' 
#' if (interactive()) {
#'
#' # Plug in taxon names
#' downstream("Apis", db = 'ncbi', downto = 'species')
#' downstream("Apis", db = 'itis', downto = 'species')
#' downstream("Apis", db = 'bold', downto = 'species')
#' downstream("Gadus", db = 'worms', downto = 'species')
#' downstream(c("Apis","Epeoloides"), db = 'itis', downto = 'species')
#' downstream("Ursus", db = 'gbif', downto = 'species')
#' downstream(get_gbifid("Ursus"), db = 'gbif', downto = 'species')
#'
#' # Many taxa
#' sp <- names_list("genus", 3)
#' downstream(sp, db = 'itis', downto = 'species')
#' downstream(sp, db = 'gbif', downto = 'species')
#'
#' # Both data sources
#' ids <- get_ids("Apis", db = c('gbif','itis'))
#' downstream(ids, downto = 'species')
#' ## same result
#' downstream(get_ids("Apis", db = c('gbif','itis')), downto = 'species')
#'
#' # Collect intermediate names
#' ## itis
#' downstream('Bangiophyceae', db="itis", downto="genus")
#' downstream('Bangiophyceae', db="itis", downto="genus", intermediate=TRUE)
#' downstream(get_tsn('Bangiophyceae'), downto="genus")
#' downstream(get_tsn('Bangiophyceae'), downto="genus", intermediate=TRUE)
#'
#' # Use the rows parameter
#' ## note how in the second function call you don't get the prompt
#' downstream("Poa", db = 'gbif', downto="species")
#' downstream("Poa", db = 'gbif', downto="species", rows=1)
#'
#' # use curl options
#' res <- downstream("Apis", db = 'gbif', downto = 'species', verbose = TRUE)
#' 
#' # Pagination
#' # GBIF limits queries to a maximum of 1000 records per request, so if
#' # there's more than 1000, use the start parameter
#' # Piper, taxonKey = 3075433
#' z1 <- downstream(3075433, db = 'gbif', downto = "species", limit=1000)
#' z2 <- downstream(3075433, db = 'gbif', downto = "species", limit=1000,
#'   start=1000)
#' z3 <- downstream(3075433, db = 'gbif', downto = "species", limit=1000,
#'   start=2000)
#' z4 <- downstream(3075433, db = 'gbif', downto = "species", limit=1000,
#'   start=3000)
#' NROW(rbind(z1[[1]], z2[[1]], z3[[1]], z4[[1]]))
#' }}
downstream <- function(...){
  UseMethod("downstream")
}

#' @export
#' @rdname downstream
downstream.default <- function(sci_id, db=NULL, downto=NULL,
                               intermediate=FALSE, rows=NA, x=NULL, ...) {
  nstop(downto, "downto")
  nstop(db)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "downstream(x)", with = "downstream(sci_id)")
    sci_id <- x
  }
  if (!is.null(x)) sci_id <- x
  switch(
    db,
    itis = {
      id <- process_stream_ids(sci_id, db, get_tsn, rows = rows, ...)
      stats::setNames(downstream(id, downto = tolower(downto),
                                 intermediate = intermediate, ...), sci_id)
    },
    gbif = {
      id <- process_stream_ids(sci_id, db, get_gbifid, rows = rows, ...)
      stats::setNames(downstream(id, downto = tolower(downto),
                                 intermediate = intermediate, ...), sci_id)
    },
    ncbi = {
      id <- process_stream_ids(sci_id, db, get_uid, rows = rows, ...)
      stats::setNames(downstream(id, downto = tolower(downto),
                                 intermediate = intermediate, ...), sci_id)
    },
    worms = {
      id <- process_stream_ids(sci_id, db, get_wormsid, rows = rows, ...)
      stats::setNames(downstream(id, downto = tolower(downto),
                                 intermediate = intermediate, ...), sci_id)
    },
    bold = {
      id <- process_stream_ids(as.character(sci_id), db, get_boldid, rows = rows, ...)
      stats::setNames(downstream(id, downto = tolower(downto),
                                 intermediate = intermediate, ...), sci_id)
    },
    stop("the provided db value was not recognised/is not supported",
         call. = FALSE)
  )
}

process_stream_ids <- function(input, db, fxn, ...){
  g <- tryCatch(as.numeric(as.character(input)), warning = function(e) e)
  if (
    inherits(g, "numeric") ||
    is.character(input) &&
    all(grepl("[[:digit:]]", input))
  ) {
    as_fxn <- switch(db, itis = as.tsn, gbif = as.gbifid,
      ncbi = as.uid, worms = as.wormsid, bold = as.boldid)
    as_fxn(input, check = FALSE)
  } else {
    eval(fxn)(input, ...)
  }
}

#' @export
#' @rdname downstream
downstream.tsn <- function(sci_id, db = NULL, downto = NULL,
                           intermediate = FALSE, ...) {
  warn_db(list(db = db), "itis")
  fun <- function(y, downto, intermediate, ...) {
    # return NA if NA is supplied
    if (is.na(y)) {
      NA
    } else {
		  itis_downstream(y, downto = downto,
		                  intermediate = intermediate, ...)
    }
  }
  out <- lapply(sci_id, fun, downto = downto, intermediate = intermediate, ...)
  structure(out, class = 'downstream', db = 'itis', .Names = sci_id)
}

#' @export
#' @rdname downstream
downstream.gbifid <- function(sci_id, db = NULL, downto = NULL,
                              intermediate = FALSE, limit = 100,
                              start = NULL, ...) {
  warn_db(list(db = db), "gbif")
  fun <- function(y, downto, intermediate, limit, start, ...){
    # return NA if NA is supplied
    if (is.na(y)) {
      NA
    } else {
      gbif_downstream(y, downto = downto,
                      intermediate = intermediate, limit = limit,
                      start = start, ...)
    }
  }
  out <- lapply(sci_id, fun, downto = downto, intermediate = intermediate,
    limit = limit, start = start, ...)
  structure(out, class = 'downstream', db = 'gbif')
}

#' @export
#' @rdname downstream
downstream.uid <- function(sci_id, db = NULL, downto = NULL,
                              intermediate = FALSE, ...) {
  warn_db(list(db = db), "ncbi")
  fun <- function(y, downto, intermediate, ...){
    # return NA if NA is supplied
    if (is.na(y)) {
      NA
    } else {
      ncbi_downstream(id = y, downto = downto,
                      intermediate = intermediate, ...)
    }
  }
  out <- lapply(sci_id, fun, downto = downto, intermediate = intermediate, ...)
  structure(out, class = 'downstream', db = 'ncbi')
}

#' @export
#' @rdname downstream
downstream.wormsid <- function(sci_id, db = NULL, downto = NULL,
                              intermediate = FALSE, ...) {
  warn_db(list(db = db), "worms")
  fun <- function(y, downto, intermediate, ...){
    # return NA if NA is supplied
    if (is.na(y)) {
      NA
    } else {
      worms_downstream(id = y, downto = downto,
                      intermediate = intermediate, ...)
    }
  }
  out <- lapply(sci_id, fun, downto = downto, intermediate = intermediate, ...)
  structure(out, class = 'downstream', db = 'worms')
}

#' @export
#' @rdname downstream
downstream.boldid <- function(sci_id, db = NULL, downto = NULL,
                              intermediate = FALSE, ...) {
  warn_db(list(db = db), "bold")
  fun <- function(y, downto, intermediate, ...){
    # return NA if NA is supplied
    if (is.na(y)) {
      NA
    } else {
      bold_downstream(id = y, downto = downto,
                      intermediate = intermediate, ...)
    }
  }
  out <- lapply(sci_id, fun, downto = downto, intermediate = intermediate, ...)
  structure(out, class = 'downstream', db = 'bold')
}

#' @export
#' @rdname downstream
downstream.ids <- function(sci_id, db = NULL, downto = NULL,
                           intermediate = FALSE, ...) {
  fun <- function(y, downto, intermediate, ...){
    # return NA if NA is supplied
    if (is.na(y)) {
      NA
    } else {
      downstream(y, downto = downto, intermediate = intermediate, ...)
    }
  }
  structure(lapply(sci_id, fun, downto = downto, intermediate = intermediate, ...),
            class = 'downstream_ids')
}

simp <- function(x) if (length(x) == 1) x[[1]] else x

#' Retrieve the lowest common taxon and rank for a given taxon name or ID
#'
#' @export
#' @param sci_id Vector of taxa names (character) or id (character or numeric) to
#' query.
#' @param db character; database to query. either `ncbi`, `itis`, `gbif`,
#' `tol`. If using ncbi, we recommend getting an API key;
#' see [taxize-authentication]
#' @param rows (numeric) Any number from 1 to infinity. If the default NA,
#' all rows are considered. Note that this parameter is ignored if you pass in
#' a taxonomic id of any of the acceptable classes: tsn, gbifid, tolid.
#' NCBI has a method for this function but rows doesn't work.
#' @param class_list (list) A list of classifications, as returned from
#' [classification()]
#' @param low_rank (character) taxonomic rank to return, of length 1
#' @param x Deprecated, see `sci_id`
#' @param ... Other arguments passed to [get_tsn()], [get_uid()],
#' [get_gbifid()], [get_tolid()]
#'
#' @return NA when no match, or a data.frame with columns
#' * name
#' * rank
#' * id
#' 
#' @section Authentication:
#' See [taxize-authentication] for help on authentication
#' 
#' @author Jimmy O'Donnell \email{jodonnellbio@@gmail.com}
#' Scott Chamberlain
#' @examples \dontrun{
#' id <- c("9031", "9823", "9606", "9470")
#' id_class <- classification(id, db = 'ncbi')
#' lowest_common(id[2:4], db = "ncbi")
#' lowest_common(id[2:4], db = "ncbi", low_rank = 'class')
#' lowest_common(id[2:4], db = "ncbi", low_rank = 'family')
#' lowest_common(id[2:4], class_list = id_class)
#' lowest_common(id[2:4], class_list = id_class, low_rank = 'class')
#' lowest_common(id[2:4], class_list = id_class, low_rank = 'family')
#'
#' # TOL
#' taxa <- c("Angraecum sesquipedale", "Dracula vampira",
#'   "Masdevallia coccinea")
#' (cls <- classification(taxa, db = "tol"))
#' lowest_common(taxa, db = "tol", class_list = cls)
#' lowest_common(get_tolid(taxa), class_list = cls)
#' xx <- get_tolid(taxa)
#' lowest_common(xx, class_list = cls)
#'
#' spp <- c("Sus scrofa", "Homo sapiens", "Nycticebus coucang")
#' lowest_common(spp, db = "ncbi")
#' lowest_common(get_uid(spp))
#'
#' lowest_common(spp, db = "itis")
#' lowest_common(get_tsn(spp))
#'
#' gbifid <- c("2704179", "3119195")
#' lowest_common(gbifid, db = "gbif")
#'
#' spp <- c("Poa annua", "Helianthus annuus")
#' lowest_common(spp, db = "gbif")
#' lowest_common(get_gbifid(spp))
#'
#' cool_orchid <- c("Angraecum sesquipedale", "Dracula vampira",
#'   "Masdevallia coccinea")
#' orchid_ncbi <- get_uid(cool_orchid)
#' orchid_gbif <- get_gbifid(cool_orchid)
#'
#' cool_orchids2 <- c("Domingoa haematochila", "Gymnadenia conopsea",
#'   "Masdevallia coccinea")
#' orchid_itis <- get_tsn(cool_orchids2)
#'
#' orchid_hier_ncbi <- classification(orchid_ncbi, db = 'ncbi')
#' orchid_hier_gbif <- classification(orchid_gbif, db = 'gbif')
#' orchid_hier_itis <- classification(orchid_itis, db = 'itis')
#'
#' lowest_common(orchid_ncbi, low_rank = 'class')
#' lowest_common(orchid_ncbi, class_list = orchid_hier_ncbi,
#'   low_rank = 'class')
#' lowest_common(orchid_gbif, low_rank = 'class')
#' lowest_common(orchid_gbif, orchid_hier_gbif, low_rank = 'class')
#' lowest_common(get_uid(cool_orchid), low_rank = 'class')
#' lowest_common(get_uid(cool_orchid), low_rank = 'family')
#'
#' lowest_common(orchid_ncbi, class_list = orchid_hier_ncbi,
#'   low_rank = 'subfamily')
#' lowest_common(orchid_gbif, class_list = orchid_hier_gbif,
#'   low_rank = 'subfamily')
#'
#' lowest_common(orchid_itis, class_list = orchid_hier_itis,
#'   low_rank = 'class')
#'
#' ## Pass in sci. names
#' nms <- c("Angraecum sesquipedale", "Dracula vampira", "Masdevallia coccinea")
#' lowest_common(x = nms, db = "ncbi")
#' lowest_common(x = nms, db = "gbif")
#' }
lowest_common <- function(...){
  UseMethod("lowest_common")
}

#' @export
#' @rdname lowest_common
#' @method lowest_common default
lowest_common.default <- function(sci_id, db = NULL, rows = NA, class_list = NULL,
                                  low_rank = NULL, x = NULL, ...) {
  if (is.null(db)) if (!is.null(class_list)) db <- attr(class_list, "db")
  nstop(db)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "lowest_common(x)", with = "lowest_common(sci_id)")
    sci_id <- x
  }
  
  if (!is.null(x)) sci_id <- x
  switch(
    db,
    itis = {
      id <- process_lowest_ids(sci_id, db, get_tsn, rows = rows, ...)
      lowest_common(id, class_list, ...)
    },
    ncbi = {
      id <- process_lowest_ids(sci_id, db, get_uid, rows = rows, ...)
      lowest_common(id, class_list, ...)
    },
    gbif = {
      id <- process_lowest_ids(sci_id, db, get_gbifid, rows = rows, ...)
      lowest_common(id, class_list, ...)
    },
    tol = {
      id <- process_lowest_ids(sci_id, db, get_tolid, rows = rows, ...)
      lowest_common(id, class_list, ...)
    },
    stop("the provided db value was not recognised", call. = FALSE)
  )
}

#' @export
#' @rdname lowest_common
lowest_common.uid <- function(sci_id, class_list = NULL, low_rank = NULL, ...) {
  check_lowest_ids(sci_id)
  class_list <- get_class(sci_id, class_list, db = "uid", ...)
  lc_helper(sci_id, class_list, low_rank, ...)
}

#' @export
#' @rdname lowest_common
lowest_common.tsn <- function(sci_id, class_list = NULL, low_rank = NULL, ...) {
  check_lowest_ids(sci_id)
  class_list <- get_class(sci_id, class_list, db = "itis", ...)
  lc_helper(sci_id, class_list, low_rank, ...)
}

#' @export
#' @rdname lowest_common
lowest_common.gbifid <- function(sci_id, class_list = NULL, low_rank = NULL, ...) {
  check_lowest_ids(sci_id)
  class_list <- get_class(sci_id, class_list, db = "gbif", ...)
  lc_helper(sci_id, class_list, low_rank, ...)
}

#' @export
#' @rdname lowest_common
lowest_common.tolid <- function(sci_id, class_list = NULL, low_rank = NULL, ...) {
  check_lowest_ids(sci_id)
  class_list <- get_class(sci_id, class_list, db = "tol", ...)
  names(class_list) <- sci_id
  lc_helper(sci_id, class_list, low_rank, ...)
}

# helpers -------------------------------------------------
lc_helper <- function(x, class_list, low_rank = NULL, ...) {
  idc <- class_list[x]
  # next line NCBI specific
  cseq <- vapply(idc, function(x) x[1, 1] != "unclassified sequences",
                 logical(1))
  idc <- idc[cseq]
  if (is.null(low_rank)) {
    x_row <- length(Reduce(intersect, lapply(idc, "[[", 1)))
    x <- idc[[1]][x_row, ]
    if (x[1, "rank"] == "no rank") {
      x[1, "rank"] <- next_best_taxon(idc[[1]][1:x_row, ])
    }
    return(x)
  } else {
    valid_ranks <- tolower(ritis::rank_names()$rankname)
    if (!(low_rank %in% valid_ranks)) {
      warning('the supplied rank is not valid')
    }
    # low_rank_names <- as.character(unique(unlist(lapply(idc, function(x) x$name[which(x$rank == low_rank)]))))
    low_rank_names <- unique(setDF(rbindlist(lapply(idc, function(x)
      x[which(x$rank == low_rank),]))))
    if (NROW(low_rank_names) == 1) {
      return(low_rank_names)
    } else {
      return(NA)
    }
  }
}

next_best_taxon <- function(x){
  paste("below-",
        tail(x[, "rank"][!duplicated(x[, "rank"])], n = 1
        ), sep = "")
}

get_class <- function(x, y, db, ...) {
  if (is.null(y)) {
    if (db == "uid") db <- "ncbi"
    classification(x, db = db, ...)
  } else {
    yattr <- sub("ncbi", "uid", attr(y, "db"))
    if (yattr != db) {
      stop(sprintf("class_list input must be of class '%s'", db), call. = FALSE)
    }
    y
  }
}

check_lowest_ids <- function(x) {
  notmiss <- na.omit(x)
  if (length(notmiss) < 2) {
    stop(length(notmiss), " found, ", length(x) - length(notmiss),
         " were NA; > 1 needed", call. = FALSE)
  }
}

process_lowest_ids <- function(input, db, fxn, ...) {
  g <- tryCatch(as.numeric(as.character(input)), warning = function(e) e)
  if (inherits(g, "condition")) eval(fxn)(input, ...)
  if (is.numeric(g) || is.character(input) && all(grepl("[[:digit:]]", input))) {
    as_fxn <- switch(db, itis = as.tsn, gbif = as.gbifid,
      ncbi = as.uid, tol = as.tolid)
    as_fxn(input, check = FALSE)
  } else {
    eval(fxn)(input, ...)
  }
}

context("lowest_common")

force_http1_1 <- list(http_version = 2L)

test_that("lowest_common works with ncbi, passing in classifications and doing internally", {
  skip_on_cran()
  skip_on_travis()

  id <- c("9031", "9823", "9606", "9470")
  vcr::use_cassette("classification_lowest_common_ncbi", {
    idc <- classification(id, db = 'ncbi', callopts = force_http1_1)
  })
  
  aa <- lowest_common(id[2:4], db = "ncbi")
  bb <- lowest_common(id[2:4], db = "ncbi", low_rank = 'class')
  cc <- lowest_common(id[2:4], db = "ncbi", class_list = idc, low_rank = 'class')

	expect_is(aa, "data.frame")
  expect_is(bb, "data.frame")
  expect_is(cc, "data.frame")
  expect_named(aa, c('name', 'rank', 'id'))
  expect_named(cc, c('name', 'rank', 'id'))

  expect_identical(aa, bb)
  expect_identical(bb, cc)

  expect_equal(NROW(aa), 1)

  # passing in classification list obs. takes less time
  expect_lt(
    system.time(lowest_common(id[2:4], db = "ncbi", class_list = idc, low_rank = 'class'))[3],
    system.time(lowest_common(id[2:4], db = "ncbi", low_rank = 'class'))[3]
  )
})

test_that("lowest_common works with itis", {
  skip_on_cran()

  #lowest_common(spp, db = "itis")
  #spp <- c("Sus scrofa", "Homo sapiens", "Nycticebus coucang")

  ids <- c("180722","180092","572890")
  vcr::use_cassette("classification_lowest_common_itis", {
    idc <- classification(ids, db = 'itis')
  }, preserve_exact_body_bytes = TRUE, match_requests_on = c("method", "query"))

  expect_identical(
    lowest_common(ids, db = "itis"),
    lowest_common(ids, db = "itis", class_list = idc)
  )

  bb <- lowest_common(ids, db = "itis", low_rank = 'class')
  cc <- lowest_common(ids, db = "itis", class_list = idc, low_rank = 'class')

  expect_is(bb, "data.frame")
  expect_is(cc, "data.frame")
  expect_named(cc, c('name', 'rank', 'id'))

  expect_identical(bb, cc)

  expect_equal(NROW(bb), 1)
})

#' Aggregate data by given taxonomic rank
#'
#' @export
#' @param data A data.frame. Column headers must have capitalized ranks (e.g.,
#'    Genus, Tribe, etc.) (data.frame)
#' @param datacol The data column (character)
#' @param rank Taxonomic rank to aggregate by (character)
#' @param fxn Arithmetic function or vector or functions (character)
#' @examples
#' if (require(vegan)) {
#' data(dune.taxon, dune, package='vegan')
#' dat <- dune.taxon
#' dat$abundance <- colSums(dune)
#' rankagg(data=dat, datacol="abundance", rank="Genus")
#' rankagg(data=dat, "abundance", rank="Family")
#' rankagg(data=dat, "abundance", rank="Genus", fxn="mean")
#' rankagg(data=dat, "abundance", rank="Subclass")
#' rankagg(data=dat, "abundance", rank="Subclass", fxn="sd")
#' }
rankagg <- function(data=NULL, datacol=NULL, rank=NULL, fxn="sum") {
  if (is.null(data) | is.null(rank))
    stop("You must specify your data.frame and taxonomic rank")
  rank <- match.arg(rank, choices = c('Species', 'Genus', 'Tribe', 'Family',
                                      'Order', 'Subclass', 'Class', 'Phylum',
                                      'Kingdom'))
  fn <- match.fun(fxn)
  out <- aggregate(data[ , datacol], list(data[ , rank]), fxn)
  names(out) = c(rank, fxn)
  return(out)
}

context("rankagg")

test_that("rankagg throws error", {
  skip_on_cran()
  skip_if_not_installed("vegan")

  data(dune.taxon, package = 'vegan')
  dat <- dune.taxon
  set.seed(1234)
  dat$abundance <- round(rlnorm(n = nrow(dat), meanlog = 5, sdlog = 2), 0)
  out <- rankagg(data = dat, datacol = "abundance", rank = "Genus")

  expect_error(rankagg(datacol = "abundance", rank = "Genus"))
  expect_error(rankagg(data = dat, datacol = "abundance"))

  expect_is(out, "data.frame")
  expect_equal(out[1, 2], 13)
})

#' Search Catalogue of Life for taxonomic classifications.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @rdname col_classification-defunct
#' @keywords internal
col_classification <- function(...) {
  .Defunct(msg = "This function is defunct. See classification().")
}

#' Return all synonyms for a taxon name with a given id.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @rdname tp_classification-defunct
#' @keywords internal
tp_classification <- function(...) {
  .Defunct(msg = "This function is defunct. See classification().")
}

#' Get full ITIS record for one or more ITIS TSN's or lsid's.
#'
#' @export
#' @param values (character) One or more TSN's (taxonomic serial number) or lsid's for a
#' taxonomic group
#' @param by (character) By "tsn" (default) or "lsid"
#' @param ... Further arguments passed on to [ritis::full_record]
#' @details You can only enter values in tsn parameter or lsid, not both.
#' @examples \dontrun{
#' # by TSN
#' itis_getrecord(202385)
#' itis_getrecord(c(202385,70340))
#'
#' # by lsid
#' itis_getrecord("urn:lsid:itis.gov:itis_tsn:202385", "lsid")
#' }

itis_getrecord <- function(values, by="tsn", ...) {
  if (!by %in% c("tsn","lsid")) stop("by must be one of 'tsn' or 'lsid'", call. = FALSE)
  temp <- switch(
    by,
    tsn = lapply(values, ritis::full_record, ...),
    lsid = lapply(values, function(x) ritis::full_record(lsid = x, ...))
  )
  if (length(values) == 1) {
    temp[[1]]
  } else {
    setNames(temp, values)
  }
}

context("itis_getrecord")

test_that("itis_getrecord returns the correct class", {
  skip_on_cran()
  vcr::use_cassette("itis_getrecord", {
      one <- itis_getrecord(202385, verbose=FALSE)
      two <- itis_getrecord(c(202385,70340), verbose=FALSE)
      three <- itis_getrecord("urn:lsid:itis.gov:itis_tsn:202385", "lsid",
                              verbose=FALSE)
  })

  expect_that(one, is_a("list"))
  expect_that(two, is_a("list"))
  expect_that(three, is_a("list"))
})

#' @title Search for taxonomic names using the Global Names Architecture
#'
#' @description Uses the Global Names Index, see http://gni.globalnames.org
#'
#' @export
#' @param sci (character) required. Name pattern you want to search
#' for. WARNING: Does not work for common names. Search term may include
#' following options:
#'
#' * `n`: A shortcut that allows to put together several elements (e.g., `n:B. bubo Linn. 1750-1800`)
#' * `g`: a genus name. (e.g. `g:B.`, `g:Bub.`, `g:Bubo`)
#' * `isp`: an infraspecies name (e.g. `sp:bubo`, `sp:gallop.`)
#' * `asp`: either species or infraspecies (all sp) (e.g. `asp:bubo`)
#' * `ds`: data-sources IDs (e.g., `ds:1,2,3`)
#' * `tx`: parent taxon . Uses classification of the first data-source from `ds`. If data-sources are not set, uses Catalogue of Life. (e.g. `tx:Aves`)
#' * `au`: author - Search by author word	(e.g. `au:Linnaeus`, `au:Linn.`)
#' * `y`: year - Search by year (e.g. `y:2005`)
#'
#' @param justtotal Return only the total results found.
#' @param parse_names If `TRUE` use [gni_parse()] on the outputs.
#' 
#' @param ... Curl options passed on to [crul::verb-GET]
#' @author Scott Chamberlain, Zachary Foster
#' @return data.frame of results.
#' @seealso [gnr_datasources()], [gna_search()]
#' @keywords globalnamesindex names taxonomy
#' @references http://gni.globalnames.org/
#' https://apidoc.globalnames.org/gnames
#' 
#' @examples \dontrun{
#' gna_search('n:B. bubo ds:1,2 au:Linn. y:1700-')
#' }
gna_search <- function(sci, justtotal = FALSE, parse_names = FALSE, ...) {
  
  query <- tc(list(search_term = sci))
  cli <- crul::HttpClient$new('https://verifier.globalnames.org',
                              headers = tx_ual, opts = list(...))
  tt <- cli$get(path = paste0('/api/v1/search/', curl::curl_escape(sci)))
  tt$raise_for_status()
  out <- jsonlite::fromJSON(tt$parse("UTF-8"), FALSE)
  if (justtotal) {
    return(out$metadata$namesNumber)
  } else {
    df <- do.call(rbind, lapply(out$names, function(x) {
      all_feilds <- unlist(x)
      names(all_feilds) <- vapply(strsplit(names(all_feilds), split = '.', fixed = TRUE), function(x) x[length(x)], FUN.VALUE = character(1))
      all_feilds <- all_feilds[! duplicated(names(all_feilds))]
      return(as.data.frame(as.list(all_feilds)))
    }))
    df <- colClasses(df, "character")
    df <- tibble::as_tibble(df)
    if (NROW(df) != 0) {
      names(df) <- c("id", "name", "cardinality", "matchType", "dataSourceId", "dataSourceTitleShort", 
                     "curation", "recordId", "outlink", "entryDate", "sortScore", 
                     "matchedNameID", "matchedName", "matchedCardinality", "currentRecordId", 
                     "currentNameId", "currentName", "currentCardinality", "currentCanonicalSimple", 
                     "currentCanonicalFull", "taxonomicStatus", "isSynonym", "classificationPath", 
                     "classificationRanks", "classificationIds", "editDistance", "stemEditDistance", 
                     "cardinalityScore", "infraSpecificRankScore", "fuzzyLessScore", 
                     "curatedDataScore", "authorMatchScore", "acceptedNameScore", 
                     "parsingQualityScore", "dataSourcesNum", "dataSourcesIds")
    }
    
    if (parse_names) {
      data.frame(df, gni_parse(as.character(df$name)), stringsAsFactors = FALSE)
    } else {
      df
    }
  }
}

#' Search for taxonomic names using the Global Names Index
#'
#' THIS FUNCTION IS DEFUNCT.
#' 
#' @export
#' @keywords internal
gni_seach <- function(names, ...) {
  .Defunct("ncbi_searcher", "traits",
           msg = "This function is defunct. See gna_search()")
}

context("gna_search")

test_that("gna_search returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("gna_search", {
    x <- gna_search('n:B. bubo ds:1,2 au:Linn. y:1700-') 
  })

	expect_is(x, "data.frame")
  expect_equal(x$id[1], "4431a0f3-e901-519a-886f-9b97e0c99d8e")
})

#' Return all distribution records for for a taxon name with a given id.
#'
#' @export
#' @param id the taxon identifier code
#' @param key Your Tropicos API key; See [taxize-authentication] 
#' for help on authentication
#' @param ... Curl options passed on to [crul::HttpClient]
#' @return List of two data.frame's, one named "location", and one "reference".
#' @references http://services.tropicos.org/help?method=GetNameDistributionsXml
#'
#' @examples \dontrun{
#' # Query using a taxon name Id
#' out <- tp_dist(id = 25509881)
#' ## just location data
#' head(out[['location']])
#' ## just reference data
#' head(out[['reference']])
#' }

tp_dist <- function(id, key=NULL, ...) {
  id <- as.numeric(as.character(id))
  if (!inherits(id, "numeric")) {
    stop("You must supply a numeric taxon name id")
  }

  url = sprintf('http://services.tropicos.org/Name/%s/Distributions', id)
	key <- getkey(key, "TROPICOS_KEY")
  args <- tc(list(format = 'json', apikey = key))
  tt <- tp_GET(url, args, ...)
  out <- jsonlite::fromJSON(tt, FALSE)
  getdata <- function(x, which) data.frame(x[[which]])
  locs <- dt2df(lapply(out, getdata, which = "Location"), idcol = FALSE)
  names(locs) <- tolower(names(locs))
  refs <- dt2df(lapply(out, getdata, which = "Reference"), idcol = FALSE)
  names(refs) <- tolower(names(refs))

  list(location = locs, reference = refs)
}

#' Return all distribution records for for a taxon name with a given id.
#'
#' Function name changed to tp_dist.
#'
#' @export
#' @keywords internal
#' @param ... ignored
#' @rdname tp_namedistributions-deprecated
tp_namedistributions <- function(...) {
  .Deprecated("tp_dist", "taxize", "Function name changed. See tp_dist", "tp_namedistributions")
}

context("tp_dist")

test_that("tp_dist returns the correct class", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tp_dist", {
    tt <- suppressMessages(tp_dist(id = 25509881))
  }, preserve_exact_body_bytes = TRUE)

	expect_that(tt, is_a("list"))
	expect_that(tt[['location']], is_a("data.frame"))
	expect_that(tt[['reference']], is_a("data.frame"))
	expect_true(any(grepl("abbreviatedtitle", names(tt[['reference']]))))
})

#' Return all accepted names for a taxon name with a given id.
#'
#' @export
#' @param id the taxon identifier code
#' @param key Your Tropicos API key; See [taxize-authentication] 
#' for help on authentication
#' @param ... Curl options passed on to [crul::verb-GET]
#' @return List or dataframe.
#' @examples \dontrun{
#' tp_accnames(id = 25503923)
#' tp_accnames(id = 25538750)
#'
#' # No accepted names found
#' tp_accnames(id = 25509881)
#' }

tp_accnames <- function(id, key = NULL, ...) {
  url = sprintf('http://services.tropicos.org/Name/%s/AcceptedNames', id)
	key <- getkey(key, "TROPICOS_KEY")
  args <- tc(list(apikey = key, format = 'json'))
  tt <- tp_GET(url, args, ...)
  res <- jsonlite::fromJSON(tt, FALSE)

  if ("Error" %in% names(res[[1]])) {
    res[[1]]
  } else {
    vvv <- lapply(res, getdata)
    syns <- dt2df(lapply(vvv, "[[", "syn"), idcol = FALSE)
    accs <- dt2df(lapply(vvv, "[[", "acc"), idcol = FALSE)
    refs <- dt2df(lapply(vvv, "[[", "ref"), idcol = FALSE)
    list(synonyms = syns, acceptednames = accs, reference = refs)
  }
}

getdata <- function(x) {
  syn <- data.frame(x$SynonymName, stringsAsFactors = FALSE)
  names(syn) <- tolower(names(syn))
  acc <- data.frame(x$AcceptedName, stringsAsFactors = FALSE)
  names(acc) <- tolower(names(acc))
  ref <- data.frame(x$Reference, stringsAsFactors = FALSE)
  names(ref) <- tolower(names(ref))
  list(syn = syn, acc = acc, ref = ref)
}

#' Return all accepted names for a taxon name with a given id.
#'
#' Function name changed to tp_accnames.
#'
#' @export
#' @keywords internal
#' @param ... ignored
#' @rdname tp_acceptednames-deprecated
tp_acceptednames <- function(...) {
  .Deprecated("tp_accnames", "taxize", "Function name changed. See tp_accnames", "tp_acceptednames")
}

context("tp_accnames")

test_that("tp_accnames returns the correct class", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tp_accnames", {
    out <- suppressMessages(tp_accnames(id = 25503923))
  }, preserve_exact_body_bytes = TRUE)

  if ("Error" %in% names(out)) skip("error in tp_accnames call - skipping")

  expect_that(out, is_a("list"))
  expect_that(out$synonyms, is_a("data.frame"))
  expect_that(out$acceptednames, is_a("data.frame"))
  expect_that(out$reference, is_a("data.frame"))
  
	expect_is(out$reference$referenceid, "integer")
	expect_that(out$reference$abbreviatedtitle, is_a("character"))
})

#' Get the NameID codes from Tropicos for taxonomic names.
#'
#' @export
#' @param sci (character) One or more scientific name's as a vector or list. Or,
#' a `taxon_state` object (see [taxon-state])
#' @param ask logical; should get_tpsid be run in interactive mode?
#' If TRUE and more than one ID is found for the species, the user is asked for
#' input. If FALSE NA is returned for multiple matches.
#' @param messages logical; If TRUE the actual taxon queried is printed on the console.
#' @param key Your API key; see [taxize-authentication]
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all rows are considered.
#' Note that this function still only gives back a tpsid class object with one to many identifiers.
#' See [get_tpsid_()] to get back all, or a subset, of the raw data that you are
#' presented during the ask process.
#' @param family (character) A family name. Optional. See `Filtering` below.
#' @param rank (character) A taxonomic rank name. See [rank_ref] for possible
#' options. Though note that some data sources use atypical ranks, so inspect the
#' data itself for options. Optional. See `Filtering` below.
#' @param sciname Deprecated, see `sci`
#' @param ... Other arguments passed to [tp_search()].
#' @param x Input to [as.tpsid()]
#' @param check logical; Check if ID matches any existing on the DB, only used in
#' [as.tpsid()]
#' @template getreturn
#'
#' @section Filtering:
#' The parameters `family` an`rank`nk are not used in the search to the data
#' provider, but are used in filtering the data down to a subset that is closer to the
#' target you want.  For all these parameters,
#' you can use regex strings since we use [grep()] internally to match.
#' Filtering narrows down to the set that matches your query, and removes the rest.
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @author Scott Chamberlain, 
#'
#' @examples \dontrun{
#' get_tpsid(sci='Poa annua')
#' get_tpsid(sci='Pinus contorta')
#'
#' get_tpsid(c("Poa annua", "Pinus contorta"))
#'
#' # specify rows to limit choices available
#' get_tpsid('Poa ann')
#' get_tpsid('Poa ann', rows=1)
#' get_tpsid('Poa ann', rows=25)
#' get_tpsid('Poa ann', rows=1:2)
#'
#' # When not found, NA given (howdy is not a species name, and Chrinomus is a fly)
#' get_tpsid("howdy")
#' get_tpsid(c("Chironomus riparius", "howdy"))
#'
#' # Narrow down results to a division or rank, or both
#' ## Satyrium example
#' ### Results w/o narrowing
#' get_tpsid("Satyrium")
#' ### w/ rank
#' get_tpsid("Satyrium", rank = "var.")
#' get_tpsid("Satyrium", rank = "sp.")
#'
#' ## w/ family
#' get_tpsid("Poa")
#' get_tpsid("Poa", family = "Iridaceae")
#' get_tpsid("Poa", family = "Orchidaceae")
#' get_tpsid("Poa", family = "Orchidaceae", rank = "gen.")
#'
#' # Fuzzy filter on any filtering fields
#' ## uses grep on the inside
#' get_tpsid("Poa", family = "orchidaceae")
#' get_tpsid("Aga", fuzzy = TRUE, parent = "*idae")
#'
#' # pass to classification function to get a taxonomic hierarchy
#' classification(get_tpsid(sci='Poa annua'))
#'
#' # Convert a tpsid without class information to a tpsid class
#' as.tpsid(get_tpsid("Pinus contorta")) # already a tpsid, returns the same
#' as.tpsid(get_tpsid(c("Chironomus riparius","Pinus contorta"))) # same
#' as.tpsid(24900183) # numeric
#' as.tpsid(c(24900183,50150089,50079838)) # numeric vector, length > 1
#' as.tpsid("24900183") # character
#' as.tpsid(c("24900183","50150089","50079838")) # character vector, length > 1
#' as.tpsid(list("24900183","50150089","50079838")) # list, either numeric or character
#' ## dont check, much faster
#' as.tpsid("24900183", check=FALSE)
#' as.tpsid(24900183, check=FALSE)
#' as.tpsid(c("24900183","50150089","50079838"), check=FALSE)
#' as.tpsid(list("24900183","50150089","50079838"), check=FALSE)
#'
#' (out <- as.tpsid(c(24900183,50150089,50079838)))
#' data.frame(out)
#' as.tpsid( data.frame(out) )
#'
#' # Get all data back
#' get_tpsid_("Poa annua")
#' get_tpsid_("Poa annua", rows=2)
#' get_tpsid_("Poa annua", rows=1:2)
#' get_tpsid_(c("asdfadfasd","Pinus contorta"), rows=1:5)
#'
#' # use curl options
#' invisible(get_tpsid("Quercus douglasii", messages = TRUE))
#' }

get_tpsid <- function(sci, ask = TRUE, messages = TRUE, key = NULL,
  rows = NA, family = NULL, rank = NULL, sciname = NULL, ...) {

  assert(sci, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(family, "character")
  assert(rank, "character")
  assert_rows(rows)
  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_tpsid(sciname)", with = "get_tpsid(sci)")
    sci <- sciname
  }
  

  if (inherits(sci, "character")) {
    tstate <- taxon_state$new(class = "tpsid", names = sci)
    items <- sci
  } else {
    assert_state(sci, "tpsid")
    tstate <- sci
    sci <- tstate$taxa_remaining()
    items <- c(sci, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci[i], "'\n")
    tmp <- tp_search(sci = sci[i], key = key, ...)
    mm <- NROW(tmp) > 1

    if (
      all(names(tmp)[[1]] == "error") ||
      all(is.na(tmp)) ||
      inherits(tmp, "character")
    ) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      att <- "not found"
    } else {
      df <- tmp[, c('nameid','scientificname','family','rankabbreviation',
                   'nomenclaturestatusname','author','displaydate')]
      names(df) <- c('tpsid','name','family','rank','status','author','date')
      id <- df$tpsid
      att <- "found"
    }

    # not found on tropicos
    if (length(id) == 0) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      att <- "not found"
    }
    # more than one found on tropicos -> user input
    if (length(id) > 1) {

        if (!is.null(family) || !is.null(rank)) {
          df <- filt(df, "family", family)
          df <- filt(df, "rank", rank)
        }

        df <- sub_rows(df, rows)
        id <- df$tpsid
        if (length(id) == 1) {
          rank_taken <- as.character(df$rank)
          direct <- TRUE
          att <- "found"
        }

        # more than one, try for direct match
        if (length(id) > 1) {
          matchtmp <- df[tolower(df$name) %in% tolower(sci[i]), "tpsid"]
          if (length(matchtmp) == 1) {
            id <- matchtmp
            direct <- TRUE
            att <- "found"
          }
        }

        if (length(id) > 1) {
          if (ask) {
            # prompt
            rownames(df) <- 1:nrow(df)
            message("\n\n")
            message("\nMore than one tpsid found for taxon '", sci[i], "'!\n
          Enter rownumber of taxon (other inputs will return 'NA'):\n")
            rownames(df) <- 1:nrow(df)
            message(paste0(utils::capture.output(df), collapse = "\n"))
            take <- scan(n = 1, quiet = TRUE, what = 'raw')

            if (length(take) == 0) {
              take <- 'notake'
              att <- 'nothing chosen'
            }
            if (take %in% seq_len(nrow(df))) {
              take <- as.numeric(take)
              message("Input accepted, took tpsid '", as.character(df$tpsid[take]), "'.\n")
              id <- as.character(df$tpsid[take])
              att <- "found"
            } else {
              id <- NA_character_
              mssg(messages, "\nReturned 'NA'!\n\n")
              att <- "not found"
            }
          } else {
            if (length(id) != 1) {
              warning(sprintf(m_more_than_one_found, "tpsid", sci[i]),
                call. = FALSE)
              id <- NA_character_
              att <- m_na_ask_false
            }
          }
        }
    }
    res <- list(id = as.character(id), att = att, multiple = mm, direct = direct)
    prog$completed(sci[i], att)
    prog$prog(att)
    tstate$add(sci[i], res)
  }
  out <- tstate$get()
  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "tpsid",
                   match = pluck_un(out, "att", ""),
                   multiple_matches = pluck_un(out, "multiple", logical(1)),
                   pattern_match = pluck_un(out, "direct", logical(1)))
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$tropicos)
}

#' @export
#' @rdname get_tpsid
as.tpsid <- function(x, check=TRUE) UseMethod("as.tpsid")

#' @export
#' @rdname get_tpsid
as.tpsid.tpsid <- function(x, check=TRUE) x

#' @export
#' @rdname get_tpsid
as.tpsid.character <- function(x, check=TRUE) if(length(x) == 1) make_tpsid(x, check) else collapse(x, make_tpsid, "tpsid", check=check)

#' @export
#' @rdname get_tpsid
as.tpsid.list <- function(x, check=TRUE) if(length(x) == 1) make_tpsid(x, check) else collapse(x, make_tpsid, "tpsid", check=check)

#' @export
#' @rdname get_tpsid
as.tpsid.numeric <- function(x, check=TRUE) as.tpsid(as.character(x), check)

#' @export
#' @rdname get_tpsid
as.tpsid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class="tpsid", match=x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri=x$uri)
}

#' @export
#' @rdname get_tpsid
as.data.frame.tpsid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "tpsid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_tpsid <- function(x, check=TRUE) make_generic(x, 'http://tropicos.org/Name/%s', "tpsid", check)

check_tpsid <- function(x){
  res <- tp_summary(x)
  !identical(names(res), "error")
}

#' @export
#' @rdname get_tpsid
get_tpsid_ <- function(sci, messages = TRUE, key = NULL, rows = NA,
  sciname = NULL, ...) {

  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_tpsid_(sciname)", with = "get_tpsid_(sci)")
    sci <- sciname
  }
  stats::setNames(lapply(sci, get_tpsid_help, messages = messages, key=key,
    rows = rows, ...), sci)
}

get_tpsid_help <- function(sci, messages, key, rows, ...){
  mssg(messages, "\nRetrieving data for taxon '", sci, "'\n")
  df <- tp_search(sci=sci, key=key, ...)
  if("error" %in% names(df)) NULL else sub_rows(df, rows)
}

context("get_tpsid")

test_that("get_tpsid returns the correct value", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("get_tpsid", {
    a <- get_tpsid(sci='Helianthus excubitor', messages=FALSE)[[1]]
    b <- get_tpsid(sci='adsf asdf asdf', messages=FALSE)[[1]]
    d <- get_tpsid(c("Helianthus excubitor", "adsf asdf asdf"), messages=FALSE)
  })
  
  expect_equal(a, "50230899")
  expect_true(is.na(b))
  expect_is(d, "tpsid")
})

test_that("get_tpsid accepts ask-argument", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("get_tpsid_ask_arg", {
    a <- get_tpsid(sci='adsf asdf asdf', ask=FALSE, messages=FALSE)[[1]]
  })

  expect_true(is.na(a))
})

test_that("get_tpsid behaves correctly on dot inputs", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("get_tpsid_warnings_dots", {
    expect_that(get_tpsid('Pinus contorta var. yukonensis', messages=FALSE),
                gives_warning("detected, being URL encoded"))
    expect_warning(get_tpsid('Pinus contorta yukonensis'), NA)
  }, preserve_exact_body_bytes = TRUE)
})

test_that("get_tpsid behaves correctly on subspecific inputs", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("get_tpsid_warnings_subspecific", {
    expect_that(get_tpsid('Poa annua var annua', messages=FALSE),
                gives_warning("Tropicos doesn't like"))
    expect_that(get_tpsid('Poa annua var. annua', messages=FALSE),
                gives_warning("Tropicos doesn't like"))
    expect_that(get_tpsid('Poa annua sp. annua', messages=FALSE),
                gives_warning("Tropicos doesn't like"))
    expect_that(get_tpsid('Poa annua ssp. annua', messages=FALSE),
                gives_warning("Tropicos doesn't like"))
    expect_that(get_tpsid('Poa annua subspecies annua', messages=FALSE),
                gives_warning("Tropicos doesn't like"))

    expect_warning(get_tpsid('Poa annua foo bar annua', messages=FALSE),
      NA)
  })
})

test_that("get_tpsid fails as expected", {
  skip_on_cran()

  expect_error(get_tpsid(), "argument \"sci\" is missing")
  expect_error(get_tpsid('Poa annua', ask = 4, messages = FALSE),
               "ask must be of class logical")

  expect_error(
    get_tpsid("Poa annua", family = TRUE, messages = FALSE),
    "family must be of class character")
  expect_error(
    get_tpsid("Poa annua", rank = TRUE, messages = FALSE),
    "rank must be of class character")

  # rows param
  expect_error(get_tpsid("Poa annua", rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_tpsid("Poa annua", rows = 0, messages = FALSE),
               "all\\(rows > 0\\) is not TRUE")
})

#' Retrieve the taxonomic hierarchy from given EOL taxonID.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @rdname eol_hierarchy-defunct
#' @keywords internal
eol_hierarchy <- function(...) {
  .Defunct(msg = "This function is defunct. See classification().")
}

#' methods for preparing/printing info for prompts for `get_*` functions
#'
#' @keywords internal
#' @examples \dontrun{
#' # nms <- c("Quercus", "Sasdsfasdf")
#' # x <- progressor$new(items = nms)
#' # x
#' # x$prog_start()
#'
#' # x$completed(nms[1], "found")
#' # x$prog_found()
#'
#' # x$completed(nms[2], "not found")
#' # x$prog_not_found()
#'
#' # x$prog_summary()
#'
#' # suppress cli::cat_line
#' # x <- progressor$new(items = nms, suppress = TRUE)
#' # x$prog_summary()
#' }
progressor <- R6::R6Class(
  "progressor",
  public = list(
    #' @field total (integer) x
    total = 0,
    #' @field found (integer) list of results when name found
    found = list(),
    #' @field not_found (integer) list of results when name not found
    not_found = list(),
    #' @field done (integer) x
    done = list(),
    #' @field suppress (integer) x
    suppress = FALSE,

    #' @description Create a new `progressor` object
    #' @param items (character) xxx
    #' @param suppress (logical) suppress messages. default: `FALSE`
    #' @return A new `progressor` object
    initialize = function(items, suppress = FALSE) {
      if (!missing(items)) self$total <- length(items)
      if (!is.null(taxize_env$options$taxon_state_messages)) {
        self$suppress <- taxize_env$options$taxon_state_messages
      } else {
        self$suppress <- suppress
      }
    },
    #' @description add results to found or not found
    #' @param name (character) vector of names
    #' @param att (character) one of "found" or "not found"
    #' @return nothing returned; adds to `$found` or `$not_found`
    completed = function(name, att) {
      private$update_suppress()
      switch(att,
        "found" = self$completed_found(name),
        "not found" = self$completed_not_found(name))
    },
    #' @description add to found results
    #' @param name (character) vector of taxon names
    #' @return nothing returned; adds to `$found`
    completed_found = function(name) self$found <- c(self$found, name),
    #' @description add to not found results
    #' @param name (character) vector of taxon names
    #' @return nothing returned; adds to `$not_found`
    completed_not_found = function(name) {
      self$not_found <- c(self$not_found, name)
    },
    #' @description print messages of total queries to do, and 
    #' percent completed
    prog_start = function() {
      private$update_suppress()
      private$sm(cli::cat_line(
        cli::rule(
          left = sprintf(" %s queries ", self$total),
          line = 2, line_col = "blue", width = 30)
      ))
      len <- length(self$found) + length(self$not_found)
      if (len > 0) {
        private$sm(cli::cat_line(cli::rule(
          left = sprintf(" %s completed ", len),
          line = 2, line_col = "lightblue", width = 25)
        ))
      }
    },
    #' @description prints message of found or not found using packages
    #' cli and crayon
    #' @param att (character) one of "found" or "not found"
    #' @return messages
    prog = function(att) {
      private$update_suppress()
      switch(att,
        "found" = self$prog_found(),
        "not found" = self$prog_not_found())
    },
    #' @description prints found message using packages cli and crayon
    #' @return messages
    prog_found = function() {
      name <- private$last(self$found)
      private$sm(cli::cat_line(
        paste(crayon::style(
          paste(private$sym$tick,  " Found: "), "green"), name)
      ))
    },
    #' @description prints not found message using packages cli and crayon
    #' @return messages
    prog_not_found = function() {
      name <- private$last(self$not_found)
      private$sm(cli::cat_line(
        paste(crayon::style(
          paste(private$sym$cross,  " Not Found: "), "red"), name)
      ))
    },
    #' @description prints summary at end of result with total found and
    #' not found
    #' @return messages
    prog_summary = function() {
      private$update_suppress()
      private$sm(cli::cat_line(
        cli::rule(left = " Results ", line = 2, line_col = "grey", width = 30),
        "\n"
      ))
      private$sm(cli::cat_line(
        paste(private$sym$bullet, "Total:",
          crayon::style(self$total, "green"), "\n"),
        paste(private$sym$bullet, "Found:",
          crayon::style(length(self$found), "green"), "\n"),
        paste(private$sym$bullet, "Not Found:",
          crayon::style(length(self$not_found), "green"))
      ))
    }
  ),

  active = list(
    #' @field p (integer) percent done
    p = function() (length(self$done) / self$total) * 100,
    #' @field d (integer) number done
    d = function() length(c(self$found, self$not_found))
  ),

  private = list(
    sym = list(
      bullet = cli::symbol$bullet,
      tick = cli::symbol$tick,
      cross = cli::symbol$cross
    ),
    last = function(x) x[length(x)],
    sm = function(x) if (!self$suppress) x,
    update_suppress = function() {
      if (!is.null(taxize_env$options$taxon_state_messages)) {
        self$suppress <- taxize_env$options$taxon_state_messages
      }
    }
  )
)

context("progressor")

taxize_options(FALSE, quiet = TRUE)

test_that("progressor works", {
  skip_on_cran()

  nms <- c("Quercus", "Sasdsfasdf")
  aa <- progressor$new(items = nms)

  expect_is(aa, "progressor")
  expect_is(aa$d, "integer")
  expect_is(aa$p, "numeric")
  expect_is(aa$completed, "function")
  expect_is(aa$completed_found, "function")
  expect_is(aa$completed_not_found, "function")
  expect_is(aa$prog_start, "function")
  expect_is(aa$prog, "function")
  expect_is(aa$prog_found, "function")
  expect_is(aa$prog_not_found, "function")
  expect_is(aa$prog_summary, "function")

  # start
  expect_output(aa$prog_start(), "2 queries")
  # expect_output(aa$prog_start(), "══")

  # completed
  expect_output(aa$prog_found(), "Found")
  aa$completed(nms[1], "found")
  expect_output(aa$prog_found(), "Quercus")
  
  aa$completed(nms[2], "not found")
  expect_output(aa$prog_not_found(), "Not Found")
  expect_output(aa$prog_not_found(), "Sasdsfasdf")
  
  # expect_output(aa$prog_summary(), "══")
  expect_output(aa$prog_summary(), "Results")
  expect_output(aa$prog_summary(), "Total: 2")
  expect_output(aa$prog_summary(), "Found: 1")
  expect_output(aa$prog_summary(), "Not Found: 1")
})

test_that("progressor fails well", {
  skip_on_cran()
  
  expect_error(progressor$new(apple = 5), "unused argument")
})

taxize_options(TRUE, quiet = TRUE)

#' @title Index Fungorum
#'
#' @description Search for taxonomic names in Index Fungorum
#'
#' @name fungorum
#' @param q (character) Query term
#' @param anywhere (logical) Default: `TRUE`
#' @param limit (integer) Number of results to return. max limit
#' value appears to be 6000, not positive about that though
#' @param key (character) A IndexFungorum taxon key
#' @param lsid (character) an LSID, e.,g. "urn:lsid:indexfungorum.org:names:81085"
#' @param date (character) Date, of the form YYYMMDD
#' @param ... Curl options passed on to [crul::verb-GET]
#' @references http://www.indexfungorum.org/, API docs:
#' http://www.indexfungorum.org/ixfwebservice/fungus.asmx
#' @return A `data.frame`, or `NULL` if no results
#' @examples \dontrun{
#' # NameSearch
#' fg_name_search(q = "Gymnopus", limit = 2, verbose = TRUE)
#' fg_name_search(q = "Gymnopus")
#'
#' # EpithetSearch
#' fg_epithet_search(q = "phalloides")
#'
#' # NameByKey
#' fg_name_by_key(17703)
#'
#' # NameFullByKey
#' fg_name_full_by_lsid("urn:lsid:indexfungorum.org:names:81085")
#'
#' # AllUpdatedNames
#' fg_all_updated_names(date = gsub("-", "", Sys.Date() - 2))
#'
#' # DeprecatedNames
#' fg_deprecated_names(date=20151001)
#'
#' # AuthorSearch
#' fg_author_search(q = "Fayod", limit = 2)
#' }

#' @export
#' @rdname fungorum
fg_name_search <- function(q, anywhere = TRUE, limit = 10, ...) {
  by_name_search("NameSearch", q, anywhere, limit, ...)
}

#' @export
#' @rdname fungorum
fg_author_search <- function(q, anywhere = TRUE, limit = 10, ...) {
  by_name_search("AuthorSearch", q, anywhere, limit, ...)
}

#' @export
#' @rdname fungorum
fg_epithet_search <- function(q, anywhere = TRUE, limit = 10, ...) {
  by_name_search("EpithetSearch", q, anywhere, limit, ...)
}

#' @export
#' @rdname fungorum
fg_name_by_key <- function(key, ...) {
  tmp <- fung_GET("NameByKey", list(NameKey = key), ...)
  fg_df(fung_parse(tmp))
}

#' @export
#' @rdname fungorum
fg_name_full_by_lsid <- function(lsid, ...) {
  tmp <- fung_GET("NameFullByKey", list(NameLsid = lsid), ...)
  xml2::xml_text(xml2::read_xml(tmp))
}

#' @export
#' @rdname fungorum
fg_all_updated_names <- function(date, ...) {
  tmp <- fung_GET("AllUpdatedNames", list(startDate = date), ...)
  xml <- fung_parse(tmp)
  (x <- setDF(rbindlist(lapply(xml, function(z) {
    vapply(xml_children(z), function(w) as.list(xml_text(w)), list(1))
  }))))
}

#' @export
#' @rdname fungorum
fg_deprecated_names <- function(date, ...) {
  tmp <- fung_GET("DeprecatedNames", list(startDate = date), ...)
  xml <- fung_parse(tmp)
  df <- setDF(rbindlist(
    lapply(xml, function(z) {
      vapply(xml_children(z), function(w) as.list(xml_text(w)), list(1))
    })
  ))
  if (NROW(df) > 0) setNames(df, c('fungusnameoldlsid', 'fungusnamenewlsid')) else df
}



# helpers -----------------
fung_base <- function() "http://www.indexfungorum.org/ixfwebservice/fungus.asmx"

fung_GET <- function(path, args, ...) {
  cli <- crul::HttpClient$new(file.path(fung_base(), path),
    headers = tx_ual, opts = list(...))
  tt <- cli$get(query = args)
  tt$raise_for_status()
  tt$parse("UTF-8")
}

fung_parse <- function(x) {
  xml <- xml2::read_xml(x)
  xml_find_all(xml, "//IndexFungorum")
}

fg_df <- function(x) {
  (x <- setDF(rbindlist(
    lapply(x, function(z) {
      data.frame(
        lapply(xml_children(z), function(w) as.list(setNames(xml_text(w), gsub("x0020_", "", tolower(xml_name(w)))))),
        stringsAsFactors = FALSE
      )
    }), use.names = TRUE, fill = TRUE
  )))
}

by_name_search <- function(path, q, anywhere, limit, ...) {
  args <- tc(list(SearchText = q, AnywhereInText = as_l(anywhere), MaxNumber = limit))
  tmp <- fung_GET(path, args, ...)
  fg_df(fung_parse(tmp))
}

context("fungorum")

test_that("fungorum - fg_name_search", {
  skip_on_cran()
  vcr::use_cassette("fg_name_search", match_requests_on = c('method', 'path'), {
    aa <- fg_name_search(q = "Gymnopus", limit = 2)
    bb <- fg_name_search(q = "Gymnopus")
  }, preserve_exact_body_bytes = TRUE)

  expect_is(aa, "data.frame")
  expect_is(bb, "data.frame")

  expect_true(any(grepl("authors", names(aa))))
  expect_equal(NROW(aa), 2)
  expect_equal(unique(aa$infraspecific_epithet), "Gymnopus")

  expect_true(any(grepl("authors", names(bb))))
  expect_equal(NROW(bb), 10)
  expect_equal(unique(bb$infraspecific_epithet)[1], "Gymnopus")
})

test_that("fungorum - fg_epithet_search", {
  skip_on_cran()
  vcr::use_cassette("fg_epithet_search", match_requests_on = c('method', 'path'), {
    aa <- fg_epithet_search(q = "phalloides", limit = 2)
  })

  expect_is(aa, "data.frame")
  expect_true(any(grepl("authors", names(aa))))
  expect_equal(NROW(aa), 2)
  expect_equal(unique(aa$infraspecific_rank)[1], "sp.")
})

test_that("fungorum - fg_name_by_key", {
  skip_on_cran()
  vcr::use_cassette("fg_name_by_key", match_requests_on = c('method', 'path'), {
    aa <- fg_name_by_key(17703)
  })

  expect_is(aa, "data.frame")
  expect_true(any(grepl("name_of_fungus", names(aa))))
  expect_true(any(grepl("uuid", names(aa))))
  expect_equal(NROW(aa), 1)
})

test_that("fungorum - fg_name_full_by_lsid", {
  skip_on_cran()
  vcr::use_cassette("fg_name_full_by_lsid", match_requests_on = c('method', 'path'), {
    aa <- fg_name_full_by_lsid("urn:lsid:indexfungorum.org:names:81085")
  })

  expect_is(aa, "character")
  expect_equal(length(aa), 1)
  expect_true(grepl("Omphalotaceae", aa))
  expect_true(grepl("1985", aa))
})

#test_that("fungorum - fg_all_updated_names", {
#  skip_on_cran()
#
#  date <- as.numeric(gsub("-", "", as.character(Sys.Date()))) - 200
#  aa <- fg_all_updated_names(date = date)
#
#  expect_is(aa, "data.frame")
#  expect_gt(NROW(aa), 1)
#  expect_match(aa[1,], "indexfungorum")
#
#  date <- as.numeric(gsub("-", "", as.character(Sys.Date() + 1)))
#  expect_equal(NROW(fg_all_updated_names(date = date)), 0)
#})

# FIXME: when date injection for vcr fixed, come back to this;
# also, the function is failing
# test_that("fungorum - fg_deprecated_names", {
#   skip_on_cran()

#   date <- as.numeric(gsub("-", "", as.character(Sys.Date() - 30)))
#   aa <- fg_deprecated_names(date = date)

#   expect_is(aa, "data.frame")
#   expect_gt(NROW(aa), 1)
#   expect_match(aa[1,1], "indexfungorum")

#   date <- as.numeric(gsub("-", "", as.character(Sys.Date() + 1)))
#   expect_equal(NROW(fg_deprecated_names(date = date)), 0)
# })

test_that("fungorum - fg_author_search", {
  skip_on_cran()
  vcr::use_cassette("fg_author_search", match_requests_on = c('method', 'path'), {
    aa <- fg_author_search(q = "Fayod", limit = 2)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(aa, "data.frame")
  expect_equal(NROW(aa), 2)
  expect_match(aa$authors, "Fayod")
})

#' Retrieve gene sequences from NCBI by taxon name and gene names.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @author Scott Chamberlain 
#' @rdname ncbi_getbyname-defunct
#' @keywords internal
ncbi_getbyname <- function(...) {
  .Defunct("ncbi_byname", "traits",
    msg = "This function is defunct. See traits::ncbi_byname()")
}

#' Retrieve gene sequences from NCBI by accession number.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @keywords internal
#' @rdname get_seqs-defunct
get_seqs <- function(...) {
  .Defunct("ncbi_getbyname", "taxize",
    "Function name changed. See ncbi_getbyname")
}

#' @title Search for taxonomic names using the Global Names Index
#'
#' @description Uses the Global Names Index, see http://gni.globalnames.org
#'
#' @export
#' @param search_term Name pattern you want to search for. WARNING: Does
#' not work for vernacular/common names. Search term may include following
#' options (Note: can, uni, gen, sp, ssp, au, yr work only for parsed names
#'
#' * *	wild card - Search by part of a word (E.g.: planta*)
#' * exact exact match	- Search for exact match of a literal string
#'  (E.g.: exact:Parus major)
#' * ns name string- Search for literal string from its beginning (other
#'  modifiers will be ignored) (E.g.: ns:parus maj*)
#' * can canonical form- Search name without authors (other modifiers will
#'  be ignored)	(E.g.: can:parus major)
#' * uni uninomial- Search for higher taxa	(E.g.: uni:parus)
#' * gen genus - Search by genus epithet of species name (E.g.: gen:parus)
#' * sp species - Search by species epithet (E.g.: sp:major)
#' * ssp subspecies - Search by infraspecies epithet (E.g.: ssp:major)
#' * au author - Search by author word	(E.g.: au:Shipunov)
#' * yr year - Search by year (E.g.: yr:2005)
#'
#' @param per_page Number of items per one page (numbers larger
#' 		than 1000 will be decreased to 1000) (default is 30).
#' @param page Page number you want to see (default is 1).
#' @param justtotal Return only the total results found.
#' @param parse_names If `TRUE`, use [gni_parse()] to parse
#' names. Default: `FALSE`
#' @param ... Curl options passed on to [crul::verb-GET]
#' @author Scott Chamberlain \email{myrmecocystus@@gmail.com}
#' @return data.frame of results.
#' @seealso [gnr_datasources()], [gni_search()]
#' @keywords globalnamesindex names taxonomy
#' @references <http://gni.globalnames.org/>,
#' <https://github.com/dimus/gni/wiki/api>
#' @details Note that you can use fuzzy searching, e.g., by attaching an
#' asterisk to the end of a search term. See the first two examples below
#' @examples \dontrun{
#' gni_search(search_term = "ani*")
#' gni_search(search_term = "ama*", per_page = 3, page = 21)
#' gni_search(search_term = "animalia", per_page = 8, page = 1)
#' gni_search(search_term = "animalia", per_page = 8, page = 1, justtotal=TRUE)
#'
#' gni_search(search_term = "Cyanistes caeruleus", parse_names=TRUE)
#'
#' # pass on curl options
#' gni_search(search_term = "ani*", verbose = TRUE)
#' }
gni_search <- function(search_term = NULL, per_page = NULL, page = NULL,
  justtotal = FALSE, parse_names = FALSE, ...) {

	query <- tc(list(search_term = search_term, per_page = per_page,
        page = page))
  cli <- crul::HttpClient$new(paste0(gni_base(), "name_strings.json"),
    headers = tx_ual, opts = list(...))
  tt <- cli$get(query = argsnull(query))
  tt$raise_for_status()
	out <- jsonlite::fromJSON(tt$parse("UTF-8"), FALSE)

	if (justtotal) {
	  out$name_strings_total
	} else {
	  df <- dt2df(lapply(out$name_strings, function(x)
	    data.frame(t(data.frame(c( checknull(x[["name"]]), checknull(x[["id"]]),
	                    checknull(x[["lsid"]]), checknull(x[["uuid_hex"]]),
	                    checknull(x[["resource_url"]]) ))))), idcol = FALSE)
	  df <- colClasses(df, "character")
	  if (NROW(df) != 0) {
	    names(df) <- c("name","id","lsid","uuid_hex","resource_url")
	  }

    if (parse_names) {
      data.frame(df, gni_parse(as.character(df$name)),
        stringsAsFactors = FALSE)
    } else {
      df
    }
	}
}

context("gni_search")

test_that("gni_search returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("gni_search", {
    x <- gni_search(search_term = "ama*", per_page = 1) 
  })

	expect_is(x, "data.frame")
  expect_equal(x[,2], "22693003")
})

#' @title ITIS hierarchy
#'
#' @description Get hierarchies from TSN values, full, upstream only, or
#' immediate downstream only
#'
#' @export
#' @param tsn One or more TSN's (taxonomic serial number). Required.
#' @param what One of full (full hierarchy), up (immediate upstream), or down
#'    (immediate downstream)
#' @param ... Further arguments passed on to [ritis::hierarchy_full()]
#' [ritis::hierarchy_up()] or [ritis::hierarchy_down()]
#' @seealso [itis_downstream()]
#' @details Note that [itis_downstream()] gets taxa downstream to a
#' particular rank, while this function only gets immediate names downstream.
#' @examples \dontrun{
#' # Get full hierarchy
#' itis_hierarchy(tsn=180543)
#'
#' # Get hierarchy upstream
#' itis_hierarchy(tsn=180543, "up")
#'
#' # Get hierarchy downstream
#' itis_hierarchy(tsn=180543, "down")
#'
#' # Many tsn's
#' itis_hierarchy(tsn=c(180543,41074,36616))
#' }

itis_hierarchy <- function(tsn, what = "full", ...) {
  temp <- switch(
    what,
    full = lapply(tsn, function(x) ritis::hierarchy_full(x, ...)),
    up = lapply(tsn, function(x) ritis::hierarchy_up(x, ...)),
    down = lapply(tsn, function(x) ritis::hierarchy_down(x, ...))
  )
  if (length(tsn) == 1) {
    tmp <- temp[[1]]
    names(tmp) <- tolower(names(tmp))
    tmp$rankname <- tolower(tmp$rankname)
    tmp
  } else {
    names(temp) <- tsn
    lapply(temp, function(x){
      names(x) <- tolower(names(x))
      x$rankname <- tolower(x$rankname)
      x
    })
  }
}

context("itis_hierarchy")

# test_that("itis_hierarchy returns the correct class", {
#   skip_on_cran()
#   vcr::use_cassette("itis_hierarchy", {
#     one <- itis_hierarchy(tsn=180543, verbose=FALSE)
#     two <- itis_hierarchy(tsn=180543, "up", verbose=FALSE)
#     three <- itis_hierarchy(tsn=180543, "down", verbose=FALSE)
#     four <-itis_hierarchy(tsn=c(180543,41074,36616), verbose=FALSE)
#   })
# 
#   expect_that(one, is_a("data.frame"))
#   expect_that(two, is_a("data.frame"))
#   expect_that(three, is_a("data.frame"))
#   expect_that(four, is_a("list"))
#   expect_that(four[[1]], is_a("data.frame"))
# })

#' Retrieve the taxonomic hierarchy for a given taxon ID.
#'
#' @export
#' @param sci_id Vector of taxa names (character) or IDs (character or numeric)
#' to query. For `db = "eol"`, EOL expects you to pass it a taxon id, called
#' `eolid` in the output of [get_eolid()]. 
#' @param db character; database to query. either `ncbi`, `itis`, `eol`,
#' `tropicos`, `gbif`, `nbn`, `worms`, `natserv`, `bold`, `wiki`, or `pow`.
#' Note that each taxonomic data source has, their own identifiers, so that
#' if you provide the wrong `db` value for the identifier you could get a
#' result, but it will likely be wrong (not what you were expecting). If using
#' ncbi, and/or tropicos, we recommend getting an API key; see
#' [taxize-authentication]
#' @param id character; identifiers, returned by [get_tsn()], [get_uid()],
#' [get_eolid()], [get_tpsid()], [get_gbifid()], [get_tolid()],
#' [get_wormsid()], [get_natservid()], [get_wormsid()], [get_wiki()],
#' [get_pow()]
#' @param callopts Curl options passed on to [crul::verb-GET]
#' @param ... For `classification`: other arguments passed to [get_tsn()],
#' [get_uid()], [get_eolid()], [get_tpsid()], [get_gbifid()],
#' [get_wormsid()], [get_natservid()], [get_wormsid()], [get_wiki()],
#' [get_pow()]. For `rbind.classification` and `cbind.classification`: one or
#' more objects of class `classification`
#' @param return_id (logical) If `TRUE` (default), return the taxon id
#' as well as the name and rank of taxa in the lineage returned.
#' Ignored for natserv as they don't return IDs in their taxonomic
#' classification data.
#' @param rows (numeric) Any number from 1 to infinity. If the default NA,
#' all rows are considered. Note that this parameter is ignored if you pass
#' in a taxonomic id instead of a name of class character.
#' @param batch_size (numeric) For NCBI queries, specify the number of IDs to
#'   lookup for each query.
#' @param max_tries (numeric) For NCBI queries, the number of times a particular
#'   query will be attempted, assuming the first does not work.
#' @param x Deprecated, see `sci_id`
#'
#' @return A named list of data.frames with the taxonomic classification of
#'    every supplied taxa.
#' @details If IDs are supplied directly (not from the `get_*` functions)
#' you must specify the type of ID. There is a timeout of 1/3 seconds between
#' queries to NCBI.
#'
#' BEWARE: Right now, NBN doesn't return the queried taxon in the
#' classification. But you can attach it yourself quite easily of course.
#' This behavior is different from the other data sources.
#'
#' @seealso [get_tsn()], [get_uid()], [get_eolid()],
#'    [get_tpsid()], [get_gbifid()], [get_wormsid()], [get_natservid()],
#'    [get_boldid()], [get_wiki()], [get_pow()]
#'
#' @section Lots of results:
#' It may happen sometimes that you get more results back from your query
#' than will show in the data.frame on screen. Our advice is to refine your
#' query in those cases. On a data source basis we can attempt to help
#' make it easier to refine queries, whether it be with the data provider
#' (unlikely to happen), or in the code in this package (more likely) -
#' let us know if you run into too many results problem and we'll see what
#' we can do.
#'
#' @section Authentication:
#' See [taxize-authentication]
#' 
#' @section EOL:
#' EOL does not have very good failure behavior. For example, if you submit
#' an ID that does not exist they'll return a 500 HTTP error, which is
#' not an appropriate error; it's probably that that ID does not exist 
#' in their database, but we can't know for sure. Isn't that fun?
#' 
#' @section NCBI Rate limits:
#' In case you run into NCBI errors due to your rate limit being exceeded, see
#' [taxize_options()], where you can set `ncbi_sleep`.
#' 
#' @section HTTP version for NCBI requests:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')` 
#'
#' @examples \dontrun{
#' # Plug in taxon IDs
#' classification(9606, db = 'ncbi')
#' classification(c(9606, 55062), db = 'ncbi')
#' classification(129313, db = 'itis')
#' classification(6985636, db = 'eol')
#' classification(126436, db = 'worms')
#' classification('Helianthus annuus', db = 'pow')
#' classification('Helianthus', db = 'pow')
#' classification('Asteraceae', db = 'pow')
#' classification("134717", db = 'natserv')
#' classification(c(2704179, 6162875, 8286319, 2441175, 731), db = 'gbif')
#' classification(25509881, db = 'tropicos')
#' classification("NBNSYS0000004786", db = 'nbn')
#' classification(as.nbnid("NBNSYS0000004786"), db = 'nbn')
#' classification(3930798, db = 'tol')
#'
#' ## works the same if IDs are in class character
#' classification(c("2704179", "2441176"), db = 'gbif')
#' classification("Agapostemon", db = "bold")
#'
#' # wikispecies
#' classification("Malus domestica", db = "wiki")
#' classification("Pinus contorta", db = "wiki")
#' classification("Pinus contorta", db = "wiki", wiki_site = "commons")
#' classification("Pinus contorta", db = "wiki", wiki_site = "pedia")
#' classification("Pinus contorta", db = "wiki", wiki_site = "pedia",
#'   wiki = "fr")
#'
#' classification(get_wiki("Malus domestica", "commons"))
#' classification(get_wiki("Malus domestica", "species"))
#' classification(c("Pinus contorta", "Malus domestica"), db = "wiki")
#'
#' # Plug in taxon names
#' ## in this case, we use get_*() fxns internally to first get taxon IDs
#' classification("Oncorhynchus mykiss", db = "eol")
#' classification(c("Chironomus riparius", "aaa vva"), db = 'ncbi')
#' classification(c("Chironomus riparius", "aaa vva"), db = 'ncbi',
#'   messages=FALSE)
#' classification(c("Chironomus riparius", "aaa vva"), db = 'itis')
#' classification(c("Chironomus riparius", "aaa vva"), db = 'itis',
#'   messages=FALSE)
#' classification(c("Chironomus riparius", "aaa vva"), db = 'eol')
#' classification("Alopias vulpinus", db = 'nbn')
#' classification('Gadus morhua', db = 'worms')
#' classification('Aquila chrysaetos', db = 'natserv')
#' classification('Gadus morhua', db = 'natserv')
#' classification('Pomatomus saltatrix', db = 'natserv')
#' classification('Aquila chrysaetos', db = 'natserv')
#' classification(c("Chironomus riparius", "asdfasdfsfdfsd"), db = 'gbif')
#' classification("Chironomus", db = 'tol')
#' classification("Poa annua", db = 'tropicos')
#'
#' # Use methods for get_uid, get_tsn, get_eolid, get_tpsid
#' classification(get_uid(c("Chironomus riparius", "Puma concolor")))
#'
#' classification(get_uid(c("Chironomus riparius", "aaa vva")))
#' classification(get_tsn(c("Chironomus riparius", "aaa vva")))
#' classification(get_tsn(c("Chironomus riparius", "aaa vva"),
#'   messages = FALSE))
#' classification(get_eolid(c("Chironomus riparius", "aaa vva")))
#' classification(get_tpsid(c("Poa annua", "aaa vva")))
#' classification(get_gbifid(c("Poa annua", "Bison bison")))
#'
#' # Pass many ids from class "ids"
#' (out <- get_ids("Puma concolor", db = c('ncbi','gbif')))
#' (cl <- classification(out))
#'
#' # Bind width-wise from class classification_ids
#' cbind(cl)
#'
#' # Bind length-wise
#' rbind(cl)
#'
#' # Many names to get_ids
#' (out <- get_ids(c("Puma concolor","Accipiter striatus"),
#'   db = c('ncbi','itis')))
#' (cl <- classification(out))
#' rbind(cl)
#' ## cbind with so many names results in some messy data
#' cbind(cl)
#' ## so you can turn off return_id
#' cbind( classification(out, return_id=FALSE) )
#'
#' (cl_uid <- classification(get_uid(c("Puma concolor",
#'   "Accipiter striatus")), return_id=FALSE))
#' rbind(cl_uid)
#' cbind(cl_uid)
#' ## cbind works a bit odd when there are lots of ranks without names
#' (cl_uid <- classification(get_uid(c("Puma concolor","Accipiter striatus")),
#'   return_id=TRUE))
#' cbind(cl_uid)
#'
#' (cl_tsn <- classification(get_tsn(c("Puma concolor","Accipiter striatus"))))
#' rbind(cl_tsn)
#' cbind(cl_tsn)
#'
#' (tsns <- get_tsn(c("Puma concolor","Accipiter striatus")))
#' (cl_tsns <- classification(tsns))
#' cbind(cl_tsns)
#'
#' # NBN data
#' (res <- classification(c("Alopias vulpinus","Pinus sylvestris"),
#'   db = 'nbn'))
#' rbind(res)
#' cbind(res)
#'
#' # Return taxonomic IDs
#' ## the return_id parameter is logical, and you can turn it on or off.
#' ## It's TRUE by default
#' classification(c("Alopias vulpinus","Pinus sylvestris"), db = 'ncbi',
#'   return_id = TRUE)
#' classification(c("Alopias vulpinus","Pinus sylvestris"), db = 'ncbi',
#'   return_id = FALSE)
#'
#' # Use rows parameter to select certain
#' classification('Poa annua', db = 'tropicos')
#' classification('Poa annua', db = 'tropicos', rows=1:4)
#' classification('Poa annua', db = 'tropicos', rows=1)
#' classification('Poa annua', db = 'tropicos', rows=6)
#' 
#' # Queries of many IDs are processed in batches for NCBI
#' ids <- c("13083", "2650392", "1547764", "230054", "353934", "656984", 
#' "271789", "126272", "184644", "73213", "662816", "1161803", "1239353", 
#' "59420", "665675", "866969", "1091219", "1431218", "1471898", 
#' "864321", "251768", "2486276", "2068772", "1825808", "2006532", 
#' "128287", "1195738", "1084683", "1886461", "508296", "377247", 
#' "1489665", "329325", "219243", "1176946", "339893", "197933", 
#' "174510", "1704048", "212897", "154842", "1239280", "260135", 
#' "405735", "1566412", "2083462", "651348", "983204", "165380", 
#' "2338856", "2068760", "167262", "34229", "1213340", "478939", 
#' "1933585", "49951", "1277794", "1671089", "1502538", "362355", 
#' "746473", "242879", "158219", "313664", "2093188", "1541232", 
#' "584742", "1331091", "147639", "284492", "75642", "1412882", 
#' "391782", "1406855", "434506", "2053357", "217315", "1444328", 
#' "329249", "2294004", "84942", "324458", "538247", "69452", "49170", 
#' "1993845", "261646", "127633", "228146", "1420004", "1629772", 
#' "577055", "697062", "231660", "648380", "554953", "746496", "2602969")
#' result <- classification(ids, db = 'ncbi')
#' }
classification <- function(...){
  UseMethod("classification")
}

#' @export
#' @rdname classification
classification.default <- function(sci_id, db = NULL, callopts = list(),
                                   return_id = TRUE, rows = NA, x = NULL, ...) {
  nstop(db)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "classification(x)", with = "classification(sci_id)")
    sci_id <- x
  }
  
  switch(
    db,
    itis = {
      id <- process_ids(sci_id, db, get_tsn, rows = rows, ...)
      stats::setNames(classification(id, return_id = return_id, ...), sci_id)
    },
    ncbi = {
      id <- process_ids(sci_id, db, get_uid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    eol = {
      id <- process_ids(sci_id, db, get_eolid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    tropicos = {
      id <- process_ids(sci_id, db, get_tpsid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    gbif = {
      id <- process_ids(sci_id, db, get_gbifid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    nbn = {
      id <- process_ids(sci_id, db, get_nbnid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    tol = {
      id <- process_ids(sci_id, db, get_tolid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    worms = {
      id <- process_ids(sci_id, db, get_wormsid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    natserv = {
      id <- process_ids(sci_id, db, get_natservid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    bold = {
      id <- process_ids(sci_id, db, get_boldid, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    wiki = {
      id <- process_ids(sci_id, db, get_wiki, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    pow = {
      id <- process_ids(sci_id, db, get_pow, rows = rows, ...)
      stats::setNames(classification(id, callopts = callopts,
        return_id = return_id, ...), sci_id)
    },
    stop("the provided db value was not recognised", call. = FALSE)
  )
}

process_ids <- function(input, db, fxn, ...){
  g <- tryCatch(as.numeric(as.character(input)), warning = function(e) e)
  if (
    inherits(g, "numeric") || # all others
    is.character(input) && all(grepl("N[HB]", input)) || # NBN
    is.character(input) && all(grepl("urn:lsid", input)) # POW
  ) {
    as_fxn <- switch(db,
           itis = as.tsn,
           ncbi = as.uid,
           eol = as.eolid,
           tropicos = as.tpsid,
           gbif = as.gbifid,
           nbn = as.nbnid,
           tol = as.tolid,
           worms = as.wormsid,
           natserv = as.natservid,
           bold = as.boldid,
           wiki = as.wiki,
           pow = as.pow)
    as_fxn(input, check = FALSE)
  } else {
    eval(fxn)(input, ...)
  }
}

#' @export
#' @rdname classification
classification.tsn <- function(id, return_id = TRUE, ...) {
  warn_db(list(...), "itis")
  fun <- function(x) {
    # return NA if NA is supplied
    if (is.na(x)) {
      out <- NA
    } else {
      out <- ritis::hierarchy_full(as.character(x), wt = "json", raw = FALSE)
      if (NROW(out) < 1) return(NA)
      # make normal data.frame
      out <- data.frame(out, stringsAsFactors = FALSE)
      # remove overhang
      out <- out[1:which(out$tsn == x), c('taxonname', 'rankname', 'tsn')]
      names(out) <- c('name', 'rank', 'id')
      # Optionally return tsn of lineage
      if (!return_id) out <- out[, c('name', 'rank')]
      out$rank <- tolower(out$rank)
    }
    return(out)
  }
  out <- lapply(id, fun)
  names(out) <- id
  structure(out, class = 'classification', db = 'itis')
}

#' @export
#' @rdname classification
classification.uid <- function(id, callopts = list(), return_id = TRUE,
  batch_size = 50, max_tries = 3, ...) {

  warn_db(list(...), "ncbi")
  fun <- function(x, callopts) {
    key <- getkey(NULL, service="entrez")
    
    query_ncbi <- function(ids) {
      query <- tc(list(db = "taxonomy", ID = paste0(ids, collapse = ','),
        api_key = key))
      cli <- crul::HttpClient$new(url = ncbi_base(),
        opts = c(http_version = 2L, callopts))
      success <- FALSE
      tries <- 1
      while (success == FALSE && tries <= max_tries) {
        res <- cli$get("entrez/eutils/efetch.fcgi", query = query)
		try({
			res$raise_for_status()
			tt <- res$parse("UTF-8")
			ttp <- xml2::read_xml(tt)
			out <- lapply(xml2::xml_find_all(ttp, '//TaxaSet/Taxon'),
			function(tax_node) {
			lin <- data.frame(
				name = xml_text_all(tax_node,
				".//LineageEx/Taxon/ScientificName"),
				rank = xml_text_all(tax_node, ".//LineageEx/Taxon/Rank"),
				id = xml_text_all(tax_node, ".//LineageEx/Taxon/TaxId"),
				stringsAsFactors = FALSE)
			targ_tax <- data.frame(
				name = xml_text_all(tax_node, "./ScientificName"),
				rank = xml_text_all(tax_node, "./Rank"),
				id = xml_text_all(tax_node, "./TaxId"),
				stringsAsFactors = FALSE)
			rbind(lin, targ_tax)
			})
			# Is not directly below root and no lineage info
			parent_id <- xml_text_all(ttp, "//TaxaSet/Taxon/ParentTaxId") %||% ""
			out[vapply(out, NROW, numeric(1)) == 0 & parent_id != "1"] <- NA
			# Add NA where the taxon ID was not found
			names(out) <- xml_text(xml2::xml_find_all(ttp,
			'//TaxaSet/Taxon/TaxId'))
			success <- ! grepl(tt, pattern = 'error', ignore.case = TRUE)
		}, silent=FALSE)
        tries <- tries + 1
        # NCBI limits requests to three per second without key or 10 per
        # second with key
        ncbi_rate_limit_pause(key)
        # Wait longer if query failed
        if (success == FALSE) {
          Sys.sleep(1)
        }
      }
      # Return NA if cannot get information
      if (!success) {
        out <- rep(list(NA), length(ids))
        warning(call. = FALSE, 'Giving up on query after ',
          max_tries, ' tries. NAs will be returned.')
      }
      return(out)
    }
    # return NA if NA is supplied
    out <- rep(list(NA), length(x))
    out[! is.na(x)] <- query_ncbi(x[! is.na(x)])
    # Optionally return taxon id of lineage taxa
    if (!return_id) {
      out[! is.na(out)] <- lapply(out[! is.na(out)],
        function(o) o[, c('name', 'rank')])
    }
    # Return ranks in all lower case
    out[! is.na(out)] <- lapply(out[! is.na(out)], function(o) {
      o$rank <- tolower(o$rank)
      return(o)
    })
    return(out)
  }
  id <- as.character(id) # force to character
  id_chunks <- split(id, ceiling(seq_along(id)/batch_size))
  out <- lapply(id_chunks, fun, callopts = callopts)
  out <- unlist(out, recursive = FALSE)
  names(out) <- id
  structure(out, class = 'classification', db = 'ncbi')
}

#' @export
#' @rdname classification
classification.eolid <- function(id, callopts = list(), return_id = TRUE, ...) {
  warn_db(list(...), "eol")
  common_names = synonyms = NULL
  fun <- function(x){
    if (is.na(x)) {
      out <- NA
    } else {
      args <- tc(list(common_names = common_names, synonyms = synonyms))
      cli <- crul::HttpClient$new(url = 'https://eol.org', opts = callopts)
      tt <- cli$get(file.path('api/hierarchy_entries/1.0', paste0(x, ".json")),
        query = args)
      tt$raise_for_status()
      res <- jsonlite::fromJSON(tt$parse("UTF-8"), FALSE)
      if (length(res$ancestors) == 0) {
        return(sprintf("No hierarchy information for %s", x))
      } else {
        fff <- lapply(res$ancestors, function(z) {
          z[sapply(z, is.null)] <- NA_character_
          data.frame(z, stringsAsFactors = FALSE)
        })
        out <- dt2df(fff, idcol = FALSE)[,c('scientificName',
          'taxonRank', 'taxonID')]
        # add querried taxon
        tr <- res$taxonRank
        out <- rbind(out,
          c(res$scientificName, if ( is.null(tr) ) NA else tr, x))
        names(out) <- c('name', 'rank', 'id')
        # Optionally return id of lineage
        if (!return_id) out <- out[, c('name', 'rank')]
        out$rank <- tolower(out$rank)
        return(out)
      }
    }
  }
  out <- lapply(id, fun)
  names(out) <- id
  structure(out, class = 'classification', db = 'eol')
}

#' @export
#' @rdname classification
classification.tpsid <- function(id, callopts = list(), return_id = TRUE, ...) {
  warn_db(list(...), "tropicos")
  fun <- function(x, callopts){
    if (is.na(x)) {
      out <- NA
    } else {
      url <- sprintf('http://services.tropicos.org/Name/%s/HigherTaxa', x)
      key <- getkey(NULL, "TROPICOS_KEY")
      args <- tc(list(format = 'json', apikey = key))
      tt <- tax_GET(url, query = args, opts = callopts)
      if ("Error" %in% names(tt)) {
        out <- data.frame(ScientificName = NA, Rank = NA)
      } else {
        out <- tt[,c('ScientificName','Rank', 'NameId')]
      }
      names(out) <- c('name', 'rank', 'id')
      # Optionally return id of lineage
      if (!return_id) out <- out[, c('name', 'rank')]
      out$rank <- tolower(out$rank)
    }
    return(out)
  }
  out <- lapply(id, fun, callopts = callopts)
  names(out) <- id
  structure(out, class = 'classification', db = 'tropicos')
}

#' @export
#' @rdname classification
classification.gbifid <- function(id, callopts = list(),
  return_id = TRUE, ...) {

  warn_db(list(...), "gbif")
  fun <- function(x, callopts){
    if (is.na(x)) {
      out <- NA
    } else {
      out <- suppressWarnings(tryCatch(
        gbif_name_usage(key = x, callopts = callopts),
        error = function(e) e))
      if (inherits(out, "simpleError")) {
        NA
      } else {
        cls = c('kingdom','phylum','class', 'order','family','genus','species')
        df1 <- stats::setNames(nmdlst2df(out[names(out) %in% cls]),
          c("rank", "name"))
        df2 <- stats::setNames(
          nmdlst2df(out[names(out) %in% paste0(cls, "Key")]),
          c(".id", "id"))
        df2$rank <- sub("Key", "", df2$.id)
        df2$.id <- NULL
        df <- merge(df1, df2, by = "rank")
        df$rank <- tolower(df$rank)
        # sort to make sure ranks are in correct descending order
        df <- df[order(vapply(df$rank, which_rank, 1)), ]
        # column order
        df <- data.frame(name = df$name, rank = df$rank, id = df$id,
          stringsAsFactors = FALSE)
        # check if target taxon is sub-specific
        if (which_rank(tolower(out$rank)) > 37) {
          df <- rbind(df,
            c(out$canonicalName, tolower(out$rank), out$key))
        }
        df$id <- as.integer(df$id)
        # Optionally return id of lineage
        if (!return_id) df[, c('name', 'rank')] else df
      }
    }
  }
  out <- lapply(id, fun, callopts = callopts)
  names(out) <- id
  structure(out, class = 'classification', db = 'gbif')
}

#' @export
#' @rdname classification
classification.nbnid <- function(id, callopts = list(),
  return_id = TRUE, ...) {

  warn_db(list(...), "nbn")
  fun <- function(x, callopts){
    if (is.na(x)) {
      out <- NA
    } else {
      out <- suppressWarnings(tryCatch(nbn_classification(id = x, callopts),
                                       error = function(e) e))
      if (inherits(out, "simpleError")) {
        NA
      } else {
        out <- out[ , c('scientificname', 'rank', 'guid')]
        names(out) <- c('name', 'rank', 'id')
        # Optionally return id of lineage
        if (!return_id) out <- out[, c('name', 'rank')]
        out$rank <- tolower(out$rank)
        return(out)
      }
    }
  }
  out <- lapply(id, fun, callopts = callopts)
  names(out) <- id
  structure(out, class = 'classification', db = 'nbn')
}

#' @export
#' @rdname classification
classification.tolid <- function(id, callopts = list(),
  return_id = TRUE, ...) {

  warn_db(list(...), "tol")
  fun <- function(x, callopts) {
    if (is.na(x)) {
      out <- NA
    } else {
      x <- as.numeric(x)
      out <- tryCatch(rotl::taxonomy_taxon_info(x,
        include_lineage = TRUE), error = function(e) e)
      if (inherits(out, "error")) {
        NA
      } else {
        outdf <- rotl::tax_lineage(out)[[1]]
        # we have to reverse row order
        outdf <- outdf[NROW(outdf):1, ]
        # tack on species searched for
        orig <- c(out[[1]]$rank, out[[1]]$name, out[[1]]$unique_name,
          out[[1]]$ott_id)
        outdf <- rbind(outdf, orig)
        row.names(outdf) <- NULL
        outdf <- outdf[ , c('name','rank', 'ott_id')]
        names(outdf) <- c('name', 'rank', 'id')
        if (!return_id) outdf <- outdf[, c('name', 'rank')]
        return(outdf)
      }
    }
  }
  out <- lapply(id, fun, callopts = callopts)
  names(out) <- id
  structure(out, class = 'classification', db = 'tol')
}

#' @export
#' @rdname classification
classification.wormsid <- function(id, callopts = list(),
  return_id = TRUE, ...) {

  warn_db(list(...), "worms")
  fun <- function(x, ...){
    if (is.na(x)) {
      out <- NA
    } else {
      out <- tryCatch(worrms::wm_classification(as.numeric(x), ...),
        error = function(e) e)
      if (inherits(out, "error")) {
        NA
      } else {
        row.names(out) <- NULL
        out <- out[ , c('scientificname','rank', 'AphiaID')]
        names(out) <- c('name', 'rank', 'id')
        if (!return_id) out <- out[, c('name', 'rank')]
        return(out)
      }
    }
  }
  out <- lapply(id, fun, ...)
  names(out) <- id
  structure(out, class = 'classification', db = 'worms')
}

#' @export
#' @rdname classification
classification.natservid <- function(id, callopts = list(),
  return_id = TRUE, ...) {

  warn_db(list(...), "natserv")
  fun <- function(x, callopts) {
    if (is.na(x)) {
      out <- NA
    } else {
      out <- tryCatch(natserv::ns_id(paste0("ELEMENT_GLOBAL.2.", x)), error = function(e) e)
      # out <- tryCatch(natserv::ns_data(x), error = function(e) e)
      if (inherits(out, "error")) {
        NA
      } else {
        zz <- c("kingdom", "phylum", "taxclass", "taxorder", "family", "genus")
        if (is.null(out$speciesGlobal)) return(NA)
        res <- tc(out$speciesGlobal[zz])
        if (length(res) == 0) return(NA)
        df <- data.frame(scientificname = unname(unlist(res)),
          rank = gsub("tax", "", names(res)), stringsAsFactors = FALSE)
        df <- rbind(df, c(out$scientificName, "species"))
        # tmp <- out[[1]]$classification
        # if (is.null(tmp)) return(NA)
        # tmp <- tmp$taxonomy$formalTaxonomy
        # if (is.null(tmp)) return(NA)
        # tmp <- tmp[names(tmp) %in% c('kingdom', 'phylum',
        #   'class', 'order', 'family', 'genus')]
        # df <- data.frame(scientificname = unname(unlist(tmp)),
        #   rank = names(tmp), stringsAsFactors = FALSE)
        # rks <- c('kingdom', 'phylum', 'class', 'order',
        #   'family', 'genus', 'species')
        # targ_taxon <- c(
        #   out[[1]]$classification$names$scientificName$unformattedName[[1]],
        #   rks[which(df$rank[length(df$rank)] == rks) + 1]
        # )
        # df <- rbind(df, targ_taxon)
        return(df)
      }
    }
  }
  out <- lapply(id, fun, callopts = callopts)
  names(out) <- id
  structure(out, class = 'classification', db = 'natserv')
}

#' @export
#' @rdname classification
classification.boldid <- function(id, callopts = list(),
  return_id = TRUE, ...) {

  warn_db(list(...), "bold")
  fun <- function(x, callopts) {
    if (is.na(x)) {
      out <- NA
    } else {
      out <- tryCatch(bold_search(id = x, includeTree = TRUE),
        error = function(e) e)
      if (inherits(out, "error")) {
        NA
      } else {
        if (is.null(out)) return(NA)
        tmp <- out[names(out) %in% c('taxid', 'taxon', 'tax_rank')]
        df <- data.frame(name = tmp$taxon, rank = tmp$tax_rank,
                          id = tmp$taxid, stringsAsFactors = FALSE)
        return(df)
      }
    }
  }
  out <- lapply(id, fun, callopts = callopts)
  names(out) <- id
  structure(out, class = 'classification', db = 'bold')
}

#' @export
#' @rdname classification
classification.wiki <- function(id, callopts = list(), return_id = TRUE, ...) {
  warn_db(list(...), "wiki")
  fun <- function(x, wiki_site = "species", wiki = "en", callopts) {
    if (is.na(x)) {
      out <- NA
    } else {
      fxn <- switch(
        wiki_site,
        species = wikitaxa::wt_wikispecies,
        commons = wikitaxa::wt_wikicommons,
        pedia = wikitaxa::wt_wikipedia
      )
      out <- tryCatch(fxn(x)$classification, error = function(e) e)
      if (inherits(out, "error")) {
        NA
      } else {
        if (is.null(out) || NROW(out) == 0) return(NA)
        df <- data.frame(name = out$name, rank = out$rank,
                         stringsAsFactors = FALSE)
        return(df)
      }
    }
  }
  out <- list()
  for (i in seq_along(id)) {
    out[[i]] <-
      fun(id[i], attr(id, "wiki_site"), attr(id, "wiki_lang"))
  }
  names(out) <- id
  structure(out, class = 'classification', db = 'wiki',
            wiki_site = attr(id, "wiki_site"), wiki = attr(id, "wiki_lang"))
}

#' @export
#' @rdname classification
classification.pow <- function(id, callopts = list(), return_id = TRUE, ...) {
  warn_db(list(...), "pow")
  fun <- function(x, callopts) {
    if (is.na(x)) {
      out <- NA
    } else {
      out <- tryCatch(pow_lookup(x), error = function(e) e)
      if (inherits(out, "error")) {
        NA
      } else {
        if (is.null(out)) return(NA)
        tmp <- out$meta$classification[,c('name', 'rank', 'fqId')]
        df <- data.frame(name = tmp$name, rank = tolower(tmp$rank),
          id = tmp$fqId, stringsAsFactors = FALSE)
        return(df)
      }
    }
  }
  out <- lapply(id, fun, callopts = callopts)
  names(out) <- id
  structure(out, class = 'classification', db = 'pow')
}

# ---------
#' @export
#' @rdname classification
classification.ids <- function(id, ...) {
  fun <- function(x, ...){
    # return NA if NA is supplied
    if (all(is.na(x))) {
      out <- NA
    } else {
      out <- classification(x, ...)
    }
    return(out)
  }
  structure(lapply(id, fun, ...), class = 'classification_ids')
}

#' @export
#' @rdname classification
cbind.classification <- function(...) {
  gethiernames <- function(x) {
    x <- data.frame(x)
    x$name <- as.character(x$name)
    x$rank <- as.character(x$rank)
    values <- setNames(data.frame(t(x[,'name']), stringsAsFactors = FALSE),
      tolower(x[,'rank']))
    if ("id" %in% names(x)) {
      x$id <- as.character(x$id)
      ids <- setNames(data.frame(t(x[,'id']), stringsAsFactors = FALSE),
        paste0(tolower(x[,'rank']),"_id") )
      data.frame(values, ids)
    } else {
      values
    }
  }
  input <- x <- c(...)
  input <- input[vapply(x, function(z) inherits(z, "data.frame"), logical(1))]
  tmp <- dt2df(lapply(input, gethiernames), idcol = FALSE)
  tmp$query <- names(x)
  tmp$db <- attr(x, "db")
  tmp
}

#' @export
#' @rdname classification
rbind.classification <- function(...) {
  input <- x <- c(...)
  db <- attr(input, "db")
  x <- input[vapply(x, function(z) inherits(z, "data.frame"), logical(1))]
  for (i in seq_along(x)) {
    x[[i]]$query <- names(x[i])
  }
  df <- dt2df(x, idcol = FALSE)
  df$db <- db
  return( df )
}

#' @export
#' @rdname classification
cbind.classification_ids <- function(...) {
  input <- c(...)

  # remove non-data.frames
  input <- input[vapply(input, function(x)
    class(x[[1]])[1], "") %in% c("data.frame", "tbl_df")]

  gethiernames <- function(x){
    x$name <- as.character(x$name)
    x$rank <- as.character(x$rank)
    values <- setNames(data.frame(t(x$name), stringsAsFactors = FALSE),
      tolower(x$rank))
    if ("id" %in% names(x)) {
      x$id <- as.character(x$id)
      ids <- setNames(data.frame(t(x$id), stringsAsFactors = FALSE),
        paste0(tolower(x$rank), "_id") )
      data.frame(values, ids)
    } else {
      values
    }
  }
  dat <- dt2df(lapply(input, function(h){
    tmp <- lapply(h, gethiernames)
    tmp <- dt2df(tmp, idcol = FALSE)
    tmp$query <- names(h)
    tmp$db <- attr(h, "db")
    tmp
  }), idcol = FALSE)
  move_col(tt = dat, y = c('query','db'))
}

#' @export
#' @rdname classification
rbind.classification_ids <- function(...) {
  input <- c(...)

  # remove non-data.frames
  input <- input[vapply(input, function(x)
    class(x[[1]])[1], "") %in% c("data.frame", "tbl_df")]

  df <- lapply(input, function(x){
    coll <- list()
    for (i in seq_along(x)) {
      coll[[i]] <- data.frame(names(x[i]), x[i][[1]], stringsAsFactors = FALSE)
    }
    coll
  })

  get <- list()
  for (i in seq_along(df[[1]])) {
    tmp <- do.call(rbind, lapply(df, "[[", i))
    source2 <- gsub("\\.[0-9]+", "", row.names(tmp))
    row.names(tmp) <- NULL
    names(tmp)[1] <- "query"
    tmp <- data.frame(db = source2, tmp, stringsAsFactors = FALSE)
    get[[i]] <- tmp
  }

  tt <- if (length(get) == 1) get[[1]] else dt2df(get, idcol = FALSE)
  move_col(tt, c('query', 'db'))
}

context("tax_name")

test_that("tax_name: ncbi", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tax_name_ncbi", {
    tmp_ncbi  <- tax_name("Baetis", get = c("family", "order"),
                          db = "ncbi", messages=FALSE)
    tmp_ncbi2 <- tax_name(c("Helianthus annuus", 'Baetis rhodani'),
                          get = c("genus", "kingdom"), db = "ncbi", messages=FALSE)
  })

	expect_is(tmp_ncbi, "data.frame")
  expect_equal(ncol(tmp_ncbi), 4)
	expect_equal(tmp_ncbi$family, "Baetidae")
	expect_equal(tmp_ncbi$order, "Ephemeroptera")
  expect_that(nrow(tmp_ncbi2), equals(2))
})

test_that("tax_name: NA's", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tax_name_na", {
    tmp_na2 <- tax_name(c("Helianthus annuus", 'xxxx'),
                        get=c("family", "order"), db="ncbi", messages=FALSE)
    tmp_na3 <- sw(tax_name(c("Helianthus annuus", 'xxxx'), rows = 1,
                        get = c("family", "order"), db="itis", messages = FALSE))
  })

  expect_is(tmp_na2, "data.frame")
  expect_is(tmp_na3, "data.frame")
  expect_equal(ncol(tmp_na2), 4)
  expect_equal(ncol(tmp_na3), 4)
  expect_true(any(is.na(tmp_na2[2, ])))
  expect_true(any(is.na(tmp_na3[2, ])))
})

test_that("tax_name: itis", {
  skip_on_cran()
  vcr::use_cassette("tax_name_itis", {
    tmp_itis <- sw(tax_name("Helianthus annuus", rows = 1,
                        get = c("family", "order"), db="itis", messages = FALSE))
  })

  expect_is(tmp_itis, "data.frame")
  expect_equal(ncol(tmp_itis), 4)
  expect_equal(tmp_itis$family, "Asteraceae")
  expect_equal(tmp_itis$order, "Asterales")
})

test_that("tax_name: rows arg", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tax_name_rows_Arg", {
    aa <- tax_name('Bacillus', c("family", "order"), db="ncbi", rows=1)
  })
  
  expect_is(aa, "data.frame")
  expect_equal(NCOL(aa), 4)
  expect_equal(NROW(aa), 1)
  expect_named(aa, c("db", "query", "family", "order"))
})

test_that("tax_name accepts ask-argument", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tax_name_arg_arg", {
    x <- sw(tax_name("Dugesia", get = "family", db = "ncbi",
                             ask = FALSE, messages = FALSE))$family
  })

  expect_true(is.na(x))
})

test_that("taxon with no data returned from classification() works", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tax_name_no_data", {
    aa <- sw(tax_name("foo bar", get = "species",
      rows = 1, messages = FALSE))
    bb <- sw(tax_name("Asterias helianthus", get = "genus", 
      messages = FALSE))
    cc <- sw(tax_name("Stellonia helianthus", get = "genus", 
      messages = FALSE))
  })
  
  expect_is(aa, "data.frame")
  expect_true(is.na(aa$species))

  expect_is(bb, "data.frame")
  expect_true(is.na(bb$genus))
  # expect_warning(tax_name("Asterias helianthus", get = "genus", messages = FALSE),
  #                "rank requested")

  expect_is(cc, "data.frame")
  expect_true(is.na(cc$genus))

  vcr::use_cassette("tax_name_throws_warning", {
    expect_warning(tax_name("Stellonia helianthus", get = "genus", messages = FALSE),
                 "rank requested")
  })
})

context("classification.gbif")

test_that("classification works for GBIF data", {
  skip_on_cran()
  vcr::use_cassette("classification_gbif", {
    clas_gbif <- classification(c("Quercus douglasii", "aaa vva"),
      db = "gbif", rows = 1, messages = FALSE)
    names(clas_gbif) <- NULL
  })

  expect_that(clas_gbif[[2]], equals(NA))
  expect_is(clas_gbif, "classification")
  expect_is(clas_gbif[[1]], "data.frame")
  expect_equal(length(clas_gbif), 2)
  # rank is lowercase
  expect_false(all(grepl("[A-Z]", clas_gbif[[1]]$rank)))

  vcr::use_cassette("classification_gbif_get_fxn", {
    gbifs <- get_gbifid(c("Quercus douglasii", "aaa vva"), rows = 1,
      messages = FALSE)
    clas_gbifs <- classification(gbifs, messages = FALSE)
    names(clas_gbifs) <- NULL
  })

  expect_identical(clas_gbifs, clas_gbif)
})

test_that("ranks are in the correct order", {
  skip_on_cran()
  vcr::use_cassette("classification_gbif_ranks_correct_order", {
    fromid_gbif <- classification(2704179, db = "gbif")
  })

  expect_is(fromid_gbif, "classification")
  expect_equal(attr(fromid_gbif, "db"), "gbif")
  rank_seq <- vapply(fromid_gbif[[1]]$rank, which_rank, 1)
  expect_true(identical(rank_seq, sort(rank_seq)))
  # species rank is expected value
  df <- fromid_gbif[[1]]
  expect_equal(
    df[df$name == "Poa annua", "id"],
    2704179
  )
})

test_that("searches for ranks below species work", {
  skip_on_cran()
  
  vcr::use_cassette("classification_gbif_ranks_below_species_subsp", {
    subsp <- classification(6162875, db = "gbif")
  })
  expect_is(subsp, "classification")
  expect_equal(attr(subsp, "db"), "gbif")
  rank_seq <- vapply(subsp[[1]]$rank, which_rank, 1)
  expect_true(identical(rank_seq, sort(rank_seq)))
  # subspecies rank is expected value
  df <- subsp[[1]]
  expect_equal(
    df[df$name == "Boa constrictor nebulosa", "id"],
    6162875
  )

  vcr::use_cassette("classification_gbif_ranks_below_species_var", {
    var <- classification(8286319, db = "gbif")
  })
  expect_is(var, "classification")
  expect_equal(attr(var, "db"), "gbif")
  rank_seq <- vapply(var[[1]]$rank, which_rank, 1)
  expect_true(identical(rank_seq, sort(rank_seq)))
  # variety rank is expected value
  df <- var[[1]]
  expect_equal(
    df[df$name == "Poa annua annua", "id"],
    8286319
  )
})

#' Return summary data a taxon name with a given id.
#'
#' @export
#' @param id the taxon identifier code
#' @param key Your Tropicos API key; See [taxize-authentication]
#' for help on authentication
#' @param ... Curl options passed on to [crul::verb-GET]
#' @return A data.frame.
#' @examples \dontrun{
#' tp_summary(id = 25509881)
#' tp_summary(id = 2700851)
#' tp_summary(id = 24900183)
#' }

tp_summary <- function(id, key = NULL, ...) {
  url <- sprintf('http://services.tropicos.org/Name/%s', id)
	key <- getkey(key, "TROPICOS_KEY")

  args <- tc(list(apikey = key, format = 'json'))
  tt <- tp_GET(url, args, ...)
  res <- jsonlite::fromJSON(tt, FALSE)
  typespec <- data.frame(res$TypeSpecimens, stringsAsFactors = FALSE)
  df <- data.frame(res[!names(res) %in% "TypeSpecimens"], stringsAsFactors = FALSE)
  if (NROW(typespec) > 0) df <- cbind(df, typespec)
  stats::setNames(df, tolower(names(df)))
}

tp_GET <- function(url, query, ...) {
  cli <- crul::HttpClient$new(url = url, headers = tx_ual, opts = list(...))
  res <- cli$get(query = query)
  res$raise_for_status()
  if (grepl("exception occurred", res$parse("UTF-8"), ignore.case = TRUE)) {
    stop("500 - a server error occurred, try again later")
  }
  res$parse("UTF-8")
}

context("tp_summary")

test_that("tp_summary returns the correct value", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tp_summary", {
    dat <- suppressMessages(tp_summary(id = 25509881))
  }, preserve_exact_body_bytes = TRUE)

  if ("error" %in% names(dat)) skip("error in tp_summary call - skipping")

  expect_match(names(dat)[[1]], ".id")
	expect_is(dat, "data.frame")
	expect_gt(NCOL(dat), 10)
})

#' Convert a list of classifications to a tree.
#'
#' This function converts a list of hierarchies for individual species into
#' a single species by taxonomic level matrix, then calculates a distance
#' matrix based on taxonomy alone, and outputs either a phylo or dist object.
#' See details for more information.
#'
#' @export
#' @param input List of classification data.frame's from the function
#' [classification()]
#' @param varstep Vary step lengths between successive levels relative to
#' proportional loss of the number of distinct classes.
#' @param check	If TRUE, remove all redundant levels which are different for
#' all rows or constant for all rows and regard each row as a different basal
#' taxon (species). If FALSE all levels are retained and basal taxa (species)
#' also must be coded as variables (columns). You will get a warning if
#' species are not coded, but you can ignore this if that was your intention.
#' @param remove_shared If `TRUE`, remove any taxa that are coarser ranks
#'   present in other taxa, such as both a genus and a species in that genus in
#'   the same tree.
#' @param ... Further arguments passed on to hclust.
#' @param x Input object to print or plot - output from class2tree function.
#' @return An object of class "classtree" with slots:
#' 
#' * phylo - The resulting object, a phylo object
#' * classification - The classification data.frame, with taxa as rows,
#'  and different classification levels as columns
#' * distmat - Distance matrix
#' * names - The names of the tips of the phylogeny
#'
#' Note that when you execute the resulting object, you only get the phylo
#' object. You can get to the other 3 slots by calling them directly, like
#' output$names, etc.
#' @details See [vegan::taxa2dist()]. Thanks to Jari Oksanen for
#' making the taxa2dist function and pointing it out, and Clarke & Warwick
#' (1998, 2001), which taxa2dist was based on.
#' The taxonomy tree created is not only based on the clustering of the 
#' taxonomy ranks (e.g. strain, species, genus, ...), but it also utilizes the
#' actual taxon clades (e.g. mammals, plants or reptiles, etc.). The process of
#' this function is as following: First, all possible taxonomy ranks and their 
#' corresponding IDs for each given taxon will be collected from the input.
#' Then, the rank vectors of all taxa will be aligned, so that they together 
#' will become a matrix where columns are ordered taxonomy ranks of all taxa and
#' rows are the rank vectors of those taxa. After that, the rank matrix will be
#' converted into taxonomy ID matrix, any missing rank will have a pseudo
#' ID from the previous rank. Finally, this taxonomy ID matrix will be used to
#' cluster taxa that have similar taxonomy hierarchy together.
#' @examples \dontrun{
#' spnames <- c('Quercus robur', 'Iris oratoria', 'Arachis paraguariensis',
#'  'Helianthus annuus','Madia elegans','Lupinus albicaulis',
#'  'Pinus lambertiana')
#' out <- classification(spnames, db='itis')
#' tr <- class2tree(out)
#' plot(tr)
#'
#' spnames <- c('Klattia flava', 'Trollius sibiricus',
#'  'Arachis paraguariensis',
#'  'Tanacetum boreale', 'Gentiana yakushimensis','Sesamum schinzianum',
#'  'Pilea verrucosa','Tibouchina striphnocalyx','Lycium dasystemum',
#'  'Berkheya echinacea','Androcymbium villosum',
#'  'Helianthus annuus','Madia elegans','Lupinus albicaulis',
#'  'Pinus lambertiana')
#' out <- classification(spnames, db='ncbi')
#' tr <- class2tree(out)
#' plot(tr)
#' }
#' 
class2tree <- function(input, varstep = TRUE, check = TRUE, remove_shared = FALSE, ...) {
  
  if (any(is.na(input))) {
    message('Removed species without classification')
    input <- input[!is.na(input)]
  }
  

  # Check that there is more than 2 taxon
  if (length(input) < 3)
    stop("Your input list of classifications must be 3 or longer.")

  if (length(unique(names(input))) < length(names(input)))
    stop("Input list of classifications contains duplicates")

  # Convert tibbles to data.frames
  input <- lapply(input, as.data.frame)
  
  # Get rank and ID list
  message('Get all ranks and their taxIDs')
  rankList <- dt2df(lapply(input, get_rank), idcol = FALSE)
  nameList <- dt2df(lapply(input, get_name), idcol = FALSE)
  strainIndex <- grep("norank", rankList$X1)
  rankList$X1[strainIndex] <- "strain"
  nameList$X1[strainIndex] <- 
    gsub("norank_[[:digit:]]+", "strain", nameList$X1[strainIndex])

  # Create taxonomy matrix
  message('Align taxonomy hierarchies...')
  df <- taxonomy_table_creator(nameList, rankList, remove_shared = remove_shared)

  if (!inherits(df, "data.frame")) {
    stop("no taxon ranks in common - try different inputs")
  }
  message('Taxonomy alignment done!')

  row.names(df) <- df[,1]
  df <- df[,-1]

  # calculate distance matrix
  taxdis <- tryCatch(taxa2dist(df, varstep = varstep, check = check),
                     error = function(e) e)

  tdf <- t(df)
  for (i in 1:ncol(tdf)){
    tdf[,i][duplicated(tdf[,i])] <- NA
  }

  # check for incorrect dimensions error
  if (is(taxdis, 'simpleError'))
    stop("Try check=FALSE, but see docs for taxa2dist function in the vegan package for details.")
  message('Calculate distance matrix')
  out <- as.phylo.hclust(hclust(taxdis, ...))
  out <- ape::di2multi(out)
  # Add node labels
  message('Add node labels')
  node_ids <- sort(unique(out$edge[,1]))
  node_labels <- sapply(phangorn::Descendants(out, node_ids), function(x) {
    sub_df <- df[out$tip.label[x],]
    node_col_i <- which(sapply(1:ncol(sub_df), function(i) {
      length(unique(sub_df[,i]))==1
    }))[1]
    if (is.na(node_col_i)) {
      NA_character_
    } else {
      unique(sub_df[,node_col_i])
    }
  })
  out$node.label <- node_labels
  
  res <- list(phylo = out, classification = as.data.frame(t(tdf)), 
    distmat = taxdis, names = names(input))
  class(res) <- 'classtree'
  return( res )
}

#' @method plot classtree
#' @export
#' @rdname class2tree
plot.classtree <- function(x, ...) {
  if (!is(x$phylo, "phylo"))
    stop("Input object must have a slot in 'phylo' of class 'phylo'")
  plot(x$phylo, ...)
}

#' @method print classtree
#' @export
#' @rdname class2tree
print.classtree <- function(x, ...) {
  if (!is(x$phylo, "phylo"))
    stop("Input object must have a slot in 'phylo' of class 'phylo'")
  print(x$phylo)
}

# Function from the vegan package
# CRAN: http://cran.rstudio.com/web/packages/vegan/
# License: GPL-2
# Maintainer:	Jari Oksanen <jari.oksanen at oulu.fi>
taxa2dist <- function(x, varstep = FALSE, check = TRUE, labels) {
  rich <- apply(x, 2, function(taxa) length(unique(taxa)))
  S <- nrow(x)
  if (check) {
    keep <- rich < S & rich > 1
    rich <- rich[keep]
    x <- x[, keep]
  }
  i <- rev(order(rich))
  x <- x[, i]
  rich <- rich[i]
  if (varstep) {
    add <- -diff(c(nrow(x), rich, 1))
    add <- add/c(S, rich)
    add <- add/sum(add) * 100
  }
  else {
    add <- rep(100/(ncol(x) + check), ncol(x) + check)
  }
  if (!is.null(names(add)))
    names(add) <- c("Base", names(add)[-length(add)])
  if (!check)
    add <- c(0, add)
  out <- matrix(add[1], nrow(x), nrow(x))
  for (i in 1:ncol(x)) {
    out <- out + add[i + 1] * outer(x[, i], x[, i], "!=")
  }
  out <- as.dist(out)
  attr(out, "method") <- "taxa2dist"
  attr(out, "steps") <- add
  if (missing(labels)) {
    attr(out, "Labels") <- rownames(x)
  }
  else {
    if (length(labels) != nrow(x))
      warning("Labels are wrong: needed ", nrow(x), " got ",
              length(labels))
    attr(out, "Labels") <- as.character(labels)
  }
  if (!check && any(out <= 0))
    warning("you used 'check=FALSE' and some distances are zero -- was this intended?")
  out
}

#' List of available NCBI taxonomy ranks
#' @noRd
#' @author Vinh Tran {tran@bio.uni-frankfurt.de}
mainTaxonomyRank <- function() {
  return(
    c(
      "isolate","strain","pathogroup","serotype","serogroup",
      "forma","formaspecialis","varietas",
      "genotype","morph","subvariety","biotype",
      "subspecies","species","speciessubgroup",
      "speciesgroup","series","section","subgenus","genus",
      "subtribe","tribe", "subfamily","family","superfamily",
      "parvorder","infraorder","suborder","order","superorder",
      "subcohort","cohort","infraclass","subclass","class","superclass",
      "subphylum","phylum","superphylum",
      "subkingdom","kingdom","superkingdom"
    )
  )
}

#' Get full taxonomy ranks and IDs
#' @noRd
#' @author Vinh Tran {tran@bio.uni-frankfurt.de}
get_rank <- function (x) {
  rank_df <- x[, 'rank']
  names(rank_df) <- x[, 'rank']

  id_df <- x[, 'id']
  joined_df <- cbind(
    data.frame(rank_df, stringsAsFactors = FALSE),
    data.frame(id_df, stringsAsFactors = FALSE)
  )
  joined_df$rank_df[!(joined_df$rank_df %in% mainTaxonomyRank())] <- 'no rank'
  joined_df <- within(
    joined_df,
    rank_df[rank_df=='no rank'] <- paste0("norank_", id_df[rank_df=='no rank'])
  )
  df <- data.frame(
    t(data.frame(rev(joined_df$rank_df))), stringsAsFactors = FALSE
  )
  outDf <- data.frame(tip = x[nrow(x), "name"], df, stringsAsFactors = FALSE)
  return(outDf)
}

#' Get taxonomy names and IDs for all ranks
#' @noRd
#' @author Vinh Tran {tran@bio.uni-frankfurt.de}
get_name <- function (x) {
  rank_df <- x[, 'rank']
  names(rank_df) <- x[, 'rank']

  nameDf <- x[, 'name']
  id_df <- x[, 'id']

  joined_df <- cbind(
    data.frame(rank_df,stringsAsFactors=FALSE),
    data.frame(nameDf,stringsAsFactors=FALSE)
  )
  joined_df$rank_df[!(joined_df$rank_df %in% mainTaxonomyRank())] <- 'no rank'
  joined_df <- within(
    joined_df,
    rank_df[rank_df=='no rank'] <- paste0("norank_",id_df[rank_df=='no rank'])
  )
  joined_df$name <- paste0(joined_df$nameDf, "##", joined_df$rank_df)

  df <- data.frame(t(data.frame(rev(joined_df$name))), stringsAsFactors = FALSE)
  outDf <- data.frame(tip = x[nrow(x), "name"], df, stringsAsFactors = FALSE)
  return(outDf)
}

#' Indexing all available ranks (including norank)
#' @noRd
#' @param rankList dataframe, whose each row is a rank list of a taxon
#' @return A dataframe containing a list of all possible ranks and their indexed
#' values.
#' @author Vinh Tran {tran@bio.uni-frankfurt.de}
rank_indexing <- function (rankList) {
  ### get all available ranks from input rankList
  uList <- unlist(rankList[seq(2, length(rankList))])
  # get final list of available ranks (remove NA items)
  allInputRank <- as.character(unique(uList))
  allInputRank <- allInputRank[!is.na(allInputRank)]

  ### initial index for main ranks
  mainRank <- mainTaxonomyRank()
  rank2index <- new.env(hash = TRUE)
  getHash <- Vectorize(get, vectorize.args = "x")
  assignHash <- Vectorize(assign, vectorize.args = c("x", "value"))
  for (i in seq_len(length(mainRank))) rank2index[[mainRank[i]]] <- i

  ### the magic happens here
  for (k in seq_len(nrow(rankList))) {
    ## get rank list for current taxon containing only ranks in allInputRank
    subList <- rankList[k,][!is.na(rankList[k,])]
    filter <- vapply(
      subList, function(x) x %in% allInputRank, FUN.VALUE = logical(1))
    subList <- subList[filter]
    
    ## indexing
    tmpEnv <- new.env(hash = TRUE)
    flag <- 0
    for (i in seq_len(length(subList))) {
      iRank <- subList[i]
      if (is.null(rank2index[[iRank]])) {
        for (j in seq_len(length(subList))) {
          if (j < i) {
            if (!is.null(tmpEnv[[subList[i - j]]])) {
              tmpEnv[[iRank]] <- tmpEnv[[subList[i - j]]] + 1
              break
            }
          } else j = j - 1
        }
      } else {
        # for old rank
        if (i > 1) {
          if (flag == 0) {
            if (!(iRank %in% ls(rank2index))) stop(paste(iRank,"not found!"))
            currentIndex <- rank2index[[iRank]]
          } else {
            if (!(iRank %in% ls(tmpEnv))) {
              tmpEnv[[iRank]] <- tmpEnv[[subList[i-1]]]
            }
            currentIndex <- tmpEnv[[iRank]]
          }
          
          if (currentIndex <= tmpEnv[[subList[i-1]]]) {
            if (flag == 0) {
              tmpEnv[[iRank]] <- tmpEnv[[subList[i-1]]] + 1
              # list of current ranks that whose index should be increased
              candidateList <- unlist(
                mget(
                  ls(rank2index)[!(ls(rank2index) %in% ls(tmpEnv))], rank2index
                )
              )
              candidateList <- candidateList[order(unlist(candidateList))]
              for (cl in seq_len(length(candidateList))) {
                r <- names(candidateList)[cl]
                fromIndex <- rank2index[[iRank]]
                if(subList[i-1] %in% ls(rank2index)) {
                  fromIndex <- rank2index[[subList[i-1]]]
                }
                
                if (rank2index[[r]] > fromIndex) {
                  tmpEnv[[r]] <-
                    rank2index[[r]] + (tmpEnv[[iRank]] - rank2index[[iRank]])
                  flag <- 1
                }
              }
            } else {
              step <- tmpEnv[[subList[i-1]]] - rank2index[[iRank]] + 1
              tmpEnv[[iRank]] <- tmpEnv[[subList[i-1]]] + 1
              for (tmpRank in ls(tmpEnv)) {
                if (tmpEnv[[tmpRank]] >= tmpEnv[[subList[i-1]]]){
                  if (!(tmpRank == iRank) && !(tmpRank == subList[i-1])){
                    tmpEnv[[tmpRank]] <- tmpEnv[[tmpRank]] + step
                  }
                }
              }
            }
            assignHash(ls(tmpEnv), getHash(ls(tmpEnv), tmpEnv), rank2index)
          } else {
            if (is.null(tmpEnv[[iRank]])) {
              tmpEnv[[iRank]] <- rank2index[[iRank]]
            }
          }
        } else {
          tmpEnv[[iRank]] <- rank2index[[iRank]]
        }
      }
    }
    assignHash(ls(tmpEnv), getHash(ls(tmpEnv), tmpEnv), rank2index)
  }
  
  # convert env into dataframe and return
  index2RankList <- lapply(
    seq_len(length(allInputRank)), function (x) {
      data.frame(
        index = rank2index[[allInputRank[x]]],
        rank = allInputRank[x], stringsAsFactors = FALSE
      )
    }
  )
  index2RankDf <- do.call(rbind, index2RankList)
  index2RankDf <- index2RankDf[with(index2RankDf, order(index2RankDf$index)),]
  return(index2RankDf)
}

#' Align NCBI taxonomy IDs of list of taxa into a sorted rank list
#' @noRd
#' @param nameList a dataframe whose each row is a rank+ID list of a taxon
#' @param rankList a dataframe whose each row is a rank list of a taxon
#' @param remove_shared If `TRUE`, remove any taxa that are coarser ranks
#'   present in other taxa, such as both a genus and a species in that genus in
#'   the same tree.
#' @return An aligned taxonomy dataframe which contains all the available
#'   taxonomy ranks from the id list and rank list
#' @author Vinh Tran {tran@bio.uni-frankfurt.de}
taxonomy_table_creator <- function (nameList, rankList, remove_shared = FALSE) {
  colnames(nameList)[1] <- "tip"
  # remove duplicated taxa (e.g. taxa with higher levels, that already belong 
  # to the taxonomy string of other taxa)
  if (remove_shared) {
    duplicatedTaxa <- lapply(
      nameList$X1,
      function (x) {
        matchs <-data.frame(which(nameList[,-1] == x, arr.ind=TRUE))
        checkHigherRank <- lapply(
          matchs$row,
          function(y) {
            if (nameList[y,]$X1 != x) return(1)
          }
        )
        if (length(unlist(checkHigherRank)) > 0) 
          return(strsplit(x, "#", fixed = TRUE)[[1]][1])
      }
    )
    if (length(unlist(duplicatedTaxa)) > 0) {
      msg <- paste("WARNING:", length(unlist(duplicatedTaxa)))
      if (length(unlist(duplicatedTaxa)) == 1)
        msg <- paste(msg, "duplicated taxon has been ignored!")
      else
        msg <- paste(msg, "duplicated taxa have been ignored!")
      if (length(unlist(duplicatedTaxa)) < 5) {
        msg <- paste(msg, "Including: ")
        msg <- c(msg, paste(unlist(duplicatedTaxa), collapse = "; "))
      }   
      message(msg)
    }
    nameList <- nameList[!(nameList$tip %in% unlist(duplicatedTaxa)),]
    rankList <- rankList[!(rankList$tip %in% unlist(duplicatedTaxa)),]
  }
  # get indexed rank list
  index2RankDf <- rank_indexing(rankList)
  # get ordered rank list
  orderedRank <- factor(index2RankDf$rank, levels = index2RankDf$rank)
  # create a dataframe containing ordered ranks
  full_rank_name_df <- data.frame(
    "rank"= matrix(unlist(orderedRank), nrow = length(orderedRank), byrow=TRUE),
    stringsAsFactors = FALSE
  )
  full_rank_name_df$index <- as.numeric(rownames(full_rank_name_df))
  
  fullRankIDdf <- data.frame(
    rank = matrix(
      unlist(orderedRank), nrow = length(orderedRank), byrow = TRUE
    ), stringsAsFactors = FALSE)
  fullRankIDdf$index <- as.numeric(rownames(fullRankIDdf))

  for (i in 1:nrow(nameList)) {
    ### get list of all IDs for this taxon
    taxonDf <- data.frame(nameList[i,])
    taxonName <- unlist(
      strsplit(as.character(nameList[i,]$tip), "##", fixed = TRUE)
    )

    ### convert into long format
    # mTaxonDf <- suppressWarnings(melt(taxonDf,id = "tip"))
    mTaxonDf <- data.table::setDF(
      suppressWarnings(
        data.table::melt(data.table::as.data.table(taxonDf), id = "tip")
      )
    )

    ### get rank names and corresponding IDs
    splitCol <- data.frame(
      do.call(
        'rbind', strsplit(as.character(mTaxonDf$value), "##", fixed=TRUE)
      )
    )
    mTaxonDf <- cbind(mTaxonDf,splitCol)

    ### remove NA cases
    mTaxonDf <- mTaxonDf[complete.cases(mTaxonDf),]

    ### subselect mTaxonDf to keep only 2 column rank id and rank name
    mTaxonDf <- mTaxonDf[,c("X1","X2")]
    if(mTaxonDf$X2[1] != index2RankDf$rank[1] &&
       !index2RankDf$rank[1] %in% mTaxonDf$X2){
      mTaxonDf <- rbind(
        data.frame("X1"=mTaxonDf$X1[1], "X2"=index2RankDf$rank[1]), mTaxonDf
      )
    }

    ### rename columns
    colnames(mTaxonDf) <- c(taxonName[1],"rank")

    ### merge with index2RankDf (Df contains all available ranks of input data)
    full_rank_name_df <- merge(
      full_rank_name_df,mTaxonDf, by=c("rank"), all.x = TRUE
    )

    ### reorder ranks
    full_rank_name_df <- full_rank_name_df[order(full_rank_name_df$index),]

    ### replace NA id by id of previous rank
    full_rank_name_df <- zoo::na.locf(full_rank_name_df)
  }

  ### remove index column
  full_rank_name_df <- subset(full_rank_name_df, select = -c(index))

  ### transpose into wide format
  t_full_rank_name_df <- transpose(full_rank_name_df)

  ### set first row to column names
  colnames(t_full_rank_name_df) = as.character(unlist(t_full_rank_name_df[1,]))
  t_full_rank_name_df <- t_full_rank_name_df[-1,]

  ### replace NA values in the dataframe t_full_rank_name_df
  if(nrow(t_full_rank_name_df[is.na(t_full_rank_name_df),]) > 0){
    t_full_rank_name_dfTMP <- 
      t_full_rank_name_df[complete.cases(t_full_rank_name_df),]
    t_full_rank_name_dfEdit <- t_full_rank_name_df[is.na(t_full_rank_name_df),]

    for(i in 1:nrow(t_full_rank_name_dfEdit)){
      for(j in 1:(ncol(t_full_rank_name_df)-1)){
        if(is.na(t_full_rank_name_dfEdit[i,j])){
          t_full_rank_name_dfEdit[i,j] <- t_full_rank_name_dfEdit[i,j+1]
        }
      }
      if(is.na(t_full_rank_name_dfEdit[i,ncol(t_full_rank_name_df)])){
        t_full_rank_name_dfEdit[i,ncol(t_full_rank_name_df)] <-
        t_full_rank_name_dfEdit[i,ncol(t_full_rank_name_df)-1]
      }
    }
    t_full_rank_name_df <- rbind(t_full_rank_name_dfEdit,t_full_rank_name_dfTMP)
  }

  ### add fullName column into t_full_rank_name_df
  fullName <- colnames(full_rank_name_df)[-c(1)]
  full_rank_name_df <- cbind(fullName,t_full_rank_name_df)

  ### return
  return(full_rank_name_df)
}

context("class2tree")

spnames <- c("Klattia flava", "Trollius sibiricus", "Arachis paraguariensis",
             "Tanacetum boreale", "Gentiana yakushimensis", "Sesamum schinzianum",
             "Pilea verrucosa", "Tibouchina striphnocalyx", "Lycium dasystemum",
             "Berkheya echinacea", "Androcymbium villosum",
             "Helianthus annuus", "Madia elegans", "Lupinus albicaulis",
             "Pinus lambertiana", "Haloarcula amylolytica JCM 13557",
             "Halomonas sp. 'Soap Lake #6'")

dupnames <- c("Mus musculus", "Escherichia coli",
              "Haloferax denitrificans", "Mus musculus")

duptaxa <- c("Haliotis", "Haliotis cracherodii", "Haliotis rufescens", 
             "Megabalanus californicus")

test_that("internal functions of class2tree", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("class2tree_internal_fxns", {
    out <- classification(spnames, db = "ncbi", messages = FALSE)
    spnames <- c("Proteus mirabilis","Citrus sinensis","Cyanophora paradoxa")
    out2 <- classification(spnames, db = "ncbi", messages = FALSE)
  })
  
  rankList <- dt2df(lapply(out, get_rank), idcol = FALSE)
  expect_equal(nrow(rankList), 17)
  
  nameList <- dt2df(lapply(out, get_name), idcol = FALSE)
  expect_true(identical(colnames(nameList), colnames(rankList)))
  
  strainIndex <- grep("norank", rankList$X1)
  rankList$X1[strainIndex] <- "strain"
  nameList$X1[strainIndex] <- 
    gsub("norank_[[:digit:]]+", "strain", nameList$X1[strainIndex])
  expect_true(length(grep("norank", rankList$X1)) == 0)
  
  indexedRank <- rank_indexing(rankList)
  expect_true(tail(indexedRank, 1)$rank == "norank_131567")
  
  taxMatrix <- taxonomy_table_creator(nameList, rankList)
  expect_is(taxMatrix, "data.frame")
  expect_true(nrow(taxMatrix) == 17)
  
  # wrong indexing
  rankList <- dt2df(lapply(out2, get_rank), idcol = FALSE)
  rankList[3,] <- c(
    "Cyanophora paradoxa","species","genus","family","superkingdom",
    "norank_131567","class",NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA,NA
  )
  indexedRank <- rank_indexing(rankList)
  expect_true(
    indexedRank$index[indexedRank$rank == "kingdom"] <
      indexedRank$index[indexedRank$rank == "class"]
  )
})

test_that("class2tree returns the correct value and class", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("class2tree_classification_call", {
    out <- classification(spnames, db = "ncbi", messages = FALSE)
  })

  out <- out[!is.na(out)]
  tr <- class2tree(out)

  expect_identical(tr$names, names(out))

  expect_is(plot(tr), "list")
  expect_is(tr, "classtree")
  expect_is(tr$phylo, "phylo")
  expect_is(tr$classification, "data.frame")
  expect_is(tr$distmat, "dist")
  expect_is(tr$names, "character")
  expect_is(tr$phylo$node.label, "character")
  expect_equal(length(tr$phylo$node.label), tr$phylo$Nnode)
  expect_equal(
    anyDuplicated(gsub("\\.\\d+$", "", names(tr$classification))), 0)
})

test_that("class2tree will abort when input contains duplicated taxa", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("class2tree_classification_dup_call", {
    out <- classification(dupnames, db = "ncbi", messages = FALSE)
  })
  expect_error(class2tree(out),
    "Input list of classifications contains duplicates")
})

test_that("class2tree detects duplicated taxa in higher levels", {
    skip_on_cran() # uses secrets
    vcr::use_cassette("class2tree_classification_dup_high_level", {
        out <- classification(duptaxa, db = "ncbi", messages = FALSE)
    })
    tree <- class2tree(out, remove_shared = TRUE)
    expect_true(nrow(tree$classification) < length(duptaxa))
})

#' Get TSN from LSID
#'
#' @export
#' @param lsid One or more lsid's
#' @param what What to retrieve. One of tsn, record, or fullrecord
#' @param ... Further arguments passed on to [ritis::lsid2tsn()],
#' [ritis::record()], or [ritis::full_record()]
#' @examples \dontrun{
#' # Get TSN
#' itis_lsid("urn:lsid:itis.gov:itis_tsn:180543")
#' itis_lsid(lsid=c("urn:lsid:itis.gov:itis_tsn:180543","urn:lsid:itis.gov:itis_tsn:28726"))
#'
#' # Get partial record
#' itis_lsid("urn:lsid:itis.gov:itis_tsn:180543", "record")
#'
#' # Get full record
#' itis_lsid("urn:lsid:itis.gov:itis_tsn:180543", "fullrecord")
#'
#' # An invalid lsid (a tsn actually)
#' itis_lsid(202385)
#' }
itis_lsid <- function(lsid=NULL, what='tsn', ...) {
  temp <- switch(
    what,
    tsn = lapply(lsid, function(x) ritis::lsid2tsn(x, ...)),
    record = lapply(lsid, function(x) ritis::record(x, ...)),
    fullrecord = lapply(lsid, function(x) ritis::full_record(lsid = x, ...))
  )
  if (length(lsid) == 1) {
    temp[[1]]
  } else {
    setNames(temp, lsid)
  }
}

context("itis_lsid")

test_that("itis_lsid returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("itis_lsid", {
    one <- itis_lsid("urn:lsid:itis.gov:itis_tsn:180543")
    two <- itis_lsid("urn:lsid:itis.gov:itis_tsn:180543", "record")
    three <- itis_lsid("urn:lsid:itis.gov:itis_tsn:202385", "fullrecord")
    four <- itis_lsid(202385)
  })

  expect_equal(one, "180543")
  expect_match(as.character(two[1, "genuspart"]), "Ursus")
  expect_match(three$acceptedNameList$tsn, "202385")
  expect_null(four)

  expect_is(one, "character")
  expect_is(two, "data.frame")
  expect_is(three, "list")
})

#' @title Search for taxonomic name details using the Global Names Index
#'
#' @description Uses the Global Names Index, see http://gni.globalnames.org/
#'
#' @export
#' @param id Name id. Required.
#' @param all_records If all_records is 1, GNI returns all records from all
#' 		repositories for the name string (takes 0, or 1 \[default\]).
#' @param ... Curl options passed on to [crul::verb-GET]
#' @author Scott Chamberlain
#' @return Data.frame of results.
#' @seealso [gnr_datasources()], [gna_search()].
#' @keywords globalnamesindex names taxonomy
#' @examples \dontrun{
#' gni_details(id = 17802847)
#'
#' # pass on curl options
#' gni_details(id = 17802847, verbose = TRUE)
#' }
gni_details <- function(id, all_records = 1, ...) {
  .Defunct(msg = "This function is defunct - See ?`taxize-defunct`")
  
#   calls <- names(sapply(match.call(), deparse))[-1]
#   calls_vec <- "url" %in% calls
#   if (any(calls_vec)) stop("The parameter url has been removed", call. = FALSE)
# 
# 	url2 <- paste0(gni_base(), "name_strings/", id, ".json")
# 	query <- tc(list(all_records = all_records))
# 	cli <- crul::HttpClient$new(url2, headers = tx_ual, opts = list(...))
# 	tt <- cli$get(query = argsnull(query))
# 	tt$raise_for_status()
#   out <- jsonlite::fromJSON(tt$parse("UTF-8"), FALSE)
# 	outdf <-
# 		dt2df(lapply(out$data, function(x)
# 			data.frame(t(c(checknull(x$records[[1]]$created_at),
# 				checknull(x$records[[1]]$updated_at),
# 				checknull(x$records[[1]]$global_id),
# 				checknull(x$records[[1]]$url),
# 				checknull(x$records[[1]]$kingdom_id),
# 				checknull(x$records[[1]]$original_name_string),
# 				checknull(x$records[[1]]$id),
# 				checknull(x$records[[1]]$name_rank_id),
# 				checknull(x$records[[1]]$name_index_id),
# 				checknull(x$records[[1]]$record_hash),
# 				checknull(x$records[[1]]$local_id),
# 				checknull(x$records[[1]]$nomenclatural_code_id) )))), idcol = FALSE)
# 	stats::setNames(outdf, c(
# 	  "created_at","updated_at","global_id","url","kingdom_id",
# 	  "original_name_string","id","name_rank_id","name_index_id","record_hash",
# 	  "local_id","nomenclatural_code_id"
# 	))
}

# context("gni_details")
# 
# test_that("gni_details returns the correct value", {
#   skip_on_cran()
#   vcr::use_cassette("gni_details", {
#     x <- gni_details(id = 17802847)
#   })
# 
#   expect_is(x, "data.frame")
#   expect_match(as.character(x[,3]), "none")
# })

context("gni_details")

test_that("gni_details returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("gni_details", {
    x <- gni_details(id = 17802847)
  })

  expect_is(x, "data.frame")
  expect_match(as.character(x[,3]), "none")
})

#' taxize parameters
#' 
#' Information on standardized parameters across the package
#' 
#' @name taxize-params
#' @section Standardized parameters:
#' 
#' - `sci`: scientific name
#' - `com`: common name
#' - `id`: name identifier
#' - `sci_com`: scientific name or common name
#' - `sci_id`: scientific name or name identifier
#' 
#' We were going to standardize parameter names for cases in which
#' a parameter accepts either of three options: scientific name,
#' common name, or name identifier. However, there was no clear
#' parameter name we could use for this case, so we've left
#' parameter names as they are for the two cases ([get_ids()] and
#' [vascan_search()])
NULL

#' @title Retrieve the upstream taxa for a given taxon name or ID.
#'
#' @description This function uses a while loop to continually collect taxa up to the
#' taxonomic rank that you specify in the `upto` parameter. You can get data
#' from ITIS (itis) only currently. There is no method exposed by itis
#' for getting taxa at a specific taxonomic rank, so we do it ourselves inside
#' the function.
#'
#' @export
#' @param sci_id Vector of taxa names (character) or IDs (character or numeric) to
#' query.
#' @param db character; database to query. One or both of `itis`. Note
#' that each taxonomic data source has their own identifiers, so that if you
#' provide the wrong `db` value for the identifier you could get a result, but
#' it will likely be wrong (not what you were expecting).
#' @param upto What taxonomic rank to go down to. One of: 'superkingdom',
#' 'kingdom', 'subkingdom','infrakingdom','phylum','division','subphylum',
#' 'subdivision','infradivision', 'superclass','class','subclass','infraclass',
#' 'superorder','order','suborder','infraorder','superfamily','family',
#' 'subfamily','tribe','subtribe','genus','subgenus', 'section','subsection',
#' 'species','subspecies','variety','form','subvariety','race', 'stirp',
#' 'morph','aberration','subform', or 'unspecified'
#' @param rows (numeric) Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this parameter is ignored if you pass in a
#' taxonomic id of any of the acceptable classes: tsn.
#' @param x Deprecated, see `sci_id`
#' @param ... Further args passed on to [itis_downstream()]
#'
#' @return A named list of data.frames with the upstream names of every
#' supplied taxa. You get an NA if there was no match in the database.
#'
#' @examples \dontrun{
#' upstream('Pinus contorta', db = 'itis', upto = 'genus')
#' }
upstream <- function(...) {
  UseMethod("upstream")
}

#' @export
#' @rdname upstream
upstream.default <- function(sci_id, db = NULL, upto = NULL, rows = NA,
  x = NULL, ...) {

  nstop(upto, "upto")
  nstop(db)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "upstream(x)", with = "upstream(sci_id)")
    sci_id <- x
  }
  
  switch(
    db,
    itis = {
      id <- process_stream_ids(sci_id, db, get_tsn, rows = rows, ...)
      setNames(upstream(id, upto = tolower(upto), ...), sci_id)
    },
    stop("the provided db value was not recognised", call. = FALSE)
  )
}

#' @export
#' @rdname upstream
upstream.tsn <- function(sci_id, db = NULL, upto = NULL, ...) {
  warn_db(list(db = db), "itis")
  fun <- function(y, ...){
    # return NA if NA is supplied
    if (is.na(y)) {
      NA
    } else {
      class <- classification(y, ...)
      toget <- class[[1]][ grep(upto, class[[1]]$rank) - 1, c("name", "id") ]
      setNames(downstream(as.tsn(toget$id), downto = upto, ...), toget$name)
    }
  }
  out <- if (length(sci_id) > 1) lapply(sci_id, fun, ...) else fun(sci_id, ...)
  structure(out, class = 'upstream', db = 'itis')
}

#' @export
#' @rdname upstream
upstream.ids <- function(sci_id, db = NULL, upto = NULL, ...) {
  fun <- function(y, ...){
    # return NA if NA is supplied
    if (is.na(y)) {
      out <- NA
    } else {
      out <- upstream(y, upto = tolower(upto), ...)
    }
    return(out)
  }
  out <- if (length(sci_id) > 1) lapply(sci_id, fun, ...) else fun(sci_id, ...)
  structure(out, class = 'downstream_ids')
}

context("upstream")

test_that("upstream basic usage works", {
  skip_on_cran()
  sw(vcr::use_cassette("upstream", {
    aa <- sw(upstream("Pinus contorta", db = 'itis', upto = 'genus',
      messages = FALSE))
  }, preserve_exact_body_bytes = TRUE))

  expect_is(aa, "upstream")
  expect_named(aa, "Pinus contorta")
  expect_is(aa$`Pinus contorta`$tsn, "character")
  expect_is(aa$`Pinus contorta`$taxonname, "character")
  expect_is(aa$`Pinus contorta`$rankid, "character")
})

# FIXME: no longer needed as only 1 option now
# test_that("warn on mismatch 'db'", {
#   skip_on_cran()
#   expect_warning(
#     upstream(
#       get_tsn('Pinus contorta', messages = FALSE), db = 'col',
#       upto = 'genus'))
# })

#' Get NatureServe taxonomic ID for a taxon name
#'
#' @export
#' @param sci_com character; A vector of common or scientific names. Or, a
#' `taxon_state` object (see [taxon-state])
#' @param searchtype character; One of 'scientific' (default) or 'common'.
#' This doesn't affect the query to NatureServe - but rather affects what
#' column of data is targeted in name filtering post data request.
#' @param ask logical; should get_natservid be run in interactive mode?
#' If `TRUE` and more than one wormsid is found for the species, the
#' user is asked for input. If `FALSE` NA is returned for
#' multiple matches. default: `TRUE`
#' @param messages logical; should progress be printed? default: `TRUE`
#' @param rows numeric; Any number from 1 to infinity. If the default NaN, all
#' rows are considered. Note that this function still only gives back a
#' natservid class object with one to many identifiers. See
#' `get_natservid_()` to get back all, or a subset, of the raw
#' data that you are presented during the ask process.
#' @param x Input to `as.natservid`
#' @param query Deprecated, see `sci_com`
#' @param ... curl options passed on to [crul::verb-POST]
#' @param check logical; Check if ID matches any existing on the DB, only
#' used in [as.natservid()]
#' @template getreturn
#' @family taxonomic-ids
#' @seealso [classification()]
#' @note Authentication no longer required
#' @examples \dontrun{
#' (x <- get_natservid("Helianthus annuus", verbose = TRUE))
#' attributes(x)
#' attr(x, "match")
#' attr(x, "multiple_matches")
#' attr(x, "pattern_match")
#' attr(x, "uri")
#'
#' get_natservid('Gadus morhua')
#' get_natservid(c("Helianthus annuus", 'Gadus morhua'))
#'
#' # specify rows to limit choices available
#' get_natservid('Ruby Quaker Moth', 'common')
#' get_natservid('Ruby*', 'common')
#' get_natservid('Ruby*', 'common', rows=1)
#' get_natservid('Ruby*', 'common', rows=1:2)
#'
#' # When not found
#' get_natservid("howdy")
#' get_natservid(c('Gadus morhua', "howdy"))
#'
#' # Convert a natservid without class information to a natservid class
#' # already a natservid, returns the same
#' as.natservid(get_natservid('Pomatomus saltatrix'))
#' # same
#' as.natservid(get_natservid(c('Gadus morhua', 'Pomatomus saltatrix')))
#' # character
#' as.natservid(101905)
#' # character vector, length > 1
#' as.natservid(c(101905, 101998))
#' # list, either numeric or character
#' as.natservid(list(101905, 101998))
#' ## dont check, much faster
#' as.natservid(101905, check = FALSE)
#' as.natservid(c(101905, 101998), check = FALSE)
#' as.natservid(list(101905, 101998), check = FALSE)
#'
#' (out <- as.natservid(c(101905, 101998), check = FALSE))
#' data.frame(out)
#' as.natservid( data.frame(out) )
#'
#' # Get all data back
#' get_natservid_("Helianthus")
#' get_natservid_("Ruby*", searchtype = "common")
#' get_natservid_("Ruby*", searchtype = "common", rows=1:3)
#' }
get_natservid <- function(sci_com, searchtype = "scientific", ask = TRUE,
                          messages = TRUE, rows = NA, query = NULL, ...) {

  assert(sci_com, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(searchtype, "character")
  assert(ask, "logical")
  assert(messages, "logical")
  assert_rows(rows)
  if (!is.null(query)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_natservid(query)", with = "get_natservid(sci_com)")
    sci_com <- query
  }
  

  if (inherits(sci_com, "character")) {
    tstate <- taxon_state$new(class = "natservid", names = sci_com)
    items <- sci_com
  } else {
    assert_state(sci_com, "natservid")
    tstate <- sci_com
    sci_com <- tstate$taxa_remaining()
    items <- c(sci_com, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci_com)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci_com[i], "'\n")

    if (!searchtype %in% c("scientific", "common")) {
      stop("'searchtype' must be one of 'scientific' or 'common'",
        call. = FALSE)
    }

    nsdf <- ns_worker(x = sci_com[i], searchtype = searchtype, ...)
    mm <- NROW(nsdf) > 1

    if (!inherits(nsdf, "tbl_df") || NROW(nsdf) == 0) {
      nsid <- NA_character_
      att <- "not found"
    } else {
      nsdf <- suppressWarnings(data.frame(nsdf))
      nsdf <- sub_rows(nsdf, rows)

      # should return NA if spec not found
      if (nrow(nsdf) == 0) {
        mssg(messages, m_not_found_sp_altclass)
        nsid <- NA_character_
        att <- 'not found'
      }

      # take the one nsid from data.frame
      if (nrow(nsdf) == 1) {
        nsid <- nsdf$id
        att <- 'found'
      }

      # check for direct match
      if (nrow(nsdf) > 1) {

        names(nsdf)[grep(searchtype, names(nsdf))] <- "target"
        direct <- match(tolower(nsdf$target), tolower(sci_com[i]))

        if (length(direct) == 1) {
          if (!all(is.na(direct))) {
            nsid <- nsdf$id[!is.na(direct)]
            direct <- TRUE
            att <- 'found'
          } else {
            direct <- FALSE
            nsid <- NA_character_
            att <- 'not found'
          }
        } else {
          direct <- FALSE
          nsid <- NA_character_
          att <- m_na_ask_false_no_direct
          warning("> 1 result; no direct match found", call. = FALSE)
        }
      }

      # multiple matches
      if (any(
        nrow(nsdf) > 1 && is.na(nsid) ||
        nrow(nsdf) > 1 && att == "found" && length(nsid) > 1
      )) {
        if (ask) {
          names(nsdf)[grep(searchtype, names(nsdf))] <- "target"

          # user prompt
          nsdf <- nsdf[order(nsdf$target), ]

          # prompt
          message("\n\n")
          rownames(nsdf) <- seq_len(NROW(nsdf))
          message(paste0(utils::capture.output(nsdf), collapse = "\n"))
          message("\nMore than one NatureServe ID found for taxon '", sci_com[i], "'!\n
                  Enter rownumber of taxon (other inputs will return 'NA'):\n") # prompt
          take <- scan(n = 1, quiet = TRUE, what = 'raw')

          if (length(take) == 0) {
            take <- 'notake'
            att <- 'nothing chosen'
          }
          if (take %in% seq_len(nrow(nsdf))) {
            take <- as.numeric(take)
            message("Input accepted, took taxon '", as.character(nsdf$target[take]), "'.\n")
            nsid <-  nsdf$id[take]
            att <- 'found'
          } else {
            nsid <- NA_character_
            mssg(messages, "\nReturned 'NA'!\n\n")
            att <- 'not found'
          }
        } else {
          if (length(nsid) != 1) {
            warning(sprintf(m_more_than_one_found, "NatureServe ID", sci_com[i]),
              call. = FALSE)
            nsid <- NA_character_
            att <- m_na_ask_false
          }
        }
      }

    }
    res <- list(id = as.character(nsid), att = att, multiple = mm,
      direct = direct)
    prog$completed(sci_com[i], att)
    prog$prog(att)
    tstate$add(sci_com[i], res)
  }
  out <- tstate$get()
  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "natservid",
                   match = pluck_un(out, "att", ""),
                   multiple_matches = pluck_un(out, "multiple", logical(1)),
                   pattern_match = pluck_un(out, "direct", logical(1)))
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$natserv)
}

#' @export
#' @rdname get_natservid
as.natservid <- function(x, check=TRUE) UseMethod("as.natservid")

#' @export
#' @rdname get_natservid
as.natservid.natservid <- function(x, check=TRUE) x

#' @export
#' @rdname get_natservid
as.natservid.character <- function(x, check=TRUE) if (length(x) == 1) make_natserv(x, check) else collapse(x, make_natserv, "natservid", check = check)

#' @export
#' @rdname get_natservid
as.natservid.list <- function(x, check=TRUE) if (length(x) == 1) make_natserv(x, check) else collapse(x, make_natserv, "natservid", check = check)

#' @export
#' @rdname get_natservid
as.natservid.numeric <- function(x, check=TRUE) as.natservid(as.character(x), check)

#' @export
#' @rdname get_natservid
as.natservid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class = "natservid", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri = x$uri)
}

#' @export
#' @rdname get_natservid
as.data.frame.natservid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "natservid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_natserv <- function(x, check=TRUE) make_generic(x, ns_base_uri(), "natservid", check)

check_natservid <- function(x){
  tt <- crul::HttpClient$new(sprintf(ns_base_uri(), x),
    headers = tx_ual)$get()$parse("UTF-8")
  !grepl("No records matched", tt)
}

#' @export
#' @rdname get_natservid
get_natservid_ <- function(sci_com, searchtype = "scientific", messages = TRUE,
  rows = NA, query = NULL, ...) {

  if (!is.null(query)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_natservid_(query)", with = "get_natservid_(sci_com)")
    sci_com <- query
  }
  stats::setNames(
    lapply(sci_com, get_natservid_help, searchtype = searchtype,
      messages = messages, rows = rows, ...),
    sci_com
  )
}

get_natservid_help <- function(sci_com, searchtype, messages, rows, ...) {
  mssg(messages, "\nRetrieving data for taxon '", sci_com, "'\n")
  df <- ns_worker(x = sci_com, searchtype = searchtype, ...)
  sub_rows(df, rows)
}

ns_base_uri <- function() "https://explorer.natureserve.org/Taxon/ELEMENT_GLOBAL.2.%s"

# x = "Helianthus annuus"; searchtype = "scientific"
# x = "Ruby*"; searchtype = "common"
ns_worker <- function(x, searchtype, ...) {
  query <- switch(searchtype,
    scientific = list(searchToken=x, matchAgainst="allScientificNames", operator="similarTo"),
    common = list(searchToken=x, matchAgainst="allCommonNames", operator="similarTo")
  )
  # FIXME: gotta have pagination, but results not returning as expected, look into natserv bug
  tmp <- tryCatch(
    # natserv::ns_search_spp(text_adv = query, page = 1, per_page = 2000L, ...),
    natserv::ns_search_spp(text_adv = query, ...),
    error = function(e) e)
  if (inherits(tmp, "error")) return(tibble::tibble())
  if (NROW(tmp$results) == 0) return(tibble::tibble())
  tmp <- tmp$results[, c("elementGlobalId", "scientificName", "primaryCommonName", "nsxUrl")]
  names(tmp) <- c('id', 'scientificname', 'commonname', 'uri')
  tmp
}

context("get_natservid")

test_that("get_natservid returns the correct value", {
  skip_on_cran()

  vcr::use_cassette("get_natservid", {
    x <- get_natservid(c('Pomatomus saltatrix', "howdy"), messages=FALSE)[2]
    w <- get_natservid(c("Helianthus annuus", 'Pomatomus saltatrix'), 
        messages=FALSE)
  })
    
  expect_true(is.na(x))
  expect_is(w, "natservid")
})

test_that("get_natservid accepts ask-argument", {
  skip_on_cran()
  
  vcr::use_cassette("get_natservid_ask_arg", {
    x <- get_natservid('howdy', ask = FALSE, messages=FALSE)
  })

  expect_true(is.na(x))
})

test_that("get_natservid fails well", {
  skip_on_cran()

  expect_true(is.na(get_natservid("asdfadsf", messages = FALSE)))

  expect_error(get_natservid(), "argument \"sci_com\" is missing")
  expect_error(get_natservid("clam", 5),
               "searchtype must be of class character")
  expect_error(get_natservid("clam", "stuff", messages = FALSE),
               "'searchtype' must be one of")
  expect_error(get_natservid("clam", ask = 4),
               "ask must be of class logical")

  # rows param
  expect_error(get_natservid('Ruby*', 'common', rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_natservid('Ruby*', 'common', rows = 0, messages = FALSE),
               "all\\(rows > 0\\) is not TRUE")
})

#' Search for terms in EOL database.
#'
#' @export
#' @param sci (character) scientific name
#' @param page A maximum of 30 results are returned per page. This parameter
#' allows you to fetch more pages of results if there are more than
#' 30 matches (Default 1)
#' @param exact Will find taxon pages if the preferred name or any synonym
#' or common name exactly matches the search term.
#' @param filter_tid Given an EOL page ID, search results will be limited
#' to members of that taxonomic group
#' @param filter_heid Given a Hierarchy Entry ID, search results will be
#' limited to members of that taxonomic group
#' @param filter_by_string Given a search term, an exact search will be made
#' and that matching page will be used as the taxonomic group against which
#' to filter search results
#' @param cache_ttl The number of seconds you wish to have the response cached.
#' @param terms Deprecated, see `sci`
#' @param ... Curl options passed on to [crul::HttpClient]
#' @details It's possible to return JSON or XML with the EOL API. However,
#' 		this function only returns JSON for now.
#'
#' @return A data frame with four columns:
#'
#' * pageid: pageid, this is the same as the eolid you can get from
#' [get_eolid()]
#' * name: taxonomic name, may or may not contain the taxonomic authority
#' * link: URL for the taxon in question
#' * content: a string of semi-colon separated names. it's not clear
#' to us what these represent exactly, but figured why not give it to users
#' in case some may find it useful
#'
#' @examples \dontrun{
#' eol_search(sci='Homo')
#' eol_search(sci='Salix', verbose = TRUE)
#' eol_search(sci='Ursus americanus')
#' eol_search('Pinus contorta')
#' }

eol_search <- function(sci, page=1, exact=NULL, filter_tid=NULL,
  filter_heid=NULL, filter_by_string=NULL, cache_ttl=NULL, terms = NULL, ...) {

  if (!is.null(terms)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "eol_search(terms)", with = "eol_search(sci)")
    sci = terms
  }
  
	query <- gsub("\\s", "+", sci)
  args <- tc(list(q = query, page = page, exact = exact,
                  filter_by_taxon_concept_id = filter_tid,
                  filter_by_hierarchy_entry_id = filter_heid,
                  filter_by_string = filter_by_string,
                  cache_ttl = cache_ttl))
  cli <- crul::HttpClient$new(
    url = paste0(eol_url("search"), ".json"),
    headers = tx_ual,
    opts = list(...)
  )
  res <- cli$get(query = argsnull(args))
  res$raise_for_status()
  tt <- res$parse("UTF-8")
  stopifnot(res$response_headers$`content-type`[1] == 'application/json; charset=utf-8')
  out <- jsonlite::fromJSON(tt, FALSE, encoding = "utf-8")
  if (out$totalResults == 0 | length(out$results) == 0) {
    data.frame(pageid = NA, name = NA, stringsAsFactors = FALSE)
  } else {
    tmp <- dt2df(
      lapply(out$results, function(z) {
        z$content <- unique(unlist(z$content))
        data.frame(z, stringsAsFactors = FALSE)
      }), idcol = FALSE)
    stats::setNames(tmp, c("pageid", "name", "link", "content"))
  }
}

# tests for eol_search fxn in taxize
context("eol_search")

test_that("eol_search returns the correct value", {
  skip_on_cran()

  vcr::use_cassette("eol_search", {
    aa <- eol_search(sci='Ursus americanus')
  })

  expect_is(aa, "data.frame")
	expect_type(aa$pageid, "integer")
  expect_type(aa$name, "character")
  expect_type(aa$link, "character")
})

#' Get kingdom names
#'
#' @export
#' @param tsn One or more TSN's (taxonomic serial number)
#' @param ... Further arguments passed on to getkingdomnamefromtsn
#' @examples \dontrun{
#' itis_kingdomnames(202385)
#' itis_kingdomnames(tsn=c(202385,183833,180543))
#' }
itis_kingdomnames <- function(tsn=NULL, ...) {
  if (is.null(tsn)) {
    ritis::kingdom_names(...)
  } else {
    sapply(
      tsn,
      function(z) {
        suppressWarnings(
          as.character(ritis::kingdom_name(z, ...)$kingdomname)
        )
      }
    )
  }
}

context("itis_kingdomnames")

test_that("itis_kingdomnames - all", {
  skip_on_cran()
  vcr::use_cassette("itis_kingdomnames", {
    aa <- itis_kingdomnames()
  })

  expect_is(aa, "tbl_df")
  expect_named(aa, c('kingdomid', 'kingdomname', 'tsn'))
  expect_true(any(grepl("Animalia", aa$kingdomname)))
  expect_is(aa$kingdomname, "character")
})

test_that("itis_kingdomnames - with TSN's", {
  skip_on_cran()
  vcr::use_cassette("itis_kingdomnames_with_tsns", {
    one <- itis_kingdomnames(202385)
    two <- itis_kingdomnames(tsn = c(202385, 183833, 180543))
  })

  expect_match(one, "Animalia")
  expect_match(two[[1]], "Animalia")

  expect_that(one, is_a("character"))
  expect_that(two, is_a("character"))
})

test_that("itis_kingdomnames returns error when not found", {
  skip_on_cran()
  vcr::use_cassette("itis_kingdomnames_fail_well", {
    expect_error(itis_kingdomnames("stuff"), class = "error")
  })
})

# name backbone
gbif_name_backbone <- function(name, rank = NULL, kingdom = NULL, phylum = NULL,
  class = NULL, order = NULL, family = NULL, genus = NULL, strict = FALSE,
  start = NULL, limit = 500, ...) {


  url = 'https://api.gbif.org/v1/species/match'
  args <- tc(list(name = name, rank = rank, kingdom = kingdom,
                  phylum = phylum, class = class, order = order, family = family,
                  genus = genus, strict = strict, verbose = TRUE, offset = start,
                  limit = limit))
  cli <- crul::HttpClient$new(url = url, headers = tx_ual)
  temp <- cli$get(query = args, ...)
  temp$raise_for_status()
  tt <- jsonlite::fromJSON(temp$parse("UTF-8"), FALSE)

  if (all(names(tt) %in% c('confidence', 'synonym', 'matchType', "note"))) {
    data.frame(NULL)
  } else {
    dd <- data.table::setDF(
      data.table::rbindlist(
        lapply(tt$alternatives,
               function(x) lapply(x, function(x) 
                if (length(x) == 0) NA else x)), use.names = TRUE, fill= TRUE))
    dat <- data.frame(tt[!names(tt) %in% c("alternatives",
                                           "note")], stringsAsFactors = FALSE)
    if (!all(names(dat) %in% c('confidence', 'synonym', 'matchType'))) {
      dd <- dt2df(list(dat, dd), idcol = FALSE)
    }
    if (limit > 0) {
      dd <- cols_move(dd, back_cols_use)
    }
    if (!is.null(dd)) dd$rank <- tolower(dd$rank)
    names(dd) <- tolower(names(dd))
    return(dd)
  }
}

# name lookup
gbif_name_lookup <- function(query = NULL, rank = NULL, 
  higherTaxonKey = NULL, status = NULL, nameType = NULL, 
  datasetKey = 'd7dddbf4-2cf0-4f39-9b2a-bb099caae36c', limit = 500, 
  start = NULL, ...) {

  url = 'https://api.gbif.org/v1/species/search'
  args <- tc(list(q = query, rank = rank, higherTaxonKey = higherTaxonKey,
                  status = status, nameType = nameType, datasetKey = datasetKey,
                  limit = limit, offset = start))
  cli <- crul::HttpClient$new(url = url, headers = tx_ual)
  temp <- cli$get(query = args, ...)
  temp$raise_for_status()
  tt <- jsonlite::fromJSON(temp$parse("UTF-8"), FALSE)
  dd <- data.table::setDF(
    data.table::rbindlist(lapply(
      tt$results,
      nlkupcleaner), use.names = TRUE, fill = TRUE)
  )
  if (limit > 0) {
    dd <- cols_move(dd, gbif_cols_use)
  }
  if (!is.null(dd)) dd$rank <- tolower(dd$rank)
  names(dd) <- tolower(names(dd))
  return(dd)
}

cols_move <- function (x, cols)  {
  other <- names(x)[!names(x) %in% cols]
  x[, c(cols, other)]
}

nlkupcleaner <- function (x) {
  tmp <- x[!names(x) %in% c("descriptions", "vernacularNames", 
    "higherClassificationMap")]
  lapply(tmp, function(x) {
    if (length(x) == 0) {
      NA
    }
    else if (length(x) > 1 || is(x, "list")) {
      paste0(x, collapse = ", ")
    }
    else {
      x
    }
  })
}

gbif_cols_use <- c("key", "canonicalName", "authorship", "rank", 
  "taxonomicStatus", "synonym")
back_cols_use <- c("usageKey", "scientificName", "rank", "status", "matchType")
gbif_cols_show_backbone <- tolower(c("gbifid", "scientificName", "rank", 
  "status", "matchtype"))
gbif_cols_show_lookup <- tolower(c("gbifid", "canonicalName", "authorship", 
  "rank", "taxonomicStatus", "synonym"))

#' uBio classification
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @rdname ubio_classification-defunct
#' @export
#' @keywords internal
#' @param ... Parameters, ignored
ubio_classification <- function(...) {
  .Defunct(msg = "the uBio API is down, for good as far as we know")
}

# tests for ubio_classification fxn in taxize
context("ubio_classification")

out <- ubio_classification(hierarchiesID = 2483153)

test_that("ubio_classification returns the correct value", {
	expect_that(names(out)[1], matches("data"))
	expect_that(as.character(out$data[,1]), matches("84"))
})

test_that("ubio_classification returns the correct dimensions", {
  expect_that(ncol(out$data), equals(8))
  expect_that(length(out), equals(5))
})

test_that("ubio_classification returns the correct class", {
	expect_that(out$data, is_a("data.frame"))
	expect_that(out, is_a("list"))
})

#' Retrieve all taxa names downstream in hierarchy for WORMS
#'
#' @export
#' @param id (integer) One or more AphiaID's
#' @param downto (character) The taxonomic level you want to go down to. 
#' See examples below. The taxonomic level IS case sensitive, and you do have 
#' to spell it correctly. See [rank_ref_zoo] for spelling.
#' @param intermediate (logical) If `TRUE`, return a list of length two 
#' with target taxon rank names, with additional list of data.frame's of
#' intermediate taxonomic groups. Default: `FALSE`
#' @param start (integer) Record number to start at
#' @param ... crul options passed on to [worrms::wm_children()], including the
#' parameters `marine_only` and `offset`, see `?worrms::wm_children` for 
#' details
#' @return data.frame of taxonomic information downstream to family from e.g.,
#' Order, Class, etc., or if `intermediated=TRUE`, list of length two,
#' with target taxon rank names, and intermediate names.
#' @examples \dontrun{
#' ## the genus Gadus
#' worms_downstream(id = 125732, downto="species")
#' worms_downstream(id = 125732, downto="species", intermediate=TRUE)
#' 
#' worms_downstream(id = 51, downto="class")
#' worms_downstream(id = 51, downto="subclass", intermediate=TRUE)
#' 
#' worms_downstream(id = 105, downto="subclass")
#' 
#' # marine_only parameter
#' worms_downstream(545470, downto = "species")
#' worms_downstream(545470, downto = "species", marine_only = FALSE)
#' }
worms_downstream <- function(id, downto, intermediate = FALSE, start = 1, 
  ...) {

  should_be('intermediate', intermediate, 'logical')

  downto <- tolower(downto)
  poss_ranks <- unique(do.call(c,
    sapply(taxize_ds$rank_ref_zoo$ranks, strsplit, split = ",",
      USE.NAMES = FALSE)))
  downto <- match.arg(downto, choices = poss_ranks)
  torank <- sapply(taxize_ds$rank_ref_zoo[which_rank(downto, zoo=TRUE), "ranks"],
    function(x) strsplit(x, ",")[[1]][[1]], USE.NAMES = FALSE)

  stop_ <- "not"
  notout <- data.frame(rank = "", stringsAsFactors = FALSE)
  out <- list()
  if (intermediate) intermed <- list()
  iter <- 0
  while (stop_ == "not") {
    iter <- iter + 1
    tt <- dt2df(lapply(id, function(x)
      worms_children(x, start = start, ...)), idcol = FALSE)
    tt <- prune_too_low(tt, downto, zoo = TRUE, ignore_missing_rank = TRUE)

    if (NROW(tt) == 0) {
      out[[iter]] <- data.frame(stringsAsFactors = FALSE)
      stop_ <- "nodata"
    } else {
      if (intermediate) intermed[[iter]] <- tt
      if (NROW(tt[tt$rank == downto, ]) > 0) {
        out[[iter]] <- tt[tt$rank == downto, ]
      }
      if (NROW(tt[!tt$rank == downto, ]) > 0) {
        notout <- tt[!tt$rank %in% torank, ]
      } else {
        notout <- data.frame(rank = downto, stringsAsFactors = FALSE)
      }

      if (all(notout$rank == downto)) {
        stop_ <- "fam"
      } else {
        id <- notout$id
        stop_ <- "not"
      }
    }
  } # end while loop

  tmp <- dt2df(out, idcol = FALSE)
  if (intermediate) {
    list(target = tmp, intermediate = intermed)
  } else {
    tmp
  }
}

worms_children <- function(x, start = 1, ...) {
  bb <- tryCatch(
    worms_children_all(x, start, ...),
    error = function(e) e
  )
  if (inherits(bb, "error")) return(data.frame(NULL))
  # handle missing rank information
  if (all(is.na(bb$rank))) {
    # drop rank column with no data
    bb$rank <- NULL
    bb <- merge(bb, taxize_ds$worrms_ranks, by.x = "taxonRankID", by.y = "id")
  }
  bb$rank <- tolower(bb$rank)
  # clean up rank names
  if (any(grepl("phylum \\(division", bb$rank))) {
    bb$rank[grep("phylum \\(division", bb$rank)] <- "phylum"
  }
  if (any(grepl("subphylum \\(subdivision", bb$rank))) {
    bb$rank[grep("subphylum \\(subdivision", bb$rank)] <- "subphylum"
  }
  names(bb)[names(bb) %in% "AphiaID"] <- "id"
  names(bb)[names(bb) %in% "scientificname"] <- "name"
  bb[c('id', 'name', 'rank')]
}

worms_children_all <- function(id, start=1, ...) {
  out <- worrms::wm_children(as.numeric(id), start=start, ...)
  out <- list(out)
  i <- 1
  while (NROW(out[[length(out)]]) == 50) {
    i <- i + 1
    tmpres <- tryCatch(
      worrms::wm_children(
        as.numeric(id),
        offset = sum(unlist(sapply(out, NROW))),
        ...
      ),
      error = function(e) e
    )
    out[[i]] <- if (inherits(tmpres, "error")) data.frame() else tmpres
  }
  res <- df2dt2tbl(out)
  return(res)
}

context("worms_downstream")

test_that("worms_downstream works", {
  skip_on_cran()
  vcr::use_cassette("worms_downstream", {
    aa <- suppressWarnings(worms_downstream(id = 125732, downto = "species"))
  })

  expect_is(aa, "data.frame")
  expect_type(aa$id, "integer")
  expect_is(aa$name, "character")
  expect_is(aa$rank, "character")

  expect_equal(unique(aa$rank), "species")
})

test_that("worms_downstream intermediate param works", {
  skip_on_cran()
  vcr::use_cassette("worms_downstream_intermediate_param", {
    cc <- suppressWarnings(worms_downstream(id = 125732, downto = "species", 
      intermediate = TRUE))
  })

  expect_is(cc, "list")
  expect_is(cc$target, "data.frame")
  expect_is(cc$intermediate, "list")

  expect_is(cc$target$name, "character")
  expect_is(cc$target$rank, "character")
  expect_is(cc$target$id, "integer")

  expect_equal(unique(cc$target$rank), "species")
})

test_that("worms_downstream fails well", {
  skip_on_cran()

  expect_error(suppressWarnings(worms_downstream(198, "adfadf")),
    "'arg' should be one of")
  expect_error(
    suppressWarnings(worms_downstream(198, "Genus", intermediate = "adf")),
               "'intermediate' should be of class 'logical'")
})

test_that("worms_downstream - start param", {
  skip_on_cran()
  vcr::use_cassette("worms_downstream_start_param", {
    aa <- suppressWarnings(worms_downstream(125732, "species", start = 1))
    bb <- suppressWarnings(worms_downstream(125732, "species", start = 3))
  })

  expect_is(aa, "data.frame")
  expect_is(bb, "data.frame")
  expect_is(aa$name, "character")
  expect_match(bb$name[1], "Gadus")
})

test_that("worms_downstream - rank names that were previously unaccounted for", {
  skip_on_cran()
  vcr::use_cassette("worms_downstream_previously_unaccounted_for_ranks", {
    # previously unaccounted for: supertribe 
    aa <- suppressWarnings(worms_downstream(123165, downto = "genus"))
    # previously unaccounted for: subterclass 
    bb <- suppressWarnings(worms_downstream(1424949, downto = "order"))
    # previously unaccounted for: subterclass 
    cc <- suppressWarnings(worms_downstream(224570, downto = "order"))
  })

  expect_is(aa, "data.frame")
  expect_is(bb, "data.frame")
  expect_is(cc, "data.frame")
})

test_that("worms_downstream - can pass through marine_only parameter", {
  skip_on_cran()
  vcr::use_cassette("worms_downstream_marine_only_para", {
    # should return one row, Didelphimorphia
    aa <- worms_downstream(1837, downto = "order")
    # should return 10 rows
    bb <- worms_downstream(1837, downto = "order", marine_only=FALSE)
  })

  expect_is(aa, "data.frame")
  expect_is(bb, "data.frame")
  expect_gt(NROW(bb), NROW(aa))
})

#' @title Search NCBI for children of a taxon
#'
#' @description Search the NCBI Taxonomy database for uids of children of taxa.
#' Taxa can be referenced by name or uid. Referencing by name is faster
#'
#' In a few cases, different taxa have the same name (e.g. Satyrium; see 
#' examples). If one of these are searched for then the children of both taxa 
#' will be returned. This can be avoided by using a uid instead of the name or 
#' specifying an ancestor. If an ancestor is provided, only children of both 
#' the taxon and its ancestor are returned. This will only fail if there are 
#' two taxa with the same name and the same specified ancestor.
#'
#' @export
#' @param name (`character`) The string to search for. Only exact matches 
#' found the name given will be returned. Not compatible with `id`.
#' @param id (`character`/`numeric`) The uid to search for. Not compatible with 
#' `name`.
#' @param start The first record to return. If omitted, the results are 
#' returned from the first record (start=0).
#' @param max_return (`numeric; length=1`) The maximum number of children
#' to return.
#' @param ancestor (`character`) The ancestor of the taxon being searched 
#' for. This is useful if there could be more than one taxon with the same 
#' name. Has no effect if `id` is used.
#' @param out_type (character) Currently either `"summary"` or `"uid"`:
#' * `summary` The output is a list of `data.frame` with children uid, name,
#' and rank.
#' * `uid` A list of character vectors of children uids
#' @param ambiguous `logical; length 1` If `FALSE`, children taxa 
#' with words like "unclassified", "unknown", "uncultured", or "sp." are 
#' removed from the output. NOTE: This option only applies when 
#' `out_type= "summary"`.
#' @param key (character) NCBI Entrez API key. optional. See Details.
#' @param ... Curl options passed on to [crul::HttpClient]
#' @return The output type depends on the value of the `out_type` 
#' parameter. Taxa that cannot be found will result in `NA`s and a lack 
#' of children results in an empty data structure.
#' @seealso [ncbi_get_taxon_summary()], [children()]
#' @author Zachary Foster \email{zacharyfoster1989@@gmail.com}
#' 
#' @section Authentication:
#' See [taxize-authentication()] for help on authentication. 
#' We strongly recommend getting an API key
#' 
#' @section HTTP version:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')` 
#' 
#' @section Rate limits:
#' In case you run into errors due to your rate limit being exceeded, see
#' [taxize_options()], where you can set `ncbi_sleep`.
#'
#' @examples \dontrun{
#' ncbi_children(name="Satyrium") #Satyrium is the name of two different genera
#' ncbi_children(name="Satyrium", ancestor="Eumaeini") # A genus of butterflies
#' ncbi_children(name="Satyrium", ancestor="Orchidaceae") # A genus of orchids
#' ncbi_children(id="266948") #"266948" is the uid for the butterfly genus
#' ncbi_children(id="62858") #"62858" is the uid for the orchid genus
#'
#' # use curl options
#' ncbi_children(name="Satyrium", ancestor="Eumaeini", verbose = TRUE)
#' }
ncbi_children <- function(name = NULL, id = NULL, start = 0, max_return = 1000,
                          ancestor = NULL, out_type = c("summary", "uid"), 
                          ambiguous = FALSE, key = NULL, ...) {

  assert(name, "character")
  assert(id, c("character", "numeric", "uid"))
  key <- getkey(key, "ENTREZ_KEY")

  # Constants -----------------------------------------------------------------
  ambiguous_regex <- paste(
     sep = "|", "unclassified", "environmental", "uncultured", "unknown",
     "unidentified", "candidate", "\\ssp\\.", "s\\.l\\.", "sensu lato", "clone",
     "miscellaneous", "candidatus", "affinis", "aff\\.", "incertae sedis",
     "mixed", "samples", "libaries")
  # Argument validation -------------------------------------------------------
  if (sum(c(is.null(name), is.null(id))) != 1) {
    stop("Either name or id must be specified, but not both")
  }
  out_type <- match.arg(out_type)
  # Get name from id ----------------------------------------------------------
  if (is.null(name)) {
    if (!inherits(id, 'uid')) attr(id, 'class') <- 'uid'
    id_taxonomy <- classification(id, db = 'ncbi')
    id_taxonomy <- lapply(id_taxonomy, function(z) {
      if (!(inherits(z, "data.frame"))) data.frame(NULL) else z
    })
    name <- vapply(id_taxonomy,
                   function(x) ifelse(nrow(x) > 0, x$name[nrow(x)], 
                    as.character(NA)), character(1))
    ancestor <- vapply(id_taxonomy,
                       function(x) ifelse(nrow(x) > 1, x$name[nrow(x) - 1], 
                        as.character(NA)), character(1))
    # set id to NULL because we're using name now
    id <- NULL
  } else if (is.null(ancestor)) {
    ancestor <- rep(NA, length(name))
  }
  # Function to search for queries one at a time ------------------------------
  single_search <- function(name, ancestor, ...) {
    if (is.na(name)) return(NA)
    # Make eutils esearch query  - - - - - - - - - - - - - - - - - - - - - - - 
    if (is.na(ancestor)) {
      ancestor_query <- NULL
    } else {
      ancestor_query <- paste0("+AND+", ancestor, "[subtree]")
    }
    
    if (is.null(id)) {
      args <- tc(list(
        db = 'taxonomy', 
        term = paste0(name, "[Next+Level]", ancestor_query),
        RetMax = max_return,
        RetStart = start,
        api_key = key
      ))
      args$term <- gsub("\\+", " ", args$term)
      
      # Search ncbi for children - - - - - - - - - - - - - - - - - - - - - - - 
      cli <- crul::HttpClient$new(ncbi_base(), headers = tx_ual,
        opts = list(http_version = 2L, ...))
      rr <- cli$get('entrez/eutils/esearch.fcgi', query = args)
      rr$raise_for_status()
      raw_results <- rr$parse("UTF-8")
      
      # Parse results  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
      results <- xml2::read_xml(raw_results)
      children_uid <- xml_text_all(results, "//eSearchResult/IdList/Id")
      
    } else {
      args <- tc(list(
        dbfrom = 'taxonomy',
        db = 'taxonomy',
        id = id,
        term = paste0(name, "[Next+Level]"),
        RetMax = max_return,
        RetStart = start,
        api_key = key
      ))
      args$term <- gsub("\\+", " ", args$term)
      
      # Search ncbi for children - - - - - - - - - - - - - - - - - - - - - - - 
      cli <- crul::HttpClient$new(ncbi_base(), headers = tx_ual,
        opts = list(http_version = 2L, ...))
      rr <- cli$get('entrez/eutils/elink.fcgi', query = args)
      rr$raise_for_status()
      raw_results <- rr$parse("UTF-8")
      
      # Parse results  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
      results <- xml2::read_xml(raw_results)
      children_uid <- xml_text_all(results, 
        "//eLinkResult/LinkSet/LinkSetDb/Link/Id")
    }
    if (length(children_uid) == 0) {
      if (out_type == "summary") {
        output <- data.frame(
          childtaxa_id     = character(),
          childtaxa_name   = character(),
          childtaxa_rank   = character(),
          stringsAsFactors = FALSE
        )
      } else {
        output <- numeric()
      }
    } else {
      if (out_type == "summary") {
        output <- ncbi_get_taxon_summary(children_uid, key = key, ...)
        names(output) <- c("childtaxa_id", "childtaxa_name", "childtaxa_rank")
        # Remove ambiguous results  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
        if (!ambiguous) {
          output <- output[!grepl(ambiguous_regex, output$childtaxa_name, 
            ignore.case = TRUE), ]
        }
      } else {
        output <- children_uid
      }
      # numeric row names can be misleading after filtering
      rownames(output) <- NULL 
    }
    # NCBI limits requests to three per second
    ncbi_rate_limit_pause(key)
    return(output)
  }
  # Combine the result of multiple searches -----------------------------------
  # output <- Map(single_search, name, ancestor)
  output <- list()
  for (i in seq_along(name)) {
    output[[i]] <- single_search(name[[i]], ancestor[[i]], ...)
  }
  if (is.null(id)) names(output) <- name else names(output) <- id
  return(output)
}

context("ncbi_children")

test_that("ncbi_children returns correct class and result", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("ncbi_children", {
    tt <- ncbi_children(id = '4751')
    tt2 <- ncbi_children(id = '4751', out_type = 'uid')
  })

  expect_is(tt, "list")
  expect_is(tt[[1]], "data.frame")
  expect_equal(ncol(tt[[1]]), 3)
  expect_is(tt2, "list")
  expect_is(tt2[[1]], "character")
  expect_error(ncbi_children(name = 'Ilex', id = '4751'))
  expect_equal(ncbi_children(name = NA)[[1]], NA)

  expect_equal(
    unname(sapply(ncbi_children('dragon', db='ncbi')[[1]], class)),
    c('character', 'character', 'character')
  )
})

test_that("ncbi_children does remove some ambiguous taxa", {
  skip_on_cran()
  vcr::use_cassette("ncbi_children_ambiguous", {
    # 28901 = "Salmonella enterica" - DOES NOT remove "subsp."
    subsp <- ncbi_children(id = '28901')
    # 2508041 = "unclassified Helianthus" - DOES remove "sp."
    sp <- ncbi_children(id = '2508041')
  })

  expect_is(subsp, "list")
  expect_is(subsp[[1]], "data.frame")
  expect_is(sp, "list")
  expect_is(sp[[1]], "data.frame")
})

test_that("ncbi_children accepts numeric and character class ids", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("ncbi_children_id_class_check", {
    tt <- ncbi_children(id = 4751)
    tt2 <- ncbi_children(id = '4751')
  })

  expect_identical(tt, tt2)
})

#' Retrieve synonyms from various sources given input taxonomic
#' names or identifiers
#'
#' @param sci_id Vector of taxa names (character) or IDs (character or numeric)
#' @param db character; database to query. either `itis`, `tropicos`,
#' `nbn`, `worms`, or `pow`. Note that each taxonomic data source has their own
#' identifiers, so that if you provide the wrong `db` value for the identifier
#' you could get a result, but it will likely be wrong (not what you were
#' expecting). If using tropicos, we  recommend getting an API key;
#' see [taxize-authentication]
#' @param id character; identifiers, returned by [get_tsn()], [get_tpsid()],
#' [get_nbnid()], [get_wormsid()], [get_pow()]
#' @param rows (numeric) Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this parameter is ignored if you pass in a
#' taxonomic id of any of the acceptable classes: tsn, tpsid, nbnid, ids.
#' @param x For `synonyms()`: deprecated, see `sci_id`. For `synonyms_df()`, 
#' the output of `synonyms()`
#' @param ... Other passed arguments to internal functions `get_*()` and
#' functions to gather synonyms.
#'
#' @return A named list of results with three types of output in each slot:
#'
#' - if the name was not found: `NA_character_`
#' - if the name was found but no synonyms found, an empty data.frame (0 rows)
#' - if the name was found, and synonyms found, a data.frames with the
#' synonyms - the column names vary by data source
#'
#' @details If IDs are supplied directly (not from the `get_*()` functions)
#' you must specify the type of ID.
#'
#' For `db = "itis"` you can pass in a parameter `accepted` to
#' toggle whether only accepted names are used `accepted = TRUE`, or if
#' all are used `accepted = FALSE`. The default is `accepted = FALSE`
#'
#' Note that IUCN requires an API key. See [rredlist::rredlist-package]
#' for help on authentiating with IUCN Redlist
#'
#' @seealso [get_tsn()] [get_tpsid()] [get_nbnid()]
#' [get_wormsid()] [get_iucn()] [get_pow()]
#'
#' @export
#' @examples \dontrun{
#' # Plug in taxon IDs
#' synonyms(183327, db="itis")
#' synonyms("25509881", db="tropicos")
#' synonyms("NBNSYS0000004629", db='nbn')
#' synonyms(105706, db='worms')
#' synonyms(12392, db='iucn')
#' synonyms('urn:lsid:ipni.org:names:358881-1', db='pow')
#'
#' # Plug in taxon names directly
#' synonyms("Pinus contorta", db="itis")
#' synonyms("Puma concolor", db="itis")
#' synonyms(c("Poa annua",'Pinus contorta','Puma concolor'), db="itis")
#' synonyms("Poa annua", db="tropicos")
#' synonyms("Pinus contorta", db="tropicos")
#' synonyms(c("Poa annua",'Pinus contorta'), db="tropicos")
#' synonyms("Pinus sylvestris", db='nbn')
#' synonyms('Pomatomus', db='worms')
#' synonyms('Pomatomus saltatrix', db='worms')
#' synonyms('Lithocarpus mindanaensis', db='pow')
#' synonyms('Poa annua', db='pow')
#' synonyms(c('Poa annua', 'Pinus contorta', 'foo bar'), db='pow')
#'
#' # not accepted names, with ITIS
#' ## looks for whether the name given is an accepted name,
#' ## and if not, uses the accepted name to look for synonyms
#' synonyms("Acer drummondii", db="itis")
#' synonyms("Spinus pinus", db="itis")
#'
#' # Use get_* methods
#' synonyms(get_tsn("Poa annua"))
#' synonyms(get_tpsid("Poa annua"))
#' synonyms(get_nbnid("Carcharodon carcharias"))
#' synonyms(get_iucn('Loxodonta africana'))
#' synonyms(get_pow('Lithocarpus mindanaensis'))
#'
#' # Pass many ids from class "ids"
#' out <- get_ids(names="Poa annua", db = c('itis','tropicos'))
#' synonyms(out)
#'
#' # Use the rows parameter to select certain rows
#' synonyms("Poa annua", db='tropicos', rows=1)
#' synonyms("Poa annua", db='tropicos', rows=1:3)
#' synonyms("Pinus sylvestris", db='nbn', rows=1:3)
#'
#' # Use curl options
#' synonyms("Poa annua", db='tropicos', rows=1, verbose = TRUE)
#' synonyms("Poa annua", db='itis', rows=1, verbose = TRUE)
#'
#'
#' # combine many outputs together
#' x <- synonyms(c("Osmia bicornis", "Osmia rufa", "Osmia"), db = "itis")
#' synonyms_df(x)
#'
#' ## note here how Pinus contorta is dropped due to no synonyms found
#' synonyms_df(x)
#'
#' ## note here that ids are taxon identifiers b/c you start with them
#' x <- synonyms(c(25509881, 13100094), db="tropicos")
#' synonyms_df(x)
#'
#' ## NBN
#' x <- synonyms(c('Aglais io', 'Usnea hirta', 'Arctostaphylos uva-ursi'),
#'   db="nbn")
#' synonyms_df(x)
#' }

synonyms <- function(...) {
  UseMethod("synonyms")
}

#' @export
#' @rdname synonyms
synonyms.default <- function(sci_id, db = NULL, rows = NA, x = NULL, ...) {
  nstop(db)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "synonyms(x)", with = "synonyms(sci_id)")
    sci_id <- x
  }
  if (!is.null(x)) sci_id <- x
  switch(
    db,
    itis = {
      id <- process_syn_ids(sci_id, db, get_tsn, rows = rows, ...)
      structure(stats::setNames(synonyms(id, ...), sci_id),
                class = "synonyms", db = "itis")
    },
    tropicos = {
      id <- process_syn_ids(sci_id, db, get_tpsid, rows = rows, ...)
      structure(stats::setNames(synonyms(id, ...), sci_id),
                class = "synonyms", db = "tropicos")
    },
    nbn = {
      id <- process_syn_ids(sci_id, db, get_nbnid, rows = rows, ...)
      structure(stats::setNames(synonyms(id, ...), sci_id),
                class = "synonyms", db = "nbn")
    },
    worms = {
      id <- process_syn_ids(sci_id, db, get_wormsid, rows = rows, ...)
      structure(stats::setNames(synonyms(id, ...), sci_id),
                class = "synonyms", db = "worms")
    },
    iucn = {
      id <- process_syn_ids(sci_id, db, get_iucn, ...)
      structure(stats::setNames(synonyms(id, ...), sci_id),
                class = "synonyms", db = "iucn")
    },
    pow = {
      id <- process_syn_ids(sci_id, db, get_pow, ...)
      structure(stats::setNames(synonyms(id, ...), sci_id),
                class = "synonyms", db = "pow")
    },
    stop("the provided db value was not recognised", call. = FALSE)
  )
}

process_syn_ids <- function(input, db, fxn, ...){
  g <- tryCatch(as.numeric(as.character(input)), warning = function(e) e)
  if (inherits(g, "condition") && all(!grepl("ipni\\.org", input))) {
    return(eval(fxn)(input, ...))
  }
  if (
    is.numeric(g) ||
    is.character(input) && all(grepl("N[HB]", input)) ||
    is.character(input) && all(grepl("ipni\\.org", input)) ||
    is.character(input) && all(grepl("[[:digit:]]", input))
  ) {
    as_fxn <- switch(db,
      itis = as.tsn,
      tropicos = as.tpsid,
      nbn = as.nbnid,
      worms = as.wormsid,
      iucn = as.iucn,
      pow = as.pow)
    if (db == "iucn") return(as_fxn(input, check = TRUE))
    return(as_fxn(input, check = FALSE))
  } else {
    eval(fxn)(input, ...)
  }
}

#' @export
#' @rdname synonyms
synonyms.tsn <- function(id, ...) {
  warn_db(list(...), "itis")
  fun <- function(x){
    if (is.na(x)) { NA_character_ } else {
      is_acc <- rit_acc_name(x, ...)
      if (all(!is.na(is_acc$acceptedName))) {
        accdf <- stats::setNames(
          data.frame(x[1], is_acc, stringsAsFactors = FALSE),
          c("sub_tsn", "acc_name", "acc_tsn", "acc_author")
        )
        x <- is_acc$acceptedTsn
        message("Accepted name(s) is/are '",
                paste0(is_acc$acceptedName, collapse = "/"), "'")
        message("Using tsn(s) ", paste0(is_acc$acceptedTsn, collapse = "/"),
                "\n")
      } else {
        accdf <- data.frame(sub_tsn = x[1], acc_tsn = x[1],
                            stringsAsFactors = FALSE)
      }

      res <- Map(function(z, w) {
        tmp <- ritis::synonym_names(z)
        if (NROW(tmp) == 0) {
          tibble::tibble()
        } else {
          tmp <- stats::setNames(tmp, c('syn_author', 'syn_name', 'syn_tsn'))
          cbind(w, tmp, row.names = NULL)
        }
      }, x, split(accdf, seq_len(NROW(accdf))))
      do.call("rbind", unname(res))
    }
  }
  stats::setNames(lapply(id, fun), id)
}

rit_acc_name <- function(x, ...) {
  tmp <- ritis::accepted_names(x, ...)
  if (NROW(tmp) == 0) {
    data.frame(submittedtsn = x[1], acceptedName = NA, acceptedTsn = x[1],
               stringsAsFactors = FALSE)
  } else {
    tmp
  }
}

#' @export
#' @rdname synonyms
synonyms.tpsid <- function(id, ...) {
  warn_db(list(...), "topicos")
  fun <- function(x) {
    if (is.na(x)) {
      NA_character_
    } else {
      res <- tp_synonyms(x, ...)$synonyms
      if (grepl("no syns found", res[1,1])) tibble::tibble() else res
    }
  }
  stats::setNames(lapply(id, fun), id)
}

#' @export
#' @rdname synonyms
synonyms.nbnid <- function(id, ...) {
  warn_db(list(...), "nbn")
  fun <- function(x){
    if (is.na(x)) {
      NA_character_
    } else {
      res <- nbn_synonyms(x, ...)
      if (length(res) == 0) tibble::tibble() else res
    }
  }
  stats::setNames(lapply(id, fun), id)
}

#' @export
#' @rdname synonyms
synonyms.wormsid <- function(id, ...) {
  warn_db(list(...), "worms")
  fun <- function(x) {
    if (is.na(x)) {
      NA_character_
    } else {
      res <- tryCatch(worrms::wm_synonyms(as.numeric(x), ...),
        error = function(e) e)
      if (inherits(res, "error")) tibble::tibble() else res
    }
  }
  stats::setNames(lapply(id, fun), id)
}

#' @export
#' @rdname synonyms
synonyms.iucn <- function(id, ...) {
  warn_db(list(...), "iucn")
  out <- vector(mode = "list", length = length(id))
  for (i in seq_along(id)) {
    if (is.na(id[[i]])) {
      out[[i]] <- NA_character_
    } else {
      res <- get_iucn_data(attr(id, "name")[i])
      out[[i]] <- if (length(res) == 0 || is.na(res)) tibble::tibble() else res[[1]]$taxon$synonyms
    }
  }
  stats::setNames(out, id)
}

#' @export
#' @rdname synonyms
synonyms.pow <- function(id, ...) {
  warn_db(list(...), "pow")
  out <- vector(mode = "list", length = length(id))
  for (i in seq_along(id)) {
    if (is.na(id[[i]])) {
      out[[i]] <- NA_character_
    } else {
      res <- pow_synonyms(id[i], ...)
      out[[i]] <- if (length(res) == 0) {
        tibble::tibble() 
      } else {
        names(res)[1] <- "id"
        res
      }
    }
  }
  stats::setNames(out, id)
}




#' @export
#' @rdname synonyms
synonyms.ids <- function(id, ...) {
  fun <- function(x){
    if (is.na(x)) {
      out <- NA_character_
    } else {
      out <- synonyms(x, ...)
    }
    return( out )
  }
  lapply(id, fun)
}

### Combine synonyms output into single data.frame -----------
#' @export
#' @rdname synonyms
synonyms_df <- function(x) {
  UseMethod("synonyms_df")
}

#' @export
synonyms_df.default <- function(x) {
  stop("no 'synonyms_df' method for ", class(x), call. = FALSE)
}

#' @export
synonyms_df.synonyms <- function(x) {
  x <- Filter(function(z) inherits(z, "data.frame"), x)
  x <- Filter(function(z) NROW(z) > 0, x)
  (data.table::setDF(
    data.table::rbindlist(x, use.names = TRUE, fill = TRUE, idcol = TRUE)
  ))
}

#' Get the page name for a Wiki taxon
#'
#' @export
#' @param sci_com (character) A vector of common or scientific names. Or, a
#' `taxon_state` object (see [taxon-state])
#' @param wiki_site (character) Wiki site. One of species (default), pedia,
#' commons
#' @param wiki (character) language. Default: en
#' @param ask logical; should get_wiki be run in interactive mode?
#' If `TRUE` and more than one wiki is found for the species, the user is
#' asked for input. If `FALSE` NA is returned for multiple matches.
#' @param messages logical; should progress be printed?
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this function still only gives back a wiki
#' class object with one to many identifiers. See [get_wiki_()] to get back
#' all, or a subset, of the raw data that you are presented during the ask
#' process.
#' @param limit (integer) number of records to return
#' @param x For `get_wiki()`: deprecated, see `sci_com`. For `as.wiki`, various,
#' see examples
#' @param ... Ignored
#' @param check logical; Check if ID matches any existing on the DB, only
#' used in [as.wiki()]
#' @template getreturn
#'
#' @details For `wiki_site = "pedia" `, we use the english language site by
#' default. Set the `wiki` parameter for a different language site.
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @examples \dontrun{
#' get_wiki(sci_com = "Quercus douglasii")
#' get_wiki(sci_com = "Quercu")
#' get_wiki(sci_com = "Quercu", "pedia")
#' get_wiki(sci_com = "Quercu", "commons")
#'
#' # diff. wikis with wikipedia
#' get_wiki("Malus domestica", "pedia")
#' get_wiki("Malus domestica", "pedia", "fr")
#'
#' # as coercion
#' as.wiki("Malus_domestica")
#' as.wiki("Malus_domestica", wiki_site = "commons")
#' as.wiki("Malus_domestica", wiki_site = "pedia")
#' as.wiki("Malus_domestica", wiki_site = "pedia", wiki = "fr")
#' as.wiki("Malus_domestica", wiki_site = "pedia", wiki = "da")
#' }

get_wiki <- function(sci_com, wiki_site = "species", wiki = "en", ask = TRUE,
                     messages = TRUE, limit = 100, rows = NA, x = NULL, ...) {

  assert(sci_com, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(wiki_site, "character")
  assert(wiki, "character")
  assert(messages, "logical")
  assert_rows(rows)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_wiki(x)", with = "get_wiki(sci_com)")
    sci_com <- x
  }
  

  if (inherits(sci_com, "character")) {
    tstate <- taxon_state$new(class = "wiki", names = sci_com)
    items <- sci_com
  } else {
    assert_state(sci_com, "wiki")
    tstate <- sci_com
    sci_com <- tstate$taxa_remaining()
    items <- c(sci_com, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci_com)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci_com[i], "'\n")
    df <- switch(
      wiki_site,
      species = wikitaxa::wt_wikispecies_search(query = sci_com[i], limit = limit, ...),
      pedia = wikitaxa::wt_wikipedia_search(query = sci_com[i], wiki = wiki,
                                            limit = limit, ...),
      commons = wikitaxa::wt_wikicommons_search(query = sci_com[i], limit = limit, ...)
    )$query$search
    mm <- NROW(df) > 1

    if (!inherits(df, "tbl_df") || NROW(df) == 0) {
      id <- NA_character_
      att <- "not found"
    } else {
      df <- df[, c("title", "size", "wordcount")]
      df <- sub_rows(df, rows)

      # should return NA if spec not found
      if (NROW(df) == 0) {
        mssg(messages, tx_msg_not_found)
        id <- NA_character_
        att <- "not found"
      }

      df$title <- gsub("\\s", "_", df$title)

      # take the one wiki from data.frame
      if (NROW(df) == 1) {
        id <- df$title
        att <- "found"
      }

      # check for direct match
      if (NROW(df) > 1) {
        df <- data.frame(df, stringsAsFactors = FALSE)
        matchtmp <- df[tolower(df$title) %in% tolower(sci_com[i]), "title"]
        if (length(matchtmp) == 1) {
          id <- matchtmp
          direct <- TRUE
          att <- "found"
        } else {
          direct <- FALSE
          id <- NA_character_
          att <- m_na_ask_false_no_direct
          warning("> 1 result; no direct match found", call. = FALSE)
        }
      }

      # multiple matches
      if (any(
        NROW(df) > 1 && is.na(id) |
        NROW(df) > 1 && att == "found" && length(id) > 1
      )) {
        if (ask) {
          # user prompt
          df <- df[order(df$title), ]
          rownames(df) <- NULL

          # prompt
          message("\n\n")
          message(paste0(utils::capture.output(df), collapse = "\n"))
          message("\nMore than one wiki ID found for taxon '", sci_com[i], "'!\n
                  Enter rownumber of taxon (other inputs will return 'NA'):\n")
          take <- scan(n = 1, quiet = TRUE, what = 'raw')

          if (length(take) == 0) {
            take <- 'notake'
            att <- 'nothing chosen'
          }
          if (take %in% seq_len(nrow(df))) {
            take <- as.numeric(take)
            message("Input accepted, took taxon '",
                    as.character(df$title[take]), "'.\n")
            id <-  df$title[take]
            att <- "found"
          } else {
            id <- NA_character_
            mssg(messages, "\nReturned 'NA'!\n\n")
            att <- "not found"
          }
        } else {
          if (length(id) != 1) {
            warning(sprintf(m_more_than_one_found, "Wiki ID", x),
              call. = FALSE)
            id <- NA_character_
            att <- m_na_ask_false
          }
        }
      }

    }

    res <- list(id = as.character(id), att = att, multiple = mm,
      direct = direct)
    prog$completed(sci_com[i], att)
    prog$prog(att)
    tstate$add(sci_com[i], res)
  }
  out <- tstate$get()
  ids <- structure(pluck_un(out, "id", ""), class = "wiki",
    match = pluck_un(out, "att", ""),
    multiple_matches = pluck_un(out, "multiple", logical(1)),
    pattern_match = pluck_un(out, "direct", logical(1))
  )
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  attr(ids, 'wiki_site') <- wiki_site
  attr(ids, 'wiki_lang') <- wiki
  if ( !all(is.na(ids)) ) {
    zz <- gsub("\\s", "_", na.omit(ids))
    base_url <- switch(
      wiki_site,
      species = 'https://species.wikimedia.org/wiki/',
      pedia = sprintf('https://%s.wikipedia.org/wiki/', wiki),
      commons = 'https://commons.wikimedia.org/wiki/'
    )
    attr(ids, 'uri') <- paste0(base_url, zz)
  }
  return(ids)
}

#' @export
#' @rdname get_wiki
as.wiki <- function(x, check=TRUE, wiki_site = "species", wiki = "en") {
  UseMethod("as.wiki")
}

#' @export
#' @rdname get_wiki
as.wiki.wiki <- function(x, check=TRUE, wiki_site = "species",
                         wiki = "en") x

#' @export
#' @rdname get_wiki
as.wiki.character <- function(x, check=TRUE, wiki_site = "species",
                              wiki = "en") {
  if (length(x) == 1) {
    make_wiki(x, check, wiki_site, wiki)
  } else {
    collapse(x, make_wiki, "wiki", check = check)
  }
}

#' @export
#' @rdname get_wiki
as.wiki.list <- function(x, check=TRUE, wiki_site = "species",
                         wiki = "en") {
  if (length(x) == 1) {
    make_wiki(x, check)
  } else {
    collapse(x, make_wiki, "wiki", check = check)
  }
}

#' @export
#' @rdname get_wiki
as.wiki.numeric <- function(x, check=TRUE, wiki_site = "species",
                            wiki = "en") {
  as.wiki(as.character(x), check)
}

#' @export
#' @rdname get_wiki
as.wiki.data.frame <- function(x, check=TRUE, wiki_site = "species",
                               wiki = "en") {

  structure(x$ids, class = "wiki", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match,
            wiki_site = x$wiki_site,
            wiki_lang = x$wiki_lang, uri = x$uri)
}

#' @export
#' @rdname get_wiki
as.data.frame.wiki <- function(x, ...){
  data.frame(ids = unclass(x),
             class = "wiki",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             wiki_site = attr(x, 'wiki_site'),
             wiki_lang = attr(x, 'wiki_lang'),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_wiki <- function(x, check = TRUE, wiki_site, wiki) {
  url <- switch(
    wiki_site,
    species = 'https://species.wikimedia.org/wiki/%s',
    pedia = paste0(sprintf('https://%s.wikipedia.org/wiki', wiki), "/%s"),
    commons = 'https://commons.wikimedia.org/wiki/%s'
  )
  make_wiki_generic(x, url, "wiki", check)
}

check_wiki <- function(x) {
  tt <- wikitaxa::wt_wiki_page(x)
  identical(tt$status_code, 200)
}

#' @export
#' @rdname get_wiki
get_wiki_ <- function(x, messages = TRUE, wiki_site = "species",
                      wiki = "en", limit = 100, rows = NA, ...) {
  stats::setNames(
    lapply(x, get_wiki_help, messages = messages, wiki_site = wiki_site,
           wiki = wiki, limit = limit, rows = rows, ...),
    x
  )
}

get_wiki_help <- function(x, messages, wiki_site = "species", wiki = "en",
                          limit = 100, rows, ...) {

  mssg(messages, "\nRetrieving data for taxon '", x, "'\n")
  assert(x, "character")
  assert(wiki_site, "character")
  assert(wiki, "character")

  df <- switch(
    wiki_site,
    species = wikitaxa::wt_wikispecies_search(query = x, limit = limit, ...),
    pedia = wikitaxa::wt_wikipedia_search(query = x, wiki = wiki,
                                          limit = limit, ...),
    commons = wikitaxa::wt_wikicommons_search(query = x, limit = limit, ...)
  )$query$search

  if (!inherits(df, "tbl_df") || NROW(df) == 0) {
    NULL
  } else {
    df <- df[, c("title", "size", "wordcount")]
    sub_rows(df, rows)
  }
}

#' iPlant name resolution
#'
#' @export
#' @param sci Vector of one or more taxonomic names (no common names)
#' @param retrieve Specifies whether to retrieve all matches for the
#' names submitted. One of 'best' (retrieves only the single best match
#' for each name submitted) or 'all' (retrieves all matches)
#' @param query Deprecated, see `sci`
#' @param ... Curl options passed on to [crul::verb-GET]
#' @return A data.frame
#' @examples \dontrun{
#' iplant_resolve(sci=c("Helianthus annuus", "Homo sapiens"))
#' iplant_resolve("Helianthusss")
#' iplant_resolve("Pooa")
#' iplant_resolve("Helianthusss", verbose = TRUE)
#' }
iplant_resolve <- function(sci, retrieve='all', query = NULL, ...) {
  if (!is.null(query)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "iplant_resolve(query)", with = "iplant_resolve(sci)")
    sci <- query
  }
  
  url <- "http://tnrs.iplantc.org/tnrsm-svc/matchNames"
  sci <- paste(sci, collapse = ",")
  args <- tc(list(names = sci, retrieve = retrieve))
  cli <- crul::HttpClient$new(url, headers = tx_ual, opts = list(...))
  out <- cli$get(query = argsnull(args))
  out$raise_for_status()
  res <- jsonlite::fromJSON(out$parse("UTF-8"), FALSE)$items
  df <- do.call(rbind, lapply(res, data.frame, stringsAsFactors = FALSE))
  if (!is.null(df)) nmslwr(df) else df
}

#' Get the OTT id for a search term
#'
#' Retrieve the Open Tree of Life Taxonomy (OTT) id of a taxon from
#' OpenTreeOfLife
#'
#' @export
#' @param sci character; one or more scientific names. Or, a `taxon_state`
#' object (see [taxon-state])
#' @param ask logical; should `get_tolid` be run in interactive mode?
#' If `TRUE` and more than one TOL is found for the species, the user is
#' asked for input. If `FALSE` NA is returned for multiple matches.
#' @param messages logical; should progress be printed?
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this function still only gives back a tol
#' class object with one to many identifiers. See [get_tolid_()]
#' to get back all, or a subset, of the raw data that you are presented during
#' the ask process.
#' @param x Input to `as.tolid`
#' @param sciname Deprecated, see `sci`
#' @param ... Ignored
#' @param check logical; Check if ID matches any existing on the DB, only
#' used in [as.tolid()]
#' @template getreturn
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @examples \dontrun{
#' get_tolid(sci = "Quercus douglasii")
#' get_tolid(sci = "Chironomus riparius")
#' get_tolid(c("Chironomus riparius","Quercus douglasii"))
#' splist <- c("annona cherimola", 'annona muricata', "quercus robur",
#' 		"shorea robusta", "pandanus patina", "oryza sativa", "durio zibethinus")
#' get_tolid(splist, messages=FALSE)
#'
#' # specify rows to limit choices available
#' get_tolid('Arni')
#' get_tolid('Arni', rows=1)
#' get_tolid('Arni', rows=1:2)
#'
#' # When not found
#' get_tolid("howdy")
#' get_tolid(c("Chironomus riparius", "howdy"))
#'
#' # Convert a tol without class information to a tol class
#' as.tolid(get_tolid("Quercus douglasii")) # already a tol, returns the same
#' as.tolid(get_tolid(c("Chironomus riparius","Pinus contorta"))) # same
#' as.tolid(5907893) # numeric
#' as.tolid(c(3930798,515712,872577)) # numeric vector, length > 1
#' as.tolid("3930798") # character
#' as.tolid(c("3930798","515712","872577")) # character vector, length > 1
#' as.tolid(list("3930798","515712","872577")) # list, either numeric or character
#' ## dont check, much faster
#' as.tolid("3930798", check=FALSE)
#' as.tolid(3930798, check=FALSE)
#' as.tolid(c("3930798","515712","872577"), check=FALSE)
#' as.tolid(list("3930798","515712","872577"), check=FALSE)
#'
#' (out <- as.tolid(c(3930798,515712,872577)))
#' data.frame(out)
#' as.tolid( data.frame(out) )
#'
#' # Get all data back
#' get_tolid_("Arni")
#' get_tolid_("Arni", rows=1)
#' get_tolid_("Arni", rows=1:2)
#' get_tolid_(c("asdfadfasd","Pinus contorta"))
#' }

get_tolid <- function(sci, ask = TRUE, messages = TRUE, rows = NA,
  sciname = NULL, ...) {

  assert(sci, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(messages, "logical")
  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_tolid(sciname)", with = "get_tolid(sci)")
    sci <- sciname
  }
  
  if (!all(is.na(rows))) {
    assert(rows, c("numeric", "integer"))
    stopifnot(rows > 0)
  }

  if (inherits(sci, "character")) {
    tstate <- taxon_state$new(class = "tolid", names = sci)
    items <- sci
  } else {
    assert_state(sci, "tolid")
    tstate <- sci
    sci <- tstate$taxa_remaining()
    items <- c(sci, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci)) {
    mssg(messages, "\nRetrieving data for taxon '", sci[i], "'\n")

    tol_df <- tryCatch(tol_resolve(sci[i], ...), error = function(e) e)

    if (
      !inherits(tol_df, "data.frame") ||
      NROW(tol_df) == 0 ||
      inherits(tol_df, "error")
    ) {
      id <- NA_character_
      att <- "not found"
    } else {
      tol_df <- tol_fetch_fuzzy(tol_df)
      tol_df <- sub_rows(tol_df, rows)

      # should return NA if spec not found
      if (NROW(tol_df) == 0) {
        mssg(messages, m_not_found_sp_altclass)
        id <- NA_character_
        att <- 'not found'
      }

      # take the one tol from data.frame
      if (NROW(tol_df) == 1) {
        id <- tol_df$ott_id
        att <- 'found'
      }

      # more than one found -> user input
      if (NROW(tol_df) > 1) {
        # check for exact match
        matchtmp <- tol_df[tolower(tol_df$unique_name) %in% tolower(sci[i]), "ott_id"]
        if (length(matchtmp) == 1) {
          id <- as.character(matchtmp)
          direct <- TRUE
          att <- "found"
        } else {
          tol_df <- sub_rows(tol_df, rows)
          if (NROW(tol_df) == 0) {
            id <- NA_character_
            att <- "not found"
          } else {
            id <- tol_df$ott_id
            if (length(id) == 1) {
              rank_taken <- tol_df$rank
              att <- "found"
            }
          }

          if (length(id) > 1) {
            if (ask) {
              # prompt
              message("\n\n")
              message("\nMore than one ToL ID found for taxon '", sci[i], "'!\n
                      Enter rownumber of taxon (other inputs will return 'NA'):\n")
              rownames(tol_df) <- 1:nrow(tol_df)
              message(paste0(utils::capture.output(tol_df), collapse = "\n"))
              take <- scan(n = 1, quiet = TRUE, what = 'raw')

              if (length(take) == 0) {
                take <- 'notake'
                att <- 'nothing chosen'
              }
              if (take %in% seq_len(nrow(tol_df))) {
                take <- as.numeric(take)
                message("Input accepted, took tol ID '",
                  as.character(tol_df$ott_id[take]), "'.\n")
                id <- as.character(tol_df$ott_id[take])
                att <- "found"
              } else {
                id <- NA_character_
                att <- "not found"
                mssg(messages, "\nReturned 'NA'!\n\n")
              }
            }
            else {
              if (length(id) != 1) {
                warning(sprintf(m_more_than_one_found, "ToL ID", sci[i]),
                  call. = FALSE)
                id <- NA_character_
                att <- m_na_ask_false
              }
            }
          }
        }
      }
    }
    res <- list(id = as.character(id), att = att, multiple = FALSE,
      direct = FALSE)
    prog$completed(sci[i], att)
    prog$prog(att)
    tstate$add(sci[i], res)
  }
  out <- tstate$get()
  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "tolid",
                   match = pluck_un(out, "att", ""),
                   multiple_matches = pluck_un(out, "multiple", logical(1)),
                   pattern_match = pluck_un(out, "direct", logical(1)))
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$tol)
}

#' @export
#' @rdname get_tolid
as.tolid <- function(x, check=TRUE) UseMethod("as.tolid")

#' @export
#' @rdname get_tolid
as.tolid.tolid <- function(x, check=TRUE) x

#' @export
#' @rdname get_tolid
as.tolid.character <- function(x, check=TRUE) {
  if (length(x) == 1) make_tol(x, check) else collapse(x, make_tol, "tolid", check = check)
}

#' @export
#' @rdname get_tolid
as.tolid.list <- function(x, check=TRUE) {
  if (length(x) == 1) make_tol(x, check) else collapse(x, make_tol, "tolid", check = check)
}

#' @export
#' @rdname get_tolid
as.tolid.numeric <- function(x, check=TRUE) as.tolid(as.character(x), check)

#' @export
#' @rdname get_tolid
as.tolid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class = "tolid", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri = x$uri)
}

#' @export
#' @rdname get_tolid
as.data.frame.tolid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "tolid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_tol <- function(x, check=TRUE) {
  make_generic(x, 'https://tree.opentreeoflife.org/opentree/argus/ottol@%s', "tolid", check)
}

check_tolid <- function(x){
  tt <- tryCatch(rotl::taxonomy_taxon_info(as.numeric(x)), error = function(e) e)
  !inherits(tt, "error")
}

#' @export
#' @rdname get_tolid
get_tolid_ <- function(sci, messages = TRUE, rows = NA, sciname = NULL) {
  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_tolid_(sciname)", with = "get_tolid_(sci)")
    sci <- sciname
  }
  stats::setNames(
    lapply(sci, get_tolid_help, messages = messages, rows = rows),
    sci
  )
}

get_tolid_help <- function(sci, messages, rows, ...){
  mssg(messages, "\nRetrieving data for taxon '", sci, "'\n")
  tol_df <- tryCatch(tol_resolve(sci, ...), error = function(e) e)
  if (!inherits(tol_df, "data.frame") || NROW(tol_df) == 0 || inherits(tol_df, "error")) {
    NULL
  } else {
    df <- tol_fetch_fuzzy(tol_df)
    sub_rows(df, rows)
  }
}

tol_fetch_fuzzy <- function(x) {
  atts <- attr(x, "original_response")
  df <- dtrbsetdf(lapply(atts$results, function(z) {
    dtrbsetdf(lapply(z$matches, function(w) {
      c(
        pop(w, "taxon"),
        lapply(w$taxon, function(m) {
          if (length(m) > 1) {
            paste0(m, collapse = ",")
          } else if (length(m) == 0) {
            ""
          } else if (length(m) == 1) {
            m[[1]]
          } else {
            m
          }
        })
      )
    }))
  }))
  if (NROW(df) == 0) return(df)
  df <- move_col_begin(df, "matched_name")
  df <- move_col_begin(df, "unique_name")
  df$tax_sources <- NULL
  df$synonyms <- NULL
  df$is_suppressed <- NULL
  df$search_string <- NULL
  df$name <- NULL
  df
}

dtrbsetdf <- function(x) {
  x <- lapply(x, function(a) {
    a[sapply(a, is.null)] <- NA_character_
    a
  })
  (out <- data.table::setDF(
    data.table::rbindlist(x, use.names = TRUE, fill = TRUE)
  ))
}

context("get_tolid")

test_that("get_tolid returns the correct value", {
  skip_on_cran()
  
  vcr::use_cassette("get_tolid", {
    x <- get_tolid("Quercus douglasii", messages = FALSE)
    z <- get_tolid(c("Chironomus", "Chaetopteryx"),
      messages = FALSE)
  })

  expect_is(x, "tolid")
  expect_is(x[1], "character")
  expect_match(attr(x, "uri"), "opentreeoflife")
  expect_equal(length(x), 1)

  expect_is(z, "tolid")
  expect_is(z[1], "character")
  expect_match(attr(z, "uri"), "opentreeoflife")
  expect_equal(length(z), 2)
})

test_that("get_tolid accepts ask-argument", {
  skip_on_cran()
  
  vcr::use_cassette("get_tolid_ask_arg", {
    x <- sw(get_tolid("Dugesia", ask = FALSE, messages = FALSE))
  })
  expect_true(is.na(x))
})

test_that("get_tolid fails as expected", {
  skip_on_cran()

  expect_error(get_tolid(), "argument \"sci\" is missing")
  expect_error(get_tolid("Satyrium", ask = 4, messages = FALSE),
               "ask must be of class logical")

  # rows param
  expect_error(get_tolid("Achlya", rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_tolid("Achlya", rows = 0, messages = FALSE),
               "rows > 0 is not TRUE")
})

#' EUBON capabilities
#'
#' @export
#' @param ... Curl options passed on to [crul::verb-GET]
#' @references https://cybertaxonomy.eu/eu-bon/utis/1.3/doc.html
#' @family eubon-methods
#' @examples \dontrun{
#' eubon_capabilities()
#' }
eubon_capabilities <- function(...) {
  res <- crul::HttpClient$new(file.path(eubon_base(), "capabilities"),
    headers = tx_ual, opts = list(...))$get()
  res$raise_for_status()
  tibble::as_tibble(
    jsonlite::fromJSON(res$parse("UTF-8"), TRUE, flatten = TRUE)
  )
}

#' Get the Catalogue of Life ID from taxonomic names.
#'
#' @export
#' @param sciname character; scientific name.
#' @param ask logical; should get_colid be run in interactive mode?
#' If TRUE and more than one ID is found for the species, the user is asked for
#' input. If FALSE NA is returned for multiple matches.
#' @param verbose logical; If TRUE the actual taxon queried is printed on the
#'    console.
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all rows are considered.
#' Note that this function still only gives back a colid class object with one to many identifiers.
#' See \code{\link[taxize]{get_colid_}} to get back all, or a subset, of the raw data that you are
#' presented during the ask process.
#' @param kingdom (character) A kingdom name. Optional. See \code{Filtering}
#' below.
#' @param phylum (character) A phylum (aka division) name. Optional. See \code{Filtering}
#' below.
#' @param class (character) A class name. Optional. See \code{Filtering} below.
#' @param order (character) An order name. Optional. See \code{Filtering} below.
#' @param family (character) A family name. Optional. See \code{Filtering} below.
#' @param rank (character) A taxonomic rank name. See \code{\link{rank_ref}} for possible
#' options. Though note that some data sources use atypical ranks, so inspect the
#' data itself for options. Optional. See \code{Filtering} below.
#' @param x Input to as.colid
#' @param ... Ignored
#' @param check logical; Check if ID matches any existing on the DB, only used in
#' \code{\link{as.colid}}
#' @template getreturn
#'
#' @section Filtering:
#' The parameters \code{kingdom}, \code{phylum}, \code{class}, \code{order}, \code{family},
#' and \code{rank} are not used in the search to the data provider, but are used in filtering
#' the data down to a subset that is closer to the target you want. For all these parameters,
#' you can use regex strings since we use \code{\link{grep}} internally to match.
#' Filtering narrows down to the set that matches your query, and removes the rest.
#'
#' @family taxonomic-ids
#' @seealso \code{\link[taxize]{classification}}
#'
#' @author Scott Chamberlain, \email{myrmecocystus@@gmail.com}
#'
#' @examples \dontrun{
#' get_colid(sciname='Poa annua')
#' get_colid(sciname='Pinus contorta')
#' get_colid(sciname='Puma concolor')
#' # get_colid(sciname="Abudefduf saxatilis")
#'
#' get_colid(c("Poa annua", "Pinus contorta"))
#'
#' # specify rows to limit choices available
#' get_colid(sciname='Poa annua')
#' get_colid(sciname='Poa annua', rows=1)
#' get_colid(sciname='Poa annua', rows=2)
#' get_colid(sciname='Poa annua', rows=1:2)
#'
#' # When not found
#' get_colid(sciname="uaudnadndj")
#' get_colid(c("Chironomus riparius", "uaudnadndj"))
#'
#' # Narrow down results to a division or rank, or both
#' ## Satyrium example
#' ### Results w/o narrowing
#' get_colid("Satyrium")
#' ### w/ division
#' get_colid("Satyrium", kingdom = "Plantae")
#' get_colid("Satyrium", kingdom = "Animalia")
#'
#' ## Rank example
#' get_colid("Poa")
#' get_colid("Poa", kingdom = "Plantae")
#' get_colid("Poa", kingdom = "Animalia")
#'
#' # Fuzzy filter on any filtering fields
#' ## uses grep on the inside
#' get_colid("Satyrium", kingdom = "p")
#'
#' # Convert a uid without class information to a uid class
#' as.colid(get_colid("Chironomus riparius")) # already a uid, returns the same
#' as.colid(get_colid(c("Chironomus riparius","Pinus contorta"))) # same
#' as.colid("714831352ad94741e4321eccdeb29f58") # character
#' # character vector, length > 1
#' as.colid(c("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d"))
#' # list, either numeric or character
#' as.colid(list("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d"))
#' ## dont check, much faster
#' as.colid("714831352ad94741e4321eccdeb29f58", check=FALSE)
#' as.colid(c("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d"),
#'  check=FALSE)
#' as.colid(list("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d"),
#'  check=FALSE)
#'
#' (out <- as.colid(c("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d")))
#' data.frame(out)
#' as.colid( data.frame(out) )
#'
#' # Get all data back
#' get_colid_("Poa annua")
#' get_colid_("Poa annua", rows=2)
#' get_colid_("Poa annua", rows=1:2)
#' get_colid_(c("asdfadfasd","Pinus contorta"))
#'
#' get_colid(sciname="Andropadus nigriceps fusciceps", rows=1)
#'
#' # use curl options
#' library("httr")
#' get_colid("Quercus douglasii", config=verbose())
#' bb <- get_colid("Quercus douglasii", config=progress())
#' }

get_colid <- function(sciname, ask = TRUE, verbose = TRUE, rows = NA,
                      kingdom = NULL, phylum = NULL, class = NULL, order = NULL,
                      family = NULL, rank = NULL, ...){

  assert(ask, "logical")
  assert(verbose, "logical")
  assert(kingdom, "character")
  assert(phylum, "character")
  assert(class, "character")
  assert(order, "character")
  assert(family, "character")
  assert(rank, "character")

  fun <- function(sciname, ask, verbose, rows, ...) {
    direct <- FALSE
    mssg(verbose, "\nRetrieving data for taxon '", sciname, "'\n")
    df <- col_search(name = sciname, response = "full", ...)[[1]]
    df <- df[, names(df) %in% c("name","rank","id","name_status","kingdom","family","acc_name")]
    mm <- NROW(df) > 1

    rank_taken <- NA
    if (NROW(df) == 0) {
      id <- NA_character_
      att <- "not found"
    } else {
      df <- stats::setNames(df, tolower(names(df)))
      df <- rename(df, c("id" = "colid"))
      colnames(df)[which(colnames(df) == 'id')] <- 'colid'
      id <- df$colid
      rank_taken <- as.character(df$rank)
      att <- "found"
    }

    # not found on col
    if (all(is.na(id))) {
      mssg(verbose, "Not found. Consider checking the spelling or alternate classification")
      id <- NA_character_
      att <- "not found"
    }
    # more than one found -> user input
    if (length(id) > 1) {
      rownames(df) <- 1:nrow(df)

      if (!is.null(kingdom) || !is.null(phylum) || !is.null(class) ||
          !is.null(order) || !is.null(family) || !is.null(rank)) {
        df <- filt(df, "kingdom", kingdom)
        df <- filt(df, "phylum", phylum)
        df <- filt(df, "class", class)
        df <- filt(df, "order", order)
        df <- filt(df, "family", family)
        df <- filt(df, "rank", rank)
      }

      df <- sub_rows(df, rows)
      id <- df$colid
      if (length(id) == 1) {
        rank_taken <- as.character(df$rank)
        direct <- TRUE
        att <- "found"
      }

      if (length(id) > 1) {
        if (ask) {
          # prompt
          rownames(df) <- 1:nrow(df)
          message("\n\n")
          message("\nMore than one colid found for taxon '", sciname, "'!\n
            Enter rownumber of taxon (other inputs will return 'NA'):\n")
          print(df)
          take <- scan(n = 1, quiet = TRUE, what = 'raw')

          if (length(take) == 0) {
            take <- 'notake'
            att <- 'nothing chosen'
          }
          if (take %in% seq_len(nrow(df))) {
            take <- as.numeric(take)
            message("Input accepted, took colid '", as.character(df$colid[take]), "'.\n")
            id <- as.character(df$colid[take])
            rank_taken <- as.character(df$rank[take])
            att <- "found"
          } else {
            id <- NA_character_
            att <- "not found"
            mssg(verbose, "\nReturned 'NA'!\n\n")
          }
        } else {
          if (length(id) != 1) {
            warning(
              sprintf("More than one id found for taxon '%s'; refine query or set ask=TRUE",
                      sciname),
              call. = FALSE
            )
            id <- NA_character_
            att <- 'NA due to ask=FALSE & > 1 result'
          }
        }
      }
    }
    list(id = id, rank = rank_taken, att = att, multiple = mm, direct = direct)
  }
  sciname <- as.character(sciname)
  out <- lapply(sciname, fun, ask = ask, verbose = verbose, rows = rows, ...)
  ids <- pluck(out, "id", "")
  atts <- pluck(out, "att", "")
  ids <- structure(ids, class = "colid", match = atts,
                   multiple_matches = pluck(out, "multiple", logical(1)),
                   pattern_match = pluck(out, "direct", logical(1)))
  if ( !all(is.na(ids)) ) {
    urls <- sapply(out, function(z){
      if (!is.na(z[['id']])) {
        if (tolower(z['rank']) == "species") {
          sprintf('http://www.catalogueoflife.org/col/details/species/id/%s', z[['id']])
        } else {
          sprintf('http://www.catalogueoflife.org/col/browse/tree/id/%s', z[['id']])
        }
      } else {
        NA
      }
    })
    attr(ids, 'uri') <- unlist(urls)
  }
  return(ids)
}

#' @export
#' @rdname get_colid
as.colid <- function(x, check=TRUE) UseMethod("as.colid")

#' @export
#' @rdname get_colid
as.colid.colid <- function(x, check=TRUE) x

#' @export
#' @rdname get_colid
as.colid.character <- function(x, check=TRUE) if (length(x) == 1) make_colid(x, check) else collapse(x, make_colid, "colid", check = check)

#' @export
#' @rdname get_colid
as.colid.list <- function(x, check=TRUE) if (length(x) == 1) make_colid(x, check) else collapse(x, make_colid, "colid", check = check)

#' @export
#' @rdname get_colid
as.colid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class = "colid", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri = x$uri)
}

#' @export
#' @rdname get_colid
as.data.frame.colid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "colid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_colid <- function(x, check=TRUE) make_generic(x, 'http://www.catalogueoflife.org/col/details/species/id/%s', "colid", check)

check_colid <- function(x){
  url <- "http://www.catalogueoflife.org/col/details/species/id/"
  res <- GET(paste0(url, x))
  !grepl("Species not found", res)
}

#' @export
#' @rdname get_colid
get_colid_ <- function(sciname, verbose = TRUE, rows = NA){
  setNames(lapply(sciname, get_colid_help, verbose = verbose, rows = rows), sciname)
}

get_colid_help <- function(sciname, verbose, rows){
  mssg(verbose, "\nRetrieving data for taxon '", sciname, "'\n")
  df <- col_search(name = sciname)[[1]]
  if (NROW(df) == 0) NULL else sub_rows(df, rows)
}

# tests for get_colid fxn in taxize
context("get_colid")

test_that("get_colid returns the correct value", {
  skip_on_cran()

  expect_that(is.na(get_colid(sciname='adsf asdf asdf', verbose=FALSE)[[1]]),
              is_true())
})

test_that("get_colid returns the correct class", {
  skip_on_cran()

  expect_that(get_colid(c("Helianthus excubitor", "adsf asdf asdf"), verbose=FALSE),
              is_a("colid"))
})

test_that("get_colid accepts ask-argument", {
  skip_on_cran()

  expect_that(is.na(get_colid(sciname='adsf asdf asdf', ask=FALSE, verbose=FALSE)[[1]]),
              is_true())
})


test_that("get_colid fails as expected", {
  skip_on_cran()

  expect_error(get_colid(), "argument \"sciname\" is missing")
  expect_error(get_colid('Poa annua', ask = 4, verbose = FALSE),
               "ask must be of class logical")
  expect_error(
    get_colid("Satyrium", kingdom = 234, verbose = FALSE),
    "kingdom must be of class character")
  expect_error(
    get_colid("Satyrium", phylum = 234, verbose = FALSE),
    "phylum must be of class character")
  expect_error(
    get_colid("Satyrium", class = 234, verbose = FALSE),
    "class must be of class character")
  expect_error(
    get_colid("Satyrium", order = 234, verbose = FALSE),
    "order must be of class character")
  expect_error(
    get_colid("Satyrium", family = 234, verbose = FALSE),
    "family must be of class character")
  expect_error(
    get_colid("Satyrium", rank = 234, verbose = FALSE),
    "rank must be of class character")

  # rows param
  expect_error(get_colid("Satyrium", rows = "foobar", verbose = FALSE),
               "'rows' must be numeric or NA")
  expect_error(get_colid("Satyrium", rows = 0, verbose = FALSE),
               "'rows' value must be an integer 1 or greater")
})

#' Retrieve all taxa names or TSNs downstream in hierarchy from given TSN.
#'
#' @export
#' @param id A taxonomic serial number.
#' @param downto The taxonomic level you want to go down to. See examples below.
#' The taxonomic level IS case sensitive, and you do have to spell it
#' correctly. See `data(rank_ref)` for spelling.
#' @param intermediate (logical) If TRUE, return a list of length two with target
#' taxon rank names, with additional list of data.frame's of intermediate
#' taxonomic groups. Default: FALSE
#' @param tsns Deprecated, see `id`
#' @param ... Further args passed on to [ritis::rank_name()] and
#' [ritis::hierarchy_down()]
#' @return Data.frame of taxonomic information downstream to family from e.g.,
#' Order, Class, etc., or if `intermediated=TRUE`, list of length two,
#' with target taxon rank names, and intermediate names.
#' @examples \dontrun{
#' ## the plant class Bangiophyceae, tsn 846509
#' itis_downstream(id = 846509, downto="genus")
#' itis_downstream(id = 846509, downto="genus", intermediate=TRUE)
#'
#' # get families downstream from Acridoidea
#' itis_downstream(id = 650497, "family")
#' ## here, intermediate leads to the same result as the target
#' itis_downstream(id = 650497, "family", intermediate=TRUE)
#'
#' # get species downstream from Ursus
#' itis_downstream(id = 180541, "species")
#'
#' # get orders down from the Division Rhodophyta (red algae)
#' itis_downstream(id = 660046, "order")
#' itis_downstream(id = 660046, "order", intermediate=TRUE)
#'
#' # get tribes down from the family Apidae
#' itis_downstream(id = 154394, downto="tribe")
#' itis_downstream(id = 154394, downto="tribe", intermediate=TRUE)
#' }

itis_downstream <- function(id, downto, intermediate = FALSE, tsns = NULL,
	...) {

  if (!is.null(tsns)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "itis_downstream(tsns)", with = "itis_downstream(id)")
    id <- tsns
  }
  
  if (!is.null(tsns)) id <- tsns
  downto <- tolower(downto)
  downto2 <- taxize_ds$rank_ref[which_rank(downto), "rankid"]
  torank_ids <- taxize_ds$rank_ref[
  	which_rank(downto):NROW(taxize_ds$rank_ref), "rankid"]

	stop_ <- "not"
	notout <- data.frame(rankname = "")
	out <- list()
	if (intermediate) intermed <- list()
	iter <- 0
	while (stop_ == "not") {
		iter <- iter + 1
		if (!nchar(as.character(notout$rankname[[1]])) > 0) {
			temp <- dt2df(lapply(as.character(id), ritis::rank_name),
				idcol = FALSE)
		} else {
		  temp <- notout
		}
		tt <- dt2df(lapply(as.character(temp$tsn), ritis::hierarchy_down),
			idcol = FALSE)
		## we need to run rank_name() to get 'rankid'
		names_ <- dt2df(lapply(split(tt, row.names(tt)), function(x) {
		  ritis::rank_name(as.character(x$tsn))[,c("rankid","rankname","tsn")]
		}), idcol = FALSE)
		##
		if (nrow(names_) == 0) {
		  out[[iter]] <- data.frame(tsn = "No data", parentname = "No data",
		                            parenttsn = "No data", taxonname = "No data",
		                            rankid = "No data", rankname = "No data")
		  stop_ <- "nodata"
		} else {
		  tt <- merge(tt, names_[,-2], by = "tsn")
		  if (intermediate) intermed[[iter]] <- tt
		  if (nrow(tt[tt$rankid == downto2, ]) > 0)
		  	out[[iter]] <- tt[tt$rankid == downto2, ]
		  if (nrow(tt[!tt$rankid == downto2, ]) > 0) {
		    notout <- tt[!tt$rankid %in% torank_ids, ]
		  } else {
		    notout <- data.frame(rankname = downto)
		  }

		  if (length(notout$rankname) > 0)
		  	notout$rankname <- tolower(notout$rankname)
		  if (all(notout$rankname == downto)) {
		    stop_ <- "fam"
		  } else {
		    id <- notout$tsn
		    stop_ <- "not"
		  }
		}
		if (intermediate) {
		  intermed[[iter]] <- stats::setNames(intermed[[iter]],
		  	tolower(names(intermed[[iter]])))
		  intermed[[iter]]$rankname <- tolower(intermed[[iter]]$rankname)
		}
	}
  tmp <- dt2df(out, idcol = FALSE)
  tmp$rankname <- tolower(tmp$rankname)
	if (intermediate) {
	  list(target = stats::setNames(tmp, tolower(names(tmp))),
	  	intermediate = intermed)
  } else {
    stats::setNames(tmp, tolower(names(tmp)))
  }
}

context("itis_downstream")

test_that("itis_downstream returns the correct value", {
  skip_on_cran()
  
  data(rank_ref, package = "taxize")
    
  vcr::use_cassette("itis_downstream", {
    dat_ <- itis_downstream(183264, "Species", verbose=FALSE)
  })

  expect_match(as.character(dat_[1,"rankname"]), "species")
	expect_that(dat_, is_a("data.frame"))
  expect_that(dim(dat_)[2], equals(6))
})

#' Get ITIS terms, i.e., tsn's, authors, common names, and scientific names.
#'
#' @export
#' @param query One or more common or scientific names, or partial names
#' @param what One of both (search common and scientific names), common
#' (search just common names), or scientific (search just scientific names)
#' @param ... Further arguments passed on to [ritis::terms()]
#' @examples \dontrun{
#' # Get terms searching both common and scientific names
#' itis_terms(query='bear')
#'
#' # Get terms searching just common names
#' itis_terms(query='tarweed', "common")
#'
#' # Get terms searching just scientific names
#' itis_terms(query='Poa annua', "scientific")
#' }
itis_terms <- function(query, what = "both", ...) {
  if (!what %in% c('both', 'scientific', 'common')) {
    stop("what must be one of 'both', 'scientific', 'common'", call. = FALSE)
  }
  temp <- switch(
    what,
    both = lapply(query, ritis::terms, ...),
    common = lapply(query, ritis::terms, what = "common", ...),
    scientific = lapply(query, ritis::terms, what = "scientific", ...)
  )
  if (length(query) == 1) {
    temp[[1]]
  } else {
    stats::setNames(temp, query)
  }
}

context("itis_terms")

# test_that("itis_terms returns the correct class", {
#   skip_on_cran()
#   vcr::use_cassette("itis_terms", {
#     one <- itis_terms(query='bear', verbose=FALSE)
#     two <- itis_terms(query='tarweed', "common", verbose=FALSE)
#     three <- itis_terms(query='Poa annua', "scientific", verbose=FALSE)
#   })
# 
#   expect_is(one, "data.frame")
#   expect_is(two, "data.frame")
#   expect_is(three, "data.frame")
# })

# failwith replacment ------------------
default_try <- function(expr, default, quiet = FALSE){
  result <- default
  if (quiet) {
    tryCatch(result <- expr, error = function(e) NULL)
  } else {
    try(result <- expr)
  }
  result
}

fail_with <- function(default = NULL, f, quiet = FALSE){
  f <- match.fun(f)
  function(...) default_try(f(...), default, quiet = quiet)
}

# ldply replacement
# x <- list(a = 5, b = 6)
# nmdlst2df(x)
nmdlst2df <- function(x, stringsAsFactors = FALSE) {
  data.frame(.id = names(x), V1 = unname(unlist(x)),
    stringsAsFactors = stringsAsFactors)
}

tax_GET <- function(url, path = NULL, query = list(), headers = list(),
  opts = list(), ...) {

  cli <- crul::HttpClient$new(url,
    headers = c(headers, tx_ual), opts = c(opts, list(...)))
  out <- cli$get(path = path, query = query)
  out <- tax_error_handle(out)
  return(out)
}

tax_GET_nocheck <- function(url, path = NULL, query = list(), headers = list(),
  opts = list(), ...) {

  cli <- crul::HttpClient$new(url,
    headers = c(headers, tx_ual), opts = c(opts, list(...)))
  out <- cli$get(path = path, query = query)
  return(out)
}

tax_POST <- function(url, path = NULL, query = list(), body = list(),
  headers = list(), opts = list(), ...) {

  cli <- crul::HttpClient$new(url,
    headers = c(headers, tx_ual), opts = c(opts, list(...)))
  out <- cli$post(path = path, query = query, body = body)
  tax_error_handle(out)
  return(out)
}

tax_error_handle <- function(x) {
  x$raise_for_status()
  txt <- x$parse("utf-8")
  xml <- tryCatch(xml2::read_xml(txt), error = function(e) e)
  json <- tryCatch(jsonlite::fromJSON(txt), error = function(e) e)
  if (x$success() && !inherits(json, "error")) return(json)
  if (inherits(xml, "error")) {
    message("running xml cntrl char replacement")
    strg <- "[[:cntrl:]]" # control characters
    xml <- tryCatch(xml2::read_xml(gsub(strg, "", txt)),
      error = function(e) e)
  }
  if (inherits(xml, "error")) stop("error parsing xml")
  errmssg <- tryCatch(xml2::xml_attr(xml, "error_message"),
    error = function(e) e)
  if (inherits(errmssg, "error")) stop("error parsing xml")
  if (nzchar(errmssg)) stop(errmssg)
  return(xml)
}

#' Return all synonyms for a taxon name with a given id.
#'
#' @export
#' @param id the taxon identifier code
#' @param key Your Tropicos API key; See [taxize-authentication] 
#' for help on authentication
#' @param ... Curl options passed on to [crul::HttpClient]
#' @return List or dataframe.
#' @examples \dontrun{
#' tp_synonyms(id = 25509881)
#' }

tp_synonyms <- function(id, key = NULL, ...) {
  url = sprintf('http://services.tropicos.org/Name/%s/Synonyms', id)
	key <- getkey(key, "TROPICOS_KEY")
  args <- tc(list(apikey = key, format = 'json'))
  tt <- tp_GET(url, args, ...)
  res <- jsonlite::fromJSON(tt, FALSE)

  if (names(res[[1]])[[1]] == "Error") {
    nonedf <- data.frame(nameid = "no syns found", scientificname = "no syns found",
                         scientificnamewithauthors = "no syns found",
                         family = "no syns found", stringsAsFactors = FALSE)
    list(accepted = nonedf, synonyms = nonedf)
  } else {
    dat <- lapply(res, function(x) lapply(x, data.frame, stringsAsFactors = FALSE))
    accepted <- dat[[1]]$AcceptedName
    df <- dt2df(lapply(dat, "[[", "SynonymName"), idcol = FALSE)
    synonyms <- df[!duplicated.data.frame(df), ]
    names(accepted) <- tolower(names(accepted))
    names(synonyms) <- tolower(names(synonyms))
    list(accepted = accepted, synonyms = synonyms)
  }
}

context("tp_synonyms")

test_that("tp_synonyms returns the correct value", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tp_synonyms", {
    dat <- suppressMessages(tp_synonyms(id = 25509881))
  }, preserve_exact_body_bytes = TRUE)

  expect_match(names(dat)[[1]], "accepted")

	expect_that(dat, is_a("list"))
	expect_that(dat[[1]], is_a("data.frame"))
	expect_that(dat[[2]], is_a("data.frame"))
	expect_that(ncol(dat[[2]]), equals(4))
})

#' @title Phylotastic Taxonomic Name Resolution Service.
#' @description Defunct - service was down far too much to be reliable
#' @keywords internal 
#' @rdname tnrs-defunct
#' @export
#' @param ... ignored
tnrs <- function(...) {
  .Defunct(msg = "This function is defunct. See ?tnrs")
}

context("tnrs")

test_that("tnrs works", {
  skip_on_cran()
  mynames <- c("Helianthus annuus", "Poa annua", "Mimulus bicolor")
  vcr::use_cassette("tnrs", {
    out <- tnrs(query = mynames, source = "iPlant_TNRS", messages = FALSE)
  })

  expect_equal(ncol(out), 7)
  expect_is(out, "data.frame")
})

test_that("tnrs returns user supplied order and row.names NULLed", {
  skip_on_cran()
  xxx <- c("Abies concolor", "Abies lasiocarpa",
           "Acer sp.", "Acer campestre",
           "Artemisia borealis", "Artemisia cana",
           "Brassica napus", "Brassica oleracea")
  vcr::use_cassette("tnrs_order_row_names", {
    out <- tnrs(xxx, source = "iPlant_TNRS", messages = FALSE)
  })

  expect_equal(NCOL(out), 7)
  expect_equal(NROW(out), 8)
  expect_is(out, "data.frame")

  # row.names are sequential
  expect_equal(as.numeric(row.names(out)), 1:8)
})

#' Retrieve accepted TSN and name
#'
#' @export
#' @param searchtsn One or more TSN for a taxon (numeric/integer)
#' @param ... Curl options passed on to[crul::verb-GET]
#'
#' @return data.frame with with row number equal to input vector length, and
#' with three columns:
#'
#' * submittedtsn (numeric) - The submitted TSN
#' * acceptedname (character) - The accepted name - if the submitted TSN
#'  is the accepted TSN, then this is `NA_character_` because ITIS does
#'  not return a name along with the TSN if it's an accepted name. We could
#'  make an extra HTTP request to ITIS, but that means additional time.
#' * acceptedtsn (numeric) - The accepted TSN
#' * author (character) - taxonomic authority
#'
#' @examples \dontrun{
#' # TSN accepted - good name
#' itis_acceptname(searchtsn = 208527)
#'
#' # TSN not accepted - input TSN is old
#' itis_acceptname(searchtsn = 504239)
#'
#' # many accepted names
#' ids <- c(18161, 18162, 18163, 18164, 18165, 18166, 46173, 46174,
#' 46178, 46181, 46186, 46193, 46196, 46197, 46200, 46201, 46204,
#' 46207, 46867, 46868)
#' itis_acceptname(searchtsn = ids)
#'
#' # many unaccepted names
#' ids <- c(39087, 46208, 46973, 46976, 46978, 46980, 47295, 47445,
#' 47448, 47512, 47515, 47527, 47546, 47622, 47783, 47786, 47787,
#' 47788, 47835, 47839)
#' itis_acceptname(searchtsn = ids)
#'
#' # many: mix of accepted and unaccepted names
#' ids <- c(18161, 18162, 47527, 47546, 47622, 46200)
#' itis_acceptname(searchtsn = ids)
#' }
itis_acceptname <- function(searchtsn, ...) {
  dt2df(lapply(searchtsn, function(z) {
    tmp <- ritis::accepted_names(z)
    if (NROW(tmp) == 0) {
      data.frame(submittedtsn = z, acceptedname = NA, acceptedtsn = z,
                 author = NA, stringsAsFactors = FALSE)
    } else {
      data.frame(
        submittedtsn = z,
        stats::setNames(tmp, tolower(names(tmp))),
        stringsAsFactors = FALSE
      )
    }
  }), idcol = FALSE)
}

context("itis_acceptname")

test_that("itis_acceptname works with accepted tsn", {
  skip_on_cran()
  vcr::use_cassette("itis_acceptname", {
    temp <- itis_acceptname(208527)
  })

  expect_is(temp, "data.frame")
	expect_named(temp, c('submittedtsn', 'acceptedname', 'acceptedtsn', 'author'))
	expect_equal(temp$submittedtsn, 208527)
	expect_true(is.na(temp$acceptedname))
})

test_that("itis_acceptname works with non-accepted tsn", {
  skip_on_cran()
  vcr::use_cassette("itis_acceptname_non_accepted_tsn", {
    temp <- itis_acceptname(504239)
  })

  expect_is(temp, "data.frame")
  expect_named(temp, c('submittedtsn', 'acceptedname', 'acceptedtsn', 'author'))
  expect_equal(temp$submittedtsn, 504239)
  expect_false(is.na(temp$acceptedname))
})

test_that("itis_acceptname fails as expected", {
  skip_on_cran()
  
  expect_error(itis_acceptname(), "\"searchtsn\" is missing")
})

#' Retrieve taxonomic rank name from given TSN.
#'
#' @export
#' @param query TSN for a taxonomic group (numeric). If query is left as default
#' (NULL), you get all possible rank names, and their TSN's (using function
#' [ritis::rank_names()]. There is slightly different terminology for
#' Monera vs. Plantae vs. Fungi vs. Animalia vs. Chromista, so there are
#' separate terminologies for each group.
#' @param ... Further arguments passed on to [ritis::rank_name()]
#' @details You can print messages by setting `verbose=FALSE`.
#' @return Taxonomic rank names or data.frame of all ranks.
#' @examples \dontrun{
#' # All ranks
#' itis_taxrank()
#'
#' # A single TSN
#' itis_taxrank(query=202385)
#'
#' # Many TSN's
#' itis_taxrank(query=c(202385,183833,180543))
#' }
itis_taxrank <- function(query = NULL, ...) {
  if (is.null(query)) {
    ritis::rank_names()
  } else {
    sapply(
      query,
      function(z) as.character(ritis::rank_name(z, ...)$rankname)
    )
  }
}

context("itis_taxrank")

test_that("itis_taxrank returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("itis_taxrank", {
    temp <- itis_taxrank(query=202385, verbose=FALSE)
  })

  expect_match(as.character(temp), "Subspecies")
	expect_is(temp, "character")
})

#' Get APG names
#'
#' Generic names and their replacements from the Angiosperm Phylogeny
#' Group III system of flowering plant classification.
#'
#' @param ... Curl args passed on to [crul::verb-GET]
#' @references http://www.mobot.org/MOBOT/research/APweb/
#' @name apg
#' @examples \dontrun{
#' head(apgOrders())
#' head(apgFamilies())
#' }

#' @export
#' @rdname apg
apgOrders <- function(...) {
  tt <- apg_GET("orders", ...)
  tmp <- strsplit(tt, "<font")[[1]]
  tmp2 <- strsplit(tmp[length(tmp)], "<br>")[[1]]
  tmp3 <- gsub("(<[^>]*>)|\r|\n|\\.$", "", tmp2)
  # remove stuff not cleaned up
  tmp4 <- tmp3[-grep("Back to", tmp3)]
  tmp5 <- sub(".*>", "", tmp4)
  # parse to a data.frame
  accorig <- acc <- grep("[Aa]ccepted", tmp5, value = TRUE)
  acc <- gsub("\\s.+", "", strtrim(acc))
  acc <- acc[nchar(acc) != 0]
  accorig <- accorig[nchar(accorig) != 0]
  accdf <- data.frame(order = acc, synonym = NA,
                      accepted = TRUE,
                      original = accorig, stringsAsFactors = FALSE)

  synorig <- syn <- grep("[Aa]ccepted", tmp5, invert = TRUE, value = TRUE)
  syn <- sapply(strsplit(syn, "=|\\s-"), strtrim)
  syn <- syn[vapply(syn, length, 1) != 0]
  synorig <- synorig[nchar(synorig) != 0]
  syndf <- dt2df(lapply(syn, function(x) {
    tmpdf <- rbind.data.frame(x)
    if (NCOL(tmpdf) == 2) {
      setNames(tmpdf, c("order", "synonym"))
    } else {
      setNames(tmpdf, c("order", "synonym", "comment"))
    }
  }), idcol = FALSE)
  syndf$accepted <- FALSE
  syndf$original <- synorig

  rbind(accdf, syndf)
}

#' @export
#' @rdname apg
apgFamilies <- function(...) {
  tt <- apg_GET("families", ...)
  tmp <- strsplit(tt, "<font")[[1]]
  tmp2 <- strsplit(tmp[length(tmp)], "<br>|<br")[[1]]
  # look for any lines missing a <br> tag, and split
  tmp2 <- unlist(lapply(tmp2, function(z) {
    if (grepl("</a>\r", z)) {
      dd <- strsplit(z, "\r\n")[[1]]
      dd[nchar(dd) != 0]
    } else {
      z
    }
  }))
  # remove html tags
  tmp3 <- gsub("(<[^>]*>)|\r|\n|\\.$", "", tmp2)
  # remove stuff not cleaned up
  tmp4 <- tmp3[-grep("Back to", tmp3)]
  tmp5 <- sub(".*>", "", tmp4)
  # parse to a data.frame
  syn <- sapply(strsplit(tmp5, "=|\\s-"), strtrim)
  syn <- syn[vapply(syn, length, 1) != 0]
  synorig <- tmp5[nchar(tmp5) != 0]
  syndf <- dt2df(lapply(syn, function(x) {
    tmpdf <- rbind.data.frame(x)
    if (NCOL(tmpdf) == 2) {
      setNames(tmpdf, c("family", "order"))
    } else {
      setNames(tmpdf, c("family", "synonym", "order"))
    }
  }), idcol = FALSE)
  syndf[] <- lapply(syndf, as.character)
  syndf$accepted <- syndf$synonym
  syndf$accepted[is.na(syndf$accepted)] <- TRUE
  syndf$accepted[ syndf$accepted != TRUE ] <- FALSE
  syndf$accepted <- as.logical(syndf$accepted)
  syndf$original <- synorig
  syndf
}

apg_GET <- function(x, ...) {
  cli <- crul::HttpClient$new(apg_base(),
    headers = tx_ual, opts = list(...))
  res <- cli$get(sprintf("MOBOT/research/APweb/top/synonymy%s.html", x))
  res$raise_for_status()
  res$parse("UTF-8")
}

apg_base <- function() "http://www.mobot.org"

context("apg* functions")

test_that("apgOrders works", {
  skip_on_cran()
  vcr::use_cassette("apgOrders", {
    orders <- apgOrders()
  })

  expect_is(orders, "data.frame")
  expect_is(orders$order, "character")
  expect_is(orders$accepted, "logical")
  expect_equal(NCOL(orders), 4)
})

test_that("apgFamilies works", {
  skip_on_cran()
  vcr::use_cassette("apgFamilies", {
    families <- apgFamilies()
  })

  expect_is(families, "data.frame")
  expect_is(families$family, "character")
  expect_is(families$accepted, "logical")
  expect_equal(NCOL(families), 5)
})

#' @title Taxonomic Information from Around the Web
#'
#' @description This package interacts with a suite of web 'APIs'
#' for taxonomic tasks, such as verifying species names, getting
#' taxonomic hierarchies, and verifying name spelling.
#'
#' @section About:
#' Allows users to search over many websites for species names (scientific and
#' common) and download up- and downstream taxonomic hierarchical information -
#' and many other things.
#'
#' The functions in the package that hit a specific API have a prefix and suffix
#' separated by an underscore. They follow the format of `service_whatitdoes`.
#' For example, `gnr_resolve` uses the Global Names Resolver API to resolve
#' species names.
#'
#' General functions in the package that don't hit a specific API don't have
#' two words separated by an underscore, e.g., `classification`
#'
#' You need API keys for some data sources. See [taxize-authentication]
#' for more information.
#'
#' @section Currently supported APIs:
#'
#' \tabular{llc}{
#' API \tab prefix \tab SOAP? \cr
#' Encyclopedia of Life (EOL)  \tab eol \tab FALSE \cr
#' Integrated Taxonomic Information Service (ITIS)  \tab itis \tab FALSE \cr
#' Global Names Resolver (from EOL/GBIF) \tab gnr \tab FALSE \cr
#' Global Names Index (from EOL/GBIF) \tab gni \tab FALSE \cr
#' IUCN Red List \tab iucn \tab FALSE \cr
#' Tropicos (from Missouri Botanical Garden) \tab tp \tab FALSE \cr
#' Theplantlist.org \tab tpl \tab FALSE \cr
#' National Center for Biotechnology Information \tab ncbi \tab FALSE \cr
#' CANADENSYS Vascan name search API \tab vascan \tab FALSE \cr
#' International Plant Names Index (IPNI) \tab ipni \tab FALSE \cr
#' World Register of Marine Species (WoRMS) \tab worms \tab TRUE \cr
#' Barcode of Life Data Systems (BOLD) \tab bold \tab FALSE \cr
#' Pan-European Species directories Infrastructure (PESI) \tab pesi \tab TRUE \cr
#' Mycobank \tab myco \tab TRUE \cr
#' National Biodiversity Network (UK)	\tab nbn \tab FALSE \cr
#' Index Fungorum \tab fg \tab FALSE \cr
#' EU BON \tab eubon \tab FALSE \cr
#' Index of Names (ION) \tab ion \tab FALSE \cr
#' Open Tree of Life (TOL) \tab tol \tab FALSE \cr
#' World Register of Marine Species (WoRMS) \tab worms \tab FALSE \cr
#' NatureServe \tab natserv \tab FALSE \cr
#' }
#'
#' If the source above has a `TRUE` in the `SOAP?` column, it is not available
#' in this package. They are available from a different package called **taxizesoap**.
#' See the GitHub repo for how to install https://github.com/ropensci/taxizesoap
#' 
#' @section Catalogue of Life (COL):
#' COL introduced rate limiting recently in 2019 - which has made the API
#' essentially unusable - CoL+ is coming soon and we'll incorporate it here
#' when it's stable. See https://github.com/ropensci/colpluz for the
#' R implementation for CoL+
#'
#' @importFrom graphics plot
#' @importFrom methods as is
#' @importFrom stats as.dist hclust na.omit setNames aggregate complete.cases
#' @importFrom crul HttpClient HttpRequest AsyncVaried upload
#' @importFrom zoo na.locf
#' @importFrom utils URLencode citation download.file read.delim write.table tail
#' @importFrom ape read.tree as.phylo.hclust plot.phylo
#' @importFrom jsonlite fromJSON toJSON
#' @importFrom data.table rbindlist setDF transpose melt dcast as.data.table
#' @importFrom xml2 xml_text xml_find_first xml_find_all xml_children read_xml
#' xml_name xml_ns as_list
#' @importFrom R6 R6Class
#' @importFrom crayon style
#' @importFrom cli symbol cat_line rule
#' @name taxize-package
#' @aliases taxize
#' @author Scott Chamberlain 
#' @author Eduard Szoecs \email{eduardszoecs@@gmail.com}
#' @author Zachary Foster \email{zacharyfoster1989@@gmail.com}
#' @author Carl Boettiger \email{cboettig@@gmail.com}
#' @author Karthik Ram \email{karthik@@ropensci.org}
#' @author Ignasi Bartomeus \email{nacho.bartomeus@@gmail.com}
#' @author John Baumgartner \email{johnbb@@student.unimelb.edu.au}
#' @author James O'Donnell \email{jodonnellbio@@gmail.com}
#' @keywords package
NULL

#' Lookup-table for IDs of taxonomic ranks
#'
#' data.frame of 46 rows, with 2 columns:
#' * rankid - a numeric rank id, consecutive
#' * ranks - a comma separated vector of names that are considered
#'  equal to one another within the row
#'
#' We use this data.frame to do data sorting/filtering based on the ordering
#' of ranks.
#'
#' Please let us know if there is a rank that occurs from one of the data
#' sources \pkg{taxize} that we don't have in `rank_ref` dataset.
#'
#' Let us know if you disagree with the ordering of ranks.
#' 
#' Note that `rankid` 280 are essentially "genetic variants"; placed just above
#' 'unspecified' to denote they're not without rank, but they're not
#' really taxonomic ranks either. As far as I know there's no way 
#' to delineate among these "genetic variant" types.
#'
#' @name rank_ref
#' @docType data
#' @keywords data
NULL

#' Lookup-table for IDs of taxonomic ranks (WoRMS)
#'
#' Same as `rank_ref` but specifically for WoRMS, where section/subsection
#' ranks are put between family/order rather than between species/genus.
#'
#' @name rank_ref_zoo
#' @docType data
#' @keywords data
NULL

#' Lookup-table for family, genus, and species names for ThePlantList
#'
#' These names are from http://www.theplantlist.org, and are from
#' version 1.1 of their data. This data is used in the function
#' [names_list()]. This is a randomly selected subset of the ~350K
#' accepted species names in Theplantlist.
#'
#' @format A data frame with 10,000 rows and 3 variables:
#' 
#'   * `family` family name
#'   * `genus` genus name
#'   * `species` specific epithet name
#' 
#' @source http://www.theplantlist.org
#' @name theplantlist
#' @docType data
#' @keywords data
NULL

#' MOBOT family names
#'
#' Family names and their replacements from the Angiosperm Phylogeny
#' Website system of flowering plant classification.
#'
#' This dataset is from Version 14, incorporated on 2020-06-03, 
#' generated using [apgFamilies()]
#' 
#' (update script in inst/ignore/apg_script.R)
#'
#' @format A data frame with 1705 rows and 6 variables:
#' 
#' * `family`: family name
#' * `synonym`: if `accepted=FALSE`, this is the accepted name;
#' if `accepted=TRUE`, this is `NA`, and the name in `family` is accepted
#' * `order`: order name for the family
#' * `accepted`: logical, if name in `family` column is accepted or not
#' * `original`: original data record from APG website, mapping
#' name in `family` column to a new name, if there is one
#' * `accepted_name`: accepted name. accepted names, combining those that
#' are accepted from `family` column, with the new name from `synonym`
#' if applicable
#' 
#' @source http://www.mobot.org/MOBOT/research/APweb/
#' @name apg_families
#' @docType data
#' @keywords data
NULL

#' MOBOT order names
#'
#' Order names and their replacements from the Angiosperm Phylogeny
#' Website system of flowering plant classification.
#'
#' This dataset is from Version 14, incorporated on 2020-06-03, 
#' generated using [apgOrders()]
#' 
#' (update script in inst/ignore/apg_script.R)
#'
#' @format A data frame with 576 rows and 5 variables:
#' 
#' * `order`: order name
#' * `synonym`: if `accepted=FALSE`, this is the accepted name;
#' if `accepted=TRUE`, this is `NA`, and the name in `order` is accepted
#' * `accepted`: logical, if name in `order` column is accepted or not
#' * `original`: original data record from APG website, mapping
#' name in `order` column to a new name, if there is one
#' * `accepted_name`: accepted name. accepted names, combining those that
#' are accepted from `order` column, with the new name from `synonym`
#' if applicable
#' 
#' @source http://www.mobot.org/MOBOT/research/APweb/
#' @name apg_orders
#' @docType data
#' @keywords data
NULL

#' Vector of plant species (genus - specific epithet) names from ThePlantList
#'
#' These names are from http://www.theplantlist.org, and are a
#' randomly chosen subset of names of the form genus/specific epithet
#' for the purpose of having some names to play with for examples in
#' this package.
#'
#' @format A vector of length 1182
#' @source http://www.theplantlist.org
#' @name plantNames
#' @docType data
#' @keywords data
NULL

#' Vector of plant genus names from ThePlantList
#'
#' These names are from http://www.theplantlist.org, and are a
#' randomly chosen subset of genera names for the purpose of having some
#' names to play with for examples in this package.
#'
#' @format A vector of length 793
#' @source http://www.theplantlist.org
#' @name plantGenusNames
#' @docType data
#' @keywords data
NULL

#' WORMS ranks
#'
#' Created using `worrms::wm_ranks_id(-1)` on 2020-02-11.
#'
#' Present in taxize in the case where WORMS does not
#' return rank names - with this dataset we can fill
#' in rank information as long as rank ids are returned
#'
#' @format A data frame with 216 rows and 2 variables:
#'
#'   * `id`: rank id
#'   * `rank`: rank name
#'
#' @name worrms_ranks
#' @docType data
#' @keywords data
NULL

#' Defunct functions in taxize
#'
#' The following functions are now defunct (no longer available):
#' * All COL functions are defunct: `as.colid, `col_children`,
#' `col_classification`, `col_downstream`, `col_search`, `get_colid`,
#' `get_colid_`, `as.data.frame.colid`, `children.colid`,
#' `classification.colid`, `downstream.colid`, `id2name.colid`,
#' `lowest_common.colid`, `synonyms.colid`, `upstream.colid`
#' * `col_classification()`: See`classification()`
#' * `tp_classification()`: See`classification()`
#' * `eol_hierarchy()`: See`classification()`
#' * [eol_invasive()]: See `eol` in the \pkg{originr} package.
#' * [use_eol()]: EOL no longer requires an API key
#' * [tpl_search()]: Use the \pkg{Taxonstand} functions `TPL` or `TPLck` directly.
#' * [get_seqs()]: This function changed name to`ncbi_getbyname()`()].
#' * [get_genes()]: This function changed name to`ncbi_getbyid()`()].
#' * [get_genes_avail()]: This function changed name to`ncbi_search()`()].
#' * [ncbi_getbyname()]: See `ncbi_byname` in the \pkg{traits} package.
#' * [ncbi_getbyid()]: See `ncbi_byid` in the \pkg{traits} package.
#' * [ncbi_search()]: See `ncbi_searcher` in the \pkg{traits} package.
#' * [gisd_isinvasive()]: See `gisd` in the \pkg{originr} package.
#' * [ubio_classification()]:  The uBio web services was down for quite a while, is
#'  now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
#' * [ubio_classification_search()]: The uBio web services was down for quite
#'  a while, is now (as of 2016-05-09) back up, but we don't trust that it will stay up
#'  and available.
#' * [ubio_id()]: The uBio web services was down for quite a while, is
#'  now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
#' * [ubio_ping()]:  The uBio web services was down for quite a while, is
#'  now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
#' * [ubio_search()]:  The uBio web services was down for quite a while, is
#'  now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
#' * [ubio_synonyms()]:  The uBio web services was down for quite a while, is
#'  now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
#' * [get_ubioid()]: The uBio web services are apparently down indefinitely.
#' * [phylomatic_tree()]: This function is defunct. See
#'  `phylomatic` in the package \pkg{brranching}
#' * [phylomatic_format()]: This function is defunct. See
#'  `phylomatic_names` in the package \pkg{brranching}
#' * [eubon()]: This function is defunct. Use [eubon_search()]
#' * [tnrs()]: This function is defunct. Was too unreliable
#' * [tnrs_sources()]: This function is defunct. Was too unreliable
#'
#' @name taxize-defunct
#' @aliases defunct
NULL

#' Species names from Species Plantarum
#'
#' These names have been compiled from
#' [*Species Plantarum*](https://en.wikipedia.org/wiki/Species_Plantarum) by
#' [Carl Linnaeus](https://en.wikipedia.org/wiki/Carl_Linnaeus) originally
#' published in 1753. It is the first work to consistently apply
#' [binomial names](https://en.wikipedia.org/wiki/Binomial_nomenclature)
#' and was the starting point for the naming of plants. The book lists every
#' species of plant known at the time, classified into
#' [genera](https://en.wikipedia.org/wiki/Genus). The dataset provides a
#' useful reference point to see how taxonomic names have changed since their
#' inception. The names were transcribed by Robert W. Kiger.
#'
#' @format A data frame with 5940 rows and 3 variables:
#' 
#' - genus First part of the binomial species name for each species
#' within the [genus](https://en.wikipedia.org/wiki/Genus)
#' 
#' - epithet specific epithet or second part of the binomial species name
#'   for each [species](https://en.wikipedia.org/wiki/Species)
#' 
#' - page_number The following abbreviations sometimes are used in the
#'   page_number field.  
#'     - "add." refers to addenda that appear on the unnumbered last page of
#'       the index in volume two.
#'     - "err." refers to the unnumbered page of errata that appears following
#'       the index in volume two.
#'     - "canc." following a page number indicates that the binomial appeared
#'       on the cancelled version of that page and does not appear on its
#'       replacement (as in the 1957-1959 facsimile edition.
#' 
#' @source [Hunt Institute for Botanical Documentation](http://fmhibd.library.cmu.edu/HIBD-DB/Species/home.php)
#' @name species_plantarum_binomials
#' @references Linnaeus, C. 1753. Species Plantarum. 2 vols. Salvius, Stockholm.
#'   \[Facsimile edition, 1957-1959, Ray Society, London.\]
#' @author Carl Linnaeus
#' @docType data
#' @keywords data
NULL

#' Get common names from scientific names.
#'
#' @export
#' @param sci character; One or more scientific names or partial names.
#' @param db character; Data source, one of `"ncbi"` (default),
#' `"itis"` `"eol"`, `"worms"`, or `"iucn"`. Note that
#' each taxonomic data source has their own identifiers,  so that if you
#' provide the wrong `db` value for the identifier you could get a
#' result, but it will likely be wrong (not what you were expecting). 
#' If using ncbi or iucn we recommend getting an API key; 
#' see [taxize-authentication]
#' @param simplify (logical) If TRUE, simplify output to a vector of names.
#' If FALSE, return variable formats from different sources, usually a
#' data.frame. Only applies to eol and itis. Specify `FALSE` to obtain
#' the language of each vernacular in the output for eol and itis.
#' @param scinames Deprecated, see `sci`
#' @param ... Further arguments passed on to functions [get_uid()],
#' [get_tsn()].
#' @param id character; identifiers, as returned by [get_tsn()],
#' [get_uid()].
#'
#' @section Authentication:
#' See [taxize-authentication] for help on authentication
#' 
#' @section HTTP version for NCBI requests:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')` 
#' 
#' @return List of character vectors, named by input taxon name,
#' or taxon ID. `character(0)` on no match
#'
#' @seealso [comm2sci()]
#'
#' @author Scott Chamberlain
#'
#' @examples \dontrun{
#' sci2comm(sci='Helianthus annuus')
#' sci2comm(sci='Helianthus annuus', db='eol')
#' sci2comm(sci=c('Helianthus annuus', 'Poa annua'))
#' sci2comm(sci='Puma concolor', db='ncbi')
#' sci2comm('Gadus morhua', db='worms')
#' sci2comm('Pomatomus saltatrix', db='worms')
#' sci2comm('Loxodonta africana', db='iucn')
#'
#' # Passing id in, works for sources: itis and ncbi, not eol
#' sci2comm(get_uid('Helianthus annuus'))
#' sci2comm(get_wormsid('Gadus morhua'))
#' sci2comm(get_iucn('Loxodonta africana'))
#'
#' # Don't simplify returned
#' sci2comm(get_iucn('Loxodonta africana'), simplify=FALSE)
#'
#' # Use curl options
#' sci2comm('Helianthus annuus', db="ncbi", verbose = TRUE)
#' }
#' @rdname sci2comm
sci2comm <- function(...){
  UseMethod("sci2comm")
}

#' @method sci2comm default
#' @export
#' @rdname sci2comm
sci2comm.default <- function(sci, db='ncbi', simplify=TRUE,
  scinames = NULL, ...) {

  if (!is.null(scinames)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "sci2comm(scinames)", with = "sci2comm(sci)")
    sci <- scinames
  }
  
  temp <- lapply(sci, getsci, db = db, simplify = simplify, ...)
  stats::setNames(temp, sci)
}

#' @export
#' @rdname sci2comm
sci2comm.uid <- function(id, ...) {
  warn_db(list(...), "ncbi")
  out <- lapply(id, function(x) ncbi_foo(x, ...))
  names(out) <- id
  return(out)
}

#' @export
#' @rdname sci2comm
sci2comm.tsn <- function(id, simplify=TRUE, ...) {
  warn_db(list(...), "itis")
  out <- lapply(id, function(x) itis_foo(x, simplify, ...))
  names(out) <- id
  return(out)
}

#' @export
#' @rdname sci2comm
sci2comm.wormsid <- function(id, simplify=TRUE, ...) {
  warn_db(list(...), "worms")
  out <- lapply(id, function(x) worms_foo(x, simplify, ...))
  names(out) <- id
  return(out)
}

#' @export
#' @rdname sci2comm
sci2comm.iucn <- function(id, simplify=TRUE, ...) {
  warn_db(list(...), "iucn")
  out <- vector("list", length(id))
  for (i in seq_along(id)) {
    out[[i]] <- iucn2comm(attr(id, "name")[i], simplify = simplify, ...)
  }
  names(out) <- id
  return(out)
}



itis2comm <- function(x, simplify, ...){
  tsn <- get_tsn(x, ...)
  itis_foo(tsn, simplify = simplify, ...)
}

eol2comm <- function(x, simplify, ...){
  tmp <- eol_search(x, ...)
  pageids <- tmp[grep(x, tmp$name, ignore.case = TRUE), "pageid"]
  dfs <- tc(
    lapply(pageids, function(x) {
      tmp <- tryCatch(
        eol_pages(taxonconceptID = x, common_names = TRUE, ...),
        error = function(e) e
      )
      if (inherits(tmp, "error")) NULL else tmp$vernacular
    })
  )
  tt <- dt2df(dfs[sapply(dfs, class) == "data.frame"], idcol = FALSE)
  tt <- tt[!duplicated(tt), ]
  if (simplify) {
    ss <- as.character(tt$vernacularname)
    ss[ !is.na(ss) ]
  } else{
    tt
  }
}

ncbi2comm <- function(x, ...){
  uid <- get_uid(x, ...)
  ncbi_foo(uid, ...)
}

worms2comm <- function(x, simplify, ...){
  id <- get_wormsid(x, ...)
  worms_foo(id, simplify = simplify, ...)
}

iucn2comm <- function(x, simplify, ...){
  # if id is not found
  if (is.na(x)) {
    return(character(0))
  } else {
    out <- get_iucn_data(x)[[1]]
    # if common name is not found
    if (NROW(out$taxon$common_names) == 0) return(character(0))
  }
  if (simplify) {
    return(as.character(out$taxon$common_names$name))
  } else{
    return(out)
  }
}

getsci <- function(nn, db, simplify, ...){
  switch(
    db,
    eol = eol2comm(nn, simplify, ...),
    itis = itis2comm(nn, simplify, ...),
    ncbi = ncbi2comm(nn, ...),
    worms = worms2comm(nn, simplify, ...),
    iucn = iucn2comm(nn, simplify, ...)
  )
}


itis_foo <- function(x, simplify=TRUE, ...){
  # if tsn is not found
  if (is.na(x)) {
    return(character(0))
  } else {
    out <- ritis::common_names(x)
    # if common name is not found
    if (nrow(out) == 0) return(character(0))
  }
  if (simplify) {
    if (!inherits(out, "tbl_df")) out else as.character(out$commonName)
  } else{
    out
  }
}

ncbi_foo <- function(x, ...){
  key <- getkey(NULL, "ENTREZ_KEY")
  query <- tc(list(db = "taxonomy", ID = x, api_key = key))
  cli <- crul::HttpClient$new(url = ncbi_base(), headers = tx_ual,
    opts = list(http_version = 2L, ...))
  res <- cli$get("entrez/eutils/efetch.fcgi", query = query)
  if (!res$success()) return(character(0))
  tt <- res$parse("UTF-8")
  ttp <- xml2::read_xml(tt)
  # common name
  out <- xml_text(
    xml_find_all(ttp,
      "//TaxaSet/Taxon/OtherNames/GenbankCommonName"))
  # NCBI limits requests to three per second
  if (is.null(key)) Sys.sleep(0.33)
  return(out)
}

worms_foo <- function(x, simplify=TRUE, ...){
  # if id is not found
  if (is.na(x)) {
    return(character(0))
  } else {
    out <- worrms::wm_common_id(as.numeric(x))
    #if common name is not found
    if (nrow(out) == 0) return(character(0))
  }
  if (simplify) {
    if (!inherits(out, "tbl_df")) out else as.character(out$vernacular)
  } else{
    out
  }
}

context("sci2comm")

test_that("sci2comm returns the correct value", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("sci2comm", {
    tt <- sci2comm(sci = 'Helianthus annuus', db = 'ncbi', messages = FALSE)
    uu <- sw(sci2comm(sci = 'Helianthus annuus', db = 'itis', rows = 1,
                      messages = FALSE))
    zz <- sci2comm(sci = 'Pomatomus saltatrix', db = 'worms',
                   messages = FALSE)
  })

  expect_that(names(tt), equals('Helianthus annuus'))
  expect_that(names(uu), equals('Helianthus annuus'))
  expect_that(names(zz), equals('Pomatomus saltatrix'))

  expect_that(tt, is_a("list"))
  expect_that(tt[[1]], is_a("character"))

  expect_that(uu, is_a("list"))
  expect_that(uu[[1]], is_a("character"))

  expect_that(zz, is_a("list"))
  expect_that(zz[[1]], is_a("character"))
})

# test_that("warn on mismatch 'db'", {
#   skip_on_cran()
#   vcr::use_cassette("sci2comm_warn_on_db_mismatch", {
#     expect_warning(
#       sci2comm(
#         get_tsn('Helianthus annuus', messages = FALSE), db = "ncbi"))
#   })
# })

get_ids_dbs <- c(
  "itis", "ncbi", "eol", "tropicos", "gbif", "nbn", "pow"
)

#' Retrieve taxonomic identifiers for a given taxon name.
#'
#' This is a convenience function to get identifiers across all data sources.
#' You can use other `get_*` functions to get identifiers from specific
#' sources if you like.
#'
#' @export
#' @param sci_com (character) Taxonomic name to query.
#' @param db (character) database to query. One or more of `ncbi`, `itis`, `eol`,
#' `tropicos`, `gbif`, `nbn`, or `pow`. By default db is set to search
#' all data sources. Note that each taxonomic data source has their own
#' identifiers, so that if you give the wrong `db` value for the identifier you
#' could get a result, it will likely be wrong (not what you were expecting).
#' If using ncbi and/or tropicos we recommend getting API keys;
#' see [taxize-authentication]
#' @param suppress (logical) suppress \pkg{cli} separators with the database
#' name being queried. default: `FALSE`
#' @param rows (numeric) Any number from 1 to infinity. If the default NA, all
#' rows are returned. When used in `get_ids` this function still only
#' gives back a ids class object with one to many identifiers. See
#' `get_ids_` to get back all, or a subset, of the raw data that you
#' are presented during the ask process.
#' @param names Deprecated, see `sci_com`
#' @param ... Other arguments passed to [get_tsn()], [get_uid()],
#' [get_eolid()], [get_tpsid()], [get_gbifid()],
#' [get_nbnid()].
#' @return A vector of taxonomic identifiers, each retaining their respective
#' S3 classes so that each element can be passed on to another function
#' (see e.g.'s).
#' @note There is a timeout of 1/3 seconds between queries to NCBI.
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @section Authentication:
#' See [taxize-authentication] for help on authentication
#'
#' @examples \dontrun{
#' # Plug in taxon names directly
#' # specify rows to limit choices available
#' get_ids("Poa annua", db="eol", rows=1)
#' get_ids("Poa annua", db="eol", rows=1:2)
#'
#' ## Or you can specify which source you want via the db parameter
#' get_ids("Chironomus riparius", db = 'ncbi')
#' get_ids("Salvelinus fontinalis", db = 'nbn')
#'
#' get_ids(c("Chironomus riparius", "Pinus contorta"), db = 'ncbi')
#' get_ids("Pinus contorta", db = c('ncbi','eol','tropicos'))
#' get_ids("ava avvva", db = c('ncbi','eol','tropicos'))
#'
#' # Pass on to other functions
#' out <- get_ids("Pinus contorta", db = c('ncbi','eol','tropicos'))
#' classification(out$ncbi)
#'
#' # Get all data back
#' get_ids_(c("Chironomus riparius", "Pinus contorta"), db = 'nbn',
#'   rows=1:10)
#' get_ids_(c("Chironomus riparius", "Pinus contorta"), db = c('nbn','gbif'),
#'   rows=1:10)
#'
#' # use curl options
#' get_ids("Agapostemon", db = "ncbi", verbose = TRUE)
#' }

get_ids <- function(sci_com,
  db = c("itis", "ncbi", "eol", "tropicos", "gbif", "nbn",
    "pow"), suppress = FALSE, names = NULL, ...) {
  
  assert(suppress, "logical")
  if (is.null(db)) stop("Must specify one or more values for db!")
  db <- match.arg(db, choices = get_ids_dbs, several.ok = TRUE)
  foo <- function(x, sci_com, ...){
    if (!suppress) cat_db(x)
    ids <- switch(x,
                  itis = get_tsn(sci_com, ...),
                  ncbi = get_uid(sci_com, ...),
                  eol = get_eolid(sci_com, ...),
                  tropicos = get_tpsid(sci_com, ...),
                  gbif = get_gbifid(sci_com, ...),
                  nbn = get_nbnid(sci_com, ...),
                  pow = get_pow(sci_com, ...))
    names(ids) <- sci_com
    return( ids )
  }

  tmp <- lapply(db, function(x) foo(x, sci_com = sci_com, ...))
  names(tmp) <- db
  class(tmp) <- "ids"
  return( tmp )
}

#' @export
#' @rdname get_ids
get_ids_ <- function(sci_com, db = get_ids_dbs, rows = NA,
  suppress = FALSE, names = NULL, ...) {

  if (is.null(db)) stop("Must specify on or more values for db!")
  db <- match.arg(db, choices = get_ids_dbs, several.ok = TRUE)
  foo <- function(x, sci_com, rows, ...){
    if (!suppress) cat_db(x)
    ids <- switch(x,
                  itis = get_tsn_(sci_com, rows = rows, ...),
                  ncbi = get_uid_(sci_com, rows = rows, ...),
                  eol = get_eolid_(sci_com, rows = rows, ...),
                  tropicos = get_tpsid_(sci_com, rows = rows, ...),
                  gbif = get_gbifid_(sci_com, rows = rows, ...),
                  nbn = get_nbnid_(sci_com, rows = rows, ...))
    stats::setNames(ids, sci_com)
  }
  structure(stats::setNames(
    lapply(db, function(x) foo(x, sci_com = sci_com, rows = rows, ...)), db),
  class = "ids")
}

cat_db <- function(x) {
  cli::cat_line(
    cli::rule(left = paste0(" db: ", x),
      line = 2, line_col = "blue", width = 30)
  )
}

#' Search Kew's Plants of the World
#'
#' @export
#' @param sci_com (character) query terms, scientific or common name
#' @param limit (integer) Number of records to return. default: 100
#' @param cursor (character) cursor string
#' @param sort (character) The field to sort by and sort order separted with
#' underscore, e.g., `sort="name_desc"`
#' @param q Deprecated, see `sci_com`
#' @param ... Further args passed on to [crul::HttpClient].
#' @return a list with slots for metadata (`meta`) with list of response
#' attributes, and data (`data`) with a data.frame of results
#' @author Scott Chamberlain, 
#' @references https://powo.science.kew.org/
#' @family pow
#' @examples \dontrun{
#' x <- pow_search(sci_com = "Quercus")
#' x$meta
#' x$meta$totalResults
#' x$meta$perPage
#' x$meta$totalPages
#' x$meta$page
#' x$meta$cursor
#' head(x$data)
#'
#' # pagination
#' pow_search(sci_com = "sunflower", limit = 2)
#'
#' # debug curl stuff
#' invisible(pow_search(sci_com = "Helianthus annuus", verbose = TRUE))
#'
#' # sort
#' desc <- pow_search(sci_com = "Helianthus", sort = "name_desc")
#' desc$data$name
#' asc <- pow_search(sci_com = "Helianthus", sort = "name_asc")
#' asc$data$name
#' }
pow_search <- function(sci_com, limit = 100, cursor = "*", sort = NULL,
  q = NULL, ...) {

  assert(sci_com, "character")
  assert(limit, c("integer", "numeric"))
  assert(cursor, "character")
  assert(sort, "character")
  if (!is.null(q)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "pow_search(q)", with = "pow_search(sci_com)")
    sci_com <- q
  }
  
  args <- tc(list(q = sci_com, perPage = limit, cursor = cursor, sort = sort))
  pow_GET(file.path(pow_base(), "api/2", "search"), args, ...)
}

#' Lookup taxa in Kew's Plants of the World
#'
#' @export
#' @param id (character) taxon id. required
#' @param include (character) vector of additional fields to include in
#' results. options include 'distribution' and 'descriptions'. optional
#' @param ... Further args passed on to [crul::HttpClient].
#' @family pow
#' @examples \dontrun{
#' pow_lookup(id = 'urn:lsid:ipni.org:names:320035-2')
#' pow_lookup(id = 'urn:lsid:ipni.org:names:320035-2',
#'   include = "distribution")
#' pow_lookup(id = 'urn:lsid:ipni.org:names:320035-2',
#'   include = c("distribution", "descriptions"))
#' }
pow_lookup <- function(id, include = NULL, ...) {
  assert(include, "character")
  if (!is.null(include)) {
    if (!all(include %in% pow_include_fields)) {
      stop("'include' must be one of ", paste0(pow_include_fields, collapse=","))
    }
  }
  if (!is.null(include)) include <- paste0(include, collapse = ",")
  args <- tc(list(fields = include))
  pow_GET(file.path(pow_base(), "api/2/taxon", id), args, ...)
}

#' Lookup synonyms in Kew's Plants of the World
#'
#' @export
#' @param id (character) taxon id. required
#' @param ... Further args passed on to [pow_lookup()]
#' @family pow
#' @examples \dontrun{
#' pow_synonyms(id = 'urn:lsid:ipni.org:names:320035-2')
#' pow_synonyms(id = 'urn:lsid:ipni.org:names:358881-1')
#' pow_synonyms(id = 'urn:lsid:ipni.org:names:359855-1')
#' }
pow_synonyms <- function(id, ...) {
  res <- pow_lookup(id, ...)
  if ("synonyms" %in% names(res$meta)) {
    tibble::as_tibble(res$meta$synonyms)
  } else {
    tibble::tibble()
  }
}

pow_include_fields <- c("distribution", "descriptions")

pow_GET <- function(url, args, ...){
  cli <- crul::HttpClient$new(url = url,
                              headers = tx_ual, opts = list(...))
  tt <- cli$get(query = argsnull(args))
  tt$raise_for_status()
  json <- jsonlite::fromJSON(tt$parse("UTF-8"))
  meta <- pop(json, "results")
  meta$message <- NULL
  list(meta = meta, data = json$results)
}

pow_base <- function() "https://powo.science.kew.org"

context("pow_search")

test_that("pow_search returns the correct class", {
  skip_on_cran()
  vcr::use_cassette("pow_search", {
    one <- pow_search(sci_com = "Quercus")
  }, preserve_exact_body_bytes = TRUE)

  expect_is(one, "list")
  expect_named(one, c("meta", "data"))
  expect_is(one$meta, "list")
  expect_is(one$meta$cursor, "character")
  expect_is(one$data, "data.frame")
})

test_that("pow_search fails well", {
  skip_on_cran()

  expect_error(pow_search(), "\"sci_com\" is missing")
  expect_error(pow_search(4), "sci_com must be of class")
  expect_error(pow_search("foo", "bar"), "limit must be of class")
  expect_error(pow_search("foo", 1, 5), "cursor must be of class")
  expect_error(pow_search("foo", 1, "bar", 5), "sort must be of class")
})


context("pow_lookup")

test_that("pow_lookup works", {
  skip_on_cran()
  vcr::use_cassette("pow_lookup", {
    one <- pow_lookup(id = "urn:lsid:ipni.org:names:320035-2")
  }, preserve_exact_body_bytes = TRUE)

  expect_is(one, "list")
  expect_named(one, c("meta", "data"))
  expect_is(one$meta, "list")
  expect_is(one$meta$synonyms, "data.frame")
  expect_null(one$meta$distribution)
  expect_null(one$data)
})

test_that("pow_lookup works", {
  skip_on_cran()
  vcr::use_cassette("pow_lookup_include_param", {
    one <- pow_lookup(id = "urn:lsid:ipni.org:names:320035-2",
      include = "distribution")
  }, preserve_exact_body_bytes = TRUE)

  expect_is(one, "list")
  expect_named(one, c("meta", "data"))
  expect_is(one$meta, "list")
  expect_is(one$meta$distribution, "list")
  expect_is(one$meta$distribution$natives, "data.frame")
  expect_is(one$meta$distribution$introduced, "data.frame")
  expect_null(one$data)
})

test_that("pow_lookup returns the correct class", {
  skip_on_cran()

  expect_error(pow_lookup(), "\"id\" is missing")
  expect_error(pow_lookup("asdfds", 5), "include must be of class")
  expect_error(
    pow_lookup(id = "urn:lsid:ipni.org:names:320035-2", include = "foo"),
    "'include' must be one of"
  )
})

#' Get jurisdiction data, i.e., native or not native in a region.
#'
#' @export
#' @param tsn One or more TSN's (taxonomic serial number)
#' @param what One of bytsn, values, or originvalues
#' @param ... Further arguments passed on to
#' [ritis::jurisdictional_origin()],
#' [ritis::jurisdiction_values()], or
#' [ritis::jurisdiction_origin_values()]
#' @examples \dontrun{
#' # Get values
#' itis_native(what="values")
#'
#' # Get origin values
#' itis_native(what="originvalues")
#'
#' # Get values by tsn
#' itis_native(tsn=180543)
#' itis_native(tsn=c(180543,41074,36616))
#' }

itis_native <- function(tsn=NULL, what="bytsn", ...) {
  switch(
    what,
    bytsn = {
      temp <- lapply(tsn, function(x) ritis::jurisdictional_origin(x, ...))
      if (length(tsn) == 1) {
        temp[[1]]
      } else {
        setNames(temp, tsn)
      }
    },
    values = ritis::jurisdiction_values(...),
    originvalues = ritis::jurisdiction_origin_values(...))
}

context("itis_native")

test_that("itis_native returns the correct class", {
  skip_on_cran()
  vcr::use_cassette("itis_native", {
    one <- itis_native(what="values")
    two <- itis_native(what="originvalues")
    three <- itis_native(tsn=180543)
    four <- itis_native(tsn=c(180543,41074,36616))
  })

  expect_is(one, "character")
  expect_is(two, "data.frame")
  expect_is(two, "tbl_df")
  expect_is(three, "data.frame")
  expect_is(three, "tbl_df")
  expect_is(four, "list")
  expect_is(four[[1]], "data.frame")
  expect_is(four[[1]], "tbl_df")
})

rename <- function (x, replace, warn_missing = TRUE) {
  names(x) <- revalue(names(x), replace, warn_missing = warn_missing)
  x
}

revalue <- function (x, replace = NULL, warn_missing = TRUE) {
  if (!is.null(x) && !is.factor(x) && !is.character(x)) {
    stop("x is not a factor or a character vector.")
  }
  mapvalues(x, from = names(replace), to = replace, warn_missing = warn_missing)
}

mapvalues <- function (x, from, to, warn_missing = TRUE) {
  if (length(from) != length(to)) {
    stop("`from` and `to` vectors are not the same length.")
  }
  if (!is.atomic(x)) {
    stop("`x` must be an atomic vector.")
  }
  if (is.factor(x)) {
    levels(x) <- mapvalues(levels(x), from, to, warn_missing)
    return(x)
  }
  mapidx <- match(x, from)
  mapidxNA <- is.na(mapidx)
  from_found <- sort(unique(mapidx))
  if (warn_missing && length(from_found) != length(from)) {
    message("The following `from` values were not present in `x`: ",
            paste(from[!(1:length(from) %in% from_found)], collapse = ", "))
  }
  x[!mapidxNA] <- to[mapidx[!mapidxNA]]
  x
}

# clade <- rank_ref[NROW(rank_ref), ]
# clade$ranks <- paste0(clade$ranks, ",clade")
# rank_ref <- rbind(rank_ref[-NROW(rank_ref), ], clade)
# rank_ref

# cohorts <- c("megacohort", "supercohort", "cohort", "subcohort", "infracohort")
# cohort_ids <- c()

# -----------------
ids <- c('01','05','10','20','25','30','40','45','50','60','70','80','83','84',
    '85','86','87','90','100','110','120','125','130','140','150','160','170',
    '180','190','200','210','215','217','220','225','230','240','250','255',
    '260','265','270','300')
ranks <- c('domain','superkingdom','kingdom','subkingdom','infrakingdom,superphylum',
    'phylum,division','subphylum,subdivision','infradivision','superclass','class',
    'subclass','infraclass','megacohort','supercohort','cohort','subcohort','infracohort',
    'superorder','order','suborder','infraorder','parvorder','superfamily','family',
    'subfamily','tribe','subtribe','genus','subgenus','section','subsection',
    'species group','species subgroup','species','infraspecies','subspecies',
    'variety,varietas','subvariety,race','stirp','form,forma,morph','aberration',
    'subform','unspecified,no rank,unranked,clade')
rank_ref <- data.frame(
  rankid = ids,
  ranks = ranks,
  stringsAsFactors = FALSE
)
save(rank_ref, file = "data/rank_ref.RData", version = 2)

# NOTE: "version = 2" is so that we don't have to require R > 3.5

#' A light wrapper around the taxonstand fxn to call Theplantlist.org database.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @rdname tpl_search-defunct
tpl_search <- function() {
  .Defunct(msg = "This function is defunct. Use the Taxonstand functions TPL or TPLck directly.")
}

#' Get the uBio id for a search term
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @rdname get_ubioid-defunct
#' @keywords internal
#' @export
get_ubioid <- function(...) .Defunct(msg = "the uBio API is gone")

#' @export
#' @rdname get_ubioid-defunct
as.ubioid <- function(...) .Defunct(msg = "the uBio API is gone")

#' @export
#' @rdname get_ubioid-defunct
get_ubioid_ <- function(...) .Defunct(msg = "the uBio API is gone")

#' EUBON children
#'
#' @export
#' @param id (character) identifier for the taxon. (LSID, DOI, URI, or any
#' other identifier used by the checklist provider)
#' @param providers (character) A list of provider id strings concatenated by
#' comma characters. The default : "pesi,bgbm-cdm-server\[col\] will
#' be used if this parameter is not set. A list of all available provider ids
#' can be obtained from the '/capabilities' service end point. Providers can be
#' nested, that is a parent provider can have sub providers. If the id of the
#' parent provider is supplied all subproviders will be queried. The query
#' can also be restricted to one or more subproviders by using the following
#' syntax: parent-id\[sub-id-1,sub-id2,...\]
#' @param timeout (numeric) The maximum of milliseconds to wait for responses
#' from any of the providers. If the timeout is exceeded the service will just
#' return the responses that have been received so far. The default timeout is
#' 0 ms (wait for ever)
#' @param ... Curl options passed on to [crul::verb-GET]
#' @references https://cybertaxonomy.eu/eu-bon/utis/1.3/doc.html
#' @return a data.frame or an empty list if no results found
#' @note There is no pagination in this method, so you may or may not be
#' getting all the results for a search. Sorry, out of our control
#' @family eubon-methods
#' @examples \dontrun{
#' x <- eubon_children(id = "urn:lsid:marinespecies.org:taxname:126141",
#'   providers = 'worms')
#' head(x)
#' }
eubon_children <- function(id, providers = NULL, timeout = 0, ...) {
  args <- tc(list(providers = paste0(providers, collapse = ","),
    timeout = timeout))
  url <- file.path(eubon_base(), "classification", id, "children")
  cli <- crul::HttpClient$new(url, headers = tx_ual, opts = list(...))
  res <- cli$get(query = args)
  eubon_error(res)
  tmp <- jsonlite::fromJSON(res$parse("UTF-8"), TRUE, flatten = TRUE)
  tmp$query$response[[1]]
}

#' Query Phylomatic for a phylogenetic tree.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @keywords internal
#' @rdname phylomatic_tree-defunct
#' @param ... Parameters, ignored
phylomatic_tree <- function(...) {
  .Defunct(msg = "This function is defunct - See ?`taxize-defunct`")
}

#' Get The Plant List csv files.
#'
#' @export
#' @param x Directory to write csv files to.
#' @param family If you want just one, or >1 family, but not all, list
#' them in a vector.
#' @param ... (list) Curl options passed on to [crul::verb-GET]
#' @details Throws a warning if you already have a directory of the one
#' provided, but still works. Writes to your home directory, change x
#' as needed.
#' @return Returns nothing to console, except a message and progress bar.
#' Writes csv files to x.
#' @author John Baumgartner \email{johnbb@@student.unimelb.edu.au}
#' @seealso [tpl_families()]
#' @references The Plant List http://www.theplantlist.org
#' @examples \dontrun{
#' # Get a few families
#' dir <- file.path(tempdir(), "abc")
#' tpl_get(dir, family = c("Platanaceae","Winteraceae"))
#' readLines(file.path(dir, "Platanaceae.csv"), n = 5)
#'
#' # You can now get Gymnosperms as well
#' dir1 <- file.path(tempdir(), "def")
#' tpl_get(dir1, family = c("Pinaceae","Taxaceae"))
#'
#' # You can get mosses too!
#' dir2 <- file.path(tempdir(), "ghi")
#' tpl_get(dir2, family = "Echinodiaceae")
#'
#' # Get all families
#' ## Beware, will take a while
#' ## dir3 <- file.path(tempdir(), "jkl")
#' ## tpl_get("dir3)
#' }

tpl_get <- function(x, family = NULL, ...) {
  cli <- crul::HttpClient$new('http://www.theplantlist.org/1.1/browse/-/',
    headers = tx_ual, opts = list(...))
  temp <- cli$get()
  temp$raise_for_status()
  temp <- xml2::read_html(temp$parse("UTF-8"), encoding = "UTF-8")
  families <- xml2::xml_text(
    xml2::xml_find_all(temp, "//ul[@id='nametree']//a"))
  csvlinks <- sprintf(
    'http://www.theplantlist.org%s%s.csv',
    xml2::xml_attr(xml2::xml_find_all(temp, "//ul[@id='nametree']//a"), 'href'),
    families)
  if (!is.null(family) && all(!family %in% families)) {
    stop(paste('Requested families not found on TPL.',
               'Use tpl_families() to list plant families indexed by TPL.'),
         call. = FALSE)
  }
  if (!is.null(family) && any(!family %in% families)) {
    warning(
      sprintf('Requested families not found on TPL: %s.\n%s',
              paste(family[!family %in% families], collapse = ', '),
              'Use tpl_families() to list plant families indexed by TPL.'),
      call. = FALSE)
  }
  if (!is.null(family)) {
    csvlinks <- csvlinks[families %in% family]
    families <- families[families %in% family]
  }

  message("Downloading csv files to ", x, "...")
  dir.create(x, showWarnings = FALSE, recursive = TRUE)
  invisible(lapply(csvlinks, getcsv, x = x))
  message("...el fin")
}

getcsv <- function(z, x) {
  invisible(crul::HttpClient$new(z)$get(disk = file.path(x, basename(z))))
}

use_fxns_urls <- list(
  tropicos = "http://services.tropicos.org/help?requestkey",
  entrez = "https://www.ncbi.nlm.nih.gov/account/"
)

#' Key helpers
#'
#' @importFrom utils browseURL
#' @title Helpers to set up authentication for the different providers.
#'
#' @description Sets up authentication to diverse providers by providing
#' the user a detailed prompt.
#'
#' @name key_helpers
#' @seealso [taxize-authentication]
NULL

#' @section `use_tropicos()`:
#' Browses to Tropicos API key request URL and provides instruction on how to
#' store the key. After filling the form you will get the key soon, but
#' not immediately.
#' @export
#' @rdname key_helpers
use_tropicos <- function() {
  if (interactive()) {
    utils::browseURL(use_fxns_urls$tropicos)
  }

  message(paste0(
    "After getting your key set it as TROPICOS_KEY in .Renviron.\n ",
    "TROPICOS_KEY='youractualkeynotthisstring'\n ",
    "For that, use usethis::edit_r_environ()"))

  invisible(use_fxns_urls$tropicos)
}

#' use eol
#' @export
#' @name use_eol-defunct
#' @keywords internal
use_eol <- function() {
  .Defunct(msg = "EOL no longer requires an API key")
}

#' @section `use_entrez()`:
#' Browse NCBI Entrez to help make an API key request and provides instruction
#' on how to store the key. There's no direct URL to request a key, one first
#' needs to log in or register and then to generate a key from one's account.
#'
#' Note that NCBI Entrez doesn't require that you use an API key,
#' but you should get higher rate limit with a key, so do get one.
#' @export
#' @rdname key_helpers
use_entrez <- function() {
  if (interactive()) {
    utils::browseURL(use_fxns_urls$entrez)
  }

  message(paste0(
    "Create your key from your (brand-new) account's. \n ",
    "After generating your key set it as ENTREZ_KEY in .Renviron.\n ",
    "ENTREZ_KEY='youractualkeynotthisstring'\n ",
    "For that, use usethis::edit_r_environ()"))

  invisible(use_fxns_urls$entrez)
}

#' @section `use_iucn()`:
#' Browse IUCN Red List API key request URL and provides instruction on how
#' to store the key. This function wraps [rredlist::rl_use_iucn()]
#' from the `rredlist` package. After filling the form you will get
#' the key soon, but not immediately.
#' @export
#' @rdname key_helpers
use_iucn <- function() {
  rredlist::rl_use_iucn()
}

context("test-key_helpers")

test_that("use_tropicos produces expected URL and message", {
  skip_on_cran()
  
  expect_equal(use_tropicos(), "http://services.tropicos.org/help?requestkey")
  expect_message(use_tropicos(), "After getting your key set")
})

test_that("use_eol is defunct", {
  expect_error(use_eol())
})

test_that("use_entrez produces expected URL and message", {
  skip_on_cran()
  
  expect_equal(use_entrez(), "https://www.ncbi.nlm.nih.gov/account/")
  expect_message(use_entrez(), "Create your key")
})

#' Capitalize the first letter of a character string.
#'
#' @export
#' @param s A character string
#' @param strict Should the algorithm be strict about capitalizing.
#'    Defaults to FALSE.
#' @param onlyfirst Capitalize only first word, lowercase all others. Useful
#'    for taxonomic names.
#' @examples
#' taxize_capwords(c("using AIC for model selection"))
#' taxize_capwords(c("using AIC for model selection"), strict=TRUE)

taxize_capwords <- function(s, strict = FALSE, onlyfirst = FALSE) {
  if (onlyfirst) {
    if (strict) {
      gsub("^(.)(.*)$", "\\U\\1\\L\\2", s, perl = TRUE)
    } else {
      gsub("^(.)", "\\U\\1", s, perl = TRUE)
    }
  } else {
    if (strict) {
      gsub("(?<=^| )(.)(.*)(?=$| )", "\\U\\1\\L\\2", s, perl = TRUE)
    } else {
      gsub("(?<=^| )(.)", "\\U\\1", s, perl = TRUE)
    }

  }
}

#' Lookup details for specific names in all taxonomies in GBIF.
#'
#'
#' This is a taxize version of the same function in the `rgbif` package so as to not have to
#' import rgbif and thus require GDAL binary installation.
#'
#' @export
#' @param rank (character) Taxonomic rank. Filters by taxonomic rank as one of:
#'     CLASS, CULTIVAR, CULTIVAR_GROUP, DOMAIN, FAMILY, FORM, GENUS, INFORMAL,
#'   	INFRAGENERIC_NAME, INFRAORDER, INFRASPECIFIC_NAME, INFRASUBSPECIFIC_NAME,
#'     KINGDOM, ORDER, PHYLUM, SECTION, SERIES, SPECIES, STRAIN, SUBCLASS, SUBFAMILY,
#'     SUBFORM, SUBGENUS, SUBKINGDOM, SUBORDER, SUBPHYLUM, SUBSECTION, SUBSERIES,
#'     SUBSPECIES, SUBTRIBE, SUBVARIETY, SUPERCLASS, SUPERFAMILY, SUPERORDER,
#'     SUPERPHYLUM, SUPRAGENERIC_NAME, TRIBE, UNRANKED, VARIETY
#' @param datasetKey (character) Filters by the dataset's key (a uuid)
#' @param uuid (character) A uuid for a dataset. Should give exact same results as datasetKey.
#' @param key (numeric) A GBIF key for a taxon
#' @param name (character) Filters by a case insensitive, canonical namestring,
#'    e.g. 'Puma concolor'
#' @param data (character) Specify an option to select what data is returned. See Description
#'    below.
#' @param language (character) Language, default is english
#' @param sourceId (numeric) Filters by the source identifier. Not used right now.
#' @param shortname (character) A short name..need more info on this?
#' @param ... Curl options passed on to [crul::HttpClient]
#' @param limit Number of records to return
#' @param start Record number to start at
#' @references https://www.gbif.org/developer/summary
#' @return A list of length two. The first element is metadata. The second is
#' either a data.frame (verbose=FALSE, default) or a list (verbose=TRUE)

gbif_name_usage <- function(key=NULL, name=NULL, data='all', language=NULL, datasetKey=NULL, uuid=NULL,
                            sourceId=NULL, rank=NULL, shortname=NULL, start=NULL, limit=20, ...)
{
  calls <- names(sapply(match.call(), deparse))[-1]
  calls_vec <- c("sourceId") %in% calls
  if(any(calls_vec))
    stop("Parameters not currently accepted: \n sourceId")


  args <- tc(list(language=language, name=name, datasetKey=datasetKey,
                  rank=rank, offset=start, limit=limit, sourceId=sourceId))
  data <- match.arg(data,
                    choices=c('all', 'verbatim', 'name', 'parents', 'children',
                              'related', 'synonyms', 'descriptions',
                              'distributions', 'images', 'references', 'species_profiles',
                              'vernacular_names', 'type_specimens', 'root'), several.ok=TRUE)

  # Define function to get data
  getdata <- function(x){
    if (!x == 'all' && is.null(key)) {
      stop('You must specify a key if data does not equal "all"', call. = FALSE)
    }

    if (x == 'all' && is.null(key)) {
      url <- 'https://api.gbif.org/v1/species'
    } else {
      if (x == 'all' && !is.null(key)) {
        url <- sprintf('https://api.gbif.org/v1/species/%s', key)
      } else
        if (x %in% c('verbatim', 'name', 'parents', 'children',
                     'related', 'synonyms', 'descriptions',
                     'distributions', 'images', 'references', 'species_profiles',
                     'vernacular_names', 'type_specimens')) {
          url <- sprintf('https://api.gbif.org/v1/species/%s/%s', key, x)
        } else
          if (x == 'root') {
            url <- sprintf('https://api.gbif.org/v1/species/root/%s/%s', uuid, shortname)
          }
    }
    cli <- crul::HttpClient$new(url = url, headers = tx_ual, opts = list(...))
    res <- cli$get(query = args)
    res$raise_for_status()
    stopifnot(res$response_headers$`content-type` == 'application/json')
    jsonlite::fromJSON(res$parse("UTF-8"), FALSE)
  }

  # Get data
  if (length(data) == 1) {
    getdata(data)
  } else {
    lapply(data, getdata)
  }
}

# NOTE: The code in this file was copied from the bold R package in order to circumvent the bold package being taken off CRAN.
#    If the bold package is put back on CRAN, this code will be deleted and the offical bold package will be used instead.
#    The code has been copied under the following license:
#
# MIT License
# 
# Copyright (c) 2021 bold authors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#   
#   The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


#' Search BOLD for taxonomy data by BOLD ID.
#'
#' @keywords internal
bold_tax_id <- function(id, dataTypes = "basic", includeTree = FALSE,
                        response = FALSE, ...) {
  .Deprecated("bold_tax_id2")
  #-- arguments check
  if (missing(id)) stop("argument 'id' is missing, with no default")
  if (!grepl("true|false", includeTree, ignore.case = TRUE))
    warning("'includeTree' should be either TRUE or FALSE.")
  if (!inherits(id, c("character", "numeric", "integer")))
    warning("'id' should be of class character, numeric or integer.")
  #-- make sure user have correct data types
  dataTypes <- .check_dataTypes_dep(dataTypes)
  if (!nzchar(dataTypes)) {
    out <- data.frame(input = id, noresults = NA)
  } else {
    #-- prep query parameters
    params <- list(dataTypes = dataTypes, includeTree = tolower(includeTree))
    #-- make URL
    URL <- b_url("API_Tax/TaxonData")
    #-- fetch data from the api
    res <- lapply(`names<-`(id, id), function(x) {
      # no need to do the call if id is NA
      if (is.na(x))
        data.frame(input = NA, noresults = NA)
      else
        .get_response_dep(args = c(taxId = x, params), url = URL, ...)
    })
    if (response) {
      out <- res
    } else {
      out <- b_rbind(mapply(FUN = .process_response_dep, x = res, y = id,
                            MoreArgs = list(z = includeTree, w = dataTypes),
                            SIMPLIFY = FALSE))
      if (NCOL(out) == 1) {
        out$noresults <- NA
      }
    }
  }
  out
}
.process_response_dep <- function(x, y, z, w){
  if (is.data.frame(x)) return(x)
  tt <- rawToChar(x$content)
  out <- if (x$status_code > 202) "stop" else jsonlite::fromJSON(tt)
  if ( length(out) == 0 || identical(out[[1]], list()) || any(out == "stop") ) {
    data.frame(input = y, stringsAsFactors = FALSE)
  } else {
    if (w %in% c("stats",'images','geo','sequencinglabs','depository')) out <- out[[1]]
    trynames <- tryCatch(as.numeric(names(out)), warning = function(w) w)
    if (!inherits(trynames, "simpleWarning")) names(out) <- NULL
    if (any(vapply(out, function(x) is.list(x) && length(x) > 0, logical(1)))) {
      out <- lapply(out, function(x) Filter(length, x))
    } else {
      out <- Filter(length, out)
    }
    if (!is.null(names(out))) {
      df <- data.frame(out, stringsAsFactors = FALSE)
    } else {
      df <- b_rbind(lapply(out, data.frame, stringsAsFactors = FALSE))
    }
    row.names(df) <- NULL
    if ("parentid" %in% names(df)) df <- df[order(df[,"parentid"]),]
    row.names(df) <- NULL
    data.frame(input = y, df, stringsAsFactors = FALSE)
  }
}
.get_response_dep <- function(args, url, ...){
  cli <- crul::HttpClient$new(url = url)
  out <- cli$get(query = args, ...)
  out$raise_for_status()
  stopifnot(out$headers$`content-type` == 'text/html; charset=utf-8')
  return(out)
}
.check_dataTypes_dep <- function(x){
  x <- b_split(x, ",", fixed = TRUE, simplify = TRUE)
  # corrects for the json typo in case the option is taken from a previous query
  x[x == "depositories"] <- "depository"
  if (length(x) > 1 && any(x == "all")) {
    x <- "all"
  } else {
    wrongType <- !x %in% b_dataTypes
    if (any(wrongType)) {
      warning(b_ennum(x[wrongType], quote = TRUE),
              if (sum(wrongType) > 1) " are not valid data types"
              else " is not a valid data type",
              if (!all(wrongType)) " and will be skipped." else ".",
              "\nChoices are:",
              b_ennum(b_dataTypes, "or", TRUE))
      x <- x[!wrongType]
    }
    x <- paste(x, collapse = ",")
  }
  x
}

#' Search BOLD for taxonomy data by BOLD ID.
#'
#' @keywords internal
bold_tax_id2 <- function(id, dataTypes = 'basic', includeTree = FALSE,
                         response = FALSE, ...) {
  #-- arguments check
  if (missing(id)) stop("argument 'id' is missing, with no default")
  # no need to do the call if id is NA
  if (length(id) == 1 && is.na(id))
    return(data.frame(input = NA, taxid = NA))
  b_assert_logical(includeTree)
  b_assert(dataTypes, "character")
  if (is.list(id)) {
    lapply(id, b_assert, what = c("character", "numeric", "integer"), name = "id")
  } else {
    b_assert(id, c("character", "numeric", "integer"))
  }
  # for compatibility with bold_tax_id
  if (length(dataTypes) == 1 && grepl(",", dataTypes)) {
    dataTypes <- c(b_split(dataTypes, ",", fixed = TRUE, simplify = TRUE))
  }
  dataTypes <- b_assert_dataTypes(dataTypes)
  #-- prep query params
  params <- list(
    dataTypes = paste(dataTypes, collapse = ","),
    includeTree = tolower(includeTree)
  )
  #-- fetch data from the api
  res <- lapply(b_nameself(id), function(x) {
    if (!length(x) || !nzchar(x)) {
      list(response = NULL, warning = "id was empty")
    } else if (is.na(x) || grepl("^NA?$", x)) {
      list(response = NULL, warning = "id was NA")
    } else {
      r <- b_GET(query = c(taxId = x, params), api = "API_Tax/TaxonData",
                 check = TRUE, ...)
      if (nzchar(r$warning) || r$response$status_code > 202) {
        w <- xml2::read_html(r$response$content)
        w <- b_lines(xml2::xml_text(w))
        w <- w[b_detect(w, "Fatal")]
        if (b_detect(w, "non-object")) w <- c(w, "Request returned no match.")
        r$warning <- c(r$warning, w)
        r$response <- NULL
      }
      r
    }
  })
  if (response) {
    res
  } else {
    #-- make data.frame with the response(s)
    #-- fixing dataTypes to match bold response
    if (length(dataTypes) == 1 && dataTypes == "all") {
      dataTypes <- b_dataTypes[1:7]
    }
    dataTypes[dataTypes == "geo"] <- "country"
    #-- parsing bold response by id
    out <- mapply(
      b_process_tax_id,
      res,
      ids = id,
      MoreArgs = list(types = dataTypes, tree = includeTree),
      SIMPLIFY = FALSE
    )
    out <- b_format_tax_id_output(out, types = dataTypes, tree = includeTree)
    # -- add attributes to output
    w <- lapply(res, `[[`, "warning")
    attr(out, "errors") <- b_rm_empty(w)
    attr(out, "params") <- list(dataTypes = dataTypes,
                                includeTree = includeTree)
    out
  }
}
b_process_tax_id <- function(x, ids, types, tree) {
  if (length(x$response)) {
    out <- b_parse(x$response, format = "json", raise = FALSE)
    if (!length(out) || identical(out[[1]], list())) {
      data.frame(taxid = NA)
    } else {
      if (!tree) {
        out <- b_format_tax_id(out, ids = ids)
      } else {
        tmp <-
          lapply(b_nameself(names(out)), function(id) b_format_tax_id(out[[id]], id))
        out <-
          lapply(b_nameself(types), b_grp_dataTypes_tree, x = tmp)
        if ("basic" %in% names(out) &&
            "parentid" %in% names(out[["basic"]])) {
          out[["basic"]] <-
            out[["basic"]][order(out[["basic"]][["parentid"]]), ]
        }
      }
      out
    }
  }
}
b_format_tax_id <- function(x, ids) {
  # 'geo' isn't group like the others for some reason
  # it's split between country and site map, default
  # was to return only 'country' (although I'm not sure why),
  # so removing the sitemap
  x$sitemap <- NULL
  # some dataTypes returns data.frames or long lists
  # simplifying all of those would make the data frame very large.
  # so adjusting for the different types :
  basic.nms <- which(names(x) %in% c("taxid", "taxon", "tax_rank", "tax_division",
                                     "parentid", "parentname", "taxonrep"))
  wiki.nms <- which(names(x) %in% c("wikipedia_summary", "wikipedia_link"))
  if (length(x$stats)) {
    x$stats <- b_format_tax_id_stats(x$stats)
  }
  if (!all(is.na(basic.nms))) {
    x$basic <- data.frame(x[basic.nms])
    x <- x[-basic.nms]
  }
  if (!all(is.na(wiki.nms))) {
    x$thirdparty <- data.frame(x[wiki.nms])
    x <- x[-wiki.nms]
  }
  for (i in which(vapply(x[!names(x) %in% c("stats", "basic", "thirdparty")], class, character(1L)) == "list")) {
    x[i] <- b_format_tax_id_list(x[i], ids = ids)
  }
  x
}
b_format_tax_id_stats <- function(x) {
  if ("publicmarkersequences" %in% names(x)) {
    markers <- which(names(x) == "publicmarkersequences")
    data.frame(as.list(c(x[-markers], x[markers], recursive = TRUE)))
  } else {
    data.frame(as.list(c(x, recursive = TRUE)))
  }
}
b_format_tax_id_list <- function(x, ids) {
  lapply(
    b_nameself(names(x)),
    function(nm) {
      out <- c(x[[nm]], recursive = TRUE, use.names = TRUE)
      out <- data.table::data.table(taxid = ids, names(out), count = out)
      data.table::setnames(out, 2, nm)
      out
    }
  )
}
b_fill_empty <- function(x, types){
  len0 <- lengths(x) == 0
  if (any(len0)) {
    filler <- list(data.frame(col2rm = NA_character_))
    filler <- list(`names<-`(rep(filler, length(types)), types))
    x[which(len0)] <- filler
  }
  x
}
b_format_tax_id_output <- function(x, types, tree) {
  types <- b_nameself(types)
  #-- to match previous behavior of passing NAs
  x <- b_fill_empty(x, types)
  if (tree && length(types) == 1 &&
      !types %in% c("basic", "thirdparty", "images", "stats")) {
    x <- b_rbind(unlist(x, recursive = FALSE, use.names = T), idcol = "input")
    x$input <- as.integer(b_extract(x$input, "^[0-9]+(?=\\.)"))
  }  else {
    x <- lapply(types, b_grp_dataTypes, x = x, idcol = "input")
  }
  if (length(x) == 1) {
    x[[1]]
  } else {
    x
  }
}
b_grp_dataTypes_tree <- function(nm, x) {
  idcol <- if (nm == "stats" || nm == "thirdparty" || nm == "images") "taxid" else NULL
  out <- b_rbind(lapply(x, `[[`, nm), idcol = idcol)
  out$taxid <- as.integer(out$taxid)
  out
}
b_grp_dataTypes <- function(nm, x, idcol = FALSE) {
  out <- b_rbind(lapply(x, `[[`, nm), idcol = idcol)
  out[["input"]] <- as.integer(out[["input"]])
  out[names(out) == "col2rm"] <- NULL
  out
}






b_assert <- function(x,
                     what,
                     name = NULL,
                     check.length = NULL) {
  if (!length(name)) {
    name <- substitute(x)
  }
  msgLen <- if (length(check.length) && !isFALSE(check.length)) {
    b_assert_length(x = x, len = check.length, name = name,
                    stopOnFail = length(x) > 0)
  } else {
    NULL
  }
  msgClass <- if (length(x)) {
    b_assert_class(x = x, what = what, name = name,
                   is2nd = length(msgLen), stopOnFail = FALSE)
  } else {
    NULL
  }
  msg <- c(msgLen, msgClass)
  if (length(msg)) {
    stop(msg, call. = FALSE)
  }
}
b_assert_class <- function(x, what, name, is2nd = FALSE, stopOnFail = TRUE) {
  .fun <- if (stopOnFail) stop else paste0
  if (!inherits(x = x, what = what)) {
    if (!is2nd)
      .fun("'", name, "' must be of class ", b_ennum(what, "or"))
    else
      .fun(" and of class ", b_ennum(what, "or"))
  } else {
    NULL
  }
}
b_assert_length <- function(x, len, name, stopOnFail = TRUE) {
  len <- as.integer(len)
  if (!is.na(len) && len >= 0) {
    .fun <- if (stopOnFail) stop else paste0
    if (len == 0 && !length(x)) {
      .fun("'", name, "' can't be empty")
    } else if (len > 0 && length(x) != len) {
      .fun("'", name, "' must be length ", len)
    }
  }
}
b_assert_logical <- function(x, name = NULL) {
  b_assert_length(x, len = 1L, name = name)
  if (!length(name)) name <- substitute(x)
  x <- tolower(x)
  if (x == "true" || x == "1")
    TRUE
  else if (x == "false" || x == "0" || x == "na")
    FALSE
  else
    stop("'", name, "' should be one of TRUE or FALSE")
}
b_validate <- function(x, choices, name){
  wrong <- !x %in% choices
  if (any(wrong)) {
    stop(
      b_ennum(x[wrong], quote = TRUE),
      if (sum(wrong) > 1)
        " are not valid "
      else
        " is not a valid ",
      name,
      "\nChoices are ",
      b_ennum(choices, join_word = "or", quote = TRUE),
      call. = FALSE
    )
  }
}
b_get_db <- function(x){
  opts <- list(case_insensitive = TRUE)
  if (b_detect(x, '^COX[1I]$', opts_regex = opts))
    "COX1"
  else if (b_detect(x, '^pub(lic)?$|_public$', opts_regex = opts))
    "COX1_SPECIES_PUBLIC"
  else if (b_detect(x, '^spe(cies)?$|_species$', opts_regex = opts))
    "COX1_SPECIES"
  else if (b_detect(x, '^(cox[1I]_)?(l640)?bp$', opts_regex = opts))
    "COX1_L640bp"
  else
    x
}
b_assert_db <- function(x){
  b_assert(x, "character", name = "db", check.length = 1L)
  x <- b_get_db(x)
  b_validate(x, choices = b_db, name = "db")
  x
}
b_get_tax_division <- function(x){
  x <- tolower(x)
  x[b_detect(x, '^animal')] <- "Animalia"
  x[b_detect(x, '^prot')] <- "Protista"
  x[b_detect(x, '^fun')] <- "Fungi"
  x[b_detect(x, '^plant')] <- "Plantae"
  x
}
b_assert_tax_division <- function(x){
  if (length(x)) {
    b_assert(x, what = "character", name = "tax_division")
    x <- b_get_tax_division(x)
    b_validate(x, choices = b_tax_division, name = "tax_division")
  }
  x
}
b_get_tax_rank <- function(x){
  x <- tolower(x)
  x[b_detect(x, '^king')] <- "kingdom"
  x[b_detect(x, '^phy')] <- "phylum"
  x[b_detect(x, '^cla')] <- "class"
  x[b_detect(x, '^ord')] <- "order"
  x[b_detect(x, '^fam')] <- "family"
  x[b_detect(x, '^subfam')] <- "subfamily"
  x[b_detect(x, '^tribe')] <- "tribe"
  x[b_detect(x, '^gen')] <- "genus"
  x[b_detect(x, '^spe')] <- "species"
  x[b_detect(x, '^subspe')] <- "subspecies"
  x
}
b_assert_tax_rank <- function(x){
  if (length(x)) {
    b_assert(x, what = "character", name = "tax_rank")
    x <- b_get_tax_rank(x)
    b_validate(x, choices = b_tax_rank, name = "tax_rank")
  }
  x
}
b_assert_format <- function(x){
  b_assert(x, what = "character", check.length = 1L, name = "format")
  x <- tolower(x)
  if (x != "xml" && x != "tsv")
    stop("'format' should be one of 'xml' or 'tsv'")
  else
    x
}
b_get_dataTypes <- function(x){
  x <- tolower(x)
  if (any(x == "all")) {
    x <- "all"
  } else {
    # corrects for the json typo in case the option is taken from a previous query
    # and for short versions/typos
    x[x == "basics"] <- "basic"
    x[x == "depo" | x == "depositories"] <- "depository"
    x[x == "labs" | x == "sequencinglab"] <- "sequencinglabs"
    x[x == "stat"] <- "stats"
    x[x == "img"] <- "images"
    x[x == "wiki"] <- "thirdparty"
  }
  x
}
b_assert_dataTypes <- function(x){
  b_assert(x, what = "character", name = "dataTypes", check.length = 0L)
  x <- b_get_dataTypes(x)
  b_validate(x, choices = b_dataTypes, name = "dataTypes")
  x
}



# -- API requests helpers
b_url <- function(api) paste0('https://v4.boldsystems.org/index.php/', api)
b_pipe_params <- function(..., paramnames = ...names(), params = list(...)) {
  params <- b_rm_empty(params)
  if (!length(params))
    stop("You must provide a non-empty value to at least one of:\n\t",
         b_ennum(paramnames, join_word = "or", quote = TRUE), call. = FALSE)
  wt <- !vapply(params, is.character, logical(1))
  if (any(wt))
    stop(b_ennum(names(wt)[wt], quote = TRUE), " must be of class character", call. = FALSE)
  if (length(params$taxon)) {
    # in case it comes from `bold_tax_name()` (#84)
    params$taxon <- b_fix_taxonName(params$taxon)
  }
  vapply(params, paste, collapse = "|", character(1))
}
b_GET <- function(query, api, check = FALSE,
                  contentType = 'text/html; charset=utf-8', ...) {
  cli <- crul::HttpClient$new(url = b_url(api))
  res <- cli$get(query = query, ...)
  if (check) b_CHECK(res = res, contentType = contentType)
  else res
}
b_CHECK <- function(res, contentType) {
  w <- ""
  # get HTTP error if any
  if (res$status_code > 202) {
    x <- res$status_http()
    w <- paste0("HTTP ", x$status_code, ": ", x$message, "\n ", x$explanation)
  }
  # check if the content returned is of the right type
  if (res$response_headers$`content-type` != contentType) {
    w <- paste0(
      "Content was type '", res$headers$`content-type`,
      "' when it should've been type '", contentType, "'")
  }
  # if warning call it now
  if (nzchar(w)) warning(w, call. = FALSE, immediate. = TRUE)
  list(response = res, warning = w)
}
b_parse <- function(res, format, raise = TRUE, cleanData = FALSE, multiple = FALSE){
  if (raise) res$raise_for_status()
  res <- {
    if (!multiple)
      rawToChar(res$content)
    else
      paste0(rawToChar(res$content, multiple = TRUE), collapse = "")
  }
  res <- enc2utf8(res)
  if (res == "") {
    NA
  } else {
    if (b_detect(res, "Fatal error")) {
      # if returning partial output for bold_seq, might as well do that here too
      warning("the request timed out, see 'If a request times out'\n",
              "returning partial output")
      res <- b_drop(str = res, pattern = "Fatal error")
      # if missing, adding closing tag so it can be read properly
      if (format == "xml" && !b_detect(res, "</bold_records")) {
        res <- paste0(res, "</bold_records>")
      }
    }
    switch(
      format,
      xml = b_read_xml(res),
      json = b_read_json(res),
      tsv = b_read_tsv(res, cleanData = cleanData),
      fasta = b_read_fasta(res)
    )
  }
}
b_cleanData <- function(x, emptyValue = NA){
  col2clean <- vapply(x, function(x) {
    any(b_detect(x, "|", max_count = 1, fixed = TRUE), na.rm = TRUE)
  }, NA)
  col2clean <- which(col2clean, useNames = FALSE)
  for (.col in col2clean) {
    x[[.col]] <- b_replace(
      x[[.col]],
      # if the same text is repeated
      # or if only "||||"
      "^([^\\|]+)(\\|\\1)+$|^\\|+$",
      # keep first text/replace with nothing
      "$1")
  }
  x[x == ""] <- emptyValue
  x
}
b_read_json <- function(x){
  jsonlite::parse_json(x, simplifyVector = TRUE, flatten = TRUE)
}
b_read_xml <- function(x) {
  # DON'T REMOVE OPTIONS
  # necessary for large request!
  xml2::read_xml(x, options = c("NOBLANKS", "HUGE"))
}
b_read_tsv <- function(x, header = TRUE, sep = "\t",
                       cleanData = FALSE, ...) {
  # -- fix for issue 104 ----------------------------------------
  # wild '\n' characters in some fields
  if (stringi::stri_detect_regex(x, "[^\r]\n[^\r]")) {
    x_names <- stringi::stri_extract_first_regex(x, "^[^\r\n]+")
    x_names <- stringi::stri_split_regex(x_names, "\t", simplify = TRUE)
    pttrn <- paste0("(?<!", x_names[length(x_names)], "|\r)\n(?!=\r)")
    x <- stringi::stri_replace_all_regex(x, pttrn, " ")   
  }
  # -------------------------------------------------------------
  x <- data.table::setDF(
    data.table::fread(
      text = x,
      header = header,
      sep = sep,
      ...
    )
  )
  if (cleanData) {
    b_cleanData(x)
  } else {
    x
  }
}
b_nameself <- function(x){
  `names<-`(x, x)
}
b_rm_empty <- function(x) {
  if (!length(x)) {
    NULL
  } else {
    x[lengths(x) > 0 & nzchar(x)]
  }
}
b_rbind <- function(x, fill = TRUE, use.names = TRUE, idcol = NULL) {
  (x <- data.table::setDF(
    data.table::rbindlist(l = x, fill = fill, use.names = use.names, idcol = idcol))
  )
}
b_ennum <- function(x, join_word = "and", quote = FALSE) {
  if (!is.character(x)) x <- as.character(x)
  x <- x[nzchar(x)]
  if (quote) {
    x <- paste0("'", x, "'")
  }
  n <- length(x)
  if (n > 1)
    paste(paste(x[-n], collapse = ", "), join_word, x[n])
  else
    x
}
b_fix_taxonName <- function(x){
  # see issue #84
  # check if supposed to:
  x <- b_replace(x,
                 # be quoted; keep quoted text
                 " ('[^']*)$",
                 # add closing quote
                 " \\\\$1\\\\'")
  x <- b_replace(x,
                 # be in parenthesis; keep parenthesis text
                 " (\\([^\\(]*)$",
                 # add closing parenthesis
                 " $1)")
  x <- b_replace(x,
                 # end with a dot (there might be more cases, but haven't seen them yet)
                 "( sp(\\. nov)?$)",
                 # add end dot
                 "$1.")
  x
}


b_extract <- function(str, pattern, mode = "all",
                      fixed = FALSE, simplify = TRUE, ...) {
  switch(mode,
         first = {
           if (!fixed)
             stringi::stri_extract_first_regex(str = str, pattern = pattern, ...)
           else
             stringi::stri_extract_first_fixed(str = str, pattern = pattern, ...)
         },
         last = {
           if (!fixed)
             stringi::stri_extract_last_regex(str = str, pattern = pattern, ...)
           else
             stringi::stri_extract_last_fixed(str = str, pattern = pattern, ...)
         },
         all = {
           if (!fixed)
             stringi::stri_extract_all_regex(str = str,
                                             pattern = pattern,
                                             simplify = simplify,
                                             ...)
           else
             stringi::stri_extract_all_fixed(str = str,
                                             pattern = pattern,
                                             simplify = simplify,
                                             ...)
         },
         str)
}
b_words <- function(str, mode = "first", ...) {
  switch(mode,
         first = stringi::stri_extract_first_words(str = str, ...),
         last = stringi::stri_extract_last_words(str = str, ...),
         all = stringi::stri_extract_all_words(str = str, ...),
         str
  )
}
b_replace <- function(str, pattern, replacement, fixed = FALSE, ...) {
  if (!fixed)
    stringi::stri_replace_all_regex(str = str, pattern = pattern,
                                    replacement = replacement, ...)
  else
    stringi::stri_replace_all_fixed(str = str, pattern = pattern,
                                    replacement = replacement, ...)
}
b_detect <- function(str, pattern, fixed = FALSE, ...) {
  if (!fixed)
    stringi::stri_detect_regex(str = str, pattern = pattern, ...)
  else
    stringi::stri_detect_fixed(str = str, pattern = pattern, ...)
}
b_split <- function(str, pattern, fixed = FALSE, ...) {
  if (!fixed)
    stringi::stri_split_regex(str = str, pattern = pattern, ...)
  else
    stringi::stri_split_fixed(str = str, pattern = pattern, ...)
}
b_count <- function(str, pattern, fixed = FALSE, ...) {
  if (!fixed)
    stringi::stri_count_regex(str = str, pattern = pattern, ...)
  else
    stringi::stri_count_fixed(str = str, pattern = pattern, ...)
}
b_drop <- function(str, pattern) {
  to <- stringi::stri_locate_first_fixed(str = str, pattern = pattern)[[1]] - 1L
  stringi::stri_sub(str = str, to = to)
}
b_lines <- function(str) {
  stringi::stri_split_lines1(str = str)
}

b_read_fasta <- function(x){
  x <- b_lines(str = x)
  if (length(x) %% 2 && # if length(x) %% 2 not 0, it's TRUE
      b_detect(x[length(x)], "^\\s*$")) {
    x <- x[-length(x)]
  }
  id_line <- b_detect(x,"^>")
  n <- which(id_line)
  id <- b_split(str = x[id_line], pattern = ">|\\|", omit_empty = TRUE, simplify = NA, n = 4)
  id <- `names<-`(as.data.frame(id), c("processid", "identification", "marker", "accession"))
  if (all(diff(c(n, length(x) + 1L)) == 2)) {
    sequence <- x[n + 1L]
    data.frame(id, sequence)
  } else if (sum(id_line) == sum(!id_line)) {
    # shouldn't happen but who knows
    warning("The file had an even number of ids and sequences, but they weren't in the proper order.",
            "\n  This shouldn't happen. Output may contain errors.",
            "\n  Please open an issue so we can see when this happens.")
    sequence <- x[!id_line]
    data.frame(id, sequence)
  } else {
    warning("The file had an uneven number of ids and sequenceuences.",
            "\n  This shouldn't happen. Returning data as a list of lines.",
            "\n  Please open an issue so we can see when this happens.")
    x
  }
}

bold_tax_name <- function(name, fuzzy = FALSE, response = FALSE,
                          tax_division = NULL, tax_rank = NULL, ...) {
  if (missing(name)) stop("argument 'name' is missing, with no default")
  b_assert(name, "character", check.length = 0L)
  # fix for #84:
  name <- b_replace(name, "(?<=[^\\\\])'", "\\\\'")
  if (!missing(response)) response <- b_assert_logical(response)
  if (!missing(fuzzy)) fuzzy <- b_assert_logical(fuzzy)
  if (!missing(tax_division)) tax_division <- b_assert_tax_division(tax_division)
  if (!missing(tax_rank)) tax_rank <- b_assert_tax_rank(tax_rank)
  res <- lapply(b_nameself(name), function(x)
    b_GET(
      query = c(taxName = x, fuzzy = tolower(fuzzy)),
      api = "API_Tax/TaxonSearch",
      check = TRUE,
      ...
    ))
  if (response) {
    res
  } else {
    out <- b_rbind(lapply(res, b_process_tax_name,
                          tax_division = tax_division,
                          tax_rank = tax_rank),
                   idcol = "input")
    w <- lapply(res, `[[`, "warning")
    # using %in% to avoid NAs
    noMatchDiv <- out$taxid %in% -1L
    noMatchRank <- out$taxid %in% -2L
    noMatch <- out$taxid %in% -3L
    out$taxid[noMatchDiv | noMatchRank | noMatch] <- NA_integer_
    w[noMatchDiv] <- "Request returned no match with supplied 'tax_division'"
    w[noMatchRank] <- "Request returned no match with supplied 'tax_rank'"
    w[noMatch] <- "Request returned no match"
    attr(out, "errors") <- b_rm_empty(w)
    attr(out, "params") <- list(fuzzy = fuzzy,
                                tax_division = tax_division,
                                tax_rank = tax_rank)
    out
  }
}

b_process_tax_name <- function(x, tax_division, tax_rank) {
  if (!nzchar(x$warning)) {
    out <- b_parse(x$response, format = "json")
    if (length(out) && length(out$top_matched_names)) {
      out <- data.frame(out$top_matched_names)
      out.nms <- names(out)
      # see issue #84
      if (any(out.nms == "taxon")) out$taxon <- b_fix_taxonName(out$taxon)
      if (any(out.nms == "parentname")) out$parentname <- b_fix_taxonName(out$parentname)
      if (any(out.nms == "taxonrep")) out$taxonrep <- b_fix_taxonName(out$taxonrep)
      if (length(tax_division) && any(out.nms == "tax_division")) {
        if (sum(out$tax_division %in% tax_division)) {
          out <- out[out$tax_division %in% tax_division,]
        } else {
          out <- data.frame(taxid = -1L) # request returned no match with supplied tax_division
        }
      }
      if (length(tax_rank)  && any(out.nms == "tax_rank")) {
        if (sum(out$tax_rank %in% tax_rank)) {
          out <- out[out$tax_rank %in% tax_rank,]
        } else {
          out <- data.frame(taxid = -2L) # request returned no match with supplied tax_rank
        }
      }
    } else {
      out <- data.frame(taxid = -3L) # request returned no match
    }
  } else {
    out <- data.frame(taxid = NA_integer_) # resquest had an error
  }
  out
}

b_dataTypes <- c(
  "basic",
  "stats",
  "geo",
  "images",
  "sequencinglabs",
  "depository",
  "thirdparty",
  "all"
)
b_tax_division <- c("Animalia", "Protista", "Fungi", "Plantae")
b_tax_rank <-
  c(
    "kingdom",
    "phylum",
    "class",
    "order",
    "family",
    "subfamily",
    "tribe",
    "genus",
    "species",
    "subspecies"
  )
b_db <- c('COX1', 'COX1_SPECIES', 'COX1_SPECIES_PUBLIC', 'COX1_L640bp')

#' Replacement function for ldply that should be faster in all cases.
#'
#' @param x A list.
#' @param convertvec Convert a vector to a data.frame before rbind is called.
#' @export
#' @keywords internal
taxize_ldfast <- function(x, convertvec=FALSE){
  convert2df <- function(x){
    if (!inherits(x, "data.frame")) {
      data.frame(rbind(x))
    } else {
      x
    }
  }

  if (convertvec) {
    dt2df(lapply(x, convert2df), idcol = FALSE)
  } else {
    dt2df(x, idcol = FALSE)
  }
}

#' Get the UK National Biodiversity Network ID from taxonomic names.
#'
#' @export
#' @param sci_com character; a vector of common or scientific names. Or, a
#' `taxon_state` object (see [taxon-state])
#' @param ask logical; should get_nbnid be run in interactive mode?
#' If TRUE and more than one ID is found for the species, the user is asked for
#' input. If FALSE NA is returned for multiple matches.
#' @param messages logical; If TRUE the actual taxon queried is printed on the
#' console.
#' @param rec_only (logical) If `TRUE` ids of recommended names are
#' returned (i.e. synonyms are removed). Defaults to `FALSE`. Remember,
#' the id of a synonym is a taxa with 'recommended' name status.
#' @param rank (character) If given, we attempt to limit the results to those
#' taxa with the matching rank.
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this function still only gives back a nbnid
#' class object with one to many identifiers. See
#' [get_nbnid_()] to get back all, or a subset, of the raw
#' data that you are presented during the ask process.
#' @param name Deprecated, see `sci_com`
#' @param ... Further args passed on to `nbn_search`
#' @param x Input to [as.nbnid()]
#' @param check logical; Check if ID matches any existing on the DB, only
#' used in [as.nbnid()]
#' @template getreturn
#'
#' @references https://api.nbnatlas.org/
#' @return an object of class nbnid, a light wrapper around a character
#' string that is the taxonomic ID - includes attributes with relavant
#' metadata
#' @family taxonomic-ids
#' @family nbn
#' @seealso [classification()]
#'
#' @author Scott Chamberlain, 
#'
#' @examples \dontrun{
#' get_nbnid(sci_com='Poa annua')
#' get_nbnid(sci_com='Poa annua', rec_only=TRUE)
#' get_nbnid(sci_com='Poa annua', rank='Species')
#' get_nbnid(sci_com='Poa annua', rec_only=TRUE, rank='Species')
#' get_nbnid(sci_com='Pinus contorta')
#'
#' # The NBN service handles common names too
#' get_nbnid(sci_com='red-winged blackbird')
#'
#' # specify rows to limit choices available
#' get_nbnid('Poa ann')
#' get_nbnid('Poa ann', rows=1)
#' get_nbnid('Poa ann', rows=25)
#' get_nbnid('Poa ann', rows=1:2)
#'
#' # When not found
#' get_nbnid(sci_com="uaudnadndj")
#' get_nbnid(c("Zootoca vivipara", "uaudnadndj"))
#' get_nbnid(c("Zootoca vivipara","Chironomus riparius", "uaudnadndj"))
#'
#' # Convert an nbnid without class information to a nbnid class
#' as.nbnid(get_nbnid("Zootoca vivipara")) # already a nbnid, returns the same
#' as.nbnid(get_nbnid(c("Zootoca vivipara","Pinus contorta"))) # same
#' as.nbnid('NHMSYS0001706186') # character
#' # character vector, length > 1
#' as.nbnid(c("NHMSYS0001706186","NHMSYS0000494848","NBNSYS0000010867"))
#' # list
#' as.nbnid(list("NHMSYS0001706186","NHMSYS0000494848","NBNSYS0000010867"))
#' ## dont check, much faster
#' as.nbnid('NHMSYS0001706186', check=FALSE)
#' as.nbnid(list("NHMSYS0001706186","NHMSYS0000494848","NBNSYS0000010867"),
#'   check=FALSE)
#'
#' (out <- as.nbnid(c("NHMSYS0001706186","NHMSYS0000494848",
#'   "NBNSYS0000010867")))
#' data.frame(out)
#' as.nbnid( data.frame(out) )
#'
#' # Get all data back
#' get_nbnid_("Zootoca vivipara")
#' get_nbnid_("Poa annua", rows=2)
#' get_nbnid_("Poa annua", rows=1:2)
#' get_nbnid_(c("asdfadfasd","Pinus contorta"), rows=1:5)
#'
#' # use curl options
#' invisible(get_nbnid("Quercus douglasii", verbose = TRUE))
#' }

get_nbnid <- function(sci_com, ask = TRUE, messages = TRUE, rec_only = FALSE,
  rank = NULL, rows = NA, name = NULL, ...) {

  assert(sci_com, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(rec_only, "logical")
  assert(rank, "character")
  assert(messages, "logical")
  assert_rows(rows)
  if (!is.null(name)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_nbnid(name)", with = "get_nbnid(sci_com)")
    sci_com <- name
  }
  
  if (inherits(sci_com, "character")) {
    tstate <- taxon_state$new(class = "nbnid", names = sci_com)
    items <- sci_com
  } else {
    assert_state(sci_com, "nbnid")
    tstate <- sci_com
    sci_com <- tstate$taxa_remaining()
    items <- c(sci_com, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci_com)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci_com[i], "'\n")
    df <- nbn_search(sci_com = sci_com[i], rows = 500, fq = "idxtype:TAXON",
      ...)$data
    if (is.null(df) || length(df) == 0) df <- data.frame(NULL)
    mm <- NROW(df) > 1

    rank_taken <- NA
    if (NROW(df) == 0) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      att <- 'not found'
    } else {
      if (rec_only) df <- df[ df$taxonomicStatus == 'accepted', ]
      if (!is.null(rank)) df <- df[ tolower(df$rank) == tolower(rank), ]
      df <- sub_rows(df, rows)
      df <- df[,c('guid', 'scientificName', 'rank', 'taxonomicStatus')]
      names(df)[1] <- 'nbnid'
      id <- df$nbnid
      rank_taken <- as.character(df$rank)
      att <- 'found'
    }

    # not found on NBN
    if (length(id) == 0) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      att <- 'not found'
    }
    # more than one, try for direct match
    if (length(id) > 1) {
      matchtmp <- df[tolower(df$scientificName) %in% tolower(sci_com[i]),]
      if (NROW(matchtmp) == 1) {
        id <- matchtmp$nbnid
        rank_taken <- as.character(matchtmp$rank)
        direct <- TRUE
        att <- "found"
      }
    }

    # more than one found -> user input
    if (length(id) > 1) {
      if (ask) {
        rownames(df) <- seq_len(NROW(df))
        # prompt
        message("\n\n")
        message("\nMore than one NBN ID found for taxon '", sci_com[i], "'!\n
            Enter rownumber of taxon (other inputs will return 'NA'):\n")
        message(paste0(utils::capture.output(df), collapse = "\n"))
        take <- scan(n = 1, quiet = TRUE, what = 'raw')

        if (length(take) == 0) {
          take <- 'notake'
          att <- 'nothing chosen'
        }
        if (take %in% seq_len(nrow(df))) {
          take <- as.numeric(take)
          message("Input accepted, took nbnid '",
                  as.character(df$nbnid[take]), "'.\n")
          id <- as.character(df$nbnid[take])
          rank_taken <- as.character(df$rank[take])
          att <- 'found'
        } else {
          id <- NA_character_
          att <- 'not found'
          mssg(messages, "\nReturned 'NA'!\n\n")
        }
      } else{
        if (length(id) != 1) {
          warning(sprintf(m_more_than_one_found, "NBN ID", sci_com[i]),
            call. = FALSE)
          id <- NA_character_
          att <- m_na_ask_false
        }
      }
    }
    # list(id = id, rank = rank_taken, att = att, multiple = mm, direct = direct)
    res <- list(id = id, rank = rank_taken, att = att, multiple = mm,
      direct = direct)
    prog$completed(sci_com[i], att)
    prog$prog(att)
    tstate$add(sci_com[i], res)
  }
  # name <- as.character(name)
  # out <- lapply(name, fun, ask = ask, messages = messages, rows = rows, ...)
  # ids <- pluck(out, "id", "")
  # atts <- pluck(out, "att", "")
  # ids <- structure(ids, class = "nbnid", match = atts,
  #                  multiple_matches = pluck(out, "multiple", logical(1)),
  #                  pattern_match = pluck(out, "direct", logical(1)))
  # if ( !all(is.na(ids)) ) {
  #   urls <- sapply(out, function(z){
  #     if (!is.na(z[['id']]))
  #       sprintf('https://species.nbnatlas.org/species/%s', z[['id']])
  #     else
  #       NA
  #   })
  #   attr(ids, 'uri') <- unlist(urls)
  # }
  # return(ids)
  out <- tstate$get()
  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "nbnid",
                   match = pluck_un(out, "att", ""),
                   multiple_matches = pluck_un(out, "multiple", logical(1)),
                   pattern_match = pluck_un(out, "direct", logical(1)))
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$nbn)
}

#' @export
#' @rdname get_nbnid
as.nbnid <- function(x, check=TRUE) UseMethod("as.nbnid")

#' @export
#' @rdname get_nbnid
as.nbnid.nbnid <- function(x, check=TRUE) x

#' @export
#' @rdname get_nbnid
as.nbnid.character <- function(x, check=TRUE) {
  if (length(x) == 1) {
    make_nbnid(x, check)
  } else {
    collapse(x, make_nbnid, "nbnid", check = check)
  }
}

#' @export
#' @rdname get_nbnid
as.nbnid.list <- function(x, check=TRUE) {
  if (length(x) == 1) {
    make_nbnid(x, check)
  } else {
    collapse(x, make_nbnid, "nbnid", check = check)
  }
}

#' @export
#' @rdname get_nbnid
as.nbnid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class = "nbnid", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri = x$uri)
}

#' @export
#' @rdname get_nbnid
as.data.frame.nbnid <- function(x, ...) {
  data.frame(ids = as.character(unclass(x)),
             class = "nbnid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_nbnid <- function(x, check=TRUE) {
  make_generic(x, 'https://species-ws.nbnatlas.org/species/%s', "nbnid", check)
}

check_nbnid <- function(x){
  url <- "https://species-ws.nbnatlas.org/species/"
  res <- tax_GET_nocheck(paste0(url, x))
  if (res$status_code == 200) TRUE else FALSE
}

#' @export
#' @rdname get_nbnid
get_nbnid_ <- function(sci_com, messages = TRUE, rec_only = FALSE, rank = NULL,
                       rows = NA, name = NULL, ...) {
  if (!is.null(name)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_nbnid_(name)", with = "get_nbnid_(sci_com)")
    sci_com <- name
  }
  stats::setNames(lapply(sci_com, get_nbnid_help, messages = messages,
                  rec_only = rec_only, rank = rank, rows = rows, ...), sci_com)
}

get_nbnid_help <- function(name, messages, rec_only, rank, rows, ...){
  mssg(messages, "\nRetrieving data for taxon '", name, "'\n")
  df <- nbn_search(sci_com = name, all = TRUE, ...)$data
  if (is.null(df)) df <- data.frame(NULL)

  if (NROW(df) == 0) {
    NULL
  } else {
    if (rec_only) df <- df[ df$taxonomicStatus == 'accepted', ]
    if (!is.null(rank)) df <- df[ df$rank == rank, ]
    df <- df[,c('guid', 'scientificName', 'rank', 'taxonomicStatus')]
    if (NROW(df) == 0) NULL else sub_rows(df, rows)
  }
}

#' Parse scientific names using EOL's name parser.
#'
#' @export
#' @param names A vector of length 1 or more of taxonomic names
#' @param ... Curl options passed on to [crul::verb-GET]
#' @return A data.frame with results, the submitted names, and the
#' parsed names with additional information.
#' @seealso [gbif_parse()], [gn_parse()]
#' @references <http://gni.globalnames.org/>
#' @examples \dontrun{
#' gni_parse("Cyanistes caeruleus")
#' gni_parse("Plantago minor")
#' gni_parse("Plantago minor minor")
#' gni_parse(c("Plantago minor minor","Helianthus annuus texanus"))
#'
#' # pass on curl options
#' gni_parse("Cyanistes caeruleus", verbose = TRUE)
#' }
gni_parse <- function(names, ...) {
  names <- paste0(names, collapse = "|")
  cli <- crul::HttpClient$new(paste0(gni_base(), "parsers.json"),
    headers = tx_ual, opts = list(...))
  tt <- cli$get(query = list(names = names))
  tt$raise_for_status()
  out <- jsonlite::fromJSON(tt$parse("UTF-8"), FALSE)
  dt2df(lapply(out, gni_parser), idcol = FALSE)
}

gni_parser <- function(x) {
  positions_names <- vapply(x$scientificName$positions, function(y)
    paste("position_", y[[1]], sep = ""), "", USE.NAMES = FALSE)
  nums <- vapply(x$scientificName$positions, function(y) y[[2]], 1,
    USE.NAMES = FALSE)
  pv <- data.frame(as.list(setNames(nums, positions_names)),
    stringsAsFactors = FALSE)

  singles <- data.frame(x$scientificName[c("verbatim","canonical",
    "normalized","hybrid","parsed")], stringsAsFactors = FALSE)
  details_ <- x$scientificName$details[[1]]
  details_ <- details_[!names(details_) %in% 'status']
  details <- dt2df(Map(function(x, y) data.frame(y, x,
    stringsAsFactors = FALSE), details_, names(details_)),
    idcol = FALSE)[,-3]
  details2 <- as.data.frame(t(data.frame(details[,2])))
  names(details2) <- details[,1]
  row.names(details2) <- NULL
  data.frame(details2, singles, pv, stringsAsFactors = FALSE)
}

context("gni_parse")

test_that("gni_parse returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("gni_parse", {
    tt <- gni_parse("Cyanistes caeruleus")
  })

  expect_match(as.character(tt[,2]), "caeruleus")
	expect_equal(tt[,"position_genus"], 9)
	expect_is(tt, "data.frame")
})

#' Return all reference records for for a taxon name with a given id.
#'
#' @export
#' @param id the taxon identifier code
#' @param key Your Tropicos API key; See [taxize-authentication]
#' for help on authentication
#' @param ... Curl options passed on to [crul::HttpClient]
#' @return List or dataframe.
#' @examples \dontrun{
#' tp_refs(id = 25509881)
#' }
tp_refs <- function(id, key = NULL, ...) {
  url = sprintf('http://services.tropicos.org/Name/%s/References', id)
	key <- getkey(key, "TROPICOS_KEY")

  args <- tc(list(apikey = key, format = 'json'))
  tt <- tp_GET(url, args, ...)
  res <- jsonlite::fromJSON(tt, FALSE)
  dt2df(lapply(res, function(x){
    x <- x$Reference
    names(x) <- tolower(names(x))
    data.frame(x, stringsAsFactors = FALSE)
  }), idcol = FALSE)
}

#' Return all reference records for for a taxon name with a given id.
#'
#' Function name changed to tp_refs.
#'
#' @export
#' @keywords internal
#' @param ... ignored
#' @rdname tp_namereferences-deprecated
tp_namereferences <- function(...) {
  .Deprecated("tp_refs", "taxize", "Function name changed. See tp_refs", "tp_namereferences")
}

context("tp_refs")

test_that("tp_refs returns the correct class", {
  skip_on_cran() # uses secrets
  vcr::use_cassette("tp_refs", {
    ttt <- suppressMessages(tp_refs(id = 25509881))
  }, preserve_exact_body_bytes = TRUE)

	expect_that(ttt, is_a("data.frame"))
})

#' Global Names Resolver Data Sources
#'
#' Retrieve data sources used in the Global Names Resolver
#'
#' @export
#' @param ... Curl options passed on to [crul::HttpClient]
#' @param todf defunct, always get a data.frame back now
#' @return data.frame/tibble
#' @seealso [gnr_resolve()], [gna_search()]
#' @keywords resolve names taxonomy
#' @references https://resolver.globalnames.org/data_sources
#' @examples \dontrun{
#' # all data sources
#' gnr_datasources()
#'
#' # give me the id for EOL
#' out <- gnr_datasources()
#' out[out$title == "EOL", "id"]
#'
#' # Fuzzy search for sources with the word zoo
#' out <- gnr_datasources()
#' out[agrep("zoo", out$title, ignore.case = TRUE), ]
#' }
gnr_datasources <- function(..., todf) {
  
  lifecycle::deprecate_warn(when = "v0.9.102", what = "gnr_datasources()", with = "gna_data_sources()")
  
  if (!missing(todf)) stop("todf is defunct", call. = FALSE)
  cli <- crul::HttpClient$new(
    url = "https://resolver.globalnames.org/data_sources.json",
    headers = tx_ual, opts = list(...))
  res <- cli$get()
  res$raise_for_status()
  tibble::as_tibble(jsonlite::fromJSON(res$parse("UTF-8")))
}

context("gnr_datasources")
test_that("gnr_datasources returns the correct class", {
  skip_on_cran()
  vcr::use_cassette("gnr_datasources", {
    tmp <- gnr_datasources()
  })
  expect_is(tmp, "data.frame")
  expect_equal(NCOL(tmp), 12)
  expect_is(tmp$title, "character")
  expect_is(tmp$updated_at, "character")
  expect_type(tmp$id, "integer")
})

test_that("gnr_datasources fails well", {
  expect_error(gnr_datasources(todf = 5),
    "todf is defunct")
})

#' Print a subset of a character vector
#'
#' Prints the start and end values for a character vector. The number of values
#' printed depend on the width of the screen by default.
#'
#' @param chars (`character`) What to print.
#' @param prefix (`character` of length 1) What to print before
#'   `chars`, on the same line.
#' @param max_chars (`numeric` of length 1) The maximum number of
#'   characters to print.
#' @param type (`"error"`, `"warning"`, `"message"`, `"cat"`, `"print"`, `"silent"`)
#'
#' @return `NULL`
#' @keywords internal
limited_print <- function(chars, prefix = "",
                          max_chars = getOption("width") - nchar(prefix) - 5,
                          type = "message") {

  if (length(chars) == 0) {
    cat(prefix)
    return(invisible(NULL))
  }


  # https://stat.ethz.ch/pipermail/r-help/2006-March/101023.html
  interleave <- function(v1,v2) {
    ord1 <- 2*(1:length(v1))-1
    ord2 <- 2*(1:length(v2))
    c(v1,v2)[order(c(ord1,ord2))]
  }

  q = "'"
  interleaved <- interleave(chars[1:(length(chars) / 2)],
                            rev(chars[(length(chars) / 2 + 1):length(chars)]))
  is_greater_than_max <- cumsum(nchar(interleaved) + 2) + 10 > max_chars
  if (all(! is_greater_than_max)) {
    max_printed <- length(chars)
  } else {
    max_printed <- which.max(is_greater_than_max)
  }
  if (max_printed < length(chars)) {
    first_part <-  chars[1:as.integer(max_printed / 2 - 0.5)]
    second_part <-
      chars[as.integer(length(chars) - (max_printed / 2) + 1.5):length(chars)]
    output <- paste0(paste0(collapse = ", ", first_part),
                     " ... ",
                     paste0(collapse = ", ", second_part),
                     "\n")
  } else {
    output <- paste0(paste0(collapse = ", ", chars), "\n")
  }
  output <- paste(prefix, output, collapse = "")

  if (type == "error") {
    stop(output)
  } else if (type == "warning") {
    warning(output)
  } else if (type == "message") {
    message(output)
  } else if (type == "cat") {
    cat(output)
  } else if (type == "print") {
    message(paste0(utils::capture.output(output), collapse = "\n"))
  } else if (type != "silent") {
    stop("invalid type option")
  }
  return(invisible(output))
}

#' Parse scientific names using Global Names Parser
#'
#' @export
#' @param names A vector of length 1 or more taxonomic names
#' @param ... Curl options passed on to [crul::verb-GET]
#' @return A data.frame with results, the submitted names, and the
#' parsed names with additional information.
#' @seealso [gbif_parse()], [gni_parse()]
#' @references http://gni.globalnames.org/
#' @examples \dontrun{
#' gna_parse("Cyanistes caeruleus")
#' gna_parse("Plantago minor")
#' gna_parse("Plantago minor minor")
#' gna_parse(c("Plantago minor minor","Helianthus annuus texanus"))
#' 
#' # if > 20 names, uses an HTTP POST request
#' x <- names_list("species", size = 30)
#' gna_parse(x)
#'
#' # pass on curl options
#' gna_parse("Cyanistes caeruleus", verbose = TRUE)
#' }
gna_parse <- function(names, ...) {
  assert(names, "character")
  method <- ifelse(length(names) <= 20, "get", "post")
  tmp <- gna_http(method, names, ...)
  out <- jsonlite::fromJSON(tmp)
  out[paste0('canonical_', colnames(out$canonical))] <- out$canonical
  out$canonical <- NULL
  tibble::as_tibble(out)
}

gna_http <- function(method, names, ...) {
  cli <- crul::HttpClient$new("https://parser.globalnames.org/api/v1/",
    headers = tx_ual, opts = list(...))
  res <- switch(method,
    get = {
      cli$get(paste0('api/v1/', paste0(names, collapse = "|")))
    },
    post = {
      cli$headers <- c(cli$headers, list(`Content-Type` = "application/json",
        accept = "application/json"))
      cli$post(body = jsonlite::toJSON(list(names = names)))
    }
  )
  res$raise_for_status()
  res$parse("UTF-8")
}


#' Parse scientific names using EOL's name parser.
#'
#' THIS FUNCTION IS DEFUNCT.
#' 
#' @export
#' @keywords internal
gni_parse <- function(names, ...) {
  .Defunct("ncbi_searcher", "traits",
           msg = "This function is defunct. See gna_parse()")
}

test_that("gna_parse", {
  skip_on_cran()
  vcr::use_cassette("gna_parse", {
    tt <- gna_parse("Cyanistes caeruleus")
  })

  expect_is(tt, "data.frame")
  expect_is(tt, "tbl")
  expect_is(tt$normalized, "character")
  expect_is(tt$canonical_stemmed, "character")
})

test_that("gn_parse fails well", {
  skip_on_cran()
  
  expect_error(gna_parse())
  expect_error(gna_parse(5))
})

#' Get rank for a given taxonomic name.
#'
#' @export
#' @param sci_id (character) Vector of one or more taxon names (character) or
#' IDs (character or numeric) to query. Or objects returned from `get_*()`
#' functions like [get_tsn()]
#' @param db (character) database to query. either `ncbi`, `itis`, `eol`,
#' `tropicos`, `gbif`,`nbn`, `worms`, `natserv`, `bold`. Note that each
#' taxonomic data source has their own identifiers, so that if you provide the
#' wrong `db` value for the identifier you may get a result, but it will
#' likely be wrong (not what you were expecting). If using ncbi we
#' recommend getting an API key; see [taxize-authentication]
#' @param rows numeric; Any number from 1 to infinity. If the default NA, 
#' all rows are considered. passed down to `get_*()` functions.
#' @param x Deprecated, see `sci_id`
#' @param ... Additional arguments to [classification()]
#' @return A named list of character vectors with ranks (all lower-cased)
#' @note While [tax_name()] returns the name of a specified
#' rank, [tax_rank()] returns the actual rank of the taxon.
#' @seealso [classification()],[tax_name()]
#' @examples \dontrun{
#' tax_rank("Helianthus annuus", db = "itis")
#' tax_rank("Helianthus annuus", db = "natserv")
#' tax_rank(get_tsn("Helianthus annuus"))
#' tax_rank(c("Helianthus", "Pinus", "Poa"), db = "itis")
#'
#' tax_rank(get_boldid("Helianthus annuus"))
#' tax_rank("421377", db = "bold")
#' tax_rank(421377, db = "bold")
#'
#' tax_rank(c("Plantae", "Helianthus annuus",
#'   "Puma", "Homo sapiens"), db = 'itis')
#' tax_rank(c("Helianthus annuus", "Quercus", "Fabaceae"), db = 'tropicos')
#'
#' tax_rank(names_list("species"), db = 'gbif')
#' tax_rank(names_list("family"), db = 'gbif')
#'
#' tax_rank(c("Gadus morhua", "Lichenopora neapolitana"),
#'   db = "worms")
#' }
tax_rank <- function(sci_id, db = NULL, rows = NA, x = NULL, ...) {
  UseMethod("tax_rank")
}

#' @export
tax_rank.default <- function(sci_id, db = NULL, rows = NA, x = NULL, ...) {
  stats::setNames(tax_rank_(sci_id, ...), sci_id)
}

#' @export
tax_rank.character <- function(sci_id, db = NULL, rows = NA, x = NULL, ...) {
  nstop(db)
  stopifnot(length(db) == 1)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "tax_rank(x)", with = "tax_rank(sci_id)")
    sci_id <- x
  }
  
  switch(
    db,
    bold = stats::setNames(tax_rank_(process_ids(sci_id, db, get_boldid,
      rows = rows), ...), sci_id),
    eol = stats::setNames(tax_rank_(process_ids(sci_id, db, get_eolid,
      rows = rows), ...), sci_id),
    gbif = stats::setNames(tax_rank_(process_ids(sci_id, db, get_gbifid,
      rows = rows), ...), sci_id),
    natserv = stats::setNames(tax_rank_(process_ids(sci_id, db, get_natservid,
      rows = rows), ...), sci_id),
    nbn = stats::setNames(tax_rank_(process_ids(sci_id, db, get_nbnid,
      rows = rows), ...), sci_id),
    tol = stats::setNames(tax_rank_(process_ids(sci_id, db, get_tolid,
      rows = rows), ...), sci_id),
    tropicos = stats::setNames(tax_rank_(process_ids(sci_id, db, get_tpsid,
      rows = rows), ...), sci_id),
    itis = stats::setNames(tax_rank_(process_ids(sci_id, db, get_tsn,
      rows = rows), ...), sci_id),
    ncbi = stats::setNames(tax_rank_(process_ids(sci_id, db, get_uid,
      rows = rows), ...), sci_id),
    worms = stats::setNames(tax_rank_(process_ids(sci_id, db, get_wormsid,
      rows = rows), ...), sci_id),
    stop("the provided db value was not recognised", call. = FALSE)
  )
}

#' @export
tax_rank.numeric <- function(sci_id, db = NULL, rows = NA, ...) {
  tax_rank(as.character(sci_id), db, rows, ...)
}

# ---------
tax_rank_ <- function(id, ...) {
  fun <- function(x, clz, ...) {
    res <- classification(x, db = clz, ...)
    if (all(is.na(res))) {
      NA_character_
    } else {
      if (NROW(res[[1]]) > 0) {
        tt <- res[[1]]
        out <- tt[tt$id == x, "rank"][[1]]
        if (length(out) == 0) NA_character_ else tolower(out)
      } else {
        NA_character_
      }
    }
  }
  lapply(id, fun, clz = dbswap(class(id)), ...)
}

context("tax_rank")

test_that("tax_rank returns the correct class", {
  skip_on_cran() # uses secrets
  # vcr::use_cassette("tax_rank", {
  #   A <- suppressMessages(tax_rank(c("Helianthus annuus", "Baetis"),
  #     db = "ncbi", messages = FALSE))
  #   B <- suppressMessages(sw(tax_rank("Helianthus", db = "itis",
  #     messages = FALSE, rows = 1)))
  #   C <- suppressMessages(sw(tax_rank(c("Helianthus annuus", "xxxxxx"),
  #     db = "ncbi", messages = FALSE)))
  # }, preserve_exact_body_bytes = TRUE, match_requests_on = c("method", "query"))
  A <- suppressMessages(tax_rank(c("Helianthus annuus", "Baetis"),
                                 db = "ncbi", messages = FALSE))
  B <- suppressMessages(sw(tax_rank("Helianthus", db = "itis",
                                    messages = FALSE, rows = 1)))
  C <- suppressMessages(sw(tax_rank(c("Helianthus annuus", "xxxxxx"),
                                    db = "ncbi", messages = FALSE)))
  
  expect_is(A, "list")
  expect_is(B, "list")
  expect_is(C, "list")
  expect_equal(names(C), c("Helianthus annuus", "xxxxxx"))

  expect_equal(A$`Helianthus annuus`, "species")
  expect_true(is.na(C$xxxxxx))

  expect_equal(length(A), 2)
  expect_equal(length(C), 2)
})

test_that("works with get_*() input", {
  skip_on_cran()
  vcr::use_cassette("tax_rank_get_star_input", {
    aa <- suppressMessages(tax_rank(get_boldid("Helianthus annuus")))
  })

  expect_is(aa, "list")
  expect_equal(names(aa), "421377")
  expect_equal(aa$`421377`, "species")
})

test_that("tax_rank fails well", {
  skip_on_cran()

  expect_error(tax_rank(), "argument \"sci_id\" is missing")
  expect_error(tax_rank("aadfd"), "Must specify db!")
  expect_error(tax_rank("Asdfadsf", db = "asdfd"),
               "the provided db value was not recognised")
  expect_error(tax_rank(NA, db = "itis"),
               "'db' not recognized")
})

#' Get scientific names from common names.
#'
#' @export
#' @param com One or more common names or partial names.
#' @param db Data source, one of *"ncbi"* (default), *"itis"*,
#' *"tropicos"*, *"eol"*, or *"worms"*. If using ncbi, we
#' recommend getting an API key; see [taxize-authentication]
#' @param itisby Search for common names across entire names (search, default),
#' at beginning of names (begin), or at end of names (end).
#' @param simplify (logical) If `TRUE`, simplify output to a vector of names.
#' If `FALSE`, return variable formats from different sources, usually a
#' data.frame.
#' @param commnames Deprecated, see `com`
#' @param ... Further arguments passed on to internal methods.
#' @param id taxon identifiers, as returned by [get_tsn()] or [get_uid()]
#' @return If `simplify=TRUE`, a list of scientific names, with list
#' labeled by your input names. If `simplify=FALSE`, a data.frame with
#' columns that vary by data source. `character(0)` on no match
#' @seealso [sci2comm()]
#' @details For data sources ITIS and NCBI you can pass in common names
#' directly, and use [get_uid()] or [get_tsn()] to get ids first, then pass in
#' to this fxn.
#'
#' For the other data sources, you can only pass in common names directly.
#'
#' @section Authentication:
#' See [taxize-authentication] for help on authentication
#' 
#' @section HTTP version for NCBI requests:
#' We hard code `http_version = 2L` to use HTTP/1.1 in HTTP requests to
#' the Entrez API. See `curl::curl_symbols('CURL_HTTP_VERSION')` 
#' 
#' @section Rate limits:
#' In case you run into errors due to your rate limit being exceeded, see
#' [taxize_options()], where you can set `ncbi_sleep`.
#'
#' @author Scott Chamberlain
#' @examples \dontrun{
#' comm2sci(com='american black bear')
#' comm2sci(com='american black bear', simplify = FALSE)
#' comm2sci(com='black bear', db='itis')
#' comm2sci(com='american black bear', db='itis')
#' comm2sci(com='annual blue grass', db='tropicos')
#' comm2sci(com=c('annual blue grass','tree of heaven'), db='tropicos')
#' comm2sci('blue whale', db = "worms")
#' comm2sci(c('blue whale', 'dwarf surfclam'), db = "worms")
#'
#' # ncbi: pass in uid's from get_uid() directly
#' x <- get_uid("western capercaillie", modifier = "Common Name")
#' comm2sci(x)
#' # itis: pass in tsn's from get_tsn() directly
#' x <- get_tsn(c("Louisiana black bear", "american crow"),
#'   searchtype = "common")
#' comm2sci(x)
#' }
comm2sci <- function(...) {
  UseMethod("comm2sci")
}


#' @method comm2sci default
#' @export
#' @rdname comm2sci
comm2sci.default <- function(com, db='ncbi', itisby='search',
                             simplify=TRUE, commnames = NULL, ...) {
  if (!is.null(commnames)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "comm2sci(commnames)", with = "comm2sci(com)")
    com <- commnames
  }
  
  assert(com, "character")
  assert(simplify, "logical")
  temp <- lapply(com, sci_from_comm, db = db, simplify = simplify,
                 itisby = itisby, ...)
  stats::setNames(temp, com)
}

sci_from_comm <- function(nn, db, simplify, itisby, ...) {
  switch(
    db,
    eol = c2s_eol(sci = nn, simplify, ...),
    itis = c2s_itis(nn, itisby, simplify, ...),
    tropicos = c2s_tp(simplify, com = nn, ...),
    ncbi = {
      ids <- get_uid(nn, modifier = "Common Name", ...)
      c2s_ncbi(ids, ...)
    },
    worms = c2s_worms(nn, simplify, ...),
    stop("'db' must be one of 'ncbi', 'itis', 'tropicos', 'eol', 'worms'",
         call. = FALSE)
  )
}

#' @export
#' @rdname comm2sci
comm2sci.tsn <- function(id, db='ncbi', itisby='search',
                         simplify=TRUE, ...) {
  warn_db(list(db = db), "itis")
  temp <- lapply(id, c2s_itis_, simplify = simplify, ...)
  stats::setNames(temp, id)
}

#' @export
#' @rdname comm2sci
comm2sci.uid <- function(id, db='ncbi', itisby='search',
                         simplify=TRUE, ...) {
  warn_db(list(db = db), "ncbi")
  temp <- lapply(id, c2s_ncbi, simplify = simplify, ...)
  stats::setNames(temp, id)
}

# helpers ------------
c2s_itis <- function(x, by='search', simplify, ...){
  tmp <- switch(
    by,
    search = ritis::search_common(x, ...),
    begin = ritis::search_common(x, from = "begin", ...),
    end = ritis::search_common(x, from = "end", ...)
  )
  # remove empty tsn slots
  tsns <- as.character(suppressWarnings(tmp$tsn))
  tsns <- tsns[!sapply(tsns, nchar, keepNA = FALSE) == 0]
  # get scientific names
  tmp <- data.table::setDF(
    data.table::rbindlist(
      lapply(tsns, ritis::scientific_name),
      fill = TRUE, use.names = TRUE
    )
  )
  if (simplify) {
    as.character(tmp$combinedname)
  } else{
    tmp
  }
}

c2s_itis_ <- function(x, by='search', simplify, ...){
  tmp <- data.table::setDF(
    data.table::rbindlist(
      lapply(x, ritis::scientific_name),
      fill = TRUE, use.names = TRUE
    )
  )
  if (simplify) {
    as.character(tmp$combinedname)
  } else {
    tmp
  }
}

c2s_ncbi <- function(x, simplify, ...) {
  key <- getkey(NULL, "ENTREZ_KEY")
  query <- tc(list(db = "taxonomy", ID = x, api_key = key))
  cli <- crul::HttpClient$new(url = ncbi_base(),
    headers = tx_ual, opts = list(http_version = 2L, ...))
  res <- cli$get("entrez/eutils/efetch.fcgi", query = query)
  if (!res$success()) return(character())
  tt <- res$parse("UTF-8")
  ttp <- xml2::read_xml(tt)
  # common name
  out <- xml_text(xml_find_all(ttp, "//TaxaSet/Taxon/ScientificName"))
  # NCBI limits requests to three per second
  ncbi_rate_limit_pause(key)
  return(out)
}

c2s_eol <- function(simplify, ...){
  tmp <- eol_search(...)
  if (all(is.na(tmp))) return(character(0))
  if (simplify) {
    as.character(tmp$name)
  } else {
    tmp
  }
}

c2s_tp <- function(simplify, ...){
  tmp <- tp_search(...)
  if (simplify) {
    as.character(tmp$scientificname)
  } else{
    tmp
  }
}

c2s_worms <- function(x, simplify, ...){
  tmp <- try_df(worrms::wm_records_common(name = x, ...))
  if (simplify) {
    as.character(tmp$scientificname)
  } else{
    tmp
  }
}

#' Get the BOLD (Barcode of Life) code for a search term.
#'
#' @export
#' @param sci character; A vector of scientific names. Or,
#' a `taxon_state` object (see [taxon-state])
#' @param fuzzy (logical) Whether to use fuzzy search or not (default: FALSE).
#' @param dataTypes (character) Specifies the datatypes that will be returned.
#' See [bold_search()] for options.
#' @param includeTree (logical) If TRUE (default: FALSE), returns a list
#' containing information for parent taxa as well as the specified taxon.
#' @param ask logical; should get_tsn be run in interactive mode?
#' If TRUE and more than one TSN is found for teh species, the user is asked for
#' input. If FALSE NA is returned for multiple matches.
#' @param messages logical; should progress be printed?
#' @param x Input to [`as.boldid()`]
#' @param ... Curl options passed on to [`crul::verb-GET`]
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all rows are
#' considered. Note that this function still only gives back a boldid class object with one
#' to many identifiers. See [get_boldid_()] to get back all, or a subset,
#' of the raw data that you are presented during the ask process.
#' @param division (character) A division (aka phylum) name. Optional. See `Filtering`
#' below.
#' @param parent (character) A parent name (i.e., the parent of the target search
#' taxon). Optional. See `Filtering` below.
#' @param rank (character) A taxonomic rank name. See [rank_ref] for possible
#' options. Though note that some data sources use atypical ranks, so inspect the
#' data itself for options. Optional. See `Filtering` below.
#' @param check logical; Check if ID matches any existing on the DB, only used in
#' [as.boldid()]
#' @param searchterm Deprecated, see `sci`
#' @template getreturn
#'
#' @section Filtering:
#' The parameters `division`, `parent`, and `rank` are not used in the search to
#' the data provider, but are used in filtering the data down to a subset that
#' is closer to the target you want.  For all these parameters, you can use
#' regex strings since we use [grep()] internally to match. Filtering narrows
#' down to the set that matches your query, and removes the rest.
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @examples \dontrun{
#' get_boldid(sci = "Agapostemon")
#' get_boldid(sci = "Chironomus riparius")
#' get_boldid(c("Chironomus riparius","Quercus douglasii"))
#' splist <- names_list('species')
#' get_boldid(splist, messages=FALSE)
#'
#' # Fuzzy searching
#' get_boldid(sci="Osmi", fuzzy=TRUE)
#'
#' # Get back a subset
#' get_boldid(sci="Osmi", fuzzy=TRUE, rows = 1)
#' get_boldid(sci="Osmi", fuzzy=TRUE, rows = 1:10)
#' get_boldid(sci=c("Osmi","Aga"), fuzzy=TRUE, rows = 1)
#' get_boldid(sci=c("Osmi","Aga"), fuzzy=TRUE, rows = 1:3)
#'
#' # found
#' get_boldid('Epicordulia princeps')
#' get_boldid('Arigomphus furcifer')
#'
#' # When not found
#' get_boldid("howdy")
#' get_boldid(c("Chironomus riparius", "howdy"))
#' get_boldid("Cordulegaster erronea")
#' get_boldid("Nasiaeshna pentacantha")
#'
#' # Narrow down results to a division or rank, or both
#' ## Satyrium example
#' ### Results w/o narrowing
#' get_boldid("Satyrium")
#' ### w/ phylum
#' get_boldid("Satyrium", division = "Plantae")
#' get_boldid("Satyrium", division = "Animalia")
#'
#' ## Rank example
#' get_boldid("Osmia", fuzzy = TRUE)
#' get_boldid("Osmia", fuzzy = TRUE, rank = "genus")
#'
#' # Fuzzy filter on any filtering fields
#' ## uses grep on the inside
#' get_boldid("Satyrium", division = "anim")
#' get_boldid("Aga", fuzzy = TRUE, parent = "*idae")
#'
#' # Convert a boldid without class information to a boldid class
#' as.boldid(get_boldid("Agapostemon")) # already a boldid, returns the same
#' as.boldid(get_boldid(c("Agapostemon","Quercus douglasii"))) # same
#' as.boldid(1973) # numeric
#' as.boldid(c(1973,101009,98597)) # numeric vector, length > 1
#' as.boldid("1973") # character
#' as.boldid(c("1973","101009","98597")) # character vector, length > 1
#' as.boldid(list("1973","101009","98597")) # list, either numeric or character
#' ## dont check, much faster
#' as.boldid("1973", check=FALSE)
#' as.boldid(1973, check=FALSE)
#' as.boldid(c("1973","101009","98597"), check=FALSE)
#' as.boldid(list("1973","101009","98597"), check=FALSE)
#'
#' (out <- as.boldid(c(1973,101009,98597)))
#' data.frame(out)
#' as.boldid( data.frame(out) )
#'
#' # Get all data back
#' get_boldid_("Osmia", fuzzy=TRUE, rows=1:5)
#' get_boldid_("Osmia", fuzzy=TRUE, rows=1)
#' get_boldid_(c("Osmi","Aga"), fuzzy=TRUE, rows = 1:3)
#' }

get_boldid <- function(sci, fuzzy = FALSE, dataTypes = 'basic',
                       includeTree = FALSE, ask = TRUE, messages = TRUE,
                       rows = NA, rank = NULL, division = NULL,
                       parent = NULL, searchterm = NULL, ...) {

  assert(sci, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(messages, "logical")
  assert(fuzzy, "logical")
  assert(dataTypes, "character")
  assert(includeTree, "logical")
  assert(rank, "character")
  assert(division, "character")
  assert(parent, "character")
  assert_rows(rows)
  if (!is.null(searchterm)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_boldid(searchterm)", with = "get_boldid(sci)")
    sci <- searchterm
  }
  

  if (inherits(sci, "character")) {
    tstate <- taxon_state$new(class = "boldid", names = sci)
    items <- sci
  } else {
    assert_state(sci, "boldid")
    tstate <- sci
    sci <- tstate$taxa_remaining()
    items <- c(sci, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci[i], "'\n")
    bold_df <- bold_search(sci = sci[i], fuzzy = fuzzy,
                           dataTypes = dataTypes,
                           includeTree = includeTree, ...)
    mm <- NROW(bold_df) > 1

    if (!is.data.frame(bold_df)) {
      boldid <- NA_character_
      att <- "not found"
    } else {
      if (all(names(bold_df) == "input")) {
        boldid <- NA_character_
        att <- "not found"
      } else {
        bold_df <- rename(bold_df, replace = c('parentname' = 'parent',
          'tax_rank' = 'rank', 'tax_division' = 'division'),
          warn_missing = FALSE
        )
        cols <- c("taxid","taxon","rank","division","parentid","parent")
        to_add <- setdiff(cols, names(bold_df))
        if (length(to_add) > 0) {
          for (i in seq_along(to_add)) {
            bold_df <- cbind(bold_df, setNames(list(NA_character_), to_add[i]))
          }
        }

        bold_df <- bold_df[, c("taxid","taxon","rank","division","parentid","parent")]

        # should return NA if spec not found
        if (nrow(bold_df) == 0) {
          mssg(messages, m_not_found_sp_altclass)
          boldid <- NA_character_
          att <- 'not found'
        }
        # take the one tsn from data.frame
        if (nrow(bold_df) == 1) {
          boldid <- bold_df$taxid
          att <- 'found'
        }
        # check for direct match
        if (nrow(bold_df) > 1) {
          names(bold_df)[grep('taxon', names(bold_df))] <- "target"
          di_rect <- match(tolower(bold_df$target), tolower(sci[i]))
          if (length(di_rect) == 1) {
            if (!all(is.na(di_rect))) {
              boldid <- bold_df$taxid[!is.na(di_rect)]
              direct <- TRUE
              att <- 'found'
            } else {
              boldid <- NA_character_
              att <- 'not found'
            }
          } else {
            direct <- FALSE
            boldid <- NA_character_
            att <- 'found'
          }
        }
        # multiple matches
        if (any(
          nrow(bold_df) > 1 && is.na(boldid) |
          nrow(bold_df) > 1 && att == "found" && length(boldid) > 1
        )) {
          names(bold_df)[grep('^taxon$', names(bold_df))] <- "target"

          if (!is.null(division) || !is.null(parent) || !is.null(rank)) {
            bold_df <- filt(bold_df, "division", division)
            bold_df <- filt(bold_df, "parent", parent)
            bold_df <- filt(bold_df, "rank", rank)

            if (NROW(bold_df) == 0) {
              boldid <- NA_character_
              att <- 'not found'
              warning("filters 'division', 'parent', or 'rank' gave no records\n",
                "  check spelling or try a different filter value")
            }
          }

          if (NROW(bold_df) > 0) {
            bold_df <- sub_rows(bold_df, rows)
            boldid <- id <- bold_df$taxid
            if (length(id) == 1) {
              direct <- TRUE
              att <- "found"
            }
          }

          if (ask) {
            # user prompt
            if (NROW(bold_df) > 0) {
              bold_df <- bold_df[order(bold_df$target), ]
              rownames(bold_df) <- 1:nrow(bold_df)
            }
            if (length(boldid) > 1 || NROW(bold_df) > 1) {
              # prompt
              message("\n\n")
              message(paste0(utils::capture.output(bold_df), collapse = "\n"))
              message("\nMore than one TSN found for taxon '", sci[i], "'!\n
            Enter rownumber of taxon (other inputs will return 'NA'):\n") # prompt
              take <- scan(n = 1, quiet = TRUE, what = 'raw')

              if (length(take) == 0)
                take <- 'notake'
              if (take %in% seq_len(nrow(bold_df))) {
                take <- as.numeric(take)
                message("Input accepted, took taxon '",
                  as.character(bold_df$target[take]), "'.\n")
                boldid <-  bold_df$taxid[take]
                att <- "found"
              } else {
                boldid <- NA_character_
                mssg(messages, "\nReturned 'NA'!\n\n")
                att <- "not found"
              }
            }
          } else {
            if (length(boldid) == 1) {
              att <- "found"
            } else {
              warning(sprintf(m_more_than_one_found, "boldid", sci[i]),
                call. = FALSE)
              boldid <- NA_character_
              att <- m_na_ask_false
            }
          }
        }
      }
    }
    res <- list(id = as.character(boldid), att = att, multiple = mm,
      direct = direct)
    prog$completed(sci[i], att)
    prog$prog(att)
    tstate$add(sci[i], res)
  }
  out <- tstate$get()
  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "boldid",
                   match = pluck_un(out, "att", ""),
                   multiple_matches = pluck_un(out, "multiple", logical(1)),
                   pattern_match = pluck_un(out, "direct", logical(1)))
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$bold)
}

#' @export
#' @rdname get_boldid
as.boldid <- function(x, check=TRUE) UseMethod("as.boldid")

#' @export
#' @rdname get_boldid
as.boldid.boldid <- function(x, check=TRUE) x

#' @export
#' @rdname get_boldid
as.boldid.character <- function(x, check=TRUE) {
  if (length(x) == 1) 
    make_boldid(x, check) 
  else 
    collapse(x, make_boldid, "boldid", check=check)
}

#' @export
#' @rdname get_boldid
as.boldid.list <- function(x, check=TRUE) {
  if (length(x) == 1) 
    make_boldid(x, check) 
  else 
    collapse(x, make_boldid, "boldid", check=check)
}

#' @export
#' @rdname get_boldid
as.boldid.numeric <- function(x, check=TRUE) as.boldid(as.character(x), check)

#' @export
#' @rdname get_boldid
as.boldid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class="boldid", match=x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri=x$uri)
}

#' @export
#' @rdname get_boldid
as.data.frame.boldid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "boldid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_boldid <- function(x, check=TRUE) {
  make_generic(x,
    'http://boldsystems.org/index.php/Taxbrowser_Taxonpage?taxid=%s',
    "boldid", check)
}

check_boldid <- function(x){
  tryid <- bold_tax_id(x)
  !identical("noresults", names(tryid)[2])
}

#' @export
#' @rdname get_boldid
get_boldid_ <- function(sci, messages = TRUE, fuzzy = FALSE,
  dataTypes='basic', includeTree=FALSE, rows = NA, searchterm = NULL, ...) {

  if (!is.null(searchterm)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_boldid_(searchterm)", with = "get_boldid_(sci)")
    sci <- searchterm
  }
  stats::setNames(lapply(sci, get_boldid_help, messages = messages,
    fuzzy = fuzzy, dataTypes=dataTypes, includeTree=includeTree,
    rows = rows, ...), sci)
}

get_boldid_help <- function(sci, messages, fuzzy, dataTypes,
  includeTree, rows, ...){
  
  mssg(messages, "\nRetrieving data for taxon '", sci, "'\n")
  df <- bold_search(name = sci, fuzzy = fuzzy, dataTypes = dataTypes,
    includeTree = includeTree)
  if(NROW(df) == 0) NULL else sub_rows(df, rows)
}

# tests for get_boldid fxn in taxize
context("get_boldid")

test_that("get_boldid returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("get_boldid", {
    expect_equal(get_boldid(sci = 'Helianthus', messages = FALSE)[[1]], '125295')
    expect_true(is.na(get_boldid(sci='adsf asdf asdf', messages=FALSE)[[1]]))
    expect_is(get_boldid(c("Helianthus excubitor", "adsf asdf asdf"), messages=FALSE), 
      "boldid")
  })
})

test_that("get_boldid accepts ask-argument", {
  skip_on_cran()
  vcr::use_cassette("get_boldid_ask_false", {
    expect_true(is.na(
      get_boldid('adsf asdf asdf', ask=FALSE, messages=FALSE)[[1]]))
  })
})

test_that("get_boldid works when there's no parent name", {
  skip_on_cran()
  vcr::use_cassette("get_boldid_no_parent_name", {
    x <- get_boldid("Chordata", messages = FALSE)
  })

  expect_is(x, "boldid")
  expect_equal(x[1], "18")
})

test_that("get_boldid fails as expected", {
  skip_on_cran()

  expect_error(get_boldid(), "argument \"sci\" is missing")
  expect_error(get_boldid("Satyrium", ask = 4, messages = FALSE),
               "ask must be of class logical")
  expect_error(
    get_boldid("Osmi", fuzzy=4, messages = FALSE),
    "fuzzy must be of class logical")
  expect_error(
    get_boldid("Osmi", dataTypes = 4, messages = FALSE),
    "dataTypes must be of class character")
  expect_error(
    get_boldid("Osmi", includeTree = 4, messages = FALSE),
    "includeTree must be of class logical")
  expect_error(
    get_boldid("Osmi", rank = 4, messages = FALSE),
    "rank must be of class character")
  expect_error(
    get_boldid("Osmi", division = 4, messages = FALSE),
    "division must be of class character")
  expect_error(
    get_boldid("Osmi", parent = 4, messages = FALSE),
    "parent must be of class character")

  # rows param
  expect_error(get_boldid("Achlya", rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_boldid("Achlya", rows = 0, messages = FALSE),
               "all\\(rows > 0\\) is not TRUE")

  # skip('BOLD API down right now, SSL issue')
  # filter param doesn't match any values - returns NA and warns
  expect_warning(
    (z=get_boldid("Satyrium", division = "Plants", messages = FALSE)),
    "check spelling")
  expect_equal(z[1], NA_character_)
})

#' Search UK National Biodiversity Network
#'
#' @export
#' @param sci_com (character) The query terms(s), a scientific or common name
#' @param fq (character) Filters to be applied to the original query. These
#' are additional params of the form fq=INDEXEDFIELD:VALUE e.g.
#' fq=rank:kingdom. See https://species-ws.nbnatlas.org/indexFields for all
#' the fields that are queryable.
#' @param rows (integer) Number of records to return
#' @param start (integer) Record offset, to enable paging
#' @param sort (character) The indexed field to sort by
#' @param order (character) Supports "asc" or "desc"
#' @param facets (list) Comma separated list of the fields to create facets
#' on e.g. facets=basis_of_record.
#' @param q Deprecated, see `sci`
#' @param ... Further args passed on to [crul::HttpClient].
#' @family nbn
#' @return a list with slots for metadata (`meta`) with list of response
#' attributes, and data (`data`) with a data.frame of results
#' @author Scott Chamberlain, 
#' @references https://api.nbnatlas.org/
#'
#' @examples \dontrun{
#' x <- nbn_search(sci_com = "Vulpes")
#' x$meta$totalRecords
#' x$meta$pageSize
#' x$meta$urlParameters
#' x$meta$queryTitle
#' head(x$data)
#'
#' nbn_search(sci_com = "blackbird", start = 4)
#'
#' # debug curl stuff
#' nbn_search(sci_com = "blackbird", verbose = TRUE)
#' }
nbn_search <- function(sci_com, fq = NULL, order = NULL, sort = NULL,
  start = 0, rows = 25, facets = NULL, q = NULL, ...) {

  if (!is.null(q)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "nbn_search(q)", with = "nbn_search(sci_com)")
    sci_com <- q
  }
  
  args <- tc(list(
    q = sci_com, fq = fq, pageSize = rows, startIndex = start, sort = sort,
    dir = order, facets = facets
  ))
  nbn_GET(file.path(nbn_base(), "search"), args, ...)
}

nbn_GET <- function(url, args, ...){
  cli <- crul::HttpClient$new(url = url, headers = tx_ual,)
  tt <- cli$get(query = argsnull(args), ...)
  tt$raise_for_status()
  json <- jsonlite::fromJSON(tt$parse("UTF-8"))$searchResults
  list(meta = pop(json, "results"), data = json$results)
}

nbn_base <- function() "https://species-ws.nbnatlas.org"

# get name from TOL ID
# tol_id2name(id = 515698)
tol_id2name <- function(id, ...) {
  cli <- crul::HttpClient$new("https://api.opentreeoflife.org", 
    headers = tx_ual, opts = list(...))
  res <- cli$post(path = "v3/taxonomy/taxon_info", encode = "json",
    body = list(ott_id = id))
  if (res$status_code > 201) {
    warning(res$status_code, ": ", res$status_http()$message, 
      ", returning empty data.frame for ", id, 
      call. = FALSE)
    return(id2name_blanks$tol)
  }
  tmp <- jsonlite::fromJSON(res$parse("UTF-8"))
  tmp[vapply(tmp, length, 1) == 0] <- NULL
  if ("tax_sources" %in% names(tmp)) {
    if (length(tmp$tax_sources) > 1) {
      ids <- lapply(tmp$tax_sources, function(z) {
        m <- strsplit(z, ":")[[1]]
        as.list(stats::setNames(m[2], paste0("tax_sources_", m[1])))
      })
      tmp$tax_sources <- NULL
      tmp <- c(tmp, unlist(ids, FALSE))
    }
  }
  tmp$unique_name <- NULL
  tmp$is_suppressed <- NULL
  names(tmp)[names(tmp) %in% "ott_id"] <- "id"
  tmp <- tmp[c('id', 'name', 'rank', grep("tax_", names(tmp), value = TRUE))]
  data.frame(tmp, stringsAsFactors = FALSE)
}

col_defunct_mssg = "COL API is no longer reliable; all COL fxns removed"

#' COL defunct functions
#' 
#' COL introduced rate limiting recently in 2019 - which has made the API
#' essentially unusable - COL+ is coming soon and we'll incorporate it here
#' when it's stable
#'
#' @name col-defunct
#' @export
#' @keywords internal 
#' @param ... ignored
as.colid <- function(...) .Defunct(msg = col_defunct_mssg)

#' @export
#' @rdname col-defunct
col_children <- function(...) .Defunct(msg = col_defunct_mssg)

#' @export
#' @rdname col-defunct
col_classification <- function(...) .Defunct(msg = col_defunct_mssg)

#' @export
#' @rdname col-defunct
col_downstream <- function(...) .Defunct(msg = col_defunct_mssg)

#' @export
#' @rdname col-defunct
col_search <- function(...) .Defunct(msg = col_defunct_mssg)

#' @export
#' @rdname col-defunct
get_colid <- function(...) .Defunct(msg = col_defunct_mssg)

#' @export
#' @rdname col-defunct
get_colid_ <- function(...) .Defunct(msg = col_defunct_mssg)

taxon_state_env <- new.env()

#' Last taxon state object from a `get_*` function call
#' 
#' @export
#' @name taxon-state
#' @details 
#' 
#' - `taxon_last()`: get the last `taxon_state` object in use
#' - `taxon_clear()`: clear any data from last `taxon_state` object
#' 
#' The `taxon_state` object is an R6 object that holds data and methods 
#' used for keeping track of results gathered within a `get_*` function.
#' You shouldn't create `taxon_state` R6 objects yourself.
#' 
#' Behaviors to be aware of:
#' 
#' - If a `taxon_state` object is not passed you don't need to worry about 
#' a previously run `get_*` function interfering with another `get_*` 
#' function call - you have to explicitly pass a `taxon_state` object
#' to use `taxon_state`
#' - The passed in `taxon_state` object must have a `$class` matching that of
#' the `get_*` function being called. For example, you can only pass a
#' `taxon_state` with `$class` of `gbifid` to `get_gbifid()`, and so on.
#' - If you run `taxon_clear()` while a `get*` function is running, you may 
#' lose track of any state known to this package before it was cleared
#' 
#' See the internal method [progressor] for information on how we control messages
#' in `get*` functions
#' 
#' @return `taxon_last()` returns an object of class `taxon_state`, the last
#' one used, else `NULL` if none found. `taxon_clear()` clears the saved state
#' 
#' @examples \dontrun{
#' spp <- names_list("species", 3)
#' res <- get_gbifid(spp)
#' z <- taxon_last()
#' z
#' z$taxa_remaining()
#' z$taxa_completed()
#' z$count # active binding; no parens needed
#' 
#' # cleanup
#' taxon_clear()
#' }
taxon_last <- function() taxon_state_env$last

#' @export
#' @rdname taxon-state
taxon_clear <- function() taxon_state_env$last <- NULL

#' Keep track of queries in `get_*` functions
#'
#' This object lives inside each `get_*` function call, maintaining
#' results as they are accumulated.
#'
#' @keywords internal
#' @examples \dontrun{
#' if (interactive()) {
#' ts <- taxon_state$new()
#' taxon_last()
#' ts
#' res <- list(
#'   id = 123456,
#'   att = "found",
#'   multiple = FALSE,
#'   direct = FALSE,
#'   class = "tsn"
#' )
#' ts$add(query = "Quercus robur", result = res)
#' ts
#' ts$get(query = "Quercus robur")
#' ts$count
#' ts$remove(query = "Quercus robur")
#' ts
#' ts$count
#'
#' res2 <- list(
#'   id = 3430834535,
#'   att = "found",
#'   multiple = FALSE,
#'   direct = FALSE,
#'   class = "gbifid"
#' )
#' ts$add(query = "Poa annua", result = res2)
#' res3 <- list(
#'   id = 1223424,
#'   att = "found",
#'   multiple = FALSE,
#'   direct = FALSE,
#'   class = "uid"
#' )
#' ts$add(query = "Puma concolor", result = res3)
#' ts
#' ts$count
#' ts$get("Puma concolor")
#' ts$get()
#'
#' # cleanup
#' ts$purge()
#' ts$count
#' }
#' }
taxon_state <- R6::R6Class(
  "taxon_state",
  public = list(
    #' @field initialized (time) time job started
    initialized = NULL,
    #' @field finalized (time) time job finished
    finalized = NULL,
    #' @field class (character) a class name (e.g., "gbif")
    class = NULL,
    #' @field names (character) one or more taxon names
    names = NULL,

    #' @description Create a new `taxon_state` object
    #' @param class (character) a class name (e.g., "gbif")
    #' @param names (character) one or more taxon names
    #' @return A new `taxon_state` object
    initialize = function(class, names) {
      taxon_state_env$last <- self
      self$initialized <- Sys.time()
      if (!missing(class)) self$class <- class
      if (!missing(names)) self$names <- names
    },

    #' @description print method for the `taxon_state` class
    #' @param x self
    #' @param ... ignored
    print = function(x, ...) {
      cat("<taxon state> ", sep = "\n")
      cat(paste0(" class: ", self$class %||% "none"), sep = "\n")
      if (!is.null(self$finalized)) {
        cat(paste0(" elapsed (sec): ",
          round(self$finalized - self$initialized, 2) %||% ""),
          sep = "\n")
      } else {
        cat(" elapsed (sec): 0", sep = "\n")
      }
      cat(paste0(" count: ", self$count %||% 0), sep = "\n")
      if (length(private$pool) > 0) {
        for (i in seq_along(private$pool)) {
          cat(sprintf("  %s: %s",
            names(private$pool)[i],
            private$pool[[i]]$id), sep = "\n")
        }
      } else {
        cat("  none ")
      }
      invisible(self)
    },
    
    #' @description add a record with it's result; duplicates allowed
    #' @param query (character), a taxon name
    #' @param result (list) a named list
    #' @return nothing returned; sets only
    add = function(query, result) {
      assert(query, "character")
      assert(result, "list")
      private$pool <- c(private$pool, stats::setNames(list(result), query))
    },
    #' @description get all records matching 'query'
    #' @param query (character), a taxon name
    #' @return a named list, with slots for the taxon id, and other attributes,
    #' named by the taxon name
    get = function(query = NULL) {
      if (is.null(query)) return(private$pool)
      private$pool[names(private$pool) %in% query]
    },
    #' @description remove's all records matching 'query'
    #' @param query (character), a taxon name
    #' @return nothing, removes records matching query
    remove = function(query) {
      private$pool[names(private$pool) %in% query] <- NULL
    },
    #' @description removes all records
    #' @return nothing returned; sets only
    purge = function() private$pool <- NULL,
    #' @description get remaining taxa
    #' @return sorted taxon names
    taxa_remaining = function() {
      done <- names(self$get())
      sort(self$names)[!sort(self$names) %in% sort(done)]
    },
    #' @description get completed taxa
    #' @return sorted taxon names
    taxa_completed = function() {
      sort(names(self$get()))
    }
  ),
  active = list(
    #' @field count (integer) count number of records
    count = function() length(private$pool),
    #' @field exit record date/time function exited
    exit = function() self$finalized <- Sys.time()
  ),
  private = list(
    pool = list()
  )
)

context("get_gbifid")

test_that("get_gbifid returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("get_gbifid", {
    z <- get_gbifid(c("Chironomus riparius", "aaasdfadsfasdf"), 
      messages = FALSE)[2]
    w <- get_gbifid(c("Chironomus riparius", "Chaetopteryx"), 
      messages = FALSE)
  }, preserve_exact_body_bytes = TRUE)

  expect_true(is.na(z))
  expect_is(w, "gbifid")
})

test_that("get_gbifid accepts ask-argument", {
  skip_on_cran()
  vcr::use_cassette("get_gbifid_ask_arg", {
    a <- sw(get_gbifid('Dugesia', ask = FALSE, messages = FALSE))
  })
  expect_true(is.na(a))
})

test_that("get_gbifid method parameter works", {
  skip_on_cran()
  vcr::use_cassette("get_gbifid_method_param", {
    mod1 <- get_gbifid_(sci = "Z*", method = "backbone", messages = FALSE, 
      rows = 1:100)
    mod2 <- get_gbifid_("Z*", method = "lookup", messages = FALSE, 
      rows = 1:100)
  }, preserve_exact_body_bytes = TRUE)

  expect_is(mod1, "list")
  expect_is(mod2, "list")
  expect_gt(NROW(mod2[[1]]), NROW(mod1[[1]]))
})

test_that("get_gbifid phylum/class/order/family parameters work", {
  skip_on_cran()
  vcr::use_cassette("get_gbifid_phylum_param", {
    aa <- get_gbifid("Satyrium", phylum = "Tracheophyta", rows = 1, 
      messages = FALSE)
    bb <- get_gbifid("Satyrium", phylum = "Arthropoda", rows = 1, 
      messages = FALSE)
  })

  expect_is(aa, "gbifid")
  expect_is(bb, "gbifid")
  expect_equal(aa[[1]], "5307264")
  expect_equal(bb[[1]], "1927718")
})

# test_that("get_gbifid rank parameter works", {
#   skip_on_cran()

#   ## Rank example
#   rf1 <- get_gbifid(sci = "Bison", rank = "genus", rows = 1, messages = FALSE)
#   rf2 <- get_gbifid(sci = "Bison bison", rank = "species", rows = 1, messages = FALSE)

#   expect_is(rf1, "gbifid")
#   expect_is(rf2, "gbifid")
#   expect_false(is.na(rf1[[1]]))
#   expect_equal(rf2[[1]], "2441176")
# })

test_that("works regardless of character or numeric GGBIF ID given back", {
  skip_on_cran()
  vcr::use_cassette("get_gbifid_class_doesnt_matter", {
    aa <- get_gbifid("Chironomus riparius", messages = FALSE)
    bb <- get_gbifid("Pinus contorta", messages = FALSE, rows = 1)
  })
  
  expect_is(aa, "gbifid")
  expect_is(aa[[1]], "character")

  expect_is(bb, "gbifid")
  expect_is(bb[[1]], "character")
})

test_that("get_gbifid fails as expected", {
  skip_on_cran()

  expect_error(get_gbifid(), "argument \"sci\" is missing")
  expect_error(get_gbifid('Poa annua', ask = 4, messages = FALSE),
               "ask must be of class logical")
  expect_error(
    get_gbifid("Satyrium", phylum = TRUE, messages = FALSE),
    "phylum must be of class character")
  expect_error(
    get_gbifid("Satyrium", class = TRUE, messages = FALSE),
    "class must be of class character")
  expect_error(
    get_gbifid("Satyrium", order = TRUE, messages = FALSE),
    "order must be of class character")
  expect_error(
    get_gbifid("Satyrium", family = TRUE, messages = FALSE),
    "family must be of class character")
  expect_error(
    get_gbifid("Satyrium", rank = TRUE, messages = FALSE),
    "rank must be of class character")
  expect_error(
    get_gbifid("Satyrium", method = 55, messages = FALSE),
    "method must be of class character")

  # rows param
  expect_error(get_gbifid("Satyrium", rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_gbifid("Satyrium", rows = 0, messages = FALSE),
               "all\\(rows > 0\\) is not TRUE")
})


test_that("get_gbifid works with state input", {
  skip_on_cran()
  
  taxon_clear()
  
  # species list
  # spp <- names_list("species", size = 3)
  spp <- c("Ilex zygophylla", "Astronia shungolensis", "Uromorus anthopophagorum")
  xx <- taxon_last()
  # make sure last taxon is null
  expect_null(xx)

  vcr::use_cassette("get_gbifid_state", {
    res <- get_gbifid(spp, messages = FALSE)
  }, preserve_exact_body_bytes = TRUE)

  # now it's not NULL
  expect_is(taxon_last(), "taxon_state")

  # & get_gbifid output is all good
  expect_is(res, "gbifid")

  # passing taxon_last to get_gbifid: already done, so should be identical
  expect_identical(res, get_gbifid(taxon_last()))
})

context("taxon_state/taxon_last/taxon_clear")

taxon_state_env$last <- NULL

test_that("taxon_last", {
  skip_on_cran()
  skip_if_net_down()

    # before anything happens, it's NULL
  expect_null(taxon_last())

  # something happens
  invisible(taxon_state$new())

  # now it's not NULL
  expect_is(taxon_last(), "taxon_state")

  # in more real context: get_gbifid call
  spp <- c("Spruceanthus marianus", "Bomarea costaricensis",
    "Rubus mindanaensis", "Columnea mira")
  vcr::use_cassette("taxon_state_get_gbifid", {
    res <- get_gbifid(spp, messages = FALSE)
  })
  tl <- taxon_last()
  expect_is(tl, "taxon_state")
  expect_equal(tl$count, 4)
  expect_equal(tl$class, "gbifid")

  # test another get_* fxn
  spp <- c("Spruceanthus marianus", "Bomarea costaricensis",
    "Rubus mindanaensis", "Columnea mira")
  vcr::use_cassette("taxon_state_get_uid", {
    res <- get_uid(spp, messages = FALSE)
  })
  tl <- taxon_last()
  expect_is(tl, "taxon_state")
  expect_equal(tl$count, 4)
  expect_equal(tl$class, "uid")
})

test_that("taxon_state works", {
  skip_on_cran()
  
  ts <- taxon_state$new()

  expect_is(ts$add, "function")
  expect_is(ts$get, "function")
  expect_is(ts$remove, "function")
  expect_is(ts$purge, "function")
  expect_is(ts$count, "integer")
  expect_is(ts$exit, "POSIXct")

  expect_equal(ts$count, 0)

  # add something
  res <- list(
    id = 123456,
    att = "found",
    multiple = FALSE,
    direct = FALSE,
    class = "tsn"
  )
  ts$add(query = "Quercus robur", result = res)
  expect_equal(ts$count, 1)

  bb <- ts$get(query = "Quercus robur")
  expect_is(bb, "list")
  expect_is(bb$`Quercus robur`$id, "numeric")
  expect_is(bb$`Quercus robur`$att, "character")
  expect_is(bb$`Quercus robur`$multiple, "logical")
  expect_is(bb$`Quercus robur`$direct, "logical")
  expect_is(bb$`Quercus robur`$class, "character")

  # adding something of the same name adds another entry
  ts$add(query = "Quercus robur", result = res)
  expect_equal(ts$count, 2)
  expect_equal(length(ts$get(query = "Quercus robur")), 2)
  
  # remove something
  ts$remove(query = "Quercus robur")
  expect_equal(ts$count, 0)

  # add many records
  res2 <- list(
    id = 3430834535,
    att = "found",
    multiple = FALSE,
    direct = FALSE,
    class = "gbifid"
  )
  ts$add(query = "Quercus robur", result = res)
  ts$add(query = "Poa annua", result = res2)
  expect_equal(ts$count, 2)

  # purge all records
  ts$purge()
  expect_equal(ts$count, 0)
})

test_that("taxon_state fails well", {
  expect_error(taxon_state$new(apple = 5), "unused argument")
})

#' Retrieve gene sequences from NCBI by accession number.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @author Scott Chamberlain
#' @rdname ncbi_getbyid-defunct
#' @keywords internal
ncbi_getbyid <- function(...) {
  .Defunct("ncbi_byid", "traits",
    msg = "This function is defunct. See traits::ncbi_byid()")
}

#' Retrieve gene sequences from NCBI by accession number.
#'
#' THIS FUNCTION IS DEFUNCT.
#'
#' @export
#' @keywords internal
#' @rdname get_genes-defunct
get_genes <- function(...) {
  .Defunct("ncbi_getbyid", "taxize", "Function name changed. See ncbi_getbyid")
}

#' Given the identifier for a data object, return all metadata about the object
#'
#' @export
#' @param id (character) The EOL data object identifier
#' @param taxonomy (logical) Whether to return any taxonomy details from
#' different taxon hierarchy providers, in an array named `taxonconcepts`
#' @param language (character) provides the results in the specified language.
#' one of ms, de, en, es, fr, gl, it, nl, nb, oc, pt-BR, sv, tl, mk, sr, uk,
#' ar, zh-Hans, zh-Hant, ko
#' @param ... Curl options passed on to [crul::HttpClient]
#' @details It's possible to return JSON or XML with the EOL API. However,
#' this function only returns JSON for now.
#' @return A list, optionally with a data.frame if `taxonomy=TRUE`
#' @examples \dontrun{
#' eol_dataobjects(id = 7561533)
#'
#' # curl options
#' eol_dataobjects(id = 7561533, verbose = TRUE)
#' }
eol_dataobjects <- function(id, taxonomy = TRUE, language = NULL, ...) {
  
  .Defunct("eol", "originr", msg = "This function is defunct since it seems that the data_objects part of the EOL API v1 does not work anymore.")

  # cli <- crul::HttpClient$new(
  #   url = file.path(eol_url("data_objects"), paste0(id, ".json")),
  #   headers = tx_ual,
  #   opts = list(...)
  # )
  # args <- argsnull(tc(list(taxonomy = as_l(taxonomy), language = language)))
  # res <- cli$get(query = args)
  # res$raise_for_status()
  # tt <- res$parse("UTF-8")
  # tmp <- jsonlite::fromJSON(tt)
  # tmp <- nmslwr(tmp)
  # if ("taxonconcepts" %in% names(tmp)) {
  #   tmp$taxonconcepts <- nmslwr(tmp$taxonconcepts)
  #   tmp$taxonconcepts$taxonrank <- tolower(tmp$taxonconcepts$taxonrank)
  # }
  # return(tmp)
}

nmslwr <- function(x) {
  stats::setNames(x, tolower(names(x)))
}

# NOTE: The EOL data_objects API does not seem to be working anymore and no documentation is available so it might be depreciated.

# context("eol_dataobjects")
# 
# test_that("eol_dataobjects with taxonomy TRUE", {
#   skip_on_cran()
# 
#   vcr::use_cassette("eol_dataobjects", {
#     temp <- sm(eol_dataobjects(id = "7561533", verbose = FALSE))
#   }, preserve_exact_body_bytes = TRUE)
# 
#   expect_is(temp, "list")
# 	expect_is(temp$taxonconcepts$scientificname, "character")
# 	expect_is(temp$taxonconcepts$identifier, "integer")
# 	expect_is(temp$taxonconcepts, "data.frame")
# })
# 
# test_that("eol_dataobjects with taxonomy FALSE", {
#   skip_on_cran()
# 
#   vcr::use_cassette("eol_dataobjects_taxonomy_param", {
#   	temp2 <- sm(eol_dataobjects(id = "7561533", 
#       taxonomy = FALSE, verbose = FALSE))
#   })
# 
# 	expect_is(temp2, "list")
# 	expect_null(temp2$taxonconcepts$scientificname)
# 	expect_null(temp2$taxonconcepts$identifier)
# })
# 
# test_that("eol_dataobjects fails well", {
#   skip_on_cran()
# 
#   expect_error(eol_dataobjects(), "argument \"id\" is missing")
# })

#' Get the GBIF backbone taxon ID from taxonomic names.
#'
#' @export
#' @param sci (character) one or more scientific names. Or, a `taxon_state`
#' object (see [taxon-state])
#' @param ask logical; should get_gbifid be run in interactive mode?
#' If TRUE and more than one ID is found for the species, the user is asked for
#' input. If FALSE NA is returned for multiple matches.
#' @param messages logical; If TRUE the actual taxon queried is printed on the console.
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all rows are considered.
#' Note that this function still only gives back a gbifid class object with one to many identifiers.
#' See [get_gbifid_()] to get back all, or a subset, of the raw data that you are
#' presented during the ask process.
#' @param phylum (character) A phylum (aka division) name. Optional. See `Filtering`
#' below.
#' @param class (character) A class name. Optional. See `Filtering` below.
#' @param order (character) An order name. Optional. See `Filtering` below.
#' @param family (character) A family name. Optional. See `Filtering` below.
#' @param rank (character) A taxonomic rank name. See [rank_ref] for possible
#' options. Though note that some data sources use atypical ranks, so inspect the
#' data itself for options. Optional. See `Filtering` below.
#' @param method (character) one of "backbone" or "lookup". See Details.
#' @param x Input to [as.gbifid()]
#' @param check logical; Check if ID matches any existing on the DB, only used in
#' [as.gbifid()]
#' @param ... Ignored
#' @param sciname Deprecated, see `sci`
#' @template getreturn
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @author Scott Chamberlain, 
#'
#' @details Internally in this function we use a function to search GBIF's taxonomy,
#' and if we find an exact match we return the ID for that match. If there isn't an
#' exact match we return the options to you to pick from.
#'
#' @section method parameter:
#' "backbone" uses the `/species/match` GBIF API route, matching against their
#' backbone taxonomy. We turn on fuzzy matching by default, as the search without
#' fuzzy against backbone is quite narrow. "lookup" uses the `/species/search`
#' GBIF API route, doing a full text search of name usages covering scientific
#' and vernacular named, species descriptions, distributions and the entire
#' classification.
#'
#' @section Filtering:
#' The parameters `phylum`, `class`, `order`, `family`, and `rank` are not used
#' in the search to the data provider, but are used in filtering the data down
#' to a subset that is closer to the target you want.  For all these parameters,
#' you can use regex strings since we use [grep()] internally to match.
#' Filtering narrows down to the set that matches your query, and removes the rest.
#'
#' @examples \dontrun{
#' get_gbifid(sci='Poa annua')
#' get_gbifid(sci='Pinus contorta')
#' get_gbifid(sci='Puma concolor')
#' 
#' #lots of queries
#' spp <- names_list("species", 10)
#' res <- get_gbifid(spp)
#' res
#' xx <- taxon_last()
#' xx
#'
#' # multiple names
#' get_gbifid(c("Poa annua", "Pinus contorta"))
#'
#' # specify rows to limit choices available
#' get_gbifid(sci='Pinus')
#' get_gbifid(sci='Pinus', rows=10)
#' get_gbifid(sci='Pinus', rows=1:3)
#'
#' # When not found, NA given
#' get_gbifid(sci="uaudnadndj")
#' get_gbifid(c("Chironomus riparius", "uaudnadndj"))
#'
#' # Narrow down results to a division or rank, or both
#' ## Satyrium example
#' ### Results w/o narrowing
#' get_gbifid("Satyrium")
#' ### w/ phylum
#' get_gbifid("Satyrium", phylum = "Tracheophyta")
#' get_gbifid("Satyrium", phylum = "Arthropoda")
#' ### w/ phylum & rank
#' get_gbifid("Satyrium", phylum = "Arthropoda", rank = "genus")
#'
#' ## Rank example
#' get_gbifid("Poa", method = "lookup")
#' get_gbifid("Poa", method = "lookup", rank = "genus")
#' get_gbifid("Poa", method = "lookup", family = "Thripidae")
#'
#' # Fuzzy filter on any filtering fields
#' ## uses grep on the inside
#' get_gbifid("Satyrium", phylum = "arthropoda")
#' get_gbifid("A*", method = "lookup", order = "*tera")
#' get_gbifid("A*", method = "lookup", order = "*ales")
#'
#' # Convert a uid without class information to a uid class
#' as.gbifid(get_gbifid("Poa annua")) # already a uid, returns the same
#' as.gbifid(get_gbifid(c("Poa annua","Puma concolor"))) # same
#' as.gbifid(2704179) # numeric
#' as.gbifid(c(2704179,2435099,3171445)) # numeric vector, length > 1
#' as.gbifid("2704179") # character
#' as.gbifid(c("2704179","2435099","3171445")) # character vector, length > 1
#' as.gbifid(list("2704179","2435099","3171445")) # list, either numeric or character
#' ## dont check, much faster
#' as.gbifid("2704179", check=FALSE)
#' as.gbifid(2704179, check=FALSE)
#' as.gbifid(2704179, check=FALSE)
#' as.gbifid(c("2704179","2435099","3171445"), check=FALSE)
#' as.gbifid(list("2704179","2435099","3171445"), check=FALSE)
#'
#' (out <- as.gbifid(c(2704179,2435099,3171445)))
#' data.frame(out)
#' as.uid( data.frame(out) )
#'
#' # Get all data back
#' get_gbifid_("Puma concolor")
#' get_gbifid_(c("Pinus", "uaudnadndj"))
#' get_gbifid_(c("Pinus", "Puma"), rows=5)
#' get_gbifid_(c("Pinus", "Puma"), rows=1:5)
#'
#' # use curl options
#' invisible(get_gbifid("Quercus douglasii", verbose = TRUE))
#' }

get_gbifid <- function(sci, ask = TRUE, messages = TRUE, rows = NA,
                       phylum = NULL, class = NULL, order = NULL,
                       family = NULL, rank = NULL, method = "backbone",
                       sciname = NULL, ...) {

  assert(sci, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(messages, "logical")
  assert(phylum, "character")
  assert(class, "character")
  assert(order, "character")
  assert(family, "character")
  assert(rank, "character")
  assert(method, "character")
  assert_rows(rows)
  
  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_gbifid(sciname)", with = "get_gbifid(sci)")
    sci <- sciname
  }
  
  if (inherits(sci, "character")) {
    tstate <- taxon_state$new(class = "gbifid", names = sci)
    items <- sci
  } else {
    assert_state(sci, "gbifid")
    tstate <- sci
    sci <- tstate$taxa_remaining()
    items <- c(sci, tstate$taxa_completed())
  }

  # Escape problematic characters
  sci <- gsub(sci, pattern = "[^\\]?'", replacement = "\\\\'")
  
  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci[i], "'\n")
    df <- switch(
      method,
      backbone = gbif_name_backbone(sci[i], ...),
      lookup = gbif_name_lookup(sci[i], ...)
    )
    mm <- NROW(df) > 1

    if (is.null(df)) df <- data.frame(NULL)

    if (nrow(df) == 0) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      att <- "not found"
    } else {
      names(df)[1] <- "gbifid"
      id <- df$gbifid
      att <- "found"
    }

    # not found
    if (length(id) == 0) {
      mssg(messages, m_not_found_sp_altclass)
      id <- NA_character_
      att <- "not found"
    }

    if (length(id) > 1) {
      # check for exact match
      matchtmp <- df[as.character(df$canonicalname) %in% sci[i], "gbifid"]
      if (length(matchtmp) == 1) {
        id <- as.character(matchtmp)
        direct <- TRUE
      } else {
        if (!is.null(phylum) || !is.null(class) || !is.null(order) ||
            !is.null(family) || !is.null(rank)) {
          df <- filt(df, "phylum", phylum)
          df <- filt(df, "class", class)
          df <- filt(df, "order", order)
          df <- filt(df, "family", family)
          df <- filt(df, "rank", rank)
        }

        df <- sub_rows(df, rows)
        if (NROW(df) == 0) {
          id <- NA_character_
          att <- "not found"
        } else {
          id <- df$gbifid
          if (length(id) == 1) {
            rank_taken <- as.character(df$rank)
            att <- "found"
          }
        }

        # more than one found -> user input
        if (length(id) > 1) {
          if (ask) {
            # limit to subset of columns for ease of use
            df <- df[, switch(method,
                              backbone = gbif_cols_show_backbone,
                              lookup = gbif_cols_show_lookup)]

            # prompt
            message("\n\n")
            message("\nMore than one GBIF ID found for taxon '", sci[i], "'!\n
            Enter rownumber of taxon (other inputs will return 'NA'):\n")
            rownames(df) <- 1:nrow(df)
            message(paste0(utils::capture.output(df), collapse = "\n"))
            take <- scan(n = 1, quiet = TRUE, what = 'raw')

            if (length(take) == 0) {
              take <- 'notake'
              att <- 'nothing chosen'
            }
            if (take %in% seq_len(nrow(df))) {
              take <- as.numeric(take)
              message("Input accepted, took gbifid '",
                      as.character(df$gbifid[take]), "'.\n")
              id <- as.character(df$gbifid[take])
              att <- "found"
            } else {
              id <- NA_character_
              att <- "not found"
              mssg(messages, "\nReturned 'NA'!\n\n")
            }
          } else {
            if (length(id) != 1) {
              warning(sprintf(m_more_than_one_found, "gbifid", sci[i]),
                call. = FALSE)
              id <- NA_character_
              att <- m_na_ask_false
            }
          }
        }
      }
    }
    res <- list(id = id, att = att, multiple = mm, direct = direct)
    prog$completed(sci[i], att)
    prog$prog(att)
    tstate$add(sci[i], res)
  }
  out <- tstate$get()
  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "gbifid",
                   match = pluck_un(out, "att", ""),
                   multiple_matches = pluck_un(out, "multiple", logical(1)),
                   pattern_match = pluck_un(out, "direct", logical(1)))
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$gbif)
}

#' @export
#' @rdname get_gbifid
as.gbifid <- function(x, check=FALSE) UseMethod("as.gbifid")

#' @export
#' @rdname get_gbifid
as.gbifid.gbifid <- function(x, check=FALSE) x

#' @export
#' @rdname get_gbifid
as.gbifid.character <- function(x, check=TRUE) if(length(x) == 1) make_gbifid(x, check) else collapse(x, make_gbifid, "gbifid", check=check)

#' @export
#' @rdname get_gbifid
as.gbifid.list <- function(x, check=TRUE) if(length(x) == 1) make_gbifid(x, check) else collapse(x, make_gbifid, "gbifid", check=check)

#' @export
#' @rdname get_gbifid
as.gbifid.numeric <- function(x, check=TRUE) as.gbifid(as.character(x), check)

#' @export
#' @rdname get_gbifid
as.gbifid.data.frame <- function(x, check = TRUE) {
  structure(x$ids, class = "gbifid", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri = x$uri)
}

#' @export
#' @rdname get_gbifid
as.data.frame.gbifid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "gbifid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_gbifid <- function(x, check=TRUE) make_generic(x, 'https://www.gbif.org/species/%s', "gbifid", check)

check_gbifid <- function(x){
  tryid <- tryCatch(gbif_name_usage(key = x), error = function(e) e)
  if ( "error" %in% class(tryid) && is.null(tryid$key) ) FALSE else TRUE
}

#' @export
#' @rdname get_gbifid
get_gbifid_ <- function(sci, messages = TRUE, rows = NA, method = "backbone",
  sciname = NULL) {

  if (!is.null(sciname)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_gbifid_(sciname)", with = "get_gbifid_(sci)")
    sci <- sciname
  }
  stats::setNames(lapply(sci, get_gbifd_help, messages = messages,
    rows = rows, method = method), sci)
}

get_gbifd_help <- function(sci, messages, rows, method){
  mssg(messages, "\nRetrieving data for taxon '", sci, "'\n")
  df <- switch(
    method,
    backbone = gbif_name_backbone(sci),
    lookup = gbif_name_lookup(sci)
  )
  if (!is.null(df)) df <- nmslwr(df)
  sub_rows(df, rows)
}

#' Return all synonyms for a taxon name with a given id from NBN
#'
#' @export
#' @param id the taxon identifier code
#' @param ... Further args passed on to [crul::verb-GET]
#' @return A data.frame
#' @family nbn
#' @references https://api.nbnatlas.org/
#' @examples \dontrun{
#' nbn_synonyms(id = 'NHMSYS0001501147')
#' nbn_synonyms(id = 'NHMSYS0000456036')
#'
#' # none
#' nbn_synonyms(id = 'NHMSYS0000502940')
#' }
nbn_synonyms <- function(id, ...) {
  url <- file.path(nbn_base(), "species", id)
  df <- nbn_GET_2(url, ...)
  df$synonyms
}

#' Get Worms ID for a taxon name
#'
#' Retrieve Worms ID of a taxon from World Register of Marine
#' Species (WORMS).
#'
#' @export
#' @param sci_com character; A vector of common or scientific names. Or, a
#' `taxon_state` object (see [taxon-state])
#' @param searchtype character; One of 'scientific' or 'common', or any unique
#' abbreviation
#' @param marine_only logical; marine only? default: `TRUE` (only used
#' when `searchtype="scientific"`); passed on to [worrms::wm_records_name()]
#' @param fuzzy logical; fuzzy search. default: `NULL` (`TRUE` for
#' `searchtype="scientific"` and `FALSE` for `searchtype="common"` to match
#' the default values for those parameters in \pkg{worrms} package); passed on
#' to [worrms::wm_records_name()] or [worrms::wm_records_common()]
#' @param accepted logical; If TRUE, removes names that are not accepted valid
#' names by WORMS. Set to `FALSE` (default) to give back both accepted
#' and unaccepted names.
#' @param ask logical; should get_wormsid be run in interactive mode?
#' If `TRUE` and more than one wormsid is found for the species, the
#' user is asked for input. If `FALSE` NA is returned for
#' multiple matches.
#' @param messages logical; should progress be printed?
#' @param rows numeric; Any number from 1 to infinity. If the default NaN, all
#' rows are considered. Note that this function still only gives back a wormsid
#' class object with one to many identifiers. See [get_wormsid_()] to get back
#' all, or a subset, of the raw data that you are presented during the ask
#' process.
#' @param query Deprecated, see `sci_com`
#' @param x Input to `as.wormsid`
#' @param ... Ignored
#' @param check logical; Check if ID matches any existing on the DB, only
#' used in [as.wormsid()]
#' @template getreturn
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @examples \dontrun{
#' (x <- get_wormsid('Gadus morhua'))
#' attributes(x)
#' attr(x, "match")
#' attr(x, "multiple_matches")
#' attr(x, "pattern_match")
#' attr(x, "uri")
#'
#' get_wormsid('Pomatomus saltatrix')
#' get_wormsid(c("Gadus morhua", "Lichenopora neapolitana"))
#'
#' # marine_only
#' get_wormsid("Apedinella", marine_only=TRUE)
#' get_wormsid("Apedinella", marine_only=FALSE)
#'
#' # fuzzy
#' ## searchtype="scientific": fuzzy is TRUE by default
#' get_wormsid("Platypro", searchtype="scientific", fuzzy=TRUE)
#' get_wormsid("Platypro", searchtype="scientific", fuzzy=FALSE)
#' ## searchtype="common": fuzzy is FALSE by default
#' get_wormsid("clam", searchtype="common", fuzzy=FALSE)
#' get_wormsid("clam", searchtype="common", fuzzy=TRUE)
#'
#' # by common name
#' get_wormsid("dolphin", 'common')
#' get_wormsid("clam", 'common')
#'
#' # specify rows to limit choices available
#' get_wormsid('Plat')
#' get_wormsid('Plat', rows=1)
#' get_wormsid('Plat', rows=1:2)
#'
#' # When not found
#' get_wormsid("howdy")
#' get_wormsid(c('Gadus morhua', "howdy"))
#'
#' # Convert a wormsid without class information to a wormsid class
#' # already a wormsid, returns the same
#' as.wormsid(get_wormsid('Gadus morhua'))
#' # same
#' as.wormsid(get_wormsid(c('Gadus morhua', 'Pomatomus saltatrix')))
#' # numeric
#' as.wormsid(126436)
#' # numeric vector, length > 1
#' as.wormsid(c(126436,151482))
#' # character
#' as.wormsid("126436")
#' # character vector, length > 1
#' as.wormsid(c("126436","151482"))
#' # list, either numeric or character
#' as.wormsid(list("126436","151482"))
#' ## dont check, much faster
#' as.wormsid("126436", check=FALSE)
#' as.wormsid(126436, check=FALSE)
#' as.wormsid(c("126436","151482"), check=FALSE)
#' as.wormsid(list("126436","151482"), check=FALSE)
#'
#' (out <- as.wormsid(c(126436,151482)))
#' data.frame(out)
#' as.wormsid( data.frame(out) )
#'
#' # Get all data back
#' get_wormsid_("Plat")
#' get_wormsid_("Plat", rows=1)
#' get_wormsid_("Plat", rows=1:2)
#' get_wormsid_("Plat", rows=1:75)
#' }
get_wormsid <- function(sci_com, searchtype = "scientific", marine_only = TRUE,
  fuzzy = NULL, accepted = FALSE, ask = TRUE, messages = TRUE,
  rows = NA, query = NULL, ...) {

  assert(sci_com, c("character", "taxon_state"))
  assert(searchtype, "character")
  assert(marine_only, "logical")
  assert(fuzzy, "logical")
  assert(accepted, "logical")
  assert(ask, "logical")
  assert_rows(rows)
  if (!is.null(query)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_wormsid(query)", with = "get_wormsid(sci_com)")
    sci_com <- query
  }
  

  if (inherits(sci_com, "character")) {
    tstate <- taxon_state$new(class = "wormsid", names = sci_com)
    items <- sci_com
  } else {
    assert_state(sci_com, "wormsid")
    tstate <- sci_com
    sci_com <- tstate$taxa_remaining()
    items <- c(sci_com, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci_com)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci_com[i], "'\n")

    if (!searchtype %in% c("scientific", "common")) {
      stop("'searchtype' must be one of 'scientific' or 'common'", call. = FALSE)
    }
    # NOTE: space replacement needed for curl problem (issue #888)
    query <- gsub(sci_com[i], pattern = ' ', replacement = '+', fixed = TRUE)
    wmdf <- switch(
      searchtype,
      scientific = worms_worker(query, worrms::wm_records_name, rows,
                                marine_only, fuzzy %||% TRUE, ...),
      common = worms_worker(query, worrms::wm_records_common, rows,
                            marine_only, fuzzy %||% FALSE, ...)
    )
    mm <- NROW(wmdf) > 1

    if (!inherits(wmdf, "tbl_df") || NROW(wmdf) == 0) {
      wmid <- NA_character_
      att <- "not found"
    } else {
      wmdf <- suppressWarnings(data.frame(wmdf))
      wmdf <- wmdf[, c("AphiaID", "scientificname", "authority", "status")]
      names(wmdf)[1] <- "id"

      if (accepted) {
        wmdf <- wmdf[ wmdf$status %in% 'accepted', ]
      }

      wmdf <- sub_rows(wmdf, rows)

      # should return NA if spec not found
      if (nrow(wmdf) == 0) {
        mssg(messages, m_not_found_sp_altclass)
        wmid <- NA_character_
        att <- "not found"
      }

      # take the one wmid from data.frame
      if (nrow(wmdf) == 1) {
        wmid <- wmdf$id
        att <- "found"
      }

      # check for direct match
      if (nrow(wmdf) > 1) {
        names(wmdf)[grep("scientificname", names(wmdf))] <- "target"
        matchtmp <- wmdf[tolower(wmdf$target) %in% tolower(sci_com[i]), "id"]
        if (length(matchtmp) == 1) {
          wmid <- matchtmp
          direct <- TRUE
          att <- "found"
        } else {
          wmid <- NA_character_
          att <- "not found"
        }
      }

      # multiple matches
      if (any(
        nrow(wmdf) > 1 && is.na(wmid) |
        nrow(wmdf) > 1 && att == "found" & length(wmid) > 1
      )) {
        if (ask) {
          names(wmdf)[grep("scientificname", names(wmdf))] <- "target"

          # user prompt
          wmdf <- wmdf[order(wmdf$target), ]

          # prompt
          message("\n\n")
          message(paste0(utils::capture.output(wmdf), collapse = "\n"))
          message("\nMore than one WORMS ID found for taxon '", sci_com[i], "'!\n
                  Enter rownumber of taxon (other inputs will return 'NA'):\n") # prompt
          take <- scan(n = 1, quiet = TRUE, what = 'raw')

          if (length(take) == 0) {
            take <- 'notake'
            att <- 'nothing chosen'
          }
          if (take %in% seq_len(nrow(wmdf))) {
            take <- as.numeric(take)
            message("Input accepted, took taxon '", as.character(wmdf$target[take]), "'.\n")
            wmid <-  wmdf$id[take]
            att <- "found"
          } else {
            wmid <- NA_character_
            mssg(messages, "\nReturned 'NA'!\n\n")
            att <- "not found"
          }
        } else {
          if (length(wmid) != 1) {
            warning(sprintf(m_more_than_one_found, "Worms ID", sci_com[i]),
              call. = FALSE)
            wmid <- NA_character_
            att <- m_na_ask_false
          }
        }
      }

    }
    res <- list(id = as.character(wmid), att = att, multiple = mm,
      direct = direct)
    prog$completed(sci_com[i], att)
    prog$prog(att)
    tstate$add(sci_com[i], res)
  }
  out <- tstate$get()
  ids <- structure(pluck_un(out, "id", ""), class = "wormsid",
    match = pluck_un(out, "att", ""),
    multiple_matches = pluck_un(out, "multiple", logical(1)),
    pattern_match = pluck_un(out, "direct", logical(1))
  )
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$worms)
}

try_df <- function(expr) {
  res <- tryCatch(expr, error = function(e) e)
  if (inherits(res, "error")) {
    data.frame(NULL)
  } else {
    res
  }
}

#' @export
#' @rdname get_wormsid
as.wormsid <- function(x, check=TRUE) UseMethod("as.wormsid")

#' @export
#' @rdname get_wormsid
as.wormsid.wormsid <- function(x, check=TRUE) x

#' @export
#' @rdname get_wormsid
as.wormsid.character <- function(x, check=TRUE) if (length(x) == 1) make_worms(x, check) else collapse(x, make_worms, "wormsid", check = check)

#' @export
#' @rdname get_wormsid
as.wormsid.list <- function(x, check=TRUE) if (length(x) == 1) make_worms(x, check) else collapse(x, make_worms, "wormsid", check = check)

#' @export
#' @rdname get_wormsid
as.wormsid.numeric <- function(x, check=TRUE) as.wormsid(as.character(x), check)

#' @export
#' @rdname get_wormsid
as.wormsid.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class = "wormsid", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri = x$uri)
}

#' @export
#' @rdname get_wormsid
as.data.frame.wormsid <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "wormsid",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_worms <- function(x, check=TRUE) {
  make_generic(x, 'https://www.marinespecies.org/aphia.php?p=taxdetails&id=%s',
    "wormsid", check)
}

check_wormsid <- function(x){
  tt <- worrms::wm_record(as.numeric(x))
  identical(as.character(tt$AphiaID), as.character(x))
}

#' @export
#' @rdname get_wormsid
get_wormsid_ <- function(sci_com, messages = TRUE, searchtype = "scientific",
  marine_only = TRUE, fuzzy = NULL, accepted = TRUE, rows = NA, query = NULL,
  ...) {

  if (!is.null(query)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_wormsid_(query)", with = "get_wormsid_(sci_com)")
    sci_com <- query
  }
  stats::setNames(
    lapply(sci_com, get_wormsid_help, messages = messages,
           searchtype = searchtype, marine_only = marine_only, fuzzy = fuzzy,
           accepted = accepted, rows = rows, ...),
    sci_com
  )
}

get_wormsid_help <- function(query, messages, searchtype, marine_only,
  fuzzy, accepted, rows, ...) {

  mssg(messages, "\nRetrieving data for taxon '", query, "'\n")
  searchtype <- match.arg(searchtype, c("scientific", "common"))
  df <- switch(
    searchtype,
    scientific = worms_worker(query, worrms::wm_records_name, rows = rows,
      marine_only = marine_only, fuzzy = fuzzy, ...),
    common = worms_worker(query, worrms::wm_records_common, rows = rows,
      marine_only = marine_only, fuzzy = fuzzy, ...)
  )
  if (!inherits(df, "tbl_df") || NROW(df) == 0) {
    NULL
  } else {
    df <- df[, c("AphiaID","scientificname","authority","status")]
    if (accepted) df <- df[ df$status %in% 'accepted', ]
    sub_rows(df, rows)
  }
}

# WORMS WORKER
# worms_worker(x = "Plat", expr = worrms::wm_records_name)
worms_worker <- function(x, expr, rows, marine_only, fuzzy, ...) {
  if (
    all(!is.na(rows)) &&
    class(rows) %in% c('numeric', 'integer') &&
    rows[length(rows)] <= 50
  ) {
    try_df(expr(x, marine_only = marine_only, fuzzy = fuzzy, ...))
  } else if (
    all(!is.na(rows)) &&
    class(rows) %in% c('numeric', 'integer') &&
    rows[length(rows)] > 50
  ) {
    out <- try_df(expr(x, marine_only = marine_only, fuzzy = fuzzy, ...))
    out <- list(out)
    i <- 1
    total <- 0
    while (NROW(out[[length(out)]]) == 50 && total < rows[length(rows)]) {
      i <- i + 1
      out[[i]] <- try_df(expr(x, marine_only = marine_only, fuzzy = fuzzy,
        offset = sum(unlist(sapply(out, NROW))), ...))
      total <- sum(unlist(sapply(out, NROW)))
    }
    df2dt2tbl(out)[rows,]
  } else {
    out <- try_df(expr(x, marine_only = marine_only, fuzzy = fuzzy, ...))
    out <- list(out)
    i <- 1
    while (NROW(out[[length(out)]]) == 50) {
      i <- i + 1
      out[[i]] <- try_df(expr(x, marine_only = marine_only, fuzzy = fuzzy,
        offset = sum(unlist(sapply(out, NROW))), ...))
    }
    df2dt2tbl(out)
  }
}

context("get_wormsid")

# test_that("get_wormsid returns the correct value", {
#   skip_on_cran()
#   vcr::use_cassette("get_wormsid", {
#     x <- sw(get_wormsid("Gadus morhua", rows = 1, messages = FALSE))
#     z <- sw(get_wormsid(c("Gadus morhua", "Lichenopora neapolitana"),
#       rows = 1, messages = FALSE))
#   })
# 
#   expect_is(x, "wormsid")
#   expect_is(x[1], "character")
#   expect_is(z, "wormsid")
#   expect_equal(length(z), 2)
#   expect_equal(sort(attr(z, "multiple_matches")), c(FALSE, TRUE))
#   expect_false(all(attr(z, "pattern_match")))
#   expect_match(attr(z, "uri"), "marinespecies.org")
# })

test_that("get_wormsid accepts ask-argument", {
  skip_on_cran()
  vcr::use_cassette("get_wormsid_ask_arg", {
    x <- sw(get_wormsid("Platanista gangetica", ask = FALSE, messages = FALSE))
    z <- get_wormsid("asdasf", ask = FALSE, messages = FALSE)
  })

  expect_is(x, "wormsid")
  expect_true(is.na(z))
})

# test_that("get_wormsid searchtype param works", {
#   skip_on_cran()
#   vcr::use_cassette("get_wormsid_query_modifiers", {
#     mod2 <- sw(get_wormsid("asiatic clam", "common", messages = FALSE))
#   })
# 
#   expect_is(mod2, "wormsid")
#   expect_equal(mod2[1], "181580")
# })

test_that("get_wormsid marine_only param works", {
  skip_on_cran()
  vcr::use_cassette("get_wormsid_marine_only", {
    a <- sw(get_wormsid("Apedinella", marine_only = TRUE, accepted = TRUE, messages = FALSE, row = 2))
    b <- sw(get_wormsid("Apedinella", marine_only = FALSE, messages = FALSE))
  })

  expect_is(a, "wormsid")
  expect_is(b, "wormsid")
  expect_equal(a[1], "248097")
  expect_equal(b[1], "248096")
})

test_that("get_wormsid fuzzy param works", {
  skip_on_cran()
  vcr::use_cassette("get_wormsid_fuzzy", {
    a <- sw(get_wormsid("Platypleu", fuzzy = TRUE, accepted = TRUE, messages = FALSE))
    b <- sw(get_wormsid("Platypleu", fuzzy = FALSE, messages = FALSE))
  })

  expect_is(a, "wormsid")
  expect_is(b, "wormsid")
  expect_equal(a[1], "105519")
  expect_true(is.na(b[1]))
})

test_that("get_wormsid fails well", {
  skip_on_cran()
  expect_error(get_wormsid(), "argument \"sci_com\" is missing")
  expect_error(get_wormsid("clam", 5, messages = FALSE),
               "searchtype must be of class character")
  expect_error(get_wormsid("clam", "stuff", messages = FALSE),
               "'searchtype' must be one of")
  expect_error(get_wormsid("clam", accepted = 4, messages = FALSE),
               "accepted must be of class logical")
  expect_error(get_wormsid("clam", ask = 4, messages = FALSE),
               "ask must be of class logical")
  expect_error(get_wormsid("clam", marine_only = 4, messages = FALSE),
               "marine_only must be of class logical")
  expect_error(get_wormsid("clam", fuzzy = 4, messages = FALSE),
               "fuzzy must be of class logical")
})

test_that("get_wormsid exact match with more than 1 exact match found", {
  skip_on_cran()
  vcr::use_cassette("get_wormsid_multiple_exact_matches_found", {
    x <- get_wormsid("Gadus", messages = FALSE, rows = 1:2)
    1
  })

  expect_is(x, "wormsid")
  expect_is(x[1], "character")
  expect_equal(length(x), 1)
  expect_true(attr(x, "multiple_matches"))
  expect_false(attr(x, "pattern_match"))
})

#' Get Kew's Plants of the World code for a taxon
#'
#' @export
#' @param sci_com character; A vector of common or scientific names. Or, a
#' `taxon_state` object (see [taxon-state])
#' @param accepted logical; If TRUE, removes names that are not accepted 
#' valid names by ITIS. Set to `FALSE` (default) to give back both 
#' accepted and unaccepted names.
#' @param ask logical; should `get_pow` be run in interactive mode?
#' If TRUE and more than one pow is found for teh species, the user is 
#' asked for input. If FALSE NA is returned for multiple matches.
#' @param messages logical; should progress be printed?
#' @param x For `get_pow()`: deprecated, see `sci_com`. For `as.pow`, various,
#' see examples
#' @param ... Curl options passed on to [crul::HttpClient]
#' @param rows numeric; Any number from 1 to infinity. If the default NA, 
#' all rows are considered. Note that this function still only gives back 
#' a pow class object with one to many identifiers. See 
#' [get_pow_()] to get back all, or a subset,
#' of the raw data that you are presented during the ask process.
#' @param family_filter (character) A division (aka phylum) name to filter
#' data after retrieved from NCBI. Optional. See `Filtering` below.
#' @param rank_filter (character) A taxonomic rank name to filter data after
#' retrieved from NCBI. See [rank_ref] for possible options.
#' Though note that some data sources use atypical ranks, so inspect the data
#' itself for options. Optional. See `Filtering` below.
#' @param check logical; Check if ID matches any existing on the DB, only 
#' used in [as.pow()]
#' @template getreturn
#' 
#' @family pow
#' 
#' @section Filtering:
#' The parameters `family_filter` an`rank_filter`er are not
#' used in the search to the data provider, but are used in filtering the data down to a
#' subset that is closer to the target you want.  For these two parameters,
#' you can use regex strings since we use [grep()] internally to match.
#' Filtering narrows down to the set that matches your query, and removes the rest.
#' 
#' @section Rate-limits:
#' As of February 2019, KEW was limiting to 5 requests per second. Note that
#' they may change that number in the future.
#' 
#' If you get errors that contain `429` you are hitting the rate limit, and you
#' can get around it by doing requests with `Sys.sleep` in between requests.
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @examples \dontrun{
#' get_pow(sci_com="Helianthus")
#' get_pow(c("Helianthus","Quercus douglasii"))
#'
#' # Get back a subset
#' get_pow(sci_com="Helianthus", rows = 1)
#' get_pow(sci_com="Helianthus", rows = 1:10)
#'
#' # When not found
#' get_pow("howdy")
#' get_pow(c("Helianthus annuus", "howdy"))
#'
#' # Narrow down results 
#' # to accepted names
#' get_pow("Helianthus", accepted = TRUE)
#' # to a kingom 
#' get_pow("Helianthus", rank_filter = "genus")
#' # to accepted names and rank
#' get_pow("Helianthus annuus", accepted = TRUE, rank_filter = "species")
#' # to a family
#' get_pow("flower", family_filter = "Acanthaceae")
#'
#' # Convert a pow without class information to a pow class
#' z <- get_pow("Helianthus annuus", accepted = TRUE, rank_filter = "species")
#' # already a pow, returns the same
#' as.pow(z)
#' as.pow("urn:lsid:ipni.org:names:119003-2")
#' # character vector, length > 1
#' ids <- c("urn:lsid:ipni.org:names:119003-2","urn:lsid:ipni.org:names:328247-2")
#' as.pow(ids)
#' # list, with character strings
#' as.pow(as.list(ids)) 
#' ## dont check, much faster
#' as.pow("urn:lsid:ipni.org:names:119003-2", check=FALSE)
#' as.pow(ids, check=FALSE)
#' as.pow(as.list(ids), check=FALSE)
#'
#' (out <- as.pow(ids))
#' data.frame(out)
#' as.pow( data.frame(out) )
#'
#' # Get all data back
#' get_pow_("Quercus", rows=1:5)
#' get_pow_("Quercus", rows=1)
#' get_pow_(c("Pinus", "Abies"), rows = 1:3)
#' }

get_pow <- function(sci_com, accepted = FALSE, ask = TRUE, messages = TRUE,
  rows = NA, family_filter = NULL, rank_filter = NULL, x = NULL, ...) {

  assert(sci_com, c("character", "taxon_state"))
  assert(accepted, "logical")
  assert(ask, "logical")
  assert(messages, "logical")
  assert(family_filter, "character")
  assert(rank_filter, "character")
  assert_rows(rows)
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_pow(x)", with = "get_pow(sci_com)")
    sci_com <- x
  }
  
  if (inherits(sci_com, "character")) {
    tstate <- taxon_state$new(class = "pow", names = sci_com)
    items <- sci_com
  } else {
    assert_state(sci_com, "pow")
    tstate <- sci_com
    sci_com <- tstate$taxa_remaining()
    items <- c(sci_com, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci_com)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci_com[i], "'\n")
    pow_df <- pow_search(sci_com = sci_com[i], ...)$data
    mm <- NROW(pow_df) > 1

    if (!inherits(pow_df, "data.frame")) {
      pow <- NA_character_
      att <- "not found"
    } else {
      pow_df <- pow_df[, c("name","rank","accepted","kingdom","family","fqId")]

      if (accepted) {
        pow_df <- pow_df[pow_df$accepted, ]
      }
      
      # should return NA if spec not found
      if (nrow(pow_df) == 0) {
        mssg(messages, "Not found. Consider checking the spelling or alternate classification")
        pow <- NA_character_
        att <- 'not found'
      }

      # take the one pow from data.frame
      if (nrow(pow_df) == 1) {
        pow <- pow_df$fqId
        att <- 'found'
      }
      # check for direct match
      if (nrow(pow_df) > 1) {
        names(pow_df)[grep('name', names(pow_df))] <- "target"
        di_rect <- pow_df[tolower(pow_df$target) %in% tolower(sci_com[i]), "fqId"]
        if (length(di_rect) == 1) {
          pow <- di_rect
          direct <- TRUE
          att <- 'found'
        } else {
          pow <- NA_character_
          direct <- FALSE
          att <- 'found'
        }
      }
      # multiple matches
      if (any(
        nrow(pow_df) > 1 && is.na(pow) |
        nrow(pow_df) > 1 && att == "found" && length(pow) > 1
      )) {
        names(pow_df)[grep('^name$', names(pow_df))] <- "target"

        if (!is.null(family_filter) || !is.null(rank_filter)) {
          pow_df <- filt(pow_df, "family", family_filter)
          pow_df <- filt(pow_df, "rank", rank_filter)
        }

        pow_df <- sub_rows(pow_df, rows)
        pow <- id <- pow_df$fqId
        if (length(id) == 1) {
          direct <- TRUE
          att <- "found"
        }

        if (ask) {
          # user prompt
          pow_df <- pow_df[order(pow_df$target), ]
          rownames(pow_df) <- 1:nrow(pow_df)
          if (length(pow) > 1 || NROW(pow_df) > 1) {
            # prompt
            message("\n\n")
            message(paste0(utils::capture.output(pow_df), collapse = "\n"))
            message("\nMore than one pow found for taxon '", sci_com[i], "'!\n
          Enter rownumber of taxon (other inputs will return 'NA'):\n") # prompt
            take <- scan(n = 1, quiet = TRUE, what = 'raw')

            if (length(take) == 0)
              take <- 'notake'
            if (take %in% seq_len(nrow(pow_df))) {
              take <- as.numeric(take)
              message("Input accepted, took taxon '", as.character(pow_df$target[take]), "'.\n")
              pow <-  pow_df$fqId[take]
              att <- 'found'
            } else {
              pow <- NA_character_
              mssg(messages, "\nReturned 'NA'!\n\n")
              att <- 'not found'
            }
          }
        } else {
          if (length(pow) == 1) {
            att <- "found"
          } else {
            warning(
              sprintf("More than one pow found for taxon '%s'; refine query or set ask=TRUE",
                      sci_com[i]),
              call. = FALSE
            )
            pow <- NA_character_
            att <- 'NA due to ask=FALSE & > 1 result'
          }
        }
      }
    }
    res <- list(id = as.character(pow), att = att, multiple = mm,
      direct = direct)
    prog$completed(sci_com[i], att)
    prog$prog(att)
    tstate$add(sci_com[i], res)
  }
  out <- tstate$get()
  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "pow",
                   match = pluck_un(out, "att", ""),
                   multiple_matches = pluck_un(out, "multiple", logical(1)),
                   pattern_match = pluck_un(out, "direct", logical(1)))
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$pow)
}

#' @export
#' @rdname get_pow
as.pow <- function(x, check=TRUE) UseMethod("as.pow")

#' @export
#' @rdname get_pow
as.pow.pow <- function(x, check=TRUE) x

#' @export
#' @rdname get_pow
as.pow.character <- function(x, check=TRUE) if(length(x) == 1) make_pow(x, check) else collapse(x, make_pow, "pow", check=check)

#' @export
#' @rdname get_pow
as.pow.list <- function(x, check=TRUE) if(length(x) == 1) make_pow(x, check) else collapse(x, make_pow, "pow", check=check)

#' @export
#' @rdname get_pow
as.pow.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class="pow", match=x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri=x$uri)
}

#' @export
#' @rdname get_pow
as.data.frame.pow <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "pow",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_pow <- function(x, check=TRUE) {
  make_generic(x, 'https://powo.science.kew.org/taxon/%s', "pow", check)
}

check_pow <- function(x){
  tryid <- tryCatch(pow_lookup(x), error = function(e) e)
  !inherits(tryid, "error")
}

#' @export
#' @rdname get_pow
get_pow_ <- function(sci_com, messages = TRUE, rows = NA, x = NULL, ...) {
  if (!is.null(x)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_pow_(x)", with = "get_pow_(sci_com)")
    sci_com <- x
  }
  
  stats::setNames(lapply(sci_com, get_pow_help, messages = messages,
    rows = rows, ...), sci_com)
}

get_pow_help <- function(x, messages, rows, ...){
  mssg(messages, "\nRetrieving data for taxon '", x, "'\n")
  df <- pow_search(sci_com = x, ...)$data
  if (NROW(df) == 0) NULL else sub_rows(df, rows)
}

context("get_pow")

test_that("get_pow: not found", {
  skip_on_cran()
  vcr::use_cassette("get_pow", {
    z <- get_pow("adsf asdf asdf", messages = FALSE)[[1]]
    w <- get_pow(c("Helianthus excubitor", "adsf asdf asdf"), 
      messages = FALSE)
  })

  expect_true(is.na(z))
  expect_is(w, "pow")
})

test_that("get_pow accepts ask-argument", {
  skip_on_cran()
  vcr::use_cassette("get_pow_ask_arg", {
    z <- get_pow("adsf asdf asdf", ask = FALSE,
      messages = FALSE)[[1]]
  })
  expect_true(is.na(z))
})

test_that("get_pow fails as expected", {
  skip_on_cran()

  expect_error(get_pow(), "argument \"sci_com\" is missing")
  expect_error(get_pow("Poa annua", ask = 4, messages = FALSE),
               "ask must be of class logical")
  expect_error(
    get_pow("Satyrium", family_filter = 234, messages = FALSE),
    "family_filter must be of class character")
  expect_error(
    get_pow("Satyrium", rank_filter = 234, messages = FALSE),
    "rank_filter must be of class character")

  # rows param
  expect_error(get_pow("Satyrium", rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_pow("Satyrium", rows = 0, messages = FALSE),
               "all\\(rows > 0\\) is not TRUE")
})

#' Search UK National Biodiversity Network database for
#' taxonomic classification
#'
#' @export
#' @param id (character) An NBN identifier.
#' @param ... Further args passed on to [crul::verb-GET]
#' @return A data.frame
#' @family nbn
#' @author Scott Chamberlain, 
#' @references https://api.nbnatlas.org/
#' @examples \dontrun{
#' nbn_classification(id="NHMSYS0000376773")
#'
#' # get id first, then pass to this fxn
#' id <- get_nbnid("Zootoca vivipara", rec_only = TRUE, rank = "Species")
#' nbn_classification(id)
#'
#' nbn_classification(id="NHMSYS0000502940", verbose = TRUE)
#' }
nbn_classification <- function(id, ...) {
  url <- file.path(nbn_base(), "classification", id)
  nbn_GET_2(url, ...)
}

nbn_GET_2 <- function(url, ...) {
  cli <- crul::HttpClient$new(url, headers = tx_ual, opts = list(...))
  res <- cli$get()
  res$raise_for_status()
  nmslwr(jsonlite::fromJSON(res$parse("UTF-8"), TRUE))
}

#' Verify a list of scientific names against biodiversity data-sources.
#'
#' This service parses incoming names, executes exact or fuzzy matching as
#' required, and returns the best-scored result. Optionally, it can also return
#' matches from data-sources selected by a user.
#'
#' @export
#'
#' @param names A `character` vector of taxon names to verify.
#' @param data_sources A `character` or `integer` vector with numbers
#'   corresponding to data sources. See the Global Names Architecture
#'   documentation for a list of available options.
#' @param all_matches When `TRUE`, return all found matches, not only the best
#'   one. Multiple results are returned in results. These results are sorted by
#'   matching quality, the first result is the same as bestResult.
#' @param capitalize When `TRUE`, capitalize the first letter of a name-string.
#' @param species_group When `TRUE`, expands the search to species group where
#'   applicable.
#' @param fuzzy_uninomial When `TRUE`, allows fuzzy matching for uninomial
#'   names.
#' @param stats When `TRUE`, finds out a kingdom and a taxon (main taxon) that
#'   contain most names. It only takes in account the names matched to the
#'   Catalogue of Life entries. This option is ignored, if the Catalogue of Life
#'   is not included in data-sources.
#' @param main_taxon_threshold A `numeric` vector from 0.5 to 1. This sets the
#'   minimal percentage for the main taxon discovery.
#' @param output_type A `character` vector of length 1, either `table` or
#'   `list`, indicating the format of the output. The tabular output only
#'   contains values that consistently appear in all results, so `list` output
#'   can have additional information. For `list` and `json` outputs, only values
#'   for unique taxon names are returned, but the `table` output has rows that
#'   correspond 1-1 with the input data.
#' @param ... Curl options passed on to [crul::HttpClient]
#'
#' @return Depends on the value of the `output_type` option
#'
#' @author Zachary S.L. Foster
#'
#' @examples \dontrun{
#' gna_verifier(c("Helianthus annuus", "Homo saapiens"))
#' gna_verifier(c("Helianthus annuus", "Homo saapiens"), all_matches = TRUE)
#' }
gna_verifier <- function(
    names,
    data_sources = c(1, 12),
    all_matches = FALSE,
    capitalize = FALSE,
    species_group = FALSE,
    fuzzy_uninomial = FALSE,
    stats = FALSE,
    main_taxon_threshold = 0.5,
    output_type = 'table',
    ...
) {
  batch_size <- 100 # How many names to lookup with each api call
  
  # Parse and verify input options
  data_sources <- as.character(data_sources)
  is_number <- grepl(data_sources, pattern = '[0-9]+')
  if (any(! is_number) || length(data_sources) == 0) {
    stop(call. = FALSE, 'The `data_sources` input must be a vector of numbers with at least one value.')
  }
  data_sources <- paste0(data_sources, collapse = '|')
  
  check_if_logical <- function(value, name) {
    if (length(value) != 1 || is.na(value) || ! is.logical(value)) {
      stop(call. = FALSE, 'The `', name, '` input must be a TRUE/FALSE vector of length 1.')
    }
  }
  check_if_logical(all_matches, 'all_matches')
  check_if_logical(capitalize, 'capitalize')
  check_if_logical(species_group, 'species_group')
  check_if_logical(fuzzy_uninomial, 'fuzzy_uninomial')
  check_if_logical(stats, 'stats')
  
  if (length(main_taxon_threshold) != 1 || ! is.numeric(main_taxon_threshold) || main_taxon_threshold < 0.5 || main_taxon_threshold > 1) {
    stop(call. = FALSE, 'The `main_taxon_threshold` input must be a single number between 0.5 to 1')
  }
  
  output_type <- match.arg(output_type, c('table', 'list', 'json'))
  
  # Convert input to unique values to avoid redundant API wprk
  unique_names <- unique(names)
  name_batches <- split(unique_names, ceiling(seq_along(unique_names) / batch_size)) 
  
  batch_data <- lapply(name_batches, function(batch) {
    # Format the API GET request
    base_url <- 'https://verifier.globalnames.org/'
    args <- c(
      data_sources = paste0(data_sources, collapse = '|'),
      all_matches = tolower(as.character(all_matches)),
      capitalize = tolower(as.character(capitalize)),
      species_group = tolower(as.character(species_group)),
      fuzzy_uninomial = tolower(as.character(fuzzy_uninomial)),
      stats = tolower(as.character(stats))
    )
    formatted_args <- paste0(paste0(names(args), '=', args), collapse = '&')
    formatted_path <- paste0(
      'api/v1/verifications/',
      paste0(batch, collapse = '|'),
      '?', formatted_args
    )
    
    # Make and parse API call
    api <- crul::HttpClient$new(base_url, headers = tx_ual, opts = list(...))
    response <- api$get(path = formatted_path)
    response$raise_for_status()
    response_json <- response$parse("UTF-8")
    response_data <- jsonlite::fromJSON(response_json, FALSE)
    return(response_data$names)
  })
 
  # Combine batch data to a single list
  response_data <- unlist(batch_data, recursive = FALSE)
  names(response_data) <- unique_names
  if (output_type == 'list') {
    return(response_data)
  }
  
  # Reformat response data to a table
  used_cols <- c(
    'submittedName',
    'dataSourceId',
    'dataSourceTitleShort',
    'curation',
    'recordId',
    'entryDate', 
    'sortScore',
    'matchedNameID',
    'matchedName', 
    'matchedCardinality',
    'matchedCanonicalSimple',
    'matchedCanonicalFull', 
    'currentRecordId', 
    'currentNameId',
    'currentName',
    'currentCardinality',
    'currentCanonicalSimple', 
    'currentCanonicalFull', 
    'taxonomicStatus',
    'isSynonym',
    'editDistance',
    'stemEditDistance',
    'matchType',
    'cardinalityScore', 
    'infraSpecificRankScore',
    'fuzzyLessScore',
    'curatedDataScore',
    'authorMatchScore',
    'acceptedNameScore',
    'parsingQualityScore'
  )
  convert_entry_to_row <- function(x, input_name) {
    if (is.null(x)) { # If there was no match
      output <- c(input_name, rep(NA, length(used_cols)))
      names(output) <- c('submittedName', used_cols)
      output['matchType'] <- 'NoMatch'
    } else {
      output <- c(input_name, unlist(x))
      names(output)[1] <- 'submittedName'
    }
    parsed_names <- vapply(strsplit(names(output), split = '\\.'),
                           function(y) y[length(y)], FUN.VALUE = character(1))
    names(output) <- parsed_names
    return(as.data.frame(as.list(output[used_cols])))
  }
  if (all_matches) {
    response_table <- do.call(rbind, lapply(response_data, function(x) {
      do.call(rbind, lapply(x$results, function(y) convert_entry_to_row(y, x$name)))
    }))    
  } else {
    response_table <- do.call(rbind, lapply(response_data, function(x) {
      convert_entry_to_row(x$bestResult, x$name)
    }))
  }
  
  # Reduplicate rows to match input
  split_table <- split(response_table, response_table$submittedName)
  submitted_names <- vapply(split_table, function(x) x$submittedName[1], FUN.VALUE = character(1))
  response_table <- do.call(rbind, split_table[match(names, submitted_names)])
  row.names(response_table) <- NULL
  
  # Format table to be more useful to the user
  numeric_cols <- c(
    'sortScore',
    'matchedCardinality',
    'currentCardinality',
    'editDistance',
    'stemEditDistance',
    'cardinalityScore',
    'infraSpecificRankScore',
    'fuzzyLessScore', 
    'curatedDataScore',
    'authorMatchScore',
    'acceptedNameScore', 
    'parsingQualityScore'
  ) 
  response_table[numeric_cols] <- lapply(response_table[numeric_cols], as.numeric)
  response_table$isSynonym <- as.logical(response_table$isSynonym)
  response_table <- tibble::as_tibble(response_table)
  
  return(response_table)
}

#' Aggregate species data to given taxonomic rank
#'
#' @export
#' @param x Community data matrix. Taxa in columns, samples in rows.
#' @param rank character; Taxonomic rank to aggregate by.
#' @param db character; taxonomic API to use, 'ncbi, 'itis' or both, see
#' [tax_name()]. Note that each taxonomic data source has
#' their own identifiers, so that if you provide the wrong `db` value
#' for the identifier you could get a result, but it will likely be wrong (not
#' what you were expecting). If using ncbi we recommend getting an API key;
#' see [taxize-authentication]
#' @param messages (logical) If FALSE (Default) suppress messages
#' @param ... Other arguments passed to [get_tsn()] or [get_uid()]
#'
#' @details `tax_agg` aggregates (sum) taxa to a specific taxonomic level.
#' If a taxon is not found in the database (ITIS or NCBI) or the supplied taxon
#' is on higher taxonomic level this taxon is not aggregated.
#'
#'
#' @return A list of class `tax_agg` with the following items:
#' * `x` Community data matrix with aggregated data.
#' * `by` A lookup-table showing which taxa were aggregated.
#' * `n_pre` Number of taxa before aggregation.
#' * `rank` Rank at which taxa have been aggregated.
#'
#' @seealso [tax_name]
#' @examples \dontrun{
#' if (requireNamespace("vegan", quietly = TRUE)) {
#'   # use dune dataset
#'   data(dune, package='vegan')
#'   species <- c("Achillea millefolium", "Agrostis stolonifera",
#'     "Aira praecox", "Alopecurus geniculatus", "Anthoxanthum odoratum",
#'     "Bellis perennis", "Bromus hordeaceus", "Chenopodium album",
#'     "Cirsium arvense", "Comarum palustre", "Eleocharis palustris",
#'     "Elymus repens", "Empetrum nigrum", "Hypochaeris radicata",
#'     "Juncus articulatus", "Juncus bufonius", "Lolium perenne",
#'     "Plantago lanceolata", "Poa pratensis", "Poa trivialis",
#'     "Ranunculus flammula", "Rumex acetosa", "Sagina procumbens",
#'     "Salix repens", "Scorzoneroides autumnalis", "Trifolium pratense",
#'     "Trifolium repens", "Vicia lathyroides", "Brachythecium rutabulum",
#'     "Calliergonella cuspidata")
#'   colnames(dune) <- species
#'
#'   # aggregate sample to families
#'   (agg <- tax_agg(dune, rank = 'family', db = 'ncbi'))
#'
#'   # extract aggregated community data matrix for further usage
#'   agg$x
#'   # check which taxa have been aggregated
#'   agg$by
#' }
#'
#' # A use case where there are different taxonomic levels in the same dataset
#' spnames <- c('Puma','Ursus americanus','Ursidae')
#' df <- data.frame(c(1,2,3), c(11,12,13), c(1,4,50))
#' names(df) <- spnames
#' out <- tax_agg(x=df, rank = 'family', db='itis')
#' out$x
#'
#' # You can input a matrix too
#' mat <- matrix(c(1,2,3, 11,12,13), nrow = 2, ncol = 3,
#'  dimnames=list(NULL, c('Puma concolor','Ursus americanus','Ailuropoda melanoleuca')))
#' tax_agg(mat, rank = 'family', db='itis')
#' }

tax_agg <- function(x, rank, db = 'ncbi', messages=FALSE, ...)
{
  if (is.matrix(x)) {
    if (is.null(colnames(x)))
      stop("The community data matrix must have named columns")
    x <- data.frame(x, check.names = FALSE)
  }
  # bring to long format
  # df_m <- data.table::melt(x)
  x$rownames <- rownames(x)
  df_m <- setDF(suppressWarnings(data.table::melt(as.data.table(x))))

  # aggregate to family level (by querying NCBI for taxonomic classification)
  uniq_tax <- as.character(unique(df_m$variable))
  agg <- tax_name(uniq_tax, get = rank, db = db, messages = messages, ...)
  lookup <- data.frame(variable = uniq_tax, agg = agg[ , 3], stringsAsFactors = FALSE)

  # merge lookup with orig.
  df_merged <- merge(lookup, df_m, by = 'variable')

  # if not found , or on higher level -> use orig. = no aggrgation
  df_merged$agg <- ifelse(is.na(df_merged$agg), df_merged$variable, df_merged$agg)

  # bring back to long format and aggregate
  df_l <- setDF(data.table::dcast(as.data.table(df_merged),
    rownames ~ agg, value.var = 'value', fun.aggregate = sum))

  rownames(df_l) <- df_l$rownames
  df_l$rownames <- NULL
  # restore order
  df_l <- df_l[x$rownames, ]
  out <- list(x = df_l, by = lookup, n_pre = ncol(x) - 1, rank = rank)
  class(out) <- 'tax_agg'
  return(out)
}

#' @method print tax_agg
#' @export
#' @rdname tax_agg
print.tax_agg <- function(x, ...)
{
  cat("\n")
  writeLines(strwrap("Aggregated community data\n",
                     prefix = "\t"))
  cat(paste("\nLevel of Aggregation:", toupper(x$rank)))
  cat(paste("\nNo. taxa before aggregation:", x$n_pre))
  cat(paste("\nNo. taxa after aggregation:", ncol(x$x)))
  cat(paste("\nNo. taxa not found:", sum(is.na(x$by$agg))))
}

context("tax_agg")

test_that("tax_agg returns the correct class", {
  skip_on_cran()
  skip_on_travis()
  skip_if_not_installed("vegan")

  suppressPackageStartupMessages(library("vegan"))
  data(dune, package='vegan')
  take <- dune[ ,1:5]
  species <- c("Bellis perennis", "Empetrum nigrum", "Juncus bufonius",
    "Juncus articulatus", "Aira praecox")
  colnames(take) <- species
  out_ncbi <- tax_agg(take, rank = 'family', db = 'ncbi', messages = FALSE)
  take2 <- take
  colnames(take2) <- NULL

  expect_that(out_ncbi, is_a("tax_agg"))
  expect_that(length(out_ncbi), equals(4))
  expect_equal(class(out_ncbi), "tax_agg")
  expect_is(out_ncbi$x, "data.frame")

  expect_that(nrow(out_ncbi$x), equals(nrow(take)))
  expect_that(nrow(out_ncbi$by), equals(length(unique(colnames(take)))))
  expect_that(out_ncbi$n_pre, equals(length(unique(colnames(take)))))
  expect_error(tax_agg(as.matrix(take2), rank = 'family', db = 'ncbi',
    messages = FALSE))
})


## Moving this test to the long running tests branch since it takes a long time
# take2 <- take
# colnames(take2)[4] <- 'xxxxx'
# out_itis <- tax_agg(take2, rank = 'family', db = 'itis', verbose = FALSE)
#
#
# test_that("Handles NAs", {
#   expect_true(is.na(out_itis$by$agg[4]))
#   expect_that(ncol(out_itis$x), equals(ncol(take2)))
# })

#' Get the TSN code for a search term.
#'
#' Retrieve the taxonomic serial numbers (TSN) of a taxon from ITIS.
#'
#' @export
#' @param sci_com character; A vector of common or scientific names.
#' Or, a `taxon_state` object (see [taxon-state])
#' @param searchtype character; One of 'scientific' or 'common', or any
#' unique abbreviation
#' @param accepted logical; If TRUE, removes names that are not accepted valid
#' names by ITIS. Set to `FALSE` (default) to give back both accepted
#' and unaccepted names.
#' @param ask logical; should get_tsn be run in interactive mode?
#' If `TRUE` and more than one TSN is found for the species, the user is
#' asked for input. If `FALSE` NA is returned for multiple matches.
#' @param messages logical; should progress be printed?
#' @param rows numeric; Any number from 1 to infinity. If the default NA, all
#' rows are considered. Note that this function still only gives back a tsn
#' class object with one to many identifiers. See
#' [get_tsn_()] to get back all, or a subset, of the raw
#' data that you are presented during the ask process.
#' @param x Input to as.tsn
#' @param searchterm Deprecated, see `sci_com`
#' @param ... Ignored
#' @param check logical; Check if ID matches any existing on the DB, only
#' used in [as.tsn()]
#' @template getreturn
#'
#' @family taxonomic-ids
#' @seealso [classification()]
#'
#' @examples \dontrun{
#' get_tsn("Quercus douglasii")
#' get_tsn("Chironomus riparius")
#' get_tsn(c("Chironomus riparius","Quercus douglasii"))
#' splist <- c("annona cherimola", 'annona muricata', "quercus robur",
#' 		"shorea robusta", "pandanus patina", "oryza sativa", "durio zibethinus")
#' get_tsn(splist, messages=FALSE)
#'
#' # specify rows to limit choices available
#' get_tsn('Arni')
#' get_tsn('Arni', rows=1)
#' get_tsn('Arni', rows=1:2)
#'
#' # When not found
#' get_tsn("howdy")
#' get_tsn(c("Chironomus riparius", "howdy"))
#'
#' # Using common names
#' get_tsn("black bear", searchtype="common")
#'
#' # Convert a tsn without class information to a tsn class
#' as.tsn(get_tsn("Quercus douglasii")) # already a tsn, returns the same
#' as.tsn(get_tsn(c("Chironomus riparius","Pinus contorta"))) # same
#' as.tsn(19322) # numeric
#' as.tsn(c(19322,129313,506198)) # numeric vector, length > 1
#' as.tsn("19322") # character
#' as.tsn(c("19322","129313","506198")) # character vector, length > 1
#' as.tsn(list("19322","129313","506198")) # list, either numeric or character
#' ## dont check, much faster
#' as.tsn("19322", check=FALSE)
#' as.tsn(19322, check=FALSE)
#' as.tsn(c("19322","129313","506198"), check=FALSE)
#' as.tsn(list("19322","129313","506198"), check=FALSE)
#'
#' (out <- as.tsn(c(19322,129313,506198)))
#' data.frame(out)
#' as.tsn( data.frame(out) )
#'
#' # Get all data back
#' get_tsn_("Arni")
#' get_tsn_("Arni", rows=1)
#' get_tsn_("Arni", rows=1:2)
#' get_tsn_(c("asdfadfasd","Pinus contorta"), rows=1:5)
#' }

get_tsn <- function(sci_com, searchtype = "scientific", accepted = FALSE,
  ask = TRUE, messages = TRUE, rows = NA, searchterm = NULL, ...) {

  assert(sci_com, c("character", "taxon_state"))
  assert(ask, "logical")
  assert(messages, "logical")
  assert(searchtype, "character")
  assert(accepted, "logical")
  assert_rows(rows)
  if (!is.null(searchterm)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_tsn(searchterm)", with = "get_tsn(sci_com)")
    sci_com <- searchterm
  }
  

  if (inherits(sci_com, "character")) {
    tstate <- taxon_state$new(class = "tsn", names = sci_com)
    items <- sci_com
  } else {
    assert_state(sci_com, "tsn")
    tstate <- sci_com
    sci_com <- tstate$taxa_remaining()
    items <- c(sci_com, tstate$taxa_completed())
  }

  prog <- progressor$new(items = items, suppress = !messages)
  done <- tstate$get()
  for (i in seq_along(done)) prog$completed(names(done)[i], done[[i]]$att)
  prog$prog_start()

  for (i in seq_along(sci_com)) {
    direct <- FALSE
    mssg(messages, "\nRetrieving data for taxon '", sci_com[i], "'\n")

    searchtype <- match.arg(searchtype, c("scientific", "common"))
    tsn_df <- ritis::terms(sci_com[i], what = searchtype, ...)
    mm <- NROW(tsn_df) > 1

    if (!inherits(tsn_df, "tbl_df") || NROW(tsn_df) == 0) {
      tsn <- NA_character_
      att <- "not found"
    } else {
      if ("commonNames" %in% names(tsn_df)) {
        tsn_df$commonNames <-
          sapply(tsn_df$commonNames, function(z) paste0(z, collapse = ","))
      }

      tsn_df <- tsn_df[, c("tsn", "scientificName", "commonNames", "nameUsage")]

      if (accepted) {
        tsn_df <- tsn_df[ tsn_df$nameUsage %in% c("valid", "accepted"), ]
      }

      tsn_df <- sub_rows(tsn_df, rows)

      # should return NA if spec not found
      if (NROW(tsn_df) == 0) {
        mssg(messages, m_not_found_sp_altclass)
        tsn <- NA_character_
        att <- "not found"
      }

      # take the one tsn from data.frame
      if (NROW(tsn_df) == 1) {
        tsn <- tsn_df$tsn
        att <- "found"
      }

      # check for direct match
      if (NROW(tsn_df) > 1) {
        tsn_df <- data.frame(tsn_df, stringsAsFactors = FALSE)
        names(tsn_df)[grep(searchtype, names(tsn_df))] <- "target"
        matchtmp <- tsn_df[tolower(tsn_df$target) %in% tolower(sci_com[i]), "tsn"]
        if (length(matchtmp) == 1) {
          tsn <- matchtmp
          direct <- TRUE
          att <- "found"
        } else {
          direct <- FALSE
          tsn <- NA_character_
          att <- m_na_ask_false_no_direct
          warning("> 1 result; no direct match found", call. = FALSE)
        }
      }

      # multiple matches
      if (any(
        NROW(tsn_df) > 1 && is.na(tsn) |
        NROW(tsn_df) > 1 && att == "found" && length(tsn) > 1
      )) {
        if (ask) {
          names(tsn_df)[grep(searchtype, names(tsn_df))] <- "target"
          # user prompt
          tsn_df <- tsn_df[order(tsn_df$target), ]
          rownames(tsn_df) <- NULL

          # prompt
          message("\n\n")
          message(paste0(utils::capture.output(tsn_df), collapse = "\n"))
          message("\nMore than one TSN found for taxon '", sci_com[i], "'!\n
            Enter rownumber of taxon (other inputs will return 'NA'):\n")
          take <- scan(n = 1, quiet = TRUE, what = "raw")

          if (length(take) == 0) {
            take <- "notake"
            att <- "nothing chosen"
          }
          if (take %in% seq_len(NROW(tsn_df))) {
            take <- as.numeric(take)
            message("Input accepted, took taxon '",
              as.character(tsn_df$target[take]), "'.\n")
            tsn <-  tsn_df$tsn[take]
            att <- "found"
          } else {
            tsn <- NA_character_
            mssg(messages, "\nReturned 'NA'!\n\n")
            att <- "not found"
          }
        } else {
          if (length(tsn) != 1) {
            warning(sprintf(m_more_than_one_found, "tsn", sci_com[i]),
              call. = FALSE)
            tsn <- NA_character_
            att <- m_na_ask_false
          }
        }
      }

    }

    res <- list(id = as.character(tsn), att = att, multiple = mm,
      direct = direct)
    prog$completed(sci_com[i], att)
    prog$prog(att)
    tstate$add(sci_com[i], res)
  }
  out <- tstate$get()
  ids <- structure(as.character(unlist(pluck(out, "id"))), class = "tsn",
                   match = pluck_un(out, "att", ""),
                   multiple_matches = pluck_un(out, "multiple", logical(1)),
                   pattern_match = pluck_un(out, "direct", logical(1)))
  on.exit(prog$prog_summary(), add = TRUE)
  on.exit(tstate$exit, add = TRUE)
  add_uri(ids, get_url_templates$itis)
}

#' @export
#' @rdname get_tsn
as.tsn <- function(x, check=TRUE) UseMethod("as.tsn")

#' @export
#' @rdname get_tsn
as.tsn.tsn <- function(x, check=TRUE) x

#' @export
#' @rdname get_tsn
as.tsn.character <- function(x, check=TRUE) if (length(x) == 1) make_tsn(x, check) else collapse(x, make_tsn, "tsn", check = check)

#' @export
#' @rdname get_tsn
as.tsn.list <- function(x, check=TRUE) if (length(x) == 1) make_tsn(x, check) else collapse(x, make_tsn, "tsn", check = check)

#' @export
#' @rdname get_tsn
as.tsn.numeric <- function(x, check=TRUE) as.tsn(as.character(x), check)

#' @export
#' @rdname get_tsn
as.tsn.data.frame <- function(x, check=TRUE) {
  structure(x$ids, class = "tsn", match = x$match,
            multiple_matches = x$multiple_matches,
            pattern_match = x$pattern_match, uri = x$uri)
}

#' @export
#' @rdname get_tsn
as.data.frame.tsn <- function(x, ...){
  data.frame(ids = as.character(unclass(x)),
             class = "tsn",
             match = attr(x, "match"),
             multiple_matches = attr(x, "multiple_matches"),
             pattern_match = attr(x, "pattern_match"),
             uri = attr(x, "uri"),
             stringsAsFactors = FALSE)
}

make_tsn <- function(x, check=TRUE) {
  make_generic(x, get_url_templates$itis, "tsn", check)
}

check_tsn <- function(x){
  tt <- suppressMessages(itis_getrecord(x))
  identical(tt$acceptedNameList$tsn, as.character(x))
}

#' @export
#' @rdname get_tsn
get_tsn_ <- function(sci_com, messages = TRUE, searchtype = "scientific",
                     accepted = TRUE, rows = NA, searchterm = NULL, ...) {
  if (!is.null(searchterm)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "get_tsn_(searchterm)", with = "get_tsn_(sci_com)")
    sci_com <- searchterm
  }
  
  stats::setNames(
    lapply(sci_com, get_tsn_help, messages = messages,
           searchtype = searchtype, accepted = accepted, rows = rows, ...),
    sci_com
  )
}

get_tsn_help <- function(sci_com, messages, searchtype, accepted,
  rows, ...) {

  mssg(messages, "\nRetrieving data for taxon '", sci_com, "'\n")
  searchtype <- match.arg(searchtype, c("scientific", "common"))
  df <- ritis::terms(sci_com, what = searchtype, ...)
  if (!inherits(df, "tbl_df") || NROW(df) == 0) {
    NULL
  } else {
    df <- df[,c("tsn", "scientificName", "commonNames", "nameUsage")]
    if ("commonNames" %in% names(df)) {
      df$commonNames <-
        sapply(df$commonNames, function(z) paste0(z, collapse = ","))
    }
    if (accepted) df <- df[ df$nameUsage %in% c("valid", "accepted"), ]
    sub_rows(df, rows)
  }
}

context("get_tsn")

test_that("get_tsn returns the correct value", {
  skip_on_cran()
  vcr::use_cassette("get_tsn", {
    x <- get_tsn("asdfasdf", messages=FALSE)[[1]]
    z <- get_tsn("Chironomus riparius", messages=FALSE)
  })

	expect_true(is.na(x))
	expect_is(z, "tsn")
})

test_that("get_tsn accepts ask and verbose arguments", {
  skip_on_cran()
  vcr::use_cassette("get_tsn_ask_verbose_args", {
    # expect_message(sw(get_tsn('Dugesia', messages=TRUE, ask = FALSE)))
    expect_true(all(is.na(sw(get_tsn('black bear', searchtype="common",
      ask=FALSE, messages=FALSE)))))
  })
})

test_that("get_tsn fails as expected", {
  skip_on_cran()

  expect_error(get_tsn(), "argument \"sci_com\" is missing")
  expect_error(get_tsn("Arni", ask = 4, messages = FALSE),
               "ask must be of class logical")
  expect_error(
    get_tsn(sci_com="black bear", searchtype=5,
            messages = FALSE),
    "searchtype must be of class character")
  expect_error(
    get_tsn("Arni", accepted = 34,
            messages = FALSE),
    "accepted must be of class logical")

  # searchtype values
  expect_error(
    get_tsn(sci_com="black bear", searchtype="asdfadf",
            messages = FALSE),
    "'arg' should be one of")

  # rows param
  expect_error(get_tsn("Achlya", rows = "foobar", messages = FALSE),
               "rows must be of class numeric, integer")
  expect_error(get_tsn("Achlya", rows = 0, messages = FALSE),
               "all\\(rows > 0\\) is not TRUE")
})

#' Get taxonomic names for a given rank
#'
#' @export
#' @param sci (character) Vector of taxonomic names to query. required.
#' @param get (character) The ranks of the taxonomic name to get, see
#' [rank_ref]. required.
#' @param db (character) The database to search from: 'itis', 'ncbi' or 'both'.
#' If 'both' both NCBI and ITIS will be queried. Result will be the union of
#' both. If using ncbi, we recommend getting an API key; see 
#' [taxize-authentication]
#' @param pref (character) If db = 'both', sets the preference for the union.
#' Either 'ncbi' (default) or 'itis'. Currently not implemented.
#' @param messages (logical) If `TRUE` the actual taxon queried is printed
#' on the console.
#' @param query Deprecated, see `sci`
#' @param ... Other arguments passed to [get_tsn()] or
#' [get_uid()].
#'
#' @return A data.frame with one column for every queried rank, in addition to
#' a column for db and queried term.
#'
#' @note While [tax_rank()] returns the actual rank of a
#' taxon, [tax_name()] searches and returns any specified rank
#' higher in taxonomy.
#' 
#' @section Authentication:
#' See [taxize-authentication] for help on authentication
#'
#' @seealso [classification()]
#'
#' @examples \dontrun{
#' # A case where itis and ncbi use the same names
#' tax_name(sci = "Helianthus annuus", get = "family", db = "itis")
#' tax_name(sci = "Helianthus annuus", get = "family", db = "ncbi")
#' tax_name(sci = "Helianthus annuus", get = c("genus","family","order"),
#'   db = "ncbi")
#'
#' # Case where itis and ncbi use different names
#' tax_name(sci = "Helianthus annuus", get = "kingdom", db = "itis")
#' tax_name(sci = "Helianthus annuus", get = "kingdom", db = "ncbi")
#'
#' # multiple rank arguments
#' tax_name(sci = c("Helianthus annuus","Baetis rhodani"), get = c("genus",
#' "kingdom"), db = "ncbi")
#' tax_name(sci = c("Helianthus annuus","Baetis rhodani"), get = c("genus",
#' "kingdom"), db = "itis")
#'
#' # query both sources
#' tax_name(sci=c("Helianthus annuus", 'Baetis rhodani'), get=c("genus",
#' "kingdom"), db="both")
#' }

tax_name <- function(sci, get, db = "itis", pref = 'ncbi', messages = TRUE,
                     query = NULL, ...) {

  if (!is.null(query)) {
    lifecycle::deprecate_warn(when = "v0.9.97", what = "tax_name(query)", with = "tax_name(sci)")
    sci <- query
  }
  
  if (missing(get)) stop("you must supply a 'get' value", call. = FALSE)
  db <- match.arg(db, c('itis', 'ncbi', 'both'))
  if (db == 'both' && !pref %in% c('ncbi', 'itis')) {
    stop("if db=both, pref must be either 'itis' or 'ncbi'!\n", call. = FALSE)
  }

  fun <- function(query, get, db, messages, ...){
    # NCBI
    if (db == "ncbi") return( do_ncbi(query, get, messages, ...) )
    # ITIS
  	if (db == "itis") return( do_itis(query, get, messages, ...) )
    # combine both
    if (db == 'both') {
      match_uid <- do_ncbi(query, get, messages, TRUE, ...)
      match_tsn <- do_itis(query, get, messages, TRUE, ...)
      stats::setNames(
        data.frame(rbind(t(c("itis", match_tsn)), t(c("ncbi", match_uid))),
                   stringsAsFactors = FALSE), c("db", "query", get))
    }
  }
  tmp = lapply(sci, fun, get = get, db = db, messages = messages, ...)
  dt2df(tmp, idcol = FALSE)
}

do_ncbi <- function(query, get, messages, both=FALSE, rows = NA, ...) {
  uid <- get_uid(query, messages = messages, rows = rows, ...)
  if (all(is.na(uid))) {
    if (messages) message("No UID found for species '", query, "'!\n")
    if (both) {
      c(query, rep(NA, length(get))) 
    } else {
      stats::setNames(
        data.frame(t(c("ncbi", query, rep(NA, length(get)))), stringsAsFactors = FALSE), 
        c("db", "query", get)
      )
    }
  } else {
    hierarchy <- classification(uid, ...)[[1]]
    if (all(is.na(hierarchy))) return(NULL)
    match <- hierarchy$name[match(tolower(get), tolower(hierarchy$rank))]
    if (both) {
      c(query, match) 
    } else {
      stats::setNames(
        data.frame(t(c("ncbi", query, match)), stringsAsFactors = FALSE), 
        c("db", "query", get)
      )
    }
  }
}

do_itis <- function(query, get, messages, both = FALSE, rows = NA, ...){
  tsn <- get_tsn(query, searchtype = "scientific", messages = messages,
                 rows = rows, ...)
  if (all(is.na(tsn))) {
    if (messages) message("No TSN found for species '", query, "'!\n")
    if (both) {
      c(query, rep(NA, length(get))) 
    } else {
      stats::setNames(
        data.frame(t(c("itis", query, rep(NA, length(get)))), stringsAsFactors = FALSE), 
        c("db", "query", get)
      )
    }
  } else {
    tt <- classification(tsn, messages = messages, ...)[[1]]
    if (all(is.na(tt))) {
      warning(sprintf("%s: no hierarchy data found in ITIS", query), call. = FALSE)
      stats::setNames(
        data.frame(t(c("itis", query, rep(NA, length(get)))), stringsAsFactors = FALSE),
        c("db", "query", get)
      )
    } else {
      if (max(unlist(Filter(function(x) length(x) > 0, sapply(get, which_rank)))) <
          which_rank(tt$rank[1])) {
        warning(sprintf("%s: highest rank of ITIS classification is '%s'", query, tt$rank[1]), 
          call. = FALSE)
      }
      if (!all(tolower(get) %in% tolower(tt$rank))) {
        warning(
          sprintf("%s: rank requested ('%s') not in ITIS classification",
                  query,
                  paste0(tolower(get)[!tolower(get) %in% tolower(tt$rank)], collapse = ", ")),
          call. = FALSE)
      }
      if (both) {
        c(query, tt$name[match(tolower(get), tolower(tt$rank))])
      } else {
        out <- tt[tolower(tt$rank) %in% tolower(get), ]$name
        if (length(out) == 0) out <- rep(NA_character_, length(get))
        rname <- tolower(tt[tolower(tt$rank) %in% tolower(get), ]$rank)
        if (length(rname) == 0) rname <- get
        stats::setNames(
          data.frame(t(c("itis", query, out)), stringsAsFactors = FALSE),
          c("db", "query", rname)
        )
      }
    }
  }
}

# tests for ubio_namebank fxn in taxize
context("ubio_namebank")

out <- ubio_search(searchName = 'elephant', sci = 1, vern = 0)

test_that("ubio_namebank returns the correct value", {
	expect_that(out$scientific[ , 1], matches("6938660"))
})

test_that("ubio_namebank returns the correct class", {
  expect_that(out, is_a("list"))
	expect_that(out$scientific, is_a("data.frame"))
	expect_that(ncol(out$scientific), equals(8))
})

# tests for ubio_search fxn in taxize
context("ubio_search")

out <- ubio_search(searchName = 'elephant', sci = 1, vern = 0)

test_that("ubio_search returns the correct dimensions", {
	expect_that(dim(out$scientific), equals(c(1,8)))
})

test_that("ubio_search returns the correct class", {
	expect_that(out$scientific, is_a("data.frame"))
	expect_that(out$scientific$basionymunit, is_a("character"))
	expect_that(out$scientific$rankname, is_a("character"))
})

test_that("gn_parse", {
  skip_on_cran()
  vcr::use_cassette("gn_parse", {
    tt <- gn_parse("Cyanistes caeruleus")
  })

  expect_is(tt, "data.frame")
  expect_is(tt, "tbl")
  expect_is(tt$normalized, "character")
  expect_is(tt$canonicalName, "data.frame")
  expect_is(tt$details, "list")
  expect_is(tt$positions, "list")
})

test_that("gn_parse fails well", {
  expect_error(gn_parse())
  expect_error(gn_parse(5))
})

test_that("rank_ref", {
  skip_on_cran()
  
  data(rank_ref, package = "taxize")

  expect_is(rank_ref, "data.frame")
  expect_named(rank_ref, c("rankid", "ranks"))
  expect_is(rank_ref$rankid, "character")
  expect_is(rank_ref$ranks, "character")
  expect_gt(NROW(rank_ref), 40)
  
  # rank ids are in order
  z <- sort(as.numeric(rank_ref$rankid))
  expect_identical(z, sort(z))

  # rank name stuff
  w <- rank_ref$ranks
  ## some have commas in them
  expect_true(any(grepl(",", w)))
  ## all are lowercase
  expect_identical(w, tolower(w))
})

---
title: Case study - resolving Species Plantarum names
author: Tim (@fozy81)
date: "2020-02-10"
output: rmarkdown::html_vignette
vignette: >
    %\VignetteIndexEntry{Case study}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
---



*Species Plantarum* by Carl Linnaeus originally published in 1753 and was the first work to consistently describe species using binominal names. It provides a reference point to chart how the usage of 5940 taxa (a subset of 500 used below) described within have changed over time. So let's use the taxize package see if we can resolve these names and see their current status. 

Firstly, let's see how many of the names still exist in modern species catologues today. We can use the `resolve` function to query the Global Name Resolver.


```r
library("taxize")
library("knitr")
# load a subset of Species Plantarum names from the taxize package
names <- species_plantarum_binomials[1:500,]
# Species Plantarum names are split into Genus and Species qualifier so need to be paste them
# together to allow the full binominal names to be sent to the global name resolver
names$species <- paste(names$genus, names$epithet)
# use resolve function to send binominal names to resolver (this may take some time)
resolved_names <- resolve(names$species)
# select only the dataframe column from the list send back from gnr
resolved_names <- resolved_names$gnr
# count the distinct names resolved per dataset source
resolved_names_source <- resolved_names[, 
  c("submitted_name", "data_source_title")
]
resolved_names_source <- unique(resolved_names_source)
resolved_names_source$count <- 1
summary_resolved <- aggregate(resolved_names_source$count,
    list(data_source = resolved_names_source$data_source_title), FUN = "sum")
summary_resolved <- summary_resolved[with(summary_resolved, order(-x)), ]
kable(head(summary_resolved, 20))
```



|   |data_source                                         |   x|
|:--|:---------------------------------------------------|---:|
|40 |uBio NameBank                                       | 499|
|19 |GBIF Backbone Taxonomy                              | 491|
|5  |Catalogue of Life                                   | 481|
|1  |Arctos                                              | 465|
|36 |The International Plant Names Index                 | 460|
|39 |Tropicos - Missouri Botanical Garden                | 422|
|21 |GRIN Taxonomy for Plants                            | 319|
|4  |BioLib.cz                                           | 288|
|11 |Encyclopedia of Life                                | 259|
|33 |Open Tree of Life Reference Taxonomy                | 249|
|35 |The Interim Register of Marine and Nonmarine Genera | 245|
|27 |ITIS                                                | 239|
|12 |EUNIS                                               | 222|
|30 |NCBI                                                | 218|
|31 |nlbif                                               | 184|
|42 |USDA NRCS PLANTS Database                           | 176|
|23 |iNaturalist                                         | 141|
|16 |Freebase                                            | 135|
|41 |Union 4                                             | 133|
|10 |Database of Vascular Plants of Canada (VASCAN)      | 123|

The table above shows none of the data sources provided can resolve all of the names in the *Species Plantarum*. Let's see if between all the data sources we can resolve all the names.


```r
gnr_names <- resolved_names[, c("submitted_name", "score")] 
gnr_names <- unique(gnr_names)

merge_gnr_plantarum <- merge(names, gnr_names,
  by.x = "species", by.y = "submitted_name", all.x = TRUE)

merge_gnr_plantarum <- aggregate(merge_gnr_plantarum$score,
  list(data_source = merge_gnr_plantarum$species), FUN = "mean")

length(merge_gnr_plantarum$data_source)
#> [1] 500
min(merge_gnr_plantarum$x, na.rm = TRUE)
#> [1] 0.75
max(merge_gnr_plantarum$x, na.rm = TRUE)
#> [1] 0.9915
```

Even though no individual name data source can resolve all the names, taken together all names in *Species Plantarum* can be resolved using modern sources. Noting that some names score poorly in terms of fuzzy matching. The poor matching could indicate imperfect matches and perhaps some matches are due to species identified after 1752 sharing similar names. There are lots of questions regarding the history of many of these original binominals and their journey into present day taxonomy.  

For example, let's see how many of these names are still the 'accepted' names according to ITIS. Not all the data providers give an accepted name. So lets use ITIS as the `get_tsn_` function returns a `nameUsage` variable which indicates if a name is still accepted or not.


```r
# find resolved species Plantarum names returned from ITIS
itis_resolved <- resolved_names$submitted_name[resolved_names$data_source_title == "ITIS"]
# select random sub-selection to save time
set.seed(42) 
itis_sample <- itis_resolved[sample(rep(1:length(unique(itis_resolved))), 200)]
# query ITIS (this may take some time)
its_names <- get_tsn_(itis_sample)
its_names_bind <- do.call("rbind", its_names)
its_names_bind$n <- 1
# count the number of names in each nameUsage category 
its_names_agg <- aggregate(its_names_bind$n,
  list(data_source = its_names_bind$nameUsage), FUN = "sum")

kable(its_names_agg)
```



|data_source |   x|
|:-----------|---:|
|accepted    | 195|
|valid       |   3|

Of all the *Species Plantarum* names found in the ITIS data source, a high number (of 200 randomly selected names) are still the 'accepted' name. This indicates Carl Linnaeus, as widely acknowledged, identified well-thought-out and pragmatic taxonomic concepts which have proved useful until the present day. 

Further research may discover why some names appear not to be used today in some data sources and where they are used, why some are no longer the accepted name.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eubon_hierarchy.R
\name{eubon_hierarchy}
\alias{eubon_hierarchy}
\title{EUBON hierarchy}
\usage{
eubon_hierarchy(id, providers = "pesi", timeout = 0, ...)
}
\arguments{
\item{id}{(character) identifier for the taxon. (LSID, DOI, URI, or any
other identifier used by the checklist provider)}

\item{providers}{(character) A list of provider id strings concatenated by
comma characters. The default : "pesi,bgbm-cdm-server[col]" will be used if
this parameter is not set. A list of all available provider ids can be
obtained from the '/capabilities' service end point. Providers can be
nested, that is a parent provider can have sub providers. If the id of the
parent provider is supplied all subproviders will be queried. The query
can also be restricted to one or more subproviders by using the following
syntax: parent-id[sub-id-1,sub-id2,...]}

\item{timeout}{(numeric) The maximum of milliseconds to wait for responses
from any of the providers. If the timeout is exceeded the service will just
return the responses that have been received so far. The default timeout is
0 ms (wait for ever)}

\item{...}{Curl options passed on to \code{\link[httr]{GET}}}
}
\description{
EUBON hierarchy
}
\examples{
\dontrun{
eubon_hierarchy(id = "urn:lsid:marinespecies.org:taxname:126141", 'worms')
eubon_hierarchy(id = "urn:lsid:marinespecies.org:taxname:274350", 'worms')
}
}
\references{
\url{http://cybertaxonomy.eu/eu-bon/utis/1.2/doc.html}
}
\seealso{
Other eubon-methods: \code{\link{eubon_capabilities}},
  \code{\link{eubon_children}}, \code{\link{eubon}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tpl_get.r
\name{tpl_get}
\alias{tpl_get}
\title{Get The Plant List csv files.}
\usage{
tpl_get(x, family = NULL, ...)
}
\arguments{
\item{x}{Directory to write csv files to.}

\item{family}{If you want just one, or >1 family, but not all, list
them in a vector.}

\item{...}{(list) Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
Returns nothing to console, except a message and progress bar.
Writes csv files to x.
}
\description{
Get The Plant List csv files.
}
\details{
Throws a warning if you already have a directory of the one
provided, but still works. Writes to your home directory, change x
as needed.
}
\examples{
\dontrun{
# Get a few families
dir <- file.path(tempdir(), "abc")
tpl_get(dir, family = c("Platanaceae","Winteraceae"))
readLines(file.path(dir, "Platanaceae.csv"), n = 5)

# You can now get Gymnosperms as well
dir1 <- file.path(tempdir(), "def")
tpl_get(dir1, family = c("Pinaceae","Taxaceae"))

# You can get mosses too!
dir2 <- file.path(tempdir(), "ghi")
tpl_get(dir2, family = "Echinodiaceae")

# Get all families
## Beware, will take a while
## dir3 <- file.path(tempdir(), "jkl")
## tpl_get("dir3)
}
}
\references{
The Plant List http://www.theplantlist.org
}
\seealso{
\code{\link[=tpl_families]{tpl_families()}}
}
\author{
John Baumgartner \email{johnbb@student.unimelb.edu.au}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eol_hierarchy.R
\name{eol_hierarchy}
\alias{eol_hierarchy}
\title{Retrieve the taxonomic hierarchy from given EOL taxonID.}
\usage{
eol_hierarchy(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tax_agg.R
\name{tax_agg}
\alias{tax_agg}
\alias{print.tax_agg}
\title{Aggregate species data to given taxonomic rank}
\usage{
tax_agg(x, rank, db = "ncbi", messages = FALSE, ...)

\method{print}{tax_agg}(x, ...)
}
\arguments{
\item{x}{Community data matrix. Taxa in columns, samples in rows.}

\item{rank}{character; Taxonomic rank to aggregate by.}

\item{db}{character; taxonomic API to use, 'ncbi, 'itis' or both, see
\code{\link[=tax_name]{tax_name()}}. Note that each taxonomic data source has
their own identifiers, so that if you provide the wrong \code{db} value
for the identifier you could get a result, but it will likely be wrong (not
what you were expecting). If using ncbi we recommend getting an API key;
see \link{taxize-authentication}}

\item{messages}{(logical) If FALSE (Default) suppress messages}

\item{...}{Other arguments passed to \code{\link[=get_tsn]{get_tsn()}} or \code{\link[=get_uid]{get_uid()}}}
}
\value{
A list of class \code{tax_agg} with the following items:
\itemize{
\item \code{x} Community data matrix with aggregated data.
\item \code{by} A lookup-table showing which taxa were aggregated.
\item \code{n_pre} Number of taxa before aggregation.
\item \code{rank} Rank at which taxa have been aggregated.
}
}
\description{
Aggregate species data to given taxonomic rank
}
\details{
\code{tax_agg} aggregates (sum) taxa to a specific taxonomic level.
If a taxon is not found in the database (ITIS or NCBI) or the supplied taxon
is on higher taxonomic level this taxon is not aggregated.
}
\examples{
\dontrun{
if (requireNamespace("vegan", quietly = TRUE)) {
  # use dune dataset
  data(dune, package='vegan')
  species <- c("Achillea millefolium", "Agrostis stolonifera",
    "Aira praecox", "Alopecurus geniculatus", "Anthoxanthum odoratum",
    "Bellis perennis", "Bromus hordeaceus", "Chenopodium album",
    "Cirsium arvense", "Comarum palustre", "Eleocharis palustris",
    "Elymus repens", "Empetrum nigrum", "Hypochaeris radicata",
    "Juncus articulatus", "Juncus bufonius", "Lolium perenne",
    "Plantago lanceolata", "Poa pratensis", "Poa trivialis",
    "Ranunculus flammula", "Rumex acetosa", "Sagina procumbens",
    "Salix repens", "Scorzoneroides autumnalis", "Trifolium pratense",
    "Trifolium repens", "Vicia lathyroides", "Brachythecium rutabulum",
    "Calliergonella cuspidata")
  colnames(dune) <- species

  # aggregate sample to families
  (agg <- tax_agg(dune, rank = 'family', db = 'ncbi'))

  # extract aggregated community data matrix for further usage
  agg$x
  # check which taxa have been aggregated
  agg$by
}

# A use case where there are different taxonomic levels in the same dataset
spnames <- c('Puma','Ursus americanus','Ursidae')
df <- data.frame(c(1,2,3), c(11,12,13), c(1,4,50))
names(df) <- spnames
out <- tax_agg(x=df, rank = 'family', db='itis')
out$x

# You can input a matrix too
mat <- matrix(c(1,2,3, 11,12,13), nrow = 2, ncol = 3,
 dimnames=list(NULL, c('Puma concolor','Ursus americanus','Ailuropoda melanoleuca')))
tax_agg(mat, rank = 'family', db='itis')
}
}
\seealso{
\link{tax_name}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iucn_getname.R
\name{iucn_getname}
\alias{iucn_getname}
\title{Get any matching IUCN species names}
\usage{
iucn_getname(name, verbose = TRUE, ...)
}
\arguments{
\item{name}{character; taxon name}

\item{verbose}{logical; should messages be printed?}

\item{...}{Further arguments passed on to \code{\link[=iucn_summary]{iucn_summary()}}, note
that you'll need an API key.}
}
\value{
Character vector of names that matched in IUCN
}
\description{
Get any matching IUCN species names
}
\details{
Beware: IUCN functions can give back incorrect data. This isn't
our fault. We do our best to get you the correct data quickly, but sometimes
IUCN gives back the wrong data, and sometimes Global Names gives back the
wrong data. We will fix these as soon as possible. In the meantime, just
make sure that the data you get back is correct.
}
\examples{
\dontrun{
iucn_getname(name = "Cyanistes caeruleus")
iucn_getname(name = "Panthera uncia")

# not found in IUCN search
iucn_getname(name = "Acacia allenii")
}
}
\seealso{
\code{\link[=iucn_summary]{iucn_summary()}} \code{\link[=iucn_status]{iucn_status()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iucn_id.R
\name{iucn_id}
\alias{iucn_id}
\title{Get an ID for a IUCN listed taxon}
\usage{
iucn_id(sciname, key = NULL, ...)
}
\arguments{
\item{sciname}{character; Scientific name. Should be cleaned and in the
format \verb{*<Genus> <Species>*}. One or more.}

\item{key}{(character) required. you IUCN Redlist API key. See
\link[rredlist:rredlist-package]{rredlist::rredlist-package} for help on authenticating with
IUCN Redlist}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
A named list (names are input taxa names) of one or more IUCN IDs.
Taxa that aren't found are silently dropped.
}
\description{
Get an ID for a IUCN listed taxon
}
\examples{
\dontrun{
iucn_id("Branta canadensis")
iucn_id("Branta bernicla")
iucn_id("Panthera uncia")
iucn_id("Lynx lynx")

# many names
iucn_id(c("Panthera uncia", "Lynx lynx"))

# many names, some not found
iucn_id(c("Panthera uncia", "Lynx lynx", "foo bar", "Gorilla gorilla gorilla"))

# a name not found
iucn_id("Foo bar")
}
}
\author{
Scott Chamberlain,
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gnr_datasources.R
\name{gnr_datasources}
\alias{gnr_datasources}
\title{Global Names Resolver Data Sources}
\usage{
gnr_datasources(..., todf)
}
\arguments{
\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}

\item{todf}{defunct, always get a data.frame back now}
}
\value{
data.frame/tibble
}
\description{
Retrieve data sources used in the Global Names Resolver
}
\examples{
\dontrun{
# all data sources
gnr_datasources()

# give me the id for EOL
out <- gnr_datasources()
out[out$title == "EOL", "id"]

# Fuzzy search for sources with the word zoo
out <- gnr_datasources()
out[agrep("zoo", out$title, ignore.case = TRUE), ]
}
}
\references{
https://resolver.globalnames.org/data_sources
}
\seealso{
\code{\link[=gnr_resolve]{gnr_resolve()}}, \code{\link[=gna_search]{gna_search()}}
}
\keyword{names}
\keyword{resolve}
\keyword{taxonomy}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/downstream.R
\name{downstream}
\alias{downstream}
\alias{downstream.default}
\alias{downstream.tsn}
\alias{downstream.gbifid}
\alias{downstream.uid}
\alias{downstream.wormsid}
\alias{downstream.boldid}
\alias{downstream.ids}
\title{Retrieve the downstream taxa for a given taxon name or ID.}
\usage{
downstream(...)

\method{downstream}{default}(
  sci_id,
  db = NULL,
  downto = NULL,
  intermediate = FALSE,
  rows = NA,
  x = NULL,
  ...
)

\method{downstream}{tsn}(sci_id, db = NULL, downto = NULL, intermediate = FALSE, ...)

\method{downstream}{gbifid}(
  sci_id,
  db = NULL,
  downto = NULL,
  intermediate = FALSE,
  limit = 100,
  start = NULL,
  ...
)

\method{downstream}{uid}(sci_id, db = NULL, downto = NULL, intermediate = FALSE, ...)

\method{downstream}{wormsid}(sci_id, db = NULL, downto = NULL, intermediate = FALSE, ...)

\method{downstream}{boldid}(sci_id, db = NULL, downto = NULL, intermediate = FALSE, ...)

\method{downstream}{ids}(sci_id, db = NULL, downto = NULL, intermediate = FALSE, ...)
}
\arguments{
\item{...}{Further args passed on to \code{\link[=itis_downstream]{itis_downstream()}},
\code{\link[=gbif_downstream]{gbif_downstream()}}, \code{\link[=ncbi_downstream]{ncbi_downstream()}},
\code{\link[=worms_downstream]{worms_downstream()}}, or \code{\link[=bold_downstream]{bold_downstream()}}}

\item{sci_id}{Vector of taxa names (character) or IDs (character or numeric)
to query.}

\item{db}{character; database to query. One or more of \code{itis}, \code{gbif},
\code{ncbi}, \code{worms}, or \code{bold}. Note that each taxonomic  data source has
their own identifiers, so that if you provide the wrong \code{db} value for
the identifier
you could get a result, but it will likely be wrong (not what you were
expecting). If using ncbi, we recommend getting an API key; see
\link{taxize-authentication}}

\item{downto}{What taxonomic rank to go down to. One of: 'superkingdom',
'kingdom', 'subkingdom','infrakingdom','phylum','division','subphylum',
'subdivision','infradivision', 'superclass','class','subclass','infraclass',
'superorder','order','suborder','infraorder','superfamily','family',
'subfamily','tribe','subtribe','genus','subgenus','section','subsection',
'species group','species','subspecies','variety','form','subvariety','race',
'stirp', 'morph','aberration','subform', 'unspecified', 'no rank'}

\item{intermediate}{(logical) If \code{TRUE}, return a list of length two
with target taxon rank names, with additional list of data.frame's of
intermediate taxonomic groups. Default: \code{FALSE}}

\item{rows}{(numeric) Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this parameter is ignored if you pass in a
taxonomic id of any of the acceptable classes: tsn.}

\item{x}{Deprecated, see \code{sci_id}}

\item{limit}{Number of records to return. Applies to gbif only.
default: 100. max: 1000. use in combination with the \code{start} parameter}

\item{start}{Record number to start at. Applies to gbif only. default: 0.
use in combination with the \code{limit} parameter}
}
\value{
A named list of data.frames with the downstream names of every
supplied taxa. You get an NA if there was no match in the database.
}
\description{
This function uses a while loop to continually collect children taxa down
to the taxonomic rank that you specify in the \code{downto} parameter. You
can get data from ITIS (itis), GBIF (gbif), NCBI (ncbi), WORMS (worms),
or BOLD (bold). There is no method exposed by these four
services for getting taxa at a specific taxonomic rank, so we do it
ourselves here.
}
\section{Authentication}{

See \link{taxize-authentication} for help on authentication
}

\section{bold}{

BEWARE: \code{db="bold"} scrapes the BOLD website, so may be unstable. That is,
one day it may work, and the next it may fail. Open an issue if you
encounter an error: https://github.com/ropensci/taxize/issues
}

\examples{
\dontrun{
# Plug in taxon IDs
downstream(125732, db = 'worms', downto = 'species')
downstream(3451, db = 'bold', downto = 'species')

if (interactive()) {

# Plug in taxon names
downstream("Apis", db = 'ncbi', downto = 'species')
downstream("Apis", db = 'itis', downto = 'species')
downstream("Apis", db = 'bold', downto = 'species')
downstream("Gadus", db = 'worms', downto = 'species')
downstream(c("Apis","Epeoloides"), db = 'itis', downto = 'species')
downstream("Ursus", db = 'gbif', downto = 'species')
downstream(get_gbifid("Ursus"), db = 'gbif', downto = 'species')

# Many taxa
sp <- names_list("genus", 3)
downstream(sp, db = 'itis', downto = 'species')
downstream(sp, db = 'gbif', downto = 'species')

# Both data sources
ids <- get_ids("Apis", db = c('gbif','itis'))
downstream(ids, downto = 'species')
## same result
downstream(get_ids("Apis", db = c('gbif','itis')), downto = 'species')

# Collect intermediate names
## itis
downstream('Bangiophyceae', db="itis", downto="genus")
downstream('Bangiophyceae', db="itis", downto="genus", intermediate=TRUE)
downstream(get_tsn('Bangiophyceae'), downto="genus")
downstream(get_tsn('Bangiophyceae'), downto="genus", intermediate=TRUE)

# Use the rows parameter
## note how in the second function call you don't get the prompt
downstream("Poa", db = 'gbif', downto="species")
downstream("Poa", db = 'gbif', downto="species", rows=1)

# use curl options
res <- downstream("Apis", db = 'gbif', downto = 'species', verbose = TRUE)

# Pagination
# GBIF limits queries to a maximum of 1000 records per request, so if
# there's more than 1000, use the start parameter
# Piper, taxonKey = 3075433
z1 <- downstream(3075433, db = 'gbif', downto = "species", limit=1000)
z2 <- downstream(3075433, db = 'gbif', downto = "species", limit=1000,
  start=1000)
z3 <- downstream(3075433, db = 'gbif', downto = "species", limit=1000,
  start=2000)
z4 <- downstream(3075433, db = 'gbif', downto = "species", limit=1000,
  start=3000)
NROW(rbind(z1[[1]], z2[[1]], z3[[1]], z4[[1]]))
}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tol_resolve.R
\name{tol_resolve}
\alias{tol_resolve}
\title{Resolve names using Open Tree of Life (OTL) resolver}
\usage{
tol_resolve(
  names = NULL,
  context_name = NULL,
  do_approximate_matching = TRUE,
  ids = NULL,
  include_suppressed = FALSE,
  ...
)
}
\arguments{
\item{names}{(character) taxon names to be queried}

\item{context_name}{name of the taxonomic context to be searched
(length-one character vector). Must match (case sensitive) one
of the values returned by \code{\link[rotl:tnrs_contexts]{rotl::tnrs_contexts()}}.}

\item{do_approximate_matching}{(logical) A logical indicating whether or not
to perform approximate string (a.k.a. \dQuote{fuzzy}) matching. Using \code{FALSE}
will greatly improve speed. Default: \code{TRUE}}

\item{ids}{An array of OTL ids to use for identifying names. These
will be assigned to each name in the names array. If ids is
provided, then ids and names must be identical in length.}

\item{include_suppressed}{(logical) Ordinarily, some quasi-taxa, such as
incertae sedis buckets and other non-OTUs, are suppressed from TNRS
results. If this parameter is true, these quasi-taxa are allowed as
possible TNRS results. Default: \code{FALSE}}

\item{...}{Curl options passed on to \code{httr::POST} within
\code{\link[rotl:tnrs_match_names]{rotl::tnrs_match_names()}}}
}
\value{
A data frame summarizing the results of the query. The
original query output is appended as an attribute to the
returned object (and can be obtained using
\code{attr(object, "original_response")}).
}
\description{
Resolve names using Open Tree of Life (OTL) resolver
}
\examples{
\dontrun{
tol_resolve(names=c("echinodermata", "xenacoelomorpha",
  "chordata", "hemichordata"))
tol_resolve(c("Hyla", "Salmo", "Diadema", "Nautilus"))
tol_resolve(c("Hyla", "Salmo", "Diadema", "Nautilus"),
  context_name = "Animals")

turducken_spp <- c("Meleagris gallopavo", "Anas platyrhynchos",
  "Gallus gallus")
tol_resolve(turducken_spp, context_name="Animals")
}
}
\references{
https://github.com/OpenTreeOfLife/germinator/wiki/TNRS-API-v3#match_names
}
\seealso{
\code{\link[=gnr_resolve]{gnr_resolve()}}, \code{\link[=tnrs]{tnrs()}}
}
\author{
Francois Michonneau \email{francois.michonneau@gmail.com}
Scott Chamberlain
}
\keyword{names}
\keyword{resolve}
\keyword{taxonomy}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gna_parse.R
\name{gni_parse}
\alias{gni_parse}
\title{Parse scientific names using EOL's name parser.}
\usage{
gni_parse(names, ...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resolve.R
\name{resolve}
\alias{resolve}
\title{Resolve names from different data sources}
\usage{
resolve(sci, db = "gnr", query = NULL, ...)
}
\arguments{
\item{sci}{Vector of one or more taxonomic names (common names not
supported)}

\item{db}{Source to check names against. One of iplant or gnr.
Default: gnr. Note that each taxonomic data source has their own
identifiers, so that if you provide the wrong \code{db} value for the
identifier you could get a result, but it will likely be wrong (not
what you were expecting).}

\item{query}{Deprecated, see \code{sci}}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET} or
\link[crul:verb-POST]{crul::verb-POST}. In addition, further named args passed
on to  each respective function. See examples}
}
\value{
A list with length equal to length of the db parameter (number
of sources requested), with each element being a data.frame or list
with results from that source.
}
\description{
Resolve names from iPlant's name resolver, and
the Global Names Resolver (GNR)
}
\examples{
\dontrun{
resolve(sci=c("Helianthus annuus", "Homo sapiens"))
resolve(sci="Quercus keloggii", db='gnr')
resolve(sci=c("Helianthus annuus", "Homo sapiens"), db=c('iplant', 'gnr'))
resolve(sci="Quercus keloggii", db=c('iplant', 'gnr'))

# pass in options specific to each source
resolve("Helianthus annuus", db = 'gnr', preferred_data_sources = c(3, 4))
resolve("Helianthus annuus", db = 'iplant', retrieve = 'best')
identical(
 resolve("Helianthus annuus", db = 'iplant', retrieve = 'best')$iplant,
 iplant_resolve("Helianthus annuus", retrieve = 'best')
)

# pass in curl options
resolve(sci="Qercuss", db = "iplant", verbose = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getkey.R
\name{getkey}
\alias{getkey}
\title{Function to get API key.}
\usage{
getkey(x = NULL, service)
}
\arguments{
\item{x}{(character) An API key, defaults to \code{NULL}}

\item{service}{(character) The API data provider, used to match to
default guest key (for Tropicos; there's no guest
key for NCBI or IUCN, for which you have to get your own)}
}
\description{
Checks first to get key from your .Rprofile or .Renviron (or similar) file
}
\examples{
\dontrun{
getkey(service="tropicos")
getkey(service="iucn")
getkey(service="entrez")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gna_search.R
\name{gna_search}
\alias{gna_search}
\title{Search for taxonomic names using the Global Names Architecture}
\usage{
gna_search(sci, justtotal = FALSE, parse_names = FALSE, ...)
}
\arguments{
\item{sci}{(character) required. Name pattern you want to search
for. WARNING: Does not work for common names. Search term may include
following options:
\itemize{
\item \code{n}: A shortcut that allows to put together several elements (e.g., \verb{n:B. bubo Linn. 1750-1800})
\item \code{g}: a genus name. (e.g. \code{g:B.}, \code{g:Bub.}, \code{g:Bubo})
\item \code{isp}: an infraspecies name (e.g. \code{sp:bubo}, \code{sp:gallop.})
\item \code{asp}: either species or infraspecies (all sp) (e.g. \code{asp:bubo})
\item \code{ds}: data-sources IDs (e.g., \verb{ds:1,2,3})
\item \code{tx}: parent taxon . Uses classification of the first data-source from \code{ds}. If data-sources are not set, uses Catalogue of Life. (e.g. \code{tx:Aves})
\item \code{au}: author - Search by author word    (e.g. \code{au:Linnaeus}, \code{au:Linn.})
\item \code{y}: year - Search by year (e.g. \code{y:2005})
}}

\item{justtotal}{Return only the total results found.}

\item{parse_names}{If \code{TRUE} use \code{\link[=gni_parse]{gni_parse()}} on the outputs.}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
data.frame of results.
}
\description{
Uses the Global Names Index, see http://gni.globalnames.org
}
\examples{
\dontrun{
gna_search('n:B. bubo ds:1,2 au:Linn. y:1700-')
}
}
\references{
http://gni.globalnames.org/
https://apidoc.globalnames.org/gnames
}
\seealso{
\code{\link[=gnr_datasources]{gnr_datasources()}}, \code{\link[=gna_search]{gna_search()}}
}
\author{
Scott Chamberlain, Zachary Foster
}
\keyword{globalnamesindex}
\keyword{names}
\keyword{taxonomy}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apg.R
\name{apg}
\alias{apg}
\alias{apgOrders}
\alias{apgFamilies}
\title{Get APG names}
\usage{
apgOrders(...)

apgFamilies(...)
}
\arguments{
\item{...}{Curl args passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\description{
Generic names and their replacements from the Angiosperm Phylogeny
Group III system of flowering plant classification.
}
\examples{
\dontrun{
head(apgOrders())
head(apgFamilies())
}
}
\references{
http://www.mobot.org/MOBOT/research/APweb/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gni_details.R
\name{gni_details}
\alias{gni_details}
\title{Search for taxonomic name details using the Global Names Index}
\usage{
gni_details(id, all_records = 1, ...)
}
\arguments{
\item{id}{Name id. Required.}

\item{all_records}{If all_records is 1, GNI returns all records from all
repositories for the name string (takes 0, or 1 [default]).}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
Data.frame of results.
}
\description{
Uses the Global Names Index, see http://gni.globalnames.org/
}
\examples{
\dontrun{
gni_details(id = 17802847)

# pass on curl options
gni_details(id = 17802847, verbose = TRUE)
}
}
\seealso{
\code{\link[=gnr_datasources]{gnr_datasources()}}, \code{\link[=gna_search]{gna_search()}}.
}
\author{
Scott Chamberlain
}
\keyword{globalnamesindex}
\keyword{names}
\keyword{taxonomy}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phylomatic_format.R
\name{phylomatic_format}
\alias{phylomatic_format}
\title{Get family names to make Phylomatic input object, and output input string
to Phylomatic for use in the function phylomatic_tree.}
\usage{
phylomatic_format(...)
}
\arguments{
\item{...}{Parameters, ignored}
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\name{taxize-package}
\alias{taxize-package}
\alias{taxize}
\title{Taxonomic Information from Around the Web}
\description{
This package interacts with a suite of web 'APIs'
for taxonomic tasks, such as verifying species names, getting
taxonomic hierarchies, and verifying name spelling.
}
\section{About}{

Allows users to search over many websites for species names (scientific and
common) and download up- and downstream taxonomic hierarchical information -
and many other things.

The functions in the package that hit a specific API have a prefix and suffix
separated by an underscore. They follow the format of \code{service_whatitdoes}.
For example, \code{gnr_resolve} uses the Global Names Resolver API to resolve
species names.

General functions in the package that don't hit a specific API don't have
two words separated by an underscore, e.g., \code{classification}

You need API keys for some data sources. See \link{taxize-authentication}
for more information.
}

\section{Currently supported APIs}{


\tabular{llc}{
API \tab prefix \tab SOAP? \cr
Encyclopedia of Life (EOL)  \tab eol \tab FALSE \cr
Integrated Taxonomic Information Service (ITIS)  \tab itis \tab FALSE \cr
Global Names Resolver (from EOL/GBIF) \tab gnr \tab FALSE \cr
Global Names Index (from EOL/GBIF) \tab gni \tab FALSE \cr
IUCN Red List \tab iucn \tab FALSE \cr
Tropicos (from Missouri Botanical Garden) \tab tp \tab FALSE \cr
Theplantlist.org \tab tpl \tab FALSE \cr
National Center for Biotechnology Information \tab ncbi \tab FALSE \cr
CANADENSYS Vascan name search API \tab vascan \tab FALSE \cr
International Plant Names Index (IPNI) \tab ipni \tab FALSE \cr
World Register of Marine Species (WoRMS) \tab worms \tab TRUE \cr
Barcode of Life Data Systems (BOLD) \tab bold \tab FALSE \cr
Pan-European Species directories Infrastructure (PESI) \tab pesi \tab TRUE \cr
Mycobank \tab myco \tab TRUE \cr
National Biodiversity Network (UK)	\tab nbn \tab FALSE \cr
Index Fungorum \tab fg \tab FALSE \cr
EU BON \tab eubon \tab FALSE \cr
Index of Names (ION) \tab ion \tab FALSE \cr
Open Tree of Life (TOL) \tab tol \tab FALSE \cr
World Register of Marine Species (WoRMS) \tab worms \tab FALSE \cr
NatureServe \tab natserv \tab FALSE \cr
}

If the source above has a \code{TRUE} in the \verb{SOAP?} column, it is not available
in this package. They are available from a different package called \strong{taxizesoap}.
See the GitHub repo for how to install https://github.com/ropensci/taxizesoap
}

\section{Catalogue of Life (COL)}{

COL introduced rate limiting recently in 2019 - which has made the API
essentially unusable - CoL+ is coming soon and we'll incorporate it here
when it's stable. See https://github.com/ropensci/colpluz for the
R implementation for CoL+
}

\author{
Scott Chamberlain

Eduard Szoecs \email{eduardszoecs@gmail.com}

Zachary Foster \email{zacharyfoster1989@gmail.com}

Carl Boettiger \email{cboettig@gmail.com}

Karthik Ram \email{karthik@ropensci.org}

Ignasi Bartomeus \email{nacho.bartomeus@gmail.com}

John Baumgartner \email{johnbb@student.unimelb.edu.au}

James O'Donnell \email{jodonnellbio@gmail.com}
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ipni_search.R
\name{ipni_search}
\alias{ipni_search}
\title{Search for names in the International Plant Names Index (IPNI).}
\usage{
ipni_search(
  family = NULL,
  infrafamily = NULL,
  genus = NULL,
  infragenus = NULL,
  species = NULL,
  infraspecies = NULL,
  publicationtitle = NULL,
  authorabbrev = NULL,
  includepublicationauthors = NULL,
  includebasionymauthors = NULL,
  geounit = NULL,
  addedsince = NULL,
  modifiedsince = NULL,
  isapnirecord = NULL,
  isgcirecord = NULL,
  isikrecord = NULL,
  ranktoreturn = NULL,
  output = "minimal",
  ...
)
}
\arguments{
\item{family}{Family name to search on (Optional)}

\item{infrafamily}{Infrafamilial name to search on (Optional)}

\item{genus}{Genus name to search on (Optional)}

\item{infragenus}{Infrageneric name to search on (Optional)}

\item{species}{Species name to search on (Optional) - Note, this is the
epithet, not the full genus - epithet name combination.}

\item{infraspecies}{Infraspecies name to search on (Optional)}

\item{publicationtitle}{Publication name or abbreviation to search on.
Again, replace any spaces with a '+' (e.g. 'J.+Bot.') (Optional)}

\item{authorabbrev}{Author standard form to search on (publishing author,
basionym author or both - see below) (Optional)}

\item{includepublicationauthors}{TRUE (default) to include the taxon author
in the search or FALSE to exclude it}

\item{includebasionymauthors}{TRUE (default) to include the basionum author
in the search or FALSE to exclude it}

\item{geounit}{Country name or other geographical unit to search on (see
the help pages for more information and warnings about the use of this
option) (Optional)}

\item{addedsince}{Date to search on in the format 'yyyy-mm-dd', e.g.
2005-08-01 for all records added since the first of August, 2005. (see the
help pages for more information and warnings about the use of this option)
(Optional. If supplied must be in format YYYY-MM-DD and must be greater
than or equal to 1984-01-01.)}

\item{modifiedsince}{Date to search on in the format 'yyyy-mm-dd', e.g.
2005-08-01 for all records edited since the first of August, 2005. (See
the help pages for more information about the use of this option)
(Optional. If supplied must be in format YYYY-MM-DD and must be greater
than or equal to 1993-01-01.)}

\item{isapnirecord}{FALSE (default) to exclude records from the Australian
Plant Name Index}

\item{isgcirecord}{FALSE (default) to exclude records from the Gray Cards
Index}

\item{isikrecord}{FALSE (default) to exclude records from the Index Kewensis}

\item{ranktoreturn}{One of a few options to choose the ranks returned. See
details.}

\item{output}{One of minimal (default), classic, short, or extended}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}
(Optional). Default: returns all ranks.}
}
\value{
a tibble (data.frame)
}
\description{
Note: This data source is also provided in the Global Names Index (GNI)
(http://gni.globalnames.org/data_sources). The interface to the data is
different among the two services though.
}
\details{
\code{ranktoreturn} options:
\itemize{
\item "all" - all records
\item "fam" - family records
\item "infrafam" - infrafamilial records
\item "gen" - generic records
\item "infragen" - infrageneric records
\item "spec" - species records
\item "infraspec" - infraspecific records
}
}
\examples{
\dontrun{
ipni_search(genus='Brintonia', isapnirecord=TRUE, isgcirecord=TRUE,
  isikrecord=TRUE)
ipni_search(genus='Ceanothus')
ipni_search(genus='Pinus', species='contorta')

# Different output formats
ipni_search(genus='Ceanothus')
ipni_search(genus='Ceanothus', output='short')
ipni_search(genus='Ceanothus', output='extended')
}
}
\references{
https://web.archive.org/web/20190501132148/http://www.ipni.org/link_to_ipni.html
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize_capwords.r
\name{taxize_capwords}
\alias{taxize_capwords}
\title{Capitalize the first letter of a character string.}
\usage{
taxize_capwords(s, strict = FALSE, onlyfirst = FALSE)
}
\arguments{
\item{s}{A character string}

\item{strict}{Should the algorithm be strict about capitalizing.
Defaults to FALSE.}

\item{onlyfirst}{Capitalize only first word, lowercase all others. Useful
for taxonomic names.}
}
\description{
Capitalize the first letter of a character string.
}
\examples{
taxize_capwords(c("using AIC for model selection"))
taxize_capwords(c("using AIC for model selection"), strict=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gisd_isinvasive.R
\name{gisd_isinvasive}
\alias{gisd_isinvasive}
\title{Check invasive species status for a set of species from GISD database}
\usage{
gisd_isinvasive(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tnrs.R
\name{tnrs}
\alias{tnrs}
\title{Phylotastic Taxonomic Name Resolution Service.}
\usage{
tnrs(...)
}
\arguments{
\item{...}{ignored}
}
\description{
Defunct - service was down far too much to be reliable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize_options.R
\name{taxize_options}
\alias{taxize_options}
\title{taxize options}
\usage{
taxize_options(taxon_state_messages = NULL, ncbi_sleep = NULL, quiet = FALSE)
}
\arguments{
\item{taxon_state_messages}{(logical) suppress messages? default: \code{NULL}
(same as setting \code{FALSE}). Set to \code{TRUE} to suppress messages, and \code{FALSE}
to not suppress messages}

\item{ncbi_sleep}{(numeric/integer) number of seconds to sleep between
NCBI ENTREZ http requests. applies to the functions: \code{\link[=classification]{classification()}},
\code{\link[=comm2sci]{comm2sci()}}, \code{\link[=genbank2uid]{genbank2uid()}}, \code{\link[=get_uid]{get_uid()}} and \code{\link[=ncbi_children]{ncbi_children()}}.
defaults: 0.334 (without API key) or 0.101 (with API key). minimum value
can not be less than 0.101}

\item{quiet}{(logical) quiet informational output from this function.
default: \code{TRUE}}
}
\description{
taxize options
}
\examples{
\dontrun{
taxize_options()
taxize_options(FALSE)
taxize_options(TRUE)
taxize_options(ncbi_sleep = 0.4)
taxize_options(taxon_state_messages = TRUE, ncbi_sleep = 0.4)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nbn_classification.R
\name{nbn_classification}
\alias{nbn_classification}
\title{Search UK National Biodiversity Network database for
taxonomic classification}
\usage{
nbn_classification(id, ...)
}
\arguments{
\item{id}{(character) An NBN identifier.}

\item{...}{Further args passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A data.frame
}
\description{
Search UK National Biodiversity Network database for
taxonomic classification
}
\examples{
\dontrun{
nbn_classification(id="NHMSYS0000376773")

# get id first, then pass to this fxn
id <- get_nbnid("Zootoca vivipara", rec_only = TRUE, rank = "Species")
nbn_classification(id)

nbn_classification(id="NHMSYS0000502940", verbose = TRUE)
}
}
\references{
https://api.nbnatlas.org/
}
\seealso{
Other nbn: 
\code{\link{get_nbnid}()},
\code{\link{nbn_search}()},
\code{\link{nbn_synonyms}()}
}
\author{
Scott Chamberlain,
}
\concept{nbn}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_acceptname.R
\name{itis_acceptname}
\alias{itis_acceptname}
\title{Retrieve accepted TSN and name}
\usage{
itis_acceptname(searchtsn, ...)
}
\arguments{
\item{searchtsn}{One or more TSN for a taxon (numeric/integer)}

\item{...}{Curl options passed on to\link[crul:verb-GET]{crul::verb-GET}}
}
\value{
data.frame with with row number equal to input vector length, and
with three columns:
\itemize{
\item submittedtsn (numeric) - The submitted TSN
\item acceptedname (character) - The accepted name - if the submitted TSN
is the accepted TSN, then this is \code{NA_character_} because ITIS does
not return a name along with the TSN if it's an accepted name. We could
make an extra HTTP request to ITIS, but that means additional time.
\item acceptedtsn (numeric) - The accepted TSN
\item author (character) - taxonomic authority
}
}
\description{
Retrieve accepted TSN and name
}
\examples{
\dontrun{
# TSN accepted - good name
itis_acceptname(searchtsn = 208527)

# TSN not accepted - input TSN is old
itis_acceptname(searchtsn = 504239)

# many accepted names
ids <- c(18161, 18162, 18163, 18164, 18165, 18166, 46173, 46174,
46178, 46181, 46186, 46193, 46196, 46197, 46200, 46201, 46204,
46207, 46867, 46868)
itis_acceptname(searchtsn = ids)

# many unaccepted names
ids <- c(39087, 46208, 46973, 46976, 46978, 46980, 47295, 47445,
47448, 47512, 47515, 47527, 47546, 47622, 47783, 47786, 47787,
47788, 47835, 47839)
itis_acceptname(searchtsn = ids)

# many: mix of accepted and unaccepted names
ids <- c(18161, 18162, 47527, 47546, 47622, 46200)
itis_acceptname(searchtsn = ids)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_tpsid.R
\name{get_tpsid}
\alias{get_tpsid}
\alias{as.tpsid}
\alias{as.tpsid.tpsid}
\alias{as.tpsid.character}
\alias{as.tpsid.list}
\alias{as.tpsid.numeric}
\alias{as.tpsid.data.frame}
\alias{as.data.frame.tpsid}
\alias{get_tpsid_}
\title{Get the NameID codes from Tropicos for taxonomic names.}
\usage{
get_tpsid(
  sci,
  ask = TRUE,
  messages = TRUE,
  key = NULL,
  rows = NA,
  family = NULL,
  rank = NULL,
  sciname = NULL,
  ...
)

as.tpsid(x, check = TRUE)

\method{as.tpsid}{tpsid}(x, check = TRUE)

\method{as.tpsid}{character}(x, check = TRUE)

\method{as.tpsid}{list}(x, check = TRUE)

\method{as.tpsid}{numeric}(x, check = TRUE)

\method{as.tpsid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{tpsid}(x, ...)

get_tpsid_(sci, messages = TRUE, key = NULL, rows = NA, sciname = NULL, ...)
}
\arguments{
\item{sci}{(character) One or more scientific name's as a vector or list. Or,
a \code{taxon_state} object (see \link{taxon-state})}

\item{ask}{logical; should get_tpsid be run in interactive mode?
If TRUE and more than one ID is found for the species, the user is asked for
input. If FALSE NA is returned for multiple matches.}

\item{messages}{logical; If TRUE the actual taxon queried is printed on the console.}

\item{key}{Your API key; see \link{taxize-authentication}}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all rows are considered.
Note that this function still only gives back a tpsid class object with one to many identifiers.
See \code{\link[=get_tpsid_]{get_tpsid_()}} to get back all, or a subset, of the raw data that you are
presented during the ask process.}

\item{family}{(character) A family name. Optional. See \code{Filtering} below.}

\item{rank}{(character) A taxonomic rank name. See \link{rank_ref} for possible
options. Though note that some data sources use atypical ranks, so inspect the
data itself for options. Optional. See \code{Filtering} below.}

\item{sciname}{Deprecated, see \code{sci}}

\item{...}{Other arguments passed to \code{\link[=tp_search]{tp_search()}}.}

\item{x}{Input to \code{\link[=as.tpsid]{as.tpsid()}}}

\item{check}{logical; Check if ID matches any existing on the DB, only used in
\code{\link[=as.tpsid]{as.tpsid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Get the NameID codes from Tropicos for taxonomic names.
}
\section{Filtering}{

The parameters \code{family} an\code{rank}nk are not used in the search to the data
provider, but are used in filtering the data down to a subset that is closer to the
target you want.  For all these parameters,
you can use regex strings since we use \code{\link[=grep]{grep()}} internally to match.
Filtering narrows down to the set that matches your query, and removes the rest.
}

\examples{
\dontrun{
get_tpsid(sci='Poa annua')
get_tpsid(sci='Pinus contorta')

get_tpsid(c("Poa annua", "Pinus contorta"))

# specify rows to limit choices available
get_tpsid('Poa ann')
get_tpsid('Poa ann', rows=1)
get_tpsid('Poa ann', rows=25)
get_tpsid('Poa ann', rows=1:2)

# When not found, NA given (howdy is not a species name, and Chrinomus is a fly)
get_tpsid("howdy")
get_tpsid(c("Chironomus riparius", "howdy"))

# Narrow down results to a division or rank, or both
## Satyrium example
### Results w/o narrowing
get_tpsid("Satyrium")
### w/ rank
get_tpsid("Satyrium", rank = "var.")
get_tpsid("Satyrium", rank = "sp.")

## w/ family
get_tpsid("Poa")
get_tpsid("Poa", family = "Iridaceae")
get_tpsid("Poa", family = "Orchidaceae")
get_tpsid("Poa", family = "Orchidaceae", rank = "gen.")

# Fuzzy filter on any filtering fields
## uses grep on the inside
get_tpsid("Poa", family = "orchidaceae")
get_tpsid("Aga", fuzzy = TRUE, parent = "*idae")

# pass to classification function to get a taxonomic hierarchy
classification(get_tpsid(sci='Poa annua'))

# Convert a tpsid without class information to a tpsid class
as.tpsid(get_tpsid("Pinus contorta")) # already a tpsid, returns the same
as.tpsid(get_tpsid(c("Chironomus riparius","Pinus contorta"))) # same
as.tpsid(24900183) # numeric
as.tpsid(c(24900183,50150089,50079838)) # numeric vector, length > 1
as.tpsid("24900183") # character
as.tpsid(c("24900183","50150089","50079838")) # character vector, length > 1
as.tpsid(list("24900183","50150089","50079838")) # list, either numeric or character
## dont check, much faster
as.tpsid("24900183", check=FALSE)
as.tpsid(24900183, check=FALSE)
as.tpsid(c("24900183","50150089","50079838"), check=FALSE)
as.tpsid(list("24900183","50150089","50079838"), check=FALSE)

(out <- as.tpsid(c(24900183,50150089,50079838)))
data.frame(out)
as.tpsid( data.frame(out) )

# Get all data back
get_tpsid_("Poa annua")
get_tpsid_("Poa annua", rows=2)
get_tpsid_("Poa annua", rows=1:2)
get_tpsid_(c("asdfadfasd","Pinus contorta"), rows=1:5)

# use curl options
invisible(get_tpsid("Quercus douglasii", messages = TRUE))
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\author{
Scott Chamberlain,
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pow_search.R
\name{pow_search}
\alias{pow_search}
\title{Search Kew's Plants of the World}
\usage{
pow_search(sci_com, limit = 100, cursor = "*", sort = NULL, q = NULL, ...)
}
\arguments{
\item{sci_com}{(character) query terms, scientific or common name}

\item{limit}{(integer) Number of records to return. default: 100}

\item{cursor}{(character) cursor string}

\item{sort}{(character) The field to sort by and sort order separted with
underscore, e.g., \code{sort="name_desc"}}

\item{q}{Deprecated, see \code{sci_com}}

\item{...}{Further args passed on to \link[crul:HttpClient]{crul::HttpClient}.}
}
\value{
a list with slots for metadata (\code{meta}) with list of response
attributes, and data (\code{data}) with a data.frame of results
}
\description{
Search Kew's Plants of the World
}
\examples{
\dontrun{
x <- pow_search(sci_com = "Quercus")
x$meta
x$meta$totalResults
x$meta$perPage
x$meta$totalPages
x$meta$page
x$meta$cursor
head(x$data)

# pagination
pow_search(sci_com = "sunflower", limit = 2)

# debug curl stuff
invisible(pow_search(sci_com = "Helianthus annuus", verbose = TRUE))

# sort
desc <- pow_search(sci_com = "Helianthus", sort = "name_desc")
desc$data$name
asc <- pow_search(sci_com = "Helianthus", sort = "name_asc")
asc$data$name
}
}
\references{
https://powo.science.kew.org/
}
\seealso{
Other pow: 
\code{\link{get_pow}()},
\code{\link{pow_lookup}()},
\code{\link{pow_synonyms}()}
}
\author{
Scott Chamberlain,
}
\concept{pow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-params.R
\name{taxize-params}
\alias{taxize-params}
\title{taxize parameters}
\description{
Information on standardized parameters across the package
}
\section{Standardized parameters}{

\itemize{
\item \code{sci}: scientific name
\item \code{com}: common name
\item \code{id}: name identifier
\item \code{sci_com}: scientific name or common name
\item \code{sci_id}: scientific name or name identifier
}

We were going to standardize parameter names for cases in which
a parameter accepts either of three options: scientific name,
common name, or name identifier. However, there was no clear
parameter name we could use for this case, so we've left
parameter names as they are for the two cases (\code{\link[=get_ids]{get_ids()}} and
\code{\link[=vascan_search]{vascan_search()}})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncbi_getbyname.R
\name{get_seqs}
\alias{get_seqs}
\title{Retrieve gene sequences from NCBI by accession number.}
\usage{
get_seqs(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_dist.R
\name{tp_namedistributions}
\alias{tp_namedistributions}
\title{Return all distribution records for for a taxon name with a given id.}
\usage{
tp_namedistributions(...)
}
\arguments{
\item{...}{ignored}
}
\description{
Function name changed to tp_dist.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_lsid.R
\name{itis_lsid}
\alias{itis_lsid}
\title{Get TSN from LSID}
\usage{
itis_lsid(lsid = NULL, what = "tsn", ...)
}
\arguments{
\item{lsid}{One or more lsid's}

\item{what}{What to retrieve. One of tsn, record, or fullrecord}

\item{...}{Further arguments passed on to \code{\link[ritis:lsid2tsn]{ritis::lsid2tsn()}},
\code{\link[ritis:record]{ritis::record()}}, or \code{\link[ritis:full_record]{ritis::full_record()}}}
}
\description{
Get TSN from LSID
}
\examples{
\dontrun{
# Get TSN
itis_lsid("urn:lsid:itis.gov:itis_tsn:180543")
itis_lsid(lsid=c("urn:lsid:itis.gov:itis_tsn:180543","urn:lsid:itis.gov:itis_tsn:28726"))

# Get partial record
itis_lsid("urn:lsid:itis.gov:itis_tsn:180543", "record")

# Get full record
itis_lsid("urn:lsid:itis.gov:itis_tsn:180543", "fullrecord")

# An invalid lsid (a tsn actually)
itis_lsid(202385)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progressor.R
\name{progressor}
\alias{progressor}
\title{methods for preparing/printing info for prompts for \verb{get_*} functions}
\description{
methods for preparing/printing info for prompts for \verb{get_*} functions

methods for preparing/printing info for prompts for \verb{get_*} functions
}
\examples{
\dontrun{
# nms <- c("Quercus", "Sasdsfasdf")
# x <- progressor$new(items = nms)
# x
# x$prog_start()

# x$completed(nms[1], "found")
# x$prog_found()

# x$completed(nms[2], "not found")
# x$prog_not_found()

# x$prog_summary()

# suppress cli::cat_line
# x <- progressor$new(items = nms, suppress = TRUE)
# x$prog_summary()
}
}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{total}}{(integer) x}

\item{\code{found}}{(integer) list of results when name found}

\item{\code{not_found}}{(integer) list of results when name not found}

\item{\code{done}}{(integer) x}

\item{\code{suppress}}{(integer) x}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{p}}{(integer) percent done}

\item{\code{d}}{(integer) number done}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-progressor-new}{\code{progressor$new()}}
\item \href{#method-progressor-completed}{\code{progressor$completed()}}
\item \href{#method-progressor-completed_found}{\code{progressor$completed_found()}}
\item \href{#method-progressor-completed_not_found}{\code{progressor$completed_not_found()}}
\item \href{#method-progressor-prog_start}{\code{progressor$prog_start()}}
\item \href{#method-progressor-prog}{\code{progressor$prog()}}
\item \href{#method-progressor-prog_found}{\code{progressor$prog_found()}}
\item \href{#method-progressor-prog_not_found}{\code{progressor$prog_not_found()}}
\item \href{#method-progressor-prog_summary}{\code{progressor$prog_summary()}}
\item \href{#method-progressor-clone}{\code{progressor$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-new"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{progressor} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$new(items, suppress = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{items}}{(character) xxx}

\item{\code{suppress}}{(logical) suppress messages. default: \code{FALSE}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{progressor} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-completed"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-completed}{}}}
\subsection{Method \code{completed()}}{
add results to found or not found
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$completed(name, att)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{(character) vector of names}

\item{\code{att}}{(character) one of "found" or "not found"}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
nothing returned; adds to \verb{$found} or \verb{$not_found}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-completed_found"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-completed_found}{}}}
\subsection{Method \code{completed_found()}}{
add to found results
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$completed_found(name)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{(character) vector of taxon names}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
nothing returned; adds to \verb{$found}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-completed_not_found"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-completed_not_found}{}}}
\subsection{Method \code{completed_not_found()}}{
add to not found results
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$completed_not_found(name)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{(character) vector of taxon names}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
nothing returned; adds to \verb{$not_found}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-prog_start"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-prog_start}{}}}
\subsection{Method \code{prog_start()}}{
print messages of total queries to do, and
percent completed
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$prog_start()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-prog"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-prog}{}}}
\subsection{Method \code{prog()}}{
prints message of found or not found using packages
cli and crayon
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$prog(att)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{att}}{(character) one of "found" or "not found"}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
messages
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-prog_found"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-prog_found}{}}}
\subsection{Method \code{prog_found()}}{
prints found message using packages cli and crayon
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$prog_found()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
messages
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-prog_not_found"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-prog_not_found}{}}}
\subsection{Method \code{prog_not_found()}}{
prints not found message using packages cli and crayon
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$prog_not_found()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
messages
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-prog_summary"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-prog_summary}{}}}
\subsection{Method \code{prog_summary()}}{
prints summary at end of result with total found and
not found
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$prog_summary()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
messages
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-progressor-clone"></a>}}
\if{latex}{\out{\hypertarget{method-progressor-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{progressor$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key_helpers.R
\name{use_eol-defunct}
\alias{use_eol-defunct}
\alias{use_eol}
\title{use eol}
\usage{
use_eol()
}
\description{
use eol
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_iucn.R
\name{get_iucn}
\alias{get_iucn}
\alias{as.iucn}
\alias{as.iucn.iucn}
\alias{as.iucn.character}
\alias{as.iucn.list}
\alias{as.iucn.numeric}
\alias{as.iucn.data.frame}
\alias{as.data.frame.iucn}
\title{Get a IUCN Redlist taxon}
\usage{
get_iucn(sci, messages = TRUE, key = NULL, x = NULL, ...)

as.iucn(x, check = TRUE, key = NULL)

\method{as.iucn}{iucn}(x, check = TRUE, key = NULL)

\method{as.iucn}{character}(x, check = TRUE, key = NULL)

\method{as.iucn}{list}(x, check = TRUE, key = NULL)

\method{as.iucn}{numeric}(x, check = TRUE, key = NULL)

\method{as.iucn}{data.frame}(x, check = TRUE, key = NULL)

\method{as.data.frame}{iucn}(x, ...)
}
\arguments{
\item{sci}{(character) A vector of scientific names. Or, a
\code{taxon_state} object (see \link{taxon-state})}

\item{messages}{logical; should progress be printed?}

\item{key}{(character) required. you IUCN Redlist API key. See
\link[rredlist:rredlist-package]{rredlist::rredlist-package} for help on authenticating with
IUCN Redlist}

\item{x}{For \code{get_iucn()}: Deprecated, see \code{sci}. For \code{as.iucn()}, various,
see examples}

\item{...}{Ignored}

\item{check}{(logical) Check if ID matches any existing on the DB, only
used in \code{\link[=as.iucn]{as.iucn()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.

Comes with the following attributes:
\itemize{
\item \emph{match} (character) - the reason for NA, either 'not found',
'found' or if \code{ask = FALSE} then 'NA due to ask=FALSE')
\item \emph{name} (character) - the taxonomic name, which is needed in
\code{\link[=synonyms]{synonyms()}} and \code{\link[=sci2comm]{sci2comm()}} methods since they
internally use \pkg{rredlist} functions which require the taxonomic name,
and not the taxonomic identifier
\item \emph{ri} (character) - The URI where more information can be
read on the taxon - includes the taxonomic identifier in the URL somewhere
}

\emph{multiple_matches} and \emph{pattern_match} do not apply here as in other \verb{get_*}
methods since there is no IUCN Redlist search, so you either get a match or
you do not get a match.
}
\description{
Get a IUCN Redlist taxon
}
\details{
There is no underscore method, because there's no real
search for IUCN, that is, where you search for a string, and get back
a bunch of results due to fuzzy matching. If that exists in the future
we'll add an underscore method here.

IUCN ids only work with \code{\link[=synonyms]{synonyms()}} and \code{\link[=sci2comm]{sci2comm()}}
methods.
}
\examples{
\dontrun{
get_iucn("Branta canadensis")
get_iucn("Branta bernicla")
get_iucn("Panthera uncia")

}
}
\seealso{
Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tnrs_sources.r
\name{tnrs_sources}
\alias{tnrs_sources}
\title{TNRS sources}
\usage{
tnrs_sources(source = NULL, ...)
}
\arguments{
\item{source}{The source to get information on, one of "iPlant_TNRS",
"NCBI", or "MSW3".}

\item{...}{Curl options to pass in \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
Sources for the TNRS API in a vector or list
}
\description{
Get sources for the Phylotastic Taxonomic Name
Resolution Service
}
\examples{
\dontrun{
# All
tnrs_sources()

# A specific source
tnrs_sources(source="NCBI")
tnrs_sources(source="MSW3")
tnrs_sources(source="iPlant_TNRS")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eol_dataobjects.R
\name{eol_dataobjects}
\alias{eol_dataobjects}
\title{Given the identifier for a data object, return all metadata about the object}
\usage{
eol_dataobjects(id, taxonomy = TRUE, language = NULL, ...)
}
\arguments{
\item{id}{(character) The EOL data object identifier}

\item{taxonomy}{(logical) Whether to return any taxonomy details from
different taxon hierarchy providers, in an array named \code{taxonconcepts}}

\item{language}{(character) provides the results in the specified language.
one of ms, de, en, es, fr, gl, it, nl, nb, oc, pt-BR, sv, tl, mk, sr, uk,
ar, zh-Hans, zh-Hant, ko}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
A list, optionally with a data.frame if \code{taxonomy=TRUE}
}
\description{
Given the identifier for a data object, return all metadata about the object
}
\details{
It's possible to return JSON or XML with the EOL API. However,
this function only returns JSON for now.
}
\examples{
\dontrun{
eol_dataobjects(id = 7561533)

# curl options
eol_dataobjects(id = 7561533, verbose = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gbif_parse.R
\name{gbif_parse}
\alias{gbif_parse}
\title{Parse taxon names using the GBIF name parser.}
\usage{
gbif_parse(scientificname, ...)
}
\arguments{
\item{scientificname}{(character) scientific names}

\item{...}{Further args passed on to \link[crul:verb-POST]{crul::verb-POST}}
}
\value{
A \code{data.frame} containing fields extracted from parsed
taxon names. Fields returned are the union of fields extracted from
all species names in \code{scientificname}.
}
\description{
Parse taxon names using the GBIF name parser.
}
\examples{
\dontrun{
gbif_parse(scientificname='x Agropogon littoralis')
gbif_parse(c('Arrhenatherum elatius var. elatius',
             'Secale cereale subsp. cereale', 'Secale cereale ssp. cereale',
             'Vanessa atalanta (Linnaeus, 1758)'))
}
}
\references{
https://www.gbif.org/tools/name-parser/about
}
\seealso{
\code{\link[=gni_parse]{gni_parse()}}, \code{\link[=gna_parse]{gna_parse()}}
}
\author{
John Baumgartner \email{johnbb@student.unimelb.edu.au}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plantminer.R
\name{plantminer}
\alias{plantminer}
\title{Search for taxonomy data from Plantminer.com}
\usage{
plantminer(plants, from = "tpl", messages = TRUE, ...)
}
\arguments{
\item{plants}{(character) Vector of plant species names. Required.}

\item{from}{(character) One of tpl (for theplantlist.com data), or
flora (for Brazilian Flora Checklist). Required. Default: \code{tpl}}

\item{messages}{(logical) informative messages or not. Default: \code{TRUE}}

\item{...}{curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
data.frame of results.
}
\description{
Search for taxonomy data from Plantminer.com
}
\note{
you used to need an API key for Plantminer; it's no longer needed
}
\examples{
\dontrun{
# A single taxon
plantminer("Ocotea pulchella")

# Many taxa
plants <- c("Myrcia lingua", "Myrcia bella", "Ocotea pulchella",
		"Miconia", "Coffea arabica var. amarella", "Bleh")
plantminer(plants)

# By deafult, tpl is used, for Theplantlist data,
# toggle the from parameter here
plantminer("Ocotea pulchella", from = "flora")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize_ldfast.R
\name{taxize_ldfast}
\alias{taxize_ldfast}
\title{Replacement function for ldply that should be faster in all cases.}
\usage{
taxize_ldfast(x, convertvec = FALSE)
}
\arguments{
\item{x}{A list.}

\item{convertvec}{Convert a vector to a data.frame before rbind is called.}
}
\description{
Replacement function for ldply that should be faster in all cases.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_terms.R
\name{itis_terms}
\alias{itis_terms}
\title{Get ITIS terms, i.e., tsn's, authors, common names, and scientific names.}
\usage{
itis_terms(query, what = "both", ...)
}
\arguments{
\item{query}{One or more common or scientific names, or partial names}

\item{what}{One of both (search common and scientific names), common
(search just common names), or scientific (search just scientific names)}

\item{...}{Further arguments passed on to \code{\link[ritis:terms]{ritis::terms()}}}
}
\description{
Get ITIS terms, i.e., tsn's, authors, common names, and scientific names.
}
\examples{
\dontrun{
# Get terms searching both common and scientific names
itis_terms(query='bear')

# Get terms searching just common names
itis_terms(query='tarweed', "common")

# Get terms searching just scientific names
itis_terms(query='Poa annua', "scientific")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_wiki.R
\name{get_wiki}
\alias{get_wiki}
\alias{as.wiki}
\alias{as.wiki.wiki}
\alias{as.wiki.character}
\alias{as.wiki.list}
\alias{as.wiki.numeric}
\alias{as.wiki.data.frame}
\alias{as.data.frame.wiki}
\alias{get_wiki_}
\title{Get the page name for a Wiki taxon}
\usage{
get_wiki(
  sci_com,
  wiki_site = "species",
  wiki = "en",
  ask = TRUE,
  messages = TRUE,
  limit = 100,
  rows = NA,
  x = NULL,
  ...
)

as.wiki(x, check = TRUE, wiki_site = "species", wiki = "en")

\method{as.wiki}{wiki}(x, check = TRUE, wiki_site = "species", wiki = "en")

\method{as.wiki}{character}(x, check = TRUE, wiki_site = "species", wiki = "en")

\method{as.wiki}{list}(x, check = TRUE, wiki_site = "species", wiki = "en")

\method{as.wiki}{numeric}(x, check = TRUE, wiki_site = "species", wiki = "en")

\method{as.wiki}{data.frame}(x, check = TRUE, wiki_site = "species", wiki = "en")

\method{as.data.frame}{wiki}(x, ...)

get_wiki_(
  x,
  messages = TRUE,
  wiki_site = "species",
  wiki = "en",
  limit = 100,
  rows = NA,
  ...
)
}
\arguments{
\item{sci_com}{(character) A vector of common or scientific names. Or, a
\code{taxon_state} object (see \link{taxon-state})}

\item{wiki_site}{(character) Wiki site. One of species (default), pedia,
commons}

\item{wiki}{(character) language. Default: en}

\item{ask}{logical; should get_wiki be run in interactive mode?
If \code{TRUE} and more than one wiki is found for the species, the user is
asked for input. If \code{FALSE} NA is returned for multiple matches.}

\item{messages}{logical; should progress be printed?}

\item{limit}{(integer) number of records to return}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this function still only gives back a wiki
class object with one to many identifiers. See \code{\link[=get_wiki_]{get_wiki_()}} to get back
all, or a subset, of the raw data that you are presented during the ask
process.}

\item{x}{For \code{get_wiki()}: deprecated, see \code{sci_com}. For \code{as.wiki}, various,
see examples}

\item{...}{Ignored}

\item{check}{logical; Check if ID matches any existing on the DB, only
used in \code{\link[=as.wiki]{as.wiki()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Get the page name for a Wiki taxon
}
\details{
For \code{wiki_site = "pedia" }, we use the english language site by
default. Set the \code{wiki} parameter for a different language site.
}
\examples{
\dontrun{
get_wiki(sci_com = "Quercus douglasii")
get_wiki(sci_com = "Quercu")
get_wiki(sci_com = "Quercu", "pedia")
get_wiki(sci_com = "Quercu", "commons")

# diff. wikis with wikipedia
get_wiki("Malus domestica", "pedia")
get_wiki("Malus domestica", "pedia", "fr")

# as coercion
as.wiki("Malus_domestica")
as.wiki("Malus_domestica", wiki_site = "commons")
as.wiki("Malus_domestica", wiki_site = "pedia")
as.wiki("Malus_domestica", wiki_site = "pedia", wiki = "fr")
as.wiki("Malus_domestica", wiki_site = "pedia", wiki = "da")
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wormsid}()}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class2tree.R
\name{class2tree}
\alias{class2tree}
\alias{plot.classtree}
\alias{print.classtree}
\title{Convert a list of classifications to a tree.}
\usage{
class2tree(input, varstep = TRUE, check = TRUE, remove_shared = FALSE, ...)

\method{plot}{classtree}(x, ...)

\method{print}{classtree}(x, ...)
}
\arguments{
\item{input}{List of classification data.frame's from the function
\code{\link[=classification]{classification()}}}

\item{varstep}{Vary step lengths between successive levels relative to
proportional loss of the number of distinct classes.}

\item{check}{If TRUE, remove all redundant levels which are different for
all rows or constant for all rows and regard each row as a different basal
taxon (species). If FALSE all levels are retained and basal taxa (species)
also must be coded as variables (columns). You will get a warning if
species are not coded, but you can ignore this if that was your intention.}

\item{remove_shared}{If \code{TRUE}, remove any taxa that are coarser ranks
present in other taxa, such as both a genus and a species in that genus in
the same tree.}

\item{...}{Further arguments passed on to hclust.}

\item{x}{Input object to print or plot - output from class2tree function.}
}
\value{
An object of class "classtree" with slots:
\itemize{
\item phylo - The resulting object, a phylo object
\item classification - The classification data.frame, with taxa as rows,
and different classification levels as columns
\item distmat - Distance matrix
\item names - The names of the tips of the phylogeny
}

Note that when you execute the resulting object, you only get the phylo
object. You can get to the other 3 slots by calling them directly, like
output$names, etc.
}
\description{
This function converts a list of hierarchies for individual species into
a single species by taxonomic level matrix, then calculates a distance
matrix based on taxonomy alone, and outputs either a phylo or dist object.
See details for more information.
}
\details{
See \code{\link[vegan:taxondive]{vegan::taxa2dist()}}. Thanks to Jari Oksanen for
making the taxa2dist function and pointing it out, and Clarke & Warwick
(1998, 2001), which taxa2dist was based on.
The taxonomy tree created is not only based on the clustering of the
taxonomy ranks (e.g. strain, species, genus, ...), but it also utilizes the
actual taxon clades (e.g. mammals, plants or reptiles, etc.). The process of
this function is as following: First, all possible taxonomy ranks and their
corresponding IDs for each given taxon will be collected from the input.
Then, the rank vectors of all taxa will be aligned, so that they together
will become a matrix where columns are ordered taxonomy ranks of all taxa and
rows are the rank vectors of those taxa. After that, the rank matrix will be
converted into taxonomy ID matrix, any missing rank will have a pseudo
ID from the previous rank. Finally, this taxonomy ID matrix will be used to
cluster taxa that have similar taxonomy hierarchy together.
}
\examples{
\dontrun{
spnames <- c('Quercus robur', 'Iris oratoria', 'Arachis paraguariensis',
 'Helianthus annuus','Madia elegans','Lupinus albicaulis',
 'Pinus lambertiana')
out <- classification(spnames, db='itis')
tr <- class2tree(out)
plot(tr)

spnames <- c('Klattia flava', 'Trollius sibiricus',
 'Arachis paraguariensis',
 'Tanacetum boreale', 'Gentiana yakushimensis','Sesamum schinzianum',
 'Pilea verrucosa','Tibouchina striphnocalyx','Lycium dasystemum',
 'Berkheya echinacea','Androcymbium villosum',
 'Helianthus annuus','Madia elegans','Lupinus albicaulis',
 'Pinus lambertiana')
out <- classification(spnames, db='ncbi')
tr <- class2tree(out)
plot(tr)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_accnames.R
\name{tp_accnames}
\alias{tp_accnames}
\title{Return all accepted names for a taxon name with a given id.}
\usage{
tp_accnames(id, key = NULL, ...)
}
\arguments{
\item{id}{the taxon identifier code}

\item{key}{Your Tropicos API key; See \link{taxize-authentication}
for help on authentication}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
List or dataframe.
}
\description{
Return all accepted names for a taxon name with a given id.
}
\examples{
\dontrun{
tp_accnames(id = 25503923)
tp_accnames(id = 25538750)

# No accepted names found
tp_accnames(id = 25509881)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-authentication.R
\name{taxize-authentication}
\alias{taxize-authentication}
\alias{authentication}
\title{taxize authentication}
\description{
Help on authentication
}
\section{What is an API?}{

An API is an Application Programming Interface. The term "API" can be used
for lots of scenarios, but in this case we're talking about web APIs,
or APIs (interfaces) to web resources. \pkg{taxize} interacts with
remote databases on the web via their APIs. You don't need to worry
about the details of how that all works; just know that some of them
require authentication and some do not.
}

\section{What are API keys?}{

For those APIs that require authentication, the way that's typically done
is through API keys: alphanumeric strings of variable lengths that are
supplied with a request to an API.

\pkg{taxize} won't get these keys for you; rather, you have to
go get a key for each service, but we do provide information on how
to get those keys. See \code{\link[=key_helpers]{key_helpers()}} for help on how to
obtain keys for this package.
}

\section{Using API keys}{

You can store API keys as R options in your \code{.Rprofile} file, or
as environment variables in either your \code{.Renviron} file or
\code{.bash_profile} file, o\code{.zshrc} file (if you use oh-my-zsh) or
similar. See \link{Startup} for help on R options and environment
variables.

Save your API keys with the following names:
\itemize{
\item Tropicos: R option or env var as 'TROPICOS_KEY'
\item IUCN: R option or env var as 'IUCN_REDLIST_KEY'
\item ENTREZ: R option or env var as 'ENTREZ_KEY'
}

If you save in .Renviron it looks like: \code{ENTREZ_KEY=somekey}

If you save in a .bash_profile, .zshrc, or similar file it looks like:
\verb{export ENTREZ_KEY=somekey}

If you save in a .Rprofile it looks like: \code{options(ENTREZ_KEY = "somekey")}

Remember to restart your R session (and to start a new shell window/tab
if you're using the shell) to take advantage of the new R options
or environment variables.

We strongly recommend using environment variables
(https://en.wikipedia.org/wiki/Environment_variable) over R options
because environment variables are widely used across programming
languages, operating systems, and computing environments; whereas
R options are specific to R.

Note that NCBI Entrez doesn't require that you use an API key,
but you do get a higher rate limit with a key (more requests per
time period), from 3 to 10 requests per second, so do get one.
}

\seealso{
\code{\link[=key_helpers]{key_helpers()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_kingdomnames.R
\name{itis_kingdomnames}
\alias{itis_kingdomnames}
\title{Get kingdom names}
\usage{
itis_kingdomnames(tsn = NULL, ...)
}
\arguments{
\item{tsn}{One or more TSN's (taxonomic serial number)}

\item{...}{Further arguments passed on to getkingdomnamefromtsn}
}
\description{
Get kingdom names
}
\examples{
\dontrun{
itis_kingdomnames(202385)
itis_kingdomnames(tsn=c(202385,183833,180543))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{rank_ref}
\alias{rank_ref}
\title{Lookup-table for IDs of taxonomic ranks}
\description{
data.frame of 46 rows, with 2 columns:
\itemize{
\item rankid - a numeric rank id, consecutive
\item ranks - a comma separated vector of names that are considered
equal to one another within the row
}
}
\details{
We use this data.frame to do data sorting/filtering based on the ordering
of ranks.

Please let us know if there is a rank that occurs from one of the data
sources \pkg{taxize} that we don't have in \code{rank_ref} dataset.

Let us know if you disagree with the ordering of ranks.

Note that \code{rankid} 280 are essentially "genetic variants"; placed just above
'unspecified' to denote they're not without rank, but they're not
really taxonomic ranks either. As far as I know there's no way
to delineate among these "genetic variant" types.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eubon_children.R
\name{eubon_children}
\alias{eubon_children}
\title{EUBON children}
\usage{
eubon_children(id, providers = NULL, timeout = 0, ...)
}
\arguments{
\item{id}{(character) identifier for the taxon. (LSID, DOI, URI, or any
other identifier used by the checklist provider)}

\item{providers}{(character) A list of provider id strings concatenated by
comma characters. The default : "pesi,bgbm-cdm-server[col] will
be used if this parameter is not set. A list of all available provider ids
can be obtained from the '/capabilities' service end point. Providers can be
nested, that is a parent provider can have sub providers. If the id of the
parent provider is supplied all subproviders will be queried. The query
can also be restricted to one or more subproviders by using the following
syntax: parent-id[sub-id-1,sub-id2,...]}

\item{timeout}{(numeric) The maximum of milliseconds to wait for responses
from any of the providers. If the timeout is exceeded the service will just
return the responses that have been received so far. The default timeout is
0 ms (wait for ever)}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
a data.frame or an empty list if no results found
}
\description{
EUBON children
}
\note{
There is no pagination in this method, so you may or may not be
getting all the results for a search. Sorry, out of our control
}
\examples{
\dontrun{
x <- eubon_children(id = "urn:lsid:marinespecies.org:taxname:126141",
  providers = 'worms')
head(x)
}
}
\references{
https://cybertaxonomy.eu/eu-bon/utis/1.3/doc.html
}
\seealso{
Other eubon-methods: 
\code{\link{eubon_capabilities}()},
\code{\link{eubon_hierarchy}()},
\code{\link{eubon_search}()}
}
\concept{eubon-methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comm2sci.R
\name{comm2sci}
\alias{comm2sci}
\alias{comm2sci.default}
\alias{comm2sci.tsn}
\alias{comm2sci.uid}
\title{Get scientific names from common names.}
\usage{
comm2sci(...)

\method{comm2sci}{default}(
  com,
  db = "ncbi",
  itisby = "search",
  simplify = TRUE,
  commnames = NULL,
  ...
)

\method{comm2sci}{tsn}(id, db = "ncbi", itisby = "search", simplify = TRUE, ...)

\method{comm2sci}{uid}(id, db = "ncbi", itisby = "search", simplify = TRUE, ...)
}
\arguments{
\item{...}{Further arguments passed on to internal methods.}

\item{com}{One or more common names or partial names.}

\item{db}{Data source, one of \emph{"ncbi"} (default), \emph{"itis"},
\emph{"tropicos"}, \emph{"eol"}, or \emph{"worms"}. If using ncbi, we
recommend getting an API key; see \link{taxize-authentication}}

\item{itisby}{Search for common names across entire names (search, default),
at beginning of names (begin), or at end of names (end).}

\item{simplify}{(logical) If \code{TRUE}, simplify output to a vector of names.
If \code{FALSE}, return variable formats from different sources, usually a
data.frame.}

\item{commnames}{Deprecated, see \code{com}}

\item{id}{taxon identifiers, as returned by \code{\link[=get_tsn]{get_tsn()}} or \code{\link[=get_uid]{get_uid()}}}
}
\value{
If \code{simplify=TRUE}, a list of scientific names, with list
labeled by your input names. If \code{simplify=FALSE}, a data.frame with
columns that vary by data source. \code{character(0)} on no match
}
\description{
Get scientific names from common names.
}
\details{
For data sources ITIS and NCBI you can pass in common names
directly, and use \code{\link[=get_uid]{get_uid()}} or \code{\link[=get_tsn]{get_tsn()}} to get ids first, then pass in
to this fxn.

For the other data sources, you can only pass in common names directly.
}
\section{Authentication}{

See \link{taxize-authentication} for help on authentication
}

\section{HTTP version for NCBI requests}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\section{Rate limits}{

In case you run into errors due to your rate limit being exceeded, see
\code{\link[=taxize_options]{taxize_options()}}, where you can set \code{ncbi_sleep}.
}

\examples{
\dontrun{
comm2sci(com='american black bear')
comm2sci(com='american black bear', simplify = FALSE)
comm2sci(com='black bear', db='itis')
comm2sci(com='american black bear', db='itis')
comm2sci(com='annual blue grass', db='tropicos')
comm2sci(com=c('annual blue grass','tree of heaven'), db='tropicos')
comm2sci('blue whale', db = "worms")
comm2sci(c('blue whale', 'dwarf surfclam'), db = "worms")

# ncbi: pass in uid's from get_uid() directly
x <- get_uid("western capercaillie", modifier = "Common Name")
comm2sci(x)
# itis: pass in tsn's from get_tsn() directly
x <- get_tsn(c("Louisiana black bear", "american crow"),
  searchtype = "common")
comm2sci(x)
}
}
\seealso{
\code{\link[=sci2comm]{sci2comm()}}
}
\author{
Scott Chamberlain
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sci2comm.R
\name{sci2comm}
\alias{sci2comm}
\alias{sci2comm.default}
\alias{sci2comm.uid}
\alias{sci2comm.tsn}
\alias{sci2comm.wormsid}
\alias{sci2comm.iucn}
\title{Get common names from scientific names.}
\usage{
sci2comm(...)

\method{sci2comm}{default}(sci, db = "ncbi", simplify = TRUE, scinames = NULL, ...)

\method{sci2comm}{uid}(id, ...)

\method{sci2comm}{tsn}(id, simplify = TRUE, ...)

\method{sci2comm}{wormsid}(id, simplify = TRUE, ...)

\method{sci2comm}{iucn}(id, simplify = TRUE, ...)
}
\arguments{
\item{...}{Further arguments passed on to functions \code{\link[=get_uid]{get_uid()}},
\code{\link[=get_tsn]{get_tsn()}}.}

\item{sci}{character; One or more scientific names or partial names.}

\item{db}{character; Data source, one of \code{"ncbi"} (default),
\code{"itis"} \code{"eol"}, \code{"worms"}, or \code{"iucn"}. Note that
each taxonomic data source has their own identifiers,  so that if you
provide the wrong \code{db} value for the identifier you could get a
result, but it will likely be wrong (not what you were expecting).
If using ncbi or iucn we recommend getting an API key;
see \link{taxize-authentication}}

\item{simplify}{(logical) If TRUE, simplify output to a vector of names.
If FALSE, return variable formats from different sources, usually a
data.frame. Only applies to eol and itis. Specify \code{FALSE} to obtain
the language of each vernacular in the output for eol and itis.}

\item{scinames}{Deprecated, see \code{sci}}

\item{id}{character; identifiers, as returned by \code{\link[=get_tsn]{get_tsn()}},
\code{\link[=get_uid]{get_uid()}}.}
}
\value{
List of character vectors, named by input taxon name,
or taxon ID. \code{character(0)} on no match
}
\description{
Get common names from scientific names.
}
\section{Authentication}{

See \link{taxize-authentication} for help on authentication
}

\section{HTTP version for NCBI requests}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\examples{
\dontrun{
sci2comm(sci='Helianthus annuus')
sci2comm(sci='Helianthus annuus', db='eol')
sci2comm(sci=c('Helianthus annuus', 'Poa annua'))
sci2comm(sci='Puma concolor', db='ncbi')
sci2comm('Gadus morhua', db='worms')
sci2comm('Pomatomus saltatrix', db='worms')
sci2comm('Loxodonta africana', db='iucn')

# Passing id in, works for sources: itis and ncbi, not eol
sci2comm(get_uid('Helianthus annuus'))
sci2comm(get_wormsid('Gadus morhua'))
sci2comm(get_iucn('Loxodonta africana'))

# Don't simplify returned
sci2comm(get_iucn('Loxodonta africana'), simplify=FALSE)

# Use curl options
sci2comm('Helianthus annuus', db="ncbi", verbose = TRUE)
}
}
\seealso{
\code{\link[=comm2sci]{comm2sci()}}
}
\author{
Scott Chamberlain
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col_children.R
\name{col_children}
\alias{col_children}
\title{Search Catalogue of Life for for direct children of a particular taxon.}
\usage{
col_children(name = NULL, id = NULL, format = NULL, start = NULL,
  checklist = NULL, ...)
}
\arguments{
\item{name}{The string to search for. Only exact matches found the name given
     will be returned, unless one or wildcards are included in the search
   string. An * (asterisk) character denotes a wildcard; a % (percentage)
character may also be used. The name must be at least 3 characters long,
not counting wildcard characters.}

\item{id}{The record ID of the specific record to return (only for scientific
names of species or infraspecific taxa)}

\item{format}{format of the results returned. Valid values are format=xml and
     format=php; if the format parameter is omitted, the results are returned
   in the default XML format. If format=php then results are returned as a
PHP array in serialized string format, which can be converted back to an
array in PHP using the unserialize command}

\item{start}{The first record to return. If omitted, the results are returned
from the first record (start=0). This is useful if the total number of
results is larger than the maximum number of results returned by a single
Web service query (currently the maximum number of results returned by a
single query is 500 for terse queries and 50 for full queries).}

\item{checklist}{The year of the checklist to query, if you want a specific
year's checklist instead of the lastest as default (numeric).}

\item{...}{Curl options passed on to \code{\link[httr]{GET}}}
}
\value{
A list of data.frame's.
}
\description{
Search Catalogue of Life for for direct children of a particular taxon.
}
\details{
You must provide one of name or id. The other parameters (format
		and start) are optional.
}
\examples{
\dontrun{
# A basic example
col_children(name="Apis")

# An example where there is no classification, results in data.frame with
# no rows
col_children(id='b2f88f382aa5568f93a97472c6be6516')

# Use a specific year's checklist
col_children(name="Apis", checklist=2012)
col_children(name="Apis", checklist=2009)

# Pass in many names or many id's
out <- col_children(name=c("Buteo","Apis","Accipiter","asdf"),
  checklist = "2012")
out$Apis # get just the output you want
library("plyr")
ldply(out) # or combine to one data.frame

# or pass many id's
ids <- c('abe977b1d27007a76dd12a5c93a637bf',
  'b2f88f382aa5568f93a97472c6be6516')
out <- col_children(id = ids, checklist=2012)
library("plyr")
ldply(out) # combine to one data.frame
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{worrms_ranks}
\alias{worrms_ranks}
\title{WORMS ranks}
\format{
A data frame with 216 rows and 2 variables:
\itemize{
\item \code{id}: rank id
\item \code{rank}: rank name
}
}
\description{
Created using \code{worrms::wm_ranks_id(-1)} on 2020-02-11.
}
\details{
Present in taxize in the case where WORMS does not
return rank names - with this dataset we can fill
in rank information as long as rank ids are returned
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_taxrank.R
\name{itis_taxrank}
\alias{itis_taxrank}
\title{Retrieve taxonomic rank name from given TSN.}
\usage{
itis_taxrank(query = NULL, ...)
}
\arguments{
\item{query}{TSN for a taxonomic group (numeric). If query is left as default
(NULL), you get all possible rank names, and their TSN's (using function
\code{\link[ritis:rank_names]{ritis::rank_names()}}. There is slightly different terminology for
Monera vs. Plantae vs. Fungi vs. Animalia vs. Chromista, so there are
separate terminologies for each group.}

\item{...}{Further arguments passed on to \code{\link[ritis:rank_name]{ritis::rank_name()}}}
}
\value{
Taxonomic rank names or data.frame of all ranks.
}
\description{
Retrieve taxonomic rank name from given TSN.
}
\details{
You can print messages by setting \code{verbose=FALSE}.
}
\examples{
\dontrun{
# All ranks
itis_taxrank()

# A single TSN
itis_taxrank(query=202385)

# Many TSN's
itis_taxrank(query=c(202385,183833,180543))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gna_search.R
\name{gni_seach}
\alias{gni_seach}
\title{Search for taxonomic names using the Global Names Index}
\usage{
gni_seach(names, ...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\name{taxize-defunct}
\alias{taxize-defunct}
\alias{defunct}
\title{Defunct functions in taxize}
\description{
The following functions are now defunct (no longer available):
\itemize{
\item All COL functions are defunct: \verb{as.colid, }col_children\verb{, }col_classification\verb{, }col_downstream\verb{, }col_search\verb{, }get_colid\verb{, }get_colid_\verb{, }as.data.frame.colid\verb{, }children.colid\verb{, }classification.colid\verb{, }downstream.colid\verb{, }id2name.colid\verb{, }lowest_common.colid\verb{, }synonyms.colid\verb{, }upstream.colid`
\item \code{col_classification()}: See\code{classification()}
\item \code{tp_classification()}: See\code{classification()}
\item \code{eol_hierarchy()}: See\code{classification()}
\item \code{\link[=eol_invasive]{eol_invasive()}}: See \code{eol} in the \pkg{originr} package.
\item \code{\link[=use_eol]{use_eol()}}: EOL no longer requires an API key
\item \code{\link[=tpl_search]{tpl_search()}}: Use the \pkg{Taxonstand} functions \code{TPL} or \code{TPLck} directly.
\item \code{\link[=get_seqs]{get_seqs()}}: This function changed name to\code{ncbi_getbyname()}()].
\item \code{\link[=get_genes]{get_genes()}}: This function changed name to\code{ncbi_getbyid()}()].
\item \code{\link[=get_genes_avail]{get_genes_avail()}}: This function changed name to\code{ncbi_search()}()].
\item \code{\link[=ncbi_getbyname]{ncbi_getbyname()}}: See \code{ncbi_byname} in the \pkg{traits} package.
\item \code{\link[=ncbi_getbyid]{ncbi_getbyid()}}: See \code{ncbi_byid} in the \pkg{traits} package.
\item \code{\link[=ncbi_search]{ncbi_search()}}: See \code{ncbi_searcher} in the \pkg{traits} package.
\item \code{\link[=gisd_isinvasive]{gisd_isinvasive()}}: See \code{gisd} in the \pkg{originr} package.
\item \code{\link[=ubio_classification]{ubio_classification()}}:  The uBio web services was down for quite a while, is
now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
\item \code{\link[=ubio_classification_search]{ubio_classification_search()}}: The uBio web services was down for quite
a while, is now (as of 2016-05-09) back up, but we don't trust that it will stay up
and available.
\item \code{\link[=ubio_id]{ubio_id()}}: The uBio web services was down for quite a while, is
now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
\item \code{\link[=ubio_ping]{ubio_ping()}}:  The uBio web services was down for quite a while, is
now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
\item \code{\link[=ubio_search]{ubio_search()}}:  The uBio web services was down for quite a while, is
now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
\item \code{\link[=ubio_synonyms]{ubio_synonyms()}}:  The uBio web services was down for quite a while, is
now (as of 2016-05-09) back up, but we don't trust that it will stay up and available.
\item \code{\link[=get_ubioid]{get_ubioid()}}: The uBio web services are apparently down indefinitely.
\item \code{\link[=phylomatic_tree]{phylomatic_tree()}}: This function is defunct. See
\code{phylomatic} in the package \pkg{brranching}
\item \code{\link[=phylomatic_format]{phylomatic_format()}}: This function is defunct. See
\code{phylomatic_names} in the package \pkg{brranching}
\item \code{\link[=eubon]{eubon()}}: This function is defunct. Use \code{\link[=eubon_search]{eubon_search()}}
\item \code{\link[=tnrs]{tnrs()}}: This function is defunct. Was too unreliable
\item \code{\link[=tnrs_sources]{tnrs_sources()}}: This function is defunct. Was too unreliable
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iucn_summary.R
\name{iucn_status}
\alias{iucn_status}
\title{Extractor functions for \code{iucn}-class.}
\usage{
iucn_status(x, ...)
}
\arguments{
\item{x}{an \code{iucn}-object as returned b\code{iucn_summary}ry}

\item{...}{Currently not used}
}
\value{
A character vector with the status.
}
\description{
Extractor functions for \code{iucn}-class.
}
\examples{
\dontrun{
ia <- iucn_summary(c("Panthera uncia", "Lynx lynx"))
iucn_status(ia)}
}
\seealso{
\code{\link[=iucn_summary]{iucn_summary()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phylomatic_tree.R
\name{phylomatic_tree}
\alias{phylomatic_tree}
\title{Query Phylomatic for a phylogenetic tree.}
\usage{
phylomatic_tree(...)
}
\arguments{
\item{...}{Parameters, ignored}
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_boldid.R
\name{get_boldid}
\alias{get_boldid}
\alias{as.boldid}
\alias{as.boldid.boldid}
\alias{as.boldid.character}
\alias{as.boldid.list}
\alias{as.boldid.numeric}
\alias{as.boldid.data.frame}
\alias{as.data.frame.boldid}
\alias{get_boldid_}
\title{Get the BOLD (Barcode of Life) code for a search term.}
\usage{
get_boldid(
  sci,
  fuzzy = FALSE,
  dataTypes = "basic",
  includeTree = FALSE,
  ask = TRUE,
  messages = TRUE,
  rows = NA,
  rank = NULL,
  division = NULL,
  parent = NULL,
  searchterm = NULL,
  ...
)

as.boldid(x, check = TRUE)

\method{as.boldid}{boldid}(x, check = TRUE)

\method{as.boldid}{character}(x, check = TRUE)

\method{as.boldid}{list}(x, check = TRUE)

\method{as.boldid}{numeric}(x, check = TRUE)

\method{as.boldid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{boldid}(x, ...)

get_boldid_(
  sci,
  messages = TRUE,
  fuzzy = FALSE,
  dataTypes = "basic",
  includeTree = FALSE,
  rows = NA,
  searchterm = NULL,
  ...
)
}
\arguments{
\item{sci}{character; A vector of scientific names. Or,
a \code{taxon_state} object (see \link{taxon-state})}

\item{fuzzy}{(logical) Whether to use fuzzy search or not (default: FALSE).}

\item{dataTypes}{(character) Specifies the datatypes that will be returned.
See \code{\link[=bold_search]{bold_search()}} for options.}

\item{includeTree}{(logical) If TRUE (default: FALSE), returns a list
containing information for parent taxa as well as the specified taxon.}

\item{ask}{logical; should get_tsn be run in interactive mode?
If TRUE and more than one TSN is found for teh species, the user is asked for
input. If FALSE NA is returned for multiple matches.}

\item{messages}{logical; should progress be printed?}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all rows are
considered. Note that this function still only gives back a boldid class object with one
to many identifiers. See \code{\link[=get_boldid_]{get_boldid_()}} to get back all, or a subset,
of the raw data that you are presented during the ask process.}

\item{rank}{(character) A taxonomic rank name. See \link{rank_ref} for possible
options. Though note that some data sources use atypical ranks, so inspect the
data itself for options. Optional. See \code{Filtering} below.}

\item{division}{(character) A division (aka phylum) name. Optional. See \code{Filtering}
below.}

\item{parent}{(character) A parent name (i.e., the parent of the target search
taxon). Optional. See \code{Filtering} below.}

\item{searchterm}{Deprecated, see \code{sci}}

\item{...}{Curl options passed on to \code{\link[crul:verb-GET]{crul::verb-GET}}}

\item{x}{Input to \code{\link[=as.boldid]{as.boldid()}}}

\item{check}{logical; Check if ID matches any existing on the DB, only used in
\code{\link[=as.boldid]{as.boldid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Get the BOLD (Barcode of Life) code for a search term.
}
\section{Filtering}{

The parameters \code{division}, \code{parent}, and \code{rank} are not used in the search to
the data provider, but are used in filtering the data down to a subset that
is closer to the target you want.  For all these parameters, you can use
regex strings since we use \code{\link[=grep]{grep()}} internally to match. Filtering narrows
down to the set that matches your query, and removes the rest.
}

\examples{
\dontrun{
get_boldid(sci = "Agapostemon")
get_boldid(sci = "Chironomus riparius")
get_boldid(c("Chironomus riparius","Quercus douglasii"))
splist <- names_list('species')
get_boldid(splist, messages=FALSE)

# Fuzzy searching
get_boldid(sci="Osmi", fuzzy=TRUE)

# Get back a subset
get_boldid(sci="Osmi", fuzzy=TRUE, rows = 1)
get_boldid(sci="Osmi", fuzzy=TRUE, rows = 1:10)
get_boldid(sci=c("Osmi","Aga"), fuzzy=TRUE, rows = 1)
get_boldid(sci=c("Osmi","Aga"), fuzzy=TRUE, rows = 1:3)

# found
get_boldid('Epicordulia princeps')
get_boldid('Arigomphus furcifer')

# When not found
get_boldid("howdy")
get_boldid(c("Chironomus riparius", "howdy"))
get_boldid("Cordulegaster erronea")
get_boldid("Nasiaeshna pentacantha")

# Narrow down results to a division or rank, or both
## Satyrium example
### Results w/o narrowing
get_boldid("Satyrium")
### w/ phylum
get_boldid("Satyrium", division = "Plantae")
get_boldid("Satyrium", division = "Animalia")

## Rank example
get_boldid("Osmia", fuzzy = TRUE)
get_boldid("Osmia", fuzzy = TRUE, rank = "genus")

# Fuzzy filter on any filtering fields
## uses grep on the inside
get_boldid("Satyrium", division = "anim")
get_boldid("Aga", fuzzy = TRUE, parent = "*idae")

# Convert a boldid without class information to a boldid class
as.boldid(get_boldid("Agapostemon")) # already a boldid, returns the same
as.boldid(get_boldid(c("Agapostemon","Quercus douglasii"))) # same
as.boldid(1973) # numeric
as.boldid(c(1973,101009,98597)) # numeric vector, length > 1
as.boldid("1973") # character
as.boldid(c("1973","101009","98597")) # character vector, length > 1
as.boldid(list("1973","101009","98597")) # list, either numeric or character
## dont check, much faster
as.boldid("1973", check=FALSE)
as.boldid(1973, check=FALSE)
as.boldid(c("1973","101009","98597"), check=FALSE)
as.boldid(list("1973","101009","98597"), check=FALSE)

(out <- as.boldid(c(1973,101009,98597)))
data.frame(out)
as.boldid( data.frame(out) )

# Get all data back
get_boldid_("Osmia", fuzzy=TRUE, rows=1:5)
get_boldid_("Osmia", fuzzy=TRUE, rows=1)
get_boldid_(c("Osmi","Aga"), fuzzy=TRUE, rows = 1:3)
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gbif_name_usage.R
\name{gbif_name_usage}
\alias{gbif_name_usage}
\title{Lookup details for specific names in all taxonomies in GBIF.}
\usage{
gbif_name_usage(
  key = NULL,
  name = NULL,
  data = "all",
  language = NULL,
  datasetKey = NULL,
  uuid = NULL,
  sourceId = NULL,
  rank = NULL,
  shortname = NULL,
  start = NULL,
  limit = 20,
  ...
)
}
\arguments{
\item{key}{(numeric) A GBIF key for a taxon}

\item{name}{(character) Filters by a case insensitive, canonical namestring,
e.g. 'Puma concolor'}

\item{data}{(character) Specify an option to select what data is returned. See Description
below.}

\item{language}{(character) Language, default is english}

\item{datasetKey}{(character) Filters by the dataset's key (a uuid)}

\item{uuid}{(character) A uuid for a dataset. Should give exact same results as datasetKey.}

\item{sourceId}{(numeric) Filters by the source identifier. Not used right now.}

\item{rank}{(character) Taxonomic rank. Filters by taxonomic rank as one of:
CLASS, CULTIVAR, CULTIVAR_GROUP, DOMAIN, FAMILY, FORM, GENUS, INFORMAL,
INFRAGENERIC_NAME, INFRAORDER, INFRASPECIFIC_NAME, INFRASUBSPECIFIC_NAME,
KINGDOM, ORDER, PHYLUM, SECTION, SERIES, SPECIES, STRAIN, SUBCLASS, SUBFAMILY,
SUBFORM, SUBGENUS, SUBKINGDOM, SUBORDER, SUBPHYLUM, SUBSECTION, SUBSERIES,
SUBSPECIES, SUBTRIBE, SUBVARIETY, SUPERCLASS, SUPERFAMILY, SUPERORDER,
SUPERPHYLUM, SUPRAGENERIC_NAME, TRIBE, UNRANKED, VARIETY}

\item{shortname}{(character) A short name..need more info on this?}

\item{start}{Record number to start at}

\item{limit}{Number of records to return}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
A list of length two. The first element is metadata. The second is
either a data.frame (verbose=FALSE, default) or a list (verbose=TRUE)
}
\description{
This is a taxize version of the same function in the \code{rgbif} package so as to not have to
import rgbif and thus require GDAL binary installation.
}
\references{
https://www.gbif.org/developer/summary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_gbifid.R
\name{get_gbifid}
\alias{get_gbifid}
\alias{as.gbifid}
\alias{as.gbifid.gbifid}
\alias{as.gbifid.character}
\alias{as.gbifid.list}
\alias{as.gbifid.numeric}
\alias{as.gbifid.data.frame}
\alias{as.data.frame.gbifid}
\alias{get_gbifid_}
\title{Get the GBIF backbone taxon ID from taxonomic names.}
\usage{
get_gbifid(
  sci,
  ask = TRUE,
  messages = TRUE,
  rows = NA,
  phylum = NULL,
  class = NULL,
  order = NULL,
  family = NULL,
  rank = NULL,
  method = "backbone",
  sciname = NULL,
  ...
)

as.gbifid(x, check = FALSE)

\method{as.gbifid}{gbifid}(x, check = FALSE)

\method{as.gbifid}{character}(x, check = TRUE)

\method{as.gbifid}{list}(x, check = TRUE)

\method{as.gbifid}{numeric}(x, check = TRUE)

\method{as.gbifid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{gbifid}(x, ...)

get_gbifid_(
  sci,
  messages = TRUE,
  rows = NA,
  method = "backbone",
  sciname = NULL
)
}
\arguments{
\item{sci}{(character) one or more scientific names. Or, a \code{taxon_state}
object (see \link{taxon-state})}

\item{ask}{logical; should get_gbifid be run in interactive mode?
If TRUE and more than one ID is found for the species, the user is asked for
input. If FALSE NA is returned for multiple matches.}

\item{messages}{logical; If TRUE the actual taxon queried is printed on the console.}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all rows are considered.
Note that this function still only gives back a gbifid class object with one to many identifiers.
See \code{\link[=get_gbifid_]{get_gbifid_()}} to get back all, or a subset, of the raw data that you are
presented during the ask process.}

\item{phylum}{(character) A phylum (aka division) name. Optional. See \code{Filtering}
below.}

\item{class}{(character) A class name. Optional. See \code{Filtering} below.}

\item{order}{(character) An order name. Optional. See \code{Filtering} below.}

\item{family}{(character) A family name. Optional. See \code{Filtering} below.}

\item{rank}{(character) A taxonomic rank name. See \link{rank_ref} for possible
options. Though note that some data sources use atypical ranks, so inspect the
data itself for options. Optional. See \code{Filtering} below.}

\item{method}{(character) one of "backbone" or "lookup". See Details.}

\item{sciname}{Deprecated, see \code{sci}}

\item{...}{Ignored}

\item{x}{Input to \code{\link[=as.gbifid]{as.gbifid()}}}

\item{check}{logical; Check if ID matches any existing on the DB, only used in
\code{\link[=as.gbifid]{as.gbifid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Get the GBIF backbone taxon ID from taxonomic names.
}
\details{
Internally in this function we use a function to search GBIF's taxonomy,
and if we find an exact match we return the ID for that match. If there isn't an
exact match we return the options to you to pick from.
}
\section{method parameter}{

"backbone" uses the \verb{/species/match} GBIF API route, matching against their
backbone taxonomy. We turn on fuzzy matching by default, as the search without
fuzzy against backbone is quite narrow. "lookup" uses the \verb{/species/search}
GBIF API route, doing a full text search of name usages covering scientific
and vernacular named, species descriptions, distributions and the entire
classification.
}

\section{Filtering}{

The parameters \code{phylum}, \code{class}, \code{order}, \code{family}, and \code{rank} are not used
in the search to the data provider, but are used in filtering the data down
to a subset that is closer to the target you want.  For all these parameters,
you can use regex strings since we use \code{\link[=grep]{grep()}} internally to match.
Filtering narrows down to the set that matches your query, and removes the rest.
}

\examples{
\dontrun{
get_gbifid(sci='Poa annua')
get_gbifid(sci='Pinus contorta')
get_gbifid(sci='Puma concolor')

#lots of queries
spp <- names_list("species", 10)
res <- get_gbifid(spp)
res
xx <- taxon_last()
xx

# multiple names
get_gbifid(c("Poa annua", "Pinus contorta"))

# specify rows to limit choices available
get_gbifid(sci='Pinus')
get_gbifid(sci='Pinus', rows=10)
get_gbifid(sci='Pinus', rows=1:3)

# When not found, NA given
get_gbifid(sci="uaudnadndj")
get_gbifid(c("Chironomus riparius", "uaudnadndj"))

# Narrow down results to a division or rank, or both
## Satyrium example
### Results w/o narrowing
get_gbifid("Satyrium")
### w/ phylum
get_gbifid("Satyrium", phylum = "Tracheophyta")
get_gbifid("Satyrium", phylum = "Arthropoda")
### w/ phylum & rank
get_gbifid("Satyrium", phylum = "Arthropoda", rank = "genus")

## Rank example
get_gbifid("Poa", method = "lookup")
get_gbifid("Poa", method = "lookup", rank = "genus")
get_gbifid("Poa", method = "lookup", family = "Thripidae")

# Fuzzy filter on any filtering fields
## uses grep on the inside
get_gbifid("Satyrium", phylum = "arthropoda")
get_gbifid("A*", method = "lookup", order = "*tera")
get_gbifid("A*", method = "lookup", order = "*ales")

# Convert a uid without class information to a uid class
as.gbifid(get_gbifid("Poa annua")) # already a uid, returns the same
as.gbifid(get_gbifid(c("Poa annua","Puma concolor"))) # same
as.gbifid(2704179) # numeric
as.gbifid(c(2704179,2435099,3171445)) # numeric vector, length > 1
as.gbifid("2704179") # character
as.gbifid(c("2704179","2435099","3171445")) # character vector, length > 1
as.gbifid(list("2704179","2435099","3171445")) # list, either numeric or character
## dont check, much faster
as.gbifid("2704179", check=FALSE)
as.gbifid(2704179, check=FALSE)
as.gbifid(2704179, check=FALSE)
as.gbifid(c("2704179","2435099","3171445"), check=FALSE)
as.gbifid(list("2704179","2435099","3171445"), check=FALSE)

(out <- as.gbifid(c(2704179,2435099,3171445)))
data.frame(out)
as.uid( data.frame(out) )

# Get all data back
get_gbifid_("Puma concolor")
get_gbifid_(c("Pinus", "uaudnadndj"))
get_gbifid_(c("Pinus", "Puma"), rows=5)
get_gbifid_(c("Pinus", "Puma"), rows=1:5)

# use curl options
invisible(get_gbifid("Quercus douglasii", verbose = TRUE))
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\author{
Scott Chamberlain,
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-id-details.R
\name{get_id_details}
\alias{get_id_details}
\title{Details on \verb{get_*()} functions}
\description{
Including outputs from \verb{get_*()} functions, as well as their
attributes, and all exception behaviors.
}
\details{
This document applies to the following functions:
\itemize{
\item \code{\link[=get_boldid]{get_boldid()}}
\item \code{\link[=get_eolid]{get_eolid()}}
\item \code{\link[=get_gbifid]{get_gbifid()}}
\item \code{\link[=get_ids]{get_ids()}}
\item \code{\link[=get_iucn]{get_iucn()}}
\item \code{\link[=get_natservid]{get_natservid()}}
\item \code{\link[=get_nbnid]{get_nbnid()}}
\item \code{\link[=get_tolid]{get_tolid()}}
\item \code{\link[=get_tpsid]{get_tpsid()}}
\item \code{\link[=get_tsn]{get_tsn()}}
\item \code{\link[=get_ubioid]{get_ubioid()}}
\item \code{\link[=get_uid]{get_uid()}}
\item \code{\link[=get_wiki]{get_wiki()}}
\item \code{\link[=get_wormsid]{get_wormsid()}}
}
}
\section{attributes}{

Each output from \verb{get_*()} functions have the following attributes:
\itemize{
\item \emph{match} (character) - the reason for NA, either 'not found', 'found' or
if \code{ask = FALSE} then 'NA due to ask=FALSE')
\item \emph{multiple_matches} (logical) - Whether multiple matches were returned by
the data source.
This can be \code{TRUE}, even if you get 1 name back because we try to pattern
match the name to see if there's any direct matches. So sometimes this
attribute is \code{TRUE}, as well as \code{pattern_match}, which then returns 1
resulting name without user prompt.
\item \emph{pattern_match} (logical) - Whether a pattern match was made.
If \code{TRUE} then\code{multiple_matches} must be \code{TRUE}, and we found a perfect match
to your name, ignoring case. If \code{FALSE}, there wasn't a direct match, and
likely you need to pick from many choices or further parameters can be used
to limit results
\item \emph{uri} (character) - The URI where more information can be read on the taxon
}
\itemize{
\item includes the taxonomic identifier in the URL somewhere. This may be missing
if the value returned is \code{NA}
}
}

\section{exceptions}{

The following are the various ways in which \verb{get_*()} functions behave:
\itemize{
\item success - the value returned is a character string or numeric
\item no matches found - you'll get an NA, refine your search or possible the
taxon searched for does not exist in the database you're using
\item more than on match and \code{ask = FALSE} - if there's more than one matching
result, and you have set \code{ask = FALSE}, then we can't determine the single
match to return, so we give back \code{NA}. However, in this case we do set the
\code{match} attribute to say \verb{NA due to ask=FALSE & > 1 result} so it's very
clear what happened - and you can even programatically check this as well
\item NA due to some other reason - some \verb{get_*()} functions have additional
parameters for filtering taxa. It's possible that even though there's
results (that is, \code{found} will say \code{TRUE}), you can get back an NA. This is
most likely if the parameter filters taxa after they are returned from the
data provider and the value passed to the parameter leads to no matches.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iucn_summary.R
\name{iucn_summary_id}
\alias{iucn_summary_id}
\title{DEFUNCT}
\usage{
iucn_summary_id(...)
}
\description{
DEFUNCT
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vascan_search.r
\name{vascan_search}
\alias{vascan_search}
\title{Search the CANADENSYS Vascan API.}
\usage{
vascan_search(q, format = "json", raw = FALSE, ...)
}
\arguments{
\item{q}{(character) Can be a scientific name, a vernacular name or a VASCAN
taxon identifier (e.g. 861)}

\item{format}{(character) One of json (default) or xml.}

\item{raw}{(logical) If TRUE, raw json or xml returned, if FALSE, parsed
data returned.}

\item{...}{(list) Further args passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
json, xml or a list.
}
\description{
Search the CANADENSYS Vascan API.
}
\details{
Note that we lowercase all outputs in data.frame's, but when a
list is given back, we don't touch the list names.
}
\examples{
\dontrun{
vascan_search(q = "Helianthus annuus")
vascan_search(q = "Helianthus annuus", raw=TRUE)
vascan_search(q = c("Helianthus annuus", "Crataegus dodgei"), raw=TRUE)

# format type
## json
c <- vascan_search(q = "Helianthus annuus", format="json", raw=TRUE)
library("jsonlite")
fromJSON(c, FALSE)

## xml
d <- vascan_search(q = "Helianthus annuus", format="xml", raw=TRUE)
library("xml2")
xml2::read_xml(d)

# lots of names, in this case 50
splist <- names_list(rank='species', size=50)
vascan_search(q = splist)

# Curl options
invisible(vascan_search(q = "Helianthus annuus", verbose = TRUE))
}
}
\references{
API docs https://data.canadensys.net/vascan/api
}
\author{
Scott Chamberlain
}
\keyword{names}
\keyword{taxonomy}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_pow.R
\name{get_pow}
\alias{get_pow}
\alias{as.pow}
\alias{as.pow.pow}
\alias{as.pow.character}
\alias{as.pow.list}
\alias{as.pow.data.frame}
\alias{as.data.frame.pow}
\alias{get_pow_}
\title{Get Kew's Plants of the World code for a taxon}
\usage{
get_pow(
  sci_com,
  accepted = FALSE,
  ask = TRUE,
  messages = TRUE,
  rows = NA,
  family_filter = NULL,
  rank_filter = NULL,
  x = NULL,
  ...
)

as.pow(x, check = TRUE)

\method{as.pow}{pow}(x, check = TRUE)

\method{as.pow}{character}(x, check = TRUE)

\method{as.pow}{list}(x, check = TRUE)

\method{as.pow}{data.frame}(x, check = TRUE)

\method{as.data.frame}{pow}(x, ...)

get_pow_(sci_com, messages = TRUE, rows = NA, x = NULL, ...)
}
\arguments{
\item{sci_com}{character; A vector of common or scientific names. Or, a
\code{taxon_state} object (see \link{taxon-state})}

\item{accepted}{logical; If TRUE, removes names that are not accepted
valid names by ITIS. Set to \code{FALSE} (default) to give back both
accepted and unaccepted names.}

\item{ask}{logical; should \code{get_pow} be run in interactive mode?
If TRUE and more than one pow is found for teh species, the user is
asked for input. If FALSE NA is returned for multiple matches.}

\item{messages}{logical; should progress be printed?}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA,
all rows are considered. Note that this function still only gives back
a pow class object with one to many identifiers. See
\code{\link[=get_pow_]{get_pow_()}} to get back all, or a subset,
of the raw data that you are presented during the ask process.}

\item{family_filter}{(character) A division (aka phylum) name to filter
data after retrieved from NCBI. Optional. See \code{Filtering} below.}

\item{rank_filter}{(character) A taxonomic rank name to filter data after
retrieved from NCBI. See \link{rank_ref} for possible options.
Though note that some data sources use atypical ranks, so inspect the data
itself for options. Optional. See \code{Filtering} below.}

\item{x}{For \code{get_pow()}: deprecated, see \code{sci_com}. For \code{as.pow}, various,
see examples}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}

\item{check}{logical; Check if ID matches any existing on the DB, only
used in \code{\link[=as.pow]{as.pow()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Get Kew's Plants of the World code for a taxon
}
\section{Filtering}{

The parameters \code{family_filter} an\code{rank_filter}er are not
used in the search to the data provider, but are used in filtering the data down to a
subset that is closer to the target you want.  For these two parameters,
you can use regex strings since we use \code{\link[=grep]{grep()}} internally to match.
Filtering narrows down to the set that matches your query, and removes the rest.
}

\section{Rate-limits}{

As of February 2019, KEW was limiting to 5 requests per second. Note that
they may change that number in the future.

If you get errors that contain \code{429} you are hitting the rate limit, and you
can get around it by doing requests with \code{Sys.sleep} in between requests.
}

\examples{
\dontrun{
get_pow(sci_com="Helianthus")
get_pow(c("Helianthus","Quercus douglasii"))

# Get back a subset
get_pow(sci_com="Helianthus", rows = 1)
get_pow(sci_com="Helianthus", rows = 1:10)

# When not found
get_pow("howdy")
get_pow(c("Helianthus annuus", "howdy"))

# Narrow down results 
# to accepted names
get_pow("Helianthus", accepted = TRUE)
# to a kingom 
get_pow("Helianthus", rank_filter = "genus")
# to accepted names and rank
get_pow("Helianthus annuus", accepted = TRUE, rank_filter = "species")
# to a family
get_pow("flower", family_filter = "Acanthaceae")

# Convert a pow without class information to a pow class
z <- get_pow("Helianthus annuus", accepted = TRUE, rank_filter = "species")
# already a pow, returns the same
as.pow(z)
as.pow("urn:lsid:ipni.org:names:119003-2")
# character vector, length > 1
ids <- c("urn:lsid:ipni.org:names:119003-2","urn:lsid:ipni.org:names:328247-2")
as.pow(ids)
# list, with character strings
as.pow(as.list(ids)) 
## dont check, much faster
as.pow("urn:lsid:ipni.org:names:119003-2", check=FALSE)
as.pow(ids, check=FALSE)
as.pow(as.list(ids), check=FALSE)

(out <- as.pow(ids))
data.frame(out)
as.pow( data.frame(out) )

# Get all data back
get_pow_("Quercus", rows=1:5)
get_pow_("Quercus", rows=1)
get_pow_(c("Pinus", "Abies"), rows = 1:3)
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other pow: 
\code{\link{pow_lookup}()},
\code{\link{pow_search}()},
\code{\link{pow_synonyms}()}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\concept{pow}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bold_downstream.R
\name{bold_downstream}
\alias{bold_downstream}
\title{Retrieve all taxa names downstream in hierarchy for BOLD}
\usage{
bold_downstream(id, downto, intermediate = FALSE, ...)
}
\arguments{
\item{id}{(integer) One or more BOLD taxonomic identifiers}

\item{downto}{(character) The taxonomic level you want to go down to.
See examples below. The taxonomic level IS case sensitive, and you do have
to spell it correctly. See \code{data(rank_ref)} for spelling.}

\item{intermediate}{(logical) If \code{TRUE}, return a list of length two
with target taxon rank names, with additional list of data.frame's of
intermediate taxonomic groups. Default: \code{FALSE}}

\item{...}{crul options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
data.frame of taxonomic information downstream to family from e.g.,
Order, Class, etc., or if \code{intermediated=TRUE}, list of length two,
with target taxon rank names, and intermediate names.
}
\description{
Retrieve all taxa names downstream in hierarchy for BOLD
}
\details{
BEWARE: This function scrapes the BOLD website, so may be unstable. That is,
one day it may work, and the next it may fail. Open an issue if you
encounter an error: https://github.com/ropensci/taxize/issues
}
\examples{
\dontrun{
## the genus Gadus
bold_downstream(id = 3451, downto="species")

bold_downstream(id = 443, downto="genus")
bold_downstream(id = 443, downto="genus", intermediate=TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/names_list.r
\name{names_list}
\alias{names_list}
\title{Get a random vector of species names.}
\usage{
names_list(rank = "genus", size = 10)
}
\arguments{
\item{rank}{(character) Taxonomic rank, one of species, genus (default),
family, order}

\item{size}{(integer/numeric) Number of names to get. Maximum depends on
the rank}
}
\value{
character vector of taxonomic names
}
\description{
Family and order names come from the APG plant names list. Genus and
species names come from Theplantlist.org.
}
\examples{
names_list()
names_list('species')
names_list('genus')
names_list('family')
names_list('order')
names_list('order', 2)
names_list('order', 15)

# You can get a lot of genus or species names if you want
nrow(theplantlist)
names_list('genus', 500)
}
\author{
Scott Chamberlain
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_hierarchy.R
\name{itis_hierarchy}
\alias{itis_hierarchy}
\title{ITIS hierarchy}
\usage{
itis_hierarchy(tsn, what = "full", ...)
}
\arguments{
\item{tsn}{One or more TSN's (taxonomic serial number). Required.}

\item{what}{One of full (full hierarchy), up (immediate upstream), or down
(immediate downstream)}

\item{...}{Further arguments passed on to \code{\link[ritis:hierarchy]{ritis::hierarchy_full()}}
\code{\link[ritis:hierarchy]{ritis::hierarchy_up()}} or \code{\link[ritis:hierarchy]{ritis::hierarchy_down()}}}
}
\description{
Get hierarchies from TSN values, full, upstream only, or
immediate downstream only
}
\details{
Note that \code{\link[=itis_downstream]{itis_downstream()}} gets taxa downstream to a
particular rank, while this function only gets immediate names downstream.
}
\examples{
\dontrun{
# Get full hierarchy
itis_hierarchy(tsn=180543)

# Get hierarchy upstream
itis_hierarchy(tsn=180543, "up")

# Get hierarchy downstream
itis_hierarchy(tsn=180543, "down")

# Many tsn's
itis_hierarchy(tsn=c(180543,41074,36616))
}
}
\seealso{
\code{\link[=itis_downstream]{itis_downstream()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ubio_synonyms.R
\name{ubio_synonyms}
\alias{ubio_synonyms}
\title{Search uBio for taxonomic synonyms by hierarchiesID.}
\usage{
ubio_synonyms(...)
}
\arguments{
\item{...}{Parameters, ignored}
}
\description{
THIS FUNCTION IS DEFUNCT.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ubio_classification_search.R
\name{ubio_classification_search}
\alias{ubio_classification_search}
\title{This function will return ClassificationBankIDs (hierarchiesIDs) that refer to the
given NamebankID}
\usage{
ubio_classification_search(...)
}
\arguments{
\item{...}{Parameters, ignored}
}
\description{
THIS FUNCTION IS DEFUNCT.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_refs.R
\name{itis_refs}
\alias{itis_refs}
\title{Get references related to a ITIS TSN.}
\usage{
itis_refs(tsn, ...)
}
\arguments{
\item{tsn}{One or more TSN's (taxonomic serial number) for a
taxonomic group (numeric)}

\item{...}{Further arguments passed on to getpublicationsfromtsn}
}
\description{
Get references related to a ITIS TSN.
}
\examples{
\dontrun{
itis_refs(202385)
itis_refs(c(202385, 70340))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxon_state.R
\name{taxon-state}
\alias{taxon-state}
\alias{taxon_last}
\alias{taxon_clear}
\title{Last taxon state object from a \verb{get_*} function call}
\usage{
taxon_last()

taxon_clear()
}
\value{
\code{taxon_last()} returns an object of class \code{taxon_state}, the last
one used, else \code{NULL} if none found. \code{taxon_clear()} clears the saved state
}
\description{
Last taxon state object from a \verb{get_*} function call
}
\details{
\itemize{
\item \code{taxon_last()}: get the last \code{taxon_state} object in use
\item \code{taxon_clear()}: clear any data from last \code{taxon_state} object
}

The \code{taxon_state} object is an R6 object that holds data and methods
used for keeping track of results gathered within a \verb{get_*} function.
You shouldn't create \code{taxon_state} R6 objects yourself.

Behaviors to be aware of:
\itemize{
\item If a \code{taxon_state} object is not passed you don't need to worry about
a previously run \verb{get_*} function interfering with another \verb{get_*}
function call - you have to explicitly pass a \code{taxon_state} object
to use \code{taxon_state}
\item The passed in \code{taxon_state} object must have a \verb{$class} matching that of
the \verb{get_*} function being called. For example, you can only pass a
\code{taxon_state} with \verb{$class} of \code{gbifid} to \code{get_gbifid()}, and so on.
\item If you run \code{taxon_clear()} while a \verb{get*} function is running, you may
lose track of any state known to this package before it was cleared
}

See the internal method \link{progressor} for information on how we control messages
in \verb{get*} functions
}
\examples{
\dontrun{
spp <- names_list("species", 3)
res <- get_gbifid(spp)
z <- taxon_last()
z
z$taxa_remaining()
z$taxa_completed()
z$count # active binding; no parens needed

# cleanup
taxon_clear()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{theplantlist}
\alias{theplantlist}
\title{Lookup-table for family, genus, and species names for ThePlantList}
\format{
A data frame with 10,000 rows and 3 variables:
\itemize{
\item \code{family} family name
\item \code{genus} genus name
\item \code{species} specific epithet name
}
}
\source{
http://www.theplantlist.org
}
\description{
These names are from http://www.theplantlist.org, and are from
version 1.1 of their data. This data is used in the function
\code{\link[=names_list]{names_list()}}. This is a randomly selected subset of the ~350K
accepted species names in Theplantlist.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pow_search.R
\name{pow_lookup}
\alias{pow_lookup}
\title{Lookup taxa in Kew's Plants of the World}
\usage{
pow_lookup(id, include = NULL, ...)
}
\arguments{
\item{id}{(character) taxon id. required}

\item{include}{(character) vector of additional fields to include in
results. options include 'distribution' and 'descriptions'. optional}

\item{...}{Further args passed on to \link[crul:HttpClient]{crul::HttpClient}.}
}
\description{
Lookup taxa in Kew's Plants of the World
}
\examples{
\dontrun{
pow_lookup(id = 'urn:lsid:ipni.org:names:320035-2')
pow_lookup(id = 'urn:lsid:ipni.org:names:320035-2',
  include = "distribution")
pow_lookup(id = 'urn:lsid:ipni.org:names:320035-2',
  include = c("distribution", "descriptions"))
}
}
\seealso{
Other pow: 
\code{\link{get_pow}()},
\code{\link{pow_search}()},
\code{\link{pow_synonyms}()}
}
\concept{pow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tnrs.R
\name{tnrs}
\alias{tnrs}
\title{Phylotastic Taxonomic Name Resolution Service.}
\usage{
tnrs(
  query = NA,
  source = NULL,
  code = NULL,
  getpost = "POST",
  sleep = 0,
  splitby = 30,
  messages = TRUE,
  ...
)
}
\arguments{
\item{query}{Vector of quoted taxonomic names to search (character).}

\item{source}{Specify the source you want to match names against. Defaults
to just retrieve data from all sources. Options: NCBI, iPlant_TNRS,
or MSW3. Only available when using getpost="POST".}

\item{code}{Nomenclatural code. One of: ICZN (zoological), ICN (algae, fungi,
and plants), ICNB (bacteria), ICBN (botanical), ICNCP (cultivated plants),
ICTV (viruses). Only available when using getpost="POST".}

\item{getpost}{Use GET or POST method to send the query. If you have more
than say 50 species or so in your query, you should probably use POST.
IMPORTANT!!!!! ->
POST is the only option for this parameter if you want to
use source or code parameters.}

\item{sleep}{Number of seconds by which to pause between calls. Defaults to 0
seconds. Use when doing many calls in a for loop ar lapply type call.}

\item{splitby}{Number by which to split species list for querying the TNRS.}

\item{messages}{Verbosity or not (default \code{TRUE})}

\item{...}{Curl options to pass in \link[crul:verb-GET]{crul::verb-GET} or \link[crul:verb-POST]{crul::verb-POST}}
}
\value{
data.frame of results from TNRS plus the name submitted, with
rows in order of user supplied names, though those with no matches are
dropped
}
\description{
Match taxonomic names using the Taxonomic Name Resolution
Service (TNRS). Returns score of the matched name, and whether it was
accepted or not.
}
\details{
If there is no match in the Taxosaurus database, nothing is
returned, so you will not get anything back for non-matches.

TNRS doesn't provide any advice about the occurrence of homonyms when
queries have no indication of a taxonomic name's authority. So if there
is any chance of a homonym, you probably want to send the authority as
well, or use \code{\link[=gnr_resolve]{gnr_resolve()}}. For example,
\code{tnrs(query="Jussiaea linearis", source="iPlant_TNRS")} gives result of
\emph{Jussiaea linearis (Willd.) Oliv. ex Kuntze}, but there is a
homonym. If you do
\code{tnrs(query="Jussiaea linearis Hochst.", source="iPlant_TNRS")} you
get a direct match for that name. So, beware that there's no indication
of homonyms.
}
\examples{
\dontrun{
mynames <- c("Helianthus annuus", "Poa annua", "Mimulus bicolor")
tnrs(query = mynames, source = "iPlant_TNRS")

# Specifying the nomenclatural code to match against
mynames <- c("Helianthus annuus", "Poa annua")
tnrs(query = mynames, code = "ICBN")

# You can specify multiple sources, by comma-separating them
mynames <- c("Panthera tigris", "Eutamias minimus", "Magnifera indica",
"Humbert humbert")
tnrs(query = mynames, source = "NCBI,MSW3")

mynames <- c("Panthera tigris", "Eutamias minimus", "Magnifera indica",
   "Humbert humbert", "Helianthus annuus", "Pinus contorta", "Poa annua",
   "Abies magnifica", "Rosa california", "Festuca arundinace",
   "Mimulus bicolor", "Sorbus occidentalis","Madia sativa", "Thymopsis thymodes",
   "Bartlettia scaposa")
tnrs(mynames, source = "NCBI")

# Pass on curl options
mynames <- c("Helianthus annuus", "Poa annua", "Mimulus bicolor")
tnrs(query = mynames, source = "iPlant_TNRS", verbose = TRUE)
}
}
\references{
\url{http://taxosaurus.org/}
}
\seealso{
\code{\link[=gnr_resolve]{gnr_resolve()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/worms_downstream.R
\name{worms_downstream}
\alias{worms_downstream}
\title{Retrieve all taxa names downstream in hierarchy for WORMS}
\usage{
worms_downstream(id, downto, intermediate = FALSE, start = 1, ...)
}
\arguments{
\item{id}{(integer) One or more AphiaID's}

\item{downto}{(character) The taxonomic level you want to go down to.
See examples below. The taxonomic level IS case sensitive, and you do have
to spell it correctly. See \link{rank_ref_zoo} for spelling.}

\item{intermediate}{(logical) If \code{TRUE}, return a list of length two
with target taxon rank names, with additional list of data.frame's of
intermediate taxonomic groups. Default: \code{FALSE}}

\item{start}{(integer) Record number to start at}

\item{...}{crul options passed on to \code{\link[worrms:wm_children]{worrms::wm_children()}}, including the
parameters \code{marine_only} and \code{offset}, see \code{?worrms::wm_children} for
details}
}
\value{
data.frame of taxonomic information downstream to family from e.g.,
Order, Class, etc., or if \code{intermediated=TRUE}, list of length two,
with target taxon rank names, and intermediate names.
}
\description{
Retrieve all taxa names downstream in hierarchy for WORMS
}
\examples{
\dontrun{
## the genus Gadus
worms_downstream(id = 125732, downto="species")
worms_downstream(id = 125732, downto="species", intermediate=TRUE)

worms_downstream(id = 51, downto="class")
worms_downstream(id = 51, downto="subclass", intermediate=TRUE)

worms_downstream(id = 105, downto="subclass")

# marine_only parameter
worms_downstream(545470, downto = "species")
worms_downstream(545470, downto = "species", marine_only = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncbi_getbyname.R
\name{ncbi_getbyname}
\alias{ncbi_getbyname}
\title{Retrieve gene sequences from NCBI by taxon name and gene names.}
\usage{
ncbi_getbyname(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\author{
Scott Chamberlain
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tnrs_sources.r
\name{tnrs_sources}
\alias{tnrs_sources}
\title{TNRS sources}
\usage{
tnrs_sources(...)
}
\arguments{
\item{...}{ignored}
}
\description{
Defunct - service was down far too much to be reliable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/synonyms.R
\name{synonyms}
\alias{synonyms}
\alias{synonyms.default}
\alias{synonyms.tsn}
\alias{synonyms.tpsid}
\alias{synonyms.nbnid}
\alias{synonyms.wormsid}
\alias{synonyms.iucn}
\alias{synonyms.pow}
\alias{synonyms.ids}
\alias{synonyms_df}
\title{Retrieve synonyms from various sources given input taxonomic
names or identifiers}
\usage{
synonyms(...)

\method{synonyms}{default}(sci_id, db = NULL, rows = NA, x = NULL, ...)

\method{synonyms}{tsn}(id, ...)

\method{synonyms}{tpsid}(id, ...)

\method{synonyms}{nbnid}(id, ...)

\method{synonyms}{wormsid}(id, ...)

\method{synonyms}{iucn}(id, ...)

\method{synonyms}{pow}(id, ...)

\method{synonyms}{ids}(id, ...)

synonyms_df(x)
}
\arguments{
\item{...}{Other passed arguments to internal functions \verb{get_*()} and
functions to gather synonyms.}

\item{sci_id}{Vector of taxa names (character) or IDs (character or numeric)}

\item{db}{character; database to query. either \code{itis}, \code{tropicos},
\code{nbn}, \code{worms}, or \code{pow}. Note that each taxonomic data source has their own
identifiers, so that if you provide the wrong \code{db} value for the identifier
you could get a result, but it will likely be wrong (not what you were
expecting). If using tropicos, we  recommend getting an API key;
see \link{taxize-authentication}}

\item{rows}{(numeric) Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this parameter is ignored if you pass in a
taxonomic id of any of the acceptable classes: tsn, tpsid, nbnid, ids.}

\item{x}{For \code{synonyms()}: deprecated, see \code{sci_id}. For \code{synonyms_df()},
the output of \code{synonyms()}}

\item{id}{character; identifiers, returned by \code{\link[=get_tsn]{get_tsn()}}, \code{\link[=get_tpsid]{get_tpsid()}},
\code{\link[=get_nbnid]{get_nbnid()}}, \code{\link[=get_wormsid]{get_wormsid()}}, \code{\link[=get_pow]{get_pow()}}}
}
\value{
A named list of results with three types of output in each slot:
\itemize{
\item if the name was not found: \code{NA_character_}
\item if the name was found but no synonyms found, an empty data.frame (0 rows)
\item if the name was found, and synonyms found, a data.frames with the
synonyms - the column names vary by data source
}
}
\description{
Retrieve synonyms from various sources given input taxonomic
names or identifiers
}
\details{
If IDs are supplied directly (not from the \verb{get_*()} functions)
you must specify the type of ID.

For \code{db = "itis"} you can pass in a parameter \code{accepted} to
toggle whether only accepted names are used \code{accepted = TRUE}, or if
all are used \code{accepted = FALSE}. The default is \code{accepted = FALSE}

Note that IUCN requires an API key. See \link[rredlist:rredlist-package]{rredlist::rredlist-package}
for help on authentiating with IUCN Redlist
}
\examples{
\dontrun{
# Plug in taxon IDs
synonyms(183327, db="itis")
synonyms("25509881", db="tropicos")
synonyms("NBNSYS0000004629", db='nbn')
synonyms(105706, db='worms')
synonyms(12392, db='iucn')
synonyms('urn:lsid:ipni.org:names:358881-1', db='pow')

# Plug in taxon names directly
synonyms("Pinus contorta", db="itis")
synonyms("Puma concolor", db="itis")
synonyms(c("Poa annua",'Pinus contorta','Puma concolor'), db="itis")
synonyms("Poa annua", db="tropicos")
synonyms("Pinus contorta", db="tropicos")
synonyms(c("Poa annua",'Pinus contorta'), db="tropicos")
synonyms("Pinus sylvestris", db='nbn')
synonyms('Pomatomus', db='worms')
synonyms('Pomatomus saltatrix', db='worms')
synonyms('Lithocarpus mindanaensis', db='pow')
synonyms('Poa annua', db='pow')
synonyms(c('Poa annua', 'Pinus contorta', 'foo bar'), db='pow')

# not accepted names, with ITIS
## looks for whether the name given is an accepted name,
## and if not, uses the accepted name to look for synonyms
synonyms("Acer drummondii", db="itis")
synonyms("Spinus pinus", db="itis")

# Use get_* methods
synonyms(get_tsn("Poa annua"))
synonyms(get_tpsid("Poa annua"))
synonyms(get_nbnid("Carcharodon carcharias"))
synonyms(get_iucn('Loxodonta africana'))
synonyms(get_pow('Lithocarpus mindanaensis'))

# Pass many ids from class "ids"
out <- get_ids(names="Poa annua", db = c('itis','tropicos'))
synonyms(out)

# Use the rows parameter to select certain rows
synonyms("Poa annua", db='tropicos', rows=1)
synonyms("Poa annua", db='tropicos', rows=1:3)
synonyms("Pinus sylvestris", db='nbn', rows=1:3)

# Use curl options
synonyms("Poa annua", db='tropicos', rows=1, verbose = TRUE)
synonyms("Poa annua", db='itis', rows=1, verbose = TRUE)


# combine many outputs together
x <- synonyms(c("Osmia bicornis", "Osmia rufa", "Osmia"), db = "itis")
synonyms_df(x)

## note here how Pinus contorta is dropped due to no synonyms found
synonyms_df(x)

## note here that ids are taxon identifiers b/c you start with them
x <- synonyms(c(25509881, 13100094), db="tropicos")
synonyms_df(x)

## NBN
x <- synonyms(c('Aglais io', 'Usnea hirta', 'Arctostaphylos uva-ursi'),
  db="nbn")
synonyms_df(x)
}
}
\seealso{
\code{\link[=get_tsn]{get_tsn()}} \code{\link[=get_tpsid]{get_tpsid()}} \code{\link[=get_nbnid]{get_nbnid()}}
\code{\link[=get_wormsid]{get_wormsid()}} \code{\link[=get_iucn]{get_iucn()}} \code{\link[=get_pow]{get_pow()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_synonyms.R
\name{tp_synonyms}
\alias{tp_synonyms}
\title{Return all synonyms for a taxon name with a given id.}
\usage{
tp_synonyms(id, key = NULL, ...)
}
\arguments{
\item{id}{the taxon identifier code}

\item{key}{Your Tropicos API key; See \link{taxize-authentication}
for help on authentication}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
List or dataframe.
}
\description{
Return all synonyms for a taxon name with a given id.
}
\examples{
\dontrun{
tp_synonyms(id = 25509881)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gni_search.R
\name{gni_search}
\alias{gni_search}
\title{Search for taxonomic names using the Global Names Index}
\usage{
gni_search(
  search_term = NULL,
  per_page = NULL,
  page = NULL,
  justtotal = FALSE,
  parse_names = FALSE,
  ...
)
}
\arguments{
\item{search_term}{Name pattern you want to search for. WARNING: Does
not work for vernacular/common names. Search term may include following
options (Note: can, uni, gen, sp, ssp, au, yr work only for parsed names
\itemize{
\item 
\itemize{
\item wild card - Search by part of a word (E.g.: planta*)
}
\item exact exact match    - Search for exact match of a literal string
(E.g.: exact:Parus major)
\item ns name string- Search for literal string from its beginning (other
modifiers will be ignored) (E.g.: ns:parus maj*)
\item can canonical form- Search name without authors (other modifiers will
be ignored)    (E.g.: can:parus major)
\item uni uninomial- Search for higher taxa    (E.g.: uni:parus)
\item gen genus - Search by genus epithet of species name (E.g.: gen:parus)
\item sp species - Search by species epithet (E.g.: sp:major)
\item ssp subspecies - Search by infraspecies epithet (E.g.: ssp:major)
\item au author - Search by author word    (E.g.: au:Shipunov)
\item yr year - Search by year (E.g.: yr:2005)
}}

\item{per_page}{Number of items per one page (numbers larger
than 1000 will be decreased to 1000) (default is 30).}

\item{page}{Page number you want to see (default is 1).}

\item{justtotal}{Return only the total results found.}

\item{parse_names}{If \code{TRUE}, use \code{\link[=gni_parse]{gni_parse()}} to parse
names. Default: \code{FALSE}}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
data.frame of results.
}
\description{
Uses the Global Names Index, see http://gni.globalnames.org
}
\details{
Note that you can use fuzzy searching, e.g., by attaching an
asterisk to the end of a search term. See the first two examples below
}
\examples{
\dontrun{
gni_search(search_term = "ani*")
gni_search(search_term = "ama*", per_page = 3, page = 21)
gni_search(search_term = "animalia", per_page = 8, page = 1)
gni_search(search_term = "animalia", per_page = 8, page = 1, justtotal=TRUE)

gni_search(search_term = "Cyanistes caeruleus", parse_names=TRUE)

# pass on curl options
gni_search(search_term = "ani*", verbose = TRUE)
}
}
\references{
\url{http://gni.globalnames.org/},
\url{https://github.com/dimus/gni/wiki/api}
}
\seealso{
\code{\link[=gnr_datasources]{gnr_datasources()}}, \code{\link[=gni_search]{gni_search()}}
}
\author{
Scott Chamberlain \email{myrmecocystus@gmail.com}
}
\keyword{globalnamesindex}
\keyword{names}
\keyword{taxonomy}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_downstream.R
\name{itis_downstream}
\alias{itis_downstream}
\title{Retrieve all taxa names or TSNs downstream in hierarchy from given TSN.}
\usage{
itis_downstream(id, downto, intermediate = FALSE, tsns = NULL, ...)
}
\arguments{
\item{id}{A taxonomic serial number.}

\item{downto}{The taxonomic level you want to go down to. See examples below.
The taxonomic level IS case sensitive, and you do have to spell it
correctly. See \code{data(rank_ref)} for spelling.}

\item{intermediate}{(logical) If TRUE, return a list of length two with target
taxon rank names, with additional list of data.frame's of intermediate
taxonomic groups. Default: FALSE}

\item{tsns}{Deprecated, see \code{id}}

\item{...}{Further args passed on to \code{\link[ritis:rank_name]{ritis::rank_name()}} and
\code{\link[ritis:hierarchy]{ritis::hierarchy_down()}}}
}
\value{
Data.frame of taxonomic information downstream to family from e.g.,
Order, Class, etc., or if \code{intermediated=TRUE}, list of length two,
with target taxon rank names, and intermediate names.
}
\description{
Retrieve all taxa names or TSNs downstream in hierarchy from given TSN.
}
\examples{
\dontrun{
## the plant class Bangiophyceae, tsn 846509
itis_downstream(id = 846509, downto="genus")
itis_downstream(id = 846509, downto="genus", intermediate=TRUE)

# get families downstream from Acridoidea
itis_downstream(id = 650497, "family")
## here, intermediate leads to the same result as the target
itis_downstream(id = 650497, "family", intermediate=TRUE)

# get species downstream from Ursus
itis_downstream(id = 180541, "species")

# get orders down from the Division Rhodophyta (red algae)
itis_downstream(id = 660046, "order")
itis_downstream(id = 660046, "order", intermediate=TRUE)

# get tribes down from the family Apidae
itis_downstream(id = 154394, downto="tribe")
itis_downstream(id = 154394, downto="tribe", intermediate=TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ids.R
\name{get_ids}
\alias{get_ids}
\alias{get_ids_}
\title{Retrieve taxonomic identifiers for a given taxon name.}
\usage{
get_ids(
  sci_com,
  db = c("itis", "ncbi", "eol", "tropicos", "gbif", "nbn", "pow"),
  suppress = FALSE,
  names = NULL,
  ...
)

get_ids_(
  sci_com,
  db = get_ids_dbs,
  rows = NA,
  suppress = FALSE,
  names = NULL,
  ...
)
}
\arguments{
\item{sci_com}{(character) Taxonomic name to query.}

\item{db}{(character) database to query. One or more of \code{ncbi}, \code{itis}, \code{eol},
\code{tropicos}, \code{gbif}, \code{nbn}, or \code{pow}. By default db is set to search
all data sources. Note that each taxonomic data source has their own
identifiers, so that if you give the wrong \code{db} value for the identifier you
could get a result, it will likely be wrong (not what you were expecting).
If using ncbi and/or tropicos we recommend getting API keys;
see \link{taxize-authentication}}

\item{suppress}{(logical) suppress \pkg{cli} separators with the database
name being queried. default: \code{FALSE}}

\item{names}{Deprecated, see \code{sci_com}}

\item{...}{Other arguments passed to \code{\link[=get_tsn]{get_tsn()}}, \code{\link[=get_uid]{get_uid()}},
\code{\link[=get_eolid]{get_eolid()}}, \code{\link[=get_tpsid]{get_tpsid()}}, \code{\link[=get_gbifid]{get_gbifid()}},
\code{\link[=get_nbnid]{get_nbnid()}}.}

\item{rows}{(numeric) Any number from 1 to infinity. If the default NA, all
rows are returned. When used in \code{get_ids} this function still only
gives back a ids class object with one to many identifiers. See
\code{get_ids_} to get back all, or a subset, of the raw data that you
are presented during the ask process.}
}
\value{
A vector of taxonomic identifiers, each retaining their respective
S3 classes so that each element can be passed on to another function
(see e.g.'s).
}
\description{
This is a convenience function to get identifiers across all data sources.
You can use other \verb{get_*} functions to get identifiers from specific
sources if you like.
}
\note{
There is a timeout of 1/3 seconds between queries to NCBI.
}
\section{Authentication}{

See \link{taxize-authentication} for help on authentication
}

\examples{
\dontrun{
# Plug in taxon names directly
# specify rows to limit choices available
get_ids("Poa annua", db="eol", rows=1)
get_ids("Poa annua", db="eol", rows=1:2)

## Or you can specify which source you want via the db parameter
get_ids("Chironomus riparius", db = 'ncbi')
get_ids("Salvelinus fontinalis", db = 'nbn')

get_ids(c("Chironomus riparius", "Pinus contorta"), db = 'ncbi')
get_ids("Pinus contorta", db = c('ncbi','eol','tropicos'))
get_ids("ava avvva", db = c('ncbi','eol','tropicos'))

# Pass on to other functions
out <- get_ids("Pinus contorta", db = c('ncbi','eol','tropicos'))
classification(out$ncbi)

# Get all data back
get_ids_(c("Chironomus riparius", "Pinus contorta"), db = 'nbn',
  rows=1:10)
get_ids_(c("Chironomus riparius", "Pinus contorta"), db = c('nbn','gbif'),
  rows=1:10)

# use curl options
get_ids("Agapostemon", db = "ncbi", verbose = TRUE)
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eubon.R
\name{eubon}
\alias{eubon}
\alias{eubon_search}
\title{EUBON taxonomy search}
\usage{
eubon(query, providers = "pesi", searchMode = "scientificNameExact",
  addSynonymy = FALSE, addParentTaxon = FALSE, timeout = 0,
  dedup = NULL, ...)

eubon_search(query, providers = "pesi", searchMode = "scientificNameExact",
  addSynonymy = FALSE, addParentTaxon = FALSE, timeout = 0,
  dedup = NULL, ...)
}
\arguments{
\item{query}{(character) The scientific name to search for. For example:
"Bellis perennis", "Prionus" or "Bolinus brandaris". This is an exact search
so wildcard characters are not supported}

\item{providers}{(character) A list of provider id strings concatenated by
comma characters. The default : "pesi,bgbm-cdm-server[col]" will be used if
this parameter is not set. A list of all available provider ids can be
obtained from the '/capabilities' service end point. Providers can be
nested, that is a parent provider can have sub providers. If the id of the
parent provider is supplied all subproviders will be queried. The query
can also be restricted to one or more subproviders by using the following
syntax: parent-id[sub-id-1,sub-id2,...]}

\item{searchMode}{(character) Specifies the searchMode. Possible search
modes are: scientificNameExact, scientificNameLike (begins with),
vernacularNameExact, vernacularNameLike
(contains), findByIdentifier. If the a provider does not support the
chosen searchMode it will be skipped and the status message in the
tnrClientStatus will be set to 'unsupported search mode' in this case.}

\item{addSynonymy}{(logical) Indicates whether the synonymy of the accepted
taxon should be included into the response. Turning this option on may
cause an increased response time. Default: \code{FALSE}}

\item{addParentTaxon}{(logical) Indicates whether the the parent taxon of
the accepted taxon should be included into the response. Turning this option
on may cause a slightly increased response time. Default: \code{FALSE}}

\item{timeout}{(numeric) The maximum of milliseconds to wait for responses
from any of the providers. If the timeout is exceeded the service will just
return the responses that have been received so far. The default timeout is
0 ms (wait for ever)}

\item{dedup}{(character) Allows to deduplicate the results by making use of
a deduplication strategy. The deduplication is done by comparing
specific properties of the taxon:
\itemize{
 \item id: compares 'taxon.identifier'
 \item id_name: compares 'taxon.identifier' AND
 'taxon.taxonName.scientificName'
 \item name: compares 'taxon.taxonName.scientificName' Using the pure
 'name' strategy is not recommended.
}}

\item{...}{Curl options passed on to \code{\link[httr]{GET}}}
}
\description{
EUBON taxonomy search
}
\details{
Note that paging is not yet implemented, so you only get the first
chunk of up to 50 results for methods that require paging. We will
implement paging here when it is available in the EU BON API.
}
\examples{
\dontrun{
eubon_search("Prionus")
eubon_search("Salmo", 'pesi')
eubon_search("Salmo", c('pesi', 'worms'))
eubon_search("Salmo", 'worms', 'scientificNameLike')
eubon_search("Salmo", 'worms', addSynonymy = TRUE)
eubon_search("Salmo", 'worms', addParentTaxon = TRUE)
}
}
\references{
\url{http://cybertaxonomy.eu/eu-bon/utis/1.2/doc.html}
}
\seealso{
Other eubon-methods: \code{\link{eubon_capabilities}},
  \code{\link{eubon_children}},
  \code{\link{eubon_hierarchy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_dist.R
\name{tp_dist}
\alias{tp_dist}
\title{Return all distribution records for for a taxon name with a given id.}
\usage{
tp_dist(id, key = NULL, ...)
}
\arguments{
\item{id}{the taxon identifier code}

\item{key}{Your Tropicos API key; See \link{taxize-authentication}
for help on authentication}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
List of two data.frame's, one named "location", and one "reference".
}
\description{
Return all distribution records for for a taxon name with a given id.
}
\examples{
\dontrun{
# Query using a taxon name Id
out <- tp_dist(id = 25509881)
## just location data
head(out[['location']])
## just reference data
head(out[['reference']])
}
}
\references{
http://services.tropicos.org/help?method=GetNameDistributionsXml
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncbi_search.R
\name{ncbi_search}
\alias{ncbi_search}
\title{Search for gene sequences available for taxa from NCBI.}
\usage{
ncbi_search(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nbn_synonyms.R
\name{nbn_synonyms}
\alias{nbn_synonyms}
\title{Return all synonyms for a taxon name with a given id from NBN}
\usage{
nbn_synonyms(id, ...)
}
\arguments{
\item{id}{the taxon identifier code}

\item{...}{Further args passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A data.frame
}
\description{
Return all synonyms for a taxon name with a given id from NBN
}
\examples{
\dontrun{
nbn_synonyms(id = 'NHMSYS0001501147')
nbn_synonyms(id = 'NHMSYS0000456036')

# none
nbn_synonyms(id = 'NHMSYS0000502940')
}
}
\references{
https://api.nbnatlas.org/
}
\seealso{
Other nbn: 
\code{\link{get_nbnid}()},
\code{\link{nbn_classification}()},
\code{\link{nbn_search}()}
}
\concept{nbn}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_name.R
\name{itis_name}
\alias{itis_name}
\title{Get taxonomic names for a given taxonomic name query.}
\usage{
itis_name(query = NULL, get = NULL)
}
\arguments{
\item{query}{TSN number (taxonomic serial number).}

\item{get}{The rank of the taxonomic name to get.}
}
\value{
Taxonomic name for the searched taxon.
}
\description{
Get taxonomic names for a given taxonomic name query.
}
\examples{
\dontrun{
itis_name(query="Helianthus annuus", get="family")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_classification.R
\name{tp_classification}
\alias{tp_classification}
\title{Return all synonyms for a taxon name with a given id.}
\usage{
tp_classification(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_colid.R
\name{get_colid}
\alias{get_colid}
\alias{as.colid}
\alias{as.colid.colid}
\alias{as.colid.character}
\alias{as.colid.list}
\alias{as.colid.data.frame}
\alias{as.data.frame.colid}
\alias{get_colid_}
\title{Get the Catalogue of Life ID from taxonomic names.}
\usage{
get_colid(sciname, ask = TRUE, verbose = TRUE, rows = NA,
  kingdom = NULL, phylum = NULL, class = NULL, order = NULL,
  family = NULL, rank = NULL, ...)

as.colid(x, check = TRUE)

\method{as.colid}{colid}(x, check = TRUE)

\method{as.colid}{character}(x, check = TRUE)

\method{as.colid}{list}(x, check = TRUE)

\method{as.colid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{colid}(x, ...)

get_colid_(sciname, verbose = TRUE, rows = NA)
}
\arguments{
\item{sciname}{character; scientific name.}

\item{ask}{logical; should get_colid be run in interactive mode?
If TRUE and more than one ID is found for the species, the user is asked for
input. If FALSE NA is returned for multiple matches.}

\item{verbose}{logical; If TRUE the actual taxon queried is printed on the
console.}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all rows are considered.
Note that this function still only gives back a colid class object with one to many identifiers.
See \code{\link[taxize]{get_colid_}} to get back all, or a subset, of the raw data that you are
presented during the ask process.}

\item{kingdom}{(character) A kingdom name. Optional. See \code{Filtering}
below.}

\item{phylum}{(character) A phylum (aka division) name. Optional. See \code{Filtering}
below.}

\item{class}{(character) A class name. Optional. See \code{Filtering} below.}

\item{order}{(character) An order name. Optional. See \code{Filtering} below.}

\item{family}{(character) A family name. Optional. See \code{Filtering} below.}

\item{rank}{(character) A taxonomic rank name. See \code{\link{rank_ref}} for possible
options. Though note that some data sources use atypical ranks, so inspect the
data itself for options. Optional. See \code{Filtering} below.}

\item{...}{Ignored}

\item{x}{Input to as.colid}

\item{check}{logical; Check if ID matches any existing on the DB, only used in
\code{\link{as.colid}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Get the Catalogue of Life ID from taxonomic names.
}
\section{Filtering}{

The parameters \code{kingdom}, \code{phylum}, \code{class}, \code{order}, \code{family},
and \code{rank} are not used in the search to the data provider, but are used in filtering
the data down to a subset that is closer to the target you want. For all these parameters,
you can use regex strings since we use \code{\link{grep}} internally to match.
Filtering narrows down to the set that matches your query, and removes the rest.
}

\examples{
\dontrun{
get_colid(sciname='Poa annua')
get_colid(sciname='Pinus contorta')
get_colid(sciname='Puma concolor')
# get_colid(sciname="Abudefduf saxatilis")

get_colid(c("Poa annua", "Pinus contorta"))

# specify rows to limit choices available
get_colid(sciname='Poa annua')
get_colid(sciname='Poa annua', rows=1)
get_colid(sciname='Poa annua', rows=2)
get_colid(sciname='Poa annua', rows=1:2)

# When not found
get_colid(sciname="uaudnadndj")
get_colid(c("Chironomus riparius", "uaudnadndj"))

# Narrow down results to a division or rank, or both
## Satyrium example
### Results w/o narrowing
get_colid("Satyrium")
### w/ division
get_colid("Satyrium", kingdom = "Plantae")
get_colid("Satyrium", kingdom = "Animalia")

## Rank example
get_colid("Poa")
get_colid("Poa", kingdom = "Plantae")
get_colid("Poa", kingdom = "Animalia")

# Fuzzy filter on any filtering fields
## uses grep on the inside
get_colid("Satyrium", kingdom = "p")

# Convert a uid without class information to a uid class
as.colid(get_colid("Chironomus riparius")) # already a uid, returns the same
as.colid(get_colid(c("Chironomus riparius","Pinus contorta"))) # same
as.colid("714831352ad94741e4321eccdeb29f58") # character
# character vector, length > 1
as.colid(c("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d"))
# list, either numeric or character
as.colid(list("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d"))
## dont check, much faster
as.colid("714831352ad94741e4321eccdeb29f58", check=FALSE)
as.colid(c("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d"),
 check=FALSE)
as.colid(list("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d"),
 check=FALSE)

(out <- as.colid(c("714831352ad94741e4321eccdeb29f58", "3b35900f74ff6e4b073ddb95c32b1f8d")))
data.frame(out)
as.colid( data.frame(out) )

# Get all data back
get_colid_("Poa annua")
get_colid_("Poa annua", rows=2)
get_colid_("Poa annua", rows=1:2)
get_colid_(c("asdfadfasd","Pinus contorta"))

get_colid(sciname="Andropadus nigriceps fusciceps", rows=1)

# use curl options
library("httr")
get_colid("Quercus douglasii", config=verbose())
bb <- get_colid("Quercus douglasii", config=progress())
}
}
\seealso{
\code{\link[taxize]{classification}}

Other taxonomic-ids: \code{\link{get_boldid}},
  \code{\link{get_eolid}}, \code{\link{get_gbifid}},
  \code{\link{get_ids}}, \code{\link{get_iucn}},
  \code{\link{get_natservid}}, \code{\link{get_nbnid}},
  \code{\link{get_tolid}}, \code{\link{get_tpsid}},
  \code{\link{get_tsn}}, \code{\link{get_uid}},
  \code{\link{get_wiki}}, \code{\link{get_wormsid}}
}
\author{
Scott Chamberlain, \email{myrmecocystus@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ubio_id.R
\name{ubio_id}
\alias{ubio_id}
\title{Search uBio by namebank ID.}
\usage{
ubio_id(...)
}
\arguments{
\item{...}{Parameters, ignored}
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apg_lookup.R
\name{apg_lookup}
\alias{apg_lookup}
\title{Lookup in the APGIII taxonomy and replace family names}
\usage{
apg_lookup(taxa, rank = "family")
}
\arguments{
\item{taxa}{(character) Taxonomic name to lookup a synonym for
in APGIII taxonomy.}

\item{rank}{(character) Taxonomic rank to lookup a synonym for. One of
family or order.}
}
\value{
A APGIII family or order name, the original name if the name
is the same as APG has, or NA if no match found
}
\description{
Lookup in the APGIII taxonomy and replace family names
}
\details{
Internally in this function, we use the datasets \link{apg_families}
and \link{apg_orders} - see their descriptions for the data in them.
The functions \code{\link[=apgOrders]{apgOrders()}} \code{\link[=apgFamilies]{apgFamilies()}} are for scraping
current content from the http://www.mobot.org/MOBOT/research/APweb/
website

The datasets used in this function are from the most recent version of
APGIII, Version 14 (http://www.mobot.org/MOBOT/research/APweb/)
}
\examples{
# New name found
apg_lookup(taxa = "Hyacinthaceae", rank = "family")
# Name is the same
apg_lookup(taxa = "Poaceae", rank = "family")
apg_lookup(taxa = "Asteraceae", rank = "family")
# Name not found
apg_lookup(taxa = "Foobar", rank = "family")

# New name found
apg_lookup(taxa = "Acerales", rank = "order")
# Name is the same
apg_lookup(taxa = "Acorales", rank = "order")
# Name not found
apg_lookup(taxa = "Foobar", rank = "order")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status_codes.R
\name{status_codes}
\alias{status_codes}
\title{Get HTTP status codes}
\usage{
status_codes()
}
\description{
Get HTTP status codes
}
\examples{
status_codes()
}
\seealso{
\code{\link[=ping]{ping()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col_classification.R
\name{col_classification}
\alias{col_classification}
\title{Search Catalogue of Life for taxonomic classifications.}
\usage{
col_classification(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{apg_families}
\alias{apg_families}
\title{MOBOT family names}
\format{
A data frame with 1705 rows and 6 variables:
\itemize{
\item \code{family}: family name
\item \code{synonym}: if \code{accepted=FALSE}, this is the accepted name;
if \code{accepted=TRUE}, this is \code{NA}, and the name in \code{family} is accepted
\item \code{order}: order name for the family
\item \code{accepted}: logical, if name in \code{family} column is accepted or not
\item \code{original}: original data record from APG website, mapping
name in \code{family} column to a new name, if there is one
\item \code{accepted_name}: accepted name. accepted names, combining those that
are accepted from \code{family} column, with the new name from \code{synonym}
if applicable
}
}
\source{
http://www.mobot.org/MOBOT/research/APweb/
}
\description{
Family names and their replacements from the Angiosperm Phylogeny
Website system of flowering plant classification.
}
\details{
This dataset is from Version 14, incorporated on 2020-06-03,
generated using \code{\link[=apgFamilies]{apgFamilies()}}

(update script in inst/ignore/apg_script.R)
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_native.R
\name{itis_native}
\alias{itis_native}
\title{Get jurisdiction data, i.e., native or not native in a region.}
\usage{
itis_native(tsn = NULL, what = "bytsn", ...)
}
\arguments{
\item{tsn}{One or more TSN's (taxonomic serial number)}

\item{what}{One of bytsn, values, or originvalues}

\item{...}{Further arguments passed on to
\code{\link[ritis:jurisdiction]{ritis::jurisdictional_origin()}},
\code{\link[ritis:jurisdiction]{ritis::jurisdiction_values()}}, or
\code{\link[ritis:jurisdiction]{ritis::jurisdiction_origin_values()}}}
}
\description{
Get jurisdiction data, i.e., native or not native in a region.
}
\examples{
\dontrun{
# Get values
itis_native(what="values")

# Get origin values
itis_native(what="originvalues")

# Get values by tsn
itis_native(tsn=180543)
itis_native(tsn=c(180543,41074,36616))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncbi_get_taxon_summary.R
\name{ncbi_get_taxon_summary}
\alias{ncbi_get_taxon_summary}
\title{NCBI taxon information from uids}
\usage{
ncbi_get_taxon_summary(id, key = NULL, ...)
}
\arguments{
\item{id}{(character) NCBI taxonomy uids to retrieve information for.
See Details.}

\item{key}{(character) NCBI Entrez API key. optional. See Details.}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A \code{data.frame} with the following columns:
\itemize{
\item \code{uid} The uid queried for
\item \code{name} The name of the taxon; a binomial name if the taxon is of rank species
\item \code{rank} The taxonomic rank (e.g. 'Genus')
}
}
\description{
Downloads summary taxon information from the NCBI taxonomy
databases for a set of taxonomy UIDs using eutils esummary.
}
\details{
If your input vector or list of NCBI IDs is longer than about
2500 characters (use \code{nchar(paste(ids, collapse = "+"))}), split
the list up into chunks since at about that number of characters you
will run into the HTTP 414 error "Request-URI Too Long".
}
\section{HTTP version}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\section{Authentication}{

See \link{taxize-authentication} for help on authentication.
We strongly recommend getting an API key
}

\examples{
\dontrun{
ncbi_get_taxon_summary(c(1430660, 4751))

# use curl options
ncbi_get_taxon_summary(c(1430660, 4751), verbose = TRUE)
}
}
\author{
Zachary Foster \email{zacharyfoster1989@Sgmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_tsn.R
\name{get_tsn}
\alias{get_tsn}
\alias{as.tsn}
\alias{as.tsn.tsn}
\alias{as.tsn.character}
\alias{as.tsn.list}
\alias{as.tsn.numeric}
\alias{as.tsn.data.frame}
\alias{as.data.frame.tsn}
\alias{get_tsn_}
\title{Get the TSN code for a search term.}
\usage{
get_tsn(
  sci_com,
  searchtype = "scientific",
  accepted = FALSE,
  ask = TRUE,
  messages = TRUE,
  rows = NA,
  searchterm = NULL,
  ...
)

as.tsn(x, check = TRUE)

\method{as.tsn}{tsn}(x, check = TRUE)

\method{as.tsn}{character}(x, check = TRUE)

\method{as.tsn}{list}(x, check = TRUE)

\method{as.tsn}{numeric}(x, check = TRUE)

\method{as.tsn}{data.frame}(x, check = TRUE)

\method{as.data.frame}{tsn}(x, ...)

get_tsn_(
  sci_com,
  messages = TRUE,
  searchtype = "scientific",
  accepted = TRUE,
  rows = NA,
  searchterm = NULL,
  ...
)
}
\arguments{
\item{sci_com}{character; A vector of common or scientific names.
Or, a \code{taxon_state} object (see \link{taxon-state})}

\item{searchtype}{character; One of 'scientific' or 'common', or any
unique abbreviation}

\item{accepted}{logical; If TRUE, removes names that are not accepted valid
names by ITIS. Set to \code{FALSE} (default) to give back both accepted
and unaccepted names.}

\item{ask}{logical; should get_tsn be run in interactive mode?
If \code{TRUE} and more than one TSN is found for the species, the user is
asked for input. If \code{FALSE} NA is returned for multiple matches.}

\item{messages}{logical; should progress be printed?}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this function still only gives back a tsn
class object with one to many identifiers. See
\code{\link[=get_tsn_]{get_tsn_()}} to get back all, or a subset, of the raw
data that you are presented during the ask process.}

\item{searchterm}{Deprecated, see \code{sci_com}}

\item{...}{Ignored}

\item{x}{Input to as.tsn}

\item{check}{logical; Check if ID matches any existing on the DB, only
used in \code{\link[=as.tsn]{as.tsn()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Retrieve the taxonomic serial numbers (TSN) of a taxon from ITIS.
}
\examples{
\dontrun{
get_tsn("Quercus douglasii")
get_tsn("Chironomus riparius")
get_tsn(c("Chironomus riparius","Quercus douglasii"))
splist <- c("annona cherimola", 'annona muricata', "quercus robur",
		"shorea robusta", "pandanus patina", "oryza sativa", "durio zibethinus")
get_tsn(splist, messages=FALSE)

# specify rows to limit choices available
get_tsn('Arni')
get_tsn('Arni', rows=1)
get_tsn('Arni', rows=1:2)

# When not found
get_tsn("howdy")
get_tsn(c("Chironomus riparius", "howdy"))

# Using common names
get_tsn("black bear", searchtype="common")

# Convert a tsn without class information to a tsn class
as.tsn(get_tsn("Quercus douglasii")) # already a tsn, returns the same
as.tsn(get_tsn(c("Chironomus riparius","Pinus contorta"))) # same
as.tsn(19322) # numeric
as.tsn(c(19322,129313,506198)) # numeric vector, length > 1
as.tsn("19322") # character
as.tsn(c("19322","129313","506198")) # character vector, length > 1
as.tsn(list("19322","129313","506198")) # list, either numeric or character
## dont check, much faster
as.tsn("19322", check=FALSE)
as.tsn(19322, check=FALSE)
as.tsn(c("19322","129313","506198"), check=FALSE)
as.tsn(list("19322","129313","506198"), check=FALSE)

(out <- as.tsn(c(19322,129313,506198)))
data.frame(out)
as.tsn( data.frame(out) )

# Get all data back
get_tsn_("Arni")
get_tsn_("Arni", rows=1)
get_tsn_("Arni", rows=1:2)
get_tsn_(c("asdfadfasd","Pinus contorta"), rows=1:5)
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tax_name.R
\name{tax_name}
\alias{tax_name}
\title{Get taxonomic names for a given rank}
\usage{
tax_name(
  sci,
  get,
  db = "itis",
  pref = "ncbi",
  messages = TRUE,
  query = NULL,
  ...
)
}
\arguments{
\item{sci}{(character) Vector of taxonomic names to query. required.}

\item{get}{(character) The ranks of the taxonomic name to get, see
\link{rank_ref}. required.}

\item{db}{(character) The database to search from: 'itis', 'ncbi' or 'both'.
If 'both' both NCBI and ITIS will be queried. Result will be the union of
both. If using ncbi, we recommend getting an API key; see
\link{taxize-authentication}}

\item{pref}{(character) If db = 'both', sets the preference for the union.
Either 'ncbi' (default) or 'itis'. Currently not implemented.}

\item{messages}{(logical) If \code{TRUE} the actual taxon queried is printed
on the console.}

\item{query}{Deprecated, see \code{sci}}

\item{...}{Other arguments passed to \code{\link[=get_tsn]{get_tsn()}} or
\code{\link[=get_uid]{get_uid()}}.}
}
\value{
A data.frame with one column for every queried rank, in addition to
a column for db and queried term.
}
\description{
Get taxonomic names for a given rank
}
\note{
While \code{\link[=tax_rank]{tax_rank()}} returns the actual rank of a
taxon, \code{\link[=tax_name]{tax_name()}} searches and returns any specified rank
higher in taxonomy.
}
\section{Authentication}{

See \link{taxize-authentication} for help on authentication
}

\examples{
\dontrun{
# A case where itis and ncbi use the same names
tax_name(sci = "Helianthus annuus", get = "family", db = "itis")
tax_name(sci = "Helianthus annuus", get = "family", db = "ncbi")
tax_name(sci = "Helianthus annuus", get = c("genus","family","order"),
  db = "ncbi")

# Case where itis and ncbi use different names
tax_name(sci = "Helianthus annuus", get = "kingdom", db = "itis")
tax_name(sci = "Helianthus annuus", get = "kingdom", db = "ncbi")

# multiple rank arguments
tax_name(sci = c("Helianthus annuus","Baetis rhodani"), get = c("genus",
"kingdom"), db = "ncbi")
tax_name(sci = c("Helianthus annuus","Baetis rhodani"), get = c("genus",
"kingdom"), db = "itis")

# query both sources
tax_name(sci=c("Helianthus annuus", 'Baetis rhodani'), get=c("genus",
"kingdom"), db="both")
}
}
\seealso{
\code{\link[=classification]{classification()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ubioid.R
\name{get_ubioid}
\alias{get_ubioid}
\alias{as.ubioid}
\alias{get_ubioid_}
\title{Get the uBio id for a search term}
\usage{
get_ubioid(...)

as.ubioid(...)

get_ubioid_(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/id2name.R
\name{id2name}
\alias{id2name}
\alias{id2name.default}
\alias{id2name.tolid}
\alias{id2name.tsn}
\alias{id2name.uid}
\alias{id2name.wormsid}
\alias{id2name.gbifid}
\alias{id2name.boldid}
\title{Taxonomic IDs to taxonomic names}
\usage{
id2name(id, db = NULL, x = NULL, ...)

\method{id2name}{default}(id, db = NULL, x = NULL, ...)

\method{id2name}{tolid}(id, ...)

\method{id2name}{tsn}(id, ...)

\method{id2name}{uid}(id, ...)

\method{id2name}{wormsid}(id, ...)

\method{id2name}{gbifid}(id, ...)

\method{id2name}{boldid}(id, ...)
}
\arguments{
\item{id}{vector of taxonomic IDs (character or numeric)}

\item{db}{(character) database to query. One or more of \code{tol}, \code{itis},
\code{ncbi}, \code{worms}, \code{gbif}, or \code{bold}. Note that each
taxonomic data source has their own  identifiers, so that if you provide
the wrong \code{db} value for the identifier you could get a result,
but it will likely be wrong (not what you were expecting). If using ncbi
we recommend getting API keys; see \link{taxize-authentication}}

\item{x}{Deprecated, see \code{id}}

\item{...}{Further args passed on to \code{tol_id2name} or
\link{itis_getrecord}, or other internal functions.
See those functions for what parameters can be passed on.}
}
\value{
A named list of data.frames, named by the input taxonomic ids
}
\description{
Taxonomic IDs to taxonomic names
}
\section{HTTP version for NCBI requests}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\examples{
\dontrun{
# ITIS
id2name(19322, db = "itis")

# TOL
id2name(515698, db = "tol")
# get NCBI ID and pass to classification()
x <- id2name(515698, db = "tol")
classification(as.uid(x[[1]]$tax_sources_ncbi))

# NCBI
id2name(315567, db = "ncbi")
id2name(3339, db = "ncbi")
id2name(9696, db = "ncbi")
id2name(c(9695, 9696), db = "ncbi")

# WORMS
id2name(105706, db = "worms")

# GBIF
id2name(2441176, db = "gbif")

# BOLD
id2name(88899, db = "bold")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_uid.R
\name{get_uid}
\alias{get_uid}
\alias{as.uid}
\alias{as.uid.uid}
\alias{as.uid.character}
\alias{as.uid.list}
\alias{as.uid.numeric}
\alias{as.uid.data.frame}
\alias{as.data.frame.uid}
\alias{get_uid_}
\title{Get the UID codes from NCBI for taxonomic names.}
\usage{
get_uid(
  sci_com,
  ask = TRUE,
  messages = TRUE,
  rows = NA,
  modifier = NULL,
  rank_query = NULL,
  division_filter = NULL,
  rank_filter = NULL,
  key = NULL,
  sciname = NULL,
  ...
)

as.uid(x, check = TRUE)

\method{as.uid}{uid}(x, check = TRUE)

\method{as.uid}{character}(x, check = TRUE)

\method{as.uid}{list}(x, check = TRUE)

\method{as.uid}{numeric}(x, check = TRUE)

\method{as.uid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{uid}(x, ...)

get_uid_(sci_com, messages = TRUE, rows = NA, key = NULL, sciname = NULL, ...)
}
\arguments{
\item{sci_com}{character; scientific or common name. Or, a \code{taxon_state}
object (see \link{taxon-state})}

\item{ask}{logical; should get_uid be run in interactive mode? If TRUE and
more than one TSN is found for the species, the user is asked for input. If
FALSE NA is returned for multiple matches.}

\item{messages}{logical; If \code{TRUE} (default) the actual taxon queried is
printed on the console.}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this function still only gives back a uid
class object with one to many identifiers. See \code{\link[=get_uid_]{get_uid_()}} to get back
all, or a subset, of the raw data that you are presented during the ask
process.}

\item{modifier}{(character) A modifier to the \code{sci_com} given. Options
include: Organism, Scientific Name, Common Name, All Names, Division,
Filter, Lineage, GC, MGC, Name Tokens, Next Level, PGC, Properties, Rank,
Subtree, Synonym, Text Word. These are not checked, so make sure they are
entered correctly, as is.}

\item{rank_query}{(character) A taxonomic rank name to modify the query sent
to NCBI. See \link{rank_ref} for possible options. Though note that
some data sources use atypical ranks, so inspect the data itself for
options. Optional. See \code{Querying} below.}

\item{division_filter}{(character) A division (aka phylum) name to filter
data after retrieved from NCBI. Optional. See \code{Filtering} below.}

\item{rank_filter}{(character) A taxonomic rank name to filter data after
retrieved from NCBI. See \link{rank_ref} for possible options.
Though note that some data sources use atypical ranks, so inspect the data
itself for options. Optional. See \code{Filtering} below.}

\item{key}{(character) NCBI Entrez API key. optional. See Details.}

\item{sciname}{Deprecated, see \code{sci_com}}

\item{...}{Ignored}

\item{x}{Input to \code{\link[=as.uid]{as.uid()}}}

\item{check}{logical; Check if ID matches any existing on the DB, only used
in \code{\link[=as.uid]{as.uid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Retrieve the Unique Identifier (UID) of a taxon from NCBI taxonomy browser.
}
\section{Rate limits}{

In case you run into errors due to your rate limit being exceeded, see
\code{\link[=taxize_options]{taxize_options()}}, where you can set \code{ncbi_sleep}.
}

\section{Querying}{
 The parameter \code{rank_query} is used in the search sent
to NCBI, whereas \code{rank_filter} filters data after it comes back. The
parameter \code{modifier} adds modifiers to the name. For example,
\code{modifier="Organism"} adds that to the name, giving e.g.,
\code{Helianthus[Organism]}.
}

\section{Filtering}{
 The parameters \code{division_filter} and
\code{rank_filter} are not used in the search to the data provider, but are
used in filtering the data down to a subset that is closer to the target
you want. For all these parameters, you can use regex strings since we use
\code{\link[=grep]{grep()}} internally to match. Filtering narrows down to the set
that matches your query, and removes the rest.
}

\section{Beware}{
 NCBI does funny things sometimes. E.g., if you search on
Fringella morel, a slight misspelling of the genus name, and a non-existent
epithet, NCBI gives back a morel fungal species. In addition, NCBI doesn't
really do fuzzy searching very well, so if there is a slight mis-spelling
in your names, you likely won't get what you are expecting. The lesson:
clean your names before using this function. Other data sources are better
about fuzzy matching.
}

\section{Authentication}{

See \link{taxize-authentication} for help on authentication

Note that even though you can't pass in your key to \code{as.uid} functions,
we still use your Entrez API key if you have it saved as an R option
or environment variable.
}

\section{HTTP version}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\examples{
\dontrun{
get_uid(c("Chironomus riparius", "Chaetopteryx"))
get_uid(c("Chironomus riparius", "aaa vva"))

# When not found
get_uid("howdy")
get_uid(c("Chironomus riparius", "howdy"))

# Narrow down results to a division or rank, or both
## By modifying the query
### w/ modifiers to the name
get_uid(sci_com = "Aratinga acuticauda", modifier = "Organism")
get_uid(sci_com = "bear", modifier = "Common Name")

### w/ rank query
get_uid(sci_com = "Pinus", rank_query = "genus")
get_uid(sci_com = "Pinus", rank_query = "subgenus")
### division query doesn't really work, for unknown reasons, so not available

## By filtering the result
## Echinacea example
### Results w/o narrowing
get_uid("Echinacea")
### w/ division
get_uid(sci_com = "Echinacea", division_filter = "eudicots")
get_uid(sci_com = "Echinacea", division_filter = "sea urchins")

## Satyrium example
### Results w/o narrowing
get_uid(sci_com = "Satyrium")
### w/ division
get_uid(sci_com = "Satyrium", division_filter = "monocots")
get_uid(sci_com = "Satyrium", division_filter = "butterflies")

## Rank example
get_uid(sci_com = "Pinus")
get_uid(sci_com = "Pinus", rank_filter = "genus")
get_uid(sci_com = "Pinus", rank_filter = "subgenus")

# Fuzzy filter on any filtering fields
## uses grep on the inside
get_uid("Satyrium", division_filter = "m")

# specify rows to limit choices available
get_uid('Dugesia') # user prompt needed
get_uid('Dugesia', rows=1) # 2 choices, so returns only 1 row, so no choices
get_uid('Dugesia', ask = FALSE) # returns NA for multiple matches

# Go to a website with more info on the taxon
res <- get_uid("Chironomus riparius")
browseURL(attr(res, "uri"))

# Convert a uid without class information to a uid class
as.uid(get_uid("Chironomus riparius")) # already a uid, returns the same
as.uid(get_uid(c("Chironomus riparius","Pinus contorta"))) # same
as.uid(315567) # numeric
as.uid(c(315567,3339,9696)) # numeric vector, length > 1
as.uid("315567") # character
as.uid(c("315567","3339","9696")) # character vector, length > 1
as.uid(list("315567","3339","9696")) # list, either numeric or character
## dont check, much faster
as.uid("315567", check=FALSE)
as.uid(315567, check=FALSE)
as.uid(c("315567","3339","9696"), check=FALSE)
as.uid(list("315567","3339","9696"), check=FALSE)

(out <- as.uid(c(315567,3339,9696)))
data.frame(out)
as.uid( data.frame(out) )

# Get all data back
get_uid_("Puma concolor")
get_uid_("Dugesia")
get_uid_("Dugesia", rows=2)
get_uid_("Dugesia", rows=1:2)
get_uid_(c("asdfadfasd","Pinus contorta"))

# use curl options
get_uid("Quercus douglasii", verbose = TRUE)
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\author{
Eduard Szoecs, \email{eduardszoecs@gmail.com}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gna_data_sources.R
\name{gna_data_sources}
\alias{gna_data_sources}
\title{Get metadata about GNA data sources}
\usage{
gna_data_sources(output_type = "table", ...)
}
\arguments{
\item{output_type}{What format of output to return. Either \code{'json'},
\code{'list'}, or \code{'table'}.}

\item{...}{Passed to \link[crul:HttpClient]{crul::HttpClient}.}
}
\description{
Downloads metadata about Global Names Architecture (GNA) data sources
available to be used in other GNA functions.
}
\examples{
\dontrun{

gna_data_sources()
}

}
\author{
Zachary S.L. Foster
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fungorum.R
\name{fungorum}
\alias{fungorum}
\alias{fg_name_search}
\alias{fg_author_search}
\alias{fg_epithet_search}
\alias{fg_name_by_key}
\alias{fg_name_full_by_lsid}
\alias{fg_all_updated_names}
\alias{fg_deprecated_names}
\title{Index Fungorum}
\usage{
fg_name_search(q, anywhere = TRUE, limit = 10, ...)

fg_author_search(q, anywhere = TRUE, limit = 10, ...)

fg_epithet_search(q, anywhere = TRUE, limit = 10, ...)

fg_name_by_key(key, ...)

fg_name_full_by_lsid(lsid, ...)

fg_all_updated_names(date, ...)

fg_deprecated_names(date, ...)
}
\arguments{
\item{q}{(character) Query term}

\item{anywhere}{(logical) Default: \code{TRUE}}

\item{limit}{(integer) Number of results to return. max limit
value appears to be 6000, not positive about that though}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}

\item{key}{(character) A IndexFungorum taxon key}

\item{lsid}{(character) an LSID, e.,g. "urn:lsid:indexfungorum.org:names:81085"}

\item{date}{(character) Date, of the form YYYMMDD}
}
\value{
A \code{data.frame}, or \code{NULL} if no results
}
\description{
Search for taxonomic names in Index Fungorum
}
\examples{
\dontrun{
# NameSearch
fg_name_search(q = "Gymnopus", limit = 2, verbose = TRUE)
fg_name_search(q = "Gymnopus")

# EpithetSearch
fg_epithet_search(q = "phalloides")

# NameByKey
fg_name_by_key(17703)

# NameFullByKey
fg_name_full_by_lsid("urn:lsid:indexfungorum.org:names:81085")

# AllUpdatedNames
fg_all_updated_names(date = gsub("-", "", Sys.Date() - 2))

# DeprecatedNames
fg_deprecated_names(date=20151001)

# AuthorSearch
fg_author_search(q = "Fayod", limit = 2)
}
}
\references{
http://www.indexfungorum.org/, API docs:
http://www.indexfungorum.org/ixfwebservice/fungus.asmx
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{plantGenusNames}
\alias{plantGenusNames}
\title{Vector of plant genus names from ThePlantList}
\format{A vector of length 793}
\source{
http://www.theplantlist.org
}
\description{
These names are from http://www.theplantlist.org, and are a
randomly chosen subset of genera names for the purpose of having some
names to play with for examples in this package.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_summary.R
\name{tp_summary}
\alias{tp_summary}
\title{Return summary data a taxon name with a given id.}
\usage{
tp_summary(id, key = NULL, ...)
}
\arguments{
\item{id}{the taxon identifier code}

\item{key}{Your Tropicos API key; See \link{taxize-authentication}
for help on authentication}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A data.frame.
}
\description{
Return summary data a taxon name with a given id.
}
\examples{
\dontrun{
tp_summary(id = 25509881)
tp_summary(id = 2700851)
tp_summary(id = 24900183)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_nbnid.R
\name{get_nbnid}
\alias{get_nbnid}
\alias{as.nbnid}
\alias{as.nbnid.nbnid}
\alias{as.nbnid.character}
\alias{as.nbnid.list}
\alias{as.nbnid.data.frame}
\alias{as.data.frame.nbnid}
\alias{get_nbnid_}
\title{Get the UK National Biodiversity Network ID from taxonomic names.}
\usage{
get_nbnid(
  sci_com,
  ask = TRUE,
  messages = TRUE,
  rec_only = FALSE,
  rank = NULL,
  rows = NA,
  name = NULL,
  ...
)

as.nbnid(x, check = TRUE)

\method{as.nbnid}{nbnid}(x, check = TRUE)

\method{as.nbnid}{character}(x, check = TRUE)

\method{as.nbnid}{list}(x, check = TRUE)

\method{as.nbnid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{nbnid}(x, ...)

get_nbnid_(
  sci_com,
  messages = TRUE,
  rec_only = FALSE,
  rank = NULL,
  rows = NA,
  name = NULL,
  ...
)
}
\arguments{
\item{sci_com}{character; a vector of common or scientific names. Or, a
\code{taxon_state} object (see \link{taxon-state})}

\item{ask}{logical; should get_nbnid be run in interactive mode?
If TRUE and more than one ID is found for the species, the user is asked for
input. If FALSE NA is returned for multiple matches.}

\item{messages}{logical; If TRUE the actual taxon queried is printed on the
console.}

\item{rec_only}{(logical) If \code{TRUE} ids of recommended names are
returned (i.e. synonyms are removed). Defaults to \code{FALSE}. Remember,
the id of a synonym is a taxa with 'recommended' name status.}

\item{rank}{(character) If given, we attempt to limit the results to those
taxa with the matching rank.}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this function still only gives back a nbnid
class object with one to many identifiers. See
\code{\link[=get_nbnid_]{get_nbnid_()}} to get back all, or a subset, of the raw
data that you are presented during the ask process.}

\item{name}{Deprecated, see \code{sci_com}}

\item{...}{Further args passed on to \code{nbn_search}}

\item{x}{Input to \code{\link[=as.nbnid]{as.nbnid()}}}

\item{check}{logical; Check if ID matches any existing on the DB, only
used in \code{\link[=as.nbnid]{as.nbnid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions

an object of class nbnid, a light wrapper around a character
string that is the taxonomic ID - includes attributes with relavant
metadata
}
\description{
Get the UK National Biodiversity Network ID from taxonomic names.
}
\examples{
\dontrun{
get_nbnid(sci_com='Poa annua')
get_nbnid(sci_com='Poa annua', rec_only=TRUE)
get_nbnid(sci_com='Poa annua', rank='Species')
get_nbnid(sci_com='Poa annua', rec_only=TRUE, rank='Species')
get_nbnid(sci_com='Pinus contorta')

# The NBN service handles common names too
get_nbnid(sci_com='red-winged blackbird')

# specify rows to limit choices available
get_nbnid('Poa ann')
get_nbnid('Poa ann', rows=1)
get_nbnid('Poa ann', rows=25)
get_nbnid('Poa ann', rows=1:2)

# When not found
get_nbnid(sci_com="uaudnadndj")
get_nbnid(c("Zootoca vivipara", "uaudnadndj"))
get_nbnid(c("Zootoca vivipara","Chironomus riparius", "uaudnadndj"))

# Convert an nbnid without class information to a nbnid class
as.nbnid(get_nbnid("Zootoca vivipara")) # already a nbnid, returns the same
as.nbnid(get_nbnid(c("Zootoca vivipara","Pinus contorta"))) # same
as.nbnid('NHMSYS0001706186') # character
# character vector, length > 1
as.nbnid(c("NHMSYS0001706186","NHMSYS0000494848","NBNSYS0000010867"))
# list
as.nbnid(list("NHMSYS0001706186","NHMSYS0000494848","NBNSYS0000010867"))
## dont check, much faster
as.nbnid('NHMSYS0001706186', check=FALSE)
as.nbnid(list("NHMSYS0001706186","NHMSYS0000494848","NBNSYS0000010867"),
  check=FALSE)

(out <- as.nbnid(c("NHMSYS0001706186","NHMSYS0000494848",
  "NBNSYS0000010867")))
data.frame(out)
as.nbnid( data.frame(out) )

# Get all data back
get_nbnid_("Zootoca vivipara")
get_nbnid_("Poa annua", rows=2)
get_nbnid_("Poa annua", rows=1:2)
get_nbnid_(c("asdfadfasd","Pinus contorta"), rows=1:5)

# use curl options
invisible(get_nbnid("Quercus douglasii", verbose = TRUE))
}
}
\references{
https://api.nbnatlas.org/
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}

Other nbn: 
\code{\link{nbn_classification}()},
\code{\link{nbn_search}()},
\code{\link{nbn_synonyms}()}
}
\author{
Scott Chamberlain,
}
\concept{nbn}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{classification}
\alias{classification}
\alias{classification.default}
\alias{classification.tsn}
\alias{classification.uid}
\alias{classification.eolid}
\alias{classification.tpsid}
\alias{classification.gbifid}
\alias{classification.nbnid}
\alias{classification.tolid}
\alias{classification.wormsid}
\alias{classification.natservid}
\alias{classification.boldid}
\alias{classification.wiki}
\alias{classification.pow}
\alias{classification.ids}
\alias{cbind.classification}
\alias{rbind.classification}
\alias{cbind.classification_ids}
\alias{rbind.classification_ids}
\title{Retrieve the taxonomic hierarchy for a given taxon ID.}
\usage{
classification(...)

\method{classification}{default}(
  sci_id,
  db = NULL,
  callopts = list(),
  return_id = TRUE,
  rows = NA,
  x = NULL,
  ...
)

\method{classification}{tsn}(id, return_id = TRUE, ...)

\method{classification}{uid}(
  id,
  callopts = list(),
  return_id = TRUE,
  batch_size = 50,
  max_tries = 3,
  ...
)

\method{classification}{eolid}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{tpsid}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{gbifid}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{nbnid}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{tolid}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{wormsid}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{natservid}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{boldid}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{wiki}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{pow}(id, callopts = list(), return_id = TRUE, ...)

\method{classification}{ids}(id, ...)

\method{cbind}{classification}(...)

\method{rbind}{classification}(...)

\method{cbind}{classification_ids}(...)

\method{rbind}{classification_ids}(...)
}
\arguments{
\item{...}{For \code{classification}: other arguments passed to \code{\link[=get_tsn]{get_tsn()}},
\code{\link[=get_uid]{get_uid()}}, \code{\link[=get_eolid]{get_eolid()}}, \code{\link[=get_tpsid]{get_tpsid()}}, \code{\link[=get_gbifid]{get_gbifid()}},
\code{\link[=get_wormsid]{get_wormsid()}}, \code{\link[=get_natservid]{get_natservid()}}, \code{\link[=get_wormsid]{get_wormsid()}}, \code{\link[=get_wiki]{get_wiki()}},
\code{\link[=get_pow]{get_pow()}}. For \code{rbind.classification} and \code{cbind.classification}: one or
more objects of class \code{classification}}

\item{sci_id}{Vector of taxa names (character) or IDs (character or numeric)
to query. For \code{db = "eol"}, EOL expects you to pass it a taxon id, called
\code{eolid} in the output of \code{\link[=get_eolid]{get_eolid()}}.}

\item{db}{character; database to query. either \code{ncbi}, \code{itis}, \code{eol},
\code{tropicos}, \code{gbif}, \code{nbn}, \code{worms}, \code{natserv}, \code{bold}, \code{wiki}, or \code{pow}.
Note that each taxonomic data source has, their own identifiers, so that
if you provide the wrong \code{db} value for the identifier you could get a
result, but it will likely be wrong (not what you were expecting). If using
ncbi, and/or tropicos, we recommend getting an API key; see
\link{taxize-authentication}}

\item{callopts}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}

\item{return_id}{(logical) If \code{TRUE} (default), return the taxon id
as well as the name and rank of taxa in the lineage returned.
Ignored for natserv as they don't return IDs in their taxonomic
classification data.}

\item{rows}{(numeric) Any number from 1 to infinity. If the default NA,
all rows are considered. Note that this parameter is ignored if you pass
in a taxonomic id instead of a name of class character.}

\item{x}{Deprecated, see \code{sci_id}}

\item{id}{character; identifiers, returned by \code{\link[=get_tsn]{get_tsn()}}, \code{\link[=get_uid]{get_uid()}},
\code{\link[=get_eolid]{get_eolid()}}, \code{\link[=get_tpsid]{get_tpsid()}}, \code{\link[=get_gbifid]{get_gbifid()}}, \code{\link[=get_tolid]{get_tolid()}},
\code{\link[=get_wormsid]{get_wormsid()}}, \code{\link[=get_natservid]{get_natservid()}}, \code{\link[=get_wormsid]{get_wormsid()}}, \code{\link[=get_wiki]{get_wiki()}},
\code{\link[=get_pow]{get_pow()}}}

\item{batch_size}{(numeric) For NCBI queries, specify the number of IDs to
lookup for each query.}

\item{max_tries}{(numeric) For NCBI queries, the number of times a particular
query will be attempted, assuming the first does not work.}
}
\value{
A named list of data.frames with the taxonomic classification of
every supplied taxa.
}
\description{
Retrieve the taxonomic hierarchy for a given taxon ID.
}
\details{
If IDs are supplied directly (not from the \verb{get_*} functions)
you must specify the type of ID. There is a timeout of 1/3 seconds between
queries to NCBI.

BEWARE: Right now, NBN doesn't return the queried taxon in the
classification. But you can attach it yourself quite easily of course.
This behavior is different from the other data sources.
}
\section{Lots of results}{

It may happen sometimes that you get more results back from your query
than will show in the data.frame on screen. Our advice is to refine your
query in those cases. On a data source basis we can attempt to help
make it easier to refine queries, whether it be with the data provider
(unlikely to happen), or in the code in this package (more likely) -
let us know if you run into too many results problem and we'll see what
we can do.
}

\section{Authentication}{

See \link{taxize-authentication}
}

\section{EOL}{

EOL does not have very good failure behavior. For example, if you submit
an ID that does not exist they'll return a 500 HTTP error, which is
not an appropriate error; it's probably that that ID does not exist
in their database, but we can't know for sure. Isn't that fun?
}

\section{NCBI Rate limits}{

In case you run into NCBI errors due to your rate limit being exceeded, see
\code{\link[=taxize_options]{taxize_options()}}, where you can set \code{ncbi_sleep}.
}

\section{HTTP version for NCBI requests}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\examples{
\dontrun{
# Plug in taxon IDs
classification(9606, db = 'ncbi')
classification(c(9606, 55062), db = 'ncbi')
classification(129313, db = 'itis')
classification(6985636, db = 'eol')
classification(126436, db = 'worms')
classification('Helianthus annuus', db = 'pow')
classification('Helianthus', db = 'pow')
classification('Asteraceae', db = 'pow')
classification("134717", db = 'natserv')
classification(c(2704179, 6162875, 8286319, 2441175, 731), db = 'gbif')
classification(25509881, db = 'tropicos')
classification("NBNSYS0000004786", db = 'nbn')
classification(as.nbnid("NBNSYS0000004786"), db = 'nbn')
classification(3930798, db = 'tol')

## works the same if IDs are in class character
classification(c("2704179", "2441176"), db = 'gbif')
classification("Agapostemon", db = "bold")

# wikispecies
classification("Malus domestica", db = "wiki")
classification("Pinus contorta", db = "wiki")
classification("Pinus contorta", db = "wiki", wiki_site = "commons")
classification("Pinus contorta", db = "wiki", wiki_site = "pedia")
classification("Pinus contorta", db = "wiki", wiki_site = "pedia",
  wiki = "fr")

classification(get_wiki("Malus domestica", "commons"))
classification(get_wiki("Malus domestica", "species"))
classification(c("Pinus contorta", "Malus domestica"), db = "wiki")

# Plug in taxon names
## in this case, we use get_*() fxns internally to first get taxon IDs
classification("Oncorhynchus mykiss", db = "eol")
classification(c("Chironomus riparius", "aaa vva"), db = 'ncbi')
classification(c("Chironomus riparius", "aaa vva"), db = 'ncbi',
  messages=FALSE)
classification(c("Chironomus riparius", "aaa vva"), db = 'itis')
classification(c("Chironomus riparius", "aaa vva"), db = 'itis',
  messages=FALSE)
classification(c("Chironomus riparius", "aaa vva"), db = 'eol')
classification("Alopias vulpinus", db = 'nbn')
classification('Gadus morhua', db = 'worms')
classification('Aquila chrysaetos', db = 'natserv')
classification('Gadus morhua', db = 'natserv')
classification('Pomatomus saltatrix', db = 'natserv')
classification('Aquila chrysaetos', db = 'natserv')
classification(c("Chironomus riparius", "asdfasdfsfdfsd"), db = 'gbif')
classification("Chironomus", db = 'tol')
classification("Poa annua", db = 'tropicos')

# Use methods for get_uid, get_tsn, get_eolid, get_tpsid
classification(get_uid(c("Chironomus riparius", "Puma concolor")))

classification(get_uid(c("Chironomus riparius", "aaa vva")))
classification(get_tsn(c("Chironomus riparius", "aaa vva")))
classification(get_tsn(c("Chironomus riparius", "aaa vva"),
  messages = FALSE))
classification(get_eolid(c("Chironomus riparius", "aaa vva")))
classification(get_tpsid(c("Poa annua", "aaa vva")))
classification(get_gbifid(c("Poa annua", "Bison bison")))

# Pass many ids from class "ids"
(out <- get_ids("Puma concolor", db = c('ncbi','gbif')))
(cl <- classification(out))

# Bind width-wise from class classification_ids
cbind(cl)

# Bind length-wise
rbind(cl)

# Many names to get_ids
(out <- get_ids(c("Puma concolor","Accipiter striatus"),
  db = c('ncbi','itis')))
(cl <- classification(out))
rbind(cl)
## cbind with so many names results in some messy data
cbind(cl)
## so you can turn off return_id
cbind( classification(out, return_id=FALSE) )

(cl_uid <- classification(get_uid(c("Puma concolor",
  "Accipiter striatus")), return_id=FALSE))
rbind(cl_uid)
cbind(cl_uid)
## cbind works a bit odd when there are lots of ranks without names
(cl_uid <- classification(get_uid(c("Puma concolor","Accipiter striatus")),
  return_id=TRUE))
cbind(cl_uid)

(cl_tsn <- classification(get_tsn(c("Puma concolor","Accipiter striatus"))))
rbind(cl_tsn)
cbind(cl_tsn)

(tsns <- get_tsn(c("Puma concolor","Accipiter striatus")))
(cl_tsns <- classification(tsns))
cbind(cl_tsns)

# NBN data
(res <- classification(c("Alopias vulpinus","Pinus sylvestris"),
  db = 'nbn'))
rbind(res)
cbind(res)

# Return taxonomic IDs
## the return_id parameter is logical, and you can turn it on or off.
## It's TRUE by default
classification(c("Alopias vulpinus","Pinus sylvestris"), db = 'ncbi',
  return_id = TRUE)
classification(c("Alopias vulpinus","Pinus sylvestris"), db = 'ncbi',
  return_id = FALSE)

# Use rows parameter to select certain
classification('Poa annua', db = 'tropicos')
classification('Poa annua', db = 'tropicos', rows=1:4)
classification('Poa annua', db = 'tropicos', rows=1)
classification('Poa annua', db = 'tropicos', rows=6)

# Queries of many IDs are processed in batches for NCBI
ids <- c("13083", "2650392", "1547764", "230054", "353934", "656984", 
"271789", "126272", "184644", "73213", "662816", "1161803", "1239353", 
"59420", "665675", "866969", "1091219", "1431218", "1471898", 
"864321", "251768", "2486276", "2068772", "1825808", "2006532", 
"128287", "1195738", "1084683", "1886461", "508296", "377247", 
"1489665", "329325", "219243", "1176946", "339893", "197933", 
"174510", "1704048", "212897", "154842", "1239280", "260135", 
"405735", "1566412", "2083462", "651348", "983204", "165380", 
"2338856", "2068760", "167262", "34229", "1213340", "478939", 
"1933585", "49951", "1277794", "1671089", "1502538", "362355", 
"746473", "242879", "158219", "313664", "2093188", "1541232", 
"584742", "1331091", "147639", "284492", "75642", "1412882", 
"391782", "1406855", "434506", "2053357", "217315", "1444328", 
"329249", "2294004", "84942", "324458", "538247", "69452", "49170", 
"1993845", "261646", "127633", "228146", "1420004", "1629772", 
"577055", "697062", "231660", "648380", "554953", "746496", "2602969")
result <- classification(ids, db = 'ncbi')
}
}
\seealso{
\code{\link[=get_tsn]{get_tsn()}}, \code{\link[=get_uid]{get_uid()}}, \code{\link[=get_eolid]{get_eolid()}},
\code{\link[=get_tpsid]{get_tpsid()}}, \code{\link[=get_gbifid]{get_gbifid()}}, \code{\link[=get_wormsid]{get_wormsid()}}, \code{\link[=get_natservid]{get_natservid()}},
\code{\link[=get_boldid]{get_boldid()}}, \code{\link[=get_wiki]{get_wiki()}}, \code{\link[=get_pow]{get_pow()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iucn_summary.R
\name{iucn_summary}
\alias{iucn_summary}
\title{Get a summary from the IUCN Red List}
\usage{
iucn_summary(x, distr_detail = FALSE, key = NULL, ...)
}
\arguments{
\item{x}{character; Scientific name. Should be cleaned and in the format
\verb{*<Genus> <Species>*}.}

\item{distr_detail}{logical; If \code{TRUE}, the geographic distribution is
returned as a list of vectors corresponding to the different range types:
native, introduced, etc.}

\item{key}{a Redlist API key, get one from
https://apiv3.iucnredlist.org/api/v3/token Required for \code{iucn_summary}.
Defaults to \code{NULL} in case you have your key stored (see \verb{Redlist Authentication} below).}

\item{...}{curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A list (for every species one entry) of data returned by
\code{\link[rredlist:rl_species_latest]{rredlist::rl_species_latest()}}.
}
\description{
Get a summary from the IUCN Red List
(https://www.iucnredlist.org/).
}
\details{
Beware: IUCN functions can give back incorrect data. This isn't our
fault. We do our best to get you the correct data quickly, but sometimes
IUCN gives back the wrong data, and sometimes Global Names gives back the
wrong data. We will fix these as soon as possible. In the meantime, just
make sure that the data you get back is correct.

\code{iucn_summary} has a default method that errors when anything's passed in
that's not \code{character} or \code{iucn} class - a \code{iucn_summary.character} method
for when you pass in taxon names - and a \code{iucn_summary.iucn} method so you
can pass in iucn class objects as output from \code{\link[=get_iucn]{get_iucn()}} or \code{\link[=as.iucn]{as.iucn()}}.
If you already have IUCN IDs, coerce them to \code{iucn} class via \code{as.iucn(..., check = FALSE)}
}
\note{
Not all data types are available for every species
and NA is returned. \code{\link[=iucn_status]{iucn_status()}} is an extractor function to easily
extract status into a vector.
}
\section{Redlist Authentication}{
 \code{iucn_summary} uses the new Redlist API for
searching for a IUCN ID, so we use the \code{\link[rredlist:rl_species]{rredlist::rl_species()}} function internally.
This function requires an API key. Get the key at
https://apiv3.iucnredlist.org/api/v3/token, and pass it to the \code{key}
parameter, or store in your \code{.Renviron} file like
\code{IUCN_REDLIST_KEY=yourkey} or in your \code{.Rprofile} file like
\code{options(iucn_redlist_key="yourkey")}. We strongly encourage you to not
pass the key in the function call but rather store it in one of those two
files. This key will also set you up to use the \pkg{rredlist} package.
}

\examples{
\dontrun{
# if you send a taxon name, an IUCN API key is required
## here, the key is being detected from a .Rprofile file
## or .Renviron file, See "Redlist Authentication" above
iucn_summary("Lutra lutra")

ia <- iucn_summary(c("Panthera uncia", "Lynx lynx"))
ia <- iucn_summary(c("Panthera uncia", "Lynx lynx", "aaa"))
iucn_summary("Muntiacus rooseveltorum/truongsonensis")
iucn_summary(c("Muntiacus rooseveltorum/truongsonensis", "Lynx lynx"))

## get detailed distribution
iac <- iucn_summary(x="Ara chloropterus", distr_detail = TRUE)
iac[[1]]$distr


# If you pass in an IUCN ID, you don't need to pass in a Redlist API Key
# extract status
iucn_status(iac)
}
}
\seealso{
\code{\link[=iucn_status]{iucn_status()}}
}
\author{
Eduard Szoecs, \email{eduardszoecs@gmail.com}

Philippe Marchand, \email{marchand.philippe@gmail.com}

Scott Chamberlain,

Zachary S.L. Foster
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col.R
\name{col-defunct}
\alias{col-defunct}
\alias{as.colid}
\alias{col_children}
\alias{col_classification}
\alias{col_downstream}
\alias{col_search}
\alias{get_colid}
\alias{get_colid_}
\title{COL defunct functions}
\usage{
as.colid(...)

col_children(...)

col_classification(...)

col_downstream(...)

col_search(...)

get_colid(...)

get_colid_(...)
}
\arguments{
\item{...}{ignored}
}
\description{
COL introduced rate limiting recently in 2019 - which has made the API
essentially unusable - COL+ is coming soon and we'll incorporate it here
when it's stable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncbi_children.R
\name{ncbi_children}
\alias{ncbi_children}
\title{Search NCBI for children of a taxon}
\usage{
ncbi_children(
  name = NULL,
  id = NULL,
  start = 0,
  max_return = 1000,
  ancestor = NULL,
  out_type = c("summary", "uid"),
  ambiguous = FALSE,
  key = NULL,
  ...
)
}
\arguments{
\item{name}{(\code{character}) The string to search for. Only exact matches
found the name given will be returned. Not compatible with \code{id}.}

\item{id}{(\code{character}/\code{numeric}) The uid to search for. Not compatible with
\code{name}.}

\item{start}{The first record to return. If omitted, the results are
returned from the first record (start=0).}

\item{max_return}{(\verb{numeric; length=1}) The maximum number of children
to return.}

\item{ancestor}{(\code{character}) The ancestor of the taxon being searched
for. This is useful if there could be more than one taxon with the same
name. Has no effect if \code{id} is used.}

\item{out_type}{(character) Currently either \code{"summary"} or \code{"uid"}:
\itemize{
\item \code{summary} The output is a list of \code{data.frame} with children uid, name,
and rank.
\item \code{uid} A list of character vectors of children uids
}}

\item{ambiguous}{\verb{logical; length 1} If \code{FALSE}, children taxa
with words like "unclassified", "unknown", "uncultured", or "sp." are
removed from the output. NOTE: This option only applies when
\code{out_type= "summary"}.}

\item{key}{(character) NCBI Entrez API key. optional. See Details.}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
The output type depends on the value of the \code{out_type}
parameter. Taxa that cannot be found will result in \code{NA}s and a lack
of children results in an empty data structure.
}
\description{
Search the NCBI Taxonomy database for uids of children of taxa.
Taxa can be referenced by name or uid. Referencing by name is faster

In a few cases, different taxa have the same name (e.g. Satyrium; see
examples). If one of these are searched for then the children of both taxa
will be returned. This can be avoided by using a uid instead of the name or
specifying an ancestor. If an ancestor is provided, only children of both
the taxon and its ancestor are returned. This will only fail if there are
two taxa with the same name and the same specified ancestor.
}
\section{Authentication}{

See \code{\link[=taxize-authentication]{taxize-authentication()}} for help on authentication.
We strongly recommend getting an API key
}

\section{HTTP version}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\section{Rate limits}{

In case you run into errors due to your rate limit being exceeded, see
\code{\link[=taxize_options]{taxize_options()}}, where you can set \code{ncbi_sleep}.
}

\examples{
\dontrun{
ncbi_children(name="Satyrium") #Satyrium is the name of two different genera
ncbi_children(name="Satyrium", ancestor="Eumaeini") # A genus of butterflies
ncbi_children(name="Satyrium", ancestor="Orchidaceae") # A genus of orchids
ncbi_children(id="266948") #"266948" is the uid for the butterfly genus
ncbi_children(id="62858") #"62858" is the uid for the orchid genus

# use curl options
ncbi_children(name="Satyrium", ancestor="Eumaeini", verbose = TRUE)
}
}
\seealso{
\code{\link[=ncbi_get_taxon_summary]{ncbi_get_taxon_summary()}}, \code{\link[=children]{children()}}
}
\author{
Zachary Foster \email{zacharyfoster1989@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gbif_downstream.R
\name{gbif_downstream}
\alias{gbif_downstream}
\title{Retrieve all taxonomic names downstream in hierarchy for GBIF}
\usage{
gbif_downstream(
  id,
  downto,
  intermediate = FALSE,
  limit = 100,
  start = NULL,
  key = NULL,
  ...
)
}
\arguments{
\item{id}{A taxonomic serial number.}

\item{downto}{The taxonomic level you want to go down to. See examples below.
The taxonomic level IS case sensitive, and you do have to spell it
correctly. See \code{data(rank_ref)} for spelling.}

\item{intermediate}{(logical) If TRUE, return a list of length two with
target taxon rank names, with additional list of data.frame's of
intermediate taxonomic groups. Default: FALSE}

\item{limit}{Number of records to return. default: 100. max: 1000. use in
combination with the \code{start} parameter}

\item{start}{Record number to start at. default: 0. use in combination
with the \code{limit} parameter}

\item{key}{Deprecated, see \code{id}}

\item{...}{Further args passed on to \code{\link[=gbif_name_usage]{gbif_name_usage()}}}
}
\value{
data.frame of taxonomic information downstream to family from e.g.,
Order, Class, etc., or if \code{intermediated=TRUE}, list of length two,
with target taxon rank names, and intermediate names.
}
\description{
Retrieve all taxonomic names downstream in hierarchy for GBIF
}
\details{
Sometimes records don't have a \code{canonicalName} entry which is
what we look for. In that case we grab the \code{scientificName} entry.
You can see the type of name colleceted in the column \code{name_type}
}
\examples{
\dontrun{
## the plant class Bangiophyceae
gbif_downstream(id = 198, downto="genus")
gbif_downstream(id = 198, downto="genus", intermediate=TRUE)

# families downstream from the family Strepsiptera (twisted wing parasites)
gbif_downstream(id = 1227, "family")
## here, intermediate leads to the same result as the target
gbif_downstream(id = 1227, "family", intermediate=TRUE)

if (interactive()) {
# Lepidoptera
gbif_downstream(id = 797, "family")

# get species downstream from the genus Ursus
gbif_downstream(id = 2433406, "species")

# get tribes down from the family Apidae
gbif_downstream(id = 7799978, downto="species")
gbif_downstream(id = 7799978, downto="species", intermediate=TRUE)

# names that don't have canonicalname entries for some results
# Myosotis: key 2925668
key <- 2925668
res <- gbif_downstream(key, downto = "species")
res2 <- downstream(key, db = "gbif", downto = "species")

# Pagination
# GBIF limits queries to a maximum of 1000 records per request, so if
# there's more than 1000, use the start parameter
# Piper, taxonKey = 3075433
x1 <- gbif_downstream(id = 3075433, downto = "species", limit=1000)
x2 <- gbif_downstream(id = 3075433, downto = "species", limit=1000,
  start=1000)
x3 <- gbif_downstream(id = 3075433, downto = "species", limit=1000,
  start=2000)
x4 <- gbif_downstream(id = 3075433, downto = "species", limit=1000,
  start=3000)
rbind(x1, x2, x3, x4)
}
}
}
\author{
Scott Chamberlain
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lowest_common.R
\name{lowest_common}
\alias{lowest_common}
\alias{lowest_common.default}
\alias{lowest_common.uid}
\alias{lowest_common.tsn}
\alias{lowest_common.gbifid}
\alias{lowest_common.tolid}
\title{Retrieve the lowest common taxon and rank for a given taxon name or ID}
\usage{
lowest_common(...)

\method{lowest_common}{default}(
  sci_id,
  db = NULL,
  rows = NA,
  class_list = NULL,
  low_rank = NULL,
  x = NULL,
  ...
)

\method{lowest_common}{uid}(sci_id, class_list = NULL, low_rank = NULL, ...)

\method{lowest_common}{tsn}(sci_id, class_list = NULL, low_rank = NULL, ...)

\method{lowest_common}{gbifid}(sci_id, class_list = NULL, low_rank = NULL, ...)

\method{lowest_common}{tolid}(sci_id, class_list = NULL, low_rank = NULL, ...)
}
\arguments{
\item{...}{Other arguments passed to \code{\link[=get_tsn]{get_tsn()}}, \code{\link[=get_uid]{get_uid()}},
\code{\link[=get_gbifid]{get_gbifid()}}, \code{\link[=get_tolid]{get_tolid()}}}

\item{sci_id}{Vector of taxa names (character) or id (character or numeric) to
query.}

\item{db}{character; database to query. either \code{ncbi}, \code{itis}, \code{gbif},
\code{tol}. If using ncbi, we recommend getting an API key;
see \link{taxize-authentication}}

\item{rows}{(numeric) Any number from 1 to infinity. If the default NA,
all rows are considered. Note that this parameter is ignored if you pass in
a taxonomic id of any of the acceptable classes: tsn, gbifid, tolid.
NCBI has a method for this function but rows doesn't work.}

\item{class_list}{(list) A list of classifications, as returned from
\code{\link[=classification]{classification()}}}

\item{low_rank}{(character) taxonomic rank to return, of length 1}

\item{x}{Deprecated, see \code{sci_id}}
}
\value{
NA when no match, or a data.frame with columns
\itemize{
\item name
\item rank
\item id
}
}
\description{
Retrieve the lowest common taxon and rank for a given taxon name or ID
}
\section{Authentication}{

See \link{taxize-authentication} for help on authentication
}

\examples{
\dontrun{
id <- c("9031", "9823", "9606", "9470")
id_class <- classification(id, db = 'ncbi')
lowest_common(id[2:4], db = "ncbi")
lowest_common(id[2:4], db = "ncbi", low_rank = 'class')
lowest_common(id[2:4], db = "ncbi", low_rank = 'family')
lowest_common(id[2:4], class_list = id_class)
lowest_common(id[2:4], class_list = id_class, low_rank = 'class')
lowest_common(id[2:4], class_list = id_class, low_rank = 'family')

# TOL
taxa <- c("Angraecum sesquipedale", "Dracula vampira",
  "Masdevallia coccinea")
(cls <- classification(taxa, db = "tol"))
lowest_common(taxa, db = "tol", class_list = cls)
lowest_common(get_tolid(taxa), class_list = cls)
xx <- get_tolid(taxa)
lowest_common(xx, class_list = cls)

spp <- c("Sus scrofa", "Homo sapiens", "Nycticebus coucang")
lowest_common(spp, db = "ncbi")
lowest_common(get_uid(spp))

lowest_common(spp, db = "itis")
lowest_common(get_tsn(spp))

gbifid <- c("2704179", "3119195")
lowest_common(gbifid, db = "gbif")

spp <- c("Poa annua", "Helianthus annuus")
lowest_common(spp, db = "gbif")
lowest_common(get_gbifid(spp))

cool_orchid <- c("Angraecum sesquipedale", "Dracula vampira",
  "Masdevallia coccinea")
orchid_ncbi <- get_uid(cool_orchid)
orchid_gbif <- get_gbifid(cool_orchid)

cool_orchids2 <- c("Domingoa haematochila", "Gymnadenia conopsea",
  "Masdevallia coccinea")
orchid_itis <- get_tsn(cool_orchids2)

orchid_hier_ncbi <- classification(orchid_ncbi, db = 'ncbi')
orchid_hier_gbif <- classification(orchid_gbif, db = 'gbif')
orchid_hier_itis <- classification(orchid_itis, db = 'itis')

lowest_common(orchid_ncbi, low_rank = 'class')
lowest_common(orchid_ncbi, class_list = orchid_hier_ncbi,
  low_rank = 'class')
lowest_common(orchid_gbif, low_rank = 'class')
lowest_common(orchid_gbif, orchid_hier_gbif, low_rank = 'class')
lowest_common(get_uid(cool_orchid), low_rank = 'class')
lowest_common(get_uid(cool_orchid), low_rank = 'family')

lowest_common(orchid_ncbi, class_list = orchid_hier_ncbi,
  low_rank = 'subfamily')
lowest_common(orchid_gbif, class_list = orchid_hier_gbif,
  low_rank = 'subfamily')

lowest_common(orchid_itis, class_list = orchid_hier_itis,
  low_rank = 'class')

## Pass in sci. names
nms <- c("Angraecum sesquipedale", "Dracula vampira", "Masdevallia coccinea")
lowest_common(x = nms, db = "ncbi")
lowest_common(x = nms, db = "gbif")
}
}
\author{
Jimmy O'Donnell \email{jodonnellbio@gmail.com}
Scott Chamberlain
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ion.R
\name{ion}
\alias{ion}
\title{ION - Index to Organism Names}
\usage{
ion(x, ...)
}
\arguments{
\item{x}{An LSID number. Required.}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A data.frame
}
\description{
ION - Index to Organism Names
}
\examples{
\dontrun{
ion(155166)
ion(298678)
ion(4796748) # ursus americanus
ion(1280626) # puma concolor
}
}
\references{
http://www.organismnames.com
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/itis_getrecord.R
\name{itis_getrecord}
\alias{itis_getrecord}
\title{Get full ITIS record for one or more ITIS TSN's or lsid's.}
\usage{
itis_getrecord(values, by = "tsn", ...)
}
\arguments{
\item{values}{(character) One or more TSN's (taxonomic serial number) or lsid's for a
taxonomic group}

\item{by}{(character) By "tsn" (default) or "lsid"}

\item{...}{Further arguments passed on to \link[ritis:full_record]{ritis::full_record}}
}
\description{
Get full ITIS record for one or more ITIS TSN's or lsid's.
}
\details{
You can only enter values in tsn parameter or lsid, not both.
}
\examples{
\dontrun{
# by TSN
itis_getrecord(202385)
itis_getrecord(c(202385,70340))

# by lsid
itis_getrecord("urn:lsid:itis.gov:itis_tsn:202385", "lsid")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gna_parse.R
\name{gna_parse}
\alias{gna_parse}
\title{Parse scientific names using Global Names Parser}
\usage{
gna_parse(names, ...)
}
\arguments{
\item{names}{A vector of length 1 or more taxonomic names}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A data.frame with results, the submitted names, and the
parsed names with additional information.
}
\description{
Parse scientific names using Global Names Parser
}
\examples{
\dontrun{
gna_parse("Cyanistes caeruleus")
gna_parse("Plantago minor")
gna_parse("Plantago minor minor")
gna_parse(c("Plantago minor minor","Helianthus annuus texanus"))

# if > 20 names, uses an HTTP POST request
x <- names_list("species", size = 30)
gna_parse(x)

# pass on curl options
gna_parse("Cyanistes caeruleus", verbose = TRUE)
}
}
\references{
http://gni.globalnames.org/
}
\seealso{
\code{\link[=gbif_parse]{gbif_parse()}}, \code{\link[=gni_parse]{gni_parse()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ping.R
\name{ping}
\alias{ping}
\alias{col_ping}
\alias{eol_ping}
\alias{itis_ping}
\alias{ncbi_ping}
\alias{tropicos_ping}
\alias{nbn_ping}
\alias{gbif_ping}
\alias{bold_ping}
\alias{ipni_ping}
\alias{vascan_ping}
\alias{fg_ping}
\title{Ping an API used in taxize to see if it's working.}
\usage{
col_ping(what = "status", ...)

eol_ping(what = "status", ...)

itis_ping(what = "status", ...)

ncbi_ping(what = "status", key = NULL, ...)

tropicos_ping(what = "status", ...)

nbn_ping(what = "status", ...)

gbif_ping(what = "status", ...)

bold_ping(what = "status", ...)

ipni_ping(what = "status", ...)

vascan_ping(what = "status", ...)

fg_ping(what = "status", ...)
}
\arguments{
\item{what}{(character) One of status (default), content, or an HTTP status
code. If status, we just check that the HTTP status code is 200, or similar
signifying the service is up. If content, we do a simple, quick check to
determine if returned content matches what's expected. If an HTTP status
code, it must match an appropriate code. See \code{\link[=status_codes]{status_codes()}}.}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}

\item{key}{(character) NCBI Entrez API key. optional. See \code{\link[=get_uid]{get_uid()}}}
}
\value{
A logical, TRUE or FALSE
}
\description{
Ping an API used in taxize to see if it's working.
}
\details{
For ITIS, see \link[ritis:description]{ritis::description}, which provides number of
scientific and common names in a character string.
}
\section{HTTP version for NCBI requests}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\examples{
\dontrun{
col_ping()
col_ping("content")
col_ping(200)
col_ping("200")
col_ping(204)

itis_ping()
eol_ping()
ncbi_ping()
tropicos_ping()
nbn_ping()

gbif_ping()
gbif_ping(200)

bold_ping()
bold_ping(200)
bold_ping("content")

ipni_ping()
ipni_ping(200)
ipni_ping("content")

vascan_ping()
vascan_ping(200)
vascan_ping("content")

# curl options
vascan_ping(verbose = TRUE)
eol_ping(500, verbose = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ubio_search.R
\name{ubio_search}
\alias{ubio_search}
\title{This function will return NameBankIDs that match given search terms}
\usage{
ubio_search(...)
}
\arguments{
\item{...}{Parameters, ignored}
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_refs.R
\name{tp_namereferences}
\alias{tp_namereferences}
\title{Return all reference records for for a taxon name with a given id.}
\usage{
tp_namereferences(...)
}
\arguments{
\item{...}{ignored}
}
\description{
Function name changed to tp_refs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncbi_downstream.R
\name{ncbi_downstream}
\alias{ncbi_downstream}
\title{Retrieve all taxa names downstream in hierarchy for NCBI}
\usage{
ncbi_downstream(id, downto, intermediate = FALSE, ...)
}
\arguments{
\item{id}{(numeric/integer) An NCBI taxonomic identifier}

\item{downto}{The taxonomic level you want to go down to. See examples below.
The taxonomic level IS case sensitive, and you do have to spell it
correctly. See \code{data(rank_ref)} for spelling.}

\item{intermediate}{(logical) If \code{TRUE}, return a list of length two
with target taxon rank names, with additional list of data.frame's of
intermediate taxonomic groups. Default: \code{FALSE}}

\item{...}{Further args passed on to \code{\link[=ncbi_children]{ncbi_children()}}}
}
\value{
Data.frame of taxonomic information downstream to family from e.g.,
Order, Class, etc., or if \code{intermediate=TRUE}, list of length two,
with target taxon rank names, and intermediate names.
}
\description{
Retrieve all taxa names downstream in hierarchy for NCBI
}
\section{No Rank}{

A sticky point with NCBI is that they can have designation for taxonomic
rank of "No Rank". So we have no way of programatically knowing what to
do with that taxon. Of course one can manually look at a name and perhaps
know what it is, or look it up on the web - but we can't do anything
programatically. So, no rank things will sometimes be missing.
}

\section{Authentication}{

See \code{\link[=taxize-authentication]{taxize-authentication()}} for help on authentication.
We strongly recommend getting an API key
}

\examples{
\dontrun{
## genus Apis
ncbi_downstream(id = 7459, downto="species")

## get intermediate taxa as a separate object
ncbi_downstream(id = 7459, downto="species", intermediate = TRUE)

## Lepidoptera
ncbi_downstream(id = 7088, downto="superfamily")

## families in the ferns (Moniliformopses)
(id <- get_uid("Moniliformopses"))
ncbi_downstream(id = id, downto = "order")
}
}
\author{
Scott Chamberlain
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/limited_print.R
\name{limited_print}
\alias{limited_print}
\title{Print a subset of a character vector}
\usage{
limited_print(
  chars,
  prefix = "",
  max_chars = getOption("width") - nchar(prefix) - 5,
  type = "message"
)
}
\arguments{
\item{chars}{(\code{character}) What to print.}

\item{prefix}{(\code{character} of length 1) What to print before
\code{chars}, on the same line.}

\item{max_chars}{(\code{numeric} of length 1) The maximum number of
characters to print.}

\item{type}{(\code{"error"}, \code{"warning"}, \code{"message"}, \code{"cat"}, \code{"print"}, \code{"silent"})}
}
\value{
\code{NULL}
}
\description{
Prints the start and end values for a character vector. The number of values
printed depend on the width of the screen by default.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_iucn.R
\name{get_iucn_data}
\alias{get_iucn_data}
\title{Get a IUCN Redlist taxon data}
\usage{
get_iucn_data(names_or_ids, messages = TRUE, key = NULL, latest = FALSE, ...)
}
\arguments{
\item{latest}{If \code{TRUE}, latest use \code{\link[rredlist:rl_species_latest]{rredlist::rl_species_latest()}} instead of \code{\link[rredlist:rl_species]{rredlist::rl_species()}}}
}
\description{
Used to get IUCN data for other functions to use.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ubio_classification.R
\name{ubio_classification}
\alias{ubio_classification}
\title{uBio classification}
\usage{
ubio_classification(...)
}
\arguments{
\item{...}{Parameters, ignored}
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rankagg.R
\name{rankagg}
\alias{rankagg}
\title{Aggregate data by given taxonomic rank}
\usage{
rankagg(data = NULL, datacol = NULL, rank = NULL, fxn = "sum")
}
\arguments{
\item{data}{A data.frame. Column headers must have capitalized ranks (e.g.,
Genus, Tribe, etc.) (data.frame)}

\item{datacol}{The data column (character)}

\item{rank}{Taxonomic rank to aggregate by (character)}

\item{fxn}{Arithmetic function or vector or functions (character)}
}
\description{
Aggregate data by given taxonomic rank
}
\examples{
if (require(vegan)) {
data(dune.taxon, dune, package='vegan')
dat <- dune.taxon
dat$abundance <- colSums(dune)
rankagg(data=dat, datacol="abundance", rank="Genus")
rankagg(data=dat, "abundance", rank="Family")
rankagg(data=dat, "abundance", rank="Genus", fxn="mean")
rankagg(data=dat, "abundance", rank="Subclass")
rankagg(data=dat, "abundance", rank="Subclass", fxn="sd")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxon_state.R
\name{taxon_state}
\alias{taxon_state}
\title{Keep track of queries in \verb{get_*} functions}
\description{
Keep track of queries in \verb{get_*} functions

Keep track of queries in \verb{get_*} functions
}
\details{
This object lives inside each \verb{get_*} function call, maintaining
results as they are accumulated.
}
\examples{
\dontrun{
if (interactive()) {
ts <- taxon_state$new()
taxon_last()
ts
res <- list(
  id = 123456,
  att = "found",
  multiple = FALSE,
  direct = FALSE,
  class = "tsn"
)
ts$add(query = "Quercus robur", result = res)
ts
ts$get(query = "Quercus robur")
ts$count
ts$remove(query = "Quercus robur")
ts
ts$count

res2 <- list(
  id = 3430834535,
  att = "found",
  multiple = FALSE,
  direct = FALSE,
  class = "gbifid"
)
ts$add(query = "Poa annua", result = res2)
res3 <- list(
  id = 1223424,
  att = "found",
  multiple = FALSE,
  direct = FALSE,
  class = "uid"
)
ts$add(query = "Puma concolor", result = res3)
ts
ts$count
ts$get("Puma concolor")
ts$get()

# cleanup
ts$purge()
ts$count
}
}
}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{initialized}}{(time) time job started}

\item{\code{finalized}}{(time) time job finished}

\item{\code{class}}{(character) a class name (e.g., "gbif")}

\item{\code{names}}{(character) one or more taxon names}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{count}}{(integer) count number of records}

\item{\code{exit}}{record date/time function exited}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-taxon_state-new}{\code{taxon_state$new()}}
\item \href{#method-taxon_state-print}{\code{taxon_state$print()}}
\item \href{#method-taxon_state-add}{\code{taxon_state$add()}}
\item \href{#method-taxon_state-get}{\code{taxon_state$get()}}
\item \href{#method-taxon_state-remove}{\code{taxon_state$remove()}}
\item \href{#method-taxon_state-purge}{\code{taxon_state$purge()}}
\item \href{#method-taxon_state-taxa_remaining}{\code{taxon_state$taxa_remaining()}}
\item \href{#method-taxon_state-taxa_completed}{\code{taxon_state$taxa_completed()}}
\item \href{#method-taxon_state-clone}{\code{taxon_state$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-new"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{taxon_state} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$new(class, names)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{class}}{(character) a class name (e.g., "gbif")}

\item{\code{names}}{(character) one or more taxon names}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{taxon_state} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-print"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-print}{}}}
\subsection{Method \code{print()}}{
print method for the \code{taxon_state} class
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$print(x, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{self}

\item{\code{...}}{ignored}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-add"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-add}{}}}
\subsection{Method \code{add()}}{
add a record with it's result; duplicates allowed
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$add(query, result)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{query}}{(character), a taxon name}

\item{\code{result}}{(list) a named list}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
nothing returned; sets only
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-get"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-get}{}}}
\subsection{Method \code{get()}}{
get all records matching 'query'
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$get(query = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{query}}{(character), a taxon name}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
a named list, with slots for the taxon id, and other attributes,
named by the taxon name
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-remove"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-remove}{}}}
\subsection{Method \code{remove()}}{
remove's all records matching 'query'
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$remove(query)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{query}}{(character), a taxon name}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
nothing, removes records matching query
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-purge"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-purge}{}}}
\subsection{Method \code{purge()}}{
removes all records
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$purge()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
nothing returned; sets only
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-taxa_remaining"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-taxa_remaining}{}}}
\subsection{Method \code{taxa_remaining()}}{
get remaining taxa
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$taxa_remaining()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
sorted taxon names
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-taxa_completed"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-taxa_completed}{}}}
\subsection{Method \code{taxa_completed()}}{
get completed taxa
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$taxa_completed()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
sorted taxon names
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-taxon_state-clone"></a>}}
\if{latex}{\out{\hypertarget{method-taxon_state-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{taxon_state$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col_search.R
\name{col_search}
\alias{col_search}
\title{Search Catalogue of Life for taxonomic IDs}
\usage{
col_search(name = NULL, id = NULL, start = NULL, checklist = NULL,
  response = "terse", ...)
}
\arguments{
\item{name}{The string to search for. Only exact matches found the name given
   will be returned, unless one or wildcards are included in the search
   string. An * (asterisk) character denotes a wildcard; a % (percentage)
character may also be used. The name must be at least 3 characters long,
not counting wildcard characters.}

\item{id}{The record ID of the specific record to return (only for scientific
names of species or infraspecific taxa)}

\item{start}{The first record to return. If omitted, the results are returned
from the first record (start=0). This is useful if the total number of
results is larger than the maximum number of results returned by a single
Web service query (currently the maximum number of results returned by a
single query is 500 for terse queries and 50 for full queries).}

\item{checklist}{The year of the checklist to query, if you want a specific
year's checklist instead of the lastest as default (numeric).}

\item{response}{(character) one of "terse" or "full"}

\item{...}{Curl options passed on to \code{\link[httr]{GET}}}
}
\value{
A list of data.frame's.
}
\description{
Search Catalogue of Life for taxonomic IDs
}
\details{
You must provide one of name or id. The other parameters (format
		and start) are optional.
}
\examples{
\dontrun{
# A basic example
col_search(name="Apis")
col_search(name="Agapostemon")
col_search(name="Poa")

# Get full response, i.e., more data
col_search(name="Apis", response="full")
col_search(name="Poa", response="full")

# Many names
col_search(name=c("Apis","Puma concolor"))
col_search(name=c("Apis","Puma concolor"), response = "full")

# An example where there is no data
col_search(id = "36c623ad9e3da39c2e978fa3576ad415")
col_search(id = "36c623ad9e3da39c2e978fa3576ad415", response = "full")
col_search(id = "787ce23969f5188c2467126d9a545be1")
col_search(id = "787ce23969f5188c2467126d9a545be1", response = "full")
col_search(id = c("36c623ad9e3da39c2e978fa3576ad415", "787ce23969f5188c2467126d9a545be1"))
## a synonym
col_search(id = "f726bdaa5924cabf8581f99889de51fc")
col_search(id = "f726bdaa5924cabf8581f99889de51fc", response = "full")
}
}
\references{
\url{http://webservice.catalogueoflife.org/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncbi_search.R
\name{get_genes_avail}
\alias{get_genes_avail}
\title{Retrieve gene sequences from NCBI by accession number.}
\usage{
get_genes_avail(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_tolid.R
\name{get_tolid}
\alias{get_tolid}
\alias{as.tolid}
\alias{as.tolid.tolid}
\alias{as.tolid.character}
\alias{as.tolid.list}
\alias{as.tolid.numeric}
\alias{as.tolid.data.frame}
\alias{as.data.frame.tolid}
\alias{get_tolid_}
\title{Get the OTT id for a search term}
\usage{
get_tolid(sci, ask = TRUE, messages = TRUE, rows = NA, sciname = NULL, ...)

as.tolid(x, check = TRUE)

\method{as.tolid}{tolid}(x, check = TRUE)

\method{as.tolid}{character}(x, check = TRUE)

\method{as.tolid}{list}(x, check = TRUE)

\method{as.tolid}{numeric}(x, check = TRUE)

\method{as.tolid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{tolid}(x, ...)

get_tolid_(sci, messages = TRUE, rows = NA, sciname = NULL)
}
\arguments{
\item{sci}{character; one or more scientific names. Or, a \code{taxon_state}
object (see \link{taxon-state})}

\item{ask}{logical; should \code{get_tolid} be run in interactive mode?
If \code{TRUE} and more than one TOL is found for the species, the user is
asked for input. If \code{FALSE} NA is returned for multiple matches.}

\item{messages}{logical; should progress be printed?}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this function still only gives back a tol
class object with one to many identifiers. See \code{\link[=get_tolid_]{get_tolid_()}}
to get back all, or a subset, of the raw data that you are presented during
the ask process.}

\item{sciname}{Deprecated, see \code{sci}}

\item{...}{Ignored}

\item{x}{Input to \code{as.tolid}}

\item{check}{logical; Check if ID matches any existing on the DB, only
used in \code{\link[=as.tolid]{as.tolid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Retrieve the Open Tree of Life Taxonomy (OTT) id of a taxon from
OpenTreeOfLife
}
\examples{
\dontrun{
get_tolid(sci = "Quercus douglasii")
get_tolid(sci = "Chironomus riparius")
get_tolid(c("Chironomus riparius","Quercus douglasii"))
splist <- c("annona cherimola", 'annona muricata', "quercus robur",
		"shorea robusta", "pandanus patina", "oryza sativa", "durio zibethinus")
get_tolid(splist, messages=FALSE)

# specify rows to limit choices available
get_tolid('Arni')
get_tolid('Arni', rows=1)
get_tolid('Arni', rows=1:2)

# When not found
get_tolid("howdy")
get_tolid(c("Chironomus riparius", "howdy"))

# Convert a tol without class information to a tol class
as.tolid(get_tolid("Quercus douglasii")) # already a tol, returns the same
as.tolid(get_tolid(c("Chironomus riparius","Pinus contorta"))) # same
as.tolid(5907893) # numeric
as.tolid(c(3930798,515712,872577)) # numeric vector, length > 1
as.tolid("3930798") # character
as.tolid(c("3930798","515712","872577")) # character vector, length > 1
as.tolid(list("3930798","515712","872577")) # list, either numeric or character
## dont check, much faster
as.tolid("3930798", check=FALSE)
as.tolid(3930798, check=FALSE)
as.tolid(c("3930798","515712","872577"), check=FALSE)
as.tolid(list("3930798","515712","872577"), check=FALSE)

(out <- as.tolid(c(3930798,515712,872577)))
data.frame(out)
as.tolid( data.frame(out) )

# Get all data back
get_tolid_("Arni")
get_tolid_("Arni", rows=1)
get_tolid_("Arni", rows=1:2)
get_tolid_(c("asdfadfasd","Pinus contorta"))
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iplant_resolve.R
\name{iplant_resolve}
\alias{iplant_resolve}
\title{iPlant name resolution}
\usage{
iplant_resolve(sci, retrieve = "all", query = NULL, ...)
}
\arguments{
\item{sci}{Vector of one or more taxonomic names (no common names)}

\item{retrieve}{Specifies whether to retrieve all matches for the
names submitted. One of 'best' (retrieves only the single best match
for each name submitted) or 'all' (retrieves all matches)}

\item{query}{Deprecated, see \code{sci}}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A data.frame
}
\description{
iPlant name resolution
}
\examples{
\dontrun{
iplant_resolve(sci=c("Helianthus annuus", "Homo sapiens"))
iplant_resolve("Helianthusss")
iplant_resolve("Pooa")
iplant_resolve("Helianthusss", verbose = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bold_search.R
\name{bold_search}
\alias{bold_search}
\title{Search Barcode of Life for taxonomic IDs}
\usage{
bold_search(
  sci = NULL,
  id = NULL,
  fuzzy = FALSE,
  dataTypes = "basic",
  includeTree = FALSE,
  response = FALSE,
  name = NULL,
  ...
)
}
\arguments{
\item{sci}{(character) One or more scientific names.}

\item{id}{(integer) One or more BOLD taxonomic identifiers.}

\item{fuzzy}{(logical) Whether to use fuzzy search or not (default: \code{FALSE}).
Only used if \code{name} passed.}

\item{dataTypes}{(character) Specifies the datatypes that will be returned.
See Details for options. This variable is ignored if \code{name} parameter is passed,
but is used if the \code{id} parameter is passed.}

\item{includeTree}{(logical) If TRUE (default: FALSE), returns a list containing
information for parent taxa as well as the specified taxon. Only used if \code{id}
passed.}

\item{response}{(logical) Note that response is the object that returns from the
curl call, useful for debugging, and getting detailed info on the API call.}

\item{name}{Deprecated, see \code{sci}}

\item{...}{named curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
A list of data.frame's.
}
\description{
Search Barcode of Life for taxonomic IDs
}
\details{
You must provide one of \code{sci} or \code{id} to this function. The other
parameters are optional. Note that when passing in \code{sci}, \code{fuzzy} can be used
as well, while if \code{id} is passed, then \code{fuzzy} is ignored, and \code{dataTypes}
\code{includeTree} can be used.

Options for \code{dataTypes} parameter:
\itemize{
\item all returns all data
\item basic returns basic taxon information
\item images returns specimen image. Includes copyright information, image URL,
image metadata.
\item stats Returns specimen and sequence statistics. Includes public species
count, public BIN count, public marker counts, public record count,
specimen count, sequenced specimen count, barcode specimen count, species
count, barcode species count.
\item geo Returns collection site information. Includes country, collection
site map.
\item sequencinglabs Returns sequencing labs. Includes lab name, record count.
\item depository Returns specimen depositories. Includes depository name,
record count.
\item thirdparty Returns information from third parties. Includes wikipedia
summary, wikipedia URL, GBIF map.
}
}
\examples{
\dontrun{
# A basic example
bold_search(sci="Apis")
bold_search(sci="Agapostemon")
bold_search(sci="Poa")

# Fuzzy search
head(bold_search(sci="Po", fuzzy=TRUE))
head(bold_search(sci="Aga", fuzzy=TRUE))

# Many names
bold_search(sci=c("Apis","Puma concolor"))
nms <- names_list('species')
bold_search(sci=nms)

# Searching by ID - dataTypes can be used, and includeTree can be used
bold_search(id=88899)
bold_search(id=88899, dataTypes="stats")
bold_search(id=88899, dataTypes="geo")
bold_search(id=88899, dataTypes="basic")
bold_search(id=88899, includeTree=TRUE)
}
}
\references{
http://www.boldsystems.org/index.php/resources/api
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pow_search.R
\name{pow_synonyms}
\alias{pow_synonyms}
\title{Lookup synonyms in Kew's Plants of the World}
\usage{
pow_synonyms(id, ...)
}
\arguments{
\item{id}{(character) taxon id. required}

\item{...}{Further args passed on to \code{\link[=pow_lookup]{pow_lookup()}}}
}
\description{
Lookup synonyms in Kew's Plants of the World
}
\examples{
\dontrun{
pow_synonyms(id = 'urn:lsid:ipni.org:names:320035-2')
pow_synonyms(id = 'urn:lsid:ipni.org:names:358881-1')
pow_synonyms(id = 'urn:lsid:ipni.org:names:359855-1')
}
}
\seealso{
Other pow: 
\code{\link{get_pow}()},
\code{\link{pow_lookup}()},
\code{\link{pow_search}()}
}
\concept{pow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{rank_ref_zoo}
\alias{rank_ref_zoo}
\title{Lookup-table for IDs of taxonomic ranks (WoRMS)}
\description{
Same as \code{rank_ref} but specifically for WoRMS, where section/subsection
ranks are put between family/order rather than between species/genus.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scrapenames.r
\name{scrapenames}
\alias{scrapenames}
\title{Find taxon names using Global Names Recognition and Discovery}
\usage{
scrapenames(
  url = NULL,
  text = NULL,
  format = "csv",
  bytes_offset = FALSE,
  return_content = FALSE,
  unique_names = TRUE,
  ambiguous_names = FALSE,
  no_bayes = FALSE,
  odds_details = FALSE,
  language = "detect",
  words_around = 0,
  verification = TRUE,
  sources = NULL,
  all_matches = FALSE,
  ...,
  file = NULL,
  unique = NULL,
  engine = NULL,
  detect_language = NULL,
  data_source_ids = NULL
)
}
\arguments{
\item{url}{(character) If text parameter is empty, and \code{url} is given,
GNfinder will process the URL and will find names in the content of its
body.}

\item{text}{(character) Contains the text which will be checked for
scientific names. If this parameter is not empty, the \code{url} parameter is
ignored.}

\item{format}{(character) Sets the output format. It can be set to: \code{"csv"}
(the default), \code{"tsv"}, or \code{"json"}.}

\item{bytes_offset}{(logical) This changes how the position of a detected
name in text is calculated. Normally a name's start and end positions are
given as the number of UTF-8 characters from the beginning of the text. If
this is \code{TRUE}, the start and end offsets are recalculated in the number of
bytes.}

\item{return_content}{(logical) If this is \code{TRUE}, the text used for the name
detection is returned back. This is especially useful if the input was not
a plain UTF-8 text and had to be prepared for name-finding. Then the
returned content can be used together with start and end fields of detected
name-strings to locate the strings in the text.}

\item{unique_names}{(logical) If this is \code{TRUE}, the output returns a list of
unique names, instead of a list of all name occurrences. Unique list of
names does not provide position information of a name in the text.}

\item{ambiguous_names}{(logical) If this is \code{TRUE}, strings which are
simultaneously scientific names and "normal" words are not filtered out
from the results. For example, generic names like America, Cancer,
Cafeteria will be returned in the results.}

\item{no_bayes}{(logical) If this is \code{TRUE}, only heuristic algorithms are
used for name detection.}

\item{odds_details}{(logical) If \code{TRUE}, the result will contain odds of all
features used for calculation of NaiveBayes odds. Odds describe probability
of a name to be 'real'. The higher the odds, the higher the probability
that a detected name is not a false positive. Odds are calculated by
multiplication of the odds of separate features. Odds details explain how
the final odds value is calculated.}

\item{language}{(character) The language of the text. Language value is used
for calculation of Bayesian odds. If this parameter is not given, \code{"eng"}
is used by default. Currently only English and German languages are
supported. Valid values are: \code{"eng"}, \code{"deu"}, and \code{"detect"}.}

\item{words_around}{(integer) Allows to see the context surrounding a
name-string. This sets the number of words located immediately before or
after a detected name. These words are then returned in the output. Default
is 0, maximum value is 5.}

\item{verification}{(character) When this \code{TRUE}, there is an additional
verification step for detected names. This step requires internet
connection and uses https://verifier.globalnames.org/api/v1 for
verification queries.}

\item{sources}{Pipe separated list of data source ids to resolve found names
against. See list of Data Sources
http://resolver.globalnames.org/data_sources}

\item{all_matches}{When this option is true all found results are returned,
not only the bestResults. The bestResult field in this case is null, and
results field should contain found results of the matches.}

\item{...}{Further args passed to \link[crul:verb-GET]{crul::verb-GET}}

\item{file}{Defunct. If you feel this is important functionality submit an
issue at "https://github.com/ropensci/taxize"}

\item{unique}{Defunct. See the \code{unique_names} option.}

\item{engine}{Defunct. The API used no longer supports this option.}

\item{detect_language}{Defunct. See the \code{language} option.}

\item{data_source_ids}{Defunct. See the \code{sources} option.}
}
\value{
A \code{\link[tibble:tibble]{tibble::tibble()}} or list representing parsed JSON output
depending on the value of the \code{format} option.
}
\description{
Uses the Global Names Recognition and Discovery service, see
http://gnrd.globalnames.org/

NOTE: This function sometimes gives data back and sometimes not. The API
that this function is using is extremely buggy.
}
\examples{
\dontrun{
# Get data from a website using its URL
scrapenames('https://en.wikipedia.org/wiki/Spider')
scrapenames('https://en.wikipedia.org/wiki/Animal')
scrapenames('https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0095068')
scrapenames('https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0080498')

scrapenames(url = 'https://en.wikipedia.org/wiki/Spider', source=c(1, 169))

# Get data from text string
scrapenames(text='A spider named Pardosa moesta Banks, 1892')

# return OCR content
scrapenames(text='A spider named Pardosa moesta Banks, 1892',
            return_content = TRUE, format = 'json')
}
}
\author{
Scott Chamberlain, Zachary Foster
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nbn_search.R
\name{nbn_search}
\alias{nbn_search}
\title{Search UK National Biodiversity Network}
\usage{
nbn_search(
  sci_com,
  fq = NULL,
  order = NULL,
  sort = NULL,
  start = 0,
  rows = 25,
  facets = NULL,
  q = NULL,
  ...
)
}
\arguments{
\item{sci_com}{(character) The query terms(s), a scientific or common name}

\item{fq}{(character) Filters to be applied to the original query. These
are additional params of the form fq=INDEXEDFIELD:VALUE e.g.
fq=rank:kingdom. See https://species-ws.nbnatlas.org/indexFields for all
the fields that are queryable.}

\item{order}{(character) Supports "asc" or "desc"}

\item{sort}{(character) The indexed field to sort by}

\item{start}{(integer) Record offset, to enable paging}

\item{rows}{(integer) Number of records to return}

\item{facets}{(list) Comma separated list of the fields to create facets
on e.g. facets=basis_of_record.}

\item{q}{Deprecated, see \code{sci}}

\item{...}{Further args passed on to \link[crul:HttpClient]{crul::HttpClient}.}
}
\value{
a list with slots for metadata (\code{meta}) with list of response
attributes, and data (\code{data}) with a data.frame of results
}
\description{
Search UK National Biodiversity Network
}
\examples{
\dontrun{
x <- nbn_search(sci_com = "Vulpes")
x$meta$totalRecords
x$meta$pageSize
x$meta$urlParameters
x$meta$queryTitle
head(x$data)

nbn_search(sci_com = "blackbird", start = 4)

# debug curl stuff
nbn_search(sci_com = "blackbird", verbose = TRUE)
}
}
\references{
https://api.nbnatlas.org/
}
\seealso{
Other nbn: 
\code{\link{get_nbnid}()},
\code{\link{nbn_classification}()},
\code{\link{nbn_synonyms}()}
}
\author{
Scott Chamberlain,
}
\concept{nbn}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/upstream.R
\name{upstream}
\alias{upstream}
\alias{upstream.default}
\alias{upstream.tsn}
\alias{upstream.ids}
\title{Retrieve the upstream taxa for a given taxon name or ID.}
\usage{
upstream(...)

\method{upstream}{default}(sci_id, db = NULL, upto = NULL, rows = NA, x = NULL, ...)

\method{upstream}{tsn}(sci_id, db = NULL, upto = NULL, ...)

\method{upstream}{ids}(sci_id, db = NULL, upto = NULL, ...)
}
\arguments{
\item{...}{Further args passed on to \code{\link[=itis_downstream]{itis_downstream()}}}

\item{sci_id}{Vector of taxa names (character) or IDs (character or numeric) to
query.}

\item{db}{character; database to query. One or both of \code{itis}. Note
that each taxonomic data source has their own identifiers, so that if you
provide the wrong \code{db} value for the identifier you could get a result, but
it will likely be wrong (not what you were expecting).}

\item{upto}{What taxonomic rank to go down to. One of: 'superkingdom',
'kingdom', 'subkingdom','infrakingdom','phylum','division','subphylum',
'subdivision','infradivision', 'superclass','class','subclass','infraclass',
'superorder','order','suborder','infraorder','superfamily','family',
'subfamily','tribe','subtribe','genus','subgenus', 'section','subsection',
'species','subspecies','variety','form','subvariety','race', 'stirp',
'morph','aberration','subform', or 'unspecified'}

\item{rows}{(numeric) Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this parameter is ignored if you pass in a
taxonomic id of any of the acceptable classes: tsn.}

\item{x}{Deprecated, see \code{sci_id}}
}
\value{
A named list of data.frames with the upstream names of every
supplied taxa. You get an NA if there was no match in the database.
}
\description{
This function uses a while loop to continually collect taxa up to the
taxonomic rank that you specify in the \code{upto} parameter. You can get data
from ITIS (itis) only currently. There is no method exposed by itis
for getting taxa at a specific taxonomic rank, so we do it ourselves inside
the function.
}
\examples{
\dontrun{
upstream('Pinus contorta', db = 'itis', upto = 'genus')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/key_helpers.R
\name{key_helpers}
\alias{key_helpers}
\alias{use_tropicos}
\alias{use_entrez}
\alias{use_iucn}
\title{Helpers to set up authentication for the different providers.}
\usage{
use_tropicos()

use_entrez()

use_iucn()
}
\description{
Sets up authentication to diverse providers by providing
the user a detailed prompt.
}
\details{
Key helpers
}
\section{\code{use_tropicos()}}{

Browses to Tropicos API key request URL and provides instruction on how to
store the key. After filling the form you will get the key soon, but
not immediately.
}

\section{\code{use_entrez()}}{

Browse NCBI Entrez to help make an API key request and provides instruction
on how to store the key. There's no direct URL to request a key, one first
needs to log in or register and then to generate a key from one's account.

Note that NCBI Entrez doesn't require that you use an API key,
but you should get higher rate limit with a key, so do get one.
}

\section{\code{use_iucn()}}{

Browse IUCN Red List API key request URL and provides instruction on how
to store the key. This function wraps \code{\link[rredlist:rl_use_iucn]{rredlist::rl_use_iucn()}}
from the \code{rredlist} package. After filling the form you will get
the key soon, but not immediately.
}

\seealso{
\link{taxize-authentication}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bold_search.R
\name{bold_children}
\alias{bold_children}
\title{Barcode of Life taxonomic children}
\usage{
bold_children(id, ...)
}
\arguments{
\item{id}{(integer) A BOLD taxonomic identifier. length=1. required}

\item{...}{named curl options passed on to \link[crul:verb-GET]{crul::verb-GET}
debugging}
}
\value{
list of data.frame's
}
\description{
BEWARE: this function scrapes data from the BOLD website, so may
be unstable. That is, one day it may work, and the next it may fail.
Open an issue if you encounter an error:
https://github.com/ropensci/taxize/issues
}
\examples{
\dontrun{
# Osmia (genus): 253 children
bold_children(id = 4940)
# Momotus (genus): 3 children
bold_children(id = 88899)
# Momotus aequatorialis (species): no children
bold_children(id = 115130)
# Osmia sp1 (species): no children
bold_children(id = 293378)
# Arthropoda (phylum): 27 children
bold_children(id = 82)
# Psocodea (order): 51 children
bold_children(id = 737139)
# Megachilinae (subfamily): 2 groups (tribes: 3, genera: 60)
bold_children(id = 4962)
# Stelis (species): 78 taxa
bold_children(id = 4952)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eol_invasive.R
\name{eol_invasive}
\alias{eol_invasive}
\title{Search for presence of taxonomic names in EOL invasive species databases.}
\usage{
eol_invasive(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col_downstream.R
\name{col_downstream}
\alias{col_downstream}
\title{Use Catalogue of Life to get downstream taxa to a given taxonomic level.}
\usage{
col_downstream(name = NULL, id = NULL, downto, format = NULL,
  start = NULL, checklist = NULL, verbose = TRUE, intermediate = FALSE,
  ...)
}
\arguments{
\item{name}{The string to search for. Only exact matches found the name given
 will be returned, unless one or wildcards are included in the search
   string. An * (asterisk) character denotes a wildcard; a % (percentage)
character may also be used. The name must be at least 3 characters long,
not counting wildcard characters.}

\item{id}{The record ID of the specific record to return (only for scientific
names of species or infraspecific taxa)}

\item{downto}{The taxonomic level you want to go down to. See examples below.
The taxonomic level IS case sensitive, and you do have to spell it
correctly. See \code{data(rank_ref)} for spelling.}

\item{format}{The returned format (default = NULL). If NULL xml is used.
Currently only xml is supported.}

\item{start}{The first record to return (default = NULL). If NULL, the
results are returned from the first record (start=0). This is useful if
the total number of results is larger than the maximum number of results
returned by a single Web service query (currently the maximum number of
results returned by a single query is 500 for terse queries and 50 for
full queries).}

\item{checklist}{The year of the checklist to query, if you want a specific
year's checklist instead of the lastest as default (numeric).}

\item{verbose}{Print or suppress messages.}

\item{intermediate}{(logical) If TRUE, return a list of length two with target
taxon rank names, with additional list of data.frame's of intermediate
taxonomic groups. Default: FALSE}

\item{...}{Curl options passed on to \code{\link[httr]{GET}}}
}
\value{
A list of data.frame's.
}
\description{
Use Catalogue of Life to get downstream taxa to a given taxonomic level.
}
\details{
Provide only names instead of id's
}
\examples{
\dontrun{
# Some basic examples
col_downstream(name="Apis", downto="species")
col_downstream(name="Bryophyta", downto="family")

# get classes down from the kingdom Animalia
col_downstream(name="Animalia", downto="class")
col_downstream(name="Animalia", downto="class", intermediate=TRUE)

# An example that takes a bit longer
col_downstream(name=c("Plantae", "Animalia"), downto="class")

# Using a checklist from a specific year
col_downstream(name="Bryophyta", downto="family", checklist=2009)

# By id
col_downstream(id='576d098d770a39d09e2bcfa1c0896b26', downto="species",
  checklist=2012)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{species_plantarum_binomials}
\alias{species_plantarum_binomials}
\title{Species names from Species Plantarum}
\format{
A data frame with 5940 rows and 3 variables:
\itemize{
\item genus First part of the binomial species name for each species
within the \href{https://en.wikipedia.org/wiki/Genus}{genus}
\item epithet specific epithet or second part of the binomial species name
for each \href{https://en.wikipedia.org/wiki/Species}{species}
\item page_number The following abbreviations sometimes are used in the
page_number field.
\itemize{
\item "add." refers to addenda that appear on the unnumbered last page of
the index in volume two.
\item "err." refers to the unnumbered page of errata that appears following
the index in volume two.
\item "canc." following a page number indicates that the binomial appeared
on the cancelled version of that page and does not appear on its
replacement (as in the 1957-1959 facsimile edition.
}
}
}
\source{
\href{http://fmhibd.library.cmu.edu/HIBD-DB/Species/home.php}{Hunt Institute for Botanical Documentation}
}
\description{
These names have been compiled from
\href{https://en.wikipedia.org/wiki/Species_Plantarum}{\emph{Species Plantarum}} by
\href{https://en.wikipedia.org/wiki/Carl_Linnaeus}{Carl Linnaeus} originally
published in 1753. It is the first work to consistently apply
\href{https://en.wikipedia.org/wiki/Binomial_nomenclature}{binomial names}
and was the starting point for the naming of plants. The book lists every
species of plant known at the time, classified into
\href{https://en.wikipedia.org/wiki/Genus}{genera}. The dataset provides a
useful reference point to see how taxonomic names have changed since their
inception. The names were transcribed by Robert W. Kiger.
}
\references{
Linnaeus, C. 1753. Species Plantarum. 2 vols. Salvius, Stockholm.
[Facsimile edition, 1957-1959, Ray Society, London.]
}
\author{
Carl Linnaeus
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tax_rank.R
\name{tax_rank}
\alias{tax_rank}
\title{Get rank for a given taxonomic name.}
\usage{
tax_rank(sci_id, db = NULL, rows = NA, x = NULL, ...)
}
\arguments{
\item{sci_id}{(character) Vector of one or more taxon names (character) or
IDs (character or numeric) to query. Or objects returned from \verb{get_*()}
functions like \code{\link[=get_tsn]{get_tsn()}}}

\item{db}{(character) database to query. either \code{ncbi}, \code{itis}, \code{eol},
\code{tropicos}, \code{gbif},\code{nbn}, \code{worms}, \code{natserv}, \code{bold}. Note that each
taxonomic data source has their own identifiers, so that if you provide the
wrong \code{db} value for the identifier you may get a result, but it will
likely be wrong (not what you were expecting). If using ncbi we
recommend getting an API key; see \link{taxize-authentication}}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA,
all rows are considered. passed down to \verb{get_*()} functions.}

\item{x}{Deprecated, see \code{sci_id}}

\item{...}{Additional arguments to \code{\link[=classification]{classification()}}}
}
\value{
A named list of character vectors with ranks (all lower-cased)
}
\description{
Get rank for a given taxonomic name.
}
\note{
While \code{\link[=tax_name]{tax_name()}} returns the name of a specified
rank, \code{\link[=tax_rank]{tax_rank()}} returns the actual rank of the taxon.
}
\examples{
\dontrun{
tax_rank("Helianthus annuus", db = "itis")
tax_rank("Helianthus annuus", db = "natserv")
tax_rank(get_tsn("Helianthus annuus"))
tax_rank(c("Helianthus", "Pinus", "Poa"), db = "itis")

tax_rank(get_boldid("Helianthus annuus"))
tax_rank("421377", db = "bold")
tax_rank(421377, db = "bold")

tax_rank(c("Plantae", "Helianthus annuus",
  "Puma", "Homo sapiens"), db = 'itis')
tax_rank(c("Helianthus annuus", "Quercus", "Fabaceae"), db = 'tropicos')

tax_rank(names_list("species"), db = 'gbif')
tax_rank(names_list("family"), db = 'gbif')

tax_rank(c("Gadus morhua", "Lichenopora neapolitana"),
  db = "worms")
}
}
\seealso{
\code{\link[=classification]{classification()}},\code{\link[=tax_name]{tax_name()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{apg_orders}
\alias{apg_orders}
\title{MOBOT order names}
\format{A data frame with 494 rows and 3 variables:
\itemize{
\item \code{original} original data record from APG website
\item \code{this} Order name
\item \code{that} Replacement order name
}}
\source{
\url{http://www.mobot.org/MOBOT/research/APweb/}
}
\description{
Order names and their replacements from the Angiosperm Phylogeny
Website system of flowering plant classification.
}
\details{
This dataset is from Version 13, incorporated on 2015-04-29.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize-package.R
\docType{data}
\name{plantNames}
\alias{plantNames}
\title{Vector of plant species (genus - specific epithet) names from ThePlantList}
\format{
A vector of length 1182
}
\source{
http://www.theplantlist.org
}
\description{
These names are from http://www.theplantlist.org, and are a
randomly chosen subset of names of the form genus/specific epithet
for the purpose of having some names to play with for examples in
this package.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genbank2uid.R
\name{genbank2uid}
\alias{genbank2uid}
\title{Get NCBI taxonomy UID from GenBankID}
\usage{
genbank2uid(id, batch_size = 100, key = NULL, ...)
}
\arguments{
\item{id}{A GenBank accession alphanumeric string, or a gi numeric string.}

\item{batch_size}{The number of queries to submit at a time.}

\item{key}{(character) NCBI Entrez API key. optional. See Details.}

\item{...}{Curl args passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
one or more NCBI taxonomic IDs
}
\description{
Get NCBI taxonomy UID from GenBankID
}
\details{
See https://www.ncbi.nlm.nih.gov/Sitemap/sequenceIDs.html for
help on why there are two identifiers, and the difference between them.
}
\section{Authentication}{

See \link{taxize-authentication} for help on authentication. We
recommend getting an API key.
}

\section{HTTP version}{

We hard code \code{http_version = 2L} to use HTTP/1.1 in HTTP requests to
the Entrez API. See \code{curl::curl_symbols('CURL_HTTP_VERSION')}
}

\section{Rate limits}{

In case you run into errors due to your rate limit being exceeded, see
\code{\link[=taxize_options]{taxize_options()}}, where you can set \code{ncbi_sleep}.
}

\examples{
\dontrun{
# with accession numbers
genbank2uid(id = 'AJ748748')
genbank2uid(id = 'Y13155')
genbank2uid(id = 'X78312')
genbank2uid(id = 'KM495596')

# with gi numbers
genbank2uid(id = 62689767)
genbank2uid(id = 22775511)
genbank2uid(id = 156446673)

# pass in many accession or gi numbers
genbank2uid(c(62689767,156446673))
genbank2uid(c('X78312','KM495596'))
genbank2uid(list('X78312',156446673))

# curl options
res <- genbank2uid(id = 156446673, verbose = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_accnames.R
\name{tp_acceptednames}
\alias{tp_acceptednames}
\title{Return all accepted names for a taxon name with a given id.}
\usage{
tp_acceptednames(id, format = "json", output = "df", key = NULL)
}
\arguments{
\item{id}{the taxon identifier code}

\item{format}{return in json or xml format (defaults to json)}

\item{output}{raw = json or xml; or df = data.frame}

\item{key}{Your Tropicos API key; loads from .Rprofile.}
}
\description{
Function name changed to tp_accnames.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eubon.R
\name{eubon_search}
\alias{eubon_search}
\title{EUBON taxonomy search}
\usage{
eubon_search(
  query,
  providers = "pesi",
  searchMode = "scientificNameExact",
  addSynonymy = FALSE,
  addParentTaxon = FALSE,
  timeout = 0,
  dedup = NULL,
  limit = 20,
  page = 1,
  ...
)
}
\arguments{
\item{query}{(character) The scientific name to search for. For example:
"Bellis perennis", "Prionus" or "Bolinus brandaris". This is an exact search
so wildcard characters are not supported}

\item{providers}{(character) A list of provider id strings concatenated by
comma characters. The default : "pesi,bgbm-cdm-server[col]" will be
used if this parameter is not set. A list of all available provider ids
can be obtained from the '/capabilities' service end point. Providers can be
nested, that is a parent provider can have sub providers. If the id of the
parent provider is supplied all subproviders will be queried. The query
can also be restricted to one or more subproviders by using the following
syntax: parent-id[sub-id-1,sub-id2,...]}

\item{searchMode}{(character) Specifies the searchMode. Possible search
modes are: \code{scientificNameExact}, \code{scientificNameLike} (begins with),
\code{vernacularNameExact}, \code{vernacularNameLike}
(contains), \code{findByIdentifier}. If the a provider does not support the
chosen searchMode it will be skipped and the status message in the
tnrClientStatus will be set to 'unsupported search mode' in this case.}

\item{addSynonymy}{(logical) Indicates whether the synonymy of the accepted
taxon should be included into the response. Turning this option on may
cause an increased response time. Default: \code{FALSE}}

\item{addParentTaxon}{(logical) Indicates whether the the parent taxon of
the accepted taxon should be included into the response. Turning this option
on may cause a slightly increased response time. Default: \code{FALSE}}

\item{timeout}{(numeric) The maximum of milliseconds to wait for responses
from any of the providers. If the timeout is exceeded the service will just
return the responses that have been received so far. The default timeout is
0 ms (wait for ever)}

\item{dedup}{(character) Allows to deduplicate the results by making use of
a deduplication strategy. The deduplication is done by comparing
specific properties of the taxon:
\itemize{
\item id: compares 'taxon.identifier'
\item id_name: compares 'taxon.identifier' AND
'taxon.taxonName.scientificName'
\item name: compares 'taxon.taxonName.scientificName' Using the pure
'name' strategy is not recommended.
}}

\item{limit}{(numeric/integer) number of records to retrieve. default: 20.
This only affects the search mode \code{scientificNameLike} and
\code{vernacularNameLike}; other search modes are expected to return only one
record per check list}

\item{page}{(numeric/integer) page to retrieve. default: 1. This only
affects the search mode \code{scientificNameLike} and \code{vernacularNameLike}; other
search modes are expected to return only one record per check list}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\description{
EUBON taxonomy search
}
\examples{
\dontrun{
eubon_search("Prionus")
eubon_search("Salmo", "pesi")
eubon_search("Salmo", c("pesi", "worms"))
eubon_search("Salmo", "worms", "scientificNameLike")
eubon_search("Salmo", "worms", "scientificNameLike", limit = 3)
eubon_search("Salmo", "worms", "scientificNameLike", limit = 20, page = 2)
eubon_search("Salmo", "worms", addSynonymy = TRUE)
eubon_search("Salmo", "worms", addParentTaxon = TRUE)
}
}
\references{
https://cybertaxonomy.eu/eu-bon/utis/1.3/doc.html
}
\seealso{
Other eubon-methods: 
\code{\link{eubon_capabilities}()},
\code{\link{eubon_children}()},
\code{\link{eubon_hierarchy}()}
}
\concept{eubon-methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_wormsid.R
\name{get_wormsid}
\alias{get_wormsid}
\alias{as.wormsid}
\alias{as.wormsid.wormsid}
\alias{as.wormsid.character}
\alias{as.wormsid.list}
\alias{as.wormsid.numeric}
\alias{as.wormsid.data.frame}
\alias{as.data.frame.wormsid}
\alias{get_wormsid_}
\title{Get Worms ID for a taxon name}
\usage{
get_wormsid(
  sci_com,
  searchtype = "scientific",
  marine_only = TRUE,
  fuzzy = NULL,
  accepted = FALSE,
  ask = TRUE,
  messages = TRUE,
  rows = NA,
  query = NULL,
  ...
)

as.wormsid(x, check = TRUE)

\method{as.wormsid}{wormsid}(x, check = TRUE)

\method{as.wormsid}{character}(x, check = TRUE)

\method{as.wormsid}{list}(x, check = TRUE)

\method{as.wormsid}{numeric}(x, check = TRUE)

\method{as.wormsid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{wormsid}(x, ...)

get_wormsid_(
  sci_com,
  messages = TRUE,
  searchtype = "scientific",
  marine_only = TRUE,
  fuzzy = NULL,
  accepted = TRUE,
  rows = NA,
  query = NULL,
  ...
)
}
\arguments{
\item{sci_com}{character; A vector of common or scientific names. Or, a
\code{taxon_state} object (see \link{taxon-state})}

\item{searchtype}{character; One of 'scientific' or 'common', or any unique
abbreviation}

\item{marine_only}{logical; marine only? default: \code{TRUE} (only used
when \code{searchtype="scientific"}); passed on to \code{\link[worrms:wm_records_name]{worrms::wm_records_name()}}}

\item{fuzzy}{logical; fuzzy search. default: \code{NULL} (\code{TRUE} for
\code{searchtype="scientific"} and \code{FALSE} for \code{searchtype="common"} to match
the default values for those parameters in \pkg{worrms} package); passed on
to \code{\link[worrms:wm_records_name]{worrms::wm_records_name()}} or \code{\link[worrms:wm_records_common]{worrms::wm_records_common()}}}

\item{accepted}{logical; If TRUE, removes names that are not accepted valid
names by WORMS. Set to \code{FALSE} (default) to give back both accepted
and unaccepted names.}

\item{ask}{logical; should get_wormsid be run in interactive mode?
If \code{TRUE} and more than one wormsid is found for the species, the
user is asked for input. If \code{FALSE} NA is returned for
multiple matches.}

\item{messages}{logical; should progress be printed?}

\item{rows}{numeric; Any number from 1 to infinity. If the default NaN, all
rows are considered. Note that this function still only gives back a wormsid
class object with one to many identifiers. See \code{\link[=get_wormsid_]{get_wormsid_()}} to get back
all, or a subset, of the raw data that you are presented during the ask
process.}

\item{query}{Deprecated, see \code{sci_com}}

\item{...}{Ignored}

\item{x}{Input to \code{as.wormsid}}

\item{check}{logical; Check if ID matches any existing on the DB, only
used in \code{\link[=as.wormsid]{as.wormsid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Retrieve Worms ID of a taxon from World Register of Marine
Species (WORMS).
}
\examples{
\dontrun{
(x <- get_wormsid('Gadus morhua'))
attributes(x)
attr(x, "match")
attr(x, "multiple_matches")
attr(x, "pattern_match")
attr(x, "uri")

get_wormsid('Pomatomus saltatrix')
get_wormsid(c("Gadus morhua", "Lichenopora neapolitana"))

# marine_only
get_wormsid("Apedinella", marine_only=TRUE)
get_wormsid("Apedinella", marine_only=FALSE)

# fuzzy
## searchtype="scientific": fuzzy is TRUE by default
get_wormsid("Platypro", searchtype="scientific", fuzzy=TRUE)
get_wormsid("Platypro", searchtype="scientific", fuzzy=FALSE)
## searchtype="common": fuzzy is FALSE by default
get_wormsid("clam", searchtype="common", fuzzy=FALSE)
get_wormsid("clam", searchtype="common", fuzzy=TRUE)

# by common name
get_wormsid("dolphin", 'common')
get_wormsid("clam", 'common')

# specify rows to limit choices available
get_wormsid('Plat')
get_wormsid('Plat', rows=1)
get_wormsid('Plat', rows=1:2)

# When not found
get_wormsid("howdy")
get_wormsid(c('Gadus morhua', "howdy"))

# Convert a wormsid without class information to a wormsid class
# already a wormsid, returns the same
as.wormsid(get_wormsid('Gadus morhua'))
# same
as.wormsid(get_wormsid(c('Gadus morhua', 'Pomatomus saltatrix')))
# numeric
as.wormsid(126436)
# numeric vector, length > 1
as.wormsid(c(126436,151482))
# character
as.wormsid("126436")
# character vector, length > 1
as.wormsid(c("126436","151482"))
# list, either numeric or character
as.wormsid(list("126436","151482"))
## dont check, much faster
as.wormsid("126436", check=FALSE)
as.wormsid(126436, check=FALSE)
as.wormsid(c("126436","151482"), check=FALSE)
as.wormsid(list("126436","151482"), check=FALSE)

(out <- as.wormsid(c(126436,151482)))
data.frame(out)
as.wormsid( data.frame(out) )

# Get all data back
get_wormsid_("Plat")
get_wormsid_("Plat", rows=1)
get_wormsid_("Plat", rows=1:2)
get_wormsid_("Plat", rows=1:75)
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxize_cite.R
\name{taxize_cite}
\alias{taxize_cite}
\title{Get citations and licenses for data sources used in taxize}
\usage{
taxize_cite(fxn = "itis", what = "citation")
}
\arguments{
\item{fxn}{Function to search on. A special case is the package name
'taxize' that will give the citations for the package.}

\item{what}{One of citation (default), license, or both.}
}
\description{
Get citations and licenses for data sources used in taxize
}
\examples{
taxize_cite(fxn='eol_search')
taxize_cite(fxn='itis_hierarchy')
taxize_cite(fxn='tp_classification')
taxize_cite(fxn='gbif_ping')
taxize_cite(fxn='plantminer')
taxize_cite(fxn='get_natservid_')
taxize_cite(fxn='as.natservid')
taxize_cite(fxn='get_wormsid')
taxize_cite(fxn='as.wormsid')

# Functions that use many data sources
taxize_cite(fxn='synonyms')
taxize_cite(fxn='classification')

# Get the taxize citation
taxize_cite(fxn='taxize')

# Get license information
taxize_cite(fxn='taxize', "license")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eubon_capabilities.R
\name{eubon_capabilities}
\alias{eubon_capabilities}
\title{EUBON capabilities}
\usage{
eubon_capabilities(...)
}
\arguments{
\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\description{
EUBON capabilities
}
\examples{
\dontrun{
eubon_capabilities()
}
}
\references{
https://cybertaxonomy.eu/eu-bon/utis/1.3/doc.html
}
\seealso{
Other eubon-methods: 
\code{\link{eubon_children}()},
\code{\link{eubon_hierarchy}()},
\code{\link{eubon_search}()}
}
\concept{eubon-methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eol_search.R
\name{eol_search}
\alias{eol_search}
\title{Search for terms in EOL database.}
\usage{
eol_search(
  sci,
  page = 1,
  exact = NULL,
  filter_tid = NULL,
  filter_heid = NULL,
  filter_by_string = NULL,
  cache_ttl = NULL,
  terms = NULL,
  ...
)
}
\arguments{
\item{sci}{(character) scientific name}

\item{page}{A maximum of 30 results are returned per page. This parameter
allows you to fetch more pages of results if there are more than
30 matches (Default 1)}

\item{exact}{Will find taxon pages if the preferred name or any synonym
or common name exactly matches the search term.}

\item{filter_tid}{Given an EOL page ID, search results will be limited
to members of that taxonomic group}

\item{filter_heid}{Given a Hierarchy Entry ID, search results will be
limited to members of that taxonomic group}

\item{filter_by_string}{Given a search term, an exact search will be made
and that matching page will be used as the taxonomic group against which
to filter search results}

\item{cache_ttl}{The number of seconds you wish to have the response cached.}

\item{terms}{Deprecated, see \code{sci}}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
A data frame with four columns:
\itemize{
\item pageid: pageid, this is the same as the eolid you can get from
\code{\link[=get_eolid]{get_eolid()}}
\item name: taxonomic name, may or may not contain the taxonomic authority
\item link: URL for the taxon in question
\item content: a string of semi-colon separated names. it's not clear
to us what these represent exactly, but figured why not give it to users
in case some may find it useful
}
}
\description{
Search for terms in EOL database.
}
\details{
It's possible to return JSON or XML with the EOL API. However,
this function only returns JSON for now.
}
\examples{
\dontrun{
eol_search(sci='Homo')
eol_search(sci='Salix', verbose = TRUE)
eol_search(sci='Ursus americanus')
eol_search('Pinus contorta')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gna_verifier.R
\name{gna_verifier}
\alias{gna_verifier}
\title{Verify a list of scientific names against biodiversity data-sources.}
\usage{
gna_verifier(
  names,
  data_sources = c(1, 12),
  all_matches = FALSE,
  capitalize = FALSE,
  species_group = FALSE,
  fuzzy_uninomial = FALSE,
  stats = FALSE,
  main_taxon_threshold = 0.5,
  output_type = "table",
  ...
)
}
\arguments{
\item{names}{A \code{character} vector of taxon names to verify.}

\item{data_sources}{A \code{character} or \code{integer} vector with numbers
corresponding to data sources. See the Global Names Architecture
documentation for a list of available options.}

\item{all_matches}{When \code{TRUE}, return all found matches, not only the best
one. Multiple results are returned in results. These results are sorted by
matching quality, the first result is the same as bestResult.}

\item{capitalize}{When \code{TRUE}, capitalize the first letter of a name-string.}

\item{species_group}{When \code{TRUE}, expands the search to species group where
applicable.}

\item{fuzzy_uninomial}{When \code{TRUE}, allows fuzzy matching for uninomial
names.}

\item{stats}{When \code{TRUE}, finds out a kingdom and a taxon (main taxon) that
contain most names. It only takes in account the names matched to the
Catalogue of Life entries. This option is ignored, if the Catalogue of Life
is not included in data-sources.}

\item{main_taxon_threshold}{A \code{numeric} vector from 0.5 to 1. This sets the
minimal percentage for the main taxon discovery.}

\item{output_type}{A \code{character} vector of length 1, either \code{table} or
\code{list}, indicating the format of the output. The tabular output only
contains values that consistently appear in all results, so \code{list} output
can have additional information. For \code{list} and \code{json} outputs, only values
for unique taxon names are returned, but the \code{table} output has rows that
correspond 1-1 with the input data.}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
Depends on the value of the \code{output_type} option
}
\description{
This service parses incoming names, executes exact or fuzzy matching as
required, and returns the best-scored result. Optionally, it can also return
matches from data-sources selected by a user.
}
\examples{
\dontrun{
gna_verifier(c("Helianthus annuus", "Homo saapiens"))
gna_verifier(c("Helianthus annuus", "Homo saapiens"), all_matches = TRUE)
}
}
\author{
Zachary S.L. Foster
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncbi_getbyid.R
\name{get_genes}
\alias{get_genes}
\title{Retrieve gene sequences from NCBI by accession number.}
\usage{
get_genes(...)
}
\description{
THIS FUNCTION IS DEFUNCT.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_eolid.R
\name{get_eolid}
\alias{get_eolid}
\alias{as.eolid}
\alias{as.eolid.eolid}
\alias{as.eolid.character}
\alias{as.eolid.list}
\alias{as.eolid.numeric}
\alias{as.eolid.data.frame}
\alias{as.data.frame.eolid}
\alias{get_eolid_}
\title{Get the EOL ID from Encyclopedia of Life from taxonomic names.}
\usage{
get_eolid(
  sci_com,
  ask = TRUE,
  messages = TRUE,
  rows = NA,
  rank = NULL,
  data_source = NULL,
  sciname = NULL,
  ...
)

as.eolid(x, check = TRUE)

\method{as.eolid}{eolid}(x, check = TRUE)

\method{as.eolid}{character}(x, check = TRUE)

\method{as.eolid}{list}(x, check = TRUE)

\method{as.eolid}{numeric}(x, check = TRUE)

\method{as.eolid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{eolid}(x, ...)

get_eolid_(sci_com, messages = TRUE, rows = NA, sciname = NULL, ...)
}
\arguments{
\item{sci_com}{character; one or more scientific or common names. Or,
a \code{taxon_state} object (see \link{taxon-state})}

\item{ask}{logical; should get_eolid be run in interactive mode?
If TRUE and more than one ID is found for the species, the user is asked for
input. If FALSE NA is returned for multiple matches.}

\item{messages}{logical; If \code{TRUE} the actual taxon queried is printed
on the console.}

\item{rows}{numeric; Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this function still only gives back a eolid
class object with one to many identifiers. See
\code{\link[=get_eolid_]{get_eolid_()}} to get back all, or a subset, of the raw
data that you are presented during the ask process.}

\item{rank}{(character) A taxonomic rank name. See \link{rank_ref}
for possible options. Though note that some data sources use atypical ranks,
so inspect the data itself for options. Optional. See \code{Filtering} below.}

\item{data_source}{(character) A data source inside of EOL. These are
longish names like e.g., "Barcode of Life Data Systems" or
"USDA PLANTS images". Optional. See \code{Filtering} below.}

\item{sciname}{Deprecated, see \code{sci_com}}

\item{...}{Further args passed on to \code{\link[=eol_search]{eol_search()}}}

\item{x}{Input to \code{\link[=as.eolid]{as.eolid()}}}

\item{check}{logical; Check if ID matches any existing on the DB, only
used in \code{\link[=as.eolid]{as.eolid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Note that EOL doesn't expose an API endpoint for directly querying for EOL
taxon ID's, so we first use the function \code{\link[=eol_search]{eol_search()}}
to find pages that deal with the species of interest, then use
\code{\link[=eol_pages]{eol_pages()}} to find the actual taxon IDs.
}
\details{
EOL is a bit odd in that they have page IDs for each taxon, but
then within that, they have taxon ids for various taxa within that page
(e.g., GBIF and NCBI each have a taxon they refer to within the
page [i.e., taxon]). And we need the taxon ids from a particular data
provider (e.g, NCBI) to do other things, like get a higher classification
tree. However, humans want the page id, not the taxon id. So, the
id returned from this function is the taxon id, not the page id. You can
get the page id for a taxon by using \code{\link[=eol_search]{eol_search()}} and
`\code{\link[=eol_pages]{eol_pages()}}, and the URI returned in the attributes for a
taxon will lead you to the taxon page, and the ID in the URL is the
page id.
}
\section{Filtering}{

The parameters \code{rank} and \code{data_source} are not used in the
search to the data provider, but are used in filtering the data down to a
subset that is closer to the target you want.  For all these parameters,
you can use regex strings since we use \code{\link[=grep]{grep()}} internally to
match. Filtering narrows down to the set that matches your query, and
removes the rest.
}

\examples{
\dontrun{
get_eolid(sci_com='Pinus contorta')
get_eolid(sci_com='Puma concolor')

get_eolid(c("Puma concolor", "Pinus contorta"))

# specify rows to limit choices available
get_eolid('Poa annua')
get_eolid('Poa annua', rows=1)
get_eolid('Poa annua', rows=2)
get_eolid('Poa annua', rows=1:2)

# When not found
get_eolid(sci_com="uaudnadndj")
get_eolid(c("Chironomus riparius", "uaudnadndj"))

# filter results to a rank or data source, or both
get_eolid("Satyrium")
get_eolid("Satyrium", rank = "genus")
get_eolid("Satyrium", data_source = "INAT")
get_eolid("Satyrium", rank = "genus",
  data_source = "North Pacific Species List")

# Convert a eolid without class information to a eolid class
# already a eolid, returns the same
as.eolid(get_eolid("Chironomus riparius"))
# same
as.eolid(get_eolid(c("Chironomus riparius","Pinus contorta")))
# numeric
as.eolid(10247706)
# numeric vector, length > 1
as.eolid(c(6985636,12188704,10247706))
# character
as.eolid("6985636")
# character vector, length > 1
as.eolid(c("6985636","12188704","10247706"))
# list, either numeric or character
as.eolid(list("6985636","12188704","10247706"))
## dont check, much faster
as.eolid("6985636", check=FALSE)
as.eolid(6985636, check=FALSE)
as.eolid(c("6985636","12188704","10247706"), check=FALSE)
as.eolid(list("6985636","12188704","10247706"), check=FALSE)

(out <- as.eolid(c(6985636,12188704,10247706)))
data.frame(out)
as.eolid( data.frame(out) )

# Get all data back
get_eolid_("Poa annua")
get_eolid_("Poa annua", rows=2)
get_eolid_("Poa annua", rows=1:2)
get_eolid_(c("asdfadfasd", "Pinus contorta"))
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_natservid}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\author{
Scott Chamberlain
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bold_copied_code.R
\name{bold_tax_id}
\alias{bold_tax_id}
\title{Search BOLD for taxonomy data by BOLD ID.}
\usage{
bold_tax_id(
  id,
  dataTypes = "basic",
  includeTree = FALSE,
  response = FALSE,
  ...
)
}
\description{
Search BOLD for taxonomy data by BOLD ID.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tpl_search.r
\name{tpl_search}
\alias{tpl_search}
\title{A light wrapper around the taxonstand fxn to call Theplantlist.org database.}
\usage{
tpl_search()
}
\description{
THIS FUNCTION IS DEFUNCT.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tp_search.R
\name{tp_search}
\alias{tp_search}
\title{Search Tropicos by scientific name, common name, or Tropicos ID.}
\usage{
tp_search(
  sci = NULL,
  com = NULL,
  nameid = NULL,
  orderby = NULL,
  sortorder = NULL,
  pagesize = NULL,
  startrow = NULL,
  type = NULL,
  key = NULL,
  name = NULL,
  commonname = NULL,
  ...
)
}
\arguments{
\item{sci}{A scientific name, e.g., "poa annua". See Details.}

\item{com}{A common name, e.g., "annual blue grass"}

\item{nameid}{Your search string. e.g., "25509881"}

\item{orderby}{Your search string. e.g., "1"}

\item{sortorder}{Your search string. e.g., "ascending"}

\item{pagesize}{Your search string. e.g., "100"}

\item{startrow}{Your search string. e.g., "1"}

\item{type}{Type of search, "wildcard" (default) will add a wildcard to
the end of your search string. "exact" will use your search string exactly.}

\item{key}{Your Tropicos API key; See \link{taxize-authentication}
for help on authentication}

\item{name}{Deprecated, see \code{sci}}

\item{commonname}{Deprecated, see \code{com}}

\item{...}{Further args passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
List or dataframe.
}
\description{
Search Tropicos by scientific name, common name, or Tropicos ID.
}
\details{
More details on the \code{name} parameter: Tropicos will fail
if you include a period (\code{.}) in your name string, e.g., \code{var.},
so we replace periods before the request is made to the Tropicos web service.
In addition, Tropicos for some reason doesn't want to see sub-specific rank
names like \code{var}/\code{subsp}, so remove those from your query.
}
\examples{
\dontrun{
tp_search(sci = 'Poa annua')
tp_search(sci = 'Poa annua subsp. annua')
tp_search(sci = 'Poa annua var. annua')
tp_search(sci = 'Poa annua var annua')
tp_search(sci = 'Poa annua annua')
}
}
\references{
http://services.tropicos.org/help?method=SearchNameXml
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_natservid.R
\name{get_natservid}
\alias{get_natservid}
\alias{as.natservid}
\alias{as.natservid.natservid}
\alias{as.natservid.character}
\alias{as.natservid.list}
\alias{as.natservid.numeric}
\alias{as.natservid.data.frame}
\alias{as.data.frame.natservid}
\alias{get_natservid_}
\title{Get NatureServe taxonomic ID for a taxon name}
\usage{
get_natservid(
  sci_com,
  searchtype = "scientific",
  ask = TRUE,
  messages = TRUE,
  rows = NA,
  query = NULL,
  ...
)

as.natservid(x, check = TRUE)

\method{as.natservid}{natservid}(x, check = TRUE)

\method{as.natservid}{character}(x, check = TRUE)

\method{as.natservid}{list}(x, check = TRUE)

\method{as.natservid}{numeric}(x, check = TRUE)

\method{as.natservid}{data.frame}(x, check = TRUE)

\method{as.data.frame}{natservid}(x, ...)

get_natservid_(
  sci_com,
  searchtype = "scientific",
  messages = TRUE,
  rows = NA,
  query = NULL,
  ...
)
}
\arguments{
\item{sci_com}{character; A vector of common or scientific names. Or, a
\code{taxon_state} object (see \link{taxon-state})}

\item{searchtype}{character; One of 'scientific' (default) or 'common'.
This doesn't affect the query to NatureServe - but rather affects what
column of data is targeted in name filtering post data request.}

\item{ask}{logical; should get_natservid be run in interactive mode?
If \code{TRUE} and more than one wormsid is found for the species, the
user is asked for input. If \code{FALSE} NA is returned for
multiple matches. default: \code{TRUE}}

\item{messages}{logical; should progress be printed? default: \code{TRUE}}

\item{rows}{numeric; Any number from 1 to infinity. If the default NaN, all
rows are considered. Note that this function still only gives back a
natservid class object with one to many identifiers. See
\code{get_natservid_()} to get back all, or a subset, of the raw
data that you are presented during the ask process.}

\item{query}{Deprecated, see \code{sci_com}}

\item{...}{curl options passed on to \link[crul:verb-POST]{crul::verb-POST}}

\item{x}{Input to \code{as.natservid}}

\item{check}{logical; Check if ID matches any existing on the DB, only
used in \code{\link[=as.natservid]{as.natservid()}}}
}
\value{
A vector of taxonomic identifiers as an S3 class.
If a taxon is not found an \code{NA} is given. If more than one identifier
is found the function asks for user input if \code{ask = TRUE}, otherwise
returns \code{NA}. If \code{ask=FALSE} and \code{rows} does not equal
\code{NA}, then a data.frame is given back, but not of the uid class, which
you can't pass on to other functions as you normally can.

See \code{\link{get_id_details}} for further details including
attributes and exceptions
}
\description{
Get NatureServe taxonomic ID for a taxon name
}
\note{
Authentication no longer required
}
\examples{
\dontrun{
(x <- get_natservid("Helianthus annuus", verbose = TRUE))
attributes(x)
attr(x, "match")
attr(x, "multiple_matches")
attr(x, "pattern_match")
attr(x, "uri")

get_natservid('Gadus morhua')
get_natservid(c("Helianthus annuus", 'Gadus morhua'))

# specify rows to limit choices available
get_natservid('Ruby Quaker Moth', 'common')
get_natservid('Ruby*', 'common')
get_natservid('Ruby*', 'common', rows=1)
get_natservid('Ruby*', 'common', rows=1:2)

# When not found
get_natservid("howdy")
get_natservid(c('Gadus morhua', "howdy"))

# Convert a natservid without class information to a natservid class
# already a natservid, returns the same
as.natservid(get_natservid('Pomatomus saltatrix'))
# same
as.natservid(get_natservid(c('Gadus morhua', 'Pomatomus saltatrix')))
# character
as.natservid(101905)
# character vector, length > 1
as.natservid(c(101905, 101998))
# list, either numeric or character
as.natservid(list(101905, 101998))
## dont check, much faster
as.natservid(101905, check = FALSE)
as.natservid(c(101905, 101998), check = FALSE)
as.natservid(list(101905, 101998), check = FALSE)

(out <- as.natservid(c(101905, 101998), check = FALSE))
data.frame(out)
as.natservid( data.frame(out) )

# Get all data back
get_natservid_("Helianthus")
get_natservid_("Ruby*", searchtype = "common")
get_natservid_("Ruby*", searchtype = "common", rows=1:3)
}
}
\seealso{
\code{\link[=classification]{classification()}}

Other taxonomic-ids: 
\code{\link{get_boldid}()},
\code{\link{get_eolid}()},
\code{\link{get_gbifid}()},
\code{\link{get_ids}()},
\code{\link{get_iucn}()},
\code{\link{get_nbnid}()},
\code{\link{get_pow}()},
\code{\link{get_tolid}()},
\code{\link{get_tpsid}()},
\code{\link{get_tsn}()},
\code{\link{get_uid}()},
\code{\link{get_wiki}()},
\code{\link{get_wormsid}()}
}
\concept{taxonomic-ids}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/children.R
\name{children}
\alias{children}
\alias{children.default}
\alias{children.tsn}
\alias{children.wormsid}
\alias{children.ids}
\alias{children.uid}
\alias{children.boldid}
\title{Retrieve immediate children taxa for a given taxon name or ID.}
\usage{
children(...)

\method{children}{default}(sci_id, db = NULL, rows = NA, x = NULL, ...)

\method{children}{tsn}(sci_id, db = NULL, ...)

\method{children}{wormsid}(sci_id, db = NULL, ...)

\method{children}{ids}(sci_id, db = NULL, ...)

\method{children}{uid}(sci_id, db = NULL, ...)

\method{children}{boldid}(sci_id, db = NULL, ...)
}
\arguments{
\item{...}{Further args passed on to \code{\link[ritis:hierarchy]{ritis::hierarchy_down()}},
\code{\link[=ncbi_children]{ncbi_children()}}, \code{\link[worrms:wm_children]{worrms::wm_children()}}, \code{\link[=bold_children]{bold_children()}}
See those functions for what parameters can be passed on.}

\item{sci_id}{Vector of taxa names (character) or IDs (character or numeric)
to query.}

\item{db}{character; database to query. One or more of \code{itis},
\code{ncbi}, \code{worms}, or \code{bold}. Note that each taxonomic data
source has their own identifiers, so that if you provide the wrong
\code{db} value for the identifier you could get a result, but it will
likely be wrong (not what you were expecting). If using ncbi, we recommend
getting an API key; see \link{taxize-authentication}}

\item{rows}{(numeric) Any number from 1 to infinity. If the default NA, all
rows are considered. Note that this parameter is ignored if you pass in a
taxonomic id of any of the acceptable classes: tsn. NCBI has a
method for this function but rows doesn't work.}

\item{x}{Deprecated, see \code{sci_id}}
}
\value{
A named list of data.frames with the children names of every
supplied taxa. You get an NA if there was no match in the database.
}
\description{
This function is different from \code{\link[=downstream]{downstream()}} in that it only
collects immediate taxonomic children, while \code{\link[=downstream]{downstream()}}
collects taxonomic names down to a specified taxonomic rank, e.g.,
getting all species in a family.
}
\section{ncbi}{

note that with \code{db = "ncbi"}, we set \code{ambiguous = TRUE}; that is, children
taxa with words like "unclassified", "unknown", "uncultured", "sp." are
NOT removed
}

\section{bold}{

BEWARE: \code{db="bold"} scrapes the BOLD website, so may be unstable. That is,
one day it may work, and the next it may fail. Open an issue if you
encounter an error: https://github.com/ropensci/taxize/issues
}

\examples{
\dontrun{
# Plug in taxonomic IDs
children(161994, db = "itis")
children(8028, db = "ncbi")
## works with numeric if as character as well
children(161994, db = "itis")
children(88899, db = "bold")
children(as.boldid(88899))

# Plug in taxon names
children("Salmo", db = 'itis')
children("Salmo", db = 'ncbi')
children("Salmo", db = 'worms')
children("Salmo", db = 'bold')

# Plug in IDs
(id <- get_wormsid("Gadus"))
children(id)

# Many taxa
sp <- c("Tragia", "Schistocarpha", "Encalypta")
children(sp, db = 'itis')

# Two data sources
(ids <- get_ids("Apis", db = c('ncbi','itis')))
children(ids)
## same result
children(get_ids("Apis", db = c('ncbi','itis')))

# Use the rows parameter
children("Poa", db = 'itis')
children("Poa", db = 'itis', rows=1)

# use curl options
res <- children("Poa", db = 'itis', rows=1, verbose = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gnr_resolve.R
\name{gnr_resolve}
\alias{gnr_resolve}
\title{Resolve names using Global Names Resolver}
\usage{
gnr_resolve(
  sci,
  data_source_ids = NULL,
  resolve_once = FALSE,
  with_context = FALSE,
  canonical = FALSE,
  highestscore = TRUE,
  best_match_only = FALSE,
  preferred_data_sources = NULL,
  with_canonical_ranks = FALSE,
  http = "get",
  cap_first = TRUE,
  fields = "minimal",
  names = NULL,
  ...
)
}
\arguments{
\item{sci}{character; taxonomic names to be resolved. Doesn't work for
vernacular/common names.}

\item{data_source_ids}{character; IDs to specify what data source is
searched. See \code{\link[=gnr_datasources]{gnr_datasources()}}.}

\item{resolve_once}{logical; Find the first available match instead of
matches across all data sources with all possible renderings of a name.
When \code{TRUE}, response is rapid but incomplete.}

\item{with_context}{logical; Reduce the likelihood of matches to taxonomic
homonyms. When \code{TRUE} a common taxonomic context is calculated for all
supplied names from matches in data sources that have classification tree
paths. Names out of determined context are penalized during score
calculation.}

\item{canonical}{logical; If \code{FALSE} (default), gives back names with
taxonomic authorities. If \code{TRUE}, returns canocial names (without tax.
authorities and abbreviations).}

\item{highestscore}{logical; Return those names with the highest score for
each searched name? Defunct}

\item{best_match_only}{(logical) If \code{TRUE}, best match only returned.
Default: \code{FALSE}}

\item{preferred_data_sources}{(character) A vector of one or more data source
IDs.}

\item{with_canonical_ranks}{(logical) Returns names with infraspecific ranks,
if present. If \code{TRUE}, we force \code{canonical=TRUE}, otherwise this parameter
would have no effect. Default: \code{FALSE}}

\item{http}{The HTTP method to use, one of "get" or "post". Default: "get".
Use \code{http="post"} with large queries. Queries with > 300 records use "post"
automatically because "get" would fail}

\item{cap_first}{(logical) For each name, fix so that the first name part is
capitalized, while others are not. This web service is sensitive to
capitalization, so you'll get different results depending on
capitalization. First name capitalized is likely what you'll want and is
the default. If \code{FALSE}, names are not modified. Default: \code{TRUE}}

\item{fields}{(character) One of minimal (default) or all. Minimal gives back
just four fields, whereas all gives all fields back.}

\item{names}{Deprecated, see \code{sci}}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
A data.frame with one attribute \code{not_known}: a character vector of
taxa unknown to the Global Names Index. Access like \code{attr(output, "not_known")}, or \code{attributes(output)$not_known}.

Columns of the output data.frame:
\itemize{
\item user_supplied_name (character) - the name you passed in to the
\code{names} parameter, unchanged.
\item submitted_name (character) - the actual name submitted to the GNR
service
\item data_source_id (integer/numeric) - data source ID
\item data_source_title (character) - data source name
\item gni_uuid (character) - Global Names Index UUID (aka identifier)
\item matched_name (character) - the matched name in the GNR service
\item matched_name2 (character) - returned if \code{canonical=TRUE}, in
which case \strong{matched_name} is not returned
\item classification_path (character) - names of the taxonomic
classification tree, with names separated by pipes (\code{|})
\item classification_path_ranks (character) - ranks of the taxonomic
classification tree, with names separated by pipes (\code{|})
\item classification_path_ids (character) - identifiers of the taxonomic
classification tree, with names separated by pipes (\code{|})
\item taxon_id (character) - taxon identifier
\item edit_distance (integer/numeric) - edit distance
\item imported_at (character) - date imported
\item match_type (integer/numeric) - match type
\item match_value (character) - description of match type
\item prescore (character) - pre score
\item score (numeric) - score
\item local_id (character) - local identifier
\item url (character) - URL for taxon
\item global_id (character) - global identifier
\item current_taxon_id (character) - current taxon id
\item current_name_string (character) - current name string

Note that names (i.e. rows) are dropped that are NA, are zero length
strings, are not character vectors, or are not found by the API.
}
}
\description{
NOTE: this function is depreciated and will be removed in a future version.
The service this function interacts with is no longer maintained and has been
replaced by GNA Verifier, which can be used with the \code{\link[=gna_verifier]{gna_verifier()}}
function.
}
\details{
See section \strong{Age of datasets in the Global Names Resolver}
}
\section{Age of datasets in the Global Names Resolver}{
 IMPORTANT: Datasets
used in the Global Names Resolver vary in how recently they've been
updated. See the \code{updated_at} field in the output of \code{\link[=gnr_datasources]{gnr_datasources()}}
for dates when each dataset was last updated.
}

\section{preferred_data_sources}{
 If \code{preferred_data_sources} is used, only
the preferred data is returned - if it has any results.
}

\examples{
\dontrun{
gnr_resolve(sci = c("Helianthus annuus", "Homo sapiens"))
gnr_resolve(sci = c("Asteraceae", "Plantae"))

# Using data source 12 (Encyclopedia of Life)
sources <- gnr_datasources()
sources
eol <- sources$id[sources$title == 'EOL']
gnr_resolve(names=c("Helianthos annuus","Homo sapians"), data_source_ids=eol)

# Two species in the NE Brazil catalogue
sps <- c('Justicia brasiliana','Schinopsis brasiliensis')
gnr_resolve(sci = sps, data_source_ids = 145)

# Best match only, compare the two
gnr_resolve(sci = "Helianthus annuus", best_match_only = FALSE)
gnr_resolve(sci = "Helianthus annuus", best_match_only = TRUE)

# Preferred data source
gnr_resolve(sci = "Helianthus annuus", preferred_data_sources = c(3,4))

# Return canonical names - default is canonical=FALSE
head(gnr_resolve(sci = "Helianthus annuus"))
head(gnr_resolve(sci = "Helianthus annuus", canonical=TRUE))

# Return canonical names with authority stripped but
# ranks still present
gnr_resolve("Scorzonera hispanica L. subsp. asphodeloides Wallr.")
## vs.
gnr_resolve("Scorzonera hispanica L. subsp. asphodeloides Wallr.",
   with_canonical_ranks = TRUE)
}
}
\references{
http://gnrd.globalnames.org/api http://gnrd.globalnames.org/
}
\seealso{
\code{\link[=gnr_datasources]{gnr_datasources()}}
}
\author{
Scott Chamberlain
}
\keyword{names}
\keyword{resolve}
\keyword{taxonomy}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tpl_families.r
\name{tpl_families}
\alias{tpl_families}
\title{Get The Plant List families.}
\usage{
tpl_families(...)
}
\arguments{
\item{...}{(list) Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
Returns a \code{data.frame} including the names of all families
indexed by The Plant List, and the major groups into which they fall
(i.e. Angiosperms, Gymnosperms, Bryophytes and Pteridophytes).
}
\description{
Get The Plant List families.
}
\details{
Requires an internet connection in order to connect to
<www.theplantlist.org>.
}
\examples{
\dontrun{
# Get a data.frame of plant families, with the group name 
# (Angiosperms, etc.)
head(tpl_families())
}
}
\seealso{
\code{\link[=tpl_get]{tpl_get()}}
}
\author{
John Baumgartner (johnbb@student.unimelb.edu.au)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eol_pages.R
\name{eol_pages}
\alias{eol_pages}
\title{Search for pages in EOL database using a taxonconceptID.}
\usage{
eol_pages(
  taxonconceptID,
  images_per_page = NULL,
  images_page = NULL,
  videos_per_page = NULL,
  videos_page = NULL,
  sounds_per_page = NULL,
  sounds_page = NULL,
  maps_per_page = NULL,
  maps_page = NULL,
  texts_per_page = NULL,
  texts_page = NULL,
  subjects = "overview",
  licenses = "all",
  details = FALSE,
  common_names = FALSE,
  synonyms = FALSE,
  references = FALSE,
  taxonomy = TRUE,
  vetted = 0,
  cache_ttl = NULL,
  ...
)
}
\arguments{
\item{taxonconceptID}{(numeric) a taxonconceptID, which is also the page
number}

\item{images_per_page}{(integer) number of returned image objects (0-75)}

\item{images_page}{(integer) images page}

\item{videos_per_page}{(integer) number of returned video objects (0-75)}

\item{videos_page}{(integer) videos page}

\item{sounds_per_page}{(integer) number of returned sound objects (0-75)}

\item{sounds_page}{(integer) sounds page}

\item{maps_per_page}{(integer) number of returned map objects (0-75)}

\item{maps_page}{(integer) maps page}

\item{texts_per_page}{(integer) number of returned text objects (0-75)}

\item{texts_page}{(integer) texts page}

\item{subjects}{'overview' (default) to return the overview text
(if exists), a pipe | delimited list of subject names from the list of EOL
accepted subjects (e.g. TaxonBiology, FossilHistory), or 'all' to get text
in any subject. Always returns an overview text as a first result (if one
exists in the given context).}

\item{licenses}{A pipe | delimited list of licenses or 'all' (default) to
get objects under any license. Licenses abbreviated cc- are all Creative
Commons licenses. Visit their site for more information on the various
licenses they offer.}

\item{details}{Include all metadata for data objects. (Default: \code{FALSE})}

\item{common_names}{Return all common names for the page's taxon
(Default: \code{FALSE})}

\item{synonyms}{Return all synonyms for the page's taxon
(Default: \code{FALSE})}

\item{references}{Return all references for the page's taxon
(Default: \code{FALSE})}

\item{taxonomy}{(logical) Whether to return any taxonomy details from
different taxon hierarchy providers, in an array named \code{taxonconcepts}
(Default: \code{TRUE})}

\item{vetted}{If 'vetted' is given a value of '1', then only trusted
content will be returned. If 'vetted' is '2', then only trusted and
unreviewed content will be returned (untrusted content will not be returned).
The default is to return all content. (Default: \code{FALSE})}

\item{cache_ttl}{The number of seconds you wish to have the response cached.}

\item{...}{Curl options passed on to \link[crul:HttpClient]{crul::HttpClient}}
}
\value{
JSON list object, or data.frame.
}
\description{
Search for pages in EOL database using a taxonconceptID.
}
\details{
It's possible to return JSON or XML with the EOL API. However,
this function only returns JSON for now.
}
\examples{
\dontrun{
(pageid <- eol_search('Pomatomus')$pageid[1])
x <- eol_pages(taxonconceptID = pageid)
x
x$scinames

z <- eol_pages(taxonconceptID = pageid, synonyms = TRUE)
z$synonyms

z <- eol_pages(taxonconceptID = pageid, common_names = TRUE)
z$vernacular
}
}
